;==============================================================================

; file: "_error.scm"

; Copyright (C) 1994-2000 by Marc Feeley, All Rights Reserved.

(##include "header.scm")

;==============================================================================

; Traps from the runtime system.

(define (##trap-err-code err-code name . args)
  (##runtime-error (##err-code->string err-code) name args))

(define (##trap-err-code* err-code name . args)
  (##runtime-error* (##err-code->string err-code) name args))

(define (##trim-absent lst)
  (let loop1 ((lst lst))
    (if (##pair? lst)
      (if (##eq? (##car lst) (macro-absent-obj))
        (loop1 (##cdr lst))
        (begin
          (let loop2 ((curr lst)
                      (next (##cdr lst)))
            (if (##pair? next)
              (if (##eq? (##car next) (macro-absent-obj))
                (loop2 curr
                       (##cdr next))
                (begin
                  (##set-cdr! curr next)
                  (loop2 next
                         (##cdr next))))
              (##set-cdr! curr next)))
          lst))
      lst)))

(define (##runtime-error err-msg name args)
  (##signal '##signal.runtime-error err-msg name (##trim-absent args)))

(define (##runtime-error* err-msg name args)

  (define (fix l)
    (if (##pair? (##cdr l)) (##cons (##car l) (fix (##cdr l))) (##car l)))

  (##signal '##signal.runtime-error err-msg name (##trim-absent (fix args))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (##default-signal-catcher s args)
  (if (##unbound? ##stderr)

    (##exit 1)

    (cond ((##eq? s '##signal.io-error)
           (##handle-simple-error
             #f
             #f
             (##car args)
             (##cdr args)
             '()))

          ((##eq? s '##signal.read-error)
           (##handle-simple-error
             'read
             (##make-locat (##port-name (##car args))
                           (##cadr args))
             (##caddr args)
             (##cdddr args)
             '()))

          ((##eq? s '##signal.unbound-dynamic-var)
           (##handle-simple-error
             #f
             #f
             "Unbound dynamic variable:"
             (##list (##car args))
            '()))

          ((##eq? s '##signal.global-unbound)
           (##handle-interpreter-error
             (##car args)
             (##cadr args)
             "Unbound variable:"
             (##list (##caddr args))
             '()))

          ((##eq? s '##signal.global-unbound-operator)
           (##handle-call-error
             (##car args)
             (##cadr args)
             "Unbound global variable in operator position"
             '()))

          ((##eq? s '##signal.global-non-procedure-operator)
           (##handle-call-error
             (##car args)
             (##cadr args)
             "Global variable in operator position is not a PROCEDURE"
             '()))

          ((##eq? s '##signal.non-procedure-jump)
           (##handle-call-error
             (let ((x (##car args)))
               (if (##self-eval? x) x (##list 'quote x)))
             (##cadr args)
             "Operator is not a PROCEDURE"
             '()))

          ((##eq? s '##signal.non-procedure-operator)
           (##handle-interpreter-error
             (##car args)
             (##cadr args)
             "Operator is not a PROCEDURE"
             '()
             (##list (##decomp (##car args)))))

          ((##eq? s '##signal.non-procedure-send)
           (##handle-interpreter-error
             (##car args)
             (##cadr args)
             "PROCEDURE expected after '=>':"
             '()
             (##list (##decomp (##car args)))))

          ((##eq? s '##signal.wrong-nb-arg)
           (##handle-call-error
             (##car args)
             (##trim-absent (##cadr args))
             "Wrong number of arguments passed to procedure"
             '()))

          ((##eq? s '##signal.unknown-keyword-arg)
           (##handle-call-error
             (##car args)
             (##cadr args)
             "Unknown keyword argument passed to procedure"
             '()))

          ((##eq? s '##signal.keyword-expected)
           (##handle-call-error
             (##car args)
             (##cadr args)
             "Keyword argument expected"
             '()))

          ((##eq? s '##signal.clam-conv-error)
           (let ((err-code (##car args))
                 (err-msg (##cadr args))
                 (proc (##caddr args))
                 (args (##cadddr args)))
             (##handle-call-error
              proc
              args
              (or err-msg (##err-code->string err-code))
              '())))

          ((##eq? s '##signal.cdef-conv-error)
           (let ((err-code (##car args))
                 (err-msg (##cadr args))
                 (proc (##caddr args)))
             (##handle-call-error
              proc
              '()
              (or err-msg (##err-code->string err-code))
              '())))

          ((##eq? s '##signal.multiple-c-return)
           (##handle-simple-error
             #f
             #f
             "Attempt to return to a C function that has already returned"
             '()
             '()))

          ((##eq? s '##signal.apply-arg-limit)
           (##handle-call-error
             (##car args)
             (##cadr args)
             "Argument count to apply exceeds implementation limit"
             '()))

          ((##eq? s '##signal.heap-overflow)
           (##handle-simple-error
             #f
             #f
             "Heap overflow"
             '()
             '()))

          ((##eq? s '##signal.stack-overflow)
           (##handle-simple-error
             #f
             #f
             "Stack overflow"
             '()
             '()))

          ((##eq? s '##signal.promise-already-determined)
           (##handle-simple-error
             #f
             #f
             "Promise already determined"
             '()
             '()))

          ((##eq? s '##signal.runtime-error)
           (##handle-call-error
             (##cadr args)
             (##caddr args)
             (##car args)
             '()))

          ((##eq? s '##signal.global-env-overflow)
           (##handle-simple-error
             #f
             #f
             "Global variable table overflow"
             '()
             '()))

          ((##eq? s '##signal.syntax-error)
           (let* ((src (##car args))
                  (locat (##source-locat src)))
             (##handle-simple-error
               #f
               locat
               (##cadr args)
               (##cddr args)
               (if locat '() (##list (##desourcify src))))))

          (else
           (##write-string "*** ERROR -- Signal not caught, " ##stderr)
           (##write s ##stderr)
           (##write-string " " ##stderr)
           (##write args ##stderr)
           (##newline ##stderr)
           (##exit 1)))))

(define (##handle-simple-error proc locat err-msg err-infos pps)
  (##sequentially (lambda ()
    (##identify-error
     "ERROR"
     proc
     locat
     err-msg
     err-infos
     pps)
    (##pop-repl))))

(define (##handle-interpreter-error $code rte err-msg err-infos pps)
  (##call-with-current-continuation
    (lambda (cont) (##sequentially (lambda ()
      (##identify-error
       "ERROR"
       (##extract-container $code rte)
       (##continuation-locat (##procedure->first-continuation cont))
       err-msg
       err-infos
       pps)
      (##debug-repl cont #t))))))

(define (##handle-call-error proc args err-msg err-infos)
  (##call-with-current-continuation
    (lambda (cont) (##sequentially (lambda ()

      (define (add-quotes l)
        (if (##pair? l)
          (let ((x (##car l)))
            (##cons (if (##self-eval? x) x (##list 'quote x))
                    (add-quotes (##cdr l))))
          '()))

      (##identify-error
       "ERROR"
       #f
       (##continuation-locat (##procedure->first-continuation cont))
       err-msg
       err-infos
       '())

      (let ((out (##repl-out)))
        (let ((call (##cons (if (##procedure? proc)
                              (##procedure-name proc)
                              proc)
                            (add-quotes args)))
              (width (##output-port-width out)))
          (let ((str (##object->limited-string call width)))
            (if (##fixnum.< (##string-length str) width)
              (##write-string str out)
              (begin
                (##write-string "(" out)
                (##write-string (##object->limited-string
                                 (##car call)
                                 (##fixnum.- width 1))
                                out)
                (##newline out)

                (let loop ((l (##cdr call)))
                  (if (##pair? l)
                    (begin
                      (##write-string "  " out)
                      (##write-string (##object->limited-string
                                       (##car l)
                                       (##fixnum.- width 2))
                                      out)
                      (##newline out)
                      (loop (##cdr l)))))

                (##write-string ")" out)))

            (##newline out)
            (##debug-repl cont #t)))))))))

(define (##identify-error kind proc locat err-msg err-infos pps)
  (let ((out (##repl-out)))
    (##write-string "*** " out)
    (##write-string kind out)
    (if (or proc locat)
      (##write-string " IN " out))
    (if proc
      (##display (if (##procedure? proc)
                   (##procedure-name proc)
                   proc)
                 out))
    (if locat
      (begin
        (if proc
          (##write-string ", " out))
        (##display-locat locat #t out)))
    (if err-msg
      (begin
        (##write-string " -- " out)
        (##display err-msg out)
        (let loop1 ((l err-infos))
          (if (##pair? l)
            (begin
              (##write-string " " out)
              (##write (##car l) out)
              (loop1 (##cdr l)))))))
    (##newline out)
    (let loop2 ((l pps))
      (if (##pair? l)
        (begin
          (##pretty-print (##car l) out)
          (loop2 (##cdr l)))))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define (##signal sig . args)
  (let ((signal-catcher
          (##dynamic-ref '##signal-catcher ##default-signal-catcher)))
    (signal-catcher sig args)))

(define (##catch-all signal-catcher thunk)
  (##dynamic-let '##signal-catcher signal-catcher thunk))

(define (##catch-signal sig signal-catcher thunk)
  (let ((parent-signal-catcher
          (##dynamic-ref '##signal-catcher ##default-signal-catcher)))
    (##catch-all (lambda (s args)
                   (if (##eq? s sig)
                     (signal-catcher s args)
                     (parent-signal-catcher s args)))
                 thunk)))

(define (##cleanup-on-signal thunk1 thunk2)
  (let ((parent-signal-catcher
          (##dynamic-ref '##signal-catcher ##default-signal-catcher)))
    (##catch-all (lambda (s args)
                   (thunk2)
                   (parent-signal-catcher s args))
                 thunk1)))

;------------------------------------------------------------------------------
