;==============================================================================

; file: "_kernel.scm"

; Copyright (C) 1994-2000 by Marc Feeley, All Rights Reserved.

;==============================================================================

(##include "header.scm")

;==============================================================================

(##declare
  (multilisp)
  (standard-bindings)
  (extended-bindings)
  (block)
  (fixnum)
)

(c-declare "

#define mem_tracezzz

/*---------------------------------------------------------------------------*/

#include \"os.h\"
#include \"setup.h\"
#include \"mem.h\"
#include \"c_intf.h\"

/*---------------------------------------------------------------------------*/

#ifndef ___CC_CMD
#define ___CC_CMD 0
#endif

#ifndef ___LD_CMD
#define ___LD_CMD 0
#endif
")

;------------------------------------------------------------------------------

; The procedure "##primordial-continuation" is only a container for
; the kernel handlers.  It must never be called.  The function
; "___setup" in the file "setup.c" is responsible for setting up the
; kernel handlers.

(define ##primordial-continuation
  (let ()

    (##declare (not inline) (not interrupts-enabled))

    (lambda ()

      (##c-code
"
/*
 * ___LBL(1)
 *
 * This is the internal return handler.  It is invoked when an
 * internal return point is returned to.
 * 
 * Internal return points are used by the compiler when some routine
 * (e.g.  garbage collector) has to be called but it would be space
 * inefficient to generate the code to construct a normal continuation
 * frame (all the live GVM registers would have to be pushed on the
 * stack).  With internal return points it is not necessary for the
 * caller to save the live GVM registers before the routine is called,
 * and restore the GVM registers when the routine returns.  Instead it
 * is the called routine which saves and restores all the registers.
 *
 * The continuation frame for an internal return point has this layout:
 *
 *              STACK
 *          |            |
 *          +------------+
 *  ___fp ->|  RESERVED  | ^    <-- added to make stack and heap length equal
 *          |<ALIGN PAD> | |    <-- words added for alignment of frame
 *          |  ret_adr   | |    <-- internal return address (back to caller)
 *          | GVM reg N  | | ADDED BY ROUTINE
 *          |    ...     | |
 *          | GVM reg 0  | |
 *          |<ALIGN PAD> | v    <-- to force known offset from ___fp to ret_adr
 *          |  slot fs   | ^
 *          |    ...     | | PUT ON STACK BY CALLER OF ROUTINE
 *          |  slot 1    | v
 *          +------------+
 *          |            |
 */

{
  int fs;
  ___WORD ira;

  /* save result in case we are returning from ##force-undetermined */

  ___ps->temp2 = ___R1;

  /* make ___fp point to internal return address */

  ___ADJFP(___RETI_RA)

  /* pop return address and all GVM registers */

  ira = ___POP;
  ___POP_REGS

  /* get number of slots put on stack by caller */

  ___RETI_GET_CFS(ira,fs)

  /* adjust ___fp so that slot fs is on top of stack */

  ___ADJFP(fs-___ROUND_TO_MULT(fs,___FRAME_ALIGN))

  /* jump to internal return point */

  ___COVER_INTERNAL_RETURN_HANDLER_END;

#if 0
if (0) { int ii;
fprintf (stderr, \"fs = %d\\n\", fs);
fprintf (stderr, \"ira = 0x%08lx  fs = %d\\n\", ira, fs);
fprintf (stderr, \"___r0 = 0x%08lx\\n\", ___r0);
fprintf (stderr, \"___r1 = 0x%08lx\\n\", ___r1);
fprintf (stderr, \"___r2 = 0x%08lx\\n\", ___r2);
fprintf (stderr, \"___r3 = 0x%08lx\\n\", ___r3);
fprintf (stderr, \"___r4 = 0x%08lx\\n\", ___r4);
for (ii=0; ii<fs; ii++)
  fprintf (stderr, \" fp[%d]=0x%08lx\", ii, ___fp[ii]);
fprintf (stderr, \"\\n\");
fflush (stderr);
}
#endif

  ___JUMPEXTPRM(___NOTHING,ira)
}
"
       (0)) ; the call "(0)" is to create the label as a return point

      (##c-code
"
/*
 * ___LBL(2)
 *
 * This is the break handler.  It is invoked when a function
 * attempts to return to its caller and the caller's stack frame
 * is not on top of the stack because it has been captured.
 *
 * At this point the callee will have cleaned up the stack so that the
 * frame pointer (___fp) points to the break frame.  The break frame
 * contains a pointer to the caller's continuation frame, which
 * contains the address in the caller where control will return (the
 * return address ret_adr1).  The caller's continuation frame can either
 * be in the stack or in the heap.  The two situations are depicted below:
 *
 *              STACK                      STACK              HEAP
 *          |            |             |            |                  caller's
 *          +------------+             +------------+    +------------+frame
 *  ___fp ->| call frame ---+  ___fp ->| call frame ---->|    HEAD    |
 *          |<ALIGN PAD> |  |          |<ALIGN PAD> |    |  ret_adr1  |
 *          +------------+  |          +------------+    |  slot fs   |
 *          |     .      |  |          |     .      |    |    ...     |
 *          |     .      |  |          |     .      |    | slot link ----+
 *          |     .      |  |          |     .      |    |    ...     |  |
 *          |     .      |  |          |     .      |    |  slot 1    |  |
 *          +------------+  |          |            |    +------------+  |
 *          |  ret_adr1  |<-+          |            |                    |
 *          |  RESERVED  |             |            |                    |
 * caller's |  slot fs   |             |            |                    |
 * frame    |    ...     |             |            |                    |
 *          |  ret_adr2  |             |            |                    |
 *          |    ...     |             |            |                    |
 *          |  slot 1    |             |     .      |    +------------+  |
 *          +------------+             |     .      |    |    HEAD    |<-+
 *          |  RESERVED  |             |     .      |    |  ret_adr2  |
 *          |    ...     |             |     .      |    |    ...    ----+
 *          +------------+             +------------+    +------------+  V
 *                                                                      ...
 *
 * These cases are distinguished by the tag on the pointer to the
 * caller's frame (i.e. 'call frame').
 *
 *
 * The break handler puts a copy of the caller's frame on the top of
 * the stack except that the slot 'link' is set to the address of the
 * break handler.  The frame pointer in the break frame is modified
 * so that it points to the frame of the caller's caller.  Finally a
 * jump to the return address in the caller's frame (ret_adr1) is
 * performed.  At that point the stack will be in the following state
 * respectively:
 *
 *              STACK                      STACK              HEAP
 *          |            |             |            |
 *          +------------+             +------------+
 *  ___fp ->|  RESERVED  |     ___fp ->|  RESERVED  | ^
 *          |  slot fs   |             |  slot fs   | |
 *          |    ...     |             |    ...     | | ADDED BY BREAK
 *          | break hdlr |             | break hdlr | | HANDLER
 *          |    ...     |             |    ...     | |
 *          |  slot 1    |             |  slot 1    | v
 *          +------------+             +------------+
 *          |new call fr.---+          |new call fr.--+  +------------+
 *          |<ALIGN PAD> |  |          |<ALIGN PAD> | |  |    HEAD    |
 *          +------------+  |          +------------+ |  |  ret_adr1  |
 *          |     .      |  |          |     .      | |  |  slot fs   |
 *          |     .      |  |          |     .      | |  |    ...     |
 *          |     .      |  |          |     .      | |  | slot link ----+
 *          |     .      |  |          |     .      | |  |    ...     |  |
 *          +------------+  |          |            | |  |  slot 1    |  |
 *          |  ret_adr1  |  |          |            | |  +------------+  |
 *          |  RESERVED  |  |          |            | |                  |
 *          |  slot fs   |  |          |            | |                  |
 *          |    ...     |  |          |            | |                  |
 *          |  ret_adr2  |  |          |            | |                  |
 *          |    ...     |  |          |            | |                  |
 *          |  slot 1    |  |          |     .      | |  +------------+  |
 *          +------------+  |          |     .      | +->|    HEAD    |<-+
 *          |  ret_adr2  |<-+          |     .      |    |  ret_adr2  |
 *          |    ...     | (see note)  |     .      |    |    ...    ----+
 *          +------------+             +------------+    +------------+  V
 *                                                                      ...
 *
 * Note that in the first case, the pointer to the caller's frame is
 * normally advanced to the frame following the caller's frame and the
 * return address ret_adr2 is saved in it.  However, if the frame
 * following the caller's frame is a break frame, then the content of
 * that break frame is copied to the topmost break frame.  This
 * ensures that break frames never contain pointers to other break
 * frames which is needed to properly implement tail-recursion.
 */

{
  int fs, link, i;
  ___WORD *fp;
  ___WORD ra1;
  ___WORD ra2;
  ___WORD cf;

  cf = ___STK(-___BREAK_FRAME_NEXT); /* pointer to caller's frame */

  if (___TYP(cf) == ___tFIXNUM)
    {
      /* caller's frame is in the stack */

      if (cf == ___FIX(0))
        ___fatal_error (\"No continuation to return to\");

      fp = ___CAST(___WORD*,cf);

      ra1 = ___FP_STK(fp,-___FRAME_STACK_RA);

      if (ra1 == ___internal_return)
        {
          ___WORD actual_ra = ___FP_STK(fp,___RETI_RA);
if (0) fprintf (stderr, \"--------------------- reti\\n\");
          ___RETI_GET_FS_LINK(actual_ra,fs,link)
          ___COVER_BREAK_HANDLER_STACK_RETI;
        }
      else
        {
if (0) fprintf (stderr, \"--------------------- retn\\n\");
          ___RETN_GET_FS_LINK(ra1,fs,link)
          ___COVER_BREAK_HANDLER_STACK_RETN;
        }

      ___FP_ADJFP(fp,-___FRAME_SPACE(fs)); /* get base of frame */

      for (i=fs; i>0; i--)
        ___SET_STK(i,___FP_STK(fp,i))

      ra2 = ___STK(link+1);

#if 0
if (0) { int ii;
fprintf (stderr, \"___internal_return = 0x%08lx\\n\", ___internal_return);
fprintf (stderr, \"ra1 = 0x%08lx\\n\", ra1);
fprintf (stderr, \"ra2 = 0x%08lx\\n\", ra2);
fprintf (stderr, \"fs = %d\\n\", fs);
fprintf (stderr, \"link = %d\\n\", link);
for (ii=1; ii<fs; ii++)
  fprintf (stderr, \" fp[%d]=0x%08lx\", -ii, ___fp[-ii]);
fprintf (stderr, \"\\n---------------------\\n\");
fflush (stderr);
}
#endif

      if (ra2 == ___ps->handler_break)
        {
          /* first frame of that section */

          ___COVER_BREAK_HANDLER_STACK_FIRST_FRAME;

          ___SET_STK(-___BREAK_FRAME_NEXT,
                     ___FP_STK(fp,-___BREAK_FRAME_NEXT))
        }
      else
        {
          /* not the first frame of that section */

          ___COVER_BREAK_HANDLER_STACK_NOT_FIRST_FRAME;

          ___FP_SET_STK(fp,-___FRAME_STACK_RA,ra2)
          ___SET_STK(-___BREAK_FRAME_NEXT,___CAST(___WORD,fp))
          ___SET_STK(link+1,___ps->handler_break)
        }
    }
  else
    {
      /* caller's frame is in the heap */

      fp = ___BODY_AS(cf,___tSUBTYPED); /* get pointer to frame's body */

      ra1 = fp[___FRAME_RA];

      if (ra1 == ___internal_return)
        {
          ___WORD actual_ra = fp[___FRAME_RETI_RA];
          ___RETI_GET_FS_LINK(actual_ra,fs,link)
          ___COVER_BREAK_HANDLER_HEAP_RETI;
        }
      else
        {
          ___RETN_GET_FS_LINK(ra1,fs,link)
          ___COVER_BREAK_HANDLER_HEAP_RETN;
        }

      fp += fs+1; /* get base of frame */

      for (i=fs; i>0; i--)
        ___SET_STK(i,___FP_STK(fp,i))

      ___SET_STK(-___BREAK_FRAME_NEXT,___STK(link+1))
      ___SET_STK(link+1,___ps->handler_break)
    }

  ___ADJFP(___FRAME_SPACE(fs))

  ___JUMPEXTPRM(___NOTHING,ra1)
}
"
       (0)) ; the call "(0)" is to create the label as a return point

      (##c-code
"
/*
 * ___LBL(3)
 *
 * This is the stack-limit handler.  It is invoked by the ___POLL(n)
 * macro when the stack limit is reached and when an interrupt is received.
 *
 * This handler checks for which reason it was invoked and dispatches
 * to one of the following procedures:
 *
 *  ##interrupt-handler
 *  ##exception.stack-overflow
 */

{
  /* setup internal return continuation frame */

  int fs;
  ___WORD ra;

  ra = ___ps->temp1;

  ___RETI_GET_CFS(ra,fs)

  ___ADJFP(___ROUND_TO_MULT(fs,___FRAME_ALIGN)-fs)

  ___PUSH_REGS /* push all GVM registers (live or not) */
  ___PUSH(ra)  /* push return address */

  ___ADJFP(-___RETI_RA)

  ___SET_R0(___internal_return)

  /* check why the handler was called */

#if 0
if (0)
{
fprintf (stderr, \"interrupt, ___fp=0x%08lx ___ps->stack_limit=0x%08lx ___POLL at %s:%d\\n\", ___fp,___ps->stack_limit, ___ps->poll_file, ___ps->poll_line);
fflush (stderr);

if (___ps->poll_line == 13201)
{
fprintf (stderr, \"danger!\\n\");
fflush (stderr);
}
}
#endif

  if (___FP_AFTER(___fp,___ps->stack_limit)
#ifdef CALL_GC_FREQUENTLY
      || --___gc_calls_to_punt < 0
#endif
     )
    {
      ___BOOL overflow;

      ___COVER_STACK_LIMIT_HANDLER_STACK_LIMIT;

      ___FRAME_STORE_RA(___R0)
      ___W_ALL
      overflow = ___stack_limit () && ___garbage_collect ();
      ___R_ALL
      ___SET_R0(___FRAME_FETCH_RA)

      if (overflow)
        {
          ___COVER_STACK_LIMIT_HANDLER_HEAP_OVERFLOW;
          ___JUMPPRM(___SET_NARGS(0),
                     ___G__23__23_exception_2e_stack_2d_overflow.prm)
        }
    }

  /* prepare for next interrupt */

  ___begin_interrupt_service ();

  if (___ps->intr_enabled)
    {
      int i;

      ___COVER_STACK_LIMIT_HANDLER_INTR_ENABLED;

      for (i=0; i<___NB_INTRS; i++)
        if (___check_interrupt (i))
          break;

      ___end_interrupt_service (i+1);

      if (i < ___NB_INTRS)
        {
          ___COVER_STACK_LIMIT_HANDLER_INTERRUPT;

          ___SET_R1(___FIX(i))
          ___JUMPPRM(___SET_NARGS(1),
                     ___G__23__23_interrupt_2d_handler.prm)
        }
    }
  else
    ___end_interrupt_service (0);

  ___COVER_STACK_LIMIT_HANDLER_END;

  ___JUMPEXTPRM(___NOTHING,___R0)
}
"
       (0)) ; the call "(0)" is to create the label as a return point

      (##c-code
"
/*
 * ___LBL(4)
 *
 * This is the heap-limit handler.  It is invoked by the ___CHECK_HEAP(n,m)
 * macro when the heap pointer reaches the end of the current msection.
 *
 * This handler simply calls ##check-heap.
 */

{
  /* setup internal return continuation frame */

  int fs;
  ___WORD ra;

  ra = ___ps->temp1;

  ___RETI_GET_CFS(ra,fs)

  ___ADJFP(___ROUND_TO_MULT(fs,___FRAME_ALIGN)-fs)

  ___PUSH_REGS /* push all GVM registers (live or not) */
  ___PUSH(ra)  /* push return address */

  ___ADJFP(-___RETI_RA)

  ___SET_R0(___internal_return)

  /* tail call to ##check-heap */

  ___COVER_HEAP_LIMIT_HANDLER_END;

  ___JUMPPRM(___SET_NARGS(0),
             ___G__23__23_check_2d_heap.prm)
}
"
       (0)) ; the call "(0)" is to create the label as a return point

      (##c-code
"
/*
 * ___LBL(5)
 *
 * This is the call to non-procedure handler.  It is invoked when
 * there is an attempt to call an object that is not a procedure.
 *
 * This handler simply tail calls ##exception.non-proc-jump (i.e.
 * the continuation will be the same as that of the faulty call).  The
 * arguments are the object that was in operator position followed by
 * the arguments of the faulty call.
 */

{
  int na;
  int i;

  na = ___ps->na;

  ___PUSH_ARGS_IN_REGS(na) /* save all arguments that are in registers */
  ___PUSH(___FIX(0))       /* make space for operator */

  for (i=0; i<na; i++)
    ___SET_STK(-i,___STK(-(i+1))) /* shift arguments up by one */

  ___SET_STK(-na,___ps->temp1) /* set operator argument */

  ___POP_ARGS_IN_REGS(na+1) /* load register arguments */

  ___COVER_NON_PROC_HANDLER_END;

  ___JUMPPRM(___SET_NARGS(na+1),
             ___G__23__23_exception_2e_non_2d_proc_2d_jump.prm)
}
"
       (0)) ; the call "(0)" is to create the label as a return point

      (##c-code
"
/*
 * ___LBL(6)
 *
 * This is the call to global non-procedure handler.  It is invoked
 * when there is an attempt to call an object that is not a procedure
 * that is bound to a global variable.
 *
 * This handler simply tail calls ##exception.global-jump or
 * ##exception.non-proc-jump (i.e. the continuation will be the same as
 * that of the faulty call).  ##exception.global-jump is called
 * when the global variable's name is known (in this case the
 * arguments are the variable's name followed by the arguments of the
 * faulty call).  Otherwise, ##exception.non-proc-jump is called
 * and the arguments are the object that was in operator position
 * followed by the arguments of the faulty call.
 */

{
  int na;
  int i;
  ___WORD sym;
  ___WORD probe;
  ___WORD result;
  ___WORD handler;

  na = ___ps->na;

  ___PUSH_ARGS_IN_REGS(na) /* save all arguments that are in registers */
  ___PUSH(___FIX(0))       /* make space for operator */

  for (i=0; i<na; i++)
    ___SET_STK(-i,___STK(-(i+1))) /* shift arguments up by one */

  result = ___CAST(___glo_struct*,___ps->temp4)->val;
  handler = ___G__23__23_exception_2e_non_2d_proc_2d_jump.prm;

  i = ___INT(___VECTORLENGTH(___symbol_table)) - 1;

  while (i >= 0)
    {
      probe = ___FIELD(___symbol_table,i);

      while (probe != ___NUL)
        {
          sym = ___CAR(probe);
          if (___ps->temp4 == ___FIELD(sym,___SYMBOL_GLOBAL))
            {
              ___COVER_GLOBAL_NON_PROC_HANDLER_FOUND;
              result = sym;
              handler = ___G__23__23_exception_2e_global_2d_jump.prm;
              break;
            }
          ___COVER_GLOBAL_NON_PROC_HANDLER_NEXT;
          probe = ___CDR(probe);
        }
      i--;
    }

  ___SET_STK(-na,result) /* set operator argument */

  ___POP_ARGS_IN_REGS(na+1) /* load register arguments */

  ___COVER_GLOBAL_NON_PROC_HANDLER_END;

  ___JUMPPRM(___SET_NARGS(na+1),handler)
}
"
       (0)) ; the call "(0)" is to create the label as a return point

      (##c-code
"
/*
 * ___LBL(7)
 *
 * This is the wrong number of arguments handler.  It is invoked when
 * a procedure is called with a number of arguments that it does not
 * accept.
 *
 * This handler simply tail calls ##exception.non-proc-jump (i.e.
 * the continuation will be the same as that of the faulty call).  The
 * arguments are the object that was in operator position followed by
 * the arguments of the faulty call.
 */

{
  int na;
  int i;

  na = ___ps->na;

  ___PUSH_ARGS_IN_REGS(na) /* save all arguments that are in registers */
  ___PUSH(___FIX(0))       /* make space for operator */

  for (i=0; i<na; i++)
    ___SET_STK(-i,___STK(-(i+1))) /* shift arguments up by one */

  /* ___ps->temp1 points to the entry point of the procedure */

  if (___HD_TYP(___HEADER(___ps->temp1)) == ___PERM)
    {
      ___COVER_WRONG_NARGS_HANDLER_NON_CLOSURE;
      ___SET_STK(-na,___ps->temp1) /* set operator argument when non-closure */
    }
  else
    {
      ___COVER_WRONG_NARGS_HANDLER_CLOSURE;
      ___SET_STK(-na,___SELF) /* set operator argument when closure */
    }

  ___POP_ARGS_IN_REGS(na+1) /* load register arguments */

  ___JUMPPRM(___SET_NARGS(na+1),
             ___G__23__23_exception_2e_wrong_2d_nb_2d_arg.prm)
}
"
       (0)) ; the call "(0)" is to create the label as a return point

      (##c-code
"
/*
 * ___LBL(8)
 *
 * This is the rest parameter handler.  It is invoked when a non-null
 * rest parameter must be constructed.
 *
 */

{
  int np;
  int na;
  int i;
  ___WORD rest_param_list;

  np = ___PRD_NBPARMS(___HEADER(___ps->temp1));
  na = ___ps->na;

  ___PUSH_ARGS_IN_REGS(na) /* save all arguments that are in registers */

  if (na < np)
    {
      ___COVER_REST_PARAM_HANDLER_WRONG_NARGS;

      ___PUSH(___FIX(0)) /* make space for operator */

      for (i=0; i<na; i++)
        ___SET_STK(-i,___STK(-(i+1))) /* shift arguments up by one */

      ___SET_STK(-na,___ps->temp1) /* set operator argument */

      ___POP_ARGS_IN_REGS(na+1) /* load register arguments */

      ___JUMPPRM(___SET_NARGS(na+1),
                 ___G__23__23_exception_2e_wrong_2d_nb_2d_arg.prm)
    }

  rest_param_list = ___NUL;

  i = na - np + 1;

  while (i > 0)
    {
      rest_param_list = ___CONS(___POP,rest_param_list);
      i--;

      if (___hp > ___ps->heap_limit)
        {
           ___BOOL need_to_gc;

           ___COVER_REST_PARAM_HANDLER_HEAP_LIMIT;

           ___W_HEAP
           need_to_gc = ___heap_limit ();
           ___R_HEAP

           if (need_to_gc)
             {
               /*
                * We need to garbage collect, but before we do we have
                * to remove all arguments from the stack so that the GC
                * only sees complete continuation frames on the stack.
                * The arguments are stored in a heap allocated vector.
                */

               ___COVER_REST_PARAM_HANDLER_NEED_TO_GC;

               while (i > 0)
                 {
                   rest_param_list = ___CONS(___POP,rest_param_list);
                   i--;
                 }

               ___BEGIN_ALLOC_VECTOR(np)
               i = np - 1;
               ___ADD_VECTOR_ELEM(i,rest_param_list)
               while (i > 0)
                 {
                   i--;
                   ___ADD_VECTOR_ELEM(i,___POP)
                 }
               ___END_ALLOC_VECTOR(np)

               ___PUSH_ARGS2(___ps->temp1,___GET_VECTOR(np))

               ___JUMPPRM(___SET_NARGS(2),
                          ___G__23__23_rest_2d_param_2d_check_2d_heap.prm)
             }
         }
    }

  ___COVER_REST_PARAM_HANDLER_NO_NEED_TO_GC;

  ___PUSH(rest_param_list) /* rest parameter is last */

  ___POP_ARGS_IN_REGS(np) /* load register arguments */

  ___JUMPEXTPRM(___SET_NARGS(-1),___ps->temp1)
}
"
       (0)) ; the call "(0)" is to create the label as a return point

      (##c-code
"
/*
 * ___LBL(9)
 *
 * This is the keyword parameter handler.  It is invoked when keyword
 * parameters must be processed.
 *
 */

{
  int np;
  int na;
  int nb_req_opt;
  int nb_key;
  int i;
  int j;
  int k;
  ___WORD key_descr;
  ___WORD key_vals[___MAX_NB_PARMS];

  np = ___PRD_NBPARMS(___HEADER(___ps->temp1));
  na = ___ps->na;
  key_descr = ___ps->temp3;
  nb_req_opt = ___ps->temp2;
  nb_key = np - nb_req_opt;

  ___PUSH_ARGS_IN_REGS(na) /* save all arguments that are in registers */

  k = na - nb_req_opt;

  if (k < 0)
    {
      ___COVER_KEYWORD_PARAM_HANDLER_WRONG_NARGS1;
      goto wrong_nb_args1;
    }

  for (j=nb_key-1; j>=0; j--)
    key_vals[j] = ___FIELD(key_descr,j*2+1);

  for (i=1; i<=k; i+=2)
    {
      ___WORD key = ___STK(-i);

      for (j=nb_key-1; j>=0; j--)
        if (key == ___FIELD(key_descr,j*2))
          {
            ___COVER_KEYWORD_PARAM_HANDLER_FOUND;
            key_vals[j] = ___STK(-i+1);
            goto continue1;
          }

      /*
       * The argument did not match any of the valid keywords for this
       * procedure, so raise an exception.
       */

      ___COVER_KEYWORD_PARAM_HANDLER_NOT_FOUND;

      ___PUSH(___FIX(0)) /* make space for operator */

      for (i=0; i<na; i++)
        ___SET_STK(-i,___STK(-(i+1))) /* shift arguments up by one */

      ___SET_STK(-na,___ps->temp1) /* set operator argument */

      ___POP_ARGS_IN_REGS(na+1) /* load register arguments */

      if (!___KEYWORDP(key))
        ___JUMPPRM(___SET_NARGS(na+1),
                   ___G__23__23_exception_2e_keyword_2d_expected.prm)
      else
        ___JUMPPRM(___SET_NARGS(na+1),
                   ___G__23__23_exception_2e_unknown_2d_keyword_2d_arg.prm)

      continue1:;
    }

  if (k & 1) /* keyword arguments must come in pairs */
    {
      ___COVER_KEYWORD_PARAM_HANDLER_WRONG_NARGS2;

      wrong_nb_args1:

      ___PUSH(___FIX(0)) /* make space for operator */

      for (i=0; i<na; i++)
        ___SET_STK(-i,___STK(-(i+1))) /* shift arguments up by one */

      ___SET_STK(-na,___ps->temp1) /* set operator argument */

      ___POP_ARGS_IN_REGS(na+1) /* load register arguments */

      ___JUMPPRM(___SET_NARGS(na+1),
                 ___G__23__23_exception_2e_wrong_2d_nb_2d_arg.prm)
    }

  ___ADJFP(-k) /* remove keyword arguments */

  for (i=0; i<nb_key; i++) /* push value of all keyword params */
    ___PUSH(key_vals[i])

  ___POP_ARGS_IN_REGS(np) /* load register arguments */

  ___COVER_KEYWORD_PARAM_HANDLER_END;

  ___JUMPEXTPRM(___SET_NARGS(-1),___ps->temp1)
}
"
       (0)) ; the call "(0)" is to create the label as a return point

      (##c-code
"
/*
 * ___LBL(10)
 *
 * This is the keyword and rest parameter handler.  It is invoked when
 * keyword parameters must be processed and a rest parameter must be
 * constructed.
 *
 */

{
  int np;
  int na;
  int nb_req_opt;
  int nb_key;
  int i;
  int j;
  int k;
  ___WORD key_descr;
  ___WORD key_vals[___MAX_NB_PARMS];
  ___WORD rest_param_list;

  np = ___PRD_NBPARMS(___HEADER(___ps->temp1));
  na = ___ps->na;
  key_descr = ___ps->temp3;
  nb_req_opt = ___ps->temp2;
  nb_key = (np-1) - nb_req_opt;

  ___PUSH_ARGS_IN_REGS(na) /* save all arguments that are in registers */

  k = na - nb_req_opt;

  if (k < 0)
    {
      ___COVER_KEYWORD_REST_PARAM_HANDLER_WRONG_NARGS1;
      goto wrong_nb_args2;
    }

  for (j=nb_key-1; j>=0; j--)
    key_vals[j] = ___FIELD(key_descr,j*2+1);

  for (i=1; i<=k; i+=2)
    {
      ___WORD key = ___STK(-i);

      for (j=nb_key-1; j>=0; j--)
        if (key == ___FIELD(key_descr,j*2))
          {
            ___COVER_KEYWORD_REST_PARAM_HANDLER_FOUND;
            key_vals[j] = ___STK(-i+1);
            goto continue2;
          }

      /*
       * The argument did not match any of the valid keywords for this
       * procedure.  An exception is raised only if the argument is not
       * a keyword.
       */

      if (!___KEYWORDP(key))
        {
          ___COVER_KEYWORD_REST_PARAM_HANDLER_NOT_FOUND;

          ___PUSH(___FIX(0)) /* make space for operator */

          for (i=0; i<na; i++)
            ___SET_STK(-i,___STK(-(i+1))) /* shift arguments up by one */

          ___SET_STK(-na,___ps->temp1) /* set operator argument */

          ___POP_ARGS_IN_REGS(na+1) /* load register arguments */

          ___JUMPPRM(___SET_NARGS(na+1),
                     ___G__23__23_exception_2e_keyword_2d_expected.prm)
        }

      continue2:;
    }

  if (k & 1) /* keyword arguments must come in pairs */
    {
      ___COVER_KEYWORD_REST_PARAM_HANDLER_WRONG_NARGS2;

      wrong_nb_args2:

      ___PUSH(___FIX(0)) /* make space for operator */

      for (i=0; i<na; i++)
        ___SET_STK(-i,___STK(-(i+1))) /* shift arguments up by one */

      ___SET_STK(-na,___ps->temp1) /* set operator argument */

      ___POP_ARGS_IN_REGS(na+1) /* load register arguments */

      ___JUMPPRM(___SET_NARGS(na+1),
                 ___G__23__23_exception_2e_wrong_2d_nb_2d_arg.prm)
    }

  rest_param_list = ___NUL;

  i = k;

  while (i > 0)
    {
      rest_param_list = ___CONS(___POP,rest_param_list);
      i--;

      if (___hp > ___ps->heap_limit)
        {
           ___BOOL need_to_gc;

           ___COVER_KEYWORD_REST_PARAM_HANDLER_HEAP_LIMIT;

           ___W_HEAP
           need_to_gc = ___heap_limit ();
           ___R_HEAP

           if (need_to_gc)
             {
               /*
                * We need to garbage collect, but before we do we have
                * to remove all arguments from the stack so that the GC
                * only sees complete continuation frames on the stack.
                * The arguments are stored in a heap allocated vector.
                */

               ___COVER_KEYWORD_REST_PARAM_HANDLER_NEED_TO_GC;

               while (i > 0)
                 {
                   rest_param_list = ___CONS(___POP,rest_param_list);
                   i--;
                 }

               for (i=0; i<nb_key; i++) /* push value of all keyword params */
                 ___PUSH(key_vals[i])

               ___BEGIN_ALLOC_VECTOR(np)
               i = np - 1;
               ___ADD_VECTOR_ELEM(i,rest_param_list)
               while (i > 0)
                 {
                   i--;
                   ___ADD_VECTOR_ELEM(i,___POP)
                 }
               ___END_ALLOC_VECTOR(np)

               ___PUSH_ARGS2(___ps->temp1,___GET_VECTOR(np))

               ___JUMPPRM(___SET_NARGS(2),
                          ___G__23__23_rest_2d_param_2d_check_2d_heap.prm)
             }
         }
    }

  ___COVER_KEYWORD_REST_PARAM_HANDLER_NO_NEED_TO_GC;

  for (i=0; i<nb_key; i++) /* push value of all keyword params */
    ___PUSH(key_vals[i])

  ___PUSH(rest_param_list) /* rest parameter is last */

  ___POP_ARGS_IN_REGS(np) /* load register arguments */

  ___JUMPEXTPRM(___SET_NARGS(-1),___ps->temp1)
}
"
       (0)) ; the call "(0)" is to create the label as a return point

      (##c-code
"
/*
 * ___LBL(11)
 *
 * This is the force handler.  It is invoked when a promise is forced.
 */

{
  ___WORD ra;
  ___WORD promise;
  ___WORD result;

  ra = ___ps->temp1;
  promise = ___ps->temp2;
  result = ___FIELD(promise,___PROMISE_RESULT);

  if (promise != result)
    {
      /* promise is determined, return cached result */

      ___COVER_FORCE_HANDLER_DETERMINED;

      ___ps->temp2 = result;
      ___JUMPEXTPRM(___NOTHING,ra)
    }
  else
    {
      /* promise is not determined */

      /* setup internal return continuation frame */

      int fs;

      ___RETI_GET_CFS(ra,fs)

      ___ADJFP(___ROUND_TO_MULT(fs,___FRAME_ALIGN)-fs)

      ___PUSH_REGS /* push all GVM registers (live or not) */
      ___PUSH(ra)  /* push return address */

      ___ADJFP(-___RETI_RA)

      ___SET_R0(___internal_return)

      /* tail call to ##force-undetermined */

      ___PUSH_ARGS2(promise,___FIELD(promise,___PROMISE_THUNK))

      ___COVER_FORCE_HANDLER_NOT_DETERMINED;

      ___JUMPPRM(___SET_NARGS(2),
                 ___G__23__23_force_2d_undetermined.prm)
    }
}
"
       (0)) ; the call "(0)" is to create the label as a return point

      (##c-code
"
/*
 * ___LBL(12)
 *
 * This is the 'c-lambda' conversion error handler.  It is invoked
 * when an error is detected by the C-interface in a 'c-lambda' during
 * a data representation conversion between Scheme and C.
 */

{
  int na;
  int i;

  na = ___ps->na;

  /* make space for 3 new arguments (one will replace return address) */

  ___ADJFP((na+1)+2-___FRAME_SPACE(na+1))

  ___SET_R0(___STK(-2)) /* get return address */

  for (i=0; i<na; i++)
    ___SET_STK(-i,___STK(-(i+3))) /* shift arguments up by three */

  ___SET_STK(-(na+2),___FIX(___ps->temp1)) /* arg 1 = error code, integer *//*************************remove ___FIX*/
  ___SET_STK(-(na+1),___ps->temp2)         /* arg 2 = error message, string or #f */
  ___SET_STK(-na,___ps->temp3)             /* arg 3 = procedure */

  ___POP_ARGS_IN_REGS(na+3) /* load register arguments */

  ___COVER_CLAM_CONV_ERROR_HANDLER;

  ___JUMPPRM(___SET_NARGS(na+3),
             ___G__23__23_exception_2e_clam_2d_conv_2d_error.prm)
}
"
       (0)) ; the call "(0)" is to create the label as a return point

      (##c-code
"
/*
 * ___LBL(13)
 *
 * This is the 'c-define' conversion error handler.  It is invoked
 * when an error is detected by the C-interface in a 'c-define' during
 * a data representation conversion between Scheme and C.
 */

{
  ___PUSH_ARGS3(___FIX(___ps->temp1),___ps->temp2,___ps->temp3)/*************************remove ___FIX*/

  ___COVER_CDEF_CONV_ERROR_HANDLER;

  ___JUMPPRM(___SET_NARGS(3),
             ___G__23__23_exception_2e_cdef_2d_conv_2d_error.prm)
}
"
       (0)) ; the call "(0)" is to create the label as a return point

      (let ((stack-marker (##first-argument 0)))
        (##c-code
"
/*
 * ___LBL(14)
 *
 * This is the return to C handler.  It is invoked when control
 * must return to C (i.e. back from the call to ___call).
 *
 * Note that the continuation frame is not removed from the stack (it
 * contains the stack marker which is needed by ___call).
 */

/************************** fix this */

#define ___RETURN_TO_C(n)___SET_R0(___LBL(n)) ___JUMPEXTPRM(___NOTHING,___ps->handler_return_to_c)


{
  ___WORD unwind_destination = ___STK(2-___FRAME_SPACE(2));

  if (___FIELD(unwind_destination,0) == ___TRU) /* first return? */
    {
      ___COVER_RETURN_TO_C_HANDLER_FIRST_RETURN;

      *___fp = ___ps->handler_return_to_c; /*****************save return address*/
      ___W_ALL
      ___propagate_error (___UNWIND_C_STACK);  /* jump back inside ___call */
    }
  else
    {
      ___COVER_RETURN_TO_C_HANDLER_MULTIPLE_RETURN;
      ___SET_R0(___ps->handler_return_to_c)
      ___JUMPPRM(___SET_NARGS(0),
                 ___G__23__23_exception_2e_multiple_2d_c_2d_return.prm)
    }
}
"
         (0)) ; the call "(0)" is to create the label as a return point
        stack-marker))))

(define ##rest-param-check-heap
  (let ()

    (##declare (not inline) (not interrupts-enabled))

    (lambda (proc args)
      (let ((overflow (##gc-without-exceptions)))
        (if overflow
          (begin
            (##exception.heap-overflow)
            (##rest-param-check-heap proc args))
          (##rest-param-resume-procedure proc args))))))

(define ##rest-param-resume-procedure
  (let ()

    (##declare (not inline) (not interrupts-enabled))

    (lambda (proc args)
      (##c-code
"
{
  ___WORD proc;
  ___WORD args;
  int np;
  int i;

  ___POP_ARGS2(proc,args)

  np = ___INT(___VECTORLENGTH(args));

  for (i=0; i<np; i++)
    ___PUSH(___FIELD(args,i))

  ___POP_ARGS_IN_REGS(np) /* load register arguments */

  ___COVER_REST_PARAM_RESUME_PROCEDURE;

  ___JUMPEXTPRM(___SET_NARGS(-1),proc)
}
"))))

(define ##gc-without-exceptions
  (let ()

    (##declare (not inline) (not interrupts-enabled))

    (lambda ()

      (##c-code
"
{
  ___BOOL overflow;

#ifdef mem_trace
printf (\"------------------------- ##gc-without-exceptions\\n\"); fflush(stdout);
#endif

  ___FRAME_STORE_RA(___R0)
  ___W_ALL
  overflow = ___garbage_collect ();
  ___R_ALL
  ___SET_R0(___FRAME_FETCH_RA)

  ___RESULT = ___BOOLEAN(overflow);

  ___COVER_GC_WITHOUT_EXCEPTIONS;
}
"))))

;------------------------------------------------------------------------------

; Memory allocation

(define ##heartbeat-interval-set!
  (let ()

    (##declare (not inline) (not interrupts-enabled))

    (lambda (seconds)
      (##c-code
"
{
  ___FLONUM_VAL(___ARG2) = ___set_heartbeat_interval (___FLONUM_VAL(___ARG1));
  ___RESULT = ___ARG2;
}
"
       seconds
       (##flonum.<-fixnum 0)))))

(define ##set-gc-report!
  (let ()

    (##declare (not inline) (not interrupts-enabled))

    (lambda (report?)
      (##c-code
"
{
  ___set_gc_report (___ARG1 != ___FAL);
  ___RESULT = ___VOID;
}
"
       report?))))

(define ##make-vector
  (let ()

    (##declare (not inline) (not interrupts-enabled))

    (lambda (n #!optional (init (macro-absent-obj)))
      (let ((v (##c-code "
{
long i;
long n = ___INT(___ARG1);
long words = n + 1;
___WORD result;
if (n > (___LMASK>>(___LF+___LWS)))
  result = ___FAL; /* abort if requested object is too big */
else if (words > ___MSECTION_BIGGEST)
  {
    ___FRAME_STORE_RA(___R0)
    ___W_ALL
    result = ___alloc_scmobj (___sVECTOR, n<<___LWS, ___STILL);
    ___R_ALL
    ___SET_R0(___FRAME_FETCH_RA)
    if (result != ___FAL)
      ___still_obj_refcount_dec (result);
  }
else
  {
    ___BOOL overflow = 0;
    ___hp += words;
    if (___hp > ___ps->heap_limit)
      {
#ifdef mem_trace
printf (\"------------------------- ##make-vector\\n\"); fflush(stdout);
#endif
        ___FRAME_STORE_RA(___R0)
        ___W_ALL
        overflow = ___heap_limit () && ___garbage_collect ();
        ___R_ALL
        ___SET_R0(___FRAME_FETCH_RA)
      }
    else
      ___hp -= words;
    if (overflow)
      result = ___FAL;
    else
      {
        result = ___TAG(___hp, ___tSUBTYPED);
        ___HEADER(result) = ___MAKE_HD_WORDS(n, ___sVECTOR);
        ___hp += words;
      }
  }
if (result != ___FAL)
  {
    ___WORD init = ___ARG2;
    if (init == ___ABSENT)
      init = ___FIX(0);
    for (i=0; i<n; i++)
      ___VECTORSET(result,___FIX(i),init)
  }
___RESULT = result;
}
" n init)))
    (or v
        (begin
          (##exception.heap-overflow)
          (##make-vector n init)))))))

(define ##make-string
  (let ()

    (##declare (not inline) (not interrupts-enabled))

    (lambda (n #!optional (init (macro-absent-obj)))
      (let ((s (##c-code "
{
long i;
long n = ___INT(___ARG1);
long words = ___WORDS((n<<___LCS)) + 1;
___WORD result;
if (n > (___LMASK>>(___LF+___LCS)))
  result = ___FAL; /* abort if requested object is too big */
else if (words > ___MSECTION_BIGGEST)
  {
    ___FRAME_STORE_RA(___R0)
    ___W_ALL
    result = ___alloc_scmobj (___sSTRING, n<<___LCS, ___STILL);
    ___R_ALL
    ___SET_R0(___FRAME_FETCH_RA)
    if (result != ___FAL)
      ___still_obj_refcount_dec (result);
  }
else
  {
    ___BOOL overflow = 0;
    ___hp += words;
    if (___hp > ___ps->heap_limit)
      {
        ___FRAME_STORE_RA(___R0)
        ___W_ALL
        overflow = ___heap_limit () && ___garbage_collect ();
        ___R_ALL
        ___SET_R0(___FRAME_FETCH_RA)
      }
    else
      ___hp -= words;
    if (overflow)
      result = ___FAL;
    else
      {
        result = ___TAG(___hp, ___tSUBTYPED);
        ___HEADER(result) = ___MAKE_HD_BYTES((n<<___LCS), ___sSTRING);
        ___hp += words;
      }
  }
if (result != ___FAL && ___ARG2 != ___ABSENT)
  {
    for (i=0; i<n; i++)
      ___STRINGSET(result,___FIX(i),___ARG2);
  }
___RESULT = result;
}
" n init)))
    (or s
        (begin
          (##exception.heap-overflow)
          (##make-string n init)))))))

(define ##make-s8vector
  (let ()

    (##declare (not inline) (not interrupts-enabled))

    (lambda (n #!optional (init (macro-absent-obj)))
      (let ((v (##c-code "
{
long i;
long n = ___INT(___ARG1);
long words = ___WORDS(n) + 1;
___WORD result;
if (n > (___LMASK>>___LF))
  result = ___FAL; /* abort if requested object is too big */
else if (words > ___MSECTION_BIGGEST)
  {
    ___FRAME_STORE_RA(___R0)
    ___W_ALL
    result = ___alloc_scmobj (___sS8VECTOR, n, ___STILL);
    ___R_ALL
    ___SET_R0(___FRAME_FETCH_RA)
    if (result != ___FAL)
      ___still_obj_refcount_dec (result);
  }
else
  {
    ___BOOL overflow = 0;
    ___hp += words;
    if (___hp > ___ps->heap_limit)
      {
        ___FRAME_STORE_RA(___R0)
        ___W_ALL
        overflow = ___heap_limit () && ___garbage_collect ();
        ___R_ALL
        ___SET_R0(___FRAME_FETCH_RA)
      }
    else
      ___hp -= words;
    if (overflow)
      result = ___FAL;
    else
      {
        result = ___TAG(___hp, ___tSUBTYPED);
        ___HEADER(result) = ___MAKE_HD_BYTES(n, ___sS8VECTOR);
        ___hp += words;
      }
  }
if (result != ___FAL && ___ARG2 != ___ABSENT)
  {
    for (i=0; i<n; i++)
      ___S8VECTORSET(result,___FIX(i),___ARG2)
  }
___RESULT = result;
}
" n init)))
    (or v
        (begin
          (##exception.heap-overflow)
          (##make-s8vector n init)))))))

(define ##make-u8vector
  (let ()

    (##declare (not inline) (not interrupts-enabled))

    (lambda (n #!optional (init (macro-absent-obj)))
      (let ((v (##c-code "
{
long i;
long n = ___INT(___ARG1);
long words = ___WORDS(n) + 1;
___WORD result;
if (n > (___LMASK>>___LF))
  result = ___FAL; /* abort if requested object is too big */
else if (words > ___MSECTION_BIGGEST)
  {
    ___FRAME_STORE_RA(___R0)
    ___W_ALL
    result = ___alloc_scmobj (___sU8VECTOR, n, ___STILL);
    ___R_ALL
    ___SET_R0(___FRAME_FETCH_RA)
    if (result != ___FAL)
      ___still_obj_refcount_dec (result);
  }
else
  {
    ___BOOL overflow = 0;
    ___hp += words;
    if (___hp > ___ps->heap_limit)
      {
        ___FRAME_STORE_RA(___R0)
        ___W_ALL
        overflow = ___heap_limit () && ___garbage_collect ();
        ___R_ALL
        ___SET_R0(___FRAME_FETCH_RA)
      }
    else
      ___hp -= words;
    if (overflow)
      result = ___FAL;
    else
      {
        result = ___TAG(___hp, ___tSUBTYPED);
        ___HEADER(result) = ___MAKE_HD_BYTES(n, ___sU8VECTOR);
        ___hp += words;
      }
  }
if (result != ___FAL && ___ARG2 != ___ABSENT)
  {
    for (i=0; i<n; i++)
      ___U8VECTORSET(result,___FIX(i),___ARG2)
  }
___RESULT = result;
}
" n init)))
    (or v
        (begin
          (##exception.heap-overflow)
          (##make-u8vector n init)))))))

(define ##make-s16vector
  (let ()

    (##declare (not inline) (not interrupts-enabled))

    (lambda (n #!optional (init (macro-absent-obj)))
      (let ((v (##c-code "
{
long i;
long n = ___INT(___ARG1);
long words = ___WORDS((n<<1)) + 1;
___WORD result;
if (n > (___LMASK>>(___LF+1)))
  result = ___FAL; /* abort if requested object is too big */
else if (words > ___MSECTION_BIGGEST)
  {
    ___FRAME_STORE_RA(___R0)
    ___W_ALL
    result = ___alloc_scmobj (___sS16VECTOR, n<<1, ___STILL);
    ___R_ALL
    ___SET_R0(___FRAME_FETCH_RA)
    if (result != ___FAL)
      ___still_obj_refcount_dec (result);
  }
else
  {
    ___BOOL overflow = 0;
    ___hp += words;
    if (___hp > ___ps->heap_limit)
      {
        ___FRAME_STORE_RA(___R0)
        ___W_ALL
        overflow = ___heap_limit () && ___garbage_collect ();
        ___R_ALL
        ___SET_R0(___FRAME_FETCH_RA)
      }
    else
      ___hp -= words;
    if (overflow)
      result = ___FAL;
    else
      {
        result = ___TAG(___hp, ___tSUBTYPED);
        ___HEADER(result) = ___MAKE_HD_BYTES((n<<1), ___sS16VECTOR);
        ___hp += words;
      }
  }
if (result != ___FAL && ___ARG2 != ___ABSENT)
  {
    for (i=0; i<n; i++)
      ___S16VECTORSET(result,___FIX(i),___ARG2)
  }
___RESULT = result;
}
" n init)))
    (or v
        (begin
          (##exception.heap-overflow)
          (##make-s16vector n init)))))))

(define ##make-u16vector
  (let ()

    (##declare (not inline) (not interrupts-enabled))

    (lambda (n #!optional (init (macro-absent-obj)))
      (let ((v (##c-code "
{
long i;
long n = ___INT(___ARG1);
long words = ___WORDS((n<<1)) + 1;
___WORD result;
if (n > (___LMASK>>(___LF+1)))
  result = ___FAL; /* abort if requested object is too big */
else if (words > ___MSECTION_BIGGEST)
  {
    ___FRAME_STORE_RA(___R0)
    ___W_ALL
    result = ___alloc_scmobj (___sU16VECTOR, n<<1, ___STILL);
    ___R_ALL
    ___SET_R0(___FRAME_FETCH_RA)
    if (result != ___FAL)
      ___still_obj_refcount_dec (result);
  }
else
  {
    ___BOOL overflow = 0;
    ___hp += words;
    if (___hp > ___ps->heap_limit)
      {
        ___FRAME_STORE_RA(___R0)
        ___W_ALL
        overflow = ___heap_limit () && ___garbage_collect ();
        ___R_ALL
        ___SET_R0(___FRAME_FETCH_RA)
      }
    else
      ___hp -= words;
    if (overflow)
      result = ___FAL;
    else
      {
        result = ___TAG(___hp, ___tSUBTYPED);
        ___HEADER(result) = ___MAKE_HD_BYTES((n<<1), ___sU16VECTOR);
        ___hp += words;
      }
  }
if (result != ___FAL && ___ARG2 != ___ABSENT)
  {
    for (i=0; i<n; i++)
      ___U16VECTORSET(result,___FIX(i),___ARG2)
  }
___RESULT = result;
}
" n init)))
    (or v
        (begin
          (##exception.heap-overflow)
          (##make-u16vector n init)))))))

(define ##make-s32vector
  (let ()

    (##declare (not inline) (not interrupts-enabled))

    (lambda (n #!optional (init (macro-absent-obj)))
      (let ((v (##c-code "
{
long i;
long n = ___INT(___ARG1);
long words = ___WORDS((n<<2)) + 1;
___WORD result;
if (n > (___LMASK>>(___LF+2)))
  result = ___FAL; /* abort if requested object is too big */
else if (words > ___MSECTION_BIGGEST)
  {
    ___FRAME_STORE_RA(___R0)
    ___W_ALL
    result = ___alloc_scmobj (___sS32VECTOR, n<<2, ___STILL);
    ___R_ALL
    ___SET_R0(___FRAME_FETCH_RA)
    if (result != ___FAL)
      ___still_obj_refcount_dec (result);
  }
else
  {
    ___BOOL overflow = 0;
    ___hp += words;
    if (___hp > ___ps->heap_limit)
      {
        ___FRAME_STORE_RA(___R0)
        ___W_ALL
        overflow = ___heap_limit () && ___garbage_collect ();
        ___R_ALL
        ___SET_R0(___FRAME_FETCH_RA)
      }
    else
      ___hp -= words;
    if (overflow)
      result = ___FAL;
    else
      {
        result = ___TAG(___hp, ___tSUBTYPED);
        ___HEADER(result) = ___MAKE_HD_BYTES((n<<2), ___sS32VECTOR);
        ___hp += words;
      }
  }
if (result != ___FAL && ___ARG2 != ___ABSENT)
  {
    for (i=0; i<n; i++)
      ___S32VECTORSET(result,___FIX(i),___ARG2)
  }
___RESULT = result;
}
" n init)))
    (or v
        (begin
          (##exception.heap-overflow)
          (##make-s32vector n init)))))))

(define ##make-u32vector
  (let ()

    (##declare (not inline) (not interrupts-enabled))

    (lambda (n #!optional (init (macro-absent-obj)))
      (let ((v (##c-code "
{
long i;
long n = ___INT(___ARG1);
long words = ___WORDS((n<<2)) + 1;
___WORD result;
if (n > (___LMASK>>(___LF+2)))
  result = ___FAL; /* abort if requested object is too big */
else if (words > ___MSECTION_BIGGEST)
  {
    ___FRAME_STORE_RA(___R0)
    ___W_ALL
    result = ___alloc_scmobj (___sU32VECTOR, n<<2, ___STILL);
    ___R_ALL
    ___SET_R0(___FRAME_FETCH_RA)
    if (result != ___FAL)
      ___still_obj_refcount_dec (result);
  }
else
  {
    ___BOOL overflow = 0;
    ___hp += words;
    if (___hp > ___ps->heap_limit)
      {
        ___FRAME_STORE_RA(___R0)
        ___W_ALL
        overflow = ___heap_limit () && ___garbage_collect ();
        ___R_ALL
        ___SET_R0(___FRAME_FETCH_RA)
      }
    else
      ___hp -= words;
    if (overflow)
      result = ___FAL;
    else
      {
        result = ___TAG(___hp, ___tSUBTYPED);
        ___HEADER(result) = ___MAKE_HD_BYTES((n<<2), ___sU32VECTOR);
        ___hp += words;
      }
  }
if (result != ___FAL && ___ARG2 != ___ABSENT)
  {
    for (i=0; i<n; i++)
      ___U32VECTORSET(result,___FIX(i),___ARG2)
  }
___RESULT = result;
}
" n init)))
    (or v
        (begin
          (##exception.heap-overflow)
          (##make-u32vector n init)))))))

(define ##make-s64vector
  (let ()

    (##declare (not inline) (not interrupts-enabled))

    (lambda (n #!optional (init (macro-absent-obj)))
      (let ((v (##c-code "
{
long i;
long n = ___INT(___ARG1);
#if ___WS == 4
long words = ___WORDS((n<<3)) + 2;
#else
long words = ___WORDS((n<<3)) + 1;
#endif
___WORD result;
if (n > (___LMASK>>(___LF+3)))
  result = ___FAL; /* abort if requested object is too big */
else if (words > ___MSECTION_BIGGEST)
  {
    ___FRAME_STORE_RA(___R0)
    ___W_ALL
    result = ___alloc_scmobj (___sS64VECTOR, n<<3, ___STILL);
    ___R_ALL
    ___SET_R0(___FRAME_FETCH_RA)
    if (result != ___FAL)
      ___still_obj_refcount_dec (result);
  }
else
  {
    ___BOOL overflow = 0;
    ___hp += words;
    if (___hp > ___ps->heap_limit)
      {
        ___FRAME_STORE_RA(___R0)
        ___W_ALL
        overflow = ___heap_limit () && ___garbage_collect ();
        ___R_ALL
        ___SET_R0(___FRAME_FETCH_RA)
      }
    else
      ___hp -= words;
    if (overflow)
      result = ___FAL;
    else
      {
#if ___WS == 4
        result = ___TAG(((___WORD*)((___WORD)(___hp+2)&~7)-1), ___tSUBTYPED);
#else
        result = ___TAG(___hp, ___tSUBTYPED);
#endif
        ___HEADER(result) = ___MAKE_HD_BYTES((n<<3), ___sS64VECTOR);
        ___hp += words;
      }
  }
if (result != ___FAL && ___ARG2 != ___ABSENT)
  {
    for (i=0; i<n; i++)
      ___S64VECTORSET(result,___FIX(i),___ARG2)
  }
___RESULT = result;
}
" n init)))
    (or v
        (begin
          (##exception.heap-overflow)
          (##make-s64vector n init)))))))

(define ##make-u64vector
  (let ()

    (##declare (not inline) (not interrupts-enabled))

    (lambda (n #!optional (init (macro-absent-obj)))
      (let ((v (##c-code "
{
long i;
long n = ___INT(___ARG1);
#if ___WS == 4
long words = ___WORDS((n<<3)) + 2;
#else
long words = ___WORDS((n<<3)) + 1;
#endif
___WORD result;
if (n > (___LMASK>>(___LF+3)))
  result = ___FAL; /* abort if requested object is too big */
else if (words > ___MSECTION_BIGGEST)
  {
    ___FRAME_STORE_RA(___R0)
    ___W_ALL
    result = ___alloc_scmobj (___sU64VECTOR, n<<3, ___STILL);
    ___R_ALL
    ___SET_R0(___FRAME_FETCH_RA)
    if (result != ___FAL)
      ___still_obj_refcount_dec (result);
  }
else
  {
    ___BOOL overflow = 0;
    ___hp += words;
    if (___hp > ___ps->heap_limit)
      {
        ___FRAME_STORE_RA(___R0)
        ___W_ALL
        overflow = ___heap_limit () && ___garbage_collect ();
        ___R_ALL
        ___SET_R0(___FRAME_FETCH_RA)
      }
    else
      ___hp -= words;
    if (overflow)
      result = ___FAL;
    else
      {
#if ___WS == 4
        result = ___TAG(((___WORD*)((___WORD)(___hp+2)&~7)-1), ___tSUBTYPED);
#else
        result = ___TAG(___hp, ___tSUBTYPED);
#endif
        ___HEADER(result) = ___MAKE_HD_BYTES((n<<3), ___sU64VECTOR);
        ___hp += words;
      }
  }
if (result != ___FAL && ___ARG2 != ___ABSENT)
  {
    for (i=0; i<n; i++)
      ___U64VECTORSET(result,___FIX(i),___ARG2)
  }
___RESULT = result;
}
" n init)))
    (or v
        (begin
          (##exception.heap-overflow)
          (##make-u64vector n init)))))))

(define ##make-f32vector
  (let ()

    (##declare (not inline) (not interrupts-enabled))

    (lambda (n #!optional (init (macro-absent-obj)))
      (let ((v (##c-code "
{
long i;
long n = ___INT(___ARG1);
long words = ___WORDS((n<<2)) + 1;
___WORD result;
if (n > (___LMASK>>(___LF+2)))
  result = ___FAL; /* abort if requested object is too big */
else if (words > ___MSECTION_BIGGEST)
  {
    ___FRAME_STORE_RA(___R0)
    ___W_ALL
    result = ___alloc_scmobj (___sF32VECTOR, n<<2, ___STILL);
    ___R_ALL
    ___SET_R0(___FRAME_FETCH_RA)
    if (result != ___FAL)
      ___still_obj_refcount_dec (result);
  }
else
  {
    ___BOOL overflow = 0;
    ___hp += words;
    if (___hp > ___ps->heap_limit)
      {
        ___FRAME_STORE_RA(___R0)
        ___W_ALL
        overflow = ___heap_limit () && ___garbage_collect ();
        ___R_ALL
        ___SET_R0(___FRAME_FETCH_RA)
      }
    else
      ___hp -= words;
    if (overflow)
      result = ___FAL;
    else
      {
        result = ___TAG(___hp, ___tSUBTYPED);
        ___HEADER(result) = ___MAKE_HD_BYTES((n<<2), ___sF32VECTOR);
        ___hp += words;
      }
  }
if (result != ___FAL && ___ARG2 != ___ABSENT)
  {
    ___F64 init = ___F64UNBOX(___ARG2);
    for (i=0; i<n; i++)
      ___F32VECTORSET(result,___FIX(i),init)
  }
___RESULT = result;
}
" n init)))
    (or v
        (begin
          (##exception.heap-overflow)
          (##make-f32vector n init)))))))

(define ##make-f64vector
  (let ()

    (##declare (not inline) (not interrupts-enabled))

    (lambda (n #!optional (init (macro-absent-obj)))
      (let ((v (##c-code "
{
long i;
long n = ___INT(___ARG1);
#if ___WS == 4
long words = ___WORDS((n<<3)) + 2;
#else
long words = ___WORDS((n<<3)) + 1;
#endif
___WORD result;
if (n > (___LMASK>>(___LF+3)))
  result = ___FAL; /* abort if requested object is too big */
else if (words > ___MSECTION_BIGGEST)
  {
    ___FRAME_STORE_RA(___R0)
    ___W_ALL
    result = ___alloc_scmobj (___sF64VECTOR, n<<3, ___STILL);
    ___R_ALL
    ___SET_R0(___FRAME_FETCH_RA)
    if (result != ___FAL)
      ___still_obj_refcount_dec (result);
  }
else
  {
    ___BOOL overflow = 0;
    ___hp += words;
    if (___hp > ___ps->heap_limit)
      {
        ___FRAME_STORE_RA(___R0)
        ___W_ALL
        overflow = ___heap_limit () && ___garbage_collect ();
        ___R_ALL
        ___SET_R0(___FRAME_FETCH_RA)
      }
    else
      ___hp -= words;
    if (overflow)
      result = ___FAL;
    else
      {
#if ___WS == 4
        result = ___TAG(((___WORD*)((___WORD)(___hp+2)&~7)-1), ___tSUBTYPED);
#else
        result = ___TAG(___hp, ___tSUBTYPED);
#endif
        ___HEADER(result) = ___MAKE_HD_BYTES((n<<3), ___sF64VECTOR);
        ___hp += words;
      }
  }
if (result != ___FAL && ___ARG2 != ___ABSENT)
  {
    ___F64 init = ___F64UNBOX(___ARG2);
    for (i=0; i<n; i++)
      ___F64VECTORSET(result,___FIX(i),init)
  }
___RESULT = result;
}
" n init)))
    (or v
        (begin
          (##exception.heap-overflow)
          (##make-f64vector n init)))))))

(define ##bignum.make
  (let ()

    (##declare (not inline) (not interrupts-enabled))

    (lambda (n x complement?)
      (let ((v (##c-code "
{
long i;
long n = ___INT(___ARG1);
#if ___BIG_ABASE_WIDTH == 32
long words = ___WORDS((n*(___BIG_ABASE_WIDTH/8))) + 1;
#else
#if ___WS == 4
long words = ___WORDS((n*(___BIG_ABASE_WIDTH/8))) + 2;
#else
long words = ___WORDS((n*(___BIG_ABASE_WIDTH/8))) + 1;
#endif
#endif
___WORD result;

if (n > (___LMASK>>___LF)/(___BIG_ABASE_WIDTH/8))
  result = ___FAL; /* abort if requested object is too big */
else if (words > ___MSECTION_BIGGEST)
  {
    ___FRAME_STORE_RA(___R0)
    ___W_ALL
#if ___BIG_ABASE_WIDTH == 32
    result = ___alloc_scmobj (___sBIGNUM, n<<2, ___STILL);
#else
    result = ___alloc_scmobj (___sBIGNUM, n<<3, ___STILL);
#endif
    ___R_ALL
    ___SET_R0(___FRAME_FETCH_RA)
    if (result != ___FAL)
      ___still_obj_refcount_dec (result);
  }
else
  {
    ___BOOL overflow = 0;
    ___hp += words;
    if (___hp > ___ps->heap_limit)
      {
        ___FRAME_STORE_RA(___R0)
        ___W_ALL
        overflow = ___heap_limit () && ___garbage_collect ();
        ___R_ALL
        ___SET_R0(___FRAME_FETCH_RA)
      }
    else
      ___hp -= words;
    if (overflow)
      result = ___FAL;
    else
      {
#if ___BIG_ABASE_WIDTH == 32
        result = ___TAG(___hp, ___tSUBTYPED);
#else
#if ___WS == 4
        result = ___TAG(((___WORD*)((___WORD)(___hp+2)&~7)-1), ___tSUBTYPED);
#else
        result = ___TAG(___hp, ___tSUBTYPED);
#endif
#endif
#if ___BIG_ABASE_WIDTH == 32
        ___HEADER(result) = ___MAKE_HD_BYTES((n<<2), ___sBIGNUM);
#else
        ___HEADER(result) = ___MAKE_HD_BYTES((n<<3), ___sBIGNUM);
#endif
        ___hp += words;
      }
  }
if (result != ___FAL)
  {
    ___WORD x = ___ARG2;
    ___WORD len;
    if (x == ___FAL)
      len = 0;
    else
      {
        len = ___INT(___BIGALENGTH(x));
        if (len > n)
          len = n;
      }
#if ___BIG_ABASE_WIDTH == 32
    if (___ARG3 == ___FAL)
      {
        for (i=0; i<len; i++)
          ___STORE_U32(___BODY_AS(result,___tSUBTYPED),i,
                       ___FETCH_U32(___BODY_AS(x,___tSUBTYPED),i));
        if (x == ___FAL ||
            ___FETCH_U32(___BODY_AS(x,___tSUBTYPED),(i-1))<___BIG_ABASE_DIV_2)
          for (; i<n; i++)
            ___STORE_U32(___BODY_AS(result,___tSUBTYPED),i,0);
        else
          for (; i<n; i++)
            ___STORE_U32(___BODY_AS(result,___tSUBTYPED),i,___BIG_ABASE_MIN_1);
      }
    else
      {
        for (i=0; i<len; i++)
          ___STORE_U32(___BODY_AS(result,___tSUBTYPED),i,
                       ~___FETCH_U32(___BODY_AS(x,___tSUBTYPED),i));
        if (x == ___FAL ||
            ___FETCH_U32(___BODY_AS(x,___tSUBTYPED),(i-1))<___BIG_ABASE_DIV_2)
          for (; i<n; i++)
            ___STORE_U32(___BODY_AS(result,___tSUBTYPED),i,___BIG_ABASE_MIN_1);
        else
          for (; i<n; i++)
            ___STORE_U32(___BODY_AS(result,___tSUBTYPED),i,0);
      }
#else
    if (___ARG3 == ___FAL)
      {
        for (i=0; i<len; i++)
          ___STORE_U64(___BODY_AS(result,___tSUBTYPED),i,
                       ___FETCH_U64(___BODY_AS(x,___tSUBTYPED),i));
        if (x == ___FAL ||
            ___FETCH_U64(___BODY_AS(x,___tSUBTYPED),(i-1))<___BIG_ABASE_DIV_2)
          for (; i<n; i++)
            ___STORE_U64(___BODY_AS(result,___tSUBTYPED),i,0);
        else
          for (; i<n; i++)
            ___STORE_U64(___BODY_AS(result,___tSUBTYPED),i,___BIG_ABASE_MIN_1);
      }
    else
      {
        for (i=0; i<len; i++)
          ___STORE_U64(___BODY_AS(result,___tSUBTYPED),i,
                       ~___FETCH_U64(___BODY_AS(x,___tSUBTYPED),i));
        if (x == ___FAL ||
            ___FETCH_U64(___BODY_AS(x,___tSUBTYPED),(i-1))<___BIG_ABASE_DIV_2)
          for (; i<n; i++)
            ___STORE_U64(___BODY_AS(result,___tSUBTYPED),i,___BIG_ABASE_MIN_1);
        else
          for (; i<n; i++)
            ___STORE_U64(___BODY_AS(result,___tSUBTYPED),i,0);
      }
#endif
  }
___RESULT = result;
}
" n x complement?)))
    (or v
        (begin
          (##exception.heap-overflow)
          (##bignum.make n x complement?)))))))

;------------------------------------------------------------------------------

; Exceptions raised by low level runtime system

(define ##exception.global-jump
  (let ()

    (##declare (not inline) (not interrupts-enabled))

    (lambda (gv . args)
      (let ((val (##global-var-ref gv)))
        (macro-force-vars (val)
          (if (##procedure? val)
            (##apply val args)
            (if (##unbound? val)
              (##signal '##signal.global-unbound-operator gv args)
              (##signal '##signal.global-non-procedure-operator gv args))))))))

(define ##exception.non-proc-jump
  (let ()

    (##declare (not inline) (not interrupts-enabled))

    (lambda (proc . args)
      (macro-force-vars (proc)
        (if (##procedure? proc)
          (##apply proc args)
          (##signal '##signal.non-procedure-jump proc args))))))

(define ##exception.wrong-nb-arg
  (let ()

    (##declare (not inline) (not interrupts-enabled))

    (lambda (proc . args)
      (##signal '##signal.wrong-nb-arg proc args))))

(define ##exception.unknown-keyword-arg
  (let ()

    (##declare (not inline) (not interrupts-enabled))

    (lambda (proc . args)
      (##signal '##signal.unknown-keyword-arg proc args))))

(define ##exception.keyword-expected
  (let ()

    (##declare (not inline) (not interrupts-enabled))

    (lambda (proc . args)
      (##signal '##signal.keyword-expected proc args))))

(define ##exception.clam-conv-error
  (let ()

    (##declare (not inline) (not interrupts-enabled))

    (lambda (err-code err-msg proc . args)
      (##signal '##signal.clam-conv-error err-code err-msg proc args))))

(define ##exception.cdef-conv-error
  (let ()

    (##declare (not inline) (not interrupts-enabled))

    (lambda (err-code err-msg proc)
      (##signal '##signal.cdef-conv-error err-code err-msg proc))))

(define ##exception.multiple-c-return
  (let ()

    (##declare (not inline) (not interrupts-enabled))

    (lambda ()
      (##signal '##signal.multiple-c-return))))

(define ##exception.apply-arg-limit
  (let ()

    (##declare (not inline) (not interrupts-enabled))

    (lambda (proc args)
      (##signal '##signal.apply-arg-limit proc args))))

(define ##exception.heap-overflow
  (let ()

    (##declare (not inline) (not interrupts-enabled))

    (lambda ()
      (##signal '##signal.heap-overflow))))

(define ##exception.stack-overflow
  (let ()

    (##declare (not inline) (not interrupts-enabled))

    (lambda ()
      (##signal '##signal.stack-overflow))))

(define ##exception.promise-already-determined
  (let ()

    (##declare (not inline) (not interrupts-enabled))

    (lambda ()
      (##signal '##signal.promise-already-determined))))

(define ##interrupt-handler
  (let ()

    (##declare (not inline) (not interrupts-enabled))

    (lambda (code)
      (let ((proc (##vector-ref ##interrupt-vector code)))
        (if (##procedure? proc)
          (proc))))))

(define ##interrupt-vector-set!
  (let ()

    (##declare (not inline) (not interrupts-enabled))

    (lambda (code handler)
      (##vector-set! ##interrupt-vector code handler))))

(define ##interrupt-vector (##vector #f #f #f))

;------------------------------------------------------------------------------

; Procedure application and continuations

(define ##apply
  (let ()

    (##declare (not inline) (not interrupts-enabled))

    (lambda (proc args)

      (##declare (interrupts-enabled)) ; make sure stack overflow is checked

      (define apply
        (lambda (proc args)
          (##c-code
"
{
  ___WORD proc;
  ___WORD args;
  ___WORD lst;
  int na;

  ___POP_ARGS2(proc,args)

  lst = args;
  na = 0;

  while (___PAIRP(lst))
    {
      ___PUSH(___CAR(lst))
      lst = ___CDR(lst);
      na++;

      if (na > ___MAX_NB_ARGS)
        {
          ___ADJFP(-na); /* remove pushed arguments */

          ___PUSH_ARGS2(proc,args)

          ___COVER_APPLY_ARG_LIMIT;

          ___JUMPPRM(___SET_NARGS(2),
                     ___G__23__23_exception_2e_apply_2d_arg_2d_limit.prm)
        }
    }

  ___POP_ARGS_IN_REGS(na) /* load register arguments */

  ___COVER_APPLY_ARG_LIMIT_END;

  ___JUMPEXTNOTSAFE(___SET_NARGS(na),proc)
}
")))

      (apply proc args))))

(define ##call-with-current-continuation
  (let ()

    (##declare (not inline))

    (lambda (receiver)
      (##continuation-capture
       (lambda (cont receiver)
         (receiver (lambda (result)
                     (##continuation-return cont result))))
       receiver))))

(define ##procedure->continuation
  (let ()

    (##declare (not inline))

    (lambda (proc)
      (##closure-ref proc 1))))

(define ##continuation-ret
  (let ()

    (##declare (not inline))

    (lambda (cont)
      (##c-code
"
{
  ___WORD frame = ___FIELD(___ARG1,___CONTINUATION_FRAME);
  ___WORD ra;

  if (___TYP(frame) == ___tSUBTYPED)
    {
      /* continuation frame is in the heap */

      ra = ___FIELD(frame,0);

      if (ra == ___internal_return)
        ra = ___FIELD(frame,___FRAME_RETI_RA);
    }
  else
    {
      /* continuation frame is in the stack */

      ___WORD *fp = ___CAST(___WORD*,frame);

      ra = *fp;

      if (ra == ___internal_return)
        ra = fp[-___RETI_RA];
    }

  ___RESULT = ra;
}
"
       cont))))

(define ##continuation-denv
  (let ()

    (##declare (not inline))

    (lambda (cont)
      (##c-code "___RESULT = ___FIELD(___ARG1,___CONTINUATION_DENV);" cont))))

(define ##continuation-fs
  (let ()

    (##declare (not inline))

    (lambda (cont)
      (##c-code
"
{
  ___WORD frame = ___FIELD(___ARG1,___CONTINUATION_FRAME);
  ___WORD ra;
  int fs;

  if (___TYP(frame) == ___tSUBTYPED)
    {
      /* continuation frame is in the heap */

      ra = ___FIELD(frame,0);

      if (ra == ___internal_return)
        ___RETI_GET_FS(___FIELD(frame,___FRAME_RETI_RA),fs)
      else
        ___RETN_GET_FS(ra,fs)
    }
  else
    {
      /* continuation frame is in the stack */

      ___WORD *fp = ___CAST(___WORD*,frame);

      ra = *fp;

      if (ra == ___internal_return)
        ___RETI_GET_FS(fp[-___RETI_RA],fs)
      else
        ___RETN_GET_FS(ra,fs)
    }

  ___RESULT = ___FIX(fs);
}
"
       cont))))

(define ##continuation-ref
  (let ()

    (##declare (not inline))

    (lambda (cont i)
      (##c-code
"
{
  ___WORD frame = ___FIELD(___ARG1,___CONTINUATION_FRAME);
  int i = ___INT(___ARG2);
  ___WORD ra;
  int fs, link;

  if (___TYP(frame) == ___tSUBTYPED)
    {
      /* continuation frame is in the heap */

      ra = ___FIELD(frame,0);

      if (ra == ___internal_return)
        ___RETI_GET_FS_LINK(___BODY_AS(frame,___tSUBTYPED)[___FRAME_RETI_RA],fs,link)
      else
        ___RETN_GET_FS_LINK(ra,fs,link)

      ___RESULT = ___BODY_AS(frame,___tSUBTYPED)[fs-i+1];  /* what if i==link and frame is first in section???? */
/*      if (i == link) ___RESULT = ___FIX(999999);/***********/
    }
  else
    {
      /* continuation frame is in the stack */

      ra = *(___WORD*)frame;

      if (ra == ___internal_return)
        ___RETI_GET_FS_LINK(((___WORD*)frame)[-___RETI_RA],fs,link)
      else
        ___RETN_GET_FS_LINK(ra,fs,link)

      ___RESULT = ((___WORD*)frame)[___FRAME_SPACE(fs)-i];  /* what if i==link and frame is first in section???? */
/*      if (i == link) ___RESULT = ___FIX(999999);/***********/
    }
}
"
       cont
       i))))

(define ##continuation-next
  (let ()

    (##declare (not inline))

    (lambda (cont)
      (let ((next (##c-code
"
{
  ___WORD f = ___ARG1;
  ___WORD frame = ___VECTORREF(f,___FIX(0));
  ___WORD denv  = ___VECTORREF(f,___FIX(1));
  ___WORD ra;
  ___WORD *fp, frame_ra, next_frame;
  int fs, link;

  if (___TYP(frame)==___tSUBTYPED)
    {
      /* continuation frame is in the heap */

      ra = ___FIELD(frame,0);

      fp = ___BODY_AS(frame,___tSUBTYPED);

      if (ra == ___internal_return)
        ___RETI_GET_FS_LINK(fp[___FRAME_RETI_RA],fs,link)
      else
        ___RETN_GET_FS_LINK(ra,fs,link)

      fp += fs+1;

      next_frame = fp[-link-1];

      if (next_frame == 0)
        ___RESULT = ___FAL;
      else
        {
          ___hp[0]=___MAKE_HD_WORDS(___CONTINUATION_SIZE,___sCONTINUATION);
          ___ADD_VECTOR_ELEM(0,next_frame)
          ___ADD_VECTOR_ELEM(1,denv)
          ___hp+=___CONTINUATION_SIZE+1;
          ___RESULT = ___GET_VECTOR(___CONTINUATION_SIZE);
        }
    }
  else
    {
      /* continuation frame is in the stack */

      ra = *(___WORD*)frame;

      if (ra == ___internal_return)
        ___RETI_GET_FS_LINK(((___WORD*)frame)[-___RETI_RA],fs,link)
      else
        ___RETN_GET_FS_LINK(ra,fs,link)

      fp = ((___WORD*)frame)+___FRAME_SPACE(fs);
      frame_ra = fp[-link-1];

      if (frame_ra == ___ps->handler_break)
        {
          /* first frame of that section */

          next_frame = fp[___BREAK_FRAME_NEXT];
        }
      else
        {
          /* not the first frame of that section */

          *fp = frame_ra;
          next_frame = (___WORD)fp;
        }

      if (next_frame == 0)
        ___RESULT = ___FAL;
      else
        {
          ___hp[0]=___MAKE_HD_WORDS(___CONTINUATION_SIZE,___sCONTINUATION);
          ___ADD_VECTOR_ELEM(0,next_frame)
          ___ADD_VECTOR_ELEM(1,denv)
          ___hp+=___CONTINUATION_SIZE+1;
          ___RESULT = ___GET_VECTOR(___CONTINUATION_SIZE);
        }
    }
}
"
                   cont)))
        (##check-heap)
        next))))

(define ##closure?
  (let ()

    (##declare (not inline))

    (lambda (proc)
      (##c-code
"
{
  if (___HD_TYP(___HEADER(___ARG1)) != ___PERM)
    ___RESULT = ___TRU;
  else
    ___RESULT = ___FAL;
}
"
       proc))))

(define ##subprocedure?
  (let ()

    (##declare (not inline))

    (lambda (proc)
      (##c-code
"
{
  if (___TYP(___ARG1) == ___tSUBTYPED &&
      ((___label_struct*)(___ARG1-___tSUBTYPED))->entry_or_descr == ___ARG1 &&
      ((___WORD *)(___ARG1-___tSUBTYPED))[-___LS] != ___MAKE_HD((___INTRO_SIZE<<___LWS),___sVECTOR,___PERM))
    ___RESULT = ___TRU;
  else
    ___RESULT = ___FAL;
}
"
       proc))))

(define ##subprocedure-id
  (let ()

    (##declare (not inline))

    (lambda (proc)
      (##c-code
"
{
  if (___TYP(___ARG1) == ___tSUBTYPED)
  {
    ___WORD *start = ((___WORD *)(___ARG1-___tSUBTYPED));
    ___WORD *ptr = start;
    while (*ptr != ___MAKE_HD((___INTRO_SIZE<<___LWS),___sVECTOR,___PERM)) ptr -= ___LS;
    ptr += ___LS;
    ___RESULT = ___FIX( (start-ptr)/___LS );
  }
  else
    ___RESULT = ___FIX(0);
}
"
       proc))))

(define ##subprocedure-parent
  (let ()

    (##declare (not inline))

    (lambda (proc)
      (##c-code
"
{
  if (___TYP(___ARG1) == ___tSUBTYPED)
  {
    ___WORD *start = ((___WORD *)(___ARG1-___tSUBTYPED));
    ___WORD *ptr = start;
    while (*ptr != ___MAKE_HD((___INTRO_SIZE<<___LWS),___sVECTOR,___PERM)) ptr -= ___LS;
    ptr += ___LS;
   ___RESULT = ___TAG(ptr,___tSUBTYPED);
  }
  else
    ___RESULT = ___FAL;
}
"
       proc))))

(define ##procedure-info
  (let ()

    (##declare (not inline))

    (lambda (proc)
      (##c-code
"
{
  if (___TYP(___ARG1) == ___tSUBTYPED)
  {
    ___WORD *start = ((___WORD *)(___ARG1-___tSUBTYPED));
    ___WORD *ptr = start;
    while (*ptr != ___MAKE_HD((___INTRO_SIZE<<___LWS),___sVECTOR,___PERM)) ptr -= ___LS;
    ___RESULT = ptr[1];
  }
  else
    ___RESULT = ___FAL;
}
"
       proc))))

;------------------------------------------------------------------------------

; Global variables

(define ##make-global-var
  (let ()

    (##declare (not inline))

    (lambda (id)
      (##c-code
"
{
  if (___FIELD(___ARG1,___SYMBOL_GLOBAL) == 0)
    {
      ___glo_struct *p = ___alloc_global_var ();
      if (p == 0)
        ___RESULT = ___FAL;
      else
        {
          p->val = ___UNB1;
          p->prm = ___FAL;
          p->next = 0;
          if (___ps->glo_list_head == 0)
            ___ps->glo_list_head = ___CAST(___WORD,p);
          else
            ___CAST(___glo_struct*,___ps->glo_list_tail)->next
              = ___CAST(___WORD,p);
          ___ps->glo_list_tail = ___CAST(___WORD,p);
          ___FIELD(___ARG1,___SYMBOL_GLOBAL) = ___CAST(___WORD,p);
          ___RESULT = ___ARG1;
        }
    }
  else
    ___RESULT = ___ARG1;
}
"
       id))))

(define ##object->global-var->identifier
  (let ()

    (##declare (not inline))

    (lambda (obj)
      (##c-code
"
{
  ___WORD p = ___ps->glo_list_head;
  while (p != 0 && ((___glo_struct*)p)->val != ___ARG1)
    p = ((___glo_struct*)p)->next;
  ___RESULT = ___FAL;
  if (p != 0)
    {
      int len = ___INT(___VECTORLENGTH(___symbol_table));
      int i;

      for (i=0; i<len; i++)
        {
          ___WORD probe = ___FIELD(___symbol_table,i);

          while (probe != ___NUL)
            {
              if (___FIELD(probe,___SYMBOL_GLOBAL) == p)
                {
                  ___RESULT = probe;
                  goto end_search;
                }
              probe = ___FIELD(probe,___SYMBOL_NEXT);
            }
        }
      end_search:;
    }
}
"
       obj))))

(define ##global-var->identifier
  (let ()

    (##declare (not inline))

    (lambda (gv)
      gv)))

(define ##make-interned-symbol
  (let ()

    (##declare (not inline))

    (lambda (name)
      (##make-interned-symkey name #t))))

(define ##make-interned-keyword
  (let ()

    (##declare (not inline))

    (lambda (name)
      (##make-interned-symkey name #f))))

(define ##make-interned-symkey
  (let ()

    (##declare (not inline))

    (lambda (name symbol?)
      (let ((obj ((c-lambda (scheme-object bool) scheme-object
"
{
  unsigned int subtype = ___arg2 ? ___sSYMBOL : ___sKEYWORD;
  ___WORD obj = ___find_symkey_from_schemestring (___arg1, subtype);
  if (obj == ___FAL)
    {
      unsigned long n = ___INT(___STRINGLENGTH(___arg1));
      obj = ___alloc_scmobj (___sSTRING, n<<___LCS, ___PERM);
      if (obj != ___FAL)
        {
          memcpy (___BODY_AS(obj,___tSUBTYPED),
                  ___BODY_AS(___arg1,___tSUBTYPED),
                  n<<___LCS);
          obj = ___new_symkey (obj, subtype);
        }
    }
  ___result = obj;
}
")
                  name
                  symbol?)))
        (or obj
            (begin
              (##exception.heap-overflow)
              (##make-interned-symkey name symbol?)))))))

(define ##symbol-hash-counter 0)

(define ##make-uninterned-symbol
  (let ()

    (##declare (not inline))

    (lambda (name) ; name must be a non-mutable string
      (let ((i ##symbol-hash-counter))
        (set! ##symbol-hash-counter
          (if (##fixnum.= i ##max-fixnum) 0 (##fixnum.+ i 1)))
        (macro-make-uninterned-symbol name i)))))

;------------------------------------------------------------------------------

; Miscellaneous

(define ##max-char
  (##c-code "___RESULT = ___FIX(___MAX_CHR);"))

(define ##min-fixnum
  (##c-code "___RESULT = ___FIX(___MIN_FIX);"))

(define ##max-fixnum
  (##c-code "___RESULT = ___FIX(___MAX_FIX);"))

(define ##fixnum-width
  (##c-code "___RESULT = ___FIX((8*___WS)-___TB);"))

(define ##bignum.adigit-width
  (##c-code "___RESULT = ___FIX(___BIG_ABASE_WIDTH);"))

(define ##bignum.mdigit-width
  (##c-code "___RESULT = ___FIX(___BIG_MBASE_WIDTH);"))

(define ##bignum.fdigit-width
  (##c-code "___RESULT = ___FIX(___BIG_FBASE_WIDTH);"))

(define ##symbol-table
  (c-lambda () scheme-object "___result = ___symbol_table;"))

(define ##keyword-table
  (c-lambda () scheme-object "___result = ___keyword_table;"))

(define ##argv
  (c-lambda () scheme-object "___result = ___arguments;"))

(define ##processed-argv '())
(set! ##processed-argv (##argv))

(define ##process-times
  (let ()

    (##declare (not inline))

    (lambda ()
      (let ((v
             (##f64vector (macro-inexact-+0)
                          (macro-inexact-+0)
                          (macro-inexact-+0))))
        (##c-code
"
{
  ___WORD result = ___ARG1;
  ___F64 user, sys, real;
  ___process_times (&user, &sys, &real);
  ___F64VECTORSET(result,___FIX(0),user)
  ___F64VECTORSET(result,___FIX(1),sys)
  ___F64VECTORSET(result,___FIX(2),real)
  ___RESULT = result;
}
"
         v)))))

(define ##get-current-time!
  (let ()

    (##declare (not inline) (not interrupts-enabled))

    (lambda (floats)
      (##c-code
"
{
  ___time t;
  ___F64 ft;

  ___time_get_current_time (&t);
  ___time_to_seconds (t, &ft);
  ___F64VECTORSET(___ARG1,___FIX(0),ft)

  ___RESULT = ___VOID;
}
"
       floats))))

(define ##condvar-select!
  (let ()

    (##declare (not inline) (not interrupts-enabled))

    (lambda (runq timeout)
      (##c-code
       "___RESULT = ___condvar_select (___ARG1, ___ARG2);"
       runq
       timeout))))

(define ##disable-interrupts!
  (let ()

    (##declare (not inline) (not interrupts-enabled))

    (lambda ()
      (##c-code "___disable_interrupts (); ___RESULT = ___VOID;"))))

(define ##enable-interrupts!
  (let ()

    (##declare (not inline) (not interrupts-enabled))

    (lambda ()
      (##c-code "___enable_interrupts (); ___RESULT = ___VOID;"))))

(define ##process-statistics
  (let ()

    (##declare (not inline) (not interrupts-enabled))

    (lambda ()
      (let ((v
             (##f64vector (macro-inexact-+0)
                          (macro-inexact-+0)
                          (macro-inexact-+0)
                          (macro-inexact-+0)
                          (macro-inexact-+0)
                          (macro-inexact-+0)
                          (macro-inexact-+0)
                          (macro-inexact-+0)
                          (macro-inexact-+0)
                          (macro-inexact-+0)
                          (macro-inexact-+0)
                          (macro-inexact-+0))))
        (##c-code
"
{
  int n;
  ___F64 user, sys, real;
  long minflt, majflt;
  ___W_ALL
  ___process_times (&user, &sys, &real);
  ___vm_stats (&minflt, &majflt);
  ___F64VECTORSET(___ARG1,___FIX(0),user)
  ___F64VECTORSET(___ARG1,___FIX(1),sys)
  ___F64VECTORSET(___ARG1,___FIX(2),real)
  ___F64VECTORSET(___ARG1,___FIX(3),___GSTATE->gc_user_time)
  ___F64VECTORSET(___ARG1,___FIX(4),___GSTATE->gc_sys_time)
  ___F64VECTORSET(___ARG1,___FIX(5),___GSTATE->gc_real_time)
  ___F64VECTORSET(___ARG1,___FIX(6),___GSTATE->nb_gcs)
  ___F64VECTORSET(___ARG1,___FIX(7),___bytes_allocated ())
  ___F64VECTORSET(___ARG1,___FIX(8),(2*(1+2)<<___LWS))
  n = (1<<___LWS) + 12 * sizeof (___F64) + (5<<___LWS);
  ___F64VECTORSET(___ARG1,___FIX(9),(2*n))
  ___F64VECTORSET(___ARG1,___FIX(10),minflt)
  ___F64VECTORSET(___ARG1,___FIX(11),majflt)
  ___R_ALL
  ___RESULT = ___ARG1;
}
"
         v)))))

(define ##check-heap
  (let ()

    (##declare (not inline) (not interrupts-enabled))

    (lambda ()
      (let ((x (##c-code
"
{
  if (___hp > ___ps->heap_limit)
    {
      ___BOOL overflow;
#ifdef mem_trace
printf (\"------------------------- ##check-heap\\n\"); fflush(stdout);
#endif
      ___FRAME_STORE_RA(___R0)
      ___W_ALL
      overflow = ___heap_limit () && ___garbage_collect ();
      ___R_ALL
      ___SET_R0(___FRAME_FETCH_RA)
      if (overflow)
        ___RESULT = ___TRU;
      else
        ___RESULT = ___FAL;
    }
  else
    ___RESULT = ___FAL;
}
")))
        (if x
          (begin
            (##exception.heap-overflow)
            (##check-heap)))))))

(define ##rest-param-heap-overflow
  (let ()

    (##declare (not inline) (not interrupts-enabled))

    (lambda (proc args)
      (##exception.heap-overflow)
      (##apply proc args))))

(define ##force-undetermined
  (let ()

    (##declare (not inline) (not interrupts-enabled))

    (lambda (promise thunk)
      (let ((result (##force (thunk))))
        (##c-code
"
{
  if (___FIELD(___ARG1,___PROMISE_RESULT) == ___ARG1)
    {
      ___FIELD(___ARG1,___PROMISE_RESULT) = ___ARG2;
      ___FIELD(___ARG1,___PROMISE_THUNK) = ___FAL;
    }
  ___RESULT = ___FIELD(___ARG1,___PROMISE_RESULT);
}
"
         promise
         result)))))

(define ##exit
  (let ()

    (##declare (not inline) (not interrupts-enabled))

    (lambda (#!optional (status 0))
      (##force-output-on-predefined)
      (##cleanup-and-exit-process status))))

(define ##cleanup-and-exit-process
  (let ()

    (##declare (not inline) (not interrupts-enabled))

    (lambda (#!optional (status 0))
      (##c-code "___cleanup_and_exit_process (___INT(___ARG1));" status))))

(define ##sequentially
  (let ()

    (##declare (not inline) (not interrupts-enabled))

    (lambda (thunk)
      (thunk))))

(define ##getenv
  (c-lambda (char-string) char-string "___getenv"))

(define ##shell-command
  (c-lambda (char-string) int "___shell_command"))

(define ##path-expand
  (let ()

    (##declare (not inline) (not interrupts-enabled))

    (lambda (path format)
      ((c-lambda (char-string int) char-string
"
{
  char path[___PATH_MAX_LENGTH+1];
  if (___path_expand (___arg1, path, ___PATH_MAX_LENGTH, ___arg2))
    ___result = path;
  else
    ___result = 0;
}
")
       path
       (case format
         ((relative) 1)
         ((shortest) -1)
         (else 0))))))

(define ##path-absolute?
  (c-lambda (char-string) bool "___path_absolute"))

(define ##path-extension
  (c-lambda (char-string) char-string "
char path[___PATH_MAX_LENGTH+1];
if (___path_extension (___arg1, path, ___PATH_MAX_LENGTH))
  ___result = path;
else
  ___result = 0;
"))

(define ##path-strip-extension
  (c-lambda (char-string) char-string "
char path[___PATH_MAX_LENGTH+1];
if (___path_strip_extension (___arg1, path, ___PATH_MAX_LENGTH))
  ___result = path;
else
  ___result = 0;
"))

(define ##path-directory
  (c-lambda (char-string) char-string "
char path[___PATH_MAX_LENGTH+1];
if (___path_directory (___arg1, path, ___PATH_MAX_LENGTH))
  ___result = path;
else
  ___result = 0;
"))

(define ##path-strip-directory
  (c-lambda (char-string) char-string "
char path[___PATH_MAX_LENGTH+1];
if (___path_strip_directory (___arg1, path, ___PATH_MAX_LENGTH))
  ___result = path;
else
  ___result = 0;
"))

(define ##path-drive
  (c-lambda (char-string) char-string "
char path[___PATH_MAX_LENGTH+1];
if (___path_drive (___arg1, path, ___PATH_MAX_LENGTH))
  ___result = path;
else
  ___result = 0;
"))

(define ##path-strip-drive
  (c-lambda (char-string) char-string "
char path[___PATH_MAX_LENGTH+1];
if (___path_strip_drive (___arg1, path, ___PATH_MAX_LENGTH))
  ___result = path;
else
  ___result = 0;
"))

(define (##load-object-file name)
  (let ((vect ((c-lambda (char-string) scheme-object "
{
  char *errmsg;
#if 0
  if (___garbage_collect ()) /* to prevent gc while loading */
    ___result = ___FAL;
  else
#endif
    {
      ___WORD result = ___load_object_file (___arg1, &errmsg);
      if (result == ___FAL)
        {
          ___err = ___CHARSTRING_to_SCMOBJ (errmsg, &result, 0);
          ___release_scmobj (result);
        }
      ___result = result;
    }
}
") name)))
    (cond ((##vector? vect)
           (lambda ()
             (let ((len (##vector-length vect)))
               (let loop ((i 0))
                 (if (##fixnum.< i len)
                   (if (##fixnum.= i (##fixnum.- len 1))
                     ((##vector-ref vect i))
                     (begin
                       ((##vector-ref vect i))
                       (loop (##fixnum.+ i 1)))))))))
          ((##string? vect)
           vect)
          (else
           (##exception.heap-overflow)
           (##load-object-file name)))))

(define ##cc-cmd
  ((c-lambda () char-string "___result = ___CC_CMD;")))

(define ##ld-cmd
  ((c-lambda () char-string "___result = ___LD_CMD;")))

(define ##ld-flip
  ((c-lambda () bool "
#ifdef ___LD_FLIP
  ___result = 1;
#else
  ___result = 0;
#endif
")))

(define ##dynamic-cc #f)

(if ##cc-cmd
  (set! ##dynamic-cc
    (lambda (root-path output-path)
      ((c-lambda (char-string char-string char-string char-string bool)
                 char-string
"
char *errmsg;
if (___dynamic_cc (___arg1, ___arg2, ___arg3, ___arg4, ___arg5, &errmsg))
  ___result = errmsg;
else
  ___result = 0;
") ##cc-cmd ##ld-cmd root-path output-path ##ld-flip))))

(define ##format-filepos
  (c-lambda (char-string long bool) char-string "___format_filepos"))

(define ##err-code->string
  (c-lambda (int)
	    char-string
"
char str[___ERR_MAX_LENGTH+1];
___err_code_to_string (___arg1, str);
___result = str;"))

(##declare (separate))

(define (##comply-to-standard-scheme?)
  (##c-code "___RESULT = ___BOOLEAN(___setup_params.standard);"))

(define (##gc)
  (let ((result (##c-code "
{
  ___BOOL overflow;
#ifdef mem_trace
printf (\"------------------------- ##gc\\n\"); fflush(stdout);
#endif
  ___FRAME_STORE_RA(___R0)
  ___W_ALL
  overflow = ___garbage_collect ();
  ___R_ALL
  ___SET_R0(___FRAME_FETCH_RA)
  ___RESULT = ___BOOLEAN(overflow);
}
")))
    (if result
      (begin
        (##exception.heap-overflow)
        (##gc))
      (##void))))

(define (##gc-finalization)
  (##declare (not interrupts-enabled))
  (let ((action (##c-code "
{
  ___WORD will = ___ps->executable_wills;
  if (___UNTAG(will) == 0) /* end of list? */
    ___RESULT = ___FAL;
  else
    {
      ___ps->executable_wills = ___BODY(will)[0];
      ___RESULT = ___BODY(will)[2];
      ___BODY(will)[2] = ___FAL; /* zap action procedure */
    }
}")))
    (if action
      (begin
        (action)
        (##gc-finalization)))))

; execute each module in sequence

(define ##module-sequencer
  (let ()
    (##declare (not inline))
    (lambda ()
      (let loop ((i 1))
        (let ((ev (##c-code "___RESULT = ___exec_vector;")))
          (let ((len (##vector-length ev)))
            (if (##fixnum.< i len)
              (begin
                ((##vector-ref ev i))
                (loop (##fixnum.+ i 1)))
              (##force-output-on-predefined))))))))

;------------------------------------------------------------------------------

(define ##err-code-EAGAIN
  (##c-code "___RESULT = ___FIX(___ERR_CODE_FROM_ERRNO(EAGAIN));"))

(define ##err-code-EINTR
  (##c-code "___RESULT = ___FIX(___ERR_CODE_FROM_ERRNO(EINTR));"))

;------------------------------------------------------------------------------

(define ##pointer-address
  (let ()

    (##declare (not inline) (not interrupts-enabled))

    (lambda (ptr)
      (let ((addr (##vector-ref ptr 0)))

        ; Beware! "addr" is a real pointer.  If the GC ever gets hold
        ; of it, thinking that it is a reference to a Scheme datum, it
        ; is bound to cause trouble.  Fortunately, because "addr" is
        ; dead after the binding of "hi" and "lo", the compiler will
        ; overwrite the register containing "addr" with "lo".

        (let* ((hi (##type-cast addr (macro-type-fixnum)))
               (lo (##type addr)))
          (##+ (##* (if (##fixnum.< hi 0) (##- hi ##bignum.2*min-fixnum) hi)
                    4)
               lo))))))

(##module-sequencer)
