;==============================================================================

; file: "_num#.scm"

; Copyright (C) 1994-2000 by Marc Feeley, All Rights Reserved.

;==============================================================================

; Error handling.

(##define-macro (macro-check-real var form expr)
  (define (flat x) (if (pair? x) (cons (car x) (flat (cdr x))) (list x)))
  `(macro-if-checks
     (if (##real? ,var)
       ,expr
       ,(if (list? form)
          `(##trap-check-real ',(car form) ,@(cdr form))
          `(##trap-check-real* ',(car form) ,@(flat (cdr form)))))
     ,expr))

(##define-macro (macro-check-inexact-real var form expr)
  (define (flat x) (if (pair? x) (cons (car x) (flat (cdr x))) (list x)))
  `(macro-if-checks
     (if (##flonum? ,var)
       ,expr
       ,(if (list? form)
          `(##trap-check-inexact-real ',(car form) ,@(cdr form))
          `(##trap-check-inexact-real* ',(car form) ,@(flat (cdr form)))))
     ,expr))

(##define-macro (macro-check-nonnegative-real var form expr)
  (define (flat x) (if (pair? x) (cons (car x) (flat (cdr x))) (list x)))
  `(macro-if-checks
     (if (and (##real? ,var)
              (##not (##negative? ,var)))
       ,expr
       ,(if (list? form)
          `(##trap-check-nonnegative-real ',(car form) ,@(cdr form))
          `(##trap-check-nonnegative-real* ',(car form) ,@(flat (cdr form)))))
     ,expr))

(##define-macro (macro-check-exact-int-non-neg var form expr)
  (define (flat x) (if (pair? x) (cons (car x) (flat (cdr x))) (list x)))
  `(macro-if-checks
     (if (##fixnum? ,var)
       (if (##not (##fixnum.< ,var 0))
         ,expr
         ,(if (list? form)
            `(##trap-check-range ',(car form) ,@(cdr form))
            `(##trap-check-range* ',(car form) ,@(flat (cdr form)))))
       (if (##bignum? ,var)
         ,(if (list? form)
            `(##trap-check-range ',(car form) ,@(cdr form))
            `(##trap-check-range* ',(car form) ,@(flat (cdr form))))
         ,(if (list? form)
            `(##trap-check-exact-int ',(car form) ,@(cdr form))
            `(##trap-check-exact-int* ',(car form) ,@(flat (cdr form))))))
     ,expr))

(##define-macro (macro-check-exact-int-range var lo hi form expr)
  (define (flat x) (if (pair? x) (cons (car x) (flat (cdr x))) (list x)))
  `(macro-if-checks
     (if (##fixnum? ,var)
       (if (and (##not (##fixnum.< ,var ,lo))
                (##fixnum.< ,var ,hi))
         ,expr
         ,(if (list? form)
            `(##trap-check-range ',(car form) ,@(cdr form))
            `(##trap-check-range* ',(car form) ,@(flat (cdr form)))))
       (if (##bignum? ,var)
         ,(if (list? form)
            `(##trap-check-range ',(car form) ,@(cdr form))
            `(##trap-check-range* ',(car form) ,@(flat (cdr form))))
         ,(if (list? form)
            `(##trap-check-exact-int ',(car form) ,@(cdr form))
            `(##trap-check-exact-int* ',(car form) ,@(flat (cdr form))))))
     ,expr))

(##define-macro (macro-check-exact-int-range-incl var lo hi form expr)
  (define (flat x) (if (pair? x) (cons (car x) (flat (cdr x))) (list x)))
  `(macro-if-checks
     (if (##fixnum? ,var)
       (if (and (##not (##fixnum.< ,var ,lo))
                (##not (##fixnum.< ,hi ,var)))
         ,expr
         ,(if (list? form)
            `(##trap-check-range ',(car form) ,@(cdr form))
            `(##trap-check-range* ',(car form) ,@(flat (cdr form)))))
       (if (##bignum? ,var)
         ,(if (list? form)
            `(##trap-check-range ',(car form) ,@(cdr form))
            `(##trap-check-range* ',(car form) ,@(flat (cdr form))))
         ,(if (list? form)
            `(##trap-check-exact-int ',(car form) ,@(cdr form))
            `(##trap-check-exact-int* ',(car form) ,@(flat (cdr form))))))
     ,expr))

(##define-macro (macro-check-exact-signed-int8 var form expr)
  (define (flat x) (if (pair? x) (cons (car x) (flat (cdr x))) (list x)))
  `(macro-if-checks
     (if (##fixnum? ,var)
       (if (and (##not (##fixnum.< ,var -128))
                (##not (##fixnum.< 127 ,var)))
         ,expr
         ,(if (list? form)
            `(##trap-check-range ',(car form) ,@(cdr form))
            `(##trap-check-range* ',(car form) ,@(flat (cdr form)))))
       (if (##bignum? ,var)
         ,(if (list? form)
            `(##trap-check-range ',(car form) ,@(cdr form))
            `(##trap-check-range* ',(car form) ,@(flat (cdr form))))
         ,(if (list? form)
            `(##trap-check-exact-int ',(car form) ,@(cdr form))
            `(##trap-check-exact-int* ',(car form) ,@(flat (cdr form))))))
     ,expr))

(##define-macro (macro-check-exact-unsigned-int8 var form expr)
  (define (flat x) (if (pair? x) (cons (car x) (flat (cdr x))) (list x)))
  `(macro-if-checks
     (if (##fixnum? ,var)
       (if (and (##not (##negative? ,var))
                (##not (##fixnum.< 255 ,var)))
         ,expr
         ,(if (list? form)
            `(##trap-check-range ',(car form) ,@(cdr form))
            `(##trap-check-range* ',(car form) ,@(flat (cdr form)))))
       (if (##bignum? ,var)
         ,(if (list? form)
            `(##trap-check-range ',(car form) ,@(cdr form))
            `(##trap-check-range* ',(car form) ,@(flat (cdr form))))
         ,(if (list? form)
            `(##trap-check-exact-int ',(car form) ,@(cdr form))
            `(##trap-check-exact-int* ',(car form) ,@(flat (cdr form))))))
     ,expr))

(##define-macro (macro-check-exact-signed-int16 var form expr)
  (define (flat x) (if (pair? x) (cons (car x) (flat (cdr x))) (list x)))
  `(macro-if-checks
     (if (##fixnum? ,var)
       (if (and (##not (##fixnum.< ,var -32768))
                (##not (##fixnum.< 32767 ,var)))
         ,expr
         ,(if (list? form)
            `(##trap-check-range ',(car form) ,@(cdr form))
            `(##trap-check-range* ',(car form) ,@(flat (cdr form)))))
       (if (##bignum? ,var)
         ,(if (list? form)
            `(##trap-check-range ',(car form) ,@(cdr form))
            `(##trap-check-range* ',(car form) ,@(flat (cdr form))))
         ,(if (list? form)
            `(##trap-check-exact-int ',(car form) ,@(cdr form))
            `(##trap-check-exact-int* ',(car form) ,@(flat (cdr form))))))
     ,expr))

(##define-macro (macro-check-exact-unsigned-int16 var form expr)
  (define (flat x) (if (pair? x) (cons (car x) (flat (cdr x))) (list x)))
  `(macro-if-checks
     (if (##fixnum? ,var)
       (if (and (##not (##negative? ,var))
                (##not (##fixnum.< 65535 ,var)))
         ,expr
         ,(if (list? form)
            `(##trap-check-range ',(car form) ,@(cdr form))
            `(##trap-check-range* ',(car form) ,@(flat (cdr form)))))
       (if (##bignum? ,var)
         ,(if (list? form)
            `(##trap-check-range ',(car form) ,@(cdr form))
            `(##trap-check-range* ',(car form) ,@(flat (cdr form))))
         ,(if (list? form)
            `(##trap-check-exact-int ',(car form) ,@(cdr form))
            `(##trap-check-exact-int* ',(car form) ,@(flat (cdr form))))))
     ,expr))

(##define-macro (macro-check-exact-signed-int32 var form expr)
  (define (flat x) (if (pair? x) (cons (car x) (flat (cdr x))) (list x)))
  `(macro-if-checks
     (if (or (##fixnum? ,var) (##bignum? ,var))
       (if (and (##not (##< ,var -2147483648))
                (##not (##< 2147483647 ,var)))
         ,expr
         ,(if (list? form)
            `(##trap-check-range ',(car form) ,@(cdr form))
            `(##trap-check-range* ',(car form) ,@(flat (cdr form)))))
       (if (##bignum? ,var)
         ,(if (list? form)
            `(##trap-check-range ',(car form) ,@(cdr form))
            `(##trap-check-range* ',(car form) ,@(flat (cdr form))))
         ,(if (list? form)
            `(##trap-check-exact-int ',(car form) ,@(cdr form))
            `(##trap-check-exact-int* ',(car form) ,@(flat (cdr form))))))
     ,expr))

(##define-macro (macro-check-exact-unsigned-int32 var form expr)
  (define (flat x) (if (pair? x) (cons (car x) (flat (cdr x))) (list x)))
  `(macro-if-checks
     (if (or (##fixnum? ,var) (##bignum? ,var))
       (if (and (##not (##negative? ,var))
                (##not (##< 4294967295 ,var)))
         ,expr
         ,(if (list? form)
            `(##trap-check-range ',(car form) ,@(cdr form))
            `(##trap-check-range* ',(car form) ,@(flat (cdr form)))))
       (if (##bignum? ,var)
         ,(if (list? form)
            `(##trap-check-range ',(car form) ,@(cdr form))
            `(##trap-check-range* ',(car form) ,@(flat (cdr form))))
         ,(if (list? form)
            `(##trap-check-exact-int ',(car form) ,@(cdr form))
            `(##trap-check-exact-int* ',(car form) ,@(flat (cdr form))))))
     ,expr))

(##define-macro (macro-check-exact-signed-int64 var form expr)
  (define (flat x) (if (pair? x) (cons (car x) (flat (cdr x))) (list x)))
  `(macro-if-checks
     (if (or (##fixnum? ,var) (##bignum? ,var))
       (if (and (##not (##< ,var -9223372036854775808))
                (##not (##< 9223372036854775807 ,var)))
         ,expr
         ,(if (list? form)
            `(##trap-check-range ',(car form) ,@(cdr form))
            `(##trap-check-range* ',(car form) ,@(flat (cdr form)))))
       (if (##bignum? ,var)
         ,(if (list? form)
            `(##trap-check-range ',(car form) ,@(cdr form))
            `(##trap-check-range* ',(car form) ,@(flat (cdr form))))
         ,(if (list? form)
            `(##trap-check-exact-int ',(car form) ,@(cdr form))
            `(##trap-check-exact-int* ',(car form) ,@(flat (cdr form))))))
     ,expr))

(##define-macro (macro-check-exact-unsigned-int64 var form expr)
  (define (flat x) (if (pair? x) (cons (car x) (flat (cdr x))) (list x)))
  `(macro-if-checks
     (if (or (##fixnum? ,var) (##bignum? ,var))
       (if (and (##not (##negative? ,var))
                (##not (##< 18446744073709551615 ,var)))
         ,expr
         ,(if (list? form)
            `(##trap-check-range ',(car form) ,@(cdr form))
            `(##trap-check-range* ',(car form) ,@(flat (cdr form)))))
       (if (##bignum? ,var)
         ,(if (list? form)
            `(##trap-check-range ',(car form) ,@(cdr form))
            `(##trap-check-range* ',(car form) ,@(flat (cdr form))))
         ,(if (list? form)
            `(##trap-check-exact-int ',(car form) ,@(cdr form))
            `(##trap-check-exact-int* ',(car form) ,@(flat (cdr form))))))
     ,expr))

;==============================================================================

; Number representation.

; There are 5 internal representations for numbers:
;
; fixnum, bignum, ratnum, flonum, cpxnum
;
; Fixnums and bignums form the class of exact-int.
; Fixnums, bignums and ratnums form the class of exact-real.
; Fixnums, bignums, ratnums and flonums form the class of noncpxnum.

; The representation has some invariants:
;
; The numerator of a ratnum is a non-zero exact-int.
; The denominator of a ratnum is an exact-int greater than 1.
; The numerator and denominator have no common divisors greater than 1.
;
; The real part of a cpxnum is a noncpxnum.
; The imaginary part of a cpxnum is a noncpxnum != fixnum 0

; The following table gives the mapping of the Scheme exact numbers to their
; internal representation:
;
;    type          representation
; exact integer  = exact-int (fixnum, bignum)
; exact rational = exact-real (fixnum, bignum, ratnum)
; exact real     = exact-real (fixnum, bignum, ratnum)
; exact complex  = exact-real or cpxnum with exact-real real and imag parts

; For inexact numbers, the representation is not quite as straightforward.
;
; There are 3 special classes of inexact representation:
; flonum-int : flonum with integer value
; cpxnum-real: cpxnum with imag part = flonum 0.0 or -0.0
; cpxnum-int : cpxnum-real with exact-int or flonum-int real part
;
; Note: cpxnum-real and cpxnum-int only exist if
; (macro-cpxnum-are-possibly-real?) returns #t.
;
; This gives the following table for Scheme's inexact numbers:
;
;      type          representation
; inexact integer  = flonum-int or cpxnum-int if it exists
; inexact rational = flonum     or cpxnum-real if it exists
; inexact real     = flonum     or cpxnum-real if it exists
; inexact complex  = flonum     or cpxnum with flonum real or imag part

(##define-macro (macro-special-case-exact-zero?) #t); (+ -0. 0)=-0.  (* 4. 0)=0
(##define-macro (macro-cpxnum-are-possibly-real?) #f)

(##define-macro (macro-exact-int? obj) ; obj can be any object
  `(or (##fixnum? ,obj)
       (##bignum? ,obj)))

(##define-macro (macro-exact-real? obj) ; obj can be any object
  `(or (macro-exact-int? ,obj)
       (##ratnum? ,obj)))

(##define-macro (macro-flonum-int? obj) ; obj must be a flonum
  `(and (##flonum.finite? ,obj)
        (##flonum.= ,obj (##flonum.floor ,obj))))

(##define-macro (macro-flonum-rational? obj) ; obj must be a flonum
  `(##flonum.finite? ,obj))

(##define-macro (macro-noncpxnum-int? obj) ; obj must be in fixnum/bignum/ratnum/flonum
  `(if (##flonum? ,obj)
     (macro-flonum-int? ,obj)
     (##not (##ratnum? ,obj))))

(##define-macro (macro-noncpxnum-rational? obj) ; obj must be in fixnum/bignum/ratnum/flonum
  `(or (##not (##flonum? ,obj))
       (macro-flonum-rational? ,obj)))

(##define-macro (macro-cpxnum-int? obj) ; obj must be a cpxnum
  `(and (macro-cpxnum-are-possibly-real?)
        (macro-cpxnum-real? ,obj)
        (let ((real (macro-cpxnum-real ,obj)))
          (macro-noncpxnum-int? real))))

(##define-macro (macro-cpxnum-rational? obj) ; obj must be a cpxnum
  `(and (macro-cpxnum-are-possibly-real?)
        (let ((imag (macro-cpxnum-imag ,obj)))
          (and (##flonum? imag)
               (##flonum.zero? imag)
               (let ((real (macro-cpxnum-real ,obj)))
                 (macro-noncpxnum-rational? real))))))

(##define-macro (macro-cpxnum-real? obj) ; obj must be a cpxnum
  `(and (macro-cpxnum-are-possibly-real?)
        (let ((imag (macro-cpxnum-imag ,obj)))
          (and (##flonum? imag)
               (##flonum.zero? imag)))))

; Dispatch for number representation

(##define-macro (macro-number-dispatch num err fix big rat flo cpx)
  `(cond ((##fixnum? ,num)                                 ,fix)
         ((##flonum? ,num)                                 ,flo)
         ((##subtyped? ,num)
          (let ((##s (##subtype ,num)))
            (cond ((##fixnum.= ##s (macro-subtype-bignum)) ,big)
                  ((##fixnum.= ##s (macro-subtype-ratnum)) ,rat)
                  ((##fixnum.= ##s (macro-subtype-cpxnum)) ,cpx)
                  (else                                    ,err))))
         (else                                             ,err)))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

; Miscellaneous constants.

(##define-macro (macro-inexact-+2)     2.0)
(##define-macro (macro-inexact--2)    -2.0)
(##define-macro (macro-inexact-+1)     1.0)
(##define-macro (macro-inexact--1)    -1.0)
(##define-macro (macro-inexact-+1/2)   0.5)
(##define-macro (macro-inexact-+0)     0.0)
(##define-macro (macro-inexact--0)    -0.0)
(##define-macro (macro-inexact-+pi)    3.141592653589793)
(##define-macro (macro-inexact--pi)   -3.141592653589793)
(##define-macro (macro-inexact-+pi/2)  1.5707963267948966)
(##define-macro (macro-inexact--pi/2) -1.5707963267948966)
(##define-macro (macro-inexact-+inf)  ''+inf.)
(##define-macro (macro-inexact--inf)  ''-inf.)
(##define-macro (macro-inexact-+nan)  ''+nan.)
(##define-macro (macro-cpxnum-+2i)    +2i)
(##define-macro (macro-cpxnum--i)     -i)
(##define-macro (macro-cpxnum-+i)     +i)

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

; Bignum related constants.

(##define-macro (macro-max-lines)                  65536)
(##define-macro (macro-max-fixnum32-div-max-lines)  8191)

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

; Flonum related constants.

(##define-macro (macro-flonum-m-bits)
  52)

(##define-macro (macro-flonum-m-bits-plus-1)
  53)

(##define-macro (macro-flonum-e-bits)
  11)

(##define-macro (macro-flonum-sign-bit) ; (expt 2 (+ (macro-flonum-e-bits) (macro-flonum-m-bits)))
  #x8000000000000000)

(##define-macro (macro-flonum-m-min) ; (expt 2.0 (macro-flonum-m-bits))
  4503599627370496.0)

(##define-macro (macro-flonum-+m-min) ; (expt 2 (macro-flonum-m-bits))
  4503599627370496)

(##define-macro (macro-flonum-+m-max) ; (- (* macro-flonum-+m-min 2) 1)
  9007199254740991)

(##define-macro (macro-flonum--m-min) ; (- (macro-flonum-+m-min))
  -4503599627370496)

(##define-macro (macro-flonum--m-max) ; (- (macro-flonum-+m-max))
  -9007199254740991)

(##define-macro (macro-flonum-e-bias) ; (- (expt 2 (- (macro-flonum-e-bits) 1)) 1)
  1023)

(##define-macro (macro-flonum-e-bias-plus-1) ; (+ (macro-flonum-e-bias) 1)
  1024)

(##define-macro (macro-flonum-e-bias-minus-1) ; (- (macro-flonum-e-bias) 1)
  1022)

(##define-macro (macro-flonum-e-min) ; (- (+ (macro-flonum-e-bias) (macro-flonum-m-bits) -1))
  -1074)

(##define-macro (macro-flonum-min-normal) ; (expt 2.0 (+ (macro-flonum-m-bits) (macro-flonum-e-min)))
  (expt 2.0 (+ 52 -1074)))

(##define-macro (macro-scale-down) ; (expt 2 (- (+ (quotient (macro-flonum-e-bias-plus-1) 2) 1)))
  (expt 2 -513))

(##define-macro (macro-inexact-scale-down) ; (expt 2.0 (- (+ (quotient (macro-flonum-e-bias-plus-1) 2) 1)))
  (expt 2.0 -513))

(##define-macro (macro-scale-up) ; (expt 2 (+ (quotient (macro-flonum-e-bias-plus-1) 2) (macro-flonum-m-bits-plus-1)))
  (expt 2 565))

(##define-macro (macro-inexact-scale-up) ; (expt 2.0 (+ (quotient (macro-flonum-e-bias-plus-1) 2) (macro-flonum-m-bits-plus-1)))
  (expt 2.0 565))

(##define-macro (macro-inexact-radix) 16384.0) ; (exact->inexact (macro-radix))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

; Ratnum objects.

; A ratnum is represented by an object vector of length 2
; slot 0 = numerator
; slot 1 = denominator

(##define-macro (macro-ratnum-make num den)
  `(##subtype-set!
    (##vector ,num ,den)
    (macro-subtype-ratnum)))

(##define-macro (macro-ratnum-numerator r)          `(macro-slot 0 ,r))
(##define-macro (macro-ratnum-numerator-set! r x)   `(macro-slot 0 ,r ,x))
(##define-macro (macro-ratnum-denominator r)        `(macro-slot 1 ,r))
(##define-macro (macro-ratnum-denominator-set! r x) `(macro-slot 1 ,r ,x))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

; Cpxnum objects.

; A cpxnum is represented by an object vector of length 2
; slot 0 = real
; slot 1 = imag

(##define-macro (macro-cpxnum-make r i)
  `(##subtype-set!
    (##vector ,r ,i)
    (macro-subtype-cpxnum)))

(##define-macro (macro-cpxnum-real c)        `(macro-slot 0 ,c))
(##define-macro (macro-cpxnum-real-set! c x) `(macro-slot 0 ,c ,x))
(##define-macro (macro-cpxnum-imag c)        `(macro-slot 1 ,c))
(##define-macro (macro-cpxnum-imag-set! c x) `(macro-slot 1 ,c ,x))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(##define-macro (macro-bignum-odd? x);;;;;;;;;;;;;;;;;;;;
  `(##fixnum.odd? (##bignum.mdigit-ref ,x 0)))

(##define-macro (macro-real->inexact x)
  `(let ((x ,x))
     (if (##flonum? x)
       x
       (##exact->inexact
        (if (macro-cpxnum-are-possibly-real?)
          (##real-part x)
          x)))))

;==============================================================================
