;==============================================================================

; file: "_std.scm"

; Copyright (C) 1994-2000 by Marc Feeley, All Rights Reserved.

;==============================================================================

(##include "header.scm")

;==============================================================================

; Error handling.

(define (##trap-list-lengths name . args)
  (##runtime-error "Lists are not of equal length" name args))

(define (##trap-list-lengths* name . args)
  (##runtime-error* "Lists are not of equal length" name args))

(define (##trap-list name . args)
  (##runtime-error "LIST expected" name args))

(define (##trap-list* name . args)
  (##runtime-error* "LIST expected" name args))

(define (##trap-check-pair name . args)
  (##runtime-error "PAIR expected" name args))

(define (##trap-check-pair* name . args)
  (##runtime-error* "PAIR expected" name args))

(define (##trap-check-char name . args)
  (##runtime-error "CHARACTER expected" name args))

(define (##trap-check-char* name . args)
  (##runtime-error* "CHARACTER expected" name args))

(define (##trap-check-symbol name . args)
  (##runtime-error "SYMBOL expected" name args))

(define (##trap-check-symbol* name . args)
  (##runtime-error* "SYMBOL expected" name args))

(define (##trap-check-keyword name . args)
  (##runtime-error "KEYWORD expected" name args))

(define (##trap-check-keyword* name . args)
  (##runtime-error* "KEYWORD expected" name args))

(define (##trap-check-string name . args)
  (##runtime-error "STRING expected" name args))

(define (##trap-check-string* name . args)
  (##runtime-error* "STRING expected" name args))

(define (##trap-check-vector name . args)
  (##runtime-error "VECTOR expected" name args))

(define (##trap-check-vector* name . args)
  (##runtime-error* "VECTOR expected" name args))

(define (##trap-check-s8vector name . args)
  (##runtime-error "S8VECTOR expected" name args))

(define (##trap-check-s8vector* name . args)
  (##runtime-error* "S8VECTOR expected" name args))

(define (##trap-check-u8vector name . args)
  (##runtime-error "U8VECTOR expected" name args))

(define (##trap-check-u8vector* name . args)
  (##runtime-error* "U8VECTOR expected" name args))

(define (##trap-check-s16vector name . args)
  (##runtime-error "S16VECTOR expected" name args))

(define (##trap-check-s16vector* name . args)
  (##runtime-error* "S16VECTOR expected" name args))

(define (##trap-check-u16vector name . args)
  (##runtime-error "U16VECTOR expected" name args))

(define (##trap-check-u16vector* name . args)
  (##runtime-error* "U16VECTOR expected" name args))

(define (##trap-check-s32vector name . args)
  (##runtime-error "S32VECTOR expected" name args))

(define (##trap-check-s32vector* name . args)
  (##runtime-error* "S32VECTOR expected" name args))

(define (##trap-check-u32vector name . args)
  (##runtime-error "U32VECTOR expected" name args))

(define (##trap-check-u32vector* name . args)
  (##runtime-error* "U32VECTOR expected" name args))

(define (##trap-check-s64vector name . args)
  (##runtime-error "S64VECTOR expected" name args))

(define (##trap-check-s64vector* name . args)
  (##runtime-error* "S64VECTOR expected" name args))

(define (##trap-check-u64vector name . args)
  (##runtime-error "U64VECTOR expected" name args))

(define (##trap-check-u64vector* name . args)
  (##runtime-error* "U64VECTOR expected" name args))

(define (##trap-check-f32vector name . args)
  (##runtime-error "F32VECTOR expected" name args))

(define (##trap-check-f32vector* name . args)
  (##runtime-error* "F32VECTOR expected" name args))

(define (##trap-check-f64vector name . args)
  (##runtime-error "F64VECTOR expected" name args))

(define (##trap-check-f64vector* name . args)
  (##runtime-error* "F64VECTOR expected" name args))

(define (##trap-check-procedure name . args)
  (##runtime-error "PROCEDURE expected" name args))

(define (##trap-check-procedure* name . args)
  (##runtime-error* "PROCEDURE expected" name args))

;------------------------------------------------------------------------------

; Definition of vector-like data types (i.e. string, vector, s8vector, ...).

(##define-macro (define-vector-procedures
                 name
                 default-elem-value
                 macro-check-elem)

  (define (sym . lst)
    (string->symbol (apply string-append (map symbol->string lst))))

  (let ()
    (define macro-check-vect (sym 'macro-check- name))

    (define ##vect?          (sym '## name '?))
    (define ##make-vect      (sym '##make- name))
    (define ##vect           (sym '## name))
    (define ##vect-length    (sym '## name '-length))
    (define ##vect-ref       (sym '## name '-ref))
    (define ##vect-set!      (sym '## name '-set!))
    (define ##vect->list     (sym '## name '->list))
    (define ##list->vect     (sym '##list-> name))
    (define ##vect-copy      (sym '## name '-copy))
    (define ##vect-fill!     (sym '## name '-fill!))
    (define ##subvect        (sym '##sub name))
    (define ##subvect-move!  (sym '##sub name '-move!))
    (define ##subvect-fill!  (sym '##sub name '-fill!))
    (define ##vect-shrink!   (sym '## name '-shrink!))

    (define vect?            (sym name '?))
    (define make-vect        (sym 'make- name))
    (define vect             (sym name))
    (define vect-length      (sym name '-length))
    (define vect-ref         (sym name '-ref))
    (define vect-set!        (sym name '-set!))
    (define vect->list       (sym name '->list))
    (define list->vect       (sym 'list-> name))
    (define vect-copy        (sym name '-copy))
    (define vect-fill!       (sym name '-fill!))
    (define subvect          (sym 'sub name))

    `(begin

       (define-system (,##vect? x))
       (define (,vect? x) (macro-force-vars (x) (,##vect? x)))

       (define-system (,##make-vect x y))

       (define (,make-vect x #!optional (y (macro-absent-obj)))
         (macro-force-vars (x y)
           (let ((init
                  (if (##eq? y (macro-absent-obj))
                    ,default-elem-value
                    y)))
             (macro-check-exact-int-non-neg x (,make-vect x y)
               (,macro-check-elem init (,make-vect x y)
                 (,##make-vect x init))))))

       (define-system (,##vect . l)
         (,##list->vect l))

       (define (,vect . l)
         (let* ((n (##length l))
                (vect (,##make-vect n ,default-elem-value)))
           (let loop ((x l) (i 0))
             (if (##pair? x)
               (let ((elem (##car x)))
                 (macro-force-vars (elem)
                   (,macro-check-elem elem (,vect . l)
                     (begin
                       (,##vect-set! vect i elem)
                       (loop (##cdr x) (##fixnum.+ i 1))))))
               vect))))

       (define-system (,##vect-length x))

       (define (,vect-length x)
         (macro-force-vars (x)
           (,macro-check-vect x (,vect-length x)
             (,##vect-length x))))

       (define-system (,##vect-ref x y))

       (define (,vect-ref x y)
         (macro-force-vars (x y)
           (,macro-check-vect x (,vect-ref x y)
             (macro-check-exact-int-range
              y
              0
              (,##vect-length x)
              (,vect-ref x y)
              (,##vect-ref x y)))))

       (define-system (,##vect-set! x y z))

       (define (,vect-set! x y z)
         (macro-force-vars (x y)
           (,macro-check-vect x (,vect-set! x y z)
             (macro-check-exact-int-range
              y
              0
              (,##vect-length x) (,vect-set! x y z)
              (,macro-check-elem z (,vect-set! x y z)
                (begin
                  (,##vect-set! x y z)
                  (##void)))))))

       (define-system (,##vect->list vect)
         (let loop ((l '()) (i (##fixnum.- (,##vect-length vect) 1)))
           (if (##fixnum.< i 0)
             l
             (loop (##cons (,##vect-ref vect i) l) (##fixnum.- i 1)))))

       (define (,vect->list vect)
         (macro-force-vars (vect)
           (,macro-check-vect vect (,vect->list vect)
             (let loop ((l '()) (i (##fixnum.- (,##vect-length vect) 1)))
               (if (##fixnum.< i 0)
                 l
                 (loop (##cons (,##vect-ref vect i) l) (##fixnum.- i 1)))))))

       (define-system (,##list->vect l)
         (let* ((n (##length l))
                (vect (,##make-vect n ,default-elem-value)))
           (let loop ((x l) (i 0))
             (if (##pair? x)
               (begin
                 (,##vect-set! vect i (##car x))
                 (loop (##cdr x) (##fixnum.+ i 1)))
               vect))))

       (define (,list->vect l)
         (let loop1 ((x l) (n 0))
           (macro-force-vars (x)
             (if (##pair? x)
               (loop1 (##cdr x) (##fixnum.+ n 1))
               (macro-check-list-end x (,list->vect l)
                 (let ((vect (,##make-vect n ,default-elem-value)))
                   (let loop2 ((x l) (i 0))
                     (macro-force-vars (x)
                       (if (##pair? x)
                         (let ((elem (##car x)))
                           (,macro-check-elem elem (,list->vect l)
                             (begin
                               (,##vect-set! vect i elem)
                               (loop2 (##cdr x) (##fixnum.+ i 1)))))
                         vect)))))))))

       (define-system (,##vect-copy vect)
         (let ((len (,##vect-length vect)))
           (,##subvect-move! vect 0 len (,##make-vect len) 0)))

       (define (,vect-copy vect)
         (macro-force-vars (vect)
           (,macro-check-vect vect (,vect-copy vect)
             (,##vect-copy vect))))

       (define-system (,##vect-fill! vect x)
         (,##subvect-fill! vect 0 (,##vect-length vect) x))

       (define (,vect-fill! vect x)
         (macro-force-vars (vect x)
           (,macro-check-vect vect (,vect-fill! vect x)
             (,macro-check-elem x (,vect-fill! vect x)
               (,##vect-fill! vect x)))))

       (define-system (,##subvect vect start end)
         (,##subvect-move!
          vect
          start
          end
          (,##make-vect (##fixnum.max (##fixnum.- end start) 0))
          0))

       (define (,subvect vect start end)
         (macro-force-vars (vect start end)
           (,macro-check-vect vect (,subvect vect start end)
             (macro-check-exact-int-range-incl
              start
              0
              (,##vect-length vect)
              (,subvect vect start end)
              (macro-check-exact-int-range-incl
               end
               start
               (,##vect-length vect)
               (,subvect vect start end)
               (,##subvect vect start end))))))

       (define (,##subvect-move! src-vect src-start src-end dst-vect dst-start)
         ; Copy direction must be selected in case src-vect and
         ; dst-vect are the same object.
         (if (##fixnum.< src-start dst-start)
           (let loop1 ((i (##fixnum.- src-end 1))
                       (j (##fixnum.-
                           (##fixnum.+ dst-start
                                       (##fixnum.- src-end src-start))
                           1)))
             (if (##fixnum.< i src-start)
               dst-vect
               (begin
                 (,##vect-set! dst-vect j (,##vect-ref src-vect i))
                 (loop1 (##fixnum.- i 1)
                        (##fixnum.- j 1)))))
           (let loop2 ((i src-start)
                       (j dst-start))
             (if (##fixnum.< i src-end)
               (begin
                 (,##vect-set! dst-vect j (,##vect-ref src-vect i))
                 (loop2 (##fixnum.+ i 1)
                        (##fixnum.+ j 1)))
               dst-vect))))

       (define-system (,##subvect-fill! vect start end x)
         (let loop ((i (##fixnum.- end 1)))
           (if (##fixnum.< i start)
             (##void)
             (begin
               (,##vect-set! vect i x)
               (loop (##fixnum.- i 1))))))

       (define-system (,##vect-shrink! x y)))))

(define-vector-procedures string    #\nul macro-check-char)
(define-vector-procedures vector    0     macro-no-check)
(define-vector-procedures s8vector  0     macro-check-exact-signed-int8)
(define-vector-procedures u8vector  0     macro-check-exact-unsigned-int8)
(define-vector-procedures s16vector 0     macro-check-exact-signed-int16)
(define-vector-procedures u16vector 0     macro-check-exact-unsigned-int16)
(define-vector-procedures s32vector 0     macro-check-exact-signed-int32)
(define-vector-procedures u32vector 0     macro-check-exact-unsigned-int32)
(define-vector-procedures s64vector 0     macro-check-exact-signed-int64)
(define-vector-procedures u64vector 0     macro-check-exact-unsigned-int64)
(define-vector-procedures f32vector 0.    macro-check-inexact-real)
(define-vector-procedures f64vector 0.    macro-check-inexact-real)

(##define-macro (define-subvector-procedures name)

  (define (sym . lst)
    (string->symbol (apply string-append (map symbol->string lst))))

  (let ()
    (define macro-check-vect (sym 'macro-check- name))

    (define ##make-vect      (sym '##make- name))
    (define ##vect-length    (sym '## name '-length))
    (define ##vect-ref       (sym '## name '-ref))
    (define ##vect-set!      (sym '## name '-set!))

    (define ##subvect-move!  (sym '##zsub name '-move!))
    (define ##subvect        (sym '##zsub name))
    (define ##vect-copy      (sym '## name '-copy))
    (define ##vect-fill!     (sym '## name '-fill!))

    (define subvect          (sym 'zsub name))

    `(begin

       (define (,##subvect-move! src src-start src-end dst dst-start)
         ; Copy direction must be selected in case src and dst are the
         ; same object.
         (if (##fixnum.< src-start dst-start)
           (let loop1 ((i (##fixnum.- src-end 1))
                       (j (##fixnum.-
                           (##fixnum.+ dst-start
                                       (##fixnum.- src-end src-start))
                           1)))
             (if (##fixnum.< i src-start)
               dst
               (begin
                 (,##vect-set! dst j (,##vect-ref src i))
                 (loop1 (##fixnum.- i 1)
                        (##fixnum.- j 1)))))
           (let loop2 ((i src-start)
                       (j dst-start))
             (if (##fixnum.< i src-end)
               (begin
                 (,##vect-set! dst j (,##vect-ref src i))
                 (loop2 (##fixnum.+ i 1)
                        (##fixnum.+ j 1)))
               dst))))

       (define-system (,##subvect src start end)
         (,##subvect-move!
          src
          start
          end
          (,##make-vect (##fixnum.max (##fixnum.- end start) 0))
          0))

       (define (,subvect src start end)
         (macro-force-vars (src start end)
           (,macro-check-vect src (,subvect src start end)
             (macro-check-exact-int-range-incl
              start
              0
              (,##vect-length src)
              (,subvect src start end)
              (macro-check-exact-int-range-incl
               end
               start
               (,##vect-length src)
               (,subvect src start end)
               (,##subvect src start end)))))))))

(define-subvector-procedures string)
(define-subvector-procedures vector)
(define-subvector-procedures u8vector)

;------------------------------------------------------------------------------

; IEEE Scheme procedures:

(define (not x)
  (macro-force-vars (x)
    (##not x)))

(define (boolean? x)
  (macro-force-vars (x)
    (or (##eq? x #t) (##eq? x #f))))

(define (eqv? x y)
  (macro-force-vars (x y)
    (let ()
      (##declare (generic)) ; do not use fixnum specific ##eqv?
      (##eqv? x y))))

(define (eq? x y)
  (macro-force-vars (x y)
    (##eq? x y)))

(define (equal? x y)
  (##equal? x y (macro-if-forces #t #f)))

(define (pair? x)
  (macro-force-vars (x)
    (##pair? x)))

(define (cons x y)
  (##cons x y))

(define (car x)
  (macro-force-vars (x)
    (macro-check-pair x (car x)
      (##car x))))

(define (cdr x)
  (macro-force-vars (x)
    (macro-check-pair x (cdr x)
      (##cdr x))))

(define (set-car! x y)
  (macro-force-vars (x)
    (macro-check-pair x (set-car! x y)
      (begin
        (##set-car! x y)
        (##void)))))

(define (set-cdr! x y)
  (macro-force-vars (x)
    (macro-check-pair x (set-cdr! x y)
      (begin
        (##set-cdr! x y)
        (##void)))))

(##define-macro (define-c...r name pattern)

  (define (gen name pattern)
    `(macro-check-pair y (,name x)
       ,(if (<= pattern 3)
          (if (= pattern 3) '(##cdr y) '(##car y))
          `(let ((y ,(if (odd? pattern) '(##cdr y) '(##car y))))
             (macro-force-vars (y)
               ,(gen name (quotient pattern 2)))))))

  `(define (,name x)
     (macro-force-vars (x)
       (let ((y x))
         ,(gen name pattern)))))

(define-c...r caar 4)
(define-c...r cadr 5)
(define-c...r cdar 6)
(define-c...r cddr 7)
(define-c...r caaar 8)
(define-c...r caadr 9)
(define-c...r cadar 10)
(define-c...r caddr 11)
(define-c...r cdaar 12)
(define-c...r cdadr 13)
(define-c...r cddar 14)
(define-c...r cdddr 15)
(define-c...r caaaar 16)
(define-c...r caaadr 17)
(define-c...r caadar 18)
(define-c...r caaddr 19)
(define-c...r cadaar 20)
(define-c...r cadadr 21)
(define-c...r caddar 22)
(define-c...r cadddr 23)
(define-c...r cdaaar 24)
(define-c...r cdaadr 25)
(define-c...r cdadar 26)
(define-c...r cdaddr 27)
(define-c...r cddaar 28)
(define-c...r cddadr 29)
(define-c...r cdddar 30)
(define-c...r cddddr 31)

(define (null? x)
  (macro-force-vars (x)
    (##null? x)))

(define (list? lst)
  (let loop ((lst1 lst) (lst2 lst))
    (macro-force-vars (lst1)
      (if (##not (##pair? lst1))
        (##null? lst1)
        (let ((lst1 (##cdr lst1)))
          (macro-force-vars (lst1 lst2)
            (cond ((##eq? lst1 lst2)
                   #f)
                  ((##pair? lst1)
                   (loop (##cdr lst1) (##cdr lst2)))
                  (else
                   (##null? lst1)))))))))

(define (list . lst)
  lst)

(define (length lst)
  (let loop ((lst1 lst) (n 0))
    (macro-force-vars (lst1)
      (if (##pair? lst1)
        (loop (##cdr lst1) (##fixnum.+ n 1))
        (macro-check-list-end lst1 (length lst)
          n)))))

(define (append #!optional (x (macro-absent-obj)) (y (macro-absent-obj)) #!rest z)

  (define (append-multiple head tail)
    (if (##null? tail)
      head
      (macro-force-vars (head)
        (if (##null? head)
          (append-multiple (##car tail) (##cdr tail))
          (list-expected-check (append-multiple-non-null head tail))))))

  (define (append-multiple-non-null lst1 lsts) ; lst1!=(), returns #f on error
    (let ((head (##car lsts))
          (tail (##cdr lsts)))
      (if (##null? tail)
        (append-2-non-null lst1 head)
        (macro-force-vars (head)
          (if (##null? head)
            (append-multiple-non-null lst1 tail)
            (let ((result (append-multiple-non-null head tail)))
              (and result
                   (append-2-non-null lst1 result))))))))

  (define (append-2-non-null lst1 lst2) ; lst1!=(), returns #f on error
    (if (##pair? lst1)
      (let ((head (##car lst1))
            (tail (##cdr lst1)))
        (macro-force-vars (tail)
          (if (##null? tail)
            (##cons head lst2)
            (let ((result (append-2-non-null tail lst2)))
              (and result
                   (##cons head result))))))
      (macro-if-checks
        #f ; error: list expected
        lst2)))

  (define (list-expected-check result)
    (macro-if-checks
      (if (##not result)
        (macro-trap-list (append x y . z))
        result)
      result))

  (cond ((##eq? x (macro-absent-obj))
         '())
        ((##eq? y (macro-absent-obj))
         x)
        ((##null? z)
         (macro-force-vars (x)
           (if (##not (##null? x))
             (list-expected-check (append-2-non-null x y))
             y)))
        (else
         (append-multiple x (##cons y z)))))

(define (reverse lst)
  (let loop ((lst1 lst) (x '()))
    (macro-force-vars (lst1)
      (if (##pair? lst1)
        (loop (##cdr lst1) (##cons (##car lst1) x))
        (macro-check-list-end lst1 (reverse lst)
          x)))))

(define (list-ref lst k)
  (macro-force-vars (k)
    (macro-check-exact-int-non-neg k (list-ref lst k)
      (let loop ((lst1 lst) (i k))
        (macro-force-vars (lst1)
          (macro-check-pair lst1 (list-ref lst k)
            (if (##fixnum.< 0 i)
              (loop (##cdr lst1) (##fixnum.- i 1))
              (##car lst1))))))))

(define (memq x lst)
  (macro-force-vars (x)
    (let loop ((lst1 lst))
      (macro-force-vars (lst1)
        (if (##pair? lst1)
          (let ((y (##car lst1)))
            (macro-force-vars (y)
              (if (##eq? x y)
                lst1
                (loop (##cdr lst1)))))
          (macro-check-list-end lst1 (memq x lst)
            #f))))))

(define (memv x lst)
  (macro-force-vars (x)
    (let loop ((lst1 lst))
      (macro-force-vars (lst1)
        (if (##pair? lst1)
          (let ((y (##car lst1)))
            (macro-force-vars (y)
              (if (let () (##declare (generic)) (##eqv? x y))
                lst1
                (loop (##cdr lst1)))))
          (macro-check-list-end lst1 (memv x lst)
            #f))))))

(define (member x lst)
  (let loop ((lst1 lst))
    (macro-force-vars (lst1)
      (if (##pair? lst1)
        (let ((y (##car lst1)))
          (if (##equal? x y (macro-if-forces #t #f))
            lst1
            (loop (##cdr lst1))))
        (macro-check-list-end lst1 (member x lst)
          #f)))))

(define (assq x lst)
  (macro-force-vars (x)
    (let loop ((lst1 lst))
      (macro-force-vars (lst1)
        (if (##pair? lst1)
          (let ((couple (##car lst1)))
            (macro-force-vars (couple)
              (macro-check-pair couple (assq x lst)
                (let ((y (##car couple)))
                  (macro-force-vars (y)
                    (if (##eq? x y)
                      couple
                      (loop (##cdr lst1))))))))
          (macro-check-list-end lst1 (assq x lst)
            #f))))))

(define (assv x lst)
  (macro-force-vars (x)
    (let loop ((lst1 lst))
      (macro-force-vars (lst1)
        (if (##pair? lst1)
          (let ((couple (##car lst1)))
            (macro-force-vars (couple)
              (macro-check-pair couple (assv x lst)
                (let ((y (##car couple)))
                  (macro-force-vars (y)
                    (if (let () (##declare (generic)) (##eqv? x y))
                      couple
                      (loop (##cdr lst1))))))))
          (macro-check-list-end lst1 (assv x lst)
            #f))))))

(define (assoc x lst)
  (let loop ((lst1 lst))
    (macro-force-vars (lst1)
      (if (##pair? lst1)
        (let ((couple (##car lst1)))
          (macro-force-vars (couple)
            (macro-check-pair couple (assoc x lst)
              (let ((y (##car couple)))
                (if (##equal? x y (macro-if-forces #t #f))
                  couple
                  (loop (##cdr lst1)))))))
        (macro-check-list-end lst1 (assoc x lst)
          #f)))))

(define (symbol? x)
  (macro-force-vars (x)
    (##symbol? x)))

(define (symbol->string sym)
  (macro-force-vars (sym)
    (macro-check-symbol sym (symbol->string sym)
      (##symbol->string sym))))

(define (string->symbol str)
  (macro-force-vars (str)
    (macro-check-string str (string->symbol str)
      (##string->symbol str))))

; Number related procedures are in "_num.scm"

(##define-macro (char-up-to-down) 32)

(define-system (##char? x))
(define (char? x) (macro-force-vars (x) (##char? x)))

(define-nary-bool (##char=? x y)
  #t
  #t
  (##char=? x y)
  macro-no-force
  macro-no-check)

(define-nary-bool (char=? x y)
  #t
  #t
  (##char=? x y)
  macro-force-vars
  macro-check-char)

(define-nary-bool (##char<? x y)
  #t
  #t
  (##char<? x y)
  macro-no-force
  macro-no-check)

(define-nary-bool (char<? x y)
  #t
  #t
  (##char<? x y)
  macro-force-vars
  macro-check-char)

(define-nary-bool (##char>? x y)
  #t
  #t
  (##char<? y x)
  macro-no-force
  macro-no-check)

(define-nary-bool (char>? x y)
  #t
  #t
  (##char<? y x)
  macro-force-vars
  macro-check-char)

(define-nary-bool (##char<=? x y)
  #t
  #t
  (##not (##char<? y x))
  macro-no-force
  macro-no-check)

(define-nary-bool (char<=? x y)
  #t
  #t
  (##not (##char<? y x))
  macro-force-vars
  macro-check-char)

(define-nary-bool (##char>=? x y)
  #t
  #t
  (##not (##char<? x y))
  macro-no-force
  macro-no-check)

(define-nary-bool (char>=? x y)
  #t
  #t
  (##not (##char<? x y))
  macro-force-vars
  macro-check-char)

(##define-macro (case-independent-char=? x y)
  `(##char=? (##char-downcase ,x) (##char-downcase ,y)))

(##define-macro (case-independent-char<? x y)
  `(##char<? (##char-downcase ,x) (##char-downcase ,y)))

(define-nary-bool (##char-ci=? x y)
  #t
  #t
  (case-independent-char=? x y)
  macro-no-force
  macro-no-check)

(define-nary-bool (char-ci=? x y)
  #t
  #t
  (case-independent-char=? x y)
  macro-force-vars
  macro-check-char)

(define-nary-bool (##char-ci<? x y)
  #t
  #t
  (case-independent-char<? x y)
  macro-no-force
  macro-no-check)

(define-nary-bool (char-ci<? x y)
  #t
  #t
  (case-independent-char<? x y)
  macro-force-vars
  macro-check-char)

(define-nary-bool (##char-ci>? x y)
  #t
  #t
  (case-independent-char<? y x)
  macro-no-force
  macro-no-check)

(define-nary-bool (char-ci>? x y)
  #t
  #t
  (case-independent-char<? y x)
  macro-force-vars
  macro-check-char)

(define-nary-bool (##char-ci<=? x y)
  #t
  #t
  (##not (case-independent-char<? y x))
  macro-no-force
  macro-no-check)

(define-nary-bool (char-ci<=? x y)
  #t
  #t
  (##not (case-independent-char<? y x))
  macro-force-vars
  macro-check-char)

(define-nary-bool (##char-ci>=? x y)
  #t
  #t
  (##not (case-independent-char<? x y))
  macro-no-force
  macro-no-check)

(define-nary-bool (char-ci>=? x y)
  #t
  #t
  (##not (case-independent-char<? x y))
  macro-force-vars
  macro-check-char)

(define-system (##char-alphabetic? c)
  (let ((x (##char-downcase c)))
    (and (##not (##char<? x #\a)) (##not (##char<? #\z x)))))

(define (char-alphabetic? c)
  (macro-force-vars (c)
    (macro-check-char c (char-alphabetic? c)
      (##char-alphabetic? c))))

(define-system (##char-numeric? c)
  (and (##not (##char<? c #\0)) (##not (##char<? #\9 c))))

(define (char-numeric? c)
  (macro-force-vars (c)
    (macro-check-char c (char-numeric? c)
      (##char-numeric? c))))

(define-system (##char-whitespace? c)
  (##char<=? c #\space))

(define (char-whitespace? c)
  (macro-force-vars (c)
    (macro-check-char c (char-whitespace? c)
      (##char-whitespace? c))))

(define-system (##char-upper-case? c)
  (and (##not (##char<? c #\A)) (##not (##char<? #\Z c))))

(define (char-upper-case? c)
  (macro-force-vars (c)
    (macro-check-char c (char-upper-case? c)
      (##char-upper-case? c))))

(define-system (##char-lower-case? c)
  (and (##not (##char<? c #\a)) (##not (##char<? #\z c))))

(define (char-lower-case? c)
  (macro-force-vars (c)
    (macro-check-char c (char-lower-case? c)
      (##char-lower-case? c))))

(define (char->integer c)
  (macro-force-vars (c)
    (macro-check-char c (char->integer c)
      (##fixnum.<-char c))))

(define (integer->char n)
  (macro-force-vars (n)
    (macro-check-exact-int-range-incl n 0 ##max-char (integer->char n)
      (##fixnum.->char n))))

(define-system (##char-upcase c)
  (if (and (##not (##char<? c #\a)) (##not (##char<? #\z c)))
    (##fixnum.->char (##fixnum.- (##fixnum.<-char c) (char-up-to-down)))
    c))

(define (char-upcase c)
  (macro-force-vars (c)
    (macro-check-char c (char-upcase c)
      (##char-upcase c))))

(define-system (##char-downcase c)
  (if (and (##not (##char<? c #\A)) (##not (##char<? #\Z c)))
    (##fixnum.->char (##fixnum.+ (##fixnum.<-char c) (char-up-to-down)))
    c))

(define (char-downcase c)
  (macro-force-vars (c)
    (macro-check-char c (char-downcase c)
      (##char-downcase c))))

(define-system (##string=? x y)
  (let ((len (##string-length x)))
    (if (##eq? len (##string-length y))
      (let loop ((i (##fixnum.- len 1)))
        (cond ((##fixnum.< i 0)
               #t)
              ((##char=? (##string-ref x i) (##string-ref y i))
               (loop (##fixnum.- i 1)))
              (else
               #f)))
      #f)))

(define-nary-bool (string=? x y)
  #t
  #t
  (##string=? x y)
  macro-force-vars
  macro-check-string)

(define-system (##string<? x y)
  (let ((lx (##string-length x))
        (ly (##string-length y)))
    (let ((n (if (##fixnum.< lx ly) lx ly)))
      (let loop ((i 0))
        (if (##fixnum.< i n)
          (let ((cx (##string-ref x i))
                (cy (##string-ref y i)))
            (if (##char=? cx cy)
              (loop (##fixnum.+ i 1))
              (##char<? cx cy)))
          (##fixnum.< n ly))))))

(define-nary-bool (string<? x y)
  #t
  #t
  (##string<? x y)
  macro-force-vars
  macro-check-string)

(define-nary-bool (string>? x y)
  #t
  #t
  (##string<? y x)
  macro-force-vars
  macro-check-string)

(define-nary-bool (string<=? x y)
  #t
  #t
  (##not (##string<? y x))
  macro-force-vars
  macro-check-string)

(define-nary-bool (string>=? x y)
  #t
  #t
  (##not (##string<? x y))
  macro-force-vars
  macro-check-string)

(define-system (##string-ci=? x y)
  (let ((len (##string-length x)))
    (if (##eq? len (##string-length y))
      (let loop ((i (##fixnum.- len 1)))
        (cond ((##fixnum.< i 0)
               #t)
              ((##char=? (##char-downcase (##string-ref x i))
                         (##char-downcase (##string-ref y i)))
               (loop (##fixnum.- i 1)))
              (else
               #f)))
      #f)))

(define-nary-bool (string-ci=? x y)
  #t
  #t
  (##string-ci=? x y)
  macro-force-vars
  macro-check-string)

(define-system (##string-ci<? x y)
  (let ((lx (##string-length x))
        (ly (##string-length y)))
    (let ((n (if (##fixnum.< lx ly) lx ly)))
      (let loop ((i 0))
        (if (##fixnum.< i n)
          (let ((cx (##char-downcase (##string-ref x i)))
                (cy (##char-downcase (##string-ref y i))))
            (if (##char=? cx cy)
              (loop (##fixnum.+ i 1))
              (##char<? cx cy)))
          (##fixnum.< n ly))))))

(define-nary-bool (string-ci<? x y)
  #t
  #t
  (##string-ci<? x y)
  macro-force-vars
  macro-check-string)

(define-nary-bool (string-ci>? x y)
  #t
  #t
  (##string-ci<? y x)
  macro-force-vars
  macro-check-string)

(define-nary-bool (string-ci<=? x y)
  #t
  #t
  (##not (##string-ci<? y x))
  macro-force-vars
  macro-check-string)

(define-nary-bool (string-ci>=? x y)
  #t
  #t
  (##not (##string-ci<? x y))
  macro-force-vars
  macro-check-string)

'(;;;;;;;;;;;

(define (##substring-move! src src-start src-end dst dst-start)
  ; Copy direction must be selected in case src and dst are the same object.
  (if (##fixnum.< src-start dst-start)
    (let loop1 ((i (##fixnum.- src-end 1))
                (j (##fixnum.- (##fixnum.+ dst-start
                                           (##fixnum.- src-end src-start))
                               1)))
      (if (##fixnum.< i src-start)
        dst
        (begin
          (##string-set! dst j (##string-ref src i))
          (loop1 (##fixnum.- i 1)
                 (##fixnum.- j 1)))))
    (let loop2 ((i src-start)
                (j dst-start))
      (if (##fixnum.< i src-end)
        (begin
          (##string-set! dst j (##string-ref src i))
          (loop2 (##fixnum.+ i 1)
                 (##fixnum.+ j 1)))
        dst))))

(define-system (##substring src start end)
  (##substring-move! src
                     start
                     end
                     (##make-string (##fixnum.max (##fixnum.- end start) 0))
                     0))

(define (substring src start end)
  (macro-force-vars (src start end)
    (macro-check-string src (substring src start end)
      (macro-check-exact-int-range-incl
       start
       0
       (##string-length src)
       (substring src start end)
       (macro-check-exact-int-range-incl
        end
        start
        (##string-length src)
        (substring src start end)
        (##substring src start end))))))
)

(define-system (##string-append . lst)
  (##append-strings lst))

(define (##append-strings lst)
  (let loop1 ((n 0)
              (lst1 lst)
              (lst2 '()))
    (if (##pair? lst1)
      (let ((s (##car lst1)))
        (loop1 (##fixnum.+ n (##string-length s))
               (##cdr lst1)
               (##cons s lst2)))
      (let ((result (##make-string n)))
        (let loop2 ((k (##fixnum.- n 1))
                    (lst2 lst2))
          (if (##pair? lst2)
            (let ((s (##car lst2)))
              (let loop3 ((i k)
                          (j (##fixnum.- (##string-length s) 1)))
                (if (##not (##fixnum.< j 0))
                  (begin
                    (##string-set! result i (##string-ref s j))
                    (loop3 (##fixnum.- i 1)
                           (##fixnum.- j 1)))
                  (loop2 i
                         (##cdr lst2)))))
            result))))))

(define (string-append . lst)
  (let loop1 ((n 0)
              (lst1 lst)
              (lst2 '()))
    (if (##pair? lst1)
      (let ((s (##car lst1)))
        (macro-force-vars (s)
          (macro-check-string s (string-append . lst)
            (loop1 (##fixnum.+ n (##string-length s))
                   (##cdr lst1)
                   (##cons s lst2)))))
      (let ((result (##make-string n)))
        (let loop2 ((k (##fixnum.- n 1))
                    (lst2 lst2))
          (if (##pair? lst2)
            (let ((s (##car lst2)))
              (let loop3 ((i k)
                          (j (##fixnum.- (##string-length s) 1)))
                (if (##not (##fixnum.< j 0))
                  (begin
                    (##string-set! result i (##string-ref s j))
                    (loop3 (##fixnum.- i 1)
                           (##fixnum.- j 1)))
                  (loop2 i
                         (##cdr lst2)))))
            result))))))

(define (procedure? x)
  (macro-force-vars (x)
    (##procedure? x)))

(define (apply p x . y)

  (define (build-arg-list arg1 other-args)
    (if (##pair? other-args)
      (let ((tail (build-arg-list (##car other-args) (##cdr other-args))))
        (macro-if-checks
          (and tail
               (##cons arg1 tail))
          (##cons arg1 tail)))
      (macro-if-forces
        (copy-proper-list arg1)
        (check-proper-list arg1))))

  (define (copy-proper-list lst)
    (macro-force-vars (lst)
      (if (##pair? lst)
        (let ((tail (copy-proper-list (##cdr lst))))
          (macro-if-checks
            (and tail
                 (##cons (##car lst) tail))
            (##cons (##car lst) tail)))
        (macro-if-checks
          (if (##null? lst)
            '()
            #f) ; error: list expected
          '()))))

  (define (check-proper-list lst)
    (macro-if-checks
      (let loop ((lst2 lst))
        (cond ((##pair? lst2)
               (loop (##cdr lst2)))
              ((##null? lst2)
               lst)
              (else
               #f))) ; error: list expected
      lst))

  (macro-force-vars (p)
    (macro-check-procedure p (apply p x . y)
      (let ((lst (build-arg-list x y)))
        (if lst
          (##apply p lst)
          (macro-trap-list (apply p x . y)))))))

(define (map p x . y)
  (macro-force-vars (p)
    (macro-check-procedure p (map p x . y)
      (let ()

        (define (proper-list-length lst)
          (let loop ((lst lst) (n 0))
            (macro-force-vars (lst)
              (cond ((##pair? lst)
                     (loop (##cdr lst) (##fixnum.+ n 1)))
                    ((##null? lst)
                     n)
                    (else
                     #f)))))

        (define (map-1 lst1)
          (macro-force-vars (lst1)
            (if (##pair? lst1)
              (let ((result (p (##car lst1))))
                (##cons result (map-1 (##cdr lst1))))
              '())))

        (define (cars lsts)
          (if (##pair? lsts)
            (let ((lst1 (##car lsts)))
              (macro-force-vars (lst1)
                (let ((head (##car lst1)))
                  (let ((tail (cars (##cdr lsts))))
                    (##cons head tail)))))
            '()))

        (define (cdrs lsts)
          (if (##pair? lsts)
            (let ((lst1 (##car lsts)))
              (macro-force-vars (lst1)
                (let ((head (##cdr lst1)))
                  (if (##pair? head)
                    (let ((tail (cdrs (##cdr lsts))))
                      (and tail
                           (##cons head tail)))
                    #f))))
            '()))

        (define (map-n lsts)
          (if lsts
            (let ((result (##apply p (cars lsts))))
              (##cons result (map-n (cdrs lsts))))
            '()))

        (cond ((##null? y)
               (macro-if-checks
                 (let ((len1 (proper-list-length x)))
                   (if len1
                     (map-1 x)
                     (macro-trap-list (map p x . y))))
                 (map-1 x)))
              (else
               (macro-if-checks
                 (let ((len1 (proper-list-length x)))
                   (if len1
                     (let loop ((lsts y))
                       (if (##null? lsts)
                         (if (##null? x)
                           '()
                           (map-n (##cons x y)))
                         (let ((len2 (proper-list-length (##car lsts))))
                           (if (##eq? len1 len2)
                             (loop (##cdr lsts))
                             (if len2
                               (macro-trap-list-lengths (map p x . y))
                               (macro-trap-list (map p x . y)))))))
                     (macro-trap-list (map p x . y))))
                 (if (##null? x)
                   '()
                   (map-n (##cons x y))))))))))

(define (for-each p x . y)
  (macro-force-vars (p)
    (macro-check-procedure p (for-each p x . y)
      (let ()

        (define (proper-list-length lst)
          (let loop ((lst lst) (n 0))
            (macro-force-vars (lst)
              (cond ((##pair? lst)
                     (loop (##cdr lst) (##fixnum.+ n 1)))
                    ((##null? lst)
                     n)
                    (else
                     #f)))))

        (define (for-each-1 lst1)
          (macro-force-vars (lst1)
            (if (##pair? lst1)
              (let ((result (p (##car lst1))))
                (for-each-1 (##cdr lst1)))
              (##void))))

        (define (cars lsts)
          (if (##pair? lsts)
            (let ((lst1 (##car lsts)))
              (macro-force-vars (lst1)
                (let ((head (##car lst1)))
                  (let ((tail (cars (##cdr lsts))))
                    (##cons head tail)))))
            '()))

        (define (cdrs lsts)
          (if (##pair? lsts)
            (let ((lst1 (##car lsts)))
              (macro-force-vars (lst1)
                (let ((head (##cdr lst1)))
                  (if (##pair? head)
                    (let ((tail (cdrs (##cdr lsts))))
                      (and tail
                           (##cons head tail)))
                    #f))))
            '()))

        (define (for-each-n lsts)
          (if lsts
            (let ((result (##apply p (cars lsts))))
              (for-each-n (cdrs lsts)))
            (##void)))

        (cond ((##null? y)
               (macro-if-checks
                 (let ((len1 (proper-list-length x)))
                   (if len1
                     (for-each-1 x)
                     (macro-trap-list (for-each p x . y))))
                 (for-each-1 x)))
              (else
               (macro-if-checks
                 (let ((len1 (proper-list-length x)))
                   (if len1
                     (let loop ((lsts y))
                       (if (##null? lsts)
                         (if (##null? x)
                           (##void)
                           (for-each-n (##cons x y)))
                         (let ((len2 (proper-list-length (##car lsts))))
                           (if (##eq? len1 len2)
                             (loop (##cdr lsts))
                             (if len2
                               (macro-trap-list-lengths (for-each p x . y))
                               (macro-trap-list (for-each p x . y)))))))
                     (macro-trap-list (for-each p x . y))))
                 (if (##null? x)
                   (##void)
                   (for-each-n (##cons x y))))))))))

(define (call-with-current-continuation p)
  (macro-force-vars (p)
    (macro-check-procedure p (call-with-current-continuation p)
      (##call-with-current-continuation p))))

; Port related procedures are in "_io.scm"

;------------------------------------------------------------------------------

; R4RS Scheme procedures:

(define (list-tail l k)
  (macro-force-vars (k)
    (macro-check-exact-int-non-neg k (list-tail l k)
      (let loop ((x l) (i k))
        (if (##fixnum.< 0 i)
          (macro-force-vars (x)
            (macro-check-pair x (list-tail l k)
              (loop (##cdr x) (##fixnum.- i 1))))
          x)))))

'(;;;;;;;;;;;;;;;;
(define-system (##string-copy str)
  (let* ((n (##string-length str))
         (result (##make-string n)))
    (let loop ((i (##fixnum.- n 1)))
      (if (##fixnum.< i 0)
        result
        (begin
          (##string-set! result i (##string-ref str i))
          (loop (##fixnum.- i 1)))))))

(define (string-copy str)
  (macro-force-vars (str)
    (macro-check-string str (string-copy str)
      (let* ((n (##string-length str))
             (result (##make-string n)))
        (let loop ((i (##fixnum.- n 1)))
          (if (##fixnum.< i 0)
            result
            (begin
              (##string-set! result i (##string-ref str i))
              (loop (##fixnum.- i 1)))))))))

(define (string-fill! str c)
  (macro-force-vars (str c)
    (macro-check-string str (string-fill! str c)
      (macro-check-char c (string-fill! str c)
        (let ((n (##string-length str)))
          (let loop ((i (##fixnum.- n 1)))
            (if (##fixnum.< i 0)
              (##void)
              (begin
                (##string-set! str i c)
                (loop (##fixnum.- i 1))))))))))

(define (vector-fill! vect x)
  (macro-force-vars (vect x)
    (macro-check-vector vect (vector-fill! vect x)
      (let ((n (##vector-length vect)))
        (let loop ((i (##fixnum.- n 1)))
          (if (##fixnum.< i 0)
            (##void)
            (begin
              (##vector-set! vect i x)
              (loop (##fixnum.- i 1)))))))))

(define-system (##vector-copy vect)
  (let* ((n (##vector-length vect))
         (result (##make-vector n 0)))
    (let loop ((i (##fixnum.- n 1)))
      (if (##fixnum.< i 0)
        result
        (begin
          (##vector-set! result i (##vector-ref vect i))
          (loop (##fixnum.- i 1)))))))
)

(define (force x)
  (##force x))

; Port related procedures are in "_io.scm"

;------------------------------------------------------------------------------

; R5RS Scheme procedures:

;(define (values . lst)
;  lst)
;
;(define (call-with-values producer consumer)
;  (apply consumer (producer)))
;
;(dynamic-wind before thunk after)
;(eval expr env)
;(scheme-report-environment version)
;(null-environment version)
;(interaction-environment)

;------------------------------------------------------------------------------

; Multilisp procedures:

(define (touch x)
  (##force x))

;------------------------------------------------------------------------------

; DSSSL procedures:

(define (keyword? x)
  (##keyword? x))

(define (keyword->string key)
  (macro-force-vars (key)
    (macro-check-keyword key (keyword->string key)
      (##keyword->string key))))

(define (string->keyword str)
  (macro-force-vars (str)
    (macro-check-string str (string->keyword str)
      (##string->keyword str))))

;==============================================================================
