;==============================================================================

; file: "_system.scm"

; Copyright (C) 1994-2000 by Marc Feeley, All Rights Reserved.

;==============================================================================

(##include "header.scm")

;==============================================================================

; System procedures

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

; Type operations.

(define-system (##type obj))
(define-system (##type-cast obj type))
(define-system (##subtype obj))
(define-system (##subtype-set! obj subtype))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

; Basic type predicates.

(define-system (##fixnum? obj)
  (##eq? (##type obj) (macro-type-fixnum)))

(define-system (##subtyped? obj)
  (##eq? (##type obj) (macro-type-subtyped)))

(define-system (##subtyped.vector? obj)
  (##eq? (##subtype obj) (macro-subtype-vector)))

(define-system (##subtyped.symbol? obj)
  (##eq? (##subtype obj) (macro-subtype-symbol)))

(define-system (##subtyped.flonum? obj)
  (##eq? (##subtype obj) (macro-subtype-flonum)))

(define-system (##subtyped.bignum? obj)
  (##eq? (##subtype obj) (macro-subtype-bignum)))

(define-system (##special? obj)
  (##eq? (##type obj) (macro-type-special)))

(define-system (##pair? obj))

; (##vector? obj) is defined in "_std.scm"

(define-system (##ratnum? obj)
  (and (##subtyped? obj)
       (##eq? (##subtype obj) (macro-subtype-ratnum))))

(define-system (##cpxnum? obj)
  (and (##subtyped? obj)
       (##eq? (##subtype obj) (macro-subtype-cpxnum))))

(define-system (##structure? obj)
  (and (##subtyped? obj)
       (##eq? (##subtype obj) (macro-subtype-structure))))

(define-system (##meroon? obj)
  (and (##subtyped? obj)
       (##eq? (##subtype obj) (macro-subtype-meroon))))

(define-system (##symbol? obj)
  (and (##subtyped? obj)
       (##eq? (##subtype obj) (macro-subtype-symbol))))

(define-system (##keyword? obj)
  (and (##subtyped? obj)
       (##eq? (##subtype obj) (macro-subtype-keyword))))

(define-system (##frame? obj)
  (and (##subtyped? obj)
       (##eq? (##subtype obj) (macro-subtype-frame))))

(define-system (##continuation? obj)
  (and (##subtyped? obj)
       (##eq? (##subtype obj) (macro-subtype-continuation))))

(define-system (##promise? obj)
  (and (##subtyped? obj)
       (##eq? (##subtype obj) (macro-subtype-promise))))

(define-system (##will? obj)
  (and (##subtyped? obj)
       (##eq? (##subtype obj) (macro-subtype-will))))

(define-system (##procedure? obj)
  (and (##subtyped? obj)
       (##eq? (##subtype obj) (macro-subtype-procedure))))

(define-system (##return? obj)
  (and (##subtyped? obj)
       (##eq? (##subtype obj) (macro-subtype-return))))

(define-system (##pointer? obj)
  (and (##subtyped? obj)
       (##eq? (##subtype obj) (macro-subtype-pointer))))

; (##string? obj) is defined in "_std.scm"
; (##s8vector? obj) is defined in "_std.scm"
; (##u8vector? obj) is defined in "_std.scm"
; (##s16vector? obj) is defined in "_std.scm"
; (##u16vector? obj) is defined in "_std.scm"
; (##s32vector? obj) is defined in "_std.scm"
; (##u32vector? obj) is defined in "_std.scm"
; (##s64vector? obj) is defined in "_std.scm"
; (##u64vector? obj) is defined in "_std.scm"
; (##f32vector? obj) is defined in "_std.scm"
; (##f64vector? obj) is defined in "_std.scm"

(define-system (##flonum? obj)
  (and (##subtyped? obj)
       (##eq? (##subtype obj) (macro-subtype-flonum))))

(define-system (##bignum? obj)
  (and (##subtyped? obj)
       (##eq? (##subtype obj) (macro-subtype-bignum))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define-system (##unbound? obj))

(define-system (##closure? obj))
(define-system (##closure-code closure))
(define-system (##closure-ref closure i))
(define-system (##closure-set! closure i val))

(define-system (##subprocedure? proc))
(define-system (##subprocedure-id subproc))
(define-system (##subprocedure-parent subproc))

(define-system (##procedure-info proc))

(define-system (##make-cell obj)
  (##cons obj '()))

(define-system (##cell-ref cell)
  (##car cell))

(define-system (##cell-set! cell val)
  (##set-car! cell val))

(define-system (##make-promise thunk))
(define-system (##force obj))

(define-system (##void))

(define (##first-argument arg1 #!optional arg2 arg3 #!rest lst)
  arg1)

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

; Variants of standard procedures.

; Most of these procedures do not force their arguments and are mostly
; of fixed arity.

(define-system (##not x)
  (if x #f #t))

; ##eqv? is defined in "_num.scm"

(define-system (##eq? x y))

(define-system (##equal? x y force?)

  (define (equal x y)

    (define (ovector=? x y)
      (let ((len-x (##vector-length x)))
        (if (##eq? len-x (##vector-length y))
          (let loop ((i (##fixnum.- len-x 1)))
            (cond ((##fixnum.< i 0)
                  #t)
                  ((equal (##vector-ref x i) (##vector-ref y i))
                   (loop (##fixnum.- i 1)))
                  (else
                   #f)))
          #f)))

    (define (structure=? x y)
      (let ((tag-x (##vector-ref x 0)))
        (and (##eq? tag-x (##vector-ref y 0))
             (let ((n (##vector-ref tag-x 1)))
               (if (and (##fixnum? n)
                        (##fixnum.< n 0))
                 #f
                 (let ((len-x (##vector-length x)))
                   (if (##eq? len-x (##vector-length y))
                     (let loop ((i (##fixnum.- len-x 1)))
                       (cond ((##fixnum.< i 1)
                             #t)
                             ((equal (##vector-ref x i) (##vector-ref y i))
                              (loop (##fixnum.- i 1)))
                             (else
                              #f)))
                     #f)))))))

    (cond ((##eq? x y)
           #t)
          ((##pair? x)
           (and (##pair? y)
                (equal (##car x) (##car y))
                (equal (##cdr x) (##cdr y))))
          ((##subtyped? x)
           (and (##subtyped? y)
                (let ((subtype-x (##subtype x)))
                  (and (##eq? subtype-x (##subtype y))
                       (cond ((macro-subtype-ovector? subtype-x)
                              (if (##eq? subtype-x (macro-subtype-structure))
                                (structure=? x y)
                                (ovector=? x y)))
                             ((macro-subtype-bvector? subtype-x)
                              (##bvector=? x y))
                             (else
                              #f))))))
          (else
           #f)))

  (define (equal* x y)

    (define (ovector=? x y)
      (let ((len-x (##vector-length x)))
        (if (##eq? len-x (##vector-length y))
          (let loop ((i (##fixnum.- len-x 1)))
            (cond ((##fixnum.< i 0)
                  #t)
                  ((equal* (##vector-ref x i) (##vector-ref y i))
                   (loop (##fixnum.- i 1)))
                  (else
                   #f)))
          #f)))

    (define (structure=? x y)
      (let ((tag-x (##vector-ref x 0)))
        (and (##eq? tag-x (##vector-ref y 0))
             (let ((n (##vector-ref tag-x 1)))
               (if (and (##fixnum? n)
                        (##fixnum.< n 0))
                 #f
                 (let ((len-x (##vector-length x)))
                   (if (##eq? len-x (##vector-length y))
                     (let loop ((i (##fixnum.- len-x 1)))
                       (cond ((##fixnum.< i 1)
                             #t)
                             ((equal* (##vector-ref x i) (##vector-ref y i))
                              (loop (##fixnum.- i 1)))
                             (else
                              #f)))
                     #f)))))))

    (let ((x (##force x))
          (y (##force y)))
      (cond ((##eq? x y)
             #t)
            ((##pair? x)
             (and (##pair? y)
                  (equal* (##car x) (##car y))
                  (equal* (##cdr x) (##cdr y))))
            ((##subtyped? x)
             (and (##subtyped? y)
                  (let ((subtype-x (##subtype x)))
                    (and (##eq? subtype-x (##subtype y))
                         (cond ((macro-subtype-ovector? subtype-x)
                                (if (##eq? subtype-x (macro-subtype-structure))
                                  (structure=? x y)
                                  (ovector=? x y)))
                               ((macro-subtype-bvector? subtype-x)
                                (##bvector=? x y))
                               (else
                                #f))))))
            (else
             #f))))

  (if force?
    (equal* x y)
    (equal x y)))

(define (##bvector=? x y)

  (define (u16vect=? x y len)
    (let loop ((i (##fixnum.- len 1)))
      (cond ((##fixnum.< i 0)
             #t)
            ((##eq? (##u16vector-ref x i) (##u16vector-ref y i))
             (loop (##fixnum.- i 1)))
            (else
             #f))))

  (let ((len-x (##u8vector-length x)))
    (and (##eq? len-x (##u8vector-length y))
         (if (##fixnum.odd? len-x)
           (let ((i (##fixnum.- len-x 1)))
             (and (##eq? (##u8vector-ref x i) (##u8vector-ref y i))
                  (u16vect=? x y (##fixnum.quotient i 2))))
           (u16vect=? x y (##fixnum.quotient len-x 2))))))

(define-system (##cons x y))

(define-system (##set-car! x y))

(define-system (##set-cdr! x y))

(define-system (##car x))

(define-system (##cdr x))

(##define-macro (define-c...r name pattern)

  (define (gen name pattern)
    (if (<= pattern 3)
       (if (= pattern 3) '(##cdr x) '(##car x))
       (let ((x (gen name (quotient pattern 2))))
         (if (odd? pattern) '(##cdr ,x) '(##car ,x)))))

  `(define-system (,name x)
     ,(gen name pattern)))

(define-c...r ##caar 4)
(define-c...r ##cadr 5)
(define-c...r ##cdar 6)
(define-c...r ##cddr 7)
(define-c...r ##caaar 8)
(define-c...r ##caadr 9)
(define-c...r ##cadar 10)
(define-c...r ##caddr 11)
(define-c...r ##cdaar 12)
(define-c...r ##cdadr 13)
(define-c...r ##cddar 14)
(define-c...r ##cdddr 15)
(define-c...r ##caaaar 16)
(define-c...r ##caaadr 17)
(define-c...r ##caadar 18)
(define-c...r ##caaddr 19)
(define-c...r ##cadaar 20)
(define-c...r ##cadadr 21)
(define-c...r ##caddar 22)
(define-c...r ##cadddr 23)
(define-c...r ##cdaaar 24)
(define-c...r ##cdaadr 25)
(define-c...r ##cdadar 26)
(define-c...r ##cdaddr 27)
(define-c...r ##cddaar 28)
(define-c...r ##cddadr 29)
(define-c...r ##cdddar 30)
(define-c...r ##cddddr 31)

(define-system (##make-will x y))
(define-system (##will-testator x))

(define-system (##null? x)
  (##eq? x '()))

(define-system (##list . l)
  l)

(define-system (##length l)
  (let loop ((l l) (n 0))
    (if (##pair? l)
      (loop (##cdr l) (##fixnum.+ n 1))
      n)))

(define-system (##append l1 l2)
  (if (##pair? l1)
    (let ((result (##cons (##car l1) '())))
      (##set-cdr!
        (let loop ((end result) (l1 (##cdr l1)))
          (if (##pair? l1)
            (let ((tail (##cons (##car l1) '())))
              (##set-cdr! end tail)
              (loop tail (##cdr l1)))
            end))
        l2)
      result)
    l2))

(define-system (##reverse l)
  (let loop ((l l) (x '()))
    (if (##pair? l)
      (loop (##cdr l) (##cons (##car l) x))
      x)))

(define-system (##memq x l)
  (let loop ((l l))
    (if (##pair? l)
      (if (##eq? x (##car l))
        l
        (loop (##cdr l)))
      #f)))

(define-system (##member x l)
  (let loop ((l l))
    (if (##pair? l)
      (if (##equal? x (##car l) #f)
        l
        (loop (##cdr l)))
      #f)))

(define-system (##assq x l)
  (let loop ((y l))
    (if (##pair? y)
      (let ((couple (##car y)))
        (if (##eq? x (##car couple))
          couple
          (loop (##cdr y))))
        #f)))

(define (##assq-cdr x l)
  (let loop ((y l))
    (if (##pair? y)
      (let ((couple (##car y)))
        (if (##eq? x (##cdr couple))
          couple
          (loop (##cdr y))))
        #f)))

(define-system (##assoc x lst)
  (let loop ((lst lst))
    (if (##pair? lst)
      (let ((couple (##car lst)))
        (let ((y (##car couple)))
          (if (##equal? x y #f)
            couple
            (loop (##cdr lst)))))
      #f)))

(define-system (##symbol->string sym)
  (macro-symbol-name sym))

(define-system (##symbol-hash sym)
  (macro-symbol-hash sym))

(define-system (##string->symbol str)
  (##make-interned-symbol str))

(define-system (##keyword->string key)
  (macro-keyword-name key))

(define-system (##keyword-hash key)
  (macro-keyword-hash key))

(define-system (##string->keyword str)
  (##make-interned-keyword str))

; numeric procedures are in "_num.scm"

; character procedures are in "_std.scm"

; string procedures are in "_std.scm"

; vector procedures are in "_std.scm"

(define-system (##apply p l))

(define-system (##map p lst)
  (let loop ((lst lst))
    (if (##pair? lst)
      (##cons (p (##car lst)) (loop (##cdr lst)))
      '())))

(define-system (##for-each p lst)
  (let loop ((lst lst))
    (if (##pair? lst)
      (begin (p (##car lst)) (loop (##cdr lst)))
      '())))

(define-system (##call-with-current-continuation p))

; input/output procedures are in "_io.scm"

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

; Procedures for front end

(define-system (##quasi-append x y)
  (macro-force-vars (x)
    (if (##pair? x)
      (let ((result (##cons (##car x) '())))
        (##set-cdr!
          (let loop ((end result) (x (##cdr x)))
            (macro-force-vars (x)
              (if (##pair? x)
                (let ((tail (##cons (##car x) '())))
                  (##set-cdr! end tail)
                  (loop tail (##cdr x)))
                end)))
          y)
        result)
      y)))

(define-system (##quasi-list . l)
  l)

(define-system (##quasi-cons x y)
  (##cons x y))

(define-system (##quasi-list->vector l)
  (let loop1 ((x l) (n 0))
    (macro-force-vars (x)
      (if (##pair? x)
        (loop1 (##cdr x) (##fixnum.+ n 1))
        (let ((vect (##make-vector n 0)))
          (let loop2 ((x l) (i 0))
            (macro-force-vars (x)
              (if (##pair? x)
                (begin
                  (##vector-set! vect i (##car x))
                  (loop2 (##cdr x) (##fixnum.+ i 1)))
                vect))))))))

(define-system (##case-memv x l)
  (macro-force-vars (x)
    (let loop ((l l))
      (if (##pair? l)
        (if (let () (##declare (generic)) (##eqv? x (##car l)))
          l
          (loop (##cdr l)))
        #f))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

; Global variables

(define-system (##make-global-var id))

(define-system (##global-var-ref gv))

(define-system (##global-var-set! gv val))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

; Jobs

(define (##make-jobs)
  (##make-queue))

(define (##add-job jobs h)
  (##queue-put! jobs h))

(define (##invoke-jobs jobs)
  (let loop ((lst (##queue-peek-list jobs)))
    (if (##pair? lst)
      (begin
        ((##car lst))
        (loop (##cdr lst))))))

(define (##make-queue)
  (let ((queue (##cons '() '())))
    (##set-car! queue queue)
    queue))

(define (##queue-put! queue x)
  (##declare (not interrupts-enabled))
  (let ((last (##cons x '())))
    (##set-cdr! (##car queue) last)
    (##set-car! queue last)
    queue))

(define (##queue-peek-list queue)
  (##cdr queue))

;------------------------------------------------------------------------------
