;==============================================================================

; file: "_thread.scm"

; Copyright (C) 1994-2000 by Marc Feeley, All Rights Reserved.

;==============================================================================

(##include "header.scm")

;==============================================================================

(##declare (not interrupts-enabled))

;------------------------------------------------------------------------------

; Procedures to support multitasking

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

; (##add-gc-interrupt-job thunk) can be called to add another job to do
; after a GC.  (##clear-gc-interrupt-jobs) clears the jobs.

(define ##gc-interrupt-jobs #f)

(define (##add-gc-interrupt-job thunk)
  (##add-job ##gc-interrupt-jobs thunk))

(define (##clear-gc-interrupt-jobs)
  (set! ##gc-interrupt-jobs (##make-jobs)))

(##clear-gc-interrupt-jobs)

(##interrupt-vector-set! 2
  (lambda ()
    (##gc-finalization)
    (##invoke-jobs ##gc-interrupt-jobs)))

; (##user-interrupt) is called on each user interrupt.

(##interrupt-vector-set! 0 ##handle-user-interrupt)

;------------------------------------------------------------------------------
;==============================================================================

; file: "thread.scm"

; Copyright (C) 2000 by Marc Feeley, All Rights Reserved.

; Version 1.3 (august 14, 2000)

(##define-macro (macro-if-gambit-3.0 yes no)
  no)

(##define-macro (macro-not-yet-implemented)
  `(##void));;;;;;;;;;;*************

;------------------------------------------------------------------------------

; Implementation of dynamic environments.

(define ##dynamic-define
  (let ()

    (##declare (not inline) (not interrupts-enabled))

    (lambda (name #!optional (val (macro-absent-obj)))
      (##env-insert!
       (macro-denv-global (macro-thread-denv (macro-current-thread)))
       name
       (if (##eq? val (macro-absent-obj))
         (##void)
         val)))))

(define ##dynamic-ref
  (let ()

    (##declare (not inline) (not interrupts-enabled))

    (lambda (name #!optional (default (macro-absent-obj)))
      (let* ((current-thread
              (macro-current-thread))
             (c1
              (macro-thread-denv-cache1 current-thread)))
        (if (##eq? name (##car c1))
          (##cdr c1)
          (let ((c2 (macro-thread-denv-cache2 current-thread)))
            (if (##eq? name (##car c2))
              (begin
                (macro-thread-denv-cache2-set! current-thread c1)
                (macro-thread-denv-cache1-set! current-thread c2)
                (##cdr c2))
              (let ((c3 (macro-thread-denv-cache3 current-thread)))
                (if (##eq? name (##car c3))
                  (begin
                    (macro-thread-denv-cache3-set! current-thread c2)
                    (macro-thread-denv-cache2-set! current-thread c1)
                    (macro-thread-denv-cache1-set! current-thread c3)
                    (##cdr c3))
                  (let* ((denv
                          (macro-thread-denv current-thread))
                         (x
                          (or (##env-lookup (macro-denv-local denv) name)
                              (##env-lookup (macro-denv-global denv) name))))
                    (if x
                      (begin
                        (macro-thread-denv-cache3-set!
                         current-thread
                         (macro-thread-denv-cache2 current-thread))
                        (macro-thread-denv-cache2-set!
                         current-thread
                         (macro-thread-denv-cache1 current-thread))
                        (macro-thread-denv-cache1-set!
                         current-thread
                         x)
                        (##cdr x))
                      (if (##eq? default (macro-absent-obj))
                        (macro-raise
                         (macro-make-unbound-dynamic-variable-exception
                          name))
                        default))))))))))))

(define ##dynamic-set!
  (let ()

    (##declare (not inline) (not interrupts-enabled))

    (lambda (name val)
      (let* ((current-thread
              (macro-current-thread))
             (c1
              (macro-thread-denv-cache1 current-thread)))
        (if (##eq? name (##car c1))
          (begin
            (##set-cdr! c1 val)
            (##void))
          (let ((c2 (macro-thread-denv-cache2 current-thread)))
            (if (##eq? name (##car c2))
              (begin
                (macro-thread-denv-cache2-set! current-thread c1)
                (macro-thread-denv-cache1-set! current-thread c2)
                (##set-cdr! c2 val)
                (##void))
              (let ((c3 (macro-thread-denv-cache3 current-thread)))
                (if (##eq? name (##car c3))
                  (begin
                    (macro-thread-denv-cache3-set! current-thread c2)
                    (macro-thread-denv-cache2-set! current-thread c1)
                    (macro-thread-denv-cache1-set! current-thread c3)
                    (##set-cdr! c3 val)
                    (##void))
                  (let* ((denv
                          (macro-thread-denv current-thread))
                         (x
                          (or (##env-lookup (macro-denv-local denv) name)
                              (##env-lookup (macro-denv-global denv) name))))
                    (if x
                      (begin
                        (macro-thread-denv-cache3-set!
                         current-thread
                         (macro-thread-denv-cache2 current-thread))
                        (macro-thread-denv-cache2-set!
                         current-thread
                         (macro-thread-denv-cache1 current-thread))
                        (macro-thread-denv-cache1-set!
                         current-thread
                         x)
                        (##set-cdr! x val)
                        (##void))
                      (macro-raise
                       (macro-make-unbound-dynamic-variable-exception
                        name)))))))))))))

(define ##dynamic-let
  (let ()

    (##declare (not inline) (not interrupts-enabled))

    (lambda (name val thunk)
      (let* ((name-val (##cons name val))
             (denv (macro-thread-denv (macro-current-thread)))
             (new-local-denv (##env-insert (macro-denv-local denv) name-val)))
        (##dynamic-env-bind
         (macro-make-denv
          (macro-denv-global denv)
          new-local-denv
          (macro-denv-interrupt-mask denv)
          (macro-denv-exception-handler denv)
          (macro-denv-debugging-settings denv)
          (macro-denv-input-port denv)
          (macro-denv-output-port denv))
         thunk)))))

(define ##dynamic-env-bind
  (let ()

    (##declare (not inline) (not interrupts-enabled))

    (lambda (denv thunk)
      (let* ((current-thread
              (macro-current-thread))
             (old-denv
              (macro-thread-denv current-thread)))
        (macro-thread-denv-set! current-thread denv)
        (let ((x (macro-env-name-val (macro-denv-local denv))))
          (macro-thread-denv-cache1-set! current-thread x)
          (macro-thread-denv-cache2-set! current-thread x)
          (macro-thread-denv-cache3-set! current-thread x)
          (let* ((result
                  (let ()
                    (##declare (interrupts-enabled))
                    (thunk)))
                 (current-thread
                  (macro-current-thread)))
            (macro-thread-denv-set! current-thread old-denv)
            (let ((x (macro-env-name-val (macro-denv-local old-denv))))
              (macro-thread-denv-cache1-set! current-thread x)
              (macro-thread-denv-cache2-set! current-thread x)
              (macro-thread-denv-cache3-set! current-thread x)
              result)))))))

(define ##env-insert
  (let ()

    (##declare (not inline))

    (lambda (env name-val)
      (let* ((name (##car name-val))
             (hash-name (macro-symbol-hash name)))

        (define (insert env)
          (if (##null? env)
            (macro-make-env
             name-val
             '()
             '())
            (let* ((x (macro-env-name-val env))
                   (name-x (##car x))
                   (hash-name-x (macro-symbol-hash name-x)))
              (cond ((##fixnum.< hash-name hash-name-x)
                     (macro-make-env
                      x
                      (insert (macro-env-left env))
                      (macro-env-right env)))
                    ((or (##fixnum.< hash-name-x hash-name)
                         (##not (##eq? name-x name)))
                     (macro-make-env
                      x
                      (macro-env-left env)
                      (insert (macro-env-right env))))
                    (else
                     (macro-make-env
                      name-val
                      (macro-env-left env)
                      (macro-env-right env)))))))

        (insert env)))))

(define ##env-insert!
  (let ()

    (##declare (not inline) (not interrupts-enabled))

    (lambda (env name val)
      (let ((hash-name (macro-symbol-hash name)))

        (define (insert! env)
          (let* ((x (macro-env-name-val env))
                 (name-x (##car x))
                 (hash-name-x (macro-symbol-hash name-x)))
            (cond ((##fixnum.< hash-name hash-name-x)
                   (let ((branch (macro-env-left env)))
                     (if (##null? branch)
                       (let ((y (##cons name val)))
                         (macro-env-left-set! env (macro-make-env y '() '()))
                         (##void))
                       (insert! branch))))
                  ((or (##fixnum.< hash-name-x hash-name)
                       (##not (##eq? name-x name)))
                   (let ((branch (macro-env-right env)))
                     (if (##null? branch)
                       (let ((y (##cons name val)))
                         (macro-env-right-set! env (macro-make-env y '() '()))
                         (##void))
                       (insert! branch))))
                  (else
                   (##set-cdr! x val)
                   (##void)))))

        (insert! env)))))

(define ##env-lookup
  (let ()

    (##declare (not inline) (not interrupts-enabled))

    (lambda (env name)
      (let ((hash-name (macro-symbol-hash name)))

        (define (lookup env)
          (if (##null? env)
            #f
            (let* ((x (macro-env-name-val env))
                   (name-x (##car x))
                   (hash-name-x (macro-symbol-hash name-x)))
              (cond ((##fixnum.< hash-name hash-name-x)
                     (lookup (macro-env-left env)))
                    ((or (##fixnum.< hash-name-x hash-name)
                         (##not (##eq? name-x name)))
                     (lookup (macro-env-right env)))
                    (else
                     x)))))

        (lookup env)))))

;------------------------------------------------------------------------------

; Implementation of time objects.

(define ##time-tag
  '#(time -1))

(define ##absrel-timeout->timeout
  (let ()

    (##declare (not inline))

    (lambda (absrel-timeout)
      (cond ((##not absrel-timeout)
             #t)
            ((macro-time? absrel-timeout)
             (macro-update-current-time!)
             (let ((current-time
                    (macro-current-time
                     (macro-thread-floats (macro-run-queue))))
                   (seconds
                    (macro-time-seconds absrel-timeout)))
               (and (##flonum.< current-time seconds)
                    seconds)))
            ((and (##fixnum? absrel-timeout)
                  (##not (##fixnum.< 0 absrel-timeout)))
             #f)
            (else
             (let ((flonum-absrel-timeout
                    (macro-real->inexact absrel-timeout)))
               (and (##flonum.positive? flonum-absrel-timeout)
                    (begin
                      (macro-update-current-time!)
                      (let ((current-time
                             (macro-current-time
                              (macro-thread-floats (macro-run-queue)))))
                        (##flonum.+ current-time
                                    flonum-absrel-timeout))))))))))

(define (##trap-check-time name . args)
  (##runtime-error "TIME expected" name args))

(define (##trap-check-time* name . args)
  (##runtime-error* "TIME expected" name args))

;------------------------------------------------------------------------------

; Implementation of thread related exceptions.

(define ##deadlock-exception
  (macro-make-structure-object '#(deadlock-exception 0)))

(define ##join-timeout-exception
  (macro-make-structure-object '#(join-timeout-exception 0)))

(define ##abandoned-mutex-exception
  (macro-make-structure-object '#(abandoned-mutex-exception 0)))

(define ##started-thread-exception
  (macro-make-structure-object '#(started-thread-exception 0)))

(define ##terminated-thread-exception
  (macro-make-structure-object '#(terminated-thread-exception 0)))

(define ##uncaught-exception-tag
  '#(uncaught-exception -1))

(define (##trap-check-uncaught-exception name . args)
  (##runtime-error "UNCAUGHT EXCEPTION object expected" name args))

(define (##trap-check-uncaught-exception* name . args)
  (##runtime-error* "UNCAUGHT EXCEPTION object expected" name args))

(define ##scheduler-exception-tag
  '#(scheduler-exception -1))

(define (##trap-check-scheduler-exception name . args)
  (##runtime-error "SCHEDULER EXCEPTION object expected" name args))

(define (##trap-check-scheduler-exception* name . args)
  (##runtime-error* "SCHEDULER EXCEPTION object expected" name args))

(define ##noncontinuable-exception-tag
  '#(noncontinuable-exception -1))

(define (##trap-check-noncontinuable-exception name . args)
  (##runtime-error "NONCONTINUABLE EXCEPTION object expected" name args))

(define (##trap-check-noncontinuable-exception* name . args)
  (##runtime-error* "NONCONTINUABLE EXCEPTION object expected" name args))

(define ##unbound-dynamic-variable-exception-tag
  '#(unbound-dynamic-variable-exception -1))

(define (##trap-check-unbound-dynamic-variable-exception name . args)
  (##runtime-error "UNBOUND DYNAMIC VARIABLE EXCEPTION object expected" name args))

(define (##trap-check-unbound-dynamic-variable-exception* name . args)
  (##runtime-error* "UNBOUND DYNAMIC VARIABLE EXCEPTION object expected" name args))

(define ##low-level-exception-tag
  '#(low-level-exception -1))

(define (##trap-check-low-level-exception name . args)
  (##runtime-error "LOW-LEVEL EXCEPTION object expected" name args))

(define (##trap-check-low-level-exception* name . args)
  (##runtime-error* "LOW-LEVEL EXCEPTION object expected" name args))

;------------------------------------------------------------------------------

; Implementation of blocked thread queues.

(macro-generate-rbtree
 macro-btq-init!
 macro-thread->btq
 ##btq-insert!
 ##btq-remove!
 ##btq-reposition!
 macro-btq-singleton?
 macro-btq-color
 macro-btq-color-set!
 macro-btq-parent
 macro-btq-parent-set!
 macro-btq-left
 macro-btq-left-set!
 macro-btq-right
 macro-btq-right-set!
 macro-thread-higher-prio?
 macro-btq-leftmost
 macro-btq-leftmost-set!
 #f
 #f
)

(define ##btq-abandon!
  (let ()

    (##declare (not inline) (not interrupts-enabled))

    (lambda (btq)
      (macro-btq-deq-remove! btq)
      (let ((leftmost (macro-btq-leftmost btq)))
        (if (##eq? leftmost btq)
          (macro-btq-unlink! btq 'abandoned)
          (if (macro-mutex? btq)
            (##mutex-signal-no-reschedule! btq leftmost #t)
            (begin
              (let ((owner (macro-btq-owner btq)))
                (if (macro-thread? owner)
                  (##thread-effective-priority-downgrade! owner)))
              (macro-btq-unlink! btq 'abandoned))))))))

; Implementation of timeout queues.

(macro-generate-rbtree
 macro-toq-init!
 macro-thread->toq
 ##toq-insert!
 ##toq-remove!
 ##toq-reposition!
 macro-toq-singleton?
 macro-toq-color
 macro-toq-color-set!
 macro-toq-parent
 macro-toq-parent-set!
 macro-toq-left
 macro-toq-left-set!
 macro-toq-right
 macro-toq-right-set!
 macro-thread-sooner-or-simultaneous-and-higher-prio?
 macro-toq-leftmost
 macro-toq-leftmost-set!
 #f
 #f
)

;------------------------------------------------------------------------------

; Implementation of threads.

(define (##trap-check-thread name . args)
  (##runtime-error "THREAD expected" name args))

(define (##trap-check-thread* name . args)
  (##runtime-error* "THREAD expected" name args))

(define ##current-thread
  (let ()

    (##declare (not inline) (not interrupts-enabled))

    (lambda ()
      (macro-current-thread))))

(define ##make-thread
  (let ()

    (##declare (not inline) (not interrupts-enabled))

    (lambda (thunk name tgroup)
      (macro-make-thread thunk name tgroup))))

(define ##thread-start!
  (let ()

    (##declare (not inline) (not interrupts-enabled))

    (lambda (thread)
     (if (and (macro-thread-end-condvar thread)
              (macro-thread-exception? thread)
              (##not (macro-thread-result thread)))
       (begin
         (macro-thread-result-set! thread ##thread-start-action!)
         (##btq-insert! (macro-run-queue) thread)
         (macro-thread-reschedule-if-needed!))
       (macro-raise (macro-make-started-thread-exception))))))

(define ##thread-base-priority-set!
  (let ()

    (##declare (not inline) (not interrupts-enabled))

    (lambda (thread base-priority)
      (let ((floats (macro-thread-floats thread)))

        ; save old boosted priority for ##thread-boosted-priority-changed!

        (macro-temp-set!
         (macro-thread-floats (macro-run-queue))
         (macro-boosted-priority floats))

        (if (##flonum.= (macro-base-priority floats)
                        (macro-boosted-priority floats))
(begin (set! pcase3 (##fixnum.+ pcase3 1));;;;;;;;;;;;;;;;;;
          (macro-boosted-priority-set!
           floats
           base-priority)
)
(begin (set! pcase4 (##fixnum.+ pcase4 1));;;;;;;;;;;;;;;;;;
          (macro-boosted-priority-set!
           floats
           (##flonum.+ base-priority
                       (macro-priority-boost floats))))
)

        (macro-base-priority-set! floats base-priority)

        (##thread-boosted-priority-changed! thread)

        ; the change of priority may have made a higher priority
        ; thread runnable, check for this

        (macro-thread-reschedule-if-needed!)))))

(define ##thread-quantum-set!
  (let ()

    (##declare (not inline) (not interrupts-enabled))

    (lambda (thread quantum)

      (macro-thread-quantum-set! thread quantum)

      ; check if the current thread's quantum is now over

      (if (and (##eq? thread (macro-current-thread))
               (##not (##flonum.< (macro-thread-quantum-used thread) quantum)))
        (##thread-yield!)
        (##void)))))

(define ##thread-priority-boost-set!
  (let ()

    (##declare (not inline) (not interrupts-enabled))

    (lambda (thread priority-boost)
      (let ((floats (macro-thread-floats thread)))

        (macro-priority-boost-set! floats priority-boost)

        (if (##flonum.= (macro-base-priority floats)
                        (macro-boosted-priority floats))

(begin (set! pcase5 (##fixnum.+ pcase5 1));;;;;;;;;;;;;;;;;;
          (##void)
)

          (begin

(set! pcase6 (##fixnum.+ pcase6 1));;;;;;;;;;;;;;;;;;

            ; save old boosted priority for ##thread-boosted-priority-changed!

            (macro-temp-set!
             (macro-thread-floats (macro-run-queue))
             (macro-boosted-priority floats))

            (macro-boosted-priority-set!
             floats
             (##flonum.+ (macro-base-priority floats)
                         priority-boost))

            (##thread-boosted-priority-changed! thread)

            ; the change of priority may have made a higher priority
            ; thread runnable, check for this

            (macro-thread-reschedule-if-needed!)))))))

(define ##thread-boosted-priority-changed!
  (let ()

    (##declare (not inline) (not interrupts-enabled))

    (lambda (thread)
      (let ((floats (macro-thread-floats thread)))
        (cond ((##flonum.<
                (macro-effective-priority floats)
                (macro-boosted-priority floats))
(set! pcase7 (##fixnum.+ pcase7 1));;;;;;;;;;;;;;;;;;
               (macro-effective-priority-set!
                floats
                (macro-boosted-priority floats))
               (##thread-effective-priority-changed! thread #t))
              ((##flonum.=
                (macro-effective-priority floats)
                (macro-temp (macro-thread-floats (macro-run-queue))))
(set! pcase8 (##fixnum.+ pcase8 1));;;;;;;;;;;;;;;;;;
               (##thread-effective-priority-downgrade! thread))
(else (set! pcase9 (##fixnum.+ pcase9 1)));;;;;;;;;;;;;;;;;;
)))))

(define ##thread-effective-priority-changed!
  (let ()

    (##declare (not inline) (not interrupts-enabled))

    (lambda (thread effective-priority-increased?)

      (if (macro-toq-parent thread)
        (##toq-reposition! thread)) ; reposition thread in the toq it is in

      (if (macro-btq-parent thread)
        (begin
          (##btq-reposition! thread) ; reposition thread in the btq it is in

          ; make sure the owner of the blocked thread queue
          ; (i.e. mutex, condvar, etc) inherits the thread's effective
          ; priority

          (let ((owner (macro-btq-owner (macro-thread->btq thread))))
            (if (macro-thread? owner)
              (if effective-priority-increased?
(begin (set! pcase10 (##fixnum.+ pcase10 1));;;;;;;;;;;;;;;;;;
                (macro-thread-inherit-priority! owner thread)
)
(begin (set! pcase11 (##fixnum.+ pcase11 1));;;;;;;;;;;;;;;;;;
                (##thread-effective-priority-downgrade! owner)))))))))
)

; (##thread-effective-priority-downgrade! thread) is called to
; recompute the effective priority of a thread.  It is only called in
; situations where the new effective priority is no more than the old
; one.

(define ##thread-effective-priority-downgrade!
  (let ()

    (##declare (not inline) (not interrupts-enabled))

    (lambda (thread)
      (let ((floats (macro-thread-floats thread)))

        ; save old effective priority for later

        (macro-temp-set!
         (macro-thread-floats (macro-run-queue))
         (macro-effective-priority floats))

        ; compute the maximum of the boosted priority and the
        ; effective priority of all the threads in blocked thread
        ; queues (i.e. mutexes, condvars, etc) owned by the thread

        (macro-effective-priority-set!
         floats
         (macro-boosted-priority floats))

        (let loop ((btq (macro-btq-deq-next thread)))
          (if (##not (##eq? btq thread))
            (let ((leftmost (macro-btq-leftmost btq)))
              (if (##not (##eq? leftmost btq))
                (let ((leftmost-floats (macro-thread-floats leftmost)))
                  (if (##flonum.< (macro-effective-priority floats)
                                  (macro-effective-priority leftmost-floats))
                    (macro-effective-priority-set!
                     floats
                     (macro-effective-priority leftmost-floats)))))
              (loop (macro-btq-deq-next btq)))))

;;;;;;;;;;;;(set! pcase14 (##cons (##cons (macro-temp (macro-thread-floats (macro-run-queue))) (macro-effective-priority floats))pcase14));;;;;;;;;;;;;;;;;;;;;;;
        (if (##not (##flonum.=
                    (macro-temp (macro-thread-floats (macro-run-queue)))
                    (macro-effective-priority floats)))
(begin (set! pcase12 (##fixnum.+ pcase12 1));;;;;;;;;;;;;;;;;;
          (##thread-effective-priority-changed! thread #f)
)
(set! pcase13 (##fixnum.+ pcase13 1));;;;;;;;;;;;;;;;;;
)))))

(define ##thread-btq-insert!
  (let ()

    (##declare (not inline) (not interrupts-enabled))

    (lambda (btq thread)
      (##btq-insert! btq thread)
      (let ((owner (macro-btq-owner btq)))
        (if (macro-thread? owner)
          (macro-thread-inherit-priority! owner thread))))))

(define ##thread-btq-remove!
  (let ()

    (##declare (not inline) (not interrupts-enabled))

    (lambda (thread)
      (let ((owner (macro-btq-owner (macro-thread->btq thread))))
        (##btq-remove! thread)
        (if (macro-thread? owner)
          (if (##flonum.= (macro-thread-effective-priority thread)
                          (macro-thread-effective-priority owner))
            (##thread-effective-priority-downgrade! owner)))))))

(define ##thread-toq-remove!
  (let ()

    (##declare (not inline) (not interrupts-enabled))

    (lambda (thread)
      (##toq-remove! thread))))

(define ##thread-check-timeouts!
  (let ()

    (##declare (not inline) (not interrupts-enabled))

    (lambda ()

      (macro-update-current-time!)

      (let ((run-queue (macro-run-queue)))
        (let loop ()
          (let ((leftmost (macro-toq-leftmost run-queue)))
            (if (and (##not (##eq? leftmost run-queue))
                     (##not (macro-thread-sooner? run-queue leftmost)))
              (begin
                (macro-thread-result-set! leftmost ##thread-timeout-action!)
                (macro-thread-btq-remove-if-in-btq! leftmost)
                (##thread-toq-remove! leftmost)
                (##btq-insert! run-queue leftmost)
                (loop)))))))))

(define ##thread-check-devices!
  (let ()

    (##declare (not inline) (not interrupts-enabled))

    (lambda (timeout)
      (let* ((run-queue (macro-run-queue))
             (code (##condvar-select! run-queue timeout)))
        (let loop ((condvar (macro-btq-deq-next run-queue)))
          (if (##eq? condvar run-queue)
            code
            (let ((next (macro-btq-deq-next condvar)))
              (if (##fixnum.odd? (macro-btq-owner condvar))
                (begin
                  (macro-btq-deq-remove! condvar)
                  (macro-btq-deq-init! condvar)
                  (##condvar-signal-no-reschedule! condvar #t)))
              (loop next))))))))

(define ##thread-heartbeat!
  (let ()

    (##declare (not inline) (not interrupts-enabled))

    (lambda ()
      (let ((code (##thread-check-devices! #f))) ; poll devices

        (##thread-check-timeouts!)

        (if (and (##fixnum.< code 0)
                 (##not (##fixnum.= code ##err-code-EINTR)))

          ; there was an error that cannot be handled, so force the
          ; primordial thread to wakeup and raise a "scheduler
          ; error" exception

          (##thread-report-scheduler-error! code))

        (let* ((run-queue
                (macro-run-queue))
               (current-thread
                (macro-current-thread))
               (run-queue-floats
                (macro-thread-floats run-queue))
               (current-thread-floats
                (macro-thread-floats current-thread))
               (quantum-used
                (##flonum.+ (macro-quantum-used current-thread-floats)
                            (macro-heartbeat-interval run-queue-floats))))

          (macro-quantum-used-set! current-thread-floats quantum-used)

          (if (##flonum.< quantum-used
                          (macro-quantum current-thread-floats))
            (macro-thread-reschedule-if-needed!)
            (##thread-yield!)))))))

(define ##thread-yield!
  (let ()

    (##declare (not inline) (not interrupts-enabled))

    (lambda ()
      (let* ((current-thread
              (macro-current-thread))
             (run-queue
              (macro-run-queue)))
        (if (##eq? (macro-btq-singleton? run-queue) current-thread)
          (begin
            ; fast case where only one thread is runnable
            (macro-thread-unboost-and-clear-quantum-used! current-thread)
            (##void))
          (macro-thread-save!
           (lambda (current-thread dummy1 dummy2 dummy3)
             (##btq-remove! current-thread)
             (macro-thread-unboost-and-clear-quantum-used! current-thread)
             (macro-thread-result-set! current-thread ##thread-void-action!)
             (##btq-insert! (macro-run-queue) current-thread)
             (##thread-schedule!))
           #f
           #f
           #f))))))

(define ##thread-reschedule!
  (let ()

    (##declare (not inline) (not interrupts-enabled))

    (lambda ()
      (macro-thread-save!
       (lambda (current-thread dummy1 dummy2 dummy3)
         (macro-thread-result-set! current-thread ##thread-void-action!)
         (##thread-schedule!))
       #f
       #f
       #f))))

(define ##thread-sleep!
  (let ()

    (##declare (not inline) (not interrupts-enabled))

    (lambda (absrel-timeout)
      (let ((timeout (##absrel-timeout->timeout absrel-timeout)))
        (if timeout
          (let loop ()
            (let ((result
                   (macro-thread-save!
                    (lambda (current-thread timeout dummy1 dummy2)
                      (##btq-remove! current-thread)
                      (macro-thread-unboost-and-clear-quantum-used!
                       current-thread)
                      (if (##not (##eq? timeout #t))
                        (begin
                          (macro-thread-timeout-set! current-thread timeout)
                          (##toq-insert! (macro-run-queue) current-thread)))
                      (##thread-schedule!))
                    timeout
                    #f
                    #f)))
              (if (##eq? result (##void))
                (loop)
                (##void))))
          (##void))))))

(define ##thread-schedule!
  (let ()

    (##declare (not inline) (not interrupts-enabled))

    (lambda ()
      (let ((run-queue (macro-run-queue)))

        ; check if there are runnable threads

        (let ((next-thread
               (macro-btq-leftmost run-queue)))
          (if (##not (##eq? next-thread run-queue))

            ; continue executing the next runnable thread

            (macro-thread-restore!
             next-thread
             (macro-thread-result next-thread))

            ; check if there are threads waiting for a timeout or for
            ; devices to become ready

            (let ((next-sleeper
                   (macro-toq-leftmost run-queue))
                  (next-condvar
                   (macro-btq-deq-next run-queue)))
              (if (or (##not (##eq? next-sleeper run-queue))
                      (##not (##eq? next-condvar run-queue)))

                ; wait for the next timeout or for a device to become
                ; ready

                (let ((code
                       (##thread-check-devices!
                        (if (##eq? next-sleeper run-queue)
                          #t ; timeout is infinite
                          (macro-thread-floats next-sleeper)))))

                  (##thread-check-timeouts!)

                  (if (##fixnum.< code 0)

                    (if (##fixnum.= code ##err-code-EINTR)

                      ; an interrupt may need to be serviced, so make
                      ; sure at least one thread is runnable

                      (let ((next-thread
                             (macro-btq-leftmost run-queue)))
                        (if (##eq? next-thread run-queue)
                          (let ((next-sleeper
                                 (macro-toq-leftmost run-queue)))
                            (if (##not (##eq? next-sleeper run-queue))

                              (##thread-interrupt!
                               next-sleeper
                               ##thread-void-action!)

                              (let ((next-condvar
                                     (macro-btq-deq-next run-queue)))
                                (if (##not (##eq? next-condvar run-queue))

                                  #f;;;;;;;;;
                                  #f))))))

                      ; there was an error that cannot be handled, so
                      ; force the primordial thread to wakeup and
                      ; raise a "scheduler error" exception

                      (##thread-report-scheduler-error! code))))

                ; no thread can possibly make further progress, so
                ; force the primordial thread to wakeup and raise a
                ; "deadlock" exception

                (##thread-interrupt!
                 (macro-primordial-thread)
                 ##thread-deadlock-action!))

              (##thread-schedule!))))))))

(define ##thread-report-scheduler-error!
  (let ()

    (##declare (not inline) (not interrupts-enabled))

    (lambda (code)
      (##thread-interrupt!
       (macro-primordial-thread)
       (lambda ()

         (macro-raise
          (macro-make-scheduler-exception
           (macro-make-low-level-exception code)))

         ; return void so that primordial thread will continue
         ; normally if exception handler returns

         (##void))))))

(define ##thread-interrupt!
  (let ()

    (##declare (not inline) (not interrupts-enabled))

    (lambda (thread action)

      ; Note: the action procedure must return void in order to
      ; restart the interrupted thread properly.

      ; remove the thread from any blocked thread queue and timeout
      ; queue it is in

      (macro-thread-btq-remove-if-in-btq! thread)
      (macro-thread-toq-remove-if-in-toq! thread)

      (macro-thread-result-set! thread action)

      (##btq-insert! (macro-run-queue) thread))))

(define ##thread-start-action!
  (let ()

    (##declare (not inline) (not interrupts-enabled))

    (lambda ()
      (let* ((current-thread
              (macro-current-thread))
             (thunk
              (macro-thread-exception? current-thread)))
        (macro-thread-exception?-set! current-thread #f)
        (let ((result (thunk)))
          (##thread-end! (macro-current-thread) #f result))))))

(define ##thread-void-action!
  (let ()

    (##declare (not inline) (not interrupts-enabled))

    (lambda ()
      (##void))))

(define ##thread-abandoned-mutex-action!
  (let ()

    (##declare (not inline) (not interrupts-enabled))

    (lambda ()
      (macro-raise (macro-make-abandoned-mutex-exception)))))

(define ##thread-locked-mutex-action!
  (let ()

    (##declare (not inline) (not interrupts-enabled))

    (lambda ()
      #t)))

(define ##thread-signaled-condvar-action!
  (let ()

    (##declare (not inline) (not interrupts-enabled))

    (lambda ()
      #t)))

(define ##thread-timeout-action!
  (let ()

    (##declare (not inline) (not interrupts-enabled))

    (lambda ()
      #f)))

(define ##thread-deadlock-action!
  (let ()

    (##declare (not inline) (not interrupts-enabled))

    (lambda ()

      (macro-raise (macro-make-deadlock-exception))

      ; return void so that primordial thread will continue normally
      ; if exception handler returns

      (##void))))

(define ##thread-suspend!
  (let ()

    (##declare (not inline) (not interrupts-enabled))

    (lambda (thread)
      (macro-not-yet-implemented))));;;;;;;;;;;;;;;;

(define ##thread-resume!
  (let ()

    (##declare (not inline) (not interrupts-enabled))

    (lambda (thread)
      (macro-not-yet-implemented))));;;;;;;;;;;;;;

(define ##thread-terminate!
  (let ()

    (##declare (not inline) (not interrupts-enabled))

    (lambda (thread)
      (##thread-end!
       thread
       #t
       (macro-make-terminated-thread-exception)))))

(define ##thread-base-exception-handler
  (let ()

    (##declare (not inline) (not interrupts-enabled))

    (lambda (exc)
      (##thread-end!
       (macro-current-thread)
       #t
       (macro-make-uncaught-exception exc)))))

(define ##thread-end!
  (let ()

    (##declare (not inline) (not interrupts-enabled))

    (lambda (thread exception? result)

      ; perform thread termination only if thread is not already
      ; terminated, or in the process of terminating

      (let ((end-condvar (macro-thread-end-condvar thread)))
        (if (##not end-condvar)
          (##void)
          (begin

            ; the thread must abandon all the blocked thread queues
            ; (i.e. mutexes, condvars, etc) it owns

            (let loop ()
              (let ((next-btq (macro-btq-deq-next thread)))
                (if (##not (##eq? next-btq thread))
                  (begin
                    (##btq-abandon! next-btq)
                    (loop)))))

            (macro-thread-end-condvar-set! thread #f)
            (macro-thread-exception?-set! thread exception?)
            (macro-thread-result-set! thread result)
            (##condvar-signal-no-reschedule! end-condvar #t)
            (macro-thread-btq-remove-if-in-btq! thread)
            (macro-thread-toq-remove-if-in-toq! thread)
;;;;;;;;            (macro-tgroup-threads-deq-remove! thread)
;;;;;;;;            (macro-tgroup-threads-deq-init! thread)
            (macro-thread-frame-set! thread 0)
            (macro-thread-denv-set! thread #f)
            (macro-thread-denv-cache1-set! thread #f)
            (macro-thread-denv-cache2-set! thread #f)
            (macro-thread-denv-cache3-set! thread #f)
            (cond ((##eq? thread (macro-primordial-thread))
                   (##exit 1))
                  ((##eq? thread (macro-current-thread))
                   (##thread-schedule!))
                  (else
                   (macro-thread-reschedule-if-needed!)))))))))

(define ##thread-join!
  (let ()

    (##declare (not inline) (not interrupts-enabled))

    (lambda (thread absrel-timeout timeout-val)
      (let ((joined-before-timeout?
             (let ((end-condvar (macro-thread-end-condvar thread)))
               (if end-condvar
                 (let ((timeout (##absrel-timeout->timeout absrel-timeout)))
                   (if timeout
                     (let loop ()
                       (let ((result
                              (macro-thread-save!
                               (lambda (current-thread thread timeout dummy1)
                                 (let ((end-condvar
                                        (macro-thread-end-condvar thread)))

                                   ; do a final check of the state of
                                   ; the thread (this double check is
                                   ; necessary because a thread-switch
                                   ; may occur during the capturing of
                                   ; the continuation by
                                   ; macro-thread-save!)

                                   (if end-condvar
                                     (begin
                                       (##btq-remove! current-thread)
                                       (macro-thread-boost-and-clear-quantum-used!
                                        current-thread)
                                       (##thread-btq-insert!
                                        end-condvar
                                        current-thread)
                                       (if (##not (##eq? timeout #t))
                                         (begin
                                           (macro-thread-timeout-set!
                                            current-thread
                                            timeout)
                                           (##toq-insert!
                                            (macro-run-queue)
                                            current-thread)))
                                       (##thread-schedule!))
                                     #t)))
                               thread
                               timeout
                               #f)))
                         (if (##eq? result (##void))
                           (loop)
                           result)))
                     #f))
                 #t))))

        (if joined-before-timeout?
          (if (macro-thread-exception? thread)
            (macro-raise (macro-thread-result thread))
            (macro-thread-result thread))
          (if (##eq? timeout-val (macro-absent-obj))
            (macro-raise (macro-make-join-timeout-exception))
            timeout-val))))))

;;;;;;;;;*********
(define ##readtable-sym #f)
(define ##error-port-sym #f)

(define ##thread-startup!
  (let ()

    (##declare (not inline) (not interrupts-enabled))

    (lambda ()
      (let* ((debugging-settings
              0);;;;;;;;;;*************
             (input-port
              ##stdin);;;;;;;;;;*************
             (output-port
              ##stdout);;;;;;;;;;*************
             (error-port
              ##stderr);;;;;;;;;;*************
             (readtable
              ##main-readtable);;;;;;;;;;*************
             (readtable-sym
              (macro-make-uninterned-symbol "readtable" 268435456)) ; (expt 2 28)
             (error-port-sym
              (macro-make-uninterned-symbol "error-port" 268435456)) ; (expt 2 28)
             (global-binding
              (##cons readtable-sym readtable))
             (local-binding
              (##cons error-port-sym error-port))
             (primordial-tgroup
              (##make-tgroup 'primordial #f))
             (primordial-thread
              (let ()

                ; these macros are defined to prevent the normal
                ; thread inheritance mechanism when the first thread
                ; is created

                (##define-macro (macro-current-thread)
                  `#f)

                (##define-macro (macro-thread-denv thread)
                  `#f)

                (##define-macro (macro-denv-global denv)
                  `(macro-make-env global-binding '() '()))

                (##define-macro (macro-denv-local denv)
                  `(macro-make-env local-binding '() '()))

                (##define-macro (macro-denv-interrupt-mask denv)
                  `0)

                (##define-macro (macro-denv-debugging-settings denv)
                  `debugging-settings)

                (##define-macro (macro-denv-input-port denv)
                  `input-port)

                (##define-macro (macro-denv-output-port denv)
                  `output-port)

                (##define-macro (macro-thread-denv-cache1 thread)
                  `local-binding)

                (##define-macro (macro-thread-denv-cache2 thread)
                  `local-binding)

                (##define-macro (macro-thread-denv-cache3 thread)
                  `local-binding)

                (##define-macro (macro-base-exception-handler)
                  `(macro-primordial-thread-base-exception-handler))

                (##define-macro (macro-thread-floats thread)
                  `#f)

                (##define-macro (macro-base-priority floats)
                  `(macro-primordial-thread-base-priority))

                (##define-macro (macro-quantum floats)
                  `(macro-primordial-thread-quantum))

                (##define-macro (macro-priority-boost floats)
                  `(macro-primordial-thread-priority-boost))

                ; create primordial thread

                (macro-make-thread #f 'primordial primordial-tgroup))))

        (set! ##readtable-sym readtable-sym)
        (set! ##error-port-sym error-port-sym)

        (macro-if-gambit-3.0;;;;;;;;;;;;;;;;;;;;;

         (begin
           (set! ##the-current-thread primordial-thread)
           (set! ##the-scheduler-state (macro-make-run-queue)))

         (##c-code
           "
           ___ps->current_thread = ___ARG1;
           ___ps->run_queue = ___ARG2;
           ___RESULT = ___FAL;
           "
           primordial-thread
           (macro-make-run-queue)))

        (macro-run-queue-primordial-thread-set!
         (macro-run-queue)
         primordial-thread)

        (##btq-insert! (macro-run-queue) primordial-thread)

        (##interrupt-vector-set! 1 ##thread-heartbeat!)

        (##void)))))

(define ##thread-heartbeat-interval-set!
  (let ()

    (##declare (not inline) (not interrupts-enabled))

    (lambda (seconds)
      (let* ((actual-interval
              (##heartbeat-interval-set! seconds))
             (run-queue
              (macro-run-queue))
             (run-queue-floats
              (macro-thread-floats run-queue)))
        (macro-heartbeat-interval-set! run-queue-floats actual-interval)
        actual-interval))))

;------------------------------------------------------------------------------

; Implementation of mutexes.

(define ##mutex-tag
  '#(mutex -1))

(define (##trap-check-mutex name . args)
  (##runtime-error "MUTEX expected" name args))

(define (##trap-check-mutex* name . args)
  (##runtime-error* "MUTEX expected" name args))

(define ##make-mutex
  (let ()

    (##declare (not inline) (not interrupts-enabled))

    (lambda (name)
      (macro-make-mutex name))))

(define ##mutex-lock-out-of-line!
  (let ()

    (##declare (not inline) (not interrupts-enabled))

    (lambda (mutex absrel-timeout new-owner)
      (let ((owner (macro-btq-owner mutex)))
        (if (##eq? owner 'abandoned)
          (begin
            (if new-owner
              (if (macro-thread-end-condvar new-owner)
                (macro-btq-link! mutex new-owner))
              (macro-btq-owner-set! mutex 'not-owned))
            (##thread-abandoned-mutex-action!))
          (let ((timeout (##absrel-timeout->timeout absrel-timeout)))
            (if timeout
              (let loop ()
                (let ((result
                       (macro-thread-save!
                        (lambda (current-thread mutex timeout new-owner)
                          (let ((owner (macro-btq-owner mutex)))

                            ; do a final check of the state of the
                            ; mutex and suspend if can't lock the
                            ; mutex (this double check is necessary
                            ; because a thread-switch may occur during
                            ; the capturing of the continuation by
                            ; macro-thread-save!)

                            (cond ((##not owner)
                                   (if new-owner
                                     (if (macro-thread-end-condvar new-owner)
                                       (macro-btq-link! mutex new-owner)
                                       (macro-btq-owner-set! mutex 'abandoned))
                                     (macro-btq-owner-set! mutex 'not-owned))
                                   #t)
                                  ((##eq? owner 'abandoned)
                                   (if new-owner
                                     (if (macro-thread-end-condvar new-owner)
                                       (macro-btq-link! mutex new-owner))
                                     (macro-btq-owner-set! mutex 'not-owned))
                                   (##thread-abandoned-mutex-action!))
                                  (else
                                   (##btq-remove! current-thread)
                                   (macro-thread-boost-and-clear-quantum-used!
                                    current-thread)
                                   (macro-thread-result-set!
                                    current-thread
                                    new-owner)
                                   (##thread-btq-insert! mutex current-thread)
                                   (if (##not (##eq? timeout #t))
                                     (begin
                                       (macro-thread-timeout-set!
                                        current-thread
                                        timeout)
                                       (##toq-insert!
                                        (macro-run-queue)
                                        current-thread)))
                                   (##thread-schedule!)))))
                        mutex
                        timeout
                        new-owner)))
                  (if (##eq? result (##void))
                    (loop)
                    result)))
              #f)))))))

(define ##mutex-signal!
  (let ()

    (##declare (not inline) (not interrupts-enabled))

    (lambda (mutex thread abandoned?)

      (##mutex-signal-no-reschedule! mutex thread abandoned?)

      ; check if the call to ##mutex-signal-no-reschedule! caused a
      ; higher priority thread to become runnable

      (macro-thread-reschedule-if-needed!))))

(define ##mutex-signal-no-reschedule!
  (let ()

    (##declare (not inline) (not interrupts-enabled))

    (lambda (mutex thread abandoned?)

      (##thread-btq-remove! thread)

      (let ((new-owner (macro-thread-result thread)))
        (if new-owner
          (if (macro-thread-end-condvar new-owner)
            (begin
              (macro-btq-link! mutex new-owner)
              (let ((new-leftmost (macro-btq-leftmost mutex)))
                (if (##not (##eq? new-leftmost mutex))
                  (macro-thread-inherit-priority! new-owner new-leftmost))))
            (macro-btq-unlink! mutex 'abandoned))
          (macro-btq-unlink! mutex 'not-owned)))

      (macro-thread-result-set!
       thread
       (if abandoned?
         ##thread-abandoned-mutex-action!
         ##thread-locked-mutex-action!))

      (macro-thread-toq-remove-if-in-toq! thread)

      (##btq-insert! (macro-run-queue) thread))))

(define ##mutex-signal-and-condvar-wait!
  (let ()

    (##declare (not inline) (not interrupts-enabled))

    (lambda (mutex condvar timeout)
      (if timeout
        (let loop ()
          (let ((result
                 (macro-thread-save!
                  (lambda (current-thread mutex condvar timeout)
                    (##btq-remove! current-thread)
                    (macro-thread-boost-and-clear-quantum-used!
                     current-thread)
                    (##thread-btq-insert! condvar current-thread)
                    (if (##not (##eq? timeout #t))
                      (begin
                        (macro-thread-timeout-set!
                         current-thread
                         timeout)
                        (##toq-insert!
                         (macro-run-queue)
                         current-thread)))
                    (macro-mutex-unlock-no-reschedule! mutex)
                    (##thread-schedule!))
                  mutex
                  condvar
                  timeout)))
            (if (##eq? result (##void))
              (loop)
              result)))
        (begin
          (macro-mutex-unlock-no-reschedule! mutex)
          (macro-thread-reschedule-if-needed!)
          #f)))))

;------------------------------------------------------------------------------

; Implementation of condition variables.

(define ##condvar-tag
  '#(condition-variable -1))

(define (##trap-check-condvar name . args)
  (##runtime-error "CONDITION VARIABLE expected" name args))

(define (##trap-check-condvar* name . args)
  (##runtime-error* "CONDITION VARIABLE expected" name args))

(define ##make-condvar
  (let ()

    (##declare (not inline) (not interrupts-enabled))

    (lambda (name)
      (macro-make-condvar name))))

(define ##condvar-signal!
  (let ()

    (##declare (not inline) (not interrupts-enabled))

    (lambda (condvar broadcast?)
      (##condvar-signal-no-reschedule! condvar broadcast?)
      (macro-thread-reschedule-if-needed!))))

(define ##condvar-signal-no-reschedule!
  (let ()

    (##declare (not inline) (not interrupts-enabled))

    (lambda (condvar broadcast?)
      (let loop ()
        (let ((leftmost (macro-btq-leftmost condvar)))
          (if (##not (##eq? leftmost condvar))
            (begin
              (macro-thread-result-set!
               leftmost
               ##thread-signaled-condvar-action!)
              (##thread-btq-remove! leftmost)
              (macro-thread-toq-remove-if-in-toq! leftmost)
              (##btq-insert! (macro-run-queue) leftmost)
;;;;;;(set! pcase14 (##cons leftmost pcase14));;;;;;;;;;;;;;;;;;;;;;;
              (if broadcast?
                (loop)
                (##void)))
            (##void)))))))

(define ##condvar-block-on-device!
  (let ()

    (##declare (not inline) (not interrupts-enabled))

    (lambda (condvar timeout)
      (let loop ()
        (let ((result
               (if timeout
                 (macro-thread-save!
                  (lambda (current-thread condvar timeout dummy1)
                    (##btq-remove! current-thread)
                    (macro-thread-boost-and-clear-quantum-used! current-thread)
                    (##thread-btq-insert! condvar current-thread)
                    (if (##not (##eq? timeout #t))
                      (begin
                        (macro-thread-timeout-set!
                         current-thread
                         timeout)
                        (##toq-insert!
                         (macro-run-queue)
                         current-thread)))
                    (macro-btq-deq-remove! condvar)
                    (macro-btq-deq-insert! (macro-run-queue) condvar)
                    (##thread-schedule!))
                  condvar
                  timeout
                  #f)
                 #f)))
          (if (##eq? result (##void))
            (loop)
            result))))))

;------------------------------------------------------------------------------

; Implementation of thread groups.

(define ##tgroup-tag
  '#(thread-group -1))

(define (##trap-check-tgroup name . args)
  (##runtime-error "THREAD GROUP expected" name args))

(define (##trap-check-tgroup* name . args)
  (##runtime-error* "THREAD GROUP expected" name args))

(define ##make-tgroup
  (let ()

    (##declare (not inline) (not interrupts-enabled))

    (lambda (name parent)
      (macro-make-tgroup name parent))))

(define ##tgroup-suspend!
  (let ()

    (##declare (not inline) (not interrupts-enabled))

    (lambda (tgroup)
      (macro-not-yet-implemented))))

(define ##tgroup-resume!
  (let ()

    (##declare (not inline) (not interrupts-enabled))

    (lambda (tgroup)
      (macro-not-yet-implemented))))

(define ##tgroup-terminate!
  (let ()

    (##declare (not inline) (not interrupts-enabled))

    (lambda (tgroup)
      (macro-not-yet-implemented))))

;------------------------------------------------------------------------------

; User accessible primitives for time objects.

(define current-time
  (let ()

    (##declare (not inline))

    (lambda ()
      (macro-update-current-time!)
      (let ((seconds
             (macro-current-time
              (macro-thread-floats (macro-run-queue)))))
        (macro-make-time seconds)))))

(define time->seconds
  (let ()

    (##declare (not inline))

    (lambda (t)
      (macro-force-vars (t)
        (macro-check-time t (time->seconds t)
          (macro-time-seconds t))))))

(define seconds->time
  (let ()

    (##declare (not inline))

    (lambda (s)
      (macro-force-vars (s)
        (macro-check-real s (seconds->time s)
          (macro-make-time (macro-real->inexact s)))))))

;------------------------------------------------------------------------------

; User accessible primitives for threads.

(define current-thread
  (let ()

    (##declare (not inline))

    (lambda ()
      (macro-current-thread))))

(define thread?
  (let ()

    (##declare (not inline))

    (lambda (obj)
      (macro-force-vars (obj)
        (macro-thread? obj)))))

(define make-thread
  (let ()

    (##declare (not inline))

    (lambda (thunk #!optional (n (macro-absent-obj)) (tg (macro-absent-obj)))
      (macro-force-vars (thunk n tg)
        (let* ((name
                (if (##eq? n (macro-absent-obj))
                  (##void)
                  n))
               (tgroup
                (if (##eq? tg (macro-absent-obj))
                  (macro-thread-tgroup (macro-current-thread))
                  tg)))
          (macro-check-procedure thunk (make-thread thunk n tg)
            (macro-check-tgroup tgroup (make-thread thunk n tg)
              (macro-make-thread thunk name tgroup))))))))

(define thread-name
  (let ()

    (##declare (not inline))

    (lambda (thread)
      (macro-force-vars (thread)
        (macro-check-thread thread (thread-name thread)
          (macro-thread-name thread))))))

(define thread-thread-group
  (let ()

    (##declare (not inline))

    (lambda (thread)
      (macro-force-vars (thread)
        (macro-check-thread thread (thread-thread-group thread)
          (macro-thread-tgroup thread))))))

(define thread-specific
  (let ()

    (##declare (not inline))

    (lambda (thread)
      (macro-force-vars (thread)
        (macro-check-thread thread (thread-specific thread)
          (macro-thread-specific thread))))))

(define thread-specific-set!
  (let ()

    (##declare (not inline))

    (lambda (thread obj)
      (macro-force-vars (thread)
        (macro-check-thread thread (thread-specific-set! thread obj)
          (macro-thread-specific-set! thread obj))))))

(define thread-base-priority
  (let ()

    (##declare (not inline))

    (lambda (thread)
      (macro-force-vars (thread)
        (macro-check-thread thread (thread-base-priority thread)
          (macro-thread-base-priority thread))))))

(define thread-base-priority-set!
  (let ()

    (##declare (not inline))

    (lambda (thread base-priority)
      (macro-force-vars (thread base-priority)
        (macro-check-thread
         thread
         (thread-base-priority-set! thread base-priority)
         (macro-check-real
          base-priority
          (thread-base-priority-set! thread base-priority)
          (##thread-base-priority-set!
           thread
           (macro-real->inexact base-priority))))))))

(define thread-quantum
  (let ()

    (##declare (not inline))

    (lambda (thread)
      (macro-force-vars (thread)
        (macro-check-thread thread (thread-quantum thread)
          (macro-thread-quantum thread))))))

(define thread-quantum-set!
  (let ()

    (##declare (not inline))

    (lambda (thread quantum)
      (macro-force-vars (thread quantum)
        (macro-check-thread thread (thread-quantum-set! thread quantum)
          (macro-check-nonnegative-real
           quantum
           (thread-quantum-set! thread quantum)
           (##thread-quantum-set!
            thread
            (macro-real->inexact quantum))))))))

(define thread-priority-boost
  (let ()

    (##declare (not inline))

    (lambda (thread)
      (macro-force-vars (thread)
        (macro-check-thread thread (thread-priority-boost thread)
          (macro-thread-priority-boost thread))))))

(define thread-priority-boost-set!
  (let ()

    (##declare (not inline))

    (lambda (thread priority-boost)
      (macro-force-vars (thread priority-boost)
        (macro-check-thread
         thread
         (thread-priority-boost-set! thread priority-boost)
         (macro-check-nonnegative-real
          priority-boost
          (thread-priority-boost-set! thread priority-boost)
          (##thread-priority-boost-set!
           thread
           (macro-real->inexact priority-boost))))))))

(define thread-start!
  (let ()

    (##declare (not inline))

    (lambda (thread)
      (macro-force-vars (thread)
        (macro-check-thread thread (thread-start! thread)
          (##thread-start! thread))))))

(define thread-yield!
  (let ()

    (##declare (not inline))

    (lambda ()
      (##thread-yield!))))

(define thread-sleep!
  (let ()

    (##declare (not inline))

    (lambda (absrel-timeout)
      (macro-force-vars (absrel-timeout)
        (macro-check-absrel-time absrel-timeout (thread-sleep! absrel-timeout)
          (##thread-sleep! absrel-timeout))))))

(define thread-suspend!
  (let ()

    (##declare (not inline))

    (lambda (thread)
      (macro-force-vars (thread)
        (macro-check-thread thread (thread-suspend! thread)
          (##thread-suspend! thread))))))

(define thread-resume!
  (let ()

    (##declare (not inline))

    (lambda (thread)
      (macro-force-vars (thread)
        (macro-check-thread thread (thread-resume! thread)
          (##thread-resume! thread))))))

(define thread-terminate!
  (let ()

    (##declare (not inline))

    (lambda (thread)
      (macro-force-vars (thread)
        (macro-check-thread thread (thread-terminate! thread)
          (##thread-terminate! thread))))))

(define thread-join!
  (let ()

    (##declare (not inline))

    (lambda (thread
             #!optional
             (absrel-timeout (macro-absent-obj))
             (timeout-val (macro-absent-obj)))
      (macro-force-vars (thread absrel-timeout timeout-val)
        (macro-check-thread
         thread
         (thread-join! thread absrel-timeout timeout-val)
          (if (or (##eq? absrel-timeout (macro-absent-obj))
                  (##not absrel-timeout))
            (##thread-join! thread #f #f)
            (macro-check-absrel-time
             absrel-timeout
             (thread-join! thread absrel-timeout timeout-val)
             (##thread-join! thread absrel-timeout timeout-val))))))))

; User accessible primitives for mutexes.

(define mutex?
  (let ()

    (##declare (not inline))

    (lambda (obj)
      (macro-force-vars (obj)
        (macro-mutex? obj)))))

(define make-mutex
  (let ()

    (##declare (not inline))

    (lambda (#!optional (n (macro-absent-obj)))
      (macro-force-vars (n)
        (let ((name
               (if (##eq? n (macro-absent-obj))
                 (##void)
                 n)))
          (macro-make-mutex name))))))

(define mutex-name
  (let ()

    (##declare (not inline))

    (lambda (mutex)
      (macro-force-vars (mutex)
        (macro-check-mutex mutex (mutex-name mutex)
          (macro-mutex-name mutex))))))

(define mutex-state
  (let ()

    (##declare (not inline))

    (lambda (mutex)
      (macro-force-vars (mutex)
        (macro-check-mutex mutex (mutex-state mutex)
          (or (macro-btq-owner mutex)
              'not-abandoned))))))

(define mutex-lock!
  (let ()

    (##declare (not inline))

    (lambda (mutex
             #!optional
             (absrel-timeout (macro-absent-obj))
             (thread (macro-absent-obj)))
      (macro-force-vars (mutex absrel-timeout thread)
        (macro-check-mutex mutex (mutex-lock! mutex absrel-timeout thread)
          (cond ((##eq? absrel-timeout (macro-absent-obj))
                 (macro-mutex-lock! mutex #f (macro-current-thread)))
                ((##not absrel-timeout)
                 (cond ((##eq? thread (macro-absent-obj))
                        (macro-mutex-lock! mutex #f (macro-current-thread)))
                       ((##not thread)
                        (macro-mutex-lock-anonymously! mutex absrel-timeout))
                       (else
                        (macro-check-thread
                         thread
                         (mutex-lock! mutex absrel-timeout thread)
                         (macro-mutex-lock! mutex absrel-timeout thread)))))
                (else
                 (macro-check-absrel-time
                  absrel-timeout
                  (mutex-lock! mutex absrel-timeout thread)
                  (cond ((##eq? thread (macro-absent-obj))
                         (macro-mutex-lock!
                          mutex
                          absrel-timeout
                          (macro-current-thread)))
                        ((##not thread)
                         (macro-mutex-lock-anonymously!
                          mutex
                          absrel-timeout))
                        (else
                         (macro-check-thread
                          thread
                          (mutex-lock! mutex absrel-timeout thread)
                          (macro-mutex-lock!
                           mutex
                           absrel-timeout
                           thread))))))))))))

(define mutex-unlock!
  (let ()

    (##declare (not inline))

    (lambda (mutex
             #!optional
             (condvar (macro-absent-obj))
             (absrel-timeout (macro-absent-obj)))
      (macro-force-vars (mutex condvar absrel-timeout)
        (macro-check-mutex mutex (mutex-unlock! mutex condvar absrel-timeout)
          (if (##eq? condvar (macro-absent-obj))
            (macro-mutex-unlock! mutex)
            (macro-check-condvar
             condvar
             (mutex-unlock! mutex condvar absrel-timeout)
             (if (##eq? absrel-timeout (macro-absent-obj))
               (##mutex-signal-and-condvar-wait! mutex condvar #t)
               (macro-check-absrel-time
                absrel-timeout
                (mutex-unlock! mutex condvar absrel-timeout)
                (let ((timeout (##absrel-timeout->timeout absrel-timeout)))
                  (##mutex-signal-and-condvar-wait!
                   mutex
                   condvar
                   timeout)))))))))))

; User accessible primitives for condition variables

(define condition-variable?
  (let ()

    (##declare (not inline))

    (lambda (obj)
      (macro-force-vars (obj)
        (macro-condvar? obj)))))

(define make-condition-variable
  (let ()

    (##declare (not inline))

    (lambda (#!optional (n (macro-absent-obj)))
      (macro-force-vars (n)
        (let ((name
               (if (##eq? n (macro-absent-obj))
                 (##void)
                 n)))
          (macro-make-condvar name))))))

(define condition-variable-name
  (let ()

    (##declare (not inline))

    (lambda (condvar)
      (macro-force-vars (condvar)
        (macro-check-condvar condvar (condition-variable-name condvar)
          (macro-condvar-name condvar))))))

(define condition-variable-signal!
  (let ()

    (##declare (not inline))

    (lambda (condvar)
      (macro-force-vars (condvar)
        (macro-check-condvar condvar (condition-variable-signal! condvar)
          (##condvar-signal! condvar #f))))))

(define condition-variable-broadcast!
  (let ()

    (##declare (not inline))

    (lambda (condvar)
      (macro-force-vars (condvar)
        (macro-check-condvar condvar (condition-variable-broadcast! condvar)
          (##condvar-signal! condvar #t))))))

; User accessible primitives for thread groups.

(define thread-group?
  (let ()

    (##declare (not inline))

    (lambda (obj)
      (macro-force-vars (obj)
        (macro-tgroup? obj)))))

(define make-thread-group
  (let ()

    (##declare (not inline))

    (lambda (#!optional (n (macro-absent-obj)) (p (macro-absent-obj)))
      (macro-force-vars (n p)
        (let ((name
               (if (##eq? n (macro-absent-obj))
                 (##void)
                 n)))
          (cond ((##eq? p (macro-absent-obj))
                 (##make-tgroup
                  name
                  (macro-thread-tgroup (macro-current-thread))))
                (p
                 (macro-check-tgroup p (make-thread-group n p)
                   (##make-tgroup name p)))
                (else
                 (##make-tgroup name #f))))))))

(define thread-group-name
  (let ()

    (##declare (not inline))

    (lambda (tgroup)
      (macro-force-vars (tgroup)
        (macro-check-tgroup tgroup (thread-group-name tgroup)
          (macro-tgroup-name tgroup))))))

(define thread-group-parent
  (let ()

    (##declare (not inline))

    (lambda (tgroup)
      (macro-force-vars (tgroup)
        (macro-check-tgroup tgroup (thread-group-parent tgroup)
          (macro-tgroup-parent tgroup))))))

(define thread-group-suspend!
  (let ()

    (##declare (not inline))

    (lambda (tgroup)
      (macro-force-vars (tgroup)
        (macro-check-tgroup tgroup (thread-group-suspend! tgroup)
          (##tgroup-suspend! tgroup))))))

(define thread-group-resume!
  (let ()

    (##declare (not inline))

    (lambda (tgroup)
      (macro-force-vars (tgroup)
        (macro-check-tgroup tgroup (thread-group-resume! tgroup)
          (##tgroup-resume! tgroup))))))

(define thread-group-terminate!
  (let ()

    (##declare (not inline))

    (lambda (tgroup)
      (macro-force-vars (tgroup)
        (macro-check-tgroup tgroup (thread-group-terminate! tgroup)
          (##tgroup-terminate! tgroup))))))

;------------------------------------------------------------------------------

; User accessible primitives for exception handling.

(define current-exception-handler
  (let ()

    (##declare (not inline))

    (lambda ()
      (macro-current-exception-handler))))

(define with-exception-handler
  (let ()

    (##declare (not inline))

    (lambda (handler thunk)
      (macro-force-vars (handler thunk)
        (macro-check-procedure handler (with-exception-handler handler thunk)
          (macro-check-procedure thunk (with-exception-handler handler thunk)
            (macro-dynamic-bind exception-handler handler thunk)))))))

(define raise
  (let ()

    (##declare (not inline))

    (lambda (obj)
      (macro-raise obj))))

(define abort
  (let ()

    (##declare (not inline))

    (lambda (obj)
      (macro-abort obj))))

; User accessible primitives for thread related exceptions.

(define deadlock-exception?
  (let ()

    (##declare (not inline))

    (lambda (obj)
      (macro-force-vars (obj)
        (macro-deadlock-exception? obj)))))

(define join-timeout-exception?
  (let ()

    (##declare (not inline))

    (lambda (obj)
      (macro-force-vars (obj)
        (macro-join-timeout-exception? obj)))))

(define abandoned-mutex-exception?
  (let ()

    (##declare (not inline))

    (lambda (obj)
      (macro-force-vars (obj)
        (macro-abandoned-mutex-exception? obj)))))

(define started-thread-exception?
  (let ()

    (##declare (not inline))

    (lambda (obj)
      (macro-force-vars (obj)
        (macro-started-thread-exception? obj)))))

(define terminated-thread-exception?
  (let ()

    (##declare (not inline))

    (lambda (obj)
      (macro-force-vars (obj)
        (macro-terminated-thread-exception? obj)))))

(define uncaught-exception?
  (let ()

    (##declare (not inline))

    (lambda (obj)
      (macro-force-vars (obj)
        (macro-uncaught-exception? obj)))))

(define uncaught-exception-reason
  (let ()

    (##declare (not inline))

    (lambda (exc)
      (macro-force-vars (exc)
        (macro-check-uncaught-exception
         exc
         (uncaught-exception-reason exc)
         (macro-uncaught-exception-reason exc))))))

(define scheduler-exception?
  (let ()

    (##declare (not inline))

    (lambda (obj)
      (macro-force-vars (obj)
        (macro-scheduler-exception? obj)))))

(define scheduler-exception-reason
  (let ()

    (##declare (not inline))

    (lambda (exc)
      (macro-force-vars (exc)
        (macro-check-scheduler-exception
         exc
         (scheduler-exception-reason exc)
         (macro-scheduler-exception-reason exc))))))

(define noncontinuable-exception?
  (let ()

    (##declare (not inline))

    (lambda (obj)
      (macro-force-vars (obj)
        (macro-noncontinuable-exception? obj)))))

(define noncontinuable-exception-reason
  (let ()

    (##declare (not inline))

    (lambda (exc)
      (macro-force-vars (exc)
        (macro-check-noncontinuable-exception
         exc
         (noncontinuable-exception-reason exc)
         (macro-noncontinuable-exception-reason exc))))))

(define unbound-dynamic-variable-exception?
  (let ()

    (##declare (not inline))

    (lambda (obj)
      (macro-force-vars (obj)
        (macro-unbound-dynamic-variable-exception? obj)))))

(define unbound-dynamic-variable-exception-name
  (let ()

    (##declare (not inline))

    (lambda (exc)
      (macro-force-vars (exc)
        (macro-check-unbound-dynamic-variable-exception
         exc
         (unbound-dynamic-variable-exception-name exc)
         (macro-unbound-dynamic-variable-exception-name exc))))))

(define low-level-exception?
  (let ()

    (##declare (not inline))

    (lambda (obj)
      (macro-force-vars (obj)
        (macro-low-level-exception? obj)))))

(define low-level-exception-err-code
  (let ()

    (##declare (not inline))

    (lambda (exc)
      (macro-force-vars (exc)
        (macro-check-low-level-exception
         exc
         (low-level-exception-err-code exc)
         (macro-low-level-exception-err-code exc))))))

;------------------------------------------------------------------------------

(define thread-boosted-priority;;;;;;;;;;;;;;;;;;;;temporary
  (let ()

    (##declare (not inline))

    (lambda (thread)
      (macro-force-vars (thread)
        (macro-check-thread thread (thread-boosted-priority thread)
          (macro-thread-boosted-priority thread))))))

(define thread-effective-priority;;;;;;;;;;;;;;;;;;;;;temporary
  (let ()

    (##declare (not inline))

    (lambda (thread)
      (macro-force-vars (thread)
        (macro-check-thread thread (thread-effective-priority thread)
          (macro-thread-effective-priority thread))))))


(define case1 0);;;;;;;;;;;;;;;;
(define case2 0)
(define case3 0)
(define dcase1 0)
(define dcase2 0)
(define dcase3 0)
(define dcase4 0)
(define ddcase1 0)
(define ddcase2 0)
(define ddcase3 0)

(define pcase1 0);;;;;;;;;;;;;;;;
(define pcase2 0)
(define pcase3 0)
(define pcase4 0)
(define pcase5 0)
(define pcase6 0)
(define pcase7 0)
(define pcase8 0)
(define pcase9 0)
(define pcase10 0)
(define pcase11 0)
(define pcase12 0)
(define pcase13 0)
(define pcase14 '())

(define (thread-stats)
  (list case1 case2 case3
        dcase1 dcase2 dcase3 dcase4
        ddcase1 ddcase2 ddcase3))

(define (thread-pstats)
  (list pcase1 pcase2 pcase3 pcase4 pcase5 pcase6 pcase7 pcase8 pcase9 pcase10 pcase11 pcase12 pcase13 pcase14))

(define (thread-pstats-reset!)
  (set! pcase1 0);;;;;;;;;;;;;;;;
  (set! pcase2 0)
  (set! pcase3 0)
  (set! pcase4 0)
  (set! pcase5 0)
  (set! pcase6 0)
  (set! pcase7 0)
  (set! pcase8 0)
  (set! pcase9 0)
  (set! pcase10 0)
  (set! pcase11 0)
  (set! pcase12 0)
  (set! pcase13 0)
  (set! pcase14 '())
)

(##enable-interrupts!)

(##thread-startup!)

(##thread-heartbeat-interval-set! (macro-inexact-+0))
