;==============================================================================

; file: "header.scm"

; Copyright (C) 1994-2000 by Marc Feeley, All Rights Reserved.

;==============================================================================

(##namespace (""
   not boolean? eqv? eq? equal? pair? cons car cdr set-car! set-cdr!
   caar cadr cdar cddr caaar caadr cadar caddr cdaar cdadr cddar cdddr
   caaaar caaadr caadar caaddr cadaar cadadr caddar cadddr cdaaar cdaadr
   cdadar cdaddr cddaar cddadr cdddar cddddr null? list? list length
   append reverse list-ref memq memv member assq assv assoc symbol?
   symbol->string string->symbol number? complex? real? rational?
   integer? exact? inexact? = < > <= >= zero? positive? negative? odd?
   even? max min + * - / abs quotient remainder modulo gcd lcm numerator
   denominator floor ceiling truncate round rationalize exp log sin cos
   tan asin acos atan sqrt expt make-rectangular make-polar real-part
   imag-part magnitude angle exact->inexact inexact->exact number->string
   string->number char? char=? char<? char>? char<=? char>=? char-ci=?
   char-ci<? char-ci>? char-ci<=? char-ci>=? char-alphabetic?
   char-numeric? char-whitespace? char-upper-case? char-lower-case?
   char->integer integer->char char-upcase char-downcase string?
   make-string string string-length string-ref string-set! string=?
   string<? string>? string<=? string>=? string-ci=? string-ci<?
   string-ci>? string-ci<=? string-ci>=? substring string-append vector?
   make-vector vector vector-length vector-ref vector-set! procedure?
   apply map for-each call-with-current-continuation call-with-input-file
   call-with-output-file input-port? output-port? current-input-port
   current-output-port open-input-file open-output-file close-input-port
   close-output-port eof-object? read read-char peek-char write display
   newline write-char

   eval error pp compile-file-to-c compile-file link-incremental link-flat

   script-arguments

   include
))

(##declare
  (multilisp)
  (extended-bindings)
  (not safe)
  (block)
  (fixnum)
  (inlining-limit 134)
)

;------------------------------------------------------------------------------

; General object representation.

; Type tags.

(##define-macro (macro-type-fixnum)   0)
(##define-macro (macro-type-subtyped) 1)
(##define-macro (macro-type-special)  2)
(##define-macro (macro-type-pair)     3)

; Subtype tags.

(##define-macro (macro-subtype-vector)       0)
(##define-macro (macro-subtype-pair)         1)
(##define-macro (macro-subtype-ratnum)       2)
(##define-macro (macro-subtype-cpxnum)       3)
(##define-macro (macro-subtype-structure)    4)
(##define-macro (macro-subtype-meroon)       6)

(##define-macro (macro-subtype-symbol)       8)
(##define-macro (macro-subtype-keyword)      9)
(##define-macro (macro-subtype-frame)        10)
(##define-macro (macro-subtype-continuation) 11)
(##define-macro (macro-subtype-promise)      12)
(##define-macro (macro-subtype-will)         13)
(##define-macro (macro-subtype-procedure)    14)
(##define-macro (macro-subtype-return)       15)
(##define-macro (macro-subtype-pointer)      16)

(##define-macro (macro-subtype-string)       18)
(##define-macro (macro-subtype-s8vector)     20)
(##define-macro (macro-subtype-u8vector)     21)
(##define-macro (macro-subtype-s16vector)    22)
(##define-macro (macro-subtype-u16vector)    23)
(##define-macro (macro-subtype-s32vector)    24)
(##define-macro (macro-subtype-u32vector)    25)
(##define-macro (macro-subtype-f32vector)    26)

; for alignment these 5 must be last:
(##define-macro (macro-subtype-s64vector)    27)
(##define-macro (macro-subtype-u64vector)    28)
(##define-macro (macro-subtype-f64vector)    29)
(##define-macro (macro-subtype-flonum)       30)
(##define-macro (macro-subtype-bignum)       31)

(##define-macro (macro-subtype-ovector? x) `(##fixnum.< ,x 8))
(##define-macro (macro-subtype-bvector? x) `(##fixnum.< 16 ,x))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

; Special objects.

(##define-macro (macro-absent-obj)
  (##namespace ("c#" absent-object))
  `',absent-object)

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

; Structure objects.

(##define-macro (macro-structure-object? obj tag)
  `(let ((obj ,obj))
     (and (##structure? obj)
          (##eq? (##vector-ref obj 0)
                 ,tag))))

(##define-macro (macro-make-structure-object tag . slots)
  `(##subtype-set!
    (##vector ,tag ,@slots)
    (macro-subtype-structure)))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

; System procedure classes.

(##define-macro (define-system form . exprs)

  (define inlinable-procs '(

##type ##type-cast ##subtype ##subtype-set!
##not ##null? ##unbound? ##eq? ##eof-object?
##fixnum? ##flonum? ##special? ##pair? ##subtyped? ##subtyped?
##subtyped.vector? ##subtyped.symbol? ##subtyped.flonum? ##subtyped.bignum?
##procedure? ##promise? ##vector? ##symbol? ##keyword? ##ratnum? ##cpxnum?
##string? ##structure? ##bignum?
##char? ;;;;  ##closure? ##subprocedure?
##fixnum.max ##fixnum.min
##fixnum.+ ##fixnum.- ##fixnum.*
##fixnum.quotient ##fixnum.remainder ##fixnum.modulo
##fixnum.logior ##fixnum.logxor ##fixnum.logand ##fixnum.lognot
##fixnum.ashr ##fixnum.lshr ##fixnum.shl
##fixnum.zero? ##fixnum.positive? ##fixnum.negative?
##fixnum.odd? ##fixnum.even?
##fixnum.= ##fixnum.< ##fixnum.> ##fixnum.<= ##fixnum.>=
##fixnum.->char ##fixnum.<-char
##flonum.->fixnum ##flonum.<-fixnum
##flonum.max ##flonum.min
##flonum.+ ##flonum.- ##flonum.*  ##flonum./
##flonum.abs ##flonum.floor ##flonum.ceiling ##flonum.round
##flonum.exp ##flonum.log
##flonum.sin ##flonum.cos ##flonum.tan
##flonum.asin ##flonum.acos ##flonum.atan
##flonum.sqrt ##flonum.copysign
##flonum.zero? ##flonum.positive? ##flonum.negative?
##flonum.finite? ##flonum.nan? ##flonum.<-fixnum-exact?
##flonum.= ##flonum.< ##flonum.> ##flonum.<= ##flonum.>=
##char=? ##char<? ##char>? ##char<=? ##char>=?
##cons ##set-car! ##set-cdr! ##car ##cdr
##caar ##cadr ##cdar ##cddr
##caaar ##caadr ##cadar ##caddr ##cdaar ##cdadr ##cddar ##cdddr
##caaaar ##caaadr ##caadar ##caaddr ##cadaar ##cadadr ##caddar ##cadddr
##cdaaar ##cdaadr ##cdadar ##cdaddr ##cddaar ##cddadr ##cdddar ##cddddr
##list
##make-cell ##cell-ref ##cell-set!
##vector
##vector-length ##vector-ref ##vector-set! ##vector-shrink!
##string
##string-length ##string-ref ##string-set! ##string-shrink!
##s8vector? ##s8vector
##s8vector-length ##s8vector-ref ##s8vector-set! ##s8vector-shrink!
##u8vector? ##u8vector
##u8vector-length ##u8vector-ref ##u8vector-set! ##u8vector-shrink!
##s16vector? ##s16vector
##s16vector-length ##s16vector-ref ##s16vector-set! ##s16vector-shrink!
##u16vector? ##u16vector
##u16vector-length ##u16vector-ref ##u16vector-set! ##u16vector-shrink!
##s32vector? ##s32vector
##s32vector-length ##s32vector-ref ##s32vector-set! ##s32vector-shrink!
##u32vector? ##u32vector
##u32vector-length ##u32vector-ref ##u32vector-set! ##u32vector-shrink!
##s64vector? ##s64vector
##s64vector-length ##s64vector-ref ##s64vector-set! ##s64vector-shrink!
##u64vector? ##u64vector
##u64vector-length ##u64vector-ref ##u64vector-set! ##u64vector-shrink!
##f32vector? ##f32vector
##f32vector-length ##f32vector-ref ##f32vector-set! ##f32vector-shrink!
##f64vector? ##f64vector
##f64vector-length ##f64vector-ref ##f64vector-set! ##f64vector-shrink!
##symbol->string ##keyword->string
##closure-code ##closure-ref ##closure-set!
;;; ##subprocedure-id ##subprocedure-parent
;;; ##procedure-info
##make-promise ##force ##void

##will? ##make-will ##will-testator

##global-var-ref
##global-var-set!

##bignum.negative?
##bignum.adigit-length
##bignum.adigit-set-ones!
##bignum.adigit-inc!
##bignum.adigit-dec!
##bignum.adigit-add!
##bignum.adigit-sub!
##bignum.mdigit-length
##bignum.mdigit-ref
##bignum.mdigit-set!
##bignum.mdigit-mul!
##bignum.mdigit-div!
##bignum.mdigit-quotient
##bignum.mdigit-remainder
##bignum.mdigit-test?

##bignum.adigit-ones?
##bignum.adigit-zero?
##bignum.adigit-negative?
##bignum.adigit-=
##bignum.adigit-<
##bignum.->fixnum
##bignum.<-fixnum
##bignum.adigit-shrink!
##bignum.adigit-set!
##bignum.adigit-cat!
##bignum.adigit-logand!
##bignum.adigit-logior!
##bignum.adigit-logxor!
##bignum.adigit-lognot!

##bignum.fdigit-length
##bignum.fdigit-ref
##bignum.fdigit-set!

))

  (define kernel-procs '(

##make-string
##make-vector
##make-s8vector
##make-u8vector
##make-s16vector
##make-u16vector
##make-s32vector
##make-u32vector
##make-s64vector
##make-u64vector
##make-f32vector
##make-f64vector
##apply
##call-with-current-continuation
##make-global-var


##closure? ##subprocedure?

##subprocedure-id ##subprocedure-parent
##procedure-info

))

  (if (memq (car form) kernel-procs)
    `(begin)
    (if (and (memq (car form) inlinable-procs)
             (list? (cdr form)))
      `(define ,form ,form)
      `(define ,form ,@exprs))))

;------------------------------------------------------------------------------

; Macro to define structure accessors.

(##define-macro (macro-slot index struct . val)
  (if (null? val)
    `(##vector-ref ,struct ,index)
    `(##vector-set! ,struct ,index ,@val)))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

; Symbol objects

; A symbol is represented by an object vector of length 3
; slot 0 = symbol name (string)
; slot 1 = hash code (non-negative fixnum)
; slot 2 = pointer to corresponding global variable (0 if none exists)

(##define-macro (macro-make-uninterned-symbol name hash)
  `(##subtype-set!
    (##vector ,name ,hash #f 0)
    (macro-subtype-symbol)))

(##define-macro (macro-symbol-name s)        `(macro-slot 0 ,s))
(##define-macro (macro-symbol-name-set! s x) `(macro-slot 0 ,s ,x))
(##define-macro (macro-symbol-hash s)        `(macro-slot 1 ,s))
(##define-macro (macro-symbol-hash-set! s x) `(macro-slot 1 ,s ,x))
(##define-macro (macro-symbol-next s)        `(macro-slot 2 ,s))
(##define-macro (macro-symbol-next-set! s x) `(macro-slot 2 ,s ,x))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

; Keyword objects

; A keyword is represented by an object vector of length 2
; slot 0 = symbol name (string) not including trailing ':'
; slot 1 = hash code (non-negative fixnum)

(##define-macro (macro-make-uninterned-keyword name hash)
  `(##subtype-set!
    (##vector ,name ,hash #f)
    (macro-subtype-keyword)))

(##define-macro (macro-keyword-name k)        `(macro-slot 0 ,k))
(##define-macro (macro-keyword-name-set! k x) `(macro-slot 0 ,k ,x))
(##define-macro (macro-keyword-hash k)        `(macro-slot 1 ,k))
(##define-macro (macro-keyword-hash-set! k x) `(macro-slot 1 ,k ,x))
(##define-macro (macro-keyword-next k)        `(macro-slot 2 ,k))
(##define-macro (macro-keyword-next-set! k x) `(macro-slot 2 ,k ,x))

;------------------------------------------------------------------------------

(##define-macro (macro-if-forces forces noforces)
  (if (memq 'force ##compilation-options)
    forces
    noforces))

(##define-macro (macro-force-vars vars expr)
  (if (memq 'force ##compilation-options)
    `(let ,(map (lambda (x) `(,x (##force ,x))) vars) ,expr)
    expr))

(##define-macro (macro-if-checks checks nochecks)
  (if (memq 'check ##compilation-options)
    checks
    nochecks))

(##define-macro (macro-no-force vars expr)
  expr)

(##define-macro (macro-no-check var form expr)
  expr)

(##define-macro
  (define-fold bool? form zero one two forcing pre-check . post-checks)
  (let* ((name-fn (car form))
         (name-param1 (cadr form))
         (name-param2 (caddr form))
         (name-others 'others)
         (name-result 'result)
         (name-folded-result 'folded-result)
         (param-name-map
          (list (cons name-param1 'param1) (cons name-param2 'param2)))
         (two-wrap
          (if (pair? (cadr two)) (car two) #f))
         (two-call
          (if (pair? (cadr two)) (cadr two) two)))

    (define (param-name sym)
      (cdr (assq sym param-name-map)))

    (define (rewrite expr)
      (let ((x (assq expr param-name-map)))
        (if x
          (cdr x)
          (if (pair? expr)
            (cons (car expr)
                  (map rewrite (cdr expr)))
            expr))))

    (define (parameter-list)
      (if (null? zero)
        ; 1 or more arguments
        (list (param-name name-param1)
              '#!optional
              (list (param-name name-param2) '(macro-absent-obj))
              '#!rest
              name-others)
        ; 0 or more arguments
        (list '#!optional
              (list (param-name name-param1) '(macro-absent-obj))
              (list (param-name name-param2) '(macro-absent-obj))
              '#!rest
              name-others)))

    (define (add-post-checks wrap expr cont)

      (define (add-wrap x)
        (if wrap
          (list wrap x)
          x))

      (define (inside lst)
        (if (null? lst)
          (if cont
            (cont (add-wrap name-result))
            (add-wrap name-result))
          (let ((check (car lst)))
            (list 'if
                  (list (car check) name-result)
                  (list (cadr check)
                        (list 'quote name-fn)
                        (param-name name-param1)
                        (param-name name-param2)
                        name-others)
                  (inside (cdr lst))))))

      (if (or cont (not (null? post-checks)))
        (list 'let
              (list (list name-result expr))
              (inside post-checks))
        (add-wrap expr)))

    (define (add-forcing names expr)
      (list forcing
            names
            expr))

    (define (add-pre-check name expr)
      (list pre-check
            name
            (cons name-fn
                  (cons (param-name name-param1)
                        (cons (param-name name-param2)
                              name-others)))
            expr))

    (define (exactly-1-arg)
      (add-forcing
       (list (param-name name-param1))
       (add-pre-check
        (param-name name-param1)
        (if (pair? one)
          (add-post-checks
           #f
           (rewrite one)
           #f)
          (rewrite one)))))

    (define (exactly-2-args)
      (add-post-checks
       two-wrap
       (rewrite two-call)
       #f))

    (define (at-least-2-args)
      (list 'let
            'loop
            (append (if bool?
                      (list (list name-folded-result #t))
                      '())
                    (list (list name-param1 (param-name name-param1))
                          (list name-param2 (param-name name-param2))
                          (list 'lst name-others)))
            (add-post-checks
             two-wrap
             two-call
             (lambda (result)
               (list 'if
                     '(##null? lst)
                     (if bool?
                       (list 'and
                             result
                             name-folded-result)
                       result)
                     (list 'let
                           '((next (##car lst)))
                           (add-forcing
                            '(next)
                            (add-pre-check
                             'next
                             (if bool?
                               (list 'loop
                                     (list 'and
                                           result
                                           name-folded-result)
                                     name-param2
                                     'next
                                     '(##cdr lst))
                               (list 'loop
                                     result
                                     'next
                                     '(##cdr lst)))))))))))

    (define (body)
      (cons 'cond
            (cons (list (list '##not
                              (list '##eq?
                                    (param-name name-param2)
                                    '(macro-absent-obj)))
                        (add-forcing
                         (list (param-name name-param1)
                               (param-name name-param2))
                         (add-pre-check
                          (param-name name-param1)
                          (add-pre-check
                           (param-name name-param2)
;                           (list 'if
;                                 (list '##null? name-others)
;                                 (exactly-2-args)
                                 (at-least-2-args)))))
;)
                  (append (if (null? zero)
                            '()
                            (list (list (list '##eq?
                                              (param-name name-param1)
                                              '(macro-absent-obj))
                                        zero)))
                          (list (list 'else
                                      (exactly-1-arg)))))))

    (list 'define
          name-fn
          (list 'lambda
                (parameter-list)
                (body)))))

(##define-macro
  (define-nary form zero one two forcing pre-check . post-checks)
  `(define-fold #f ,form ,zero ,one ,two ,forcing ,pre-check ,@post-checks))

(##define-macro
  (define-nary-bool form zero one two forcing pre-check . post-checks)
  `(define-fold #t ,form ,zero ,one ,two ,forcing ,pre-check ,@post-checks))

;------------------------------------------------------------------------------

; Macros for "_eval.scm" and "_repl.scm".

(##define-macro (make-global-var id) `(##make-global-var ,id))
(##define-macro (global-var-ref gv) `(##global-var-ref ,gv))
(##define-macro (global-var-set! gv val) `(##global-var-set! ,gv ,val))
(##define-macro (global-var->identifier gv) `(##global-var->identifier ,gv))

(##define-macro (quasi-list->vector x) `(##quasi-list->vector ,x))
(##define-macro (quasi-append x y) `(##quasi-append ,x ,y))
(##define-macro (quasi-cons x y) `(##quasi-cons ,x ,y))

(##define-macro (true? x) x)
(##define-macro (unbound? x) `(##unbound? ,x))

(##define-macro (self-var)     '##self-var)
(##define-macro (selector-var) '##selector-var)
(##define-macro (do-loop-var)  '##do-loop-var)

(##define-macro (rt-error-unbound-global-var code rte var)
  `(##signal '##signal.global-unbound ,code ,rte ,var))

(##define-macro (rt-error-non-procedure-send code rte)
  `(##signal '##signal.non-procedure-send ,code ,rte))

(##define-macro (rt-error-non-procedure-oper code rte)
  `(##signal '##signal.non-procedure-operator ,code ,rte))

(##define-macro (rt-error-wrong-nb-args proc args)
  `(##signal '##signal.wrong-nb-arg ,proc ,args))

(##define-macro (rt-error-unknown-keyword-arg proc args)
  `(##signal '##signal.unknown-keyword-arg ,proc ,args))

(##define-macro (rt-error-keyword-expected proc args)
  `(##signal '##signal.keyword-expected ,proc ,args))

(##define-macro (ct-error-global-env-overflow var)
  `(##signal '##signal.global-env-overflow ,var))

(##define-macro (ct-error-syntax src msg . args)
  `(##signal '##signal.syntax-error ,src ,msg ,@args))

; Macro to create a node of executable code.

(##define-macro (mk-code code-prc cte src stepper subcodes . lst)
  `(let (($code (##vector #f ,code-prc ,cte ,src ,stepper ,@subcodes ,@lst)))
     ,@(let loop ((l subcodes) (i 5) (r '()))
         (if (pair? l)
           (loop (cdr l)
                 (+ i 1)
                 (cons `(##vector-set! (##vector-ref $code ,i) 0 $code) r))
           (reverse r)))
     $code))

(##define-macro (code-link c)         `(##vector-ref ,c 0))
(##define-macro (code-cprc c)         `(##vector-ref ,c 1))
(##define-macro (code-cte c)          `(##vector-ref ,c 2))
(##define-macro (code-locat c)        `(##vector-ref ,c 3))
(##define-macro (code-locat-set! c l) `(##vector-set! ,c 3 ,l))
(##define-macro (code-stepper c)      `(##vector-ref ,c 4))
(##define-macro (code-length c)       `(##fixnum.- (##vector-length ,c) 5))
(##define-macro (code-ref c n)        `(##vector-ref ,c (##fixnum.+ ,n 5)))
(##define-macro (code-set! c n x)     `(##vector-set! ,c (##fixnum.+ ,n 5) ,x))
(##define-macro (^ n)                 `(##vector-ref $code ,(+ n 5)))

(##define-macro (is-child-code? child parent)
  `(let ((child ,child)
         (parent ,parent))
     (and (##vector? child)
          (##fixnum.< 3 (##vector-length child))
          (##eq? (code-link child) parent))))

(##define-macro (code-run c)
  `(let (($$code ,c))
     ((##vector-ref $$code 1) $$code rte)))

; Macro to create the "code procedure" associated with a code node.

(##define-macro (mk-cprc . body)
  `(let ()
     (##declare (not inline) (not interrupts-enabled) (environment-map))
     (lambda ($code rte)
       (let (($$continue
              (lambda ($code rte)
                (##declare (inline))
                ,@body)))
         (##declare (interrupts-enabled))
         ($$continue $code rte)))))

(##define-macro (mk-gen params . def)
  `(let ()
     (##declare (not inline))
     (lambda (cte tail? src ,@params) ,@def)))

(##define-macro (gen proc src . args)
  `(,proc cte tail? ,src ,@args))

(##define-macro (call-step! vars . body)      `(step! #t 1 ,vars ,@body))
(##define-macro (future-step! vars . body)    `(step! #f 2 ,vars ,@body))
(##define-macro (delay-step! vars . body)     `(step! #f 2 ,vars ,@body))
(##define-macro (lambda-step! vars . body)    `(step! #f 3 ,vars ,@body))
(##define-macro (define-step! vars . body)    `(step! #f 4 ,vars ,@body))
(##define-macro (set!-step! vars . body)      `(step! #f 5 ,vars ,@body))
(##define-macro (reference-step! vars . body) `(step! #f 6 ,vars ,@body))
(##define-macro (constant-step! vars . body)  `(step! #f 7 ,vars ,@body))

(##define-macro (make-no-stepper)
  ''#(#(#f #f #f #f #f #f #f) #f #f #f #f #f #f #f))

(##define-macro (make-step-handlers)
  '(##vector ##step-handler
             ##step-handler
             ##step-handler
             ##step-handler
             ##step-handler
             ##step-handler
             ##step-handler))

(##define-macro (make-main-stepper)
  '(##vector (##vector-copy ##step-handlers)
             #f
             #f
             #f
             #f
             #f
             #f
             #f))

(##define-macro (step! leapable? handler-index vars . body)
  `(let* (($$execute-body
           (lambda ($code rte ,@vars) ,@body))
          ($$handler
           (##vector-ref (code-stepper $code) ,handler-index)))
     (if $$handler
       ($$handler ,leapable?
                  $code
                  rte
                  (lambda ($code rte ,@vars)
                    ($$execute-body $code rte ,@vars))
                  ,@vars)
       ($$execute-body $code rte ,@vars))))

;------------------------------------------------------------------------------

; Macros to manipulate the runtime environment

(##define-macro (mk-rte rte . lst)
  `(##vector ,rte ,@lst))

(##define-macro (mk-rte* rte ns)
  `(let (($rte (##make-vector (##fixnum.+ ,ns 1) (##void))))
     (##vector-set! $rte 0 ,rte)
     $rte))

(##define-macro (rte-up rte)         `(##vector-ref ,rte 0))
(##define-macro (rte-ref rte i)      `(##vector-ref ,rte ,i))
(##define-macro (rte-set! rte i val) `(##vector-set! ,rte ,i ,val))

;==============================================================================

(##include "../lib/_kernel#.scm")
(##include "../lib/_thread#.scm")
(##include "../lib/_system#.scm")
(##include "../lib/_error#.scm")
(##include "../lib/_num#.scm")
(##include "../lib/_io#.scm")
(##include "../lib/_eval#.scm")
(##include "../lib/_repl#.scm")
(##include "../lib/_std#.scm")
(##include "../lib/_nonstd#.scm")
