/* file: "main.c" */

/* Copyright (C) 1994-2000 by Marc Feeley, All Rights Reserved. */

/* This is the driver of the Gambit-C system */

#define ___INCLUDED_FROM_MAIN
#define ___VERSION_MAJOR 21
#include "gambit.h"

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>

#include "os.h"


___HIDDEN void usage_err
   ___P((char **argv),(argv)
char **argv;)
{
  fprintf (stderr, "Usage: %s [-:OPT,OPT...] ...\n", argv[0]);
  fprintf (stderr, "  where OPT can be one of:\n");
  fprintf (stderr, "    mMIN_HEAP_SIZE_IN_KILOBYTES\n");
  fprintf (stderr, "    hMAX_HEAP_SIZE_IN_KILOBYTES\n");
  fprintf (stderr, "    lPERCENT_LIVE_AFTER_GC\n");
  fprintf (stderr, "    s\n");
  fprintf (stderr, "    d[level]\n");
  ___exit_process (1);
}


___EXP_FUNC(int,___main)
  ___P((int argc,
        char **argv,
        ___mod_or_lnk (*linker)(___global_state_struct*)),
       (argc,
        argv,
        linker)
int argc;
char **argv;
___mod_or_lnk (*linker)();)
{
  long min_heap_len;
  long max_heap_len;
  int live_percent;
  ___BOOL standard;
  int debug_level;

  /* handle arguments to runtime */

  min_heap_len = 0;
  max_heap_len = 0;
  live_percent = 0;
  standard = 0;
  debug_level = 0;

  if (argc > 1)
    {
      char *s, *arg = argv[1];
      if (*arg++ == '-' && *arg++ == ':')
        {
          do
            {
              s = arg++;
              switch (*s)
                {
                case 'm':
                case 'h':
                case 'l':
                case 'd':
                  {
                    char temp;
                    int argval;
                    while (*arg >= '0' && *arg <= '9') arg++;
                    temp = *arg;
                    *arg = '\0';
                    if (arg == s+1 && *s == 'd')
                      argval = 1;
                    else
                      argval = atoi (s+1);
                    *arg = temp;
                    switch (*s)
                      {
                      case 'm': min_heap_len = argval<<10;
                                break;
                      case 'h': max_heap_len = argval<<10;
                                break;
                      case 'l': live_percent = argval;
                                break;
                      case 'd': debug_level = argval;
                                break;
                      }
                    break;
                  }
                case 's':
                  standard = 1;
                  break;
                default:
                  usage_err (argv);
                }
            }
          while (*arg++ == ',');
          if (*--arg != '\0')
            usage_err (argv);
          argv[1] = argv[0];
          argc--;
          argv++;
        }
    }

#ifdef FORCE_MAX_HEAP
  if (max_heap_len == 0 || max_heap_len > FORCE_MAX_HEAP<<10)
    max_heap_len = FORCE_MAX_HEAP<<10;
#endif

  /* Setup program, run it and perform any cleanup necessary. */

  {
    ___setup_params_struct setup_params;

    ___setup_params_reset (&setup_params);

    setup_params.version_major = ___VERSION_MAJOR;
    setup_params.version_minor = ___VERSION_MINOR;
    setup_params.argc          = argc;
    setup_params.argv          = argv;
    setup_params.min_heap      = min_heap_len;
    setup_params.max_heap      = max_heap_len;
    setup_params.live_percent  = live_percent;
    setup_params.standard      = standard;
    setup_params.debug_level   = debug_level;
    setup_params.linker        = linker;

    ___setup (&setup_params);

    ___cleanup ();
  }

  /* Exit program. */

  ___exit_process (0);

  return 0; /* satisfy some compilers */
}
