/* file: "os.h" */

/* Copyright (C) 1994-2000 by Marc Feeley, All Rights Reserved. */

#ifndef ___OS_H
#define ___OS_H

/*---------------------------------------------------------------------------*/

/* Select OS routines to use. */

#ifdef sparc
#ifndef __linux__
#define USE_UNIX
#define USE_dlopen
#define USE_sigaction_SIGINT
#define USE_sigaction
#define USE_setitimer
#endif
#endif

#ifdef sgi
#define USE_UNIX
#define USE_dlopen
#define USE_signal_SIGINT
#define USE_signal
#define USE_setitimer
#endif

#ifdef __alpha
#define USE_UNIX
#define USE_dlopen
#define USE_sigaction_SIGINT
#define USE_sigaction
#define USE_setitimer
#endif

#ifdef hpux
#define USE_UNIX
#define USE_shl_load
#define USE_sigaction_SIGINT
#define USE_sigaction
#define USE_setitimer
#endif

#ifdef _AIX
#define USE_UNIX
#define USE_sigaction_SIGINT
#define USE_sigaction
#define USE_setitimer
#endif

#ifdef __linux__
#define USE_UNIX
#define USE_dlopen
#define USE_sigaction_SIGINT
#define USE_sigaction
#define USE_setitimer
#define USE__FPU_SETCW
#endif

#ifdef __OpenBSD__
#ifdef i386
#define USE_UNIX
#define USE_dlopen
#define USE_sigaction_SIGINT
#define USE_sigaction
#define USE_setitimer
#endif
#endif

#ifdef __NetBSD__
#ifdef i386
#define USE_UNIX
#define USE_dlopen
#define USE_sigaction_SIGINT
#define USE_sigaction
#define USE_setitimer
#endif
#endif

#ifdef __FreeBSD__
#ifdef i386
#define USE_UNIX
#define USE_dlopen
#define USE_sigaction_SIGINT
#define USE_sigaction
#define USE_setitimer
#endif
#endif

#ifdef __DJGPP__
#ifdef __i386__
#define USE_DJGPP
#define USE_signal_SIGINT
#define USE_signal
#define USE_setitimer
#endif
#endif

#ifdef _MSC_VER
#ifndef __MWERKS__
#if _MSC_VER < 1000
#ifdef _M_I86
#define USE_WIN32
#define USE_fullpath
#define USE_ftime
#endif
#else
#ifdef _M_IX86
#define USE_WIN32
#define USE_fullpath
#define USE_ftime
#endif
#endif
#ifndef S_ISDIR
#define S_ISDIR(mode)(mode&_S_IFDIR)
#endif
#define popen(command,mode)_popen(command,mode)
#define pclose(stream)_pclose(stream)
#define isatty(descr)_isatty(descr)
#define USE_autocad_gui
#endif
#endif

#ifdef USE_autocad_gui
#define RTNORM  5100
#define RTCAN  -5002
extern int acedPrompt (const char *str);
extern int acedGetString (int cronly, const char *prompt, char *result);
extern int acedGetStringB (int cronly, const char *prompt,
                           char *result, int bufsize);
extern int acedUsrBrk (void);
extern int acedCommand (int rtype, ...);
#endif

#ifdef __BORLANDC__
#ifdef _M_IX86
#define USE_WIN32
#define USE_fullpath
#define USE_ftime
#define popen(command,mode)_popen(command,mode)
#define pclose(stream)_pclose(stream)
#endif
#endif

#ifdef __WATCOMC__
#ifdef __386__
#define USE_WIN32
#define USE_fullpath
#define USE_ftime
#ifdef __SW_BW
#define INCLUDE_wdefwin_h
#endif
#define popen(command,mode)_popen(command,mode)
#define pclose(stream)_pclose(stream)
#endif
#endif

#ifdef __MWERKS__
#ifdef __INTEL__
#define USE_WIN32
#define USE_time
#define fileno(stream)_fileno(stream)
#define isatty(descr)_isatty(descr)
#else
#define USE_MACOS
#define USE_GetDateTime
#define USE_TickCount
#define USE_CWIDE
#define EIO (EINTR+1)
#define BUFFERED_STDOUT
#if __MWERKS__ < 0x2000
#define kPrivateCFragCopy kNewCFragCopy
#endif
#if __MWERKS__ < 0x1800
#define USE_SIOUX
#else
#define USE_mac_gui
#endif
#endif
#define EINTR 9999
#define popen(command,mode) 0
#define pclose(stream) -1
#endif

#ifdef __MACHTEN__
#define USE_UNIX
#endif

#ifdef __LCC__
#ifdef _X86_
#ifdef _M_IX86
#define USE_fullpath
#define USE_ftime
#ifndef S_ISDIR
#define S_ISDIR(mode)(mode&_S_IFDIR)
#endif
#endif
#endif
#endif

#ifdef _CRAY
#ifdef _CRAY1
#define USE_UNICOS
#endif
#endif

/*---------------------------------------------------------------------------*/

/* Determine which header files to include. */

#ifdef USE_UNIX
#define USE_getrusage
#define USE_gettimeofday
#define INCLUDE_unistd_h
#define INCLUDE_pwd_h
#define INCLUDE_sys_types_h
#define INCLUDE_sys_time_h
#define INCLUDE_time_h
#endif

#ifdef USE_DJGPP
#define USE_gettimeofday
#define USE_times
#define USE_DOS_PATHS
#define USE_dxe_load
#define USE_control87
#endif

#ifdef USE_WIN32
#define USE_clock
#define USE_DOS
#define USE_LoadLibrary
#define USE_signal_SIGINT
#define USE_CreateThread
#define INCLUDE_windows_h
#define INCLUDE_io_h
#define USE_control87
#define USE_FormatMessage
#endif

#ifdef USE_OS2
#define USE_DOS
#define USE_fullpath
#define USE_DosStartTimer
#define USE_DosLoadModule
#define INCLUDE_sys_types_h
#define INCLUDE_sys_stat_h
#define INCLUDE_signal_h
#define INCL_DOSPROCESS
#define INCL_DOSERRORS
#define INCL_DOSFILEMGR
#define INCL_DOSMISC
#define INCL_DOSDATETIME
#define INCLUDE_os2_h
#endif

#ifdef USE_DOS
#define USE_DOS_PATHS
#ifndef USE_WIN32
#define INCLUDE_dos_h
#endif
#define INCLUDE_direct_h
#define INCLUDE_time_h
#endif

#ifdef USE_MACOS
#define USE_GetDiskFragment
#define USE_signal_SIGINT
#define USE_VInstall
#define INCLUDE_Retrace_h
#define INCLUDE_Files_h
#define INCLUDE_Finder_h
#define INCLUDE_Errors_h
#define INCLUDE_Folders_h
#define INCLUDE_OSUtils_h
#define INCLUDE_Power_h
#define INCLUDE_CodeFragments_h
#define INCLUDE_signal_h
#define INCLUDE_stat_h
#ifdef USE_SIOUX
#define INCLUDE_SIOUX_h
#define INCLUDE_unistd_h
#define INCLUDE_unix_h
#endif
#ifdef USE_mac_gui
#define INCLUDE_mac_gui_h
#endif
#endif

#ifdef USE_UNICOS
#define USE_gettimeofday
#define USE_times
#define USE_signal_SIGINT
#define USE_t_fork
#define INCLUDE_unistd_h
#define INCLUDE_sys_types_h
#define INCLUDE_sys_time_h
#define INCLUDE_time_h
#endif

#ifdef USE_time
#define INCLUDE_time_h
#endif

#ifdef USE_ftime
#define INCLUDE_sys_types_h
#define INCLUDE_sys_timeb_h
#endif

#ifdef USE_gettimeofday
#define INCLUDE_sys_time_h
#endif

#ifdef USE_times
#define INCLUDE_sys_times_h
#endif

#ifdef USE_clock_gettime
#define INCLUDE_time_h
#endif

#ifdef USE_getclock
#define INCLUDE_sys_timers_h
#endif

#ifdef USE_getrusage
#define INCLUDE_sys_resource_h
#endif

#ifdef USE_dlopen
#define INCLUDE_dlfcn_h
#endif

#ifdef USE_shl_load
#define INCLUDE_dl_h
#endif

#ifdef USE_dxe_load
#define INCLUDE_sys_dxe_h
#endif

#ifdef USE_signal
#undef INCLUDE_signal_h
#define INCLUDE_signal_h
#endif

#ifdef USE_sigaction
#undef INCLUDE_signal_h
#define INCLUDE_signal_h
#endif

#ifdef USE_fullpath
#define INCLUDE_sys_stat_h
#endif

#ifdef USE_control87
#define INCLUDE_float_h
#endif

#ifdef USE__FPU_SETCW
#ifdef HAS_fpu_control_h
#define INCLUDE_fpu_control_h
#else
/*
 * Some Linux distributions don't have the file "fpu_control.h".
 * Instead of including that file we explicitly define the macros we
 * need.
 */
#ifndef _FPU_CONTROL_H
#define _FPU_CONTROL_H
#define _FPU_MASK_IM  0x01
#define _FPU_MASK_DM  0x02
#define _FPU_MASK_ZM  0x04
#define _FPU_MASK_OM  0x08
#define _FPU_MASK_UM  0x10
#define _FPU_MASK_PM  0x20
#define _FPU_DOUBLE   0x200
#define _FPU_RC_NEAREST 0x0
typedef unsigned int fpu_control_t __attribute__ ((__mode__ (__HI__)));
#define _FPU_SETCW(cw) __asm__ ("fldcw %0" : : "m" (*&cw))
#endif
#endif
#endif

#ifdef USE_t_fork
#define INCLUDE_tfork_h
#define INCLUDE_sys_wait_h
#endif


/*---------------------------------------------------------------------------*/

/* Inclusion of header files. */

#define _STDIO_UNLOCK_CHAR_IO

#include <stdio.h>
#include <errno.h>

#ifndef ESUCCESS
#define ESUCCESS 0
#endif

#ifndef EINTR
#define EINTR -1
#endif

#include <stdlib.h>

#include <string.h>

#ifdef INCLUDE_unistd_h
#include <unistd.h>
#endif

#ifdef INCLUDE_pwd_h
#include <pwd.h>
#endif

#ifdef INCLUDE_dlfcn_h
#include <dlfcn.h>
#endif

#ifdef INCLUDE_dl_h
#include <dl.h>
#endif

#ifdef INCLUDE_sys_dxe_h
#include <sys/dxe.h>
#endif

#ifdef INCLUDE_sys_types_h
#include <sys/types.h>
#endif

#ifdef INCLUDE_sys_time_h
#include <sys/time.h>
#endif

#ifdef INCLUDE_sys_times_h
#include <sys/times.h>
#endif

#ifdef INCLUDE_sys_timeb_h
#include <sys/timeb.h>
#endif

#ifdef INCLUDE_sys_timers_h
#include <sys/timers.h>
#endif

#ifdef INCLUDE_time_h
#include <time.h>
#endif

#ifdef INCLUDE_sys_resource_h
#include <sys/resource.h>
#endif

#ifdef INCLUDE_sys_stat_h
#include <sys/stat.h>
#endif

#ifdef INCLUDE_sys_wait_h
#include <sys/wait.h>
#endif

#ifdef INCLUDE_stat_h
#include <stat.h>
#endif

#ifdef INCLUDE_signal_h
#include <signal.h>
#endif

#ifdef INCLUDE_windows_h
#include <windows.h>
#endif

#ifdef INCLUDE_io_h
#include <io.h>
#endif

#ifdef INCLUDE_float_h
#include <float.h>
#endif

#ifdef INCLUDE_fpu_control_h
#include <fpu_control.h>
#endif

#ifdef INCLUDE_os2_h
#include <os2.h>
#endif

#ifdef INCLUDE_dos_h
#include <dos.h>
#endif

#ifdef INCLUDE_direct_h
#include <direct.h>
#endif

#ifdef INCLUDE_Retrace_h
#include <Retrace.h>
#endif

#ifdef INCLUDE_Files_h
#include <Files.h>
#endif

#ifdef INCLUDE_Finder_h
#include <Finder.h>
#endif

#ifdef INCLUDE_Errors_h
#include <Errors.h>
#endif

#ifdef INCLUDE_Folders_h
#include <Folders.h>
#endif

#ifdef INCLUDE_OSUtils_h
#include <OSUtils.h>
#endif

#ifdef INCLUDE_Power_h
#include <Power.h>
#endif

#ifdef INCLUDE_CodeFragments_h
#include <CodeFragments.h>
#endif

#ifdef INCLUDE_SIOUX_h
#include <SIOUX.h>
#endif

#ifdef INCLUDE_mac_gui_h
#include "mac_gui.h"
#endif

#ifdef INCLUDE_unix_h
#include <unix.h>
#endif

#ifdef INCLUDE_wdefwin_h
#include <wdefwin.h>
#endif

#ifdef INCLUDE_tfork_h
#include <tfork.h>
#endif


/*---------------------------------------------------------------------------*/

/* Max length of a path, not including null. */

#define ___PATH_MAX_LENGTH 1024


/* Max length of a shell command, not including null. */

#define ___CMD_MAX_LENGTH 1024


/* Max length of an error message, not including null. */

#define ___ERR_MAX_LENGTH 1024


/*
 * Does "___alloc_mem" allocate upwards (___ALLOC_MEM_UP) or downwards
 * (___ALLOC_MEM_DOWN).  This setting only affects performance.
 */

#define ___ALLOC_MEM_UP


/* Process termination. */

extern void ___exit_process
   ___P((int status),
        ());


/* Fatal error handling. */

/* 
 * The procedure "___fatal_error" is called by the runtime system when
 * an unrecoverable error has occured.
 */

extern void ___fatal_error
   ___P((char *msg),
        ());


/* Conversion of error codes to error messages. */

extern void ___err_code_to_string
   ___P((___ERR_CODE err,
         char *str),
        ());


/* Low-level memory allocation. */

extern void *___alloc_mem
   ___P((unsigned long bytes),
        ());

extern void ___free_mem
   ___P((void *ptr),
        ());


/* Time management. */

#define ___FLOAT_TIME_REPRESENTATION

#ifdef ___FLOAT_TIME_REPRESENTATION

typedef ___F64 ___time;

#endif

#ifdef ___INT_TIME_REPRESENTATION

typedef struct ___time_struct
  {
    ___SM32 secs;
    ___SM32 nsecs;
  } ___time;

#endif

extern ___time ___time_pos_infinity;
extern ___time ___time_neg_infinity;
extern ___time ___time_zero;

extern void ___time_from_nsecs
   ___P((___time *tim,
         ___SM32 secs,
         ___SM32 nsecs),
        ());

extern void ___time_add
   ___P((___time *tim1,
         ___time tim2),
        ());

extern void ___time_subtract
   ___P((___time *tim1,
         ___time tim2),
        ());

extern ___BOOL ___time_equal
   ___P((___time tim1,
         ___time tim2),
        ());

extern ___BOOL ___time_less
   ___P((___time tim1,
         ___time tim2),
        ());

extern ___BOOL ___time_positive
   ___P((___time tim),
        ());

extern void ___time_get_current_time
   ___P((___time *tim),
        ());

extern void ___time_to_seconds
   ___P((___time tim,
         ___F64 *seconds),
        ());

extern void ___time_from_seconds
   ___P((___time *tim,
         ___F64 seconds),
        ());


/* Process times (user time, system time and real time). */

extern void ___process_times
   ___P((___F64 *user,
         ___F64 *sys,
         ___F64 *real),
        ());


/* Heartbeat interrupt handling. */

extern ___F64 ___set_heartbeat_interval
   ___P((___F64 seconds),
        ());


/* Access to environment variables. */

extern char *___getenv
   ___P((char *name),
        ());


/* Shell command. */

extern int ___shell_command
   ___P((char *cmd),
        ());


/* Dynamic loading. */

extern void *___dynamic_load
   ___P((char *path,
         char *name,
         char **errmsg),
        ());


/* Dynamic C compilation. */

extern ___BOOL ___dynamic_cc
   ___P((char *cc_cmd,
         char *ld_cmd,
         char *root_path,
         char *output_path,
         int ld_flip,
         char **errmsg),
        ());


/* Virtual memory statistics. */

extern void ___vm_stats
   ___P((long *minflt,
         long *majflt),
        ());


/* Filesystem path expansion. */

#define ___PATH_ABSOLUTE 0
#define ___PATH_RELATIVE 1
#define ___PATH_SHORTEST -1

extern ___BOOL ___path_expand
   ___P((char *path,
         char *new_path,
         long max_length,
         int format),
        ());

extern ___BOOL ___path_absolute
   ___P((char *path),
        ());

extern ___BOOL ___path_extension
   ___P((char *path,
         char *ext,
         long max_length),
        ());

extern ___BOOL ___path_strip_extension
   ___P((char *path,
         char *new_path,
         long max_length),
        ());

extern ___BOOL ___path_directory
   ___P((char *path,
         char *dir,
         long max_length),
        ());

extern ___BOOL ___path_strip_directory
   ___P((char *path,
         char *new_path,
         long max_length),
        ());

extern ___BOOL ___path_drive
   ___P((char *path,
         char *drive,
         long max_length),
        ());

extern ___BOOL ___path_strip_drive
   ___P((char *path,
         char *new_path,
         long max_length),
        ());


/* Formatting of source code position. */

extern char *___format_filepos
   ___P((char *path,
         long filepos,
         ___BOOL pinpoint),
        ());


/* OS initialization/finalization. */

extern void ___setup_os ___PVOID;
extern void ___cleanup_os ___PVOID;


/* Utilities for machine encoding of characters. */

/*
 * For now a LATIN-1 encoding of characters is assumed for 'unsigned
 * char'.  This means that the mapping to and from the ISO/IEC
 * 10646-1:1993 character set is the identity function.  To support
 * other machine representations, e.g. EBCDIC, these macros would have
 * to actually translate the characters.
 */

#define ucs4_to_uchar(u)u
#define uchar_to_ucs4(c)c


/*---------------------------------------------------------------------------*/

#define USE_POSIX
#define USE_WIN32zzzzzzzz
#define USE_MACOSzzzzzzzz
#define USE_TERMINFO
#define USE_TERMCAPzzzzzzzzzz

#define USE_gettimeofday
#define USE_select
#define USE_nanosleepzzzzzzzz
#define USE_errno




#include <stdio.h>
#include <stdlib.h>

#include <errno.h>


#ifdef USE_POSIX

#include <curses.h>
#include <term.h>

#include <signal.h>
#include <sys/ioctl.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include <sys/time.h>

#include <termios.h>
#include <unistd.h>

#include <termios.h>

#include <sys/socket.h>
#include <netinet/in.h>

#include <dirent.h>

#endif


#ifdef USE_WIN32

#include <winsock2.h>
#include <windows.h>
#include <tchar.h>
#include <sys/types.h>
#include <sys/timeb.h>

#undef COLOR_CYAN
#undef COLOR_BLUE
#define COLOR_CYAN 6
#define COLOR_BLUE 4

#endif

/*****************************************************************************/

/* Devices. */

typedef struct ___device_group_struct
  {
    struct ___device_struct *list; /* list of devices in this group */
  } ___device_group;

#define ___NONE_KIND              0
#define ___OBJECT_KIND            1
#define ___CHAR_KIND              3
#define ___BYTE_KIND              7
#define ___DEVICE_KIND            15
#define ___FILE_DEVICE_KIND       (___DEVICE_KIND+16)
#define ___TTY_DEVICE_KIND        (___DEVICE_KIND+32)
#define ___TCP_CLIENT_DEVICE_KIND (___DEVICE_KIND+64)
#define ___TCP_SERVER_KIND        (___OBJECT_KIND+128)
#define ___PIPE_DEVICE_KIND       (___DEVICE_KIND+256)
#define ___SERIAL_DEVICE_KIND     (___DEVICE_KIND+512)
#define ___DIRECTORY_KIND         (___OBJECT_KIND+1024)
#define ___EVENT_QUEUE_KIND       (___OBJECT_KIND+2048)
#define ___TIMER_KIND             (___OBJECT_KIND+4096)

#define ___CHAR_ENCODING(x)     ((x)&(15<<0))
#define ___CHAR_ENCODING_ASCII  (0<<0)
#define ___CHAR_ENCODING_LATIN1 (1<<0)
#define ___CHAR_ENCODING_UTF8   (2<<0)
#define ___CHAR_ENCODING_UCS2   (3<<0)
#define ___CHAR_ENCODING_UCS2BE (4<<0)
#define ___CHAR_ENCODING_UCS2LE (5<<0)
#define ___CHAR_ENCODING_UCS4   (6<<0)
#define ___CHAR_ENCODING_UCS4BE (7<<0)
#define ___CHAR_ENCODING_UCS4LE (8<<0)
#define ___CHAR_ENCODING_NATIVE (9<<0)

#ifdef ___CPU_BIGEND
#define ___CHAR_ENCODING_U8  ___CHAR_ENCODING_LATIN1
#define ___CHAR_ENCODING_U16 ___CHAR_ENCODING_UCS2BE
#define ___CHAR_ENCODING_U32 ___CHAR_ENCODING_UCS4BE
#else
#define ___CHAR_ENCODING_U8  ___CHAR_ENCODING_LATIN1
#define ___CHAR_ENCODING_U16 ___CHAR_ENCODING_UCS2LE
#define ___CHAR_ENCODING_U32 ___CHAR_ENCODING_UCS4LE
#endif

#define ___EOL_ENCODING(x)   ((x)&(7<<4))
#define ___EOL_ENCODING_LF   (0<<4)
#define ___EOL_ENCODING_CR   (1<<4)
#define ___EOL_ENCODING_CRLF (2<<4)
#define ___EOL_ENCODING_LFCR (3<<4)

#define ___DECODE_STATE(x)   ((x)&(3<<7))
#define ___DECODE_STATE_NONE (0<<7)
#define ___DECODE_STATE_LF   (1<<7)
#define ___DECODE_STATE_CR   (2<<7)

#define ___OPEN_STATE(x)     ((x)&(1<<9))
#define ___OPEN_STATE_OPEN   (0<<9)
#define ___OPEN_STATE_CLOSED (1<<9)

#define ___LINE_BUFFERED(x) ((x)<___BUFFERING)
#define ___NO_BUFFERING     (0<<12)
#define ___LINE_BUFFERING   (1<<12)
#define ___BUFFERING        (2<<12)

#define ___DIRECTION_RD 1
#define ___DIRECTION_WR 2

#define ___STAGE_OPEN     0
#define ___STAGE_CLOSING1 1
#define ___STAGE_CLOSING2 2
#define ___STAGE_CLOSED   3

#define ___SELECT_PASS_CHECK 0
#define ___SELECT_PASS_1     1
#define ___SELECT_PASS_2     2
#define ___SELECT_PASS_3     3

#define ___STREAM_DEFAULT_OPTIONS(ice,iee,ib,oce,oee,ob) \
___FIX((ice+iee+ib)+((oce+oee+ob)<<14))

typedef struct ___device_struct
  {
    void *vtbl;

#ifdef USE_PUMPS
    LONG refcount;                 /* device structure is released when zero */
#else
    int refcount;                  /* device structure is released when zero */
#endif

    ___device_group *group;        /* device group this device belongs to */
    struct ___device_struct *prev; /* bidirectional list pointer to previous */
    struct ___device_struct *next; /* bidirectional list pointer to next */
    int direction;                 /* ___DIRECTION_RD and/or ___DIRECTION_WR */
    int read_stage;                /* ___STAGE_OPEN ... ___STAGE_CLOSED */
    int write_stage;               /* ___STAGE_OPEN ... ___STAGE_CLOSED */
    int select_step;               /* for implementing "___device_select" */
  } ___device;

typedef struct ___device_select_state_struct
  {
    ___device **devs; /* devices to select on */

    ___time timeout;  /* timeout of the select */

#ifdef USE_select
    int highest_fd_plus_1;
    fd_set readfds;
    fd_set writefds;
    fd_set exceptfds;
    int select_result;
#endif

#ifdef USE_MsgWaitForMultipleObjects
    DWORD timeout;/*********************/
    int timeout_dev_pos;
    DWORD message_queue_mask;
    int message_queue_dev_pos;
    int nb_wait_objs;
    HANDLE wait_objs_buffer[MAXIMUM_WAIT_OBJECTS];
    int wait_obj_to_dev_pos[MAXIMUM_WAIT_OBJECTS];
#endif

#ifdef USE_MACOS
    /*********************/
#endif
  } ___device_select_state;

typedef struct ___device_vtbl_struct
  {
#define ___device_kind(self) \
___CAST(___device_vtbl*,(self)->vtbl)->kind(self)

    int (*kind) ___P((___device *self),());

#define ___device_select_virt(self,for_writing,i,pass,state) \
___CAST(___device_vtbl*,(self)->vtbl)->select_virt(self,for_writing,i,pass,state)

    ___WORD (*select_virt)
       ___P((___device *self,
             ___BOOL for_writing,
             int i,
             int pass,
             ___device_select_state *state),
            ());

#define ___device_release_virt(self) \
___CAST(___device_vtbl*,(self)->vtbl)->release_virt(self)

    ___WORD (*release_virt)
       ___P((___device *self),
            ());

#define ___device_flush_write_virt(self) \
___CAST(___device_vtbl*,(self)->vtbl)->flush_write_virt(self)

    ___WORD (*flush_write_virt)
       ___P((___device *self),
            ());

#define ___device_close_virt(self,direction) \
___CAST(___device_vtbl*,(self)->vtbl)->close_virt(self,direction)

    ___WORD (*close_virt)
       ___P((___device *self,
             int direction),
            ());
  } ___device_vtbl;

/* Timer device. */

typedef struct ___device_timer_struct
  {
    ___device base;
    ___time expiry;
  } ___device_timer;

typedef struct ___device_timer_vtbl_struct
  {
    ___device_vtbl base;
  } ___device_timer_vtbl;

extern ___WORD ___device_timer_setup
   ___P((___device_timer **dev,
         ___device_group *dgroup),
        ());

extern void ___device_timer_set_expiry
   ___P((___device_timer *dev,
         ___time expiry),
        ());

/* Byte stream device. */

typedef int ___stream_index;

typedef struct ___device_stream_struct
  {
    ___device base;

#ifdef USE_PUMPS
    ___device_stream_pump *read_pump;
    ___device_stream_pump *write_pump;
#endif
  } ___device_stream;

typedef struct ___device_stream_vtbl_struct
  {
    ___device_vtbl base;

#define ___device_stream_select_raw_virt(self,for_writing,i,pass,state) \
___CAST(___device_stream_vtbl*,(self)->base.vtbl)->select_raw_virt(self,for_writing,i,pass,state)

    ___WORD (*select_raw_virt)
       ___P((___device_stream *self,
             ___BOOL for_writing,
             int i,
             int pass,
             ___device_select_state *state),
            ());

#define ___device_stream_release_raw_virt(self) \
___CAST(___device_stream_vtbl*,(self)->base.vtbl)->release_raw_virt(self)

    ___WORD (*release_raw_virt)
       ___P((___device_stream *self),
            ());

#define ___device_stream_flush_write_raw_virt(self) \
___CAST(___device_stream_vtbl*,(self)->base.vtbl)->flush_write_raw_virt(self)

    ___WORD (*flush_write_raw_virt)
       ___P((___device_stream *self),
            ());

#define ___device_stream_seek_raw_virt(self,pos) \
___CAST(___device_stream_vtbl*,(self)->base.vtbl)->seek_raw_virt(self,pos)

    ___WORD (*seek_raw_virt)
       ___P((___device_stream *self,
             ___stream_index pos),
            ());

#define ___device_stream_read_raw_virt(self,buf,len,len_done) \
___CAST(___device_stream_vtbl*,(self)->base.vtbl)->read_raw_virt(self,buf,len,len_done)

    ___WORD (*read_raw_virt)
       ___P((___device_stream *self,
             ___U8 *buf,
             ___stream_index len,
             ___stream_index *len_done),
            ());

#define ___device_stream_write_raw_virt(self,buf,len,len_done) \
___CAST(___device_stream_vtbl*,(self)->base.vtbl)->write_raw_virt(self,buf,len,len_done)

    ___WORD (*write_raw_virt)
       ___P((___device_stream *self,
             ___U8 *buf,
             ___stream_index len,
             ___stream_index *len_done),
            ());

#define ___device_stream_width(self) \
___CAST(___device_stream_vtbl*,(self)->base.vtbl)->width(self)

    ___WORD (*width)
       ___P((___device_stream *self),
            ());

#define ___device_stream_default_options(self) \
___CAST(___device_stream_vtbl*,(self)->base.vtbl)->default_options(self)

    ___WORD (*default_options)
       ___P((___device_stream *self),
            ());
  } ___device_stream_vtbl;

extern ___WORD ___device_stream_seek
   ___P((___device_stream *self,
         ___stream_index pos),
        ());

extern ___WORD ___device_stream_read
   ___P((___device_stream *self,
         ___U8 *buf,
         ___stream_index len,
         ___stream_index *len_done),
        ());

extern ___WORD ___device_stream_write
   ___P((___device_stream *self,
         ___U8 *buf,
         ___stream_index len,
         ___stream_index *len_done),
        ());

extern ___WORD ___device_stream_setup
   ___P((___device_stream *dev,
         ___device_group *dgroup,
         int direction,
         int pumps_on),
        ());


extern ___WORD ___device_stream_setup_from_fd
   ___P((___device_stream **dev,
         ___device_group *dgroup,
         int fd,
         int type,
         int direction),
        ());

extern ___WORD ___device_stream_setup_from_path
   ___P((___device_stream **dev,
         ___device_group *dgroup,
         char *path,
         int flags,
         int mode),
        ());


#if 0
/* Tty stream device. */

typedef struct ___device_tty_struct
  {
    ___device_stream base;
    tty t;
  } ___device_tty;

typedef struct ___device_tty_vtbl_struct
  {
    ___device_stream_vtbl base;
  } ___device_tty_vtbl;

extern ___WORD ___device_tty_setup
   ___P((___device_tty **dev,
         ___device_group *dgroup,
         int fd,
         int direction),
        ());
#endif




extern ___WORD ___device_group_setup
   ___P((___device_group **dgroup),
        ());

extern ___WORD ___device_group_cleanup
   ___P((___device_group *dgroup),
        ());





extern ___WORD ___device_select
   ___P((___device **devs,
         int nb_read_devs,
         int nb_write_devs,
         ___time timeout),
        ());

extern ___WORD ___device_flush_write
   ___P((___device *self),
        ());

extern ___WORD ___device_close
   ___P((___device *self,
         int direction),
        ());

extern ___WORD ___device_cleanup
   ___P((___device *self),
        ());



extern ___WORD ___condvar_select
   ___P((___WORD runq,
         ___WORD timeout),
        ());




extern ___WORD ___dev_kind
   ___P((___WORD dev),
        ());

extern ___WORD ___dev_force_output
   ___P((___WORD dev),
        ());

extern ___WORD ___dev_close
   ___P((___WORD dev,
         ___WORD direction),
        ());

extern ___WORD ___dev_stream_open_predefined
   ___P((___WORD index,
         ___WORD flags),
        ());

extern ___WORD ___dev_stream_open_path
   ___P((___WORD path,
         ___WORD flags,
         ___WORD mode),
        ());

extern ___WORD ___dev_stream_read
   ___P((___WORD dev,
         ___WORD buffer,
         ___WORD lo,
         ___WORD hi),
        ());

extern ___WORD ___dev_stream_write
   ___P((___WORD dev,
         ___WORD buffer,
         ___WORD lo,
         ___WORD hi),
        ());

extern ___WORD ___dev_stream_width
   ___P((___WORD dev),
        ());

extern ___WORD ___dev_stream_default_options
   ___P((___WORD dev),
        ());




extern ___WORD ___dev_tty_type_set
   ___P((___WORD dev,
         ___WORD term_type,
         ___WORD emacs_bindings),
        ());

extern ___WORD ___dev_tty_text_attributes_set
   ___P((___WORD dev,
         ___WORD input,
         ___WORD output),
        ());

extern ___WORD ___dev_tty_history
   ___P((___WORD dev),
        ());

extern ___WORD ___dev_tty_history_set
   ___P((___WORD dev,
         ___WORD history),
        ());

extern ___WORD ___dev_tty_max_history_length_set
   ___P((___WORD dev,
         ___WORD max_length),
        ());

extern ___WORD ___dev_tty_paren_balance_duration_set
   ___P((___WORD dev,
         ___WORD duration),
        ());

extern ___WORD ___dev_tty_mode_set
   ___P((___WORD dev,
         ___WORD mode),
        ());

extern ___WORD ___get_host_by_name
   ___P((___WORD host_name),
        ());

extern ___WORD ___dev_tcp_client_open
   ___P((___WORD server_addr,
         ___WORD port_num),
        ());

extern ___WORD ___dev_tcp_server_open
   ___P((___WORD port_num,
         ___WORD backlog),
        ());

extern ___WORD ___dev_tcp_server_read
   ___P((___WORD dev),
        ());

extern ___WORD ___dev_directory_open_path
   ___P((___WORD path),
        ());

extern ___WORD ___dev_directory_read
   ___P((___WORD dev),
        ());

extern ___WORD ___port_decode_chars
   ___P((___WORD port,
         ___WORD eof),
        ());

extern ___WORD ___port_encode_chars
   ___P((___WORD port),
        ());

#endif
