;==============================================================================

; file: "_num.scm"

; Copyright (C) 1994-2000 by Marc Feeley, All Rights Reserved.

;==============================================================================

(##include "header.scm")

;==============================================================================

; Error handling.

(define (##trap-check-number name . args)
  (##runtime-error "NUMBER expected" name args))

(define (##trap-check-number* name . args)
  (##runtime-error* "NUMBER expected" name args))

(define (##trap-check-real name . args)
  (##runtime-error "REAL expected" name args))

(define (##trap-check-real* name . args)
  (##runtime-error* "REAL expected" name args))

(define (##trap-check-inexact-real name . args)
  (##runtime-error "Inexact REAL expected" name args))

(define (##trap-check-inexact-real* name . args)
  (##runtime-error* "Inexact REAL expected" name args))

(define (##trap-check-nonnegative-real name . args)
  (##runtime-error "Nonnegative REAL expected" name args))

(define (##trap-check-nonnegative-real* name . args)
  (##runtime-error* "Nonnegative REAL expected" name args))

(define (##trap-check-finite-real name . args)
  (##runtime-error "Finite REAL expected" name args))

(define (##trap-check-finite-real* name . args)
  (##runtime-error* "Finite REAL expected" name args))

(define (##trap-check-rational name . args)
  (##runtime-error "RATIONAL expected" name args))

(define (##trap-check-rational* name . args)
  (##runtime-error* "RATIONAL expected" name args))

(define (##trap-check-integer name . args)
  (##runtime-error "INTEGER expected" name args))

(define (##trap-check-integer* name . args)
  (##runtime-error* "INTEGER expected" name args))

(define (##trap-check-exact-int name . args)
  (##runtime-error "Exact INTEGER expected" name args))

(define (##trap-check-exact-int* name . args)
  (##runtime-error* "Exact INTEGER expected" name args))

(define (##trap-check-range name . args)
  (##runtime-error "Out of range" name args))

(define (##trap-check-range* name . args)
  (##runtime-error* "Out of range" name args))

(define (##trap-divide-by-zero name . args)
  (##runtime-error "Division by zero" name args))

(define (##trap-divide-by-zero* name . args)
  (##runtime-error* "Division by zero" name args))

;==============================================================================

; Numerical comparison predicates.

(define (##eqv? x y)
  (macro-number-dispatch x (##eq? x y)
    (and (##fixnum? y) (##fixnum.= x y)) ; x = fixnum
    (and (##bignum? y) (##bignum.= x y)) ; x = bignum
    (and (##ratnum? y) (##ratnum.= x y)) ; x = ratnum
    (and (##flonum? y) (##bvector=? x y)) ; x = flonum (all bits must be equal)
    (and (##cpxnum? y) ; x = cpxnum
         (##eqv? (macro-cpxnum-real x) (macro-cpxnum-real y))
         (##eqv? (macro-cpxnum-imag x) (macro-cpxnum-imag y)))))

(define (##= x y)

  (define (type-error) '())

  (macro-number-dispatch x (type-error)

    (macro-number-dispatch y (type-error) ; x = fixnum
      (##fixnum.= x y)
      #f
      #f
      (if (##flonum.<-fixnum-exact? x)
        (##flonum.= (##flonum.<-fixnum x) y)
        (and (##flonum.finite? y)
             (##ratnum.= (##ratnum.<-exact-int x) (##flonum.->ratnum y))))
      (##cpxnum.= (##cpxnum.<-noncpxnum x) y))

    (macro-number-dispatch y (type-error) ; x = bignum
      #f
      (##bignum.= x y)
      #f
      (and (##flonum.finite? y)
           (##ratnum.= (##ratnum.<-exact-int x) (##flonum.->ratnum y)))
      (##cpxnum.= (##cpxnum.<-noncpxnum x) y))

    (macro-number-dispatch y (type-error) ; x = ratnum
      #f
      #f
      (##ratnum.= x y)
      (and (##flonum.finite? y)
           (##ratnum.= x (##flonum.->ratnum y)))
      (##cpxnum.= (##cpxnum.<-noncpxnum x) y))

    (macro-number-dispatch y (type-error) ; x = flonum
      (if (##flonum.<-fixnum-exact? y)
        (##flonum.= x (##flonum.<-fixnum y))
        (and (##flonum.finite? x)
             (##ratnum.= (##flonum.->ratnum x) (##ratnum.<-exact-int y))))
      (and (##flonum.finite? x)
           (##ratnum.= (##flonum.->ratnum x) (##ratnum.<-exact-int y)))
      (and (##flonum.finite? x)
           (##ratnum.= (##flonum.->ratnum x) y))
      (##flonum.= x y)
      (##cpxnum.= (##cpxnum.<-noncpxnum x) y))

    (macro-number-dispatch y (type-error) ; x = cpxnum
      (##cpxnum.= x (##cpxnum.<-noncpxnum y))
      (##cpxnum.= x (##cpxnum.<-noncpxnum y))
      (##cpxnum.= x (##cpxnum.<-noncpxnum y))
      (##cpxnum.= x (##cpxnum.<-noncpxnum y))
      (##cpxnum.= x y))))

(define-nary-bool (= x y)
  #t
  (if (##complex? x) #t '())
  (##= x y)
  macro-force-vars
  macro-no-check
  (##null? ##trap-check-number*))

(define (##< x y #!optional (nan-result #f))

  (define (type-error) '())

  (macro-number-dispatch x (type-error)

    (macro-number-dispatch y (type-error) ; x = fixnum
      (##fixnum.< x y)
      (##not (##bignum.negative? y))
      (##ratnum.< (##ratnum.<-exact-int x) y)
      (cond ((##flonum.finite? y)
             (if (##flonum.<-fixnum-exact? x)
               (##flonum.< (##flonum.<-fixnum x) y)
               (##ratnum.< (##ratnum.<-exact-int x) (##flonum.->ratnum y))))
            ((##flonum.nan? y)
             nan-result)
            (else
             (##flonum.positive? y)))
      (if (macro-cpxnum-real? y)
        (##< x (macro-cpxnum-real y) nan-result)
        (type-error)))

    (macro-number-dispatch y (type-error) ; x = bignum
      (##bignum.negative? x)
      (##bignum.< x y)
      (##ratnum.< (##ratnum.<-exact-int x) y)
      (cond ((##flonum.finite? y)
             (##ratnum.< (##ratnum.<-exact-int x) (##flonum.->ratnum y)))
            ((##flonum.nan? y)
             nan-result)
            (else
             (##flonum.positive? y)))
      (if (macro-cpxnum-real? y)
        (##< x (macro-cpxnum-real y) nan-result)
        (type-error)))

    (macro-number-dispatch y (type-error) ; x = ratnum
      (##ratnum.< x (##ratnum.<-exact-int y))
      (##ratnum.< x (##ratnum.<-exact-int y))
      (##ratnum.< x y)
      (cond ((##flonum.finite? y)
             (##ratnum.< x (##flonum.->ratnum y)))
            ((##flonum.nan? y)
             nan-result)
            (else
             (##flonum.positive? y)))
      (if (macro-cpxnum-real? y)
        (##< x (macro-cpxnum-real y) nan-result)
        (type-error)))

    (macro-number-dispatch y (type-error) ; x = flonum
      (cond ((##flonum.finite? x)
             (if (##flonum.<-fixnum-exact? y)
               (##flonum.< x (##flonum.<-fixnum y))
               (##ratnum.< (##flonum.->ratnum x) (##ratnum.<-exact-int y))))
            ((##flonum.nan? x)
             nan-result)
            (else
             (##flonum.negative? x)))
      (cond ((##flonum.finite? x)
             (##ratnum.< (##flonum.->ratnum x) (##ratnum.<-exact-int y)))
            ((##flonum.nan? x)
             nan-result)
            (else
             (##flonum.negative? x)))
      (cond ((##flonum.finite? x)
             (##ratnum.< (##flonum.->ratnum x) y))
            ((##flonum.nan? x)
             nan-result)
            (else
             (##flonum.negative? x)))
      (if (or (##flonum.nan? x) (##flonum.nan? y))
        nan-result
        (##flonum.< x y))
      (if (macro-cpxnum-real? y)
        (##< x (macro-cpxnum-real y) nan-result)
        (type-error)))

    (if (macro-cpxnum-real? x) ; x = cpxnum
      (macro-number-dispatch y (type-error)
        (##< (macro-cpxnum-real x) y nan-result)
        (##< (macro-cpxnum-real x) y nan-result)
        (##< (macro-cpxnum-real x) y nan-result)
        (##< (macro-cpxnum-real x) y nan-result)
        (if (macro-cpxnum-real? y)
          (##< (macro-cpxnum-real x) (macro-cpxnum-real y) nan-result)
          (type-error)))
      (type-error))))

(define-nary-bool (< x y)
  #t
  (if (##real? x) #t '())
  (##< x y #f)
  macro-force-vars
  macro-no-check
  (##null? ##trap-check-real*))

(define-nary-bool (> x y)
  #t
  (if (##real? x) #t '())
  (##< y x #f)
  macro-force-vars
  macro-no-check
  (##null? ##trap-check-real*))

(define-nary-bool (<= x y)
  #t
  (if (##real? x) #t '())
  (##not (##< y x #t))
  macro-force-vars
  macro-no-check
  (##null? ##trap-check-real*))

(define-nary-bool (>= x y)
  #t
  (if (##real? x) #t '())
  (##not (##< x y #t))
  macro-force-vars
  macro-no-check
  (##null? ##trap-check-real*))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

; Max and min

(define (##max x y)

  (define (type-error) '())

  (macro-number-dispatch x (type-error)

    (macro-number-dispatch y (type-error) ; x = fixnum
      (##fixnum.max x y)
      (if (##< x y) y x)
      (if (##< x y) y x)
      (##flonum.max (##flonum.<-fixnum x) y)
      (if (macro-cpxnum-real? y)
        (##max x (macro-cpxnum-real y))
        (type-error)))

    (macro-number-dispatch y (type-error) ; x = bignum
      (if (##< x y) y x)
      (if (##< x y) y x)
      (if (##< x y) y x)
      (##flonum.max (##flonum.<-exact-int x) y)
      (if (macro-cpxnum-real? y)
        (##max x (macro-cpxnum-real y))
        (type-error)))

    (macro-number-dispatch y (type-error) ; x = ratnum
      (if (##< x y) y x)
      (if (##< x y) y x)
      (if (##< x y) y x)
      (##flonum.max (##flonum.<-ratnum x) y)
      (if (macro-cpxnum-real? y)
        (##max x (macro-cpxnum-real y))
        (type-error)))

    (macro-number-dispatch y (type-error) ; x = flonum
      (##flonum.max x (##flonum.<-fixnum y))
      (##flonum.max x (##flonum.<-exact-int y))
      (##flonum.max x (##flonum.<-ratnum y))
      (##flonum.max x y)
      (if (macro-cpxnum-real? y)
        (##max x (macro-cpxnum-real y))
        (type-error)))

    (if (macro-cpxnum-real? x) ; x = cpxnum
      (macro-number-dispatch y (type-error)
        (##max (macro-cpxnum-real x) y)
        (##max (macro-cpxnum-real x) y)
        (##max (macro-cpxnum-real x) y)
        (##max (macro-cpxnum-real x) y)
        (if (macro-cpxnum-real? y)
          (##max (macro-cpxnum-real x) (macro-cpxnum-real y))
          (type-error)))
      (type-error))))

(define-nary (max x y)
  ()
  (if (##real? x) x '())
  (##max x y)
  macro-force-vars
  macro-no-check
  (##null? ##trap-check-real*))

(define (##min x y)

  (define (type-error) '())

  (macro-number-dispatch x (type-error)

    (macro-number-dispatch y (type-error) ; x = fixnum
      (##fixnum.min x y)
      (if (##< x y) x y)
      (if (##< x y) x y)
      (##flonum.min (##flonum.<-fixnum x) y)
      (if (macro-cpxnum-real? y)
        (##min x (macro-cpxnum-real y))
        (type-error)))

    (macro-number-dispatch y (type-error) ; x = bignum
      (if (##< x y) x y)
      (if (##< x y) x y)
      (if (##< x y) x y)
      (##flonum.min (##flonum.<-exact-int x) y)
      (if (macro-cpxnum-real? y)
        (##min x (macro-cpxnum-real y))
        (type-error)))

    (macro-number-dispatch y (type-error) ; x = ratnum
      (if (##< x y) x y)
      (if (##< x y) x y)
      (if (##< x y) x y)
      (##flonum.min (##flonum.<-ratnum x) y)
      (if (macro-cpxnum-real? y)
        (##min x (macro-cpxnum-real y))
        (type-error)))

    (macro-number-dispatch y (type-error) ; x = flonum
      (##flonum.min x (##flonum.<-fixnum y))
      (##flonum.min x (##flonum.<-exact-int y))
      (##flonum.min x (##flonum.<-ratnum y))
      (##flonum.min x y)
      (if (macro-cpxnum-real? y)
        (##min x (macro-cpxnum-real y))
        (type-error)))

    (if (macro-cpxnum-real? x) ; x = cpxnum
      (macro-number-dispatch y (type-error)
        (##min (macro-cpxnum-real x) y)
        (##min (macro-cpxnum-real x) y)
        (##min (macro-cpxnum-real x) y)
        (##min (macro-cpxnum-real x) y)
        (if (macro-cpxnum-real? y)
          (##min (macro-cpxnum-real x) (macro-cpxnum-real y))
          (type-error)))
      (type-error))))

(define-nary (min x y)
  ()
  (if (##real? x) x '())
  (##min x y)
  macro-force-vars
  macro-no-check
  (##null? ##trap-check-real*))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

; +, *, -, /

(define (##+ x y)

  (define (type-error) '())

  (macro-number-dispatch x (type-error)

    (macro-number-dispatch y (type-error) ; x = fixnum
      (if (##fixnum.negative? x)
        (if (##fixnum.negative? y)
          (let ((r (##fixnum.+ x y)))
            (if (##fixnum.negative? r)
              r
              (##bignum.+ (##bignum.<-fixnum r) ##bignum.2*min-fixnum)))
          (##fixnum.+ x y))
        (if (##fixnum.negative? y)
          (##fixnum.+ x y)
          (let ((r (##fixnum.+ x y)))
            (if (##fixnum.negative? r)
              (##bignum.- (##bignum.<-fixnum r) ##bignum.2*min-fixnum)
              r))))
      (##bignum.+ (##bignum.<-fixnum x) y)
      (##ratnum.+ (##ratnum.<-exact-int x) y)
      (if (and (macro-special-case-exact-zero?) (##fixnum.zero? x))
        y
        (##flonum.+ (##flonum.<-fixnum x) y))
      (##cpxnum.+ (##cpxnum.<-noncpxnum x) y))

    (macro-number-dispatch y (type-error) ; x = bignum
      (##bignum.+ x (##bignum.<-fixnum y))
      (##bignum.+ x y)
      (##ratnum.+ (##ratnum.<-exact-int x) y)
      (##flonum.+ (##flonum.<-exact-int x) y)
      (##cpxnum.+ (##cpxnum.<-noncpxnum x) y))

    (macro-number-dispatch y (type-error) ; x = ratnum
      (##ratnum.+ x (##ratnum.<-exact-int y))
      (##ratnum.+ x (##ratnum.<-exact-int y))
      (##ratnum.+ x y)
      (##flonum.+ (##flonum.<-ratnum x) y)
      (##cpxnum.+ (##cpxnum.<-noncpxnum x) y))

    (macro-number-dispatch y (type-error) ; x = flonum
      (if (and (macro-special-case-exact-zero?) (##fixnum.zero? y))
        x
        (##flonum.+ x (##flonum.<-fixnum y)))
      (##flonum.+ x (##flonum.<-exact-int y))
      (##flonum.+ x (##flonum.<-ratnum y))
      (##flonum.+ x y)
      (##cpxnum.+ (##cpxnum.<-noncpxnum x) y))

    (macro-number-dispatch y (type-error) ; x = cpxnum
      (##cpxnum.+ x (##cpxnum.<-noncpxnum y))
      (##cpxnum.+ x (##cpxnum.<-noncpxnum y))
      (##cpxnum.+ x (##cpxnum.<-noncpxnum y))
      (##cpxnum.+ x (##cpxnum.<-noncpxnum y))
      (##cpxnum.+ x y))))

(define-nary (+ x y)
  0
  (if (##complex? x) x '())
  (##+ x y)
  macro-force-vars
  macro-no-check
  (##null? ##trap-check-number*))

(define (##* x y)

  (define (type-error) '())

  (macro-number-dispatch x (type-error)

    (macro-number-dispatch y (type-error) ; x = fixnum
      (if (or (##fixnum.< x ##bignum.min-fixnum-div-mdigit-base)
              (##fixnum.< ##bignum.max-fixnum-div-mdigit-base x)
              (##not (##fixnum.< ##bignum.minus-mdigit-base y))
              (##fixnum.< ##bignum.mdigit-base y))
        (##bignum.* (##bignum.<-fixnum x) (##bignum.<-fixnum y))
        (##fixnum.* x y))
      (##bignum.* (##bignum.<-fixnum x) y)
      (##ratnum.* (##ratnum.<-exact-int x) y)
      (if (and (macro-special-case-exact-zero?) (##fixnum.zero? x))
        x
        (##flonum.* (##flonum.<-fixnum x) y))
      (##cpxnum.* (##cpxnum.<-noncpxnum x) y))

    (macro-number-dispatch y (type-error) ; x = bignum
      (##bignum.* x (##bignum.<-fixnum y))
      (##bignum.* x y)
      (##ratnum.* (##ratnum.<-exact-int x) y)
      (##flonum.* (##flonum.<-exact-int x) y)
      (##cpxnum.* (##cpxnum.<-noncpxnum x) y))

    (macro-number-dispatch y (type-error) ; x = ratnum
      (##ratnum.* x (##ratnum.<-exact-int y))
      (##ratnum.* x (##ratnum.<-exact-int y))
      (##ratnum.* x y)
      (##flonum.* (##flonum.<-ratnum x) y)
      (##cpxnum.* (##cpxnum.<-noncpxnum x) y))

    (macro-number-dispatch y (type-error) ; x = flonum
      (if (and (macro-special-case-exact-zero?) (##fixnum.zero? y))
        y
        (##flonum.* x (##flonum.<-fixnum y)))
      (##flonum.* x (##flonum.<-exact-int y))
      (##flonum.* x (##flonum.<-ratnum y))
      (##flonum.* x y)
      (##cpxnum.* (##cpxnum.<-noncpxnum x) y))

    (macro-number-dispatch y (type-error) ; x = cpxnum
      (##cpxnum.* x (##cpxnum.<-noncpxnum y))
      (##cpxnum.* x (##cpxnum.<-noncpxnum y))
      (##cpxnum.* x (##cpxnum.<-noncpxnum y))
      (##cpxnum.* x (##cpxnum.<-noncpxnum y))
      (##cpxnum.* x y))))

(define-nary (* x y)
  1
  (if (##complex? x) x '())
  (##* x y)
  macro-force-vars
  macro-no-check
  (##null? ##trap-check-number*))

(define (##negate x)

  (define (type-error) '())

  (macro-number-dispatch x (type-error)
    (if (##fixnum.negative? x)
      (let ((r (##fixnum.- x)))
        (if (##fixnum.negative? r)
          (##bignum.- (##bignum.<-fixnum r) ##bignum.2*min-fixnum)
          r))
      (##fixnum.- x))
    (##bignum.- (##bignum.<-fixnum 0) x)
    (macro-ratnum-make (##negate (macro-ratnum-numerator x))
                       (macro-ratnum-denominator x))
    (##flonum.- x)
    (##make-rectangular (##negate (macro-cpxnum-real x))
                        (##negate (macro-cpxnum-imag x)))))

(define (##- x y)

  (define (type-error) '())

  (macro-number-dispatch x (type-error)

    (macro-number-dispatch y (type-error) ; x = fixnum
      (if (##fixnum.negative? x)
        (if (##fixnum.negative? y)
          (##fixnum.- x y)
          (let ((r (##fixnum.- x y)))
            (if (##fixnum.negative? r)
              r
              (##bignum.+ (##bignum.<-fixnum r) ##bignum.2*min-fixnum))))
        (if (##fixnum.negative? y)
          (let ((r (##fixnum.- x y)))
            (if (##fixnum.negative? r)
              (##bignum.- (##bignum.<-fixnum r) ##bignum.2*min-fixnum)
              r))
          (##fixnum.- x y)))
      (##bignum.- (##bignum.<-fixnum x) y)
      (##ratnum.- (##ratnum.<-exact-int x) y)
      (if (and (macro-special-case-exact-zero?) (##fixnum.zero? x))
        (##flonum.- y)
        (##flonum.- (##flonum.<-fixnum x) y))
      (##cpxnum.- (##cpxnum.<-noncpxnum x) y))

    (macro-number-dispatch y (type-error) ; x = bignum
      (##bignum.- x (##bignum.<-fixnum y))
      (##bignum.- x y)
      (##ratnum.- (##ratnum.<-exact-int x) y)
      (##flonum.- (##flonum.<-exact-int x) y)
      (##cpxnum.- (##cpxnum.<-noncpxnum x) y))

    (macro-number-dispatch y (type-error) ; x = ratnum
      (##ratnum.- x (##ratnum.<-exact-int y))
      (##ratnum.- x (##ratnum.<-exact-int y))
      (##ratnum.- x y)
      (##flonum.- (##flonum.<-ratnum x) y)
      (##cpxnum.- (##cpxnum.<-noncpxnum x) y))

    (macro-number-dispatch y (type-error) ; x = flonum
      (if (and (macro-special-case-exact-zero?) (##fixnum.zero? y))
        x
        (##flonum.- x (##flonum.<-fixnum y)))
      (##flonum.- x (##flonum.<-exact-int y))
      (##flonum.- x (##flonum.<-ratnum y))
      (##flonum.- x y)
      (##cpxnum.- (##cpxnum.<-noncpxnum x) y))

    (macro-number-dispatch y (type-error) ; x = cpxnum
      (##cpxnum.- x (##cpxnum.<-noncpxnum y))
      (##cpxnum.- x (##cpxnum.<-noncpxnum y))
      (##cpxnum.- x (##cpxnum.<-noncpxnum y))
      (##cpxnum.- x (##cpxnum.<-noncpxnum y))
      (##cpxnum.- x y))))

(define-nary (- x y)
  ()
  (##negate x)
  (##- x y)
  macro-force-vars
  macro-no-check
  (##null? ##trap-check-number*))

(define (##/ x y)

  (define (type-error) '())

  (define (divide-by-zero-error) #f)

  (macro-number-dispatch y (type-error)

    (macro-number-dispatch x (type-error) ; y = fixnum
      (if (##fixnum.zero? y)
        (divide-by-zero-error)
        (##ratnum./ (##ratnum.<-exact-int x) (##ratnum.<-exact-int y)))
      (if (##fixnum.zero? y)
        (divide-by-zero-error)
        (##ratnum./ (##ratnum.<-exact-int x) (##ratnum.<-exact-int y)))
      (if (##fixnum.zero? y)
        (divide-by-zero-error)
        (##ratnum./ x (##ratnum.<-exact-int y)))
      (if (##fixnum.zero? y)
        (divide-by-zero-error)
        (##flonum./ x (##flonum.<-fixnum y)))
      (if (##fixnum.zero? y)
        (divide-by-zero-error)
        (##cpxnum./ x (##cpxnum.<-noncpxnum y))))

    (macro-number-dispatch x (type-error) ; y = bignum
      (##ratnum./ (##ratnum.<-exact-int x) (##ratnum.<-exact-int y))
      (##ratnum./ (##ratnum.<-exact-int x) (##ratnum.<-exact-int y))
      (##ratnum./ x (##ratnum.<-exact-int y))
      (##flonum./ x (##flonum.<-exact-int y))
      (##cpxnum./ x (##cpxnum.<-noncpxnum y)))

    (macro-number-dispatch x (type-error) ; y = ratnum
      (##ratnum./ (##ratnum.<-exact-int x) y)
      (##ratnum./ (##ratnum.<-exact-int x) y)
      (##ratnum./ x y)
      (##flonum./ x (##flonum.<-ratnum y))
      (##cpxnum./ x (##cpxnum.<-noncpxnum y)))

    (macro-number-dispatch x (type-error) ; y = flonum, no error possible
      (if (and (macro-special-case-exact-zero?) (##fixnum.zero? x))
        x
        (##flonum./ (##flonum.<-fixnum x) y))
      (##flonum./ (##flonum.<-exact-int x) y)
      (##flonum./ (##flonum.<-ratnum x) y)
      (##flonum./ x y)
      (##cpxnum./ x (##cpxnum.<-noncpxnum y)))

    (macro-number-dispatch x (type-error) ; y = cpxnum
      (##cpxnum./ (##cpxnum.<-noncpxnum x) y)
      (##cpxnum./ (##cpxnum.<-noncpxnum x) y)
      (##cpxnum./ (##cpxnum.<-noncpxnum x) y)
      (##cpxnum./ (##cpxnum.<-noncpxnum x) y)
      (##cpxnum./ x y))))

(define-nary (/ x y)
  ()
  (##/ 1 x)
  (##/ x y)
  macro-force-vars
  macro-no-check
  (##null? ##trap-check-number*)
  (##not ##trap-divide-by-zero*))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

; abs

(define (##abs x)

  (define (type-error)
    (##trap-check-real 'abs x))

  (macro-number-dispatch x (type-error)
    (if (##fixnum.negative? x) (##negate x) x)
    (if (##bignum.negative? x) (##negate x) x)
    (macro-ratnum-make (##abs (macro-ratnum-numerator x))
                       (macro-ratnum-denominator x))
    (##flonum.abs x)
    (if (macro-cpxnum-real? x)
      (##make-rectangular (##abs (macro-cpxnum-real x))
                          (##abs (macro-cpxnum-imag x)))
      (type-error))))

(define (abs x) (macro-force-vars (x) (##abs x)))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

; quotient, remainder, modulo

(define (##quotient x y)

  (define (type-error)
    (##trap-check-integer 'quotient x y))

  (define (divide-by-zero-error)
    (##trap-divide-by-zero 'quotient x y))

  (define (exact-quotient x y)
    (##car (##exact-int.div x y)))

  (define (inexact-quotient x y)
    (let ((exact-y (##inexact->exact y)))
      (if (##eq? exact-y 0)
        (divide-by-zero-error)
        (##exact->inexact
         (##quotient (##inexact->exact x) exact-y)))))

  (macro-number-dispatch y (type-error)

    (macro-number-dispatch x (type-error) ; y = fixnum
      (cond ((##fixnum.= y 0)
             (divide-by-zero-error))
            ((##fixnum.= y -1) ; (quotient ##min-fixnum -1) is a bignum
             (##negate x))
            (else
             (##fixnum.quotient x y)))
      (cond ((##fixnum.= y 0)
             (divide-by-zero-error))
            (else
             (exact-quotient x y)))
      (type-error)
      (if (macro-flonum-int? x)
        (inexact-quotient x y)
        (type-error))
      (if (macro-cpxnum-int? x)
        (inexact-quotient x y)
        (type-error)))

    (macro-number-dispatch x (type-error) ; y = bignum
      (exact-quotient x y)
      (exact-quotient x y)
      (type-error)
      (if (macro-flonum-int? x)
        (inexact-quotient x y)
        (type-error))
      (if (macro-cpxnum-int? x)
        (inexact-quotient x y)
        (type-error)))

    (type-error) ; y = ratnum

    (macro-number-dispatch x (type-error) ; y = flonum
      (if (macro-flonum-int? y)
        (inexact-quotient x y)
        (type-error))
      (if (macro-flonum-int? y)
        (inexact-quotient x y)
        (type-error))
      (type-error)
      (if (macro-flonum-int? x)
        (if (macro-flonum-int? y)
          (inexact-quotient x y)
          (type-error))
        (type-error))
      (if (macro-cpxnum-int? x)
        (if (macro-flonum-int? y)
          (inexact-quotient x y)
          (type-error))
        (type-error)))

    (if (macro-cpxnum-int? y) ; y = cpxnum
      (macro-number-dispatch x (type-error)
        (inexact-quotient x y)
        (inexact-quotient x y)
        (type-error)
        (if (macro-flonum-int? x)
          (inexact-quotient x y)
          (type-error))
        (if (macro-cpxnum-int? x)
          (inexact-quotient x y)
          (type-error)))
      (type-error))))

(define (quotient x y) (macro-force-vars (x y) (##quotient x y)))

(define (##remainder x y)

  (define (type-error)
    (##trap-check-integer 'remainder x y))

  (define (divide-by-zero-error)
    (##trap-divide-by-zero 'remainder x y))

  (define (exact-remainder x y)
    (##cdr (##exact-int.div x y)))

  (define (inexact-remainder x y)
    (let ((exact-y (##inexact->exact y)))
      (if (##eq? exact-y 0)
        (divide-by-zero-error)
        (##exact->inexact
         (##remainder (##inexact->exact x) exact-y)))))

  (macro-number-dispatch y (type-error)

    (macro-number-dispatch x (type-error) ; y = fixnum
      (cond ((##fixnum.= y 0)
             (divide-by-zero-error))
            (else
             (##fixnum.remainder x y)))
      (cond ((##fixnum.= y 0)
             (divide-by-zero-error))
            (else
             (exact-remainder x y)))
      (type-error)
      (if (macro-flonum-int? x)
        (inexact-remainder x y)
        (type-error))
      (if (macro-cpxnum-int? x)
        (inexact-remainder x y)
        (type-error)))

    (macro-number-dispatch x (type-error) ; y = bignum
      (exact-remainder x y)
      (exact-remainder x y)
      (type-error)
      (if (macro-flonum-int? x)
        (inexact-remainder x y)
        (type-error))
      (if (macro-cpxnum-int? x)
        (inexact-remainder x y)
        (type-error)))

    (type-error) ; y = ratnum

    (macro-number-dispatch x (type-error) ; y = flonum
      (if (macro-flonum-int? y)
        (inexact-remainder x y)
        (type-error))
      (if (macro-flonum-int? y)
        (inexact-remainder x y)
        (type-error))
      (type-error)
      (if (macro-flonum-int? x)
        (if (macro-flonum-int? y)
          (inexact-remainder x y)
          (type-error))
        (type-error))
      (if (macro-cpxnum-int? x)
        (if (macro-flonum-int? y)
          (inexact-remainder x y)
          (type-error))
        (type-error)))

    (if (macro-cpxnum-int? y) ; y = cpxnum
      (macro-number-dispatch x (type-error)
        (inexact-remainder x y)
        (inexact-remainder x y)
        (type-error)
        (if (macro-flonum-int? x)
          (inexact-remainder x y)
          (type-error))
        (if (macro-cpxnum-int? x)
          (inexact-remainder x y)
          (type-error)))
      (type-error))))

(define (remainder x y) (macro-force-vars (x y) (##remainder x y)))

(define (##modulo x y)

  (define (type-error)
    (##trap-check-integer 'modulo x y))

  (define (divide-by-zero-error)
    (##trap-divide-by-zero 'modulo x y))

  (define (exact-modulo x y)
    (let ((r (##cdr (##exact-int.div x y))))
      (if (##eq? r 0)
        0
        (if (##eq? (##negative? x) (##negative? y))
          r
          (##+ r y)))))

  (define (inexact-modulo x y)
    (let ((exact-y (##inexact->exact y)))
      (if (##eq? exact-y 0)
        (divide-by-zero-error)
        (##exact->inexact
         (##modulo (##inexact->exact x) exact-y)))))

  (macro-number-dispatch y (type-error)

    (macro-number-dispatch x (type-error) ; y = fixnum
      (cond ((##fixnum.= y 0)
             (divide-by-zero-error))
            (else
             (##fixnum.modulo x y)))
      (cond ((##fixnum.= y 0)
             (divide-by-zero-error))
            (else
             (exact-modulo x y)))
      (type-error)
      (if (macro-flonum-int? x)
        (inexact-modulo x y)
        (type-error))
      (if (macro-cpxnum-int? x)
        (inexact-modulo x y)
        (type-error)))

    (macro-number-dispatch x (type-error) ; y = bignum
      (exact-modulo x y)
      (exact-modulo x y)
      (type-error)
      (if (macro-flonum-int? x)
        (inexact-modulo x y)
        (type-error))
      (if (macro-cpxnum-int? x)
        (inexact-modulo x y)
        (type-error)))

    (type-error) ; y = ratnum

    (macro-number-dispatch x (type-error) ; y = flonum
      (if (macro-flonum-int? y)
        (inexact-modulo x y)
        (type-error))
      (if (macro-flonum-int? y)
        (inexact-modulo x y)
        (type-error))
      (type-error)
      (if (macro-flonum-int? x)
        (if (macro-flonum-int? y)
          (inexact-modulo x y)
          (type-error))
        (type-error))
      (if (macro-cpxnum-int? x)
        (if (macro-flonum-int? y)
          (inexact-modulo x y)
          (type-error))
        (type-error)))

    (if (macro-cpxnum-int? y) ; y = cpxnum
      (macro-number-dispatch x (type-error)
        (inexact-modulo x y)
        (inexact-modulo x y)
        (type-error)
        (if (macro-flonum-int? x)
          (inexact-modulo x y)
          (type-error))
        (if (macro-cpxnum-int? x)
          (inexact-modulo x y)
          (type-error)))
      (type-error))))

(define (modulo x y) (macro-force-vars (x y) (##modulo x y)))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

; gcd, lcm

(define (##gcd x y)

  (define (type-error) '())

  (define (exact-gcd x y)
    (let loop ((x (##abs x))
               (y (##abs y)))
      (if (##eq? y 0)
        x
        (loop y
              (##remainder x y)))))

  (define (inexact-gcd x y)
    (##exact->inexact
     (exact-gcd (##inexact->exact x)
                (##inexact->exact y))))

  (if (and (##integer? x) (##integer? y))
    (if (and (##exact? x) (##exact? y))
      (exact-gcd x y)
      (inexact-gcd x y))
    (type-error)))

(define-nary (gcd x y)
  0
  (if (##integer? x) x '())
  (##gcd x y)
  macro-force-vars
  macro-no-check
  (##null? ##trap-check-integer*))

(define (##lcm x y)

  (define (type-error) '())

  (define (exact-lcm x y)
    (if (or (##eq? x 0) (##eq? y 0))
      0
      (##quotient (##abs (##* x y))
                  (##gcd x y))))

  (define (inexact-lcm x y)
    (##exact->inexact
     (exact-lcm (##inexact->exact x)
                (##inexact->exact y))))

  (if (and (##integer? x) (##integer? y))
    (if (and (##exact? x) (##exact? y))
      (exact-lcm x y)
      (inexact-lcm x y))
    (type-error)))

(define-nary (lcm x y)
  1
  (if (##integer? x) x '())
  (##lcm x y)
  macro-force-vars
  macro-no-check
  (##null? ##trap-check-integer*))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

; numerator, denominator

(define (##numerator x)

  (define (type-error)
    (##trap-check-rational 'numerator x))

  (macro-number-dispatch x (type-error)
    x
    x
    (macro-ratnum-numerator x)
    (cond ((##flonum.zero? x)
           x)
          ((macro-flonum-rational? x)
           (##exact->inexact (##numerator (##flonum.inexact->exact x))))
          (else
           (type-error)))
    (if (macro-cpxnum-rational? x)
      (##numerator (macro-cpxnum-real x))
      (type-error))))

(define (numerator x) (macro-force-vars (x) (##numerator x)))

(define (##denominator x)

  (define (type-error)
    (##trap-check-rational 'denominator x))

  (macro-number-dispatch x (type-error)
    1
    1
    (macro-ratnum-denominator x)
    (if (macro-flonum-rational? x)
      (##exact->inexact (##denominator (##flonum.inexact->exact x)))
      (type-error))
    (if (macro-cpxnum-rational? x)
      (##denominator (macro-cpxnum-real x))
      (type-error))))

(define (denominator x) (macro-force-vars (x) (##denominator x)))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

; floor, ceiling, truncate, round

(define (##floor x)

  (define (type-error)
    (##trap-check-finite-real 'floor x))

  (macro-number-dispatch x (type-error)
    x
    x
    (let ((num (macro-ratnum-numerator x))
          (den (macro-ratnum-denominator x)))
      (if (##negative? num)
        (##quotient (##- num (##- den 1)) den)
        (##quotient num den)))
    (if (##flonum.finite? x)
      (##flonum.floor x)
      (type-error))
    (if (macro-cpxnum-real? x)
      (##floor (macro-cpxnum-real x))
      (type-error))))

(define (floor x) (macro-force-vars (x) (##floor x)))

(define (##ceiling x)

  (define (type-error)
    (##trap-check-finite-real 'ceiling x))

  (macro-number-dispatch x (type-error)
    x
    x
    (let ((num (macro-ratnum-numerator x))
          (den (macro-ratnum-denominator x)))
      (if (##negative? num)
        (##quotient num den)
        (##quotient (##+ num (##- den 1)) den)))
    (if (##flonum.finite? x)
      (##flonum.ceiling x)
      (type-error))
    (if (macro-cpxnum-real? x)
      (##ceiling (macro-cpxnum-real x))
      (type-error))))

(define (ceiling x) (macro-force-vars (x) (##ceiling x)))

(define (##truncate x)

  (define (type-error)
    (##trap-check-finite-real 'truncate x))

  (macro-number-dispatch x (type-error)
    x
    x
    (##quotient (macro-ratnum-numerator x)
                (macro-ratnum-denominator x))
    (if (##flonum.finite? x)
      (##flonum.truncate x)
      (type-error))
    (if (macro-cpxnum-real? x)
      (##truncate (macro-cpxnum-real x))
      (type-error))))

(define (truncate x) (macro-force-vars (x) (##truncate x)))

(define (##round x)

  (define (type-error)
    (##trap-check-finite-real 'round x))

  (macro-number-dispatch x (type-error)
    x
    x
    (let ((num (macro-ratnum-numerator x))
          (den (macro-ratnum-denominator x)))
      (if (##eq? den 2)
        (##exact-int.shift (##exact-int.shift (##+ num 1) -2) 1)
        (##floor
         (##ratnum.normalize
          (##+ (##exact-int.shift num 1) den)
          (##exact-int.shift den 1)))))
    (if (##flonum.finite? x)
      (##flonum.round x)
      (type-error))
    (if (macro-cpxnum-real? x)
      (##round (macro-cpxnum-real x))
      (type-error))))

(define (round x) (macro-force-vars (x) (##round x)))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

; rationalize

(define (##rationalize x y)

  (define (type-error)
    (##trap-check-finite-real 'rationalize x y))

  (define (simplest-rational1 x y)
    (if (##< y x)
      (simplest-rational2 y x)
      (simplest-rational2 x y)))

  (define (simplest-rational2 x y)
    (cond ((##not (##< x y))
           x)
          ((##positive? x)
           (simplest-rational3 x y))
          ((##negative? y)
           (##negate (simplest-rational3 (##negate y) (##negate x))))
          (else
           0)))

  (define (simplest-rational3 x y)
    (let ((fx (##floor x))
          (fy (##floor y)))
      (cond ((##not (##< fx x))
             fx)
            ((##= fx fy)
             (##+ fx
                  (##/ 1
                       (simplest-rational3
                        (##/ 1 (##- y fy))
                        (##/ 1 (##- x fx))))))
            (else
             (##+ fx 1)))))

  (cond ((##not (##rational? x)) 
         (type-error))
        ((and (##flonum? y)
              (##flonum.= y (macro-inexact-+inf)))
         (macro-inexact-+0))
        ((or (##not (##rational? y))
             (##negative? y))
         (##trap-check-nonnegative-real 'rationalize x y))
        ((and (##exact? x) (##exact? y))
         (simplest-rational1 (##- x y) (##+ x y)))
        (else
         (let ((exact-x (##inexact->exact x))
               (exact-y (##inexact->exact y)))
           (##exact->inexact
            (simplest-rational1 (##- exact-x exact-y)
                                (##+ exact-x exact-y)))))))

(define (rationalize x y) (macro-force-vars (x y) (##rationalize x y)))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

; trigonometry and complex numbers

(define (##exp x)

  (define (type-error)
    (##trap-check-number 'exp x))

  (macro-number-dispatch x (type-error)
    (if (##fixnum.zero? x)
      1
      (##flonum.exp (##flonum.<-fixnum x)))
    (##flonum.exp (##flonum.<-exact-int x))
    (##flonum.exp (##flonum.<-ratnum x))
    (##flonum.exp x)
    (##make-polar (##exp (macro-cpxnum-real x))
                  (macro-cpxnum-imag x))))

(define (exp x) (macro-force-vars (x) (##exp x)))

(define (##log x)

  (define (type-error)
    (##trap-check-number 'log x))

  (define (range-error)
    (##trap-check-range 'log x))

  (define (negative-log x)
    (##make-rectangular (##log (##negate x)) (macro-inexact-+pi)))

  (macro-number-dispatch x (type-error)
    (if (##fixnum.zero? x)
      (range-error)
      (if (##fixnum.negative? x)
        (negative-log x)
        (if (##eq? x 1)
          0
          (##flonum.log (##flonum.<-fixnum x)))))
    (if (##bignum.negative? x)
      (negative-log x)
      (##flonum.log (##flonum.<-exact-int x)))
    (if (##negative? (macro-ratnum-numerator x))
      (negative-log x)
      (##flonum.log (##flonum.<-ratnum x)))
    (if (and (##not (##flonum.nan? x))
             (##flonum.negative? (##flonum.copysign (macro-inexact-+1) x)))
      (negative-log x)
      (##flonum.log x))
    (##make-rectangular (##log (##magnitude x)) (##angle x))))

(define (log x) (macro-force-vars (x) (##log x)))

(define (##sin x)

  (define (type-error)
    (##trap-check-number 'sin x))

  (macro-number-dispatch x (type-error)
    (if (##fixnum.zero? x)
      0
      (##flonum.sin (##flonum.<-fixnum x)))
    (##flonum.sin (##flonum.<-exact-int x))
    (##flonum.sin (##flonum.<-ratnum x))
    (##flonum.sin x)
    (##/ (##- (##exp (##make-rectangular
                      (##negate (macro-cpxnum-imag x))
                      (macro-cpxnum-real x)))
              (##exp (##make-rectangular
                      (macro-cpxnum-imag x)
                      (##negate (macro-cpxnum-real x)))))
         (macro-cpxnum-+2i))))

(define (sin x) (macro-force-vars (x) (##sin x)))

(define (##cos x)

  (define (type-error)
    (##trap-check-number 'cos x))

  (macro-number-dispatch x (type-error)
    (if (##fixnum.zero? x)
      1
      (##flonum.cos (##flonum.<-fixnum x)))
    (##flonum.cos (##flonum.<-exact-int x))
    (##flonum.cos (##flonum.<-ratnum x))
    (##flonum.cos x)
    (##/ (##+ (##exp (##make-rectangular
                      (##negate (macro-cpxnum-imag x))
                      (macro-cpxnum-real x)))
              (##exp (##make-rectangular
                      (macro-cpxnum-imag x)
                      (##negate (macro-cpxnum-real x)))))
         2)))

(define (cos x) (macro-force-vars (x) (##cos x)))

(define (##tan x)

  (define (type-error)
    (##trap-check-number 'tan x))

  (macro-number-dispatch x (type-error)
    (if (##fixnum.zero? x)
      0
      (##flonum.tan (##flonum.<-fixnum x)))
    (##flonum.tan (##flonum.<-exact-int x))
    (##flonum.tan (##flonum.<-ratnum x))
    (##flonum.tan x)
    (let ((a (##exp (##make-rectangular
                     (##negate (macro-cpxnum-imag x))
                     (macro-cpxnum-real x))))
          (b (##exp (##make-rectangular
                     (macro-cpxnum-imag x)
                     (##negate (macro-cpxnum-real x))))))
      (let ((c (##/ (##- a b) (##+ a b))))
        (##make-rectangular (##imag-part c) (##negate (##real-part c)))))))

(define (tan x) (macro-force-vars (x) (##tan x)))

(define (##asin x)

  (define (type-error)
    (##trap-check-number 'asin x))

  (define (safe-case x)
    (##* (macro-cpxnum--i)
         (##log (##+ (##* (macro-cpxnum-+i) x)
                     (##sqrt (##- 1 (##* x x)))))))

  (define (unsafe-case x)
    (##negate (safe-case (##negate x))))

  (define (real-case x)
    (cond ((##< x -1)
           (unsafe-case x))
          ((##< 1 x)
           (safe-case x))
          (else
           (##flonum.asin (##exact->inexact x)))))

  (macro-number-dispatch x (type-error)
    (if (##fixnum.zero? x)
      0
      (real-case x))
    (real-case x)
    (real-case x)
    (real-case x)
    (let ((imag (macro-cpxnum-imag x)))
      (if (or (##positive? imag)
              (and (##flonum? imag)
                   (##flonum.zero? imag)
                   (##negative? (macro-cpxnum-real x))))
        (unsafe-case x)
        (safe-case x)))))

(define (asin x) (macro-force-vars (x) (##asin x)))

(define (##acos x)

  (define (type-error)
    (##trap-check-number 'acos x))

  (define (complex-case x)
    (##* (macro-cpxnum--i)
         (##log (##+ x
                     (##* (macro-cpxnum-+i) (##sqrt (##- 1 (##* x x))))))))

  (define (real-case x)
    (if (or (##< x -1) (##< 1 x))
      (complex-case x)
      (##flonum.acos (##exact->inexact x))))

  (macro-number-dispatch x (type-error)
    (if (##fixnum.zero? x)
      (macro-inexact-+pi/2)
      (real-case x))
    (real-case x)
    (real-case x)
    (real-case x)
    (complex-case x)))

(define (acos x) (macro-force-vars (x) (##acos x)))

(define (##atan x)

  (define (type-error)
    (##trap-check-number 'atan x))

  (define (range-error)
    (##trap-check-range 'atan x))

  (macro-number-dispatch x (type-error)
    (if (##fixnum.zero? x)
      0
      (##flonum.atan (##flonum.<-fixnum x)))
    (##flonum.atan (##flonum.<-exact-int x))
    (##flonum.atan (##flonum.<-ratnum x))
    (##flonum.atan x)
    (let ((real (macro-cpxnum-real x))
          (imag (macro-cpxnum-imag x)))
      (if (and (##eq? real 0) (##eq? imag 1))
        (range-error)
        (let ((a (##make-rectangular (##negate imag) real)))
          (##/ (##- (##log (##+ a 1)) (##log (##- 1 a)))
               (macro-cpxnum-+2i)))))))

(define (##atan2 y x)

  (define (type-error)
    (##trap-check-real 'atan y x))

  (if (and (##real? x) (##real? y))
    (##flonum.atan (##exact->inexact (##real-part y))
                   (##exact->inexact (##real-part x)))
    (type-error)))

(define (atan x #!optional (y (macro-absent-obj)))
  (macro-force-vars (x)
    (if (##eq? y (macro-absent-obj))
      (##atan x)
      (macro-force-vars (y)
        (##atan2 x y)))))

(define (##sqrt x)

  (define (type-error)
    (##trap-check-number 'sqrt x))

  (define (exact-int-sqrt x)
    (cond ((##eq? x 0)
           0)
          ((##negative? x)
           (##make-rectangular 0 (exact-int-sqrt (##negate x))))
          (else
           (let ((y (##exact-int.root x 2)))
             (if (##= x (##exact-int.square y))
               y
               (##flonum.sqrt (##flonum.<-exact-int x)))))))

  (macro-number-dispatch x (type-error)
    (exact-int-sqrt x)
    (exact-int-sqrt x)
    (##/ (exact-int-sqrt (macro-ratnum-numerator x))
         (exact-int-sqrt (macro-ratnum-denominator x)))
    (if (##flonum.negative? x)
      (##make-rectangular 0 (##flonum.sqrt (##flonum.- x)))
      (##flonum.sqrt x))
    (let ((real (##real-part x))
          (imag (##imag-part x)))
      (if (and (##flonum? imag)
               (##flonum.zero? imag))
        (if (##flonum.positive? (##flonum.copysign (macro-inexact-+1) imag))
          (cond ((##negative? real)
                 (##make-rectangular (macro-inexact-+0) 
                                     (##exact->inexact 
                                      (##sqrt (##negate real)))))
                ((and (##flonum? real)
                      (##flonum.nan? real))
                 (##make-rectangular real real))
                (else
                 (##make-rectangular (##exact->inexact (##sqrt real))
                                     (macro-inexact-+0))))
          (cond ((##negative? real)
                 (##make-rectangular (macro-inexact-+0) 
                                     (##exact->inexact 
                                      (##negate (##sqrt (##negate real))))))
                ((and (##flonum? real)
                      (##flonum.nan? real))
                 (##make-rectangular real real))
                (else
                 (##make-rectangular (##exact->inexact (##sqrt real))
                                     (macro-inexact--0)))))
        (##make-polar (##sqrt (##magnitude x))
                      (##/ (##angle x) 2))))))

(define (sqrt x) (macro-force-vars (x) (##sqrt x)))

(define (##expt x y)

  (define (type-error)
    (##trap-check-number 'expt x y))

  (define (exact-int-expt x y)
    ; x is a number, y is an exact-int
    ; y >= 0
    (cond ((##eq? y 0)
           1)
          ((or (##zero? x) (##= x 1))
           x)
          (else
           (let loop ((x x) (y y) (result 1))
             (if (##eq? y 1)
               (##* x result)
               (loop (##* x x)
                     (##exact-int.shift y -1)
                     (if (##odd? y)
                       (##* x result)
                       result)))))))

  (define (real-expt x y) ; y > 0
    (cond ((##zero? y)
           (if (##eq? y 0) 1 (macro-inexact-+1)))
          ((##zero? x)
           (if (##eq? x 0) 0 (macro-inexact-+0)))
          ((##eq? x 1)
           1)
          ((and (##flonum? y) (macro-flonum-int? y))
           (let ((result (exact-int-expt x (##flonum.->exact-int y))))
             (if (##real? result)
               (##exact->inexact result)
               (let ((real (##real-part result))
                     (imag (##imag-part result)))
                 (##make-rectangular (if (##eq? real 0)
                                       0
                                       (##exact->inexact real))
                                     (##exact->inexact imag))))))
          (else
           (complex-expt x y))))

  (define (complex-expt x y)
    (##exp (##* (##log x) y)))

  (define (invert z)
    (let ((result (##/ 1 z)))
      (if (##not result)
        (##trap-check-range 'expt x y)
        result)))

  (if (##complex? x)
    (macro-number-dispatch y (type-error)
      (if (and (##flonum? x) (##flonum.nan? x))
        x
        (if (##fixnum.negative? y)
          (invert (exact-int-expt x (##negate y)))
          (exact-int-expt x y)))
      (if (and (##flonum? x) (##flonum.nan? x))
        x
        (if (##bignum.negative? y)
          (invert (exact-int-expt x (##negate y)))
          (exact-int-expt x y)))
      (if (and (##flonum? x) (##flonum.nan? x))
        x
        (if (##negative? (macro-ratnum-numerator y))
          (invert (real-expt x (##negate y)))
          (real-expt x y)))
      (if (and (##flonum? x) (##flonum.nan? x))
        x
        (if (##flonum.nan? y)
          y
          (if (##flonum.negative? y)
            (invert (real-expt x (##flonum.- y)))
            (real-expt x y))))
      (if (and (##flonum? x) (##flonum.nan? x))
        x
        (complex-expt x y)))
    (type-error)))

(define (expt x y) (macro-force-vars (x y) (##expt x y)))

(define (##make-rectangular x y)

  (define (type-error)
    (##trap-check-real 'make-rectangular x y))

  (if (and (##real? x) (##real? y))
    (let ((real (##real-part x))
          (imag (##real-part y)))
      (if (##eq? imag 0)
        real
        (macro-cpxnum-make real imag)))
    (type-error)))

(define (make-rectangular x y)
  (macro-force-vars (x y) (##make-rectangular x y)))

(define (##make-polar x y)

  (define (type-error)
    (##trap-check-real 'make-polar x y))

  (if (and (##real? x) (##real? y))
    (let ((real-x (##real-part x))
          (real-y (##real-part y)))
      (##make-rectangular (##* real-x (##cos real-y))
                          (##* real-x (##sin real-y))))
    (type-error)))

(define (make-polar x y) (macro-force-vars (x y) (##make-polar x y)))

(define (##real-part x)

  (define (type-error)
    (##trap-check-number 'real-part x))

  (macro-number-dispatch x (type-error)
    x x x x (macro-cpxnum-real x)))

(define (real-part x) (macro-force-vars (x) (##real-part x)))

(define (##imag-part x)

  (define (type-error)
    (##trap-check-number 'imag-part x))

  (macro-number-dispatch x (type-error)
    0 0 0 0 (macro-cpxnum-imag x)))

(define (imag-part x) (macro-force-vars (x) (##imag-part x)))

(define (##magnitude x)

  (define (type-error)
    (##trap-check-number 'magnitude x))

  (macro-number-dispatch x (type-error)
    (if (##fixnum.negative? x) (##negate x) x)
    (if (##bignum.negative? x) (##negate x) x)
    (macro-ratnum-make (##abs (macro-ratnum-numerator x))
                       (macro-ratnum-denominator x))
    (##flonum.abs x)
    (let ((abs-r (##abs (##real-part x)))
          (abs-i (##abs (##imag-part x))))

      (define (complex-magn a b)
        (cond ((##eq? a 0)
               b)
              ((and (##flonum? a) (##flonum.zero? a))
               (##exact->inexact b))
              (else
               (let ((c (##/ a b)))
                 (##* b (##sqrt (##+ (##* c c) 1)))))))

      (cond ((or (and (##flonum? abs-r)
                      (##flonum.= abs-r (macro-inexact-+inf)))
                 (and (##flonum? abs-i)
                      (##flonum.= abs-i (macro-inexact-+inf))))
             (macro-inexact-+inf))
            ((and (##flonum? abs-r) (##flonum.nan? abs-r))
             abs-r)
            ((and (##flonum? abs-i) (##flonum.nan? abs-i))
             abs-i)
            (else
             (if (##< abs-r abs-i)
               (complex-magn abs-r abs-i)
               (complex-magn abs-i abs-r)))))))

(define (magnitude x) (macro-force-vars (x) (##magnitude x)))

(define (##angle x)

  (define (type-error)
    (##trap-check-number 'angle x))

  (macro-number-dispatch x (type-error)
    (if (##fixnum.negative? x)
      (macro-inexact-+pi)
      0)
    (if (##bignum.negative? x)
      (macro-inexact-+pi)
      0)
    (if (##negative? (macro-ratnum-numerator x))
      (macro-inexact-+pi)
      0)
    (if (##flonum.negative? (##flonum.copysign (macro-inexact-+1) x))
      (macro-inexact-+pi)
      (macro-inexact-+0))
    (##atan2 (macro-cpxnum-imag x) (macro-cpxnum-real x))))

(define (angle x) (macro-force-vars (x) (##angle x)))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

; exact->inexact, inexact->exact

(define (##exact->inexact x)

  (define (type-error)
    (##trap-check-number 'exact->inexact x))

  (macro-number-dispatch x (type-error)
    (##flonum.<-fixnum x)
    (##flonum.<-exact-int x)
    (##flonum.<-ratnum x)
    x
    (##make-rectangular (##exact->inexact (macro-cpxnum-real x))
                        (##exact->inexact (macro-cpxnum-imag x)))))

(define (exact->inexact x)
  (macro-force-vars (x) (##exact->inexact x)))

(define (##inexact->exact x)

  (define (type-error)
    (##trap-check-number 'inexact->exact x))

  (define (range-error)
    (##trap-check-range 'inexact->exact x))

  (macro-number-dispatch x (type-error)
    x
    x
    x
    (if (macro-flonum-rational? x)
      (##flonum.inexact->exact x)
      (range-error))
    (let ((real (macro-cpxnum-real x))
          (imag (macro-cpxnum-imag x)))
      (if (and (macro-noncpxnum-rational? real)
               (macro-noncpxnum-rational? imag))
        (##make-rectangular (##inexact->exact real)
                            (##inexact->exact imag))
        (range-error)))))

(define (inexact->exact x)
  (macro-force-vars (x)
    (##inexact->exact x)))

;------------------------------------------------------------------------------



;------------------------------------------------------------------------------

; Numerical type predicates.

(define (##complex? x)
  (macro-number-dispatch x #f
    #t ; x = fixnum
    #t ; x = bignum
    #t ; x = ratnum
    #t ; x = flonum
    #t)) ; x = cpxnum

(define (number? x)  (macro-force-vars (x) (##complex? x)))
(define (complex? x) (macro-force-vars (x) (##complex? x)))

(define (##real? x)
  (macro-number-dispatch x #f
    #t ; x = fixnum
    #t ; x = bignum
    #t ; x = ratnum
    #t ; x = flonum
    (macro-cpxnum-real? x))) ; x = cpxnum

(define (real? x) (macro-force-vars (x) (##real? x)))

(define (##rational? x)
  (macro-number-dispatch x #f
    #t ; x = fixnum
    #t ; x = bignum
    #t ; x = ratnum
    (macro-flonum-rational? x) ; x = flonum
    (macro-cpxnum-rational? x))) ; x = cpxnum

(define (rational? x) (macro-force-vars (x) (##rational? x)))

(define (##integer? x)
  (macro-number-dispatch x #f
    #t ; x = fixnum
    #t ; x = bignum
    #f ; x = ratnum
    (macro-flonum-int? x) ; x = flonum
    (macro-cpxnum-int? x))) ; x = cpxnum

(define (integer? x) (macro-force-vars (x) (##integer? x)))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

; Exactness predicates.

(define (##exact? x)

  (define (type-error) #f)

  (macro-number-dispatch x (type-error)
    #t ; x = fixnum
    #t ; x = bignum
    #t ; x = ratnum
    #f ; x = flonum
    (and (##not (##flonum? (macro-cpxnum-real x))) ; x = cpxnum
         (##not (##flonum? (macro-cpxnum-imag x))))))

(define (exact? x)
  (macro-force-vars (x)
    (let ()

      (define (type-error)
        (##trap-check-number 'exact? x))

      (macro-number-dispatch x (type-error)
        #t ; x = fixnum
        #t ; x = bignum
        #t ; x = ratnum
        #f ; x = flonum
        (and (##not (##flonum? (macro-cpxnum-real x))) ; x = cpxnum
             (##not (##flonum? (macro-cpxnum-imag x))))))))

(define (inexact? x)
  (macro-force-vars (x)
    (let ()

      (define (type-error)
        (##trap-check-number 'inexact? x))

      (macro-number-dispatch x (type-error)
        #f ; x = fixnum
        #f ; x = bignum
        #f ; x = ratnum
        #t ; x = flonum
        (or (##flonum? (macro-cpxnum-real x)) ; x = cpxnum
            (##flonum? (macro-cpxnum-imag x)))))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

; Numerical comparison predicates.

(define (##zero? x)

  (define (type-error)
    (##trap-check-number 'zero? x))

  (macro-number-dispatch x (type-error)
    (##fixnum.zero? x)
    #f
    #f
    (##flonum.zero? x)
    (and (let ((imag (macro-cpxnum-imag x)))
           (and (##flonum? imag) (##flonum.zero? imag)))
         (let ((real (macro-cpxnum-real x)))
           (if (##fixnum? real)
             (##fixnum.zero? real)
             (and (##flonum? real) (##flonum.zero? real)))))))

(define (zero? x) (macro-force-vars (x) (##zero? x)))

(define (##positive? x)

  (define (type-error)
    (##trap-check-real 'positive? x))

  (macro-number-dispatch x (type-error)
    (##fixnum.positive? x)
    (##not (##bignum.negative? x))
    (##positive? (macro-ratnum-numerator x))
    (##flonum.positive? x)
    (if (macro-cpxnum-real? x)
      (##positive? (macro-cpxnum-real x))
      (type-error))))

(define (positive? x) (macro-force-vars (x) (##positive? x)))

(define (##negative? x)

  (define (type-error)
    (##trap-check-real 'negative? x))

  (macro-number-dispatch x (type-error)
    (##fixnum.negative? x)
    (##bignum.negative? x)
    (##negative? (macro-ratnum-numerator x))
    (##flonum.negative? x)
    (if (macro-cpxnum-real? x)
      (##negative? (macro-cpxnum-real x))
      (type-error))))

(define (negative? x) (macro-force-vars (x) (##negative? x)))

(define (##odd? x)

  (define (type-error)
    (##trap-check-integer 'odd? x))

  (macro-number-dispatch x (type-error)
    (##fixnum.odd? x)
    (macro-bignum-odd? x)
    (type-error)
    (if (macro-flonum-int? x)
      (##odd? (##flonum.->exact-int x))
      (type-error))
    (if (macro-cpxnum-int? x)
      (##odd? (##inexact->exact (macro-cpxnum-real x)))
      (type-error))))

(define (odd? x) (macro-force-vars (x) (##odd? x)))

(define (##even? x)

  (define (type-error)
    (##trap-check-integer 'even? x))

  (macro-number-dispatch x (type-error)
    (##not (##fixnum.odd? x))
    (##not (macro-bignum-odd? x))
    (type-error)
    (if (macro-flonum-int? x)
      (##even? (##flonum.->exact-int x))
      (type-error))
    (if (macro-cpxnum-int? x)
      (##even? (##inexact->exact (macro-cpxnum-real x)))
      (type-error))))

(define (even? x) (macro-force-vars (x) (##even? x)))

;------------------------------------------------------------------------------

; number->string, string->number

(define (##exact-int.number->string x rad force-sign?)

  (##define-macro (macro-make-block-size)
    (let* ((max-rad 16)
           (t (make-vector (+ max-rad 1) 0)))

      (define max-fixnum 536870911) ; OK to be conservative

      (define (block-size-for rad)
        (let loop ((i 0) (rad^i 1))
          (let ((new-rad^i (* rad^i rad)))
            (if (<= new-rad^i max-fixnum)
              (loop (+ i 1) new-rad^i)
              i))))

      (let loop ((i max-rad))
        (if (< 1 i)
          (begin
            (vector-set! t i (block-size-for i))
            (loop (- i 1)))))

      `',t))

  (define block-size (macro-make-block-size))

  (##define-macro (macro-make-rad^block-size)
    (let* ((max-rad 16)
           (t (make-vector (+ max-rad 1) 0)))

      (define max-fixnum 536870911) ; OK to be conservative

      (define (rad^block-size-for rad)
        (let loop ((i 0) (rad^i 1))
          (let ((new-rad^i (* rad^i rad)))
            (if (<= new-rad^i max-fixnum)
              (loop (+ i 1) new-rad^i)
              rad^i))))

      (let loop ((i max-rad))
        (if (< 1 i)
          (begin
            (vector-set! t i (rad^block-size-for i))
            (loop (- i 1)))))

      `',t))

  (define rad^block-size (macro-make-rad^block-size))

  (define (make-string-from-last-fixnum rad x len pos)
    (let loop ((x x) (len len) (pos pos))
      (if (##fixnum.= x 0)
        (##make-string len)
        (let* ((new-pos
                (##fixnum.+ pos 1))
               (s
                (loop (##fixnum.quotient x rad)
                      (##fixnum.+ len 1)
                      new-pos)))
          (##string-set!
           s
           (##fixnum.- (##string-length s) new-pos)
           (##string-ref ##digit-to-char-table
                         (##fixnum.- (##fixnum.remainder x rad))))
          s))))

  (define (convert-non-last-fixnum s rad x pos)
    (let loop ((x x)
               (size (##vector-ref block-size rad))
               (i (##fixnum.- (##string-length s) pos)))
      (if (##fixnum.< 0 size)
        (let ((new-i (##fixnum.- i 1)))
          (##string-set!
           s
           new-i
           (##string-ref ##digit-to-char-table
                         (##fixnum.remainder x rad)))
          (loop (##fixnum.quotient x rad)
                (##fixnum.- size 1)
                new-i)))))

  (define (make-string-from-fixnums rad lst len pos)
    (let loop ((lst lst) (pos pos))
      (let ((new-lst (##cdr lst)))
        (if (##null? new-lst)
          (make-string-from-last-fixnum
           rad
           (##fixnum.- (##car lst))
           (##fixnum.+ len pos)
           pos)
          (let* ((size
                  (##vector-ref block-size rad))
                 (new-pos
                  (##fixnum.+ pos size))
                 (s
                  (loop new-lst new-pos)))
            (convert-non-last-fixnum s rad (##car lst) pos)
            s)))))

  (define (uinteger->fixnums level sqs x lst)
    (cond ((and (##null? lst) (##eq? x 0))
           lst)
          ((##fixnum.= level 0)
           (##cons x lst))
          (else
           (let* ((qr (##exact-int.div x (##car sqs)))
                  (new-level (##fixnum.- level 1))
                  (new-sqs (##cdr sqs))
                  (q (##car qr))
                  (r (##cdr qr)))
             (uinteger->fixnums
              new-level
              new-sqs
              r
              (uinteger->fixnums new-level new-sqs q lst))))))

  (define (uinteger->string x rad len)
    (make-string-from-fixnums
     rad
     (let ((rad^size
            (##vector-ref rad^block-size rad)))
       (let loop ((level 0)
                  (sqs '())
                  (rad^size^2^level rad^size))
         (let ((new-level
                (##fixnum.+ level 1))
               (new-sqs
                (##cons rad^size^2^level sqs))
               (new-rad^size^2^level
                (##exact-int.square rad^size^2^level)))
           (if (##< x new-rad^size^2^level)
             (uinteger->fixnums new-level new-sqs x '())
             (loop new-level
                   new-sqs
                   new-rad^size^2^level)))))
     len
     0))

  (if (##fixnum? x)

    (cond ((##fixnum.negative? x)
           (let ((s (make-string-from-last-fixnum rad x 1 0)))
             (##string-set! s 0 #\-)
             s))
          ((##fixnum.zero? x)
           (if force-sign?
             (##string #\+ #\0)
             (##string #\0)))
          (else
           (if force-sign?
             (let ((s (make-string-from-last-fixnum rad (##fixnum.- x) 1 0)))
               (##string-set! s 0 #\+)
               s)
             (make-string-from-last-fixnum rad (##fixnum.- x) 0 0))))

    (cond ((##bignum.negative? x)
           (let ((s (uinteger->string (##negate x) rad 1)))
             (##string-set! s 0 #\-)
             s))
          (else
           (if force-sign?
             (let ((s (uinteger->string x rad 1)))
               (##string-set! s 0 #\+)
               s)
             (uinteger->string x rad 0))))))

(define ##digit-to-char-table "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ")

(define (##ratnum.number->string x rad force-sign?)
  (##string-append
   (##exact-int.number->string (macro-ratnum-numerator x) rad force-sign?)
   "/"
   (##exact-int.number->string (macro-ratnum-denominator x) rad #f)))

(##define-macro (macro-chez-fp-syntax) #f)

(##define-macro (macro-make-10^constants)
  (define n 326)
  (let ((v (make-vector n)))
    (let loop ((i 0) (x 1))
      (if (< i n)
        (begin
          (vector-set! v i x)
          (loop (+ i 1) (* x 10)))))
    `',v))

(define ##10^-constants (macro-make-10^constants))

(define (##flonum.printout v sign-prefix)

  ; This algorithm is derived from the paper "Printing Floating-Point
  ; Numbers Quickly and Accurately" by Robert G. Burger and R. Kent Dybvig,
  ; SIGPLAN'96 Conference on Programming Language Design an Implementation.

  ; v is a flonum
  ; f is an exact integer (fixnum or bignum)
  ; e is an exact integer (fixnum only)

  (define (10^ n) ; 0 <= n < 326
    (##vector-ref ##10^-constants n))

  (define (base-10-log x)
    (##define-macro (1/log10) `',(/ (log 10)))
    (##flonum.* (##flonum.log x) (1/log10)))

  (##define-macro (epsilon)
    1e-10)

  (define (scale r s m+ m- round? v)

    ; r is an exact integer (fixnum or bignum)
    ; s is an exact integer (fixnum or bignum)
    ; m+ is an exact integer (fixnum or bignum)
    ; m- is an exact integer (fixnum or bignum)
    ; round? is a boolean
    ; v is a flonum

    (let ((est
           (##flonum.->fixnum
            (##flonum.ceiling (##flonum.- (base-10-log v) (epsilon))))))
      (if (##fixnum.negative? est)
        (let ((factor (10^ (##fixnum.- est))))
          (fixup (##* r factor)
                 s
                 (##* m+ factor)
                 (##* m- factor)
                 est
                 round?))
        (let ((factor (10^ est)))
          (fixup r
                 (##* s factor)
                 m+
                 m-
                 est
                 round?)))))

  (define (fixup r s m+ m- k round?)
    (if (if round?
          (##not (##< (##+ r m+) s))
          (##< s (##+ r m+)))
      (##cons (##fixnum.+ k 1)
              (generate r
                        s
                        m+
                        m-
                        round?
                        0))
      (##cons k
              (generate (##* r 10)
                        s
                        (##* m+ 10)
                        (##* m- 10)
                        round?
                        0))))

  (define (generate r s m+ m- round? n)
    (let* ((dr (##exact-int.div r s))
           (d (##car dr))
           (r (##cdr dr))
           (tc (if round?
                 (##not (##< (##+ r m+) s))
                 (##< s (##+ r m+)))))
      (if (if round? (##not (##< m- r)) (##< r m-))
        (let* ((last-digit
                (if tc
                  (let ((r*2 (##exact-int.shift r 1)))
                    (if (or (and (##fixnum.even? d)
                                 (##= r*2 s)) ; tie, round d to even
                            (##< r*2 s))
                      d
                      (##fixnum.+ d 1)))
                  d))
               (str
                (##make-string (##fixnum.+ n 1))))
          (##string-set! str n
            (##string-ref ##digit-to-char-table last-digit))
          str)
        (if tc
          (let ((str
                 (##make-string (##fixnum.+ n 1))))
            (##string-set! str n
              (##string-ref ##digit-to-char-table (##fixnum.+ d 1)))
            str)
          (let ((str
                 (generate (##* r 10)
                           s
                           (##* m+ 10)
                           (##* m- 10)
                           round?
                           (##fixnum.+ n 1))))
            (##string-set! str n
              (##string-ref ##digit-to-char-table d))
            str)))))

  (define (flonum->exponent-and-digits v)
    (let* ((x (##flonum.->exact-exponential-format v))
           (f (##vector-ref x 0))
           (e (##vector-ref x 1))
           (round? (##not (##odd? f))))
      (if (##fixnum.negative? e)
        (if (and (##not (##fixnum.= e (macro-flonum-e-min)))
                 (##= f (macro-flonum-+m-min)))
          (scale (##exact-int.shift f 2)
                 (##exact-int.shift 1 (##fixnum.- 2 e))
                 2
                 1
                 round?
                 v)
          (scale (##exact-int.shift f 1)
                 (##exact-int.shift 1 (##fixnum.- 1 e))
                 1
                 1
                 round?
                 v))
        (let ((2^e (##exact-int.shift 1 e)))
          (if (##= f (macro-flonum-+m-min))
            (scale (##exact-int.shift f (##fixnum.+ e 2))
                   4
                   (##exact-int.shift 1 (##fixnum.+ e 1))
                   2^e
                   round?
                   v)
            (scale (##exact-int.shift f (##fixnum.+ e 1))
                   2
                   2^e
                   2^e
                   round?
                   v))))))

  (let* ((x (flonum->exponent-and-digits v))
         (e (##car x))
         (d (##cdr x))            ; d = digits
         (n (##string-length d))) ; n = number of digits

    (cond ((and (##not (##fixnum.< e 0)) ; 0<=e<=10
                (##not (##fixnum.< 10 e)))

           (cond ((##fixnum.= e 0) ; e=0

                  ; Format 1: .DDD    (0.DDD in chez-fp-syntax)

                  (##string-append sign-prefix
                                   (if (macro-chez-fp-syntax) "0." ".")
                                   d))

                 ((##fixnum.< e n) ; e<n

                  ; Format 2: D.DDD up to DDD.D

                  (##string-append sign-prefix
                                   (##substring d 0 e)
                                   "."
                                   (##substring d e n)))

                 ((##fixnum.= e n) ; e=n

                  ; Format 3: DDD.    (DDD.0 in chez-fp-syntax)

                  (##string-append sign-prefix
                                   d
                                   (if (macro-chez-fp-syntax) ".0" ".")))

                 (else ; e>n

                  ; Format 4: DDD000000.    (DDD000000.0 in chez-fp-syntax)

                  (##string-append sign-prefix
                                   d
                                   (##make-string (##fixnum.- e n) #\0)
                                   (if (macro-chez-fp-syntax) ".0" ".")))))

          ((and (##not (##fixnum.< e -2)) ; -2<=e<=-1
                (##not (##fixnum.< -1 e)))

           ; Format 5: .0DDD or .00DDD    (0.0DDD or 0.00DDD in chez-fp-syntax)

           (##string-append sign-prefix
                            (if (macro-chez-fp-syntax) "0." ".")
                            (##make-string (##fixnum.- e) #\0)
                            d))

          (else

           ; Format 6: D.DDDeEEE
           ;
           ; This is the most general format.  We insert a period after
           ; the first digit (unless there is only one digit) and add
           ; an exponent.

           (##string-append sign-prefix
                            (##substring d 0 1)
                            (if (##fixnum.= n 1) "" ".")
                            (##substring d 1 n)
                            "e"
                            (##number->string (##fixnum.- e 1) 10))))))

(define (##flonum.number->string x rad force-sign?)

  (define (non-neg-num->str x rad sign-prefix)
    (if (##flonum.zero? x)
      (##string-append sign-prefix (if (macro-chez-fp-syntax) "0.0" "0."))
      (##flonum.printout x sign-prefix)))

  (cond ((##flonum.nan? x)
         (##string-copy (if (macro-chez-fp-syntax) "+nan.0" "+nan.")))
        ((##flonum.negative? (##flonum.copysign (macro-inexact-+1) x))
         (let ((abs-x (##flonum.copysign x (macro-inexact-+1))))
           (cond ((##flonum.= abs-x (macro-inexact-+inf))
                  (##string-copy (if (macro-chez-fp-syntax) "-inf.0" "-inf.")))
                 (else
                  (non-neg-num->str abs-x rad "-")))))
        (else
         (cond ((##flonum.= x (macro-inexact-+inf))
                (##string-copy (if (macro-chez-fp-syntax) "+inf.0" "+inf.")))
               (force-sign?
                (non-neg-num->str x rad "+"))
               (else
                (non-neg-num->str x rad ""))))))

(define (##cpxnum.number->string x rad force-sign?)
  (let* ((real
          (macro-cpxnum-real x))
         (real-str
          (if (##eq? real 0) "" (##number->string real rad force-sign?))))
    (let ((imag (macro-cpxnum-imag x)))
      (cond ((##eq? imag 1)
             (##string-append real-str "+i"))
            ((##eq? imag -1)
             (##string-append real-str "-i"))
            (else
             (##string-append real-str
                              (##number->string imag rad #t)
                              "i"))))))

(define (##number->string x #!optional (rad 10) (force-sign? #f))
  (macro-number-dispatch x '()
    (##exact-int.number->string x rad force-sign?)
    (##exact-int.number->string x rad force-sign?)
    (##ratnum.number->string x rad force-sign?)
    (##flonum.number->string x rad force-sign?)
    (##cpxnum.number->string x rad force-sign?)))

(define (number->string n #!optional (r (macro-absent-obj)))
  (macro-force-vars (n r)
    (let ((rad (if (##eq? r (macro-absent-obj)) 10 r)))

      (define (range-error)
        (##trap-check-range 'number->string n r))

      (define (type-error1)
        (##trap-check-number 'number->string n r))

      (define (type-error2)
        (##trap-check-exact-int 'number->string n r))

      (if (macro-exact-int? rad)
        (if (or (##eq? rad 2)
                (##eq? rad 8)
                (##eq? rad 10)
                (##eq? rad 16))
          (let ((result (##number->string n rad #f)))
            (if (##null? result)
              (type-error1)
              result))
          (range-error))
        (type-error2)))))

(##define-macro (macro-make-char-to-digit-table)
  (let ((t (make-vector 128 99)))
    (vector-set! t (char->integer #\#) 0) ; #\# counts as 0
    (let loop1 ((i 9))
      (if (not (< i 0))
        (begin
          (vector-set! t (+ (char->integer #\0) i) i)
          (loop1 (- i 1)))))
    (let loop2 ((i 25))
      (if (not (< i 0))
        (begin
          (vector-set! t (+ (char->integer #\A) i) (+ i 10))
          (vector-set! t (+ (char->integer #\a) i) (+ i 10))
          (loop2 (- i 1)))))
    `',(list->u8vector (vector->list t))))

(define ##char-to-digit-table (macro-make-char-to-digit-table))

(define (##string->number str #!optional (rad 10) (check-only? #f))

  ; The number grammar parsed by this procedure is:
  ;
  ; <num R E> : <prefix R E> <complex R E>
  ; 
  ; <complex R E> : <real R E>
  ;               | <real R E> @ <real R E>
  ;               | <real R E> <sign> <ureal R> i
  ;               | <real R E> <sign-inf-nan R E> i
  ;               | <real R E> <sign> i
  ;               | <sign> <ureal R> i
  ;               | <sign-inf-nan R E> i
  ;               | <sign> i
  ; 
  ; <real R E> : <ureal R>
  ;            | <sign> <ureal R>
  ;            | <sign-inf-nan R E>
  ; 
  ; <sign-inf-nan R i> : +inf.
  ;                    | -inf.
  ;                    | +nan.
  ; <sign-inf-nan R empty> : <sign-inf-nan R i>
  ; 
  ; <ureal R> : <uinteger R>
  ;           | <uinteger R> / <uinteger R>
  ;           | <decimal R>
  ; 
  ; <decimal 10> : <uinteger 10> <suffix>
  ;              | . <digit 10>+ #* <suffix>
  ;              | <digit 10>+ . <digit 10>* #* <suffix>
  ;              | <digit 10>+ #+ . #* <suffix>
  ; 
  ; <uinteger R> : <digit R>+ #*
  ; 
  ; <prefix R E> : <radix R E> <exactness E>
  ;              | <exactness E> <radix R E>
  ; 
  ; <suffix> : <empty> 
  ;          | <exponent marker> <digit 10>+
  ;          | <exponent marker> <sign> <digit 10>+
  ; 
  ; <exponent marker> : e | s | f | d | l
  ; <sign> : + | -
  ; <exactness empty> : <empty>
  ; <exactness i> : #i
  ; <exactness e> : #e
  ; <radix 2> : #b
  ; <radix 8> : #o
  ; <radix 10> : <empty> | #d
  ; <radix 16> : #x
  ; <digit 2> : 0 | 1
  ; <digit 8> : 0 | 1 | 2 | 3 | 4 | 5 | 6 | 7
  ; <digit 10> : 0 | 1 | 2 | 3 | 4 | 5 | 6 | 7 | 8 | 9
  ; <digit 16> : <digit 10> | a | b | c | d | e | f

  (define max-fixnum32/10 53687091) ; OK to be conservative

  (##define-macro (macro-make-exact-10^n-table)

    (define max-exact-power-of-10 22) ; (floor (inexact->exact (/ (log (expt 2 (macro-flonum-m-bits-plus-1))) (log 5))))

    (let ((t (make-vector (+ max-exact-power-of-10 1))))

      (let loop ((i max-exact-power-of-10))
        (if (not (< i 0))
          (begin
            (vector-set! t i (exact->inexact (expt 10 i)))
            (loop (- i 1)))))

      `',(list->f64vector (vector->list t))))

  (define exact-10^n-table (macro-make-exact-10^n-table))

  (##define-macro (macro-make-block-size)
    (let* ((max-rad 16)
           (t (make-vector (+ max-rad 1) 0)))

      (define max-fixnum 536870911) ; OK to be conservative

      (define (block-size-for rad)
        (let loop ((i 0) (rad^i 1))
          (let ((new-rad^i (* rad^i rad)))
            (if (<= new-rad^i max-fixnum)
              (loop (+ i 1) new-rad^i)
              i))))

      (let loop ((i max-rad))
        (if (< 1 i)
          (begin
            (vector-set! t i (block-size-for i))
            (loop (- i 1)))))

      `',t))

  (define block-size (macro-make-block-size))

  (##define-macro (macro-make-rad^block-size)
    (let* ((max-rad 16)
           (t (make-vector (+ max-rad 1) 0)))

      (define max-fixnum 536870911) ; OK to be conservative

      (define (rad^block-size-for rad)
        (let loop ((i 0) (rad^i 1))
          (let ((new-rad^i (* rad^i rad)))
            (if (<= new-rad^i max-fixnum)
              (loop (+ i 1) new-rad^i)
              rad^i))))

      (let loop ((i max-rad))
        (if (< 1 i)
          (begin
            (vector-set! t i (rad^block-size-for i))
            (loop (- i 1)))))

      `',t))

  (define rad^block-size (macro-make-rad^block-size))

  (define (substring->uinteger-fixnum str rad i j)

    ; Simple case: result is known to fit in a fixnum.

    (let loop ((i i) (n 0))
      (if (##fixnum.< i j)
        (let ((c (##string-ref str i)))
          (if (##char<? c 128)
            (loop (##fixnum.+ i 1)
                  (##fixnum.+ (##fixnum.* n rad)
                              (##u8vector-ref ##char-to-digit-table c)))
            (loop (##fixnum.+ i 1)
                  (##fixnum.* n rad))))
        n)))

  (define (substring->uinteger-aux sqs width str rad i j)

    ; Divide-and-conquer algorithm (fast for large bignums if bignum
    ; multiplication is fast).

    (if (##null? sqs)
      (substring->uinteger-fixnum str rad i j)
      (let* ((new-sqs (##cdr sqs))
             (new-width (##fixnum.quotient width 2))
             (mid (##fixnum.- j new-width)))
        (if (##fixnum.< i mid)
          (let* ((a (substring->uinteger-aux new-sqs new-width str rad i mid))
                 (b (substring->uinteger-aux new-sqs new-width str rad mid j)))
            (##+ (##* a (##car sqs)) b))
          (substring->uinteger-aux new-sqs new-width str rad i j)))))

  (define (squares rad n)
    (let loop ((rad rad) (n n) (lst '()))
      (if (##fixnum.= n 1)
        (##cons rad lst)
        (loop (##exact-int.square rad)
              (##fixnum.- n 1)
              (##cons rad lst)))))

  (define (substring->uinteger str rad i j)

    ; Converts a substring into an unsigned integer.  Selects a fast
    ; conversion algorithm when result fits in a fixnum.

    (let ((len (##fixnum.- j i))
          (size (##vector-ref block-size rad)))
      (if (##fixnum.< size len)
        (let ((levels
               (##exact-int.width
                (##fixnum.quotient (##fixnum.- len 1) size))))
          (substring->uinteger-aux
           (squares (##vector-ref rad^block-size rad) levels)
           (##fixnum.shl size levels)
           str
           rad
           i
           j))
        (substring->uinteger-fixnum str rad i j))))

  (define (float-substring->uinteger str i j)

    ; Converts a substring containing the decimals of a floating-point
    ; number into an unsigned integer (any period is simply skipped).

    (let loop1 ((i i) (n 0))
      (if (##not (##fixnum.< i j))
        n
        (let ((c (##string-ref str i)))
          (if (##char=? c #\.)
            (loop1 (##fixnum.+ i 1) n)
            (let ((new-n
                   (##fixnum.+ (##fixnum.* n 10)
                               (if (##char<? c 128)
                                 (##u8vector-ref ##char-to-digit-table c)
                                 0))))
              (if (##fixnum.< new-n max-fixnum32/10)
                (loop1 (##fixnum.+ i 1) new-n)
                (let loop2 ((i i) (n n))
                  (if (##not (##fixnum.< i j))
                    n
                    (let ((c (##string-ref str i)))
                      (if (##char=? c #\.)
                        (loop2 (##fixnum.+ i 1) n)
                        (let ((new-n
                               (##+
                                (##* n 10)
                                (if (##char<? c 128)
                                  (##u8vector-ref ##char-to-digit-table c)
                                  0))))
                          (loop2 (##fixnum.+ i 1) new-n)))))))))))))

  (define (uinteger str rad i)
    (and (##fixnum.< i (##string-length str))
         (let ((c (##string-ref str i)))
           (and (##char<? c 128)
                (##not (##char=? c #\#))
                (##fixnum.< (##u8vector-ref ##char-to-digit-table c) rad)
                (digits-and-sharps str rad (##fixnum.+ i 1))))))

  (define (digits-and-sharps str rad i)
    (let loop ((i i))
      (if (##fixnum.< i (##string-length str))
        (let ((c (##string-ref str i)))
          (if (##char<? c 128)
            (if (##char=? c #\#)
              (sharps str (##fixnum.+ i 1))
              (if (##fixnum.< (##u8vector-ref ##char-to-digit-table c) rad)
                (loop (##fixnum.+ i 1))
                i))
            i))
        i)))

  (define (sharps str i)
    (let loop ((i i))
      (if (##fixnum.< i (##string-length str))
        (if (##char=? (##string-ref str i) #\#)
          (loop (##fixnum.+ i 1))
          i)
        i)))

  (define (suffix str i1)
    (if (##fixnum.< (##fixnum.+ i1 1) (##string-length str))
      (let ((c1 (##string-ref str i1)))
        (if (or (##char=? c1 #\e) (##char=? c1 #\E)
                (##char=? c1 #\s) (##char=? c1 #\S)
                (##char=? c1 #\f) (##char=? c1 #\F)
                (##char=? c1 #\d) (##char=? c1 #\D)
                (##char=? c1 #\l) (##char=? c1 #\L))
          (let ((c2 (##string-ref str (##fixnum.+ i1 1))))
            (let ((i2
                   (if (or (##char=? c2 #\+) (##char=? c2 #\-))
                     (uinteger str 10 (##fixnum.+ i1 2))
                     (uinteger str 10 (##fixnum.+ i1 1)))))
              (if (and i2
                       (##not (##char=? (##string-ref str (##fixnum.- i2 1))
                                        #\#)))
                i2
                i1)))
          i1))
      i1))

  (define (ureal str rad i1)
    (let ((i2 (uinteger str rad i1)))
      (if i2
        (if (##fixnum.< i2 (##string-length str))
          (let ((c (##string-ref str i2)))
            (cond ((##char=? c #\/)
                   (let ((i3 (uinteger str rad (##fixnum.+ i2 1))))
                     (and i3
                          (##vector i3 i2))))
                  ((##fixnum.= rad 10)
                   (if (##char=? c #\.)
                     (let ((i3
                            (if (##char=? (##string-ref str (##fixnum.- i2 1))
                                          #\#)
                              (sharps str (##fixnum.+ i2 1))
                              (digits-and-sharps str 10 (##fixnum.+ i2 1)))))
                       (and i3
                            (let ((i4 (suffix str i3)))
                              (##vector i4 i3 i2))))
                     (let ((i3 (suffix str i2)))
                       (if (##fixnum.= i2 i3)
                         i2
                         (##vector i3 i2 i2)))))
                  (else
                   i2)))
          i2)
        (and (##fixnum.= rad 10)
             (##fixnum.< i1 (##string-length str))
             (##char=? (##string-ref str i1) #\.)
             (let ((i3 (uinteger str rad (##fixnum.+ i1 1))))
               (and i3
                    (let ((i4 (suffix str i3)))
                      (##vector i4 i3 i1))))))))

  (define (inf-nan str sign i e)
    (and (##not (##eq? e 'e))
         (if (##fixnum.< (##fixnum.+ i (if (macro-chez-fp-syntax) 4 3))
                         (##string-length str))
           (and (##char=? (##string-ref str (##fixnum.+ i 3)) #\.)
                (if (macro-chez-fp-syntax)
                  (##char=? (##string-ref str (##fixnum.+ i 4)) #\0)
                  #t)
                (or (and (let ((c (##string-ref str i)))
                           (or (##char=? c #\i) (##char=? c #\I)))
                         (let ((c (##string-ref str (##fixnum.+ i 1))))
                           (or (##char=? c #\n) (##char=? c #\N)))
                         (let ((c (##string-ref str (##fixnum.+ i 2))))
                           (or (##char=? c #\f) (##char=? c #\F))))
                    (and (##not (##char=? sign #\-))
                         (let ((c (##string-ref str i)))
                           (or (##char=? c #\n) (##char=? c #\N)))
                         (let ((c (##string-ref str (##fixnum.+ i 1))))
                           (or (##char=? c #\a) (##char=? c #\A)))
                         (let ((c (##string-ref str (##fixnum.+ i 2))))
                           (or (##char=? c #\n) (##char=? c #\N)))))
                (##vector (##fixnum.+ i (if (macro-chez-fp-syntax) 5 4))))
           #f)))

  (define (make-rec x y)
    (##make-rectangular x y))

  (define (make-pol x y e)
    (let ((n (##make-polar x y)))
      (if (##eq? e 'e)
        (##inexact->exact n)
        n)))

  (define (make-inexact-real sign uinteger exponent)
    (let ((n
           (if (and (##fixnum? uinteger)
                    (##flonum.<-fixnum-exact? uinteger)
                    (##fixnum? exponent)
                    (##fixnum.< (##fixnum.- exponent)
                                (##f64vector-length exact-10^n-table))
                    (##fixnum.< exponent
                                (##f64vector-length exact-10^n-table)))
             (if (##fixnum.< exponent 0)
               (##flonum./ (##flonum.<-fixnum uinteger)
                           (##f64vector-ref exact-10^n-table
                                            (##fixnum.- exponent)))
               (##flonum.* (##flonum.<-fixnum uinteger)
                           (##f64vector-ref exact-10^n-table
                                            exponent)))
             (##exact->inexact
              (##* uinteger (##expt 10 exponent))))))
      (if (##char=? sign #\-)
        (##flonum.copysign n (macro-inexact--1))
        n)))

  (define (get-zero e)
    (if (##eq? e 'i)
      (macro-inexact-+0)
      0))

  (define (get-one sign e)
    (if (##eq? e 'i)
      (if (##char=? sign #\-) (macro-inexact--1) (macro-inexact-+1))
      (if (##char=? sign #\-) -1 1)))

  (define (get-real start sign str rad e i)
    (if (##fixnum? i)
      (let* ((abs-n
              (substring->uinteger str rad start i))
             (n
              (if (##char=? sign #\-)
                (##negate abs-n)
                abs-n)))
        (if (or (##eq? e 'i)
                (and (##not e)
                     (##char=? (##string-ref str (##fixnum.- i 1)) #\#)))
          (##exact->inexact n)
          n))
      (let ((j (##vector-ref i 0))
            (len (##vector-length i)))
        (cond ((##fixnum.= len 3) ; xxx.yyyEzzz
               (let* ((after-frac-part
                       (##vector-ref i 1))
                      (unadjusted-exponent
                       (if (##fixnum.= after-frac-part j) ; no exponent part?
                         0
                         (let* ((c
                                 (##string-ref
                                  str
                                  (##fixnum.+ after-frac-part 1)))
                                (n
                                 (substring->uinteger
                                  str
                                  10
                                  (if (or (##char=? c #\+) (##char=? c #\-))
                                    (##fixnum.+ after-frac-part 2)
                                    (##fixnum.+ after-frac-part 1))
                                  j)))
                           (if (##char=? c #\-)
                             (##negate n)
                             n))))
                      (c
                       (##string-ref str start))
                      (uinteger
                       (float-substring->uinteger str start after-frac-part))
                      (decimals-after-point
                       (##fixnum.-
                        (##fixnum.- after-frac-part (##vector-ref i 2))
                        1))
                      (exponent
                       (if (##fixnum.< 0 decimals-after-point)
                         (if (and (##fixnum? unadjusted-exponent)
                                  (##fixnum.< (##fixnum.- unadjusted-exponent
                                                          decimals-after-point)
                                              unadjusted-exponent))
                           (##fixnum.- unadjusted-exponent
                                       decimals-after-point)
                           (##- unadjusted-exponent
                                decimals-after-point))
                         unadjusted-exponent)))
                 (if (##eq? e 'e)
                   (##*
                    (if (##char=? sign #\-)
                      (##negate uinteger)
                      uinteger)
                    (##expt 10 exponent))
                   (make-inexact-real sign uinteger exponent))))
              ((##fixnum.= len 2) ; xxx/yyy
               (let* ((after-num
                       (##vector-ref i 1))
                      (abs-num
                       (substring->uinteger str rad start after-num))
                      (num
                       (if (##char=? sign #\-)
                         (##negate abs-num)
                         abs-num))
                      (den
                       (substring->uinteger str
                                            rad
                                            (##fixnum.+ after-num 1)
                                            j))
                      (num/den
                       (##/ num den)))
                 (if (or (##eq? e 'i)
                         (and (##not e)
                              (or (##char=? (##string-ref
                                             str
                                             (##fixnum.- after-num 1))
                                            #\#)
                                  (##char=? (##string-ref
                                             str
                                             (##fixnum.- j 1))
                                            #\#))))
                   (##exact->inexact num/den)
                   num/den)))
              (else ; (##fixnum.= len 1) ; inf or nan
               (let* ((c
                       (##string-ref str start))
                      (n
                       (if (or (##char=? c #\i) (##char=? c #\I))
                         (macro-inexact-+inf)
                         (macro-inexact-+nan))))
                 (if (##char=? sign #\-)
                   (##flonum.copysign n (macro-inexact--1))
                   n)))))))

  (define (i-end str i)
    (and (##fixnum.= (##fixnum.+ i 1) (##string-length str))
         (let ((c (##string-ref str i)))
           (or (##char=? c #\i) (##char=? c #\I)))))

  (define (complex start sign str rad e i)
    (let ((j (if (##fixnum? i) i (##vector-ref i 0))))
      (let ((c (##string-ref str j)))
        (cond ((##char=? c #\@)
               (let ((j+1 (##fixnum.+ j 1)))
                 (if (##fixnum.< j+1 (##string-length str))
                   (let* ((sign2
                           (##string-ref str j+1))
                          (start2
                           (if (or (##char=? sign2 #\+) (##char=? sign2 #\-))
                             (##fixnum.+ j+1 1)
                             j+1))
                          (k
                           (or (ureal str rad start2)
                               (and (##fixnum.< j+1 start2)
                                    (inf-nan str sign2 start2 e)))))
                     (and k
                          (let ((l (if (##fixnum? k) k (##vector-ref k 0))))
                            (and (##fixnum.= l (##string-length str))
                                 (or check-only?
                                     (make-pol
                                      (get-real start sign str rad e i)
                                      (get-real start2 sign2 str rad e k)
                                      e))))))
                   #f)))
              ((or (##char=? c #\+) (##char=? c #\-))
               (let* ((start2
                       (##fixnum.+ j 1))
                      (k
                       (or (ureal str rad start2)
                           (inf-nan str c start2 e))))
                 (if (##not k)
                   (if (i-end str start2)
                     (or check-only?
                         (make-rec
                          (get-real start sign str rad e i)
                          (get-one c e)))
                     #f)
                   (let ((l (if (##fixnum? k) k (##vector-ref k 0))))
                     (and (i-end str l)
                          (or check-only?
                              (make-rec
                               (get-real start sign str rad e i)
                               (get-real start2 c str rad e k))))))))
              (else
               #f)))))

  (define (after-prefix start str rad e)

    ; invariant: start = 0, 2 or 4, (string-length str) > start

    (let ((c (##string-ref str start)))
      (if (or (##char=? c #\+) (##char=? c #\-))
        (let ((i (or (ureal str rad (##fixnum.+ start 1))
                     (inf-nan str c (##fixnum.+ start 1) e))))
          (if (##not i)
            (if (i-end str (##fixnum.+ start 1))
              (or check-only?
                  (make-rec
                   (get-zero e)
                   (get-one c e)))
              #f)
            (let ((j (if (##fixnum? i) i (##vector-ref i 0))))
              (cond ((##fixnum.= j (##string-length str))
                     (or check-only?
                         (get-real (##fixnum.+ start 1) c str rad e i)))
                    ((i-end str j)
                     (or check-only?
                         (make-rec
                          (get-zero e)
                          (get-real (##fixnum.+ start 1) c str rad e i))))
                    (else
                     (complex (##fixnum.+ start 1) c str rad e i))))))
        (let ((i (ureal str rad start)))
          (if (##not i)
            #f
            (let ((j (if (##fixnum? i) i (##vector-ref i 0))))
              (cond ((##fixnum.= j (##string-length str))
                     (or check-only?
                         (get-real start #\+ str rad e i)))
                    (else
                     (complex start #\+ str rad e i)))))))))

  (define (radix-prefix c)
    (cond ((or (##char=? c #\b) (##char=? c #\B))  2)
          ((or (##char=? c #\o) (##char=? c #\O))  8)
          ((or (##char=? c #\d) (##char=? c #\D)) 10)
          ((or (##char=? c #\x) (##char=? c #\X)) 16)
          (else                                   #f)))

  (define (exactness-prefix c)
    (cond ((or (##char=? c #\i) (##char=? c #\I)) 'i)
          ((or (##char=? c #\e) (##char=? c #\E)) 'e)
          (else                                   #f)))

  (cond ((##fixnum.< 2 (##string-length str)) ; >= 3 chars
         (if (##char=? (##string-ref str 0) #\#)
           (let ((rad1 (radix-prefix (##string-ref str 1))))
             (if rad1
               (if (and (##fixnum.< 4 (##string-length str)) ; >= 5 chars
                        (##char=? (##string-ref str 2) #\#))
                 (let ((e1 (exactness-prefix (##string-ref str 3))))
                   (if e1
                     (after-prefix 4 str rad1 e1)
                     #f))
                 (after-prefix 2 str rad1 #f))
               (let ((e2 (exactness-prefix (##string-ref str 1))))
                 (if e2
                   (if (and (##fixnum.< 4 (##string-length str)) ; >= 5 chars
                            (##char=? (##string-ref str 2) #\#))
                     (let ((rad2 (radix-prefix (##string-ref str 3))))
                       (if rad2
                         (after-prefix 4 str rad2 e2)
                         #f))
                     (after-prefix 2 str rad e2))
                   #f))))
           (after-prefix 0 str rad #f)))
        ((##fixnum.< 0 (##string-length str)) ; >= 1 char
         (after-prefix 0 str rad #f))
        (else
         #f)))

(define (string->number str #!optional (r (macro-absent-obj)))
  (macro-force-vars (str r)
    (macro-check-string str (string->number str r)
      (let ((rad (if (##eq? r (macro-absent-obj)) 10 r)))

        (define (range-error)
          (##trap-check-range 'string->number str r))

        (define (type-error)
          (##trap-check-exact-int 'string->number str r))

        (if (macro-exact-int? rad)
          (if (or (##eq? rad 2)
                  (##eq? rad 8)
                  (##eq? rad 10)
                  (##eq? rad 16))
            (##string->number str rad #f)
            (range-error))
          (type-error))))))

;------------------------------------------------------------------------------


;==============================================================================

; file: "_num2.scm"

; Copyright (C) 1994-2000 by Marc Feeley, All Rights Reserved.

(##include "header.scm")

;------------------------------------------------------------------------------

; Fixnum operations
; -----------------

(define-nary (##fixnum.max x y)
  ()
  x
  (##fixnum.max x y)
  macro-no-force
  macro-no-check)

(define-nary (##fixnum.min x y)
  ()
  x
  (##fixnum.min x y)
  macro-no-force
  macro-no-check)

(define-nary (##fixnum.+ x y)
  0
  x
  (##fixnum.+ x y)
  macro-no-force
  macro-no-check)

(define-nary (##fixnum.* x y)
  1
  x
  (##fixnum.* x y)
  macro-no-force
  macro-no-check)

(define-nary (##fixnum.- x y)
  ()
  (##fixnum.- x)
  (##fixnum.- x y)
  macro-no-force
  macro-no-check)

(define-system (##fixnum.quotient x y))

(define-system (##fixnum.remainder x y))

(define-system (##fixnum.modulo x y)
  (let ((r (##fixnum.remainder x y)))
    (if (##fixnum.zero? r)
      0
      (if (##eq? (##fixnum.negative? x) (##fixnum.negative? y))
        r
        (##fixnum.+ r y)))))

(define-nary (##fixnum.logior x y)
  0
  x
  (##fixnum.logior x y)
  macro-no-force
  macro-no-check)

(define-nary (##fixnum.logxor x y)
  0
  x
  (##fixnum.logxor x y)
  macro-no-force
  macro-no-check)

(define-nary (##fixnum.logand x y)
  -1
  x
  (##fixnum.logand x y)
  macro-no-force
  macro-no-check)

(define-system (##fixnum.lognot x)
  (##fixnum.- -1 x))

(define-system (##fixnum.ashr x y))
(define-system (##fixnum.lshr x y))
(define-system (##fixnum.shl x y))

(define-system (##fixnum.zero? x)
  (##eq? x 0))

(define-system (##fixnum.positive? x)
  (##fixnum.< 0 x))

(define-system (##fixnum.negative? x)
  (##fixnum.< x 0))

(define-system (##fixnum.odd? x)
  (##fixnum.= (##fixnum.logand x 1) 1))

(define-system (##fixnum.even? x)
  (##fixnum.= (##fixnum.logand x 1) 0))

(define-nary-bool (##fixnum.= x y)
  #t
  #t
  (##fixnum.= x y)
  macro-no-force
  macro-no-check)

(define-nary-bool (##fixnum.< x y)
  #t
  #t
  (##fixnum.< x y)
  macro-no-force
  macro-no-check)

(define-nary-bool (##fixnum.> x y)
  #t
  #t
  (##fixnum.< y x)
  macro-no-force
  macro-no-check)

(define-nary-bool (##fixnum.<= x y)
  #t
  #t
  (##not (##fixnum.< y x))
  macro-no-force
  macro-no-check)

(define-nary-bool (##fixnum.>= x y)
  #t
  #t
  (##not (##fixnum.< x y))
  macro-no-force
  macro-no-check)

(define-system (##fixnum.->char x))
(define-system (##fixnum.<-char x))

;------------------------------------------------------------------------------

; Bignum operations
; -----------------

; The bignum operations were mostly implemented by the "ber numerical
; analyst" Brad Lucier (http://www.math.purdue.edu/~lucier) with some
; guidance from Marc Feeley.

; Bignums are represented with 'adigit' vectors.  Each element is an
; integer containing ##bignum.adigit-width bits (typically 64 bits).
; These bits encode an integer in two's complement representation.
; The first element contains the least significant bits and the most
; significant bit of the last element is the sign (0=positive,
; 1=negative).

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define-system (##bignum.negative? x))
(define-system (##bignum.adigit-length x))
(define-system (##bignum.adigit-set-ones! x i))
(define-system (##bignum.adigit-inc! x i))
(define-system (##bignum.adigit-dec! x i))
(define-system (##bignum.adigit-add! x i y j carry))
(define-system (##bignum.adigit-sub! x i y j borrow))
(define-system (##bignum.mdigit-length x))
(define-system (##bignum.mdigit-ref x i))
(define-system (##bignum.mdigit-set! x i mdigit))
(define-system (##bignum.mdigit-mul! x i y j multiplier carry))
(define-system (##bignum.mdigit-div! x i y j quotient borrow))
(define-system (##bignum.mdigit-quotient u j v_n-1))
(define-system (##bignum.mdigit-remainder u j v_n-1 q-hat))
(define-system (##bignum.mdigit-test? q-hat v_n-2 r-hat u_j-2))

(define-system (##bignum.adigit-ones? x i))
(define-system (##bignum.adigit-zero? x i))
(define-system (##bignum.adigit-negative? x i))
(define-system (##bignum.adigit-= x y i))
(define-system (##bignum.adigit-< x y i))
(define-system (##bignum.->fixnum x))
(define-system (##bignum.<-fixnum x))
(define-system (##bignum.adigit-shrink! x n))
(define-system (##bignum.adigit-set! x i y j))
(define-system (##bignum.adigit-cat! x i hi j lo k divider))
(define-system (##bignum.adigit-logand! x i y j))
(define-system (##bignum.adigit-logior! x i y j))
(define-system (##bignum.adigit-logxor! x i y j))
(define-system (##bignum.adigit-lognot! x i))

(define-system (##bignum.fdigit-length x))
(define-system (##bignum.fdigit-ref x i))
(define-system (##bignum.fdigit-set! x i fdigit))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

; Bignum related constants.

(define ##bignum.adigit-ones #xffffffffffffffff)
(define ##bignum.adigit-zeros #x10000000000000000)

(define ##bignum.fdigit-base
  (##fixnum.shl 1 ##bignum.fdigit-width))

(define ##bignum.mdigit-base
  (##fixnum.shl 1 ##bignum.mdigit-width))

(define ##bignum.inexact-mdigit-base
  (##flonum.<-fixnum ##bignum.mdigit-base))

(define ##bignum.mdigit-base-minus-1
  (##fixnum.- ##bignum.mdigit-base 1))

(define ##bignum.minus-mdigit-base
  (##fixnum.- ##bignum.mdigit-base))

(define ##bignum.max-fixnum-div-mdigit-base
  (##fixnum.quotient ##max-fixnum ##bignum.mdigit-base))

(define ##bignum.min-fixnum-div-mdigit-base
  (##fixnum.quotient ##min-fixnum ##bignum.mdigit-base))

(define ##bignum.2*min-fixnum
  (if (##fixnum? -1073741824)
    -4611686018427387904 ; (- (expt 2 62))
    -1073741824))        ; (- (expt 2 30))

; The following global variables control when each of the three
; multiplication algorithms are used.

(define ##bignum.naive-mul-max-width 1400)
(set! ##bignum.naive-mul-max-width ##bignum.naive-mul-max-width)

(define ##bignum.fft-mul-min-width 6800)
(set! ##bignum.fft-mul-min-width ##bignum.fft-mul-min-width)

(define ##bignum.fft-mul-max-width 8388608) ; (* ##bignum.fdigit-width (expt 2 20))
(set! ##bignum.fft-mul-max-width ##bignum.fft-mul-max-width)

; For FFT the analysis is as follows: the algorithm must compute sums
; of up to 2^21 products of 8 bit numbers; this requires 21+16=37 bits
; accuracy.  Each FFT loses about sqrt(21) ~= 5 bits, or 10 bits total
; for the direct and inverse transforms.  That's 47 bits, and one more
; bit is lost for the multiplication of the FFT's.  So 48 bits are
; needed, which leaves 5 bits extra accuracy when using 64 bit IEEE
; floating point numbers.  That's more than necessary.

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

; Operations where arguments are in bignum format

; Bignum comparison.

(define (##bignum.< x y)

  ; x is a normalized bignum, y is a normalized bignum

  (##declare (not interrupts-enabled))

  (define (loop i)
    (and (##not (##fixnum.< i 0))
         (or (##bignum.adigit-< x y i)
             (and (##bignum.adigit-= x y i)
                  (loop (##fixnum.- i 1))))))

  (if (##bignum.negative? x)
    (if (##bignum.negative? y)
      (let ((x-length (##bignum.adigit-length x))
            (y-length (##bignum.adigit-length y)))
        (or (##fixnum.< y-length x-length)
            (and (##fixnum.= x-length y-length)
                 (loop (##fixnum.- x-length 1)))))
      #t)
    (if (##bignum.negative? y)
      #f
      (let ((x-length (##bignum.adigit-length x))
            (y-length (##bignum.adigit-length y)))
        (or (##fixnum.< x-length y-length)
            (and (##fixnum.= x-length y-length)
                 (loop (##fixnum.- x-length 1))))))))

(define (##bignum.= x y)

  ; x is a normalized bignum, y is a normalized bignum

  (##declare (not interrupts-enabled))

  (let ((x-length (##bignum.adigit-length x)))
    (and (##fixnum.= x-length (##bignum.adigit-length y))
         (let loop ((i (##fixnum.- x-length 1)))
           (or (##fixnum.< i 0)
               (and (##bignum.adigit-= x y i)
                    (loop (##fixnum.- i 1))))))))

; Bignum addition and subtraction.

(define (##bignum.+ x y)

  ; x is an unnormalized bignum, y is an unnormalized bignum

  (##declare (not interrupts-enabled))

  (define (add x x-length y y-length)
    (let* ((result-length
            (##fixnum.+ y-length
                        (if (##eq? (##bignum.negative? x)
                                   (##bignum.negative? y)) 
                          1 
                          0)))
           (result
            (##bignum.make result-length y #f)))
      (let loop ((i 0)
                 (carry 0))
        (if (##fixnum.< i x-length)
          (loop (##fixnum.+ i 1)
                (##bignum.adigit-add! result i x i carry))
          (##bignum.propagate-carry-and-normalize!
           result
           result-length
           x-length
           (##bignum.negative? x)
           (##fixnum.zero? carry))))))

  (let ((x-length (##bignum.adigit-length x))
        (y-length (##bignum.adigit-length y)))
    (if (##fixnum.< x-length y-length)
      (add x x-length y y-length)
      (add y y-length x x-length))))

(define (##bignum.- x y)

  ; x is an unnormalized bignum, y is an unnormalized bignum

  (##declare (not interrupts-enabled))

  (let ((x-length (##bignum.adigit-length x))
        (y-length (##bignum.adigit-length y)))
    (if (##fixnum.< x-length y-length)

      (let* ((result-length
              (##fixnum.+ y-length
                          (if (##eq? (##bignum.negative? x) 
                                     (##bignum.negative? y))
                            0
                            1)))
             (result
              (##bignum.make result-length y #t)))
        (let loop1 ((i 0)
                    (carry 1))
          (if (##fixnum.< i x-length)
            (loop1 (##fixnum.+ i 1)
                   (##bignum.adigit-add! result i x i carry))
            (##bignum.propagate-carry-and-normalize!
             result
             result-length
             x-length
             (##bignum.negative? x)
             (##fixnum.zero? carry)))))

      (let* ((result-length
              (##fixnum.+ x-length
                          (if (##eq? (##bignum.negative? x) 
                                     (##bignum.negative? y))
                            0
                            1)))
             (result
              (##bignum.make result-length x #f)))
        (let loop2 ((i 0)
                    (borrow 0))
          (if (##fixnum.< i y-length)
            (loop2 (##fixnum.+ i 1)
                   (##bignum.adigit-sub! result i y i borrow))
            (##bignum.propagate-carry-and-normalize!
             result
             result-length
             y-length
             (##not (##bignum.negative? y))
             (##not (##fixnum.zero? borrow)))))))))

(define (##bignum.propagate-carry-and-normalize!
         result
         result-length
         i
         borrow?
         propagate?)

  (##declare (not interrupts-enabled))

  (if (##eq? borrow? propagate?)
    (if borrow?

      (let loop1 ((i i)
                  (borrow 1))
        (if (and (##not (##fixnum.zero? borrow))
                 (##fixnum.< i result-length))
          (loop1 (##fixnum.+ i 1)
                 (##bignum.adigit-dec! result i))
          (##bignum.normalize! result)))

      (let loop2 ((i i)
                  (carry 1))
        (if (and (##not (##fixnum.zero? carry))
                 (##fixnum.< i result-length))
          (loop2 (##fixnum.+ i 1)
                 (##bignum.adigit-inc! result i))
          (##bignum.normalize! result))))

    (##bignum.normalize! result)))

(define (##bignum.normalize! result)

  (##declare (not interrupts-enabled))

  (let ((n (##fixnum.- (##bignum.adigit-length result) 1)))

    (cond ((##bignum.adigit-zero? result n)
           (let loop1 ((i (##fixnum.- n 1)))
             (cond ((##fixnum.< i 0)
                    0)
                   ((##bignum.adigit-zero? result i)
                    (loop1 (##fixnum.- i 1)))
                   ((##bignum.adigit-negative? result i)
                    (##bignum.adigit-shrink! result (##fixnum.+ i 2)))
                   (else
                    (or (and (##fixnum.= i 0)
                             (##bignum.->fixnum result))
                        (##bignum.adigit-shrink! result (##fixnum.+ i 1)))))))

          ((##bignum.adigit-ones? result n)
           (let loop2 ((i (##fixnum.- n 1)))
             (cond ((##fixnum.< i 0)
                    -1)
                   ((##bignum.adigit-ones? result i)
                    (loop2 (##fixnum.- i 1)))
                   ((##not (##bignum.adigit-negative? result i))
                    (##bignum.adigit-shrink! result (##fixnum.+ i 2)))
                   (else
                    (or (and (##fixnum.= i 0)
                             (##bignum.->fixnum result))
                        (##bignum.adigit-shrink! result (##fixnum.+ i 1)))))))

          ((and (##fixnum.= n 0)
                (##bignum.->fixnum result)) 
           =>
           (lambda (x) x))

          (else
           result))))

; Bignum multiplication.

(define (##bignum.* x y)

  ; Routines for real FFT translated from Ooura's code.

  (define (makewt nw ip ipcontrol w)

    (##declare (not interrupts-enabled))

    (##vector-set! ipcontrol 0 nw)
    (##vector-set! ipcontrol 1 1)
    (if (##fixnum.< 2 nw)
      (let* ((nwh
              (##fixnum.ashr nw 1))
             (delta
              (##flonum./ (##flonum.atan (macro-inexact-+1))
                          (##flonum.<-fixnum nwh))))
        (##f64vector-set! w 0 (macro-inexact-+1))
        (##f64vector-set! w 1 (macro-inexact-+0))
        (##f64vector-set! w nwh
          (##flonum.cos (##flonum.* delta (##flonum.<-fixnum nwh))))
        (##f64vector-set! w (##fixnum.+ nwh 1)
          (##f64vector-ref w nwh))
        (let loop ((j 2))
          (if (##fixnum.< j nwh)
            (let ((arg (##flonum.* delta (##flonum.<-fixnum j))))
              (let ((x (##flonum.cos arg))
                    (y (##flonum.sin arg)))
                (##f64vector-set! w j x)
                (##f64vector-set! w (##fixnum.+ j 1) y)
                (##f64vector-set! w (##fixnum.- nw j) y)
                (##f64vector-set! w (##fixnum.- (##fixnum.+ nw 1) j) x)
                (loop (##fixnum.+ j 2))))
            (bitrv2 nw ip w))))))

  (define (makect nc ipcontrol c)

    (##declare (not interrupts-enabled))

    (##vector-set! ipcontrol 1 nc)
    (if (##fixnum.< 1 nc)
      (let* ((nch
              (##fixnum.ashr nc 1))
             (delta
              (##flonum./ (##flonum.atan (macro-inexact-+1))
                          (##flonum.<-fixnum nch))))
        (##f64vector-set! c 0
          (##flonum.cos (##flonum.* delta (##flonum.<-fixnum nch))))
        (##f64vector-set! c nch
          (##flonum.* (macro-inexact-+1/2) (##f64vector-ref c 0)))
        (let loop ((j 1))
          (if (##fixnum.< j nch)
            (let ((arg (##flonum.* delta (##flonum.<-fixnum j))))
              (##f64vector-set! c j
                (##flonum.* (macro-inexact-+1/2) (##flonum.cos arg)))
              (##f64vector-set! c (##fixnum.- nc j)
                (##flonum.* (macro-inexact-+1/2) (##flonum.sin arg)))
              (loop (##fixnum.+ j 1))))))))

  (define (bitrv2 n ip a)

    (##declare (not interrupts-enabled))

    (##vector-set! ip 0 0)

    (let loop1 ((l n) (m 1))

      (cond ((##fixnum.< (##fixnum.shl m 2) l)
             (let ((l (##fixnum.ashr l 1)))
               (let loop2 ((j 0))
                 (if (##fixnum.< j m)
                   (begin
                     (##vector-set! ip (##fixnum.+ m j)
                       (##fixnum.+ (##vector-ref ip j) l))
                     (loop2 (##fixnum.+ j 1)))
                   (loop1 l (##fixnum.shl m 1))))))

            ((##fixnum.< l (##fixnum.shl m 2))
             (let loop3 ((k 1))
               (if (##fixnum.< k m)
                 (let loop4 ((j 0))
                   (if (##fixnum.< j k)
                     (let ((j1
                            (##fixnum.+ (##fixnum.shl j 1)
                                        (##vector-ref ip k)))
                           (k1
                            (##fixnum.+ (##fixnum.shl k 1)
                                        (##vector-ref ip j))))
                       (let ((xr (##f64vector-ref a (##fixnum.+ j1 0)))
                             (xi (##f64vector-ref a (##fixnum.+ j1 1))))
                         (##f64vector-set! a (##fixnum.+ j1 0)
                           (##f64vector-ref a (##fixnum.+ k1 0)))
                         (##f64vector-set! a (##fixnum.+ j1 1)
                           (##f64vector-ref a (##fixnum.+ k1 1)))
                         (##f64vector-set! a (##fixnum.+ k1 0) xr)
                         (##f64vector-set! a (##fixnum.+ k1 1) xi))
                       (loop4 (##fixnum.+ j 1)))
                     (loop3 (##fixnum.+ k 1)))))))

            (else 
             (let ((m2 (##fixnum.shl m 1)))
               (let loop5 ((k 1))
                 (if (##fixnum.< k m)
                   (let loop6 ((j 0))
                     (if (##fixnum.< j k)
                       (let ((j1
                              (##fixnum.+ (##fixnum.shl j 1)
                                          (##vector-ref ip k)))
                             (k1
                              (##fixnum.+ (##fixnum.shl k 1)
                                          (##vector-ref ip j))))
                         (let ((xr (##f64vector-ref a (##fixnum.+ j1 0)))
                               (xi (##f64vector-ref a (##fixnum.+ j1 1))))
                           (##f64vector-set! a (##fixnum.+ j1 0)
                             (##f64vector-ref a (##fixnum.+ k1 0)))
                           (##f64vector-set! a (##fixnum.+ j1 1)
                             (##f64vector-ref a (##fixnum.+ k1 1)))
                           (##f64vector-set! a (##fixnum.+ k1 0) xr)
                           (##f64vector-set! a (##fixnum.+ k1 1) xi))
                         (let ((j1 (##fixnum.+ j1 m2))
                               (k1 (##fixnum.+ k1 m2)))
                           (let ((xr (##f64vector-ref a (##fixnum.+ j1 0)))
                                 (xi (##f64vector-ref a (##fixnum.+ j1 1))))
                             (##f64vector-set! a (##fixnum.+ j1 0)
                               (##f64vector-ref a (##fixnum.+ k1 0)))
                             (##f64vector-set! a (##fixnum.+ j1 1)
                               (##f64vector-ref a (##fixnum.+ k1 1)))
                             (##f64vector-set! a (##fixnum.+ k1 0) xr)
                             (##f64vector-set! a (##fixnum.+ k1 1) xi)))
                         (loop6 (##fixnum.+ j 1)))
                       (loop5 (##fixnum.+ k 1)))))))))))

  (define (cftfsub n a w)
  
    (define (cleanup l)

      (##declare (not interrupts-enabled))

      (if (##fixnum.= (##fixnum.shl l 2) n)

        (let loop1 ((j 0))
          (if (##fixnum.< j l)
            (let* ((j1 (##fixnum.+ j  l))
                   (j2 (##fixnum.+ j1 l))
                   (j3 (##fixnum.+ j2 l)))
              (let ((aj00 (##f64vector-ref a (##fixnum.+ j  0)))
                    (aj01 (##f64vector-ref a (##fixnum.+ j  1)))
                    (aj10 (##f64vector-ref a (##fixnum.+ j1 0)))
                    (aj11 (##f64vector-ref a (##fixnum.+ j1 1)))
                    (aj20 (##f64vector-ref a (##fixnum.+ j2 0)))
                    (aj21 (##f64vector-ref a (##fixnum.+ j2 1)))
                    (aj30 (##f64vector-ref a (##fixnum.+ j3 0)))
                    (aj31 (##f64vector-ref a (##fixnum.+ j3 1))))
                (let ((x0r (##flonum.+ aj00 aj10))
                      (x0i (##flonum.+ aj01 aj11))
                      (x1r (##flonum.- aj00 aj10))
                      (x1i (##flonum.- aj01 aj11))
                      (x2r (##flonum.+ aj20 aj30))
                      (x2i (##flonum.+ aj21 aj31))
                      (x3r (##flonum.- aj20 aj30))
                      (x3i (##flonum.- aj21 aj31)))
                  (##f64vector-set! a (##fixnum.+ j  0) (##flonum.+ x0r x2r))
                  (##f64vector-set! a (##fixnum.+ j  1) (##flonum.+ x0i x2i))
                  (##f64vector-set! a (##fixnum.+ j2 0) (##flonum.- x0r x2r))
                  (##f64vector-set! a (##fixnum.+ j2 1) (##flonum.- x0i x2i))
                  (##f64vector-set! a (##fixnum.+ j1 0) (##flonum.- x1r x3i))
                  (##f64vector-set! a (##fixnum.+ j1 1) (##flonum.+ x1i x3r))
                  (##f64vector-set! a (##fixnum.+ j3 0) (##flonum.+ x1r x3i))
                  (##f64vector-set! a (##fixnum.+ j3 1) (##flonum.- x1i x3r))))
              (loop1 (##fixnum.+ j 2)))))

        (let loop2 ((j 0))
          (if (##fixnum.< j l)
            (let ((j1 (##fixnum.+ j l)))
              (let ((aj00 (##f64vector-ref a (##fixnum.+ j  0)))
                    (aj01 (##f64vector-ref a (##fixnum.+ j  1)))
                    (aj10 (##f64vector-ref a (##fixnum.+ j1 0)))
                    (aj11 (##f64vector-ref a (##fixnum.+ j1 1))))
                (let ((x0r (##flonum.- aj00 aj10))
                      (x0i (##flonum.- aj01 aj11)))
                  (##f64vector-set! a (##fixnum.+ j 0) (##flonum.+ aj00 aj10))
                  (##f64vector-set! a (##fixnum.+ j 1) (##flonum.+ aj01 aj11))
                  (##f64vector-set! a (##fixnum.+ j1 0) x0r)
                  (##f64vector-set! a (##fixnum.+ j1 1) x0i)))
              (loop2 (##fixnum.+ j 2)))))))

    (if (##fixnum.< 8 n)
      (begin
        (cft1st n a w)
        (let loop ((l 8))
          (if (##fixnum.< (##fixnum.shl l 2) n)
            (begin
              (cftmdl n l a w)
              (loop (##fixnum.shl l 2)))
            (cleanup l))))
      (cleanup 2)))

  (define (cftbsub n a w)
  
    (define (cleanup l)

      (##declare (not interrupts-enabled))

      (if (##fixnum.= (##fixnum.shl l 2) n)

        (let loop1 ((j 0))
          (if (##fixnum.< j l)
            (let* ((j1 (##fixnum.+ j  l))
                   (j2 (##fixnum.+ j1 l))
                   (j3 (##fixnum.+ j2 l)))
              (let ((aj00 (##f64vector-ref a (##fixnum.+ j  0)))
                    (aj01 (##f64vector-ref a (##fixnum.+ j  1)))
                    (aj10 (##f64vector-ref a (##fixnum.+ j1 0)))
                    (aj11 (##f64vector-ref a (##fixnum.+ j1 1)))
                    (aj20 (##f64vector-ref a (##fixnum.+ j2 0)))
                    (aj21 (##f64vector-ref a (##fixnum.+ j2 1)))
                    (aj30 (##f64vector-ref a (##fixnum.+ j3 0)))
                    (aj31 (##f64vector-ref a (##fixnum.+ j3 1))))
                (let ((x0r             (##flonum.+ aj00 aj10))
                      (x0i (##flonum.- (##flonum.+ aj01 aj11)))
                      (x1r             (##flonum.- aj00 aj10))
                      (x1i (##flonum.- (##flonum.- aj01 aj11)))
                      (x2r             (##flonum.+ aj20 aj30))
                      (x2i             (##flonum.+ aj21 aj31))
                      (x3r             (##flonum.- aj20 aj30))
                      (x3i             (##flonum.- aj21 aj31)))
                  (##f64vector-set! a (##fixnum.+ j  0) (##flonum.+ x0r x2r))
                  (##f64vector-set! a (##fixnum.+ j  1) (##flonum.- x0i x2i))
                  (##f64vector-set! a (##fixnum.+ j2 0) (##flonum.- x0r x2r))
                  (##f64vector-set! a (##fixnum.+ j2 1) (##flonum.+ x0i x2i))
                  (##f64vector-set! a (##fixnum.+ j1 0) (##flonum.- x1r x3i))
                  (##f64vector-set! a (##fixnum.+ j1 1) (##flonum.- x1i x3r))
                  (##f64vector-set! a (##fixnum.+ j3 0) (##flonum.+ x1r x3i))
                  (##f64vector-set! a (##fixnum.+ j3 1) (##flonum.+ x1i x3r))))
              (loop1 (##fixnum.+ j 2)))))

        (let loop2 ((j 0))
          (if (##fixnum.< j l)
            (let ((j1 (##fixnum.+ j l)))
              (let ((aj00 (##f64vector-ref a (##fixnum.+ j  0)))
                    (aj01 (##f64vector-ref a (##fixnum.+ j  1)))
                    (aj10 (##f64vector-ref a (##fixnum.+ j1 0)))
                    (aj11 (##f64vector-ref a (##fixnum.+ j1 1))))
                (let ((x0r             (##flonum.- aj00 aj10))
                      (x0i (##flonum.- (##flonum.- aj01 aj11))))
                  (##f64vector-set! a (##fixnum.+ j 0)
                    (##flonum.+ aj00 aj10))
                  (##f64vector-set! a (##fixnum.+ j 1)
                    (##flonum.- (##flonum.+ aj01 aj11)))
                  (##f64vector-set! a (##fixnum.+ j1 0) x0r)
                  (##f64vector-set! a (##fixnum.+ j1 1) x0i)))
              (loop2 (##fixnum.+ j 2)))))))

    (if (##fixnum.< 8 n)
      (begin
        (cft1st n a w)
        (let loop ((l 8))
          (if (##fixnum.< (##fixnum.shl l 2) n)
            (begin
              (cftmdl n l a w)
              (loop (##fixnum.shl l 2)))
            (cleanup l))))
      (cleanup 2)))

  (define (cft1st n a w)

    (##declare (not interrupts-enabled))

    (let ((a0 (##f64vector-ref a 0))
          (a1 (##f64vector-ref a 1))
          (a2 (##f64vector-ref a 2))
          (a3 (##f64vector-ref a 3))
          (a4 (##f64vector-ref a 4))
          (a5 (##f64vector-ref a 5))
          (a6 (##f64vector-ref a 6))
          (a7 (##f64vector-ref a 7)))
      (let ((x0r (##flonum.+ a0 a2))
            (x0i (##flonum.+ a1 a3))
            (x1r (##flonum.- a0 a2))
            (x1i (##flonum.- a1 a3))
            (x2r (##flonum.+ a4 a6))
            (x2i (##flonum.+ a5 a7))
            (x3r (##flonum.- a4 a6))
            (x3i (##flonum.- a5 a7)))
        (##f64vector-set! a 0 (##flonum.+ x0r x2r))
        (##f64vector-set! a 1 (##flonum.+ x0i x2i))
        (##f64vector-set! a 4 (##flonum.- x0r x2r))
        (##f64vector-set! a 5 (##flonum.- x0i x2i))
        (##f64vector-set! a 2 (##flonum.- x1r x3i))
        (##f64vector-set! a 3 (##flonum.+ x1i x3r))
        (##f64vector-set! a 6 (##flonum.+ x1r x3i))
        (##f64vector-set! a 7 (##flonum.- x1i x3r))))
    (let ((a8 (##f64vector-ref a 8))
          (a9 (##f64vector-ref a 9))
          (a10 (##f64vector-ref a 10))
          (a11 (##f64vector-ref a 11))
          (a12 (##f64vector-ref a 12))
          (a13 (##f64vector-ref a 13))
          (a14 (##f64vector-ref a 14))
          (a15 (##f64vector-ref a 15)))
      (let ((wk1r (##f64vector-ref w 2))
            (x0r (##flonum.+ a8 a10))
            (x0i (##flonum.+ a9 a11))
            (x1r (##flonum.- a8 a10))
            (x1i (##flonum.- a9 a11))
            (x2r (##flonum.+ a12 a14))
            (x2i (##flonum.+ a13 a15))
            (x3r (##flonum.- a12 a14))
            (x3i (##flonum.- a13 a15)))
        (##f64vector-set! a 8 (##flonum.+ x0r x2r))
        (##f64vector-set! a 9 (##flonum.+ x0i x2i))
        (##f64vector-set! a 12 (##flonum.- x2i x0i))
        (##f64vector-set! a 13 (##flonum.- x0r x2r))
        (let ((x0r (##flonum.- x1r x3i))
              (x0i (##flonum.+ x1i x3r)))
          (##f64vector-set! a 10 (##flonum.* wk1r (##flonum.- x0r x0i)))
          (##f64vector-set! a 11 (##flonum.* wk1r (##flonum.+ x0r x0i))))
        (let ((x0r (##flonum.+ x3i x1r))
              (x0i (##flonum.- x3r x1i)))
          (##f64vector-set! a 14 (##flonum.* wk1r (##flonum.- x0i x0r)))
          (##f64vector-set! a 15 (##flonum.* wk1r (##flonum.+ x0i x0r))))))
    (let loop ((j 16) (k1 2))
      (if (##fixnum.< j n)
        (let ((k2 (##fixnum.shl k1 1))
              (wk2r (##f64vector-ref w k1))
              (wk2i (##f64vector-ref w (##fixnum.+ k1 1))))
          (let ((wk1r (##f64vector-ref w k2))
                (wk1i (##f64vector-ref w (##fixnum.+ k2 1))))
            (let ((a0 (##f64vector-ref a (##fixnum.+ j 0)))
                  (a1 (##f64vector-ref a (##fixnum.+ j 1)))
                  (a2 (##f64vector-ref a (##fixnum.+ j 2)))
                  (a3 (##f64vector-ref a (##fixnum.+ j 3)))
                  (a4 (##f64vector-ref a (##fixnum.+ j 4)))
                  (a5 (##f64vector-ref a (##fixnum.+ j 5)))
                  (a6 (##f64vector-ref a (##fixnum.+ j 6)))
                  (a7 (##f64vector-ref a (##fixnum.+ j 7))))
              (let ((wk3r
                     (##flonum.- wk1r
                                 (##flonum.* (macro-inexact-+2)
                                             (##flonum.* wk2i wk1i))))
                    (wk3i
                     (##flonum.- (##flonum.* (macro-inexact-+2)
                                             (##flonum.* wk2i wk1r))
                                 wk1i))
                    (x0r (##flonum.+ a0 a2))
                    (x0i (##flonum.+ a1 a3))
                    (x1r (##flonum.- a0 a2))
                    (x1i (##flonum.- a1 a3))
                    (x2r (##flonum.+ a4 a6))
                    (x2i (##flonum.+ a5 a7))
                    (x3r (##flonum.- a4 a6))
                    (x3i (##flonum.- a5 a7)))
                (##f64vector-set! a (##fixnum.+ j 0) (##flonum.+ x0r x2r))
                (##f64vector-set! a (##fixnum.+ j 1) (##flonum.+ x0i x2i))
                (let ((x0r (##flonum.- x0r x2r))
                      (x0i (##flonum.- x0i x2i)))
                  (##f64vector-set! a (##fixnum.+ j 4)
                    (##flonum.- (##flonum.* wk2r x0r)
                                (##flonum.* wk2i x0i)))
                  (##f64vector-set! a (##fixnum.+ j 5)
                    (##flonum.+ (##flonum.* wk2r x0i)
                                (##flonum.* wk2i x0r))))
                (let ((x0r (##flonum.- x1r x3i))
                      (x0i (##flonum.+ x1i x3r)))
                  (##f64vector-set! a (##fixnum.+ j 2)
                    (##flonum.- (##flonum.* wk1r x0r)
                                (##flonum.* wk1i x0i)))
                  (##f64vector-set! a (##fixnum.+ j 3)
                    (##flonum.+ (##flonum.* wk1r x0i)
                                (##flonum.* wk1i x0r))))
                (let ((x0r (##flonum.+ x1r x3i))
                      (x0i (##flonum.- x1i x3r)))
                  (##f64vector-set! a (##fixnum.+ j 6)
                    (##flonum.- (##flonum.* wk3r x0r)
                                (##flonum.* wk3i x0i)))
                  (##f64vector-set! a (##fixnum.+ j 7)
                    (##flonum.+ (##flonum.* wk3r x0i)
                                (##flonum.* wk3i x0r)))))))
          (let ((wk1r (##f64vector-ref w (##fixnum.+ k2 2)))
                (wk1i (##f64vector-ref w (##fixnum.+ k2 3))))
            (let (( a8 (##f64vector-ref a (##fixnum.+ j  8)))
                  ( a9 (##f64vector-ref a (##fixnum.+ j  9)))
                  (a10 (##f64vector-ref a (##fixnum.+ j 10)))
                  (a11 (##f64vector-ref a (##fixnum.+ j 11)))
                  (a12 (##f64vector-ref a (##fixnum.+ j 12)))
                  (a13 (##f64vector-ref a (##fixnum.+ j 13)))
                  (a14 (##f64vector-ref a (##fixnum.+ j 14)))
                  (a15 (##f64vector-ref a (##fixnum.+ j 15))))
              (let ((wk3r
                     (##flonum.- wk1r
                                 (##flonum.* (macro-inexact-+2)
                                             (##flonum.* wk2r wk1i))))
                    (wk3i
                     (##flonum.- (##flonum.* (macro-inexact-+2)
                                             (##flonum.* wk2r wk1r))
                                 wk1i))
                    (x0r (##flonum.+ a8 a10))
                    (x0i (##flonum.+ a9 a11))
                    (x1r (##flonum.- a8 a10))
                    (x1i (##flonum.- a9 a11))
                    (x2r (##flonum.+ a12 a14))
                    (x2i (##flonum.+ a13 a15))
                    (x3r (##flonum.- a12 a14))
                    (x3i (##flonum.- a13 a15)))
                (##f64vector-set! a (##fixnum.+ j 8) (##flonum.+ x0r x2r))
                (##f64vector-set! a (##fixnum.+ j 9) (##flonum.+ x0i x2i))
                (let ((x0r (##flonum.- x0r x2r))
                      (x0i (##flonum.- x0i x2i)))
                  (##f64vector-set! a (##fixnum.+ j 12)
                    (##flonum.- (##flonum.+ (##flonum.* wk2i x0r)
                                            (##flonum.* wk2r x0i))))
                  (##f64vector-set! a (##fixnum.+ j 13)
                    (##flonum.- (##flonum.- (##flonum.* wk2i x0i)
                                            (##flonum.* wk2r x0r)))))
                (let ((x0r (##flonum.- x1r x3i))
                      (x0i (##flonum.+ x1i x3r)))
                  (##f64vector-set! a (##fixnum.+ j 10)
                    (##flonum.- (##flonum.* wk1r x0r)
                                (##flonum.* wk1i x0i)))
                  (##f64vector-set! a (##fixnum.+ j 11)
                    (##flonum.+ (##flonum.* wk1r x0i)
                                (##flonum.* wk1i x0r))))
                (let ((x0r (##flonum.+ x1r x3i))
                      (x0i (##flonum.- x1i x3r)))
                  (##f64vector-set! a (##fixnum.+ j 14)
                    (##flonum.- (##flonum.* wk3r x0r)
                                (##flonum.* wk3i x0i)))
                  (##f64vector-set! a (##fixnum.+ j 15)
                    (##flonum.+ (##flonum.* wk3r x0i)
                                (##flonum.* wk3i x0r)))))))
          (loop (##fixnum.+ j 16) (##fixnum.+ k1 2))))))

  (define (cftmdl n l a w)

    (##declare (not interrupts-enabled))

    (let ((m (##fixnum.shl l 2)))

      (let loop1 ((j 0))
        (if (##fixnum.< j l)
          (let* ((j1 (##fixnum.+ j l))
                 (j2 (##fixnum.+ j1 l))
                 (j3 (##fixnum.+ j2 l)))
            (let ((aj00 (##f64vector-ref a (##fixnum.+ j  0)))
                  (aj01 (##f64vector-ref a (##fixnum.+ j  1)))
                  (aj10 (##f64vector-ref a (##fixnum.+ j1 0)))
                  (aj11 (##f64vector-ref a (##fixnum.+ j1 1)))
                  (aj20 (##f64vector-ref a (##fixnum.+ j2 0)))
                  (aj21 (##f64vector-ref a (##fixnum.+ j2 1)))
                  (aj30 (##f64vector-ref a (##fixnum.+ j3 0)))
                  (aj31 (##f64vector-ref a (##fixnum.+ j3 1))))
              (let ((x0r (##flonum.+ aj00 aj10))
                    (x0i (##flonum.+ aj01 aj11))
                    (x1r (##flonum.- aj00 aj10))
                    (x1i (##flonum.- aj01 aj11))
                    (x2r (##flonum.+ aj20 aj30))
                    (x2i (##flonum.+ aj21 aj31))
                    (x3r (##flonum.- aj20 aj30))
                    (x3i (##flonum.- aj21 aj31)))
                (##f64vector-set! a (##fixnum.+ j  0) (##flonum.+ x0r x2r))
                (##f64vector-set! a (##fixnum.+ j  1) (##flonum.+ x0i x2i))
                (##f64vector-set! a (##fixnum.+ j2 0) (##flonum.- x0r x2r))
                (##f64vector-set! a (##fixnum.+ j2 1) (##flonum.- x0i x2i))
                (##f64vector-set! a (##fixnum.+ j1 0) (##flonum.- x1r x3i))
                (##f64vector-set! a (##fixnum.+ j1 1) (##flonum.+ x1i x3r))
                (##f64vector-set! a (##fixnum.+ j3 0) (##flonum.+ x1r x3i))
                (##f64vector-set! a (##fixnum.+ j3 1) (##flonum.- x1i x3r))
                (loop1 (##fixnum.+ j 2)))))))

      (let ((wk1r (##f64vector-ref w 2)))
        (let loop2 ((j m))
          (if (##fixnum.< j (##fixnum.+ l m))
            (let* ((j1 (##fixnum.+ j l))
                   (j2 (##fixnum.+ j1 l))
                   (j3 (##fixnum.+ j2 l)))
              (let ((aj00 (##f64vector-ref a (##fixnum.+ j  0)))
                    (aj01 (##f64vector-ref a (##fixnum.+ j  1)))
                    (aj10 (##f64vector-ref a (##fixnum.+ j1 0)))
                    (aj11 (##f64vector-ref a (##fixnum.+ j1 1)))
                    (aj20 (##f64vector-ref a (##fixnum.+ j2 0)))
                    (aj21 (##f64vector-ref a (##fixnum.+ j2 1)))
                    (aj30 (##f64vector-ref a (##fixnum.+ j3 0)))
                    (aj31 (##f64vector-ref a (##fixnum.+ j3 1))))
                (let ((x0r (##flonum.+ aj00 aj10))
                      (x0i (##flonum.+ aj01 aj11))
                      (x1r (##flonum.- aj00 aj10))
                      (x1i (##flonum.- aj01 aj11))
                      (x2r (##flonum.+ aj20 aj30))
                      (x2i (##flonum.+ aj21 aj31))
                      (x3r (##flonum.- aj20 aj30))
                      (x3i (##flonum.- aj21 aj31)))
                  (##f64vector-set! a (##fixnum.+ j  0) (##flonum.+ x0r x2r))
                  (##f64vector-set! a (##fixnum.+ j  1) (##flonum.+ x0i x2i))
                  (##f64vector-set! a (##fixnum.+ j2 0) (##flonum.- x2i x0i))
                  (##f64vector-set! a (##fixnum.+ j2 1) (##flonum.- x0r x2r))
                  (let ((x0r (##flonum.- x1r x3i))
                        (x0i (##flonum.+ x1i x3r)))
                    (##f64vector-set! a (##fixnum.+ j1 0)
                      (##flonum.* wk1r (##flonum.- x0r x0i)))
                    (##f64vector-set! a (##fixnum.+ j1 1)
                      (##flonum.* wk1r (##flonum.+ x0r x0i))))
                  (let ((x0r (##flonum.+ x3i x1r))
                        (x0i (##flonum.- x3r x1i)))
                    (##f64vector-set! a (##fixnum.+ j3 0)
                      (##flonum.* wk1r (##flonum.- x0i x0r)))
                    (##f64vector-set! a (##fixnum.+ j3 1)
                      (##flonum.* wk1r (##flonum.+ x0i x0r))))
                  (loop2 (##fixnum.+ j 2))))))))

      (let ((m2 (##fixnum.shl m 1)))
        (let loop3 ((k m2) (k1 2))
          (if (##fixnum.< k n)
            (let* ((k2 (##fixnum.shl k1 1))
                   (wk2r (##f64vector-ref w (##fixnum.+ k1 0)))
                   (wk2i (##f64vector-ref w (##fixnum.+ k1 1)))
                   (wk1r (##f64vector-ref w (##fixnum.+ k2 0)))
                   (wk1i (##f64vector-ref w (##fixnum.+ k2 1)))
                   (wk3r
                    (##flonum.- wk1r
                                (##flonum.* (macro-inexact-+2)
                                            (##flonum.* wk2i wk1i))))
                   (wk3i
                    (##flonum.- (##flonum.* (macro-inexact-+2)
                                            (##flonum.* wk2i wk1r))
                                wk1i)))
              (let loop4 ((j k))
                (if (##fixnum.< j (##fixnum.+ l k))

                  (let* ((j1 (##fixnum.+ j l))
                         (j2 (##fixnum.+ j1 l))
                         (j3 (##fixnum.+ j2 l)))
                    (let ((aj00 (##f64vector-ref a (##fixnum.+ j  0)))
                          (aj01 (##f64vector-ref a (##fixnum.+ j  1)))
                          (aj10 (##f64vector-ref a (##fixnum.+ j1 0)))
                          (aj11 (##f64vector-ref a (##fixnum.+ j1 1)))
                          (aj20 (##f64vector-ref a (##fixnum.+ j2 0)))
                          (aj21 (##f64vector-ref a (##fixnum.+ j2 1)))
                          (aj30 (##f64vector-ref a (##fixnum.+ j3 0)))
                          (aj31 (##f64vector-ref a (##fixnum.+ j3 1))))
                      (let ((x0r (##flonum.+ aj00 aj10))
                            (x0i (##flonum.+ aj01 aj11))
                            (x1r (##flonum.- aj00 aj10))
                            (x1i (##flonum.- aj01 aj11))
                            (x2r (##flonum.+ aj20 aj30))
                            (x2i (##flonum.+ aj21 aj31))
                            (x3r (##flonum.- aj20 aj30))
                            (x3i (##flonum.- aj21 aj31)))
                        (##f64vector-set! a (##fixnum.+ j 0)
                          (##flonum.+ x0r x2r))
                        (##f64vector-set! a (##fixnum.+ j 1)
                          (##flonum.+ x0i x2i))
                        (let ((x0r (##flonum.- x0r x2r))
                              (x0i (##flonum.- x0i x2i)))
                          (##f64vector-set! a (##fixnum.+ j2 0)
                            (##flonum.- (##flonum.* wk2r x0r)
                                        (##flonum.* wk2i x0i)))
                          (##f64vector-set! a (##fixnum.+ j2 1)
                            (##flonum.+ (##flonum.* wk2r x0i)
                                        (##flonum.* wk2i x0r))))
                        (let ((x0r (##flonum.- x1r x3i))
                              (x0i (##flonum.+ x1i x3r)))
                          (##f64vector-set! a (##fixnum.+ j1 0)
                            (##flonum.- (##flonum.* wk1r x0r)
                                        (##flonum.* wk1i x0i)))
                          (##f64vector-set! a (##fixnum.+ j1 1)
                            (##flonum.+ (##flonum.* wk1r x0i)
                                        (##flonum.* wk1i x0r))))
                        (let ((x0r (##flonum.+ x1r x3i))
                              (x0i (##flonum.- x1i x3r)))
                          (##f64vector-set! a (##fixnum.+ j3 0)
                            (##flonum.- (##flonum.* wk3r x0r)
                                        (##flonum.* wk3i x0i)))
                          (##f64vector-set! a (##fixnum.+ j3 1)
                            (##flonum.+ (##flonum.* wk3r x0i)
                                        (##flonum.* wk3i x0r))))
                        (loop4 (##fixnum.+ j 2)))))

                  (let* ((wk1r
                          (##f64vector-ref w (##fixnum.+ k2 2)))
                         (wk1i
                          (##f64vector-ref w (##fixnum.+ k2 3)))
                         (wk3r
                          (##flonum.- wk1r
                                      (##flonum.* (macro-inexact-+2)
                                                  (##flonum.* wk2r wk1i))))
                         (wk3i
                          (##flonum.- (##flonum.* (macro-inexact-+2)
                                                  (##flonum.* wk2r wk1r))
                                      wk1i)))

                    (let loop5 ((j (##fixnum.+ k m)))
                      (if (##fixnum.< j (##fixnum.+ l k m))
                        (let* ((j1 (##fixnum.+ j l))
                               (j2 (##fixnum.+ j1 l))
                               (j3 (##fixnum.+ j2 l)))
                          (let ((aj00 (##f64vector-ref a (##fixnum.+ j  0)))
                                (aj01 (##f64vector-ref a (##fixnum.+ j  1)))
                                (aj10 (##f64vector-ref a (##fixnum.+ j1 0)))
                                (aj11 (##f64vector-ref a (##fixnum.+ j1 1)))
                                (aj20 (##f64vector-ref a (##fixnum.+ j2 0)))
                                (aj21 (##f64vector-ref a (##fixnum.+ j2 1)))
                                (aj30 (##f64vector-ref a (##fixnum.+ j3 0)))
                                (aj31 (##f64vector-ref a (##fixnum.+ j3 1))))
                            (let ((x0r (##flonum.+ aj00 aj10))
                                  (x0i (##flonum.+ aj01 aj11))
                                  (x1r (##flonum.- aj00 aj10))
                                  (x1i (##flonum.- aj01 aj11))
                                  (x2r (##flonum.+ aj20 aj30))
                                  (x2i (##flonum.+ aj21 aj31))
                                  (x3r (##flonum.- aj20 aj30))
                                  (x3i (##flonum.- aj21 aj31)))
                              (##f64vector-set! a (##fixnum.+ j 0)
                                (##flonum.+ x0r x2r))
                              (##f64vector-set! a (##fixnum.+ j 1)
                                (##flonum.+ x0i x2i))
                              (let ((x0r (##flonum.- x0r x2r))
                                    (x0i (##flonum.- x0i x2i)))
                                (##f64vector-set! a (##fixnum.+ j2 0)
                                  (##flonum.-
                                   (##flonum.+ (##flonum.* wk2i x0r)
                                               (##flonum.* wk2r x0i))))
                                (##f64vector-set! a (##fixnum.+ j2 1)
                                  (##flonum.- (##flonum.* wk2r x0r)
                                              (##flonum.* wk2i x0i))))
                              (let ((x0r (##flonum.- x1r x3i))
                                    (x0i (##flonum.+ x1i x3r)))
                                (##f64vector-set! a (##fixnum.+ j1 0)
                                  (##flonum.- (##flonum.* wk1r x0r)
                                              (##flonum.* wk1i x0i)))
                                (##f64vector-set! a (##fixnum.+ j1 1)
                                  (##flonum.+ (##flonum.* wk1r x0i)
                                              (##flonum.* wk1i x0r))))
                              (let ((x0r (##flonum.+ x1r x3i))
                                    (x0i (##flonum.- x1i x3r)))
                                (##f64vector-set! a (##fixnum.+ j3 0)
                                  (##flonum.- (##flonum.* wk3r x0r)
                                              (##flonum.* wk3i x0i)))
                                (##f64vector-set! a (##fixnum.+ j3 1)
                                  (##flonum.+ (##flonum.* wk3r x0i)
                                              (##flonum.* wk3i x0r))))
                              (loop5 (##fixnum.+ j 2)))))))

                    (loop3 (##fixnum.+ k m2) (##fixnum.+ k1 2)))))))))))

  (define (rftfsub n a nc c)

    (##declare (not interrupts-enabled))

    (let ((ks (##fixnum.quotient (##fixnum.shl nc 2) n))
          (m (##fixnum.ashr n 1)))
      (let loop ((k 2)
                 (kk ks))
        (if (##fixnum.< k m)
          (let ((j (##fixnum.- n k)))
            (let ((wkr
                   (##flonum.- (macro-inexact-+1/2)
                               (##f64vector-ref c (##fixnum.- nc kk))))
                  (wki (##f64vector-ref c kk))
                  (aj0 (##f64vector-ref a (##fixnum.+ j 0)))
                  (aj1 (##f64vector-ref a (##fixnum.+ j 1)))
                  (ak0 (##f64vector-ref a (##fixnum.+ k 0)))
                  (ak1 (##f64vector-ref a (##fixnum.+ k 1))))
              (let ((xr (##flonum.- ak0 aj0))
                    (xi (##flonum.+ ak1 aj1)))
                (let ((yr
                       (##flonum.- (##flonum.* wkr xr) (##flonum.* wki xi)))
                      (yi
                       (##flonum.+ (##flonum.* wkr xi) (##flonum.* wki xr))))
                  (##f64vector-set! a (##fixnum.+ k 0) (##flonum.- ak0 yr))
                  (##f64vector-set! a (##fixnum.+ k 1) (##flonum.- ak1 yi))
                  (##f64vector-set! a (##fixnum.+ j 0) (##flonum.+ aj0 yr))
                  (##f64vector-set! a (##fixnum.+ j 1) (##flonum.- aj1 yi)))))
            (loop (##fixnum.+ k 2)
                  (##fixnum.+ kk ks)))))))

  (define (rftbsub n a nc c)

    (##declare (not interrupts-enabled))

    (let ((ks (##fixnum.quotient (##fixnum.shl nc 2) n))
          (m (##fixnum.ashr n 1)))
      (##f64vector-set! a 1 (##flonum.- (##f64vector-ref a 1)))
      (let loop ((k 2)
                 (kk ks))
        (if (##fixnum.< k m)
          (let ((j (##fixnum.- n k)))
            (let ((wkr
                   (##flonum.- (macro-inexact-+1/2)
                               (##f64vector-ref c (##fixnum.- nc kk))))
                  (wki (##f64vector-ref c kk))
                  (aj0 (##f64vector-ref a (##fixnum.+ j 0)))
                  (aj1 (##f64vector-ref a (##fixnum.+ j 1)))
                  (ak0 (##f64vector-ref a (##fixnum.+ k 0)))
                  (ak1 (##f64vector-ref a (##fixnum.+ k 1))))
              (let ((xr (##flonum.- ak0 aj0))
                    (xi (##flonum.+ ak1 aj1)))
                (let ((yr
                       (##flonum.+ (##flonum.* wkr xr) (##flonum.* wki xi)))
                      (yi
                       (##flonum.- (##flonum.* wkr xi) (##flonum.* wki xr))))
                  (##f64vector-set! a (##fixnum.+ k 0) (##flonum.- ak0 yr))
                  (##f64vector-set! a (##fixnum.+ k 1) (##flonum.- yi ak1))
                  (##f64vector-set! a (##fixnum.+ j 0) (##flonum.+ aj0 yr))
                  (##f64vector-set! a (##fixnum.+ j 1) (##flonum.- yi aj1)))))
            (loop (##fixnum.+ k 2)
                  (##fixnum.+ kk ks)))
          (##f64vector-set! a (##fixnum.+ m 1)
            (##flonum.- (##f64vector-ref a (##fixnum.+ m 1))))))))

  (define (rdft n isgn a ip ipcontrol w wc)

    (##declare (not interrupts-enabled))

    (let* ((nw0
            (##vector-ref ipcontrol 0))
           (nw
            (if (##fixnum.< (##fixnum.shl nw0 2) n)
              (let ((nw (##fixnum.ashr n 2)))
                (makewt nw ip ipcontrol w)
                nw)
              nw0))
           (nc0
            (##vector-ref ipcontrol 1))
           (nc
            (if (##fixnum.< (##fixnum.shl nc0 2) n)
              (let ((nc (##fixnum.ashr n 2)))
                (makect nc ipcontrol wc)
                nc)
              nc0)))

      (if (##fixnum.< isgn 0)
        (begin
          (##f64vector-set! a 1
            (##flonum.* (macro-inexact-+1/2)
                        (##flonum.- (##f64vector-ref a 0)
                                    (##f64vector-ref a 1))))
          (##f64vector-set! a 0
            (##flonum.- (##f64vector-ref a 0)
                        (##f64vector-ref a 1)))
          (if (##fixnum.< 4 n)
            (begin
              (rftbsub n a nc wc)
              (bitrv2 n ip a)
              (cftbsub n a w))
            (if (##fixnum.= n 4)
              (cftfsub n a w))))
        (begin
          (if (##fixnum.< 4 n)
            (begin
              (bitrv2 n ip a)
              (cftfsub n a w)
              (rftfsub n a nc wc))
            (if (##fixnum.= n 4)
              (cftfsub n a w)))
          (let ((xi (##flonum.- (##f64vector-ref a 0) (##f64vector-ref a 1))))
            (##f64vector-set! a 0
              (##flonum.+ (##f64vector-ref a 0)
                          (##f64vector-ref a 1)))
            (##f64vector-set! a 1 xi))))))

  (define (fft-mul x y)

    (##declare (not interrupts-enabled))

    (define-structure fft-state two-n ipcontrol w wc ip a b)

    (define (make-new-state two-n)
      (make-fft-state
       two-n
       (##vector 0 0)
       (##make-f64vector (##fixnum.ashr two-n 2) (macro-inexact-+0))
       (##make-f64vector (##fixnum.ashr two-n 2) (macro-inexact-+0))
       (##make-vector (##fixnum.+ 
                       2 
                       (##flonum.->fixnum 
                        (##flonum.sqrt 
                         (##flonum.<-fixnum two-n))))
                      0)
       (##make-f64vector two-n (macro-inexact-+0))
       (##make-f64vector two-n (macro-inexact-+0))))

    (let* ((x-length
            (##bignum.fdigit-length x))
           (y-length 
            (##bignum.fdigit-length y))
           (mn
            (##fixnum.max x-length y-length))
           (two-n
            (##fixnum.shl 2 (##exact-int.width (##fixnum.- mn 1))))
           (normalizer
            (##flonum./ (##flonum.<-fixnum (##fixnum.ashr two-n 1))))
           (internal-state
            (make-new-state two-n)))

      (let ((ipcontrol (fft-state-ipcontrol internal-state))
            (w         (fft-state-w         internal-state))
            (wc        (fft-state-wc        internal-state))
            (ip        (fft-state-ip        internal-state))
            (a         (fft-state-a         internal-state)))

        (let loop1 ((i 0))
          (if (##fixnum.< i x-length)
            (begin
              (##f64vector-set! a i
                (##flonum.<-fixnum (##bignum.fdigit-ref x i)))
              (loop1 (##fixnum.+ i 1)))))

        (rdft two-n 1 a ip ipcontrol w wc)

        (if (##eq? x y) ; fast handling of squaring an integer

          (begin
            (##f64vector-set! a 0
              (##flonum.* (##f64vector-ref a 0) (##f64vector-ref a 0)))
            (##f64vector-set! a 1
              (##flonum.* (##f64vector-ref a 1) (##f64vector-ref a 1)))
            (let loop2 ((i 2))
              (if (##fixnum.< i two-n)
                (let ((ai (##f64vector-ref a i))
                      (ai+1 (##f64vector-ref a (##fixnum.+ i 1))))
                  (##f64vector-set! a i
                    (##flonum.- (##flonum.* ai ai)   (##flonum.* ai+1 ai+1)))
                  (##f64vector-set! a (##fixnum.+ i 1)
                    (##flonum.+ (##flonum.* ai ai+1) (##flonum.* ai   ai+1)))
                  (loop2 (##fixnum.+ i 2))))))

            (let ((b (fft-state-b internal-state)))
              (let loop3 ((i 0))
                (if (##fixnum.< i y-length)
                  (begin
                    (##f64vector-set! b i
                      (##flonum.<-fixnum (##bignum.fdigit-ref y i)))
                    (loop3 (##fixnum.+ i 1)))))
              (rdft two-n 1 b ip ipcontrol w wc)
              (##f64vector-set! a 0
                (##flonum.* (##f64vector-ref b 0) (##f64vector-ref a 0)))
              (##f64vector-set! a 1
                (##flonum.* (##f64vector-ref b 1) (##f64vector-ref a 1)))
              (let loop4 ((i 2))
                (if (##fixnum.< i two-n)
                  (let ((ai (##f64vector-ref a i))
                        (ai+1 (##f64vector-ref a (##fixnum.+ i 1)))
                        (bi (##f64vector-ref b i))
                        (bi+1 (##f64vector-ref b (##fixnum.+ i 1))))
                    (##f64vector-set! a i
                      (##flonum.- (##flonum.* bi ai)   (##flonum.* ai+1 bi+1)))
                    (##f64vector-set! a (##fixnum.+ i 1)
                      (##flonum.+ (##flonum.* bi ai+1) (##flonum.* ai   bi+1)))
                    (loop4 (##fixnum.+ i 2)))))))

        (rdft two-n -1 a ip ipcontrol w wc)

        (let* ((result-length
                (##fixnum.+ x-length y-length))
               (fbase
                (##flonum.<-fixnum ##bignum.fdigit-base))
               (result
                (##bignum.make
                 (##fixnum.quotient
                  result-length
                  (##fixnum.quotient ##bignum.adigit-width
                                     ##bignum.fdigit-width))
                 #f 
                 #f)))
          (let ((loop-carry (##f64vector (macro-inexact-+0))))
            (let loop5 ((i 0))
              (if (##fixnum.< i result-length)
                  (let* ((t
                          (##flonum.+ (##flonum.+ (##f64vector-ref loop-carry 0)
                                                  (macro-inexact-+1/2))
                                      (##flonum.* (##f64vector-ref a i)
                                                  normalizer)))
                         (carry
                          (##flonum.floor (##flonum./ t fbase)))
                         (digit
                          (##flonum.- t (##flonum.* carry fbase))))
                    (##bignum.fdigit-set! result i (##flonum.->fixnum digit))
                    (##f64vector-set! loop-carry 0 carry)
                    (loop5 (##fixnum.+ i 1)))
                  (cleanup x y result))))))))

  (define (naive-mul x x-length y y-length)
    (let ((result
           (##bignum.make
            (##fixnum.+ (##bignum.adigit-length x) (##bignum.adigit-length y))
            #f 
            #f)))

      (let loop1 ((k 0))
        (if (##fixnum.< k y-length)
          (let ((multiplier (##bignum.mdigit-ref y k)))
            (if (##not (##eq? multiplier 0))
              (let loop2 ((i 0)
                          (j k)
                          (carry 0))

                (##declare (not interrupts-enabled))

                (if (##fixnum.< i x-length)
                  (loop2 (##fixnum.+ i 2)
                         (##fixnum.+ j 2)
                         (##bignum.mdigit-mul! result
                                               (##fixnum.+ j 1)
                                               x
                                               (##fixnum.+ i 1)
                                               multiplier 
                                               (##bignum.mdigit-mul! result
                                                                     j
                                                                     x
                                                                     i
                                                                     multiplier
                                                                     carry)))
                  (##bignum.mdigit-set! result j carry))))
            (loop1 (##fixnum.+ k 1)))))

      (cleanup x y result)))
  
  (define (cleanup x y result)

    ; Both naive-mul and fft-mul do unsigned multiplies, fix that here.

    (##declare (not interrupts-enabled))

    (define (fix x y result)
      (if (##bignum.negative? y)
        (let ((x-length (##bignum.adigit-length x)))
          (let loop ((i 0)
                     (j (##bignum.adigit-length y))
                     (borrow 0))
            (if (##fixnum.< i x-length)
              (loop (##fixnum.+ i 1)
                    (##fixnum.+ j 1)
                    (##bignum.adigit-sub! result j x i borrow)))))))

    (fix x y result)
    (fix y x result)
    (##bignum.normalize! result))

  (define (karatsuba-mul x y)
    (let* ((x-length
            (##bignum.adigit-length x))
           (y-length
            (##bignum.adigit-length y))
           (shift-digits
            (##fixnum.ashr y-length 1))
           (shift-bits
            (##fixnum.* shift-digits ##bignum.adigit-width))
           (y-high
            (##bignum.shift y (##fixnum.- shift-bits)))
           (y-low
            (let ((result ; make sure top digit is 0
                   (##bignum.make (##fixnum.+ shift-digits 1) y #f)))
              (##bignum.adigit-set! result
                                    shift-digits
                                    ##bignum.adigit-zeros
                                    0)
              (##bignum.normalize! result))))
      (if (##fixnum.< shift-digits x-length)

        (let ((x-high
               (##bignum.shift x (##fixnum.- shift-bits)))
              (x-low
               (let ((result ; make sure top digit is 0
                      (##bignum.make (##fixnum.+ shift-digits 1) x #f)))
                 (##bignum.adigit-set! result
                                       shift-digits
                                       ##bignum.adigit-zeros
                                       0)
                 (##bignum.normalize! result))))
          (let ((high-term
                 (##* x-high y-high))
                (low-term
                 (##* x-low y-low))
                (mid-term
                 (##* (##- x-high x-low)
                      (##- y-low y-high))))
            (##+
             (##exact-int.shift high-term (##fixnum.+ shift-bits shift-bits))
             (##+ (##exact-int.shift
                   (##+ high-term
                        (##+ mid-term low-term))
                   shift-bits)
                  low-term))))

        (##+ (##exact-int.shift (##* x y-high) shift-bits)
             (##* x y-low)))))

  (define (mul x x-length y y-length) ; x-length <= y-length
    (let ((x-width (##fixnum.* x-length ##bignum.mdigit-width)))
      (if (##fixnum.< ##bignum.naive-mul-max-width x-width)
        (let ((y-width (##fixnum.* y-length ##bignum.mdigit-width)))
          (if (or (##fixnum.< y-width ##bignum.fft-mul-min-width)
                  (##fixnum.< ##bignum.fft-mul-max-width y-width))
            (karatsuba-mul x y)
            (fft-mul x y)))
        (naive-mul y y-length x x-length))))

  (let ((x-length (##bignum.mdigit-length x))
        (y-length (##bignum.mdigit-length y)))
    (if (##fixnum.< x-length y-length)
      (mul x x-length y y-length)
      (mul y y-length x x-length))))

(define (##bignum.shift x shift)
  (let* ((digit-shift
          (if (##fixnum.< shift 0)
            (##fixnum.- (##fixnum.quotient (##fixnum.+ shift 1)
                                           ##bignum.adigit-width)
                        1)
            (##fixnum.quotient shift ##bignum.adigit-width)))
         (bit-shift
          (##fixnum.modulo shift ##bignum.adigit-width))
         (x-length
          (##bignum.adigit-length x))
         (result-length
          (##fixnum.+ (##fixnum.+ x-length digit-shift)
                      (if (##fixnum.zero? bit-shift) 0 1))))
    (if (##fixnum.< 0 result-length)

      (let ((result (##bignum.make result-length #f #f)))

        (##declare (not interrupts-enabled))

        (if (##fixnum.zero? bit-shift)
          (let ((smallest-i (##fixnum.max 0 digit-shift)))
            (let loop1 ((i (##fixnum.- result-length 1))
                        (j (##fixnum.- x-length 1)))
              (if (##fixnum.< i smallest-i)
                (##bignum.normalize! result)
                (begin
                  (##bignum.adigit-set! result i x j)
                  (loop1 (##fixnum.- i 1)
                         (##fixnum.- j 1))))))
          (let ((left-fill (if (##bignum.negative? x)
                             ##bignum.adigit-ones
                             ##bignum.adigit-zeros))
                (i (##fixnum.- result-length 1))
                (j (##fixnum.- x-length 1))
                (divider bit-shift)
                (smallest-i (##fixnum.max 0 (##fixnum.+ digit-shift 1))))
            (##bignum.adigit-cat! result i left-fill 0 x j divider)
            (let loop2 ((i (##fixnum.- i 1))
                        (j (##fixnum.- j 1)))
              (if (##fixnum.< i smallest-i)
                (begin
                  (if (##not (##fixnum.< i 0))
                    (##bignum.adigit-cat! result
                                          i
                                          x
                                          (##fixnum.+ j 1) 
                                          ##bignum.adigit-zeros
                                          0
                                          divider))
                  (##bignum.normalize! result))
                (begin
                  (##bignum.adigit-cat! result
                                        i
                                        x
                                        (##fixnum.+ j 1)
                                        x
                                        j
                                        divider)
                  (loop2 (##fixnum.- i 1)
                         (##fixnum.- j 1))))))))

      (if (##bignum.negative? x)
        -1
        0))))

; Bignum division.

(define (##bignum.div u v)

  ; u is an unnormalized bignum, v is a normalized exact-int
  ; 0 < v <= u
  
  (define (naive-div u v) 

    ; u is a normalized bignum, v is an unnormalized bignum
    ; v >= ##bignum.mdigit-base

    (let ((n
           (let loop1 ((i (##fixnum.- (##bignum.mdigit-length v) 1)))
             (if (##fixnum.< 0 (##bignum.mdigit-ref v i))
               (##fixnum.+ i 1)
               (loop1 (##fixnum.- i 1))))))
      (let ((normalizing-bit-shift
             (##fixnum.- ##bignum.mdigit-width
                         (##exact-int.width
                          (##bignum.mdigit-ref v (##fixnum.- n 1))))))
        (let ((u (##bignum.shift u normalizing-bit-shift)) 
              (v (##bignum.shift v normalizing-bit-shift)))
          (let ((q
                 (##bignum.make
                  (##fixnum.+ (##fixnum.- (##bignum.adigit-length u)
                                          (##bignum.adigit-length v))
                              2) ; 1 is not always sufficient...
                  #f
                  #f))
                (v_n-1
                 (##bignum.mdigit-ref v (##fixnum.- n 1)))
                (v_n-2 (##bignum.mdigit-ref v (##fixnum.- n 2))))
            (let ((m
                   (let loop2 ((i (##fixnum.- (##bignum.mdigit-length u) 1)))
                     (let ((u_i (##bignum.mdigit-ref u i)))
                       (if (##fixnum.< 0 u_i)
                         (if (##not (##fixnum.< u_i v_n-1))
                           (##fixnum.- (##fixnum.+ i 1) n)
                           (##fixnum.- i n))
                         (loop2 (##fixnum.- i 1)))))))
              (let loop3 ((j m))
                (if (##not (##fixnum.< j 0))
                  (let ((q-hat
                         (let ((q-hat
                                (##bignum.mdigit-quotient
                                 u
                                 (##fixnum.+ j n)
                                 v_n-1))
                               (u_n+j-2
                                (##bignum.mdigit-ref
                                 u
                                 (##fixnum.+ (##fixnum.- j 2) n)
                                 )))
                           (let ((r-hat
                                  (##bignum.mdigit-remainder
                                   u
                                   (##fixnum.+ j n)
                                   v_n-1
                                   q-hat)))
                             (if (or (##fixnum.= q-hat ##bignum.mdigit-base)
                                     (##bignum.mdigit-test?
                                      q-hat
                                      v_n-2
                                      r-hat
                                      u_n+j-2))
                               (let ((q-hat (##fixnum.- q-hat 1))
                                     (r-hat (##fixnum.+ r-hat v_n-1)))
                                 (if (and (##fixnum.< r-hat
                                                      ##bignum.mdigit-base)
                                          (##bignum.mdigit-test?
                                           q-hat
                                           v_n-2
                                           r-hat
                                           u_n+j-2))
                                   (##fixnum.- q-hat 1)
                                   q-hat))
                               q-hat)))))

                    (##declare (not interrupts-enabled))

                    (let loop4 ((i j)
                                (k 0)
                                (borrow 0))
                      (if (##fixnum.< k n)
                        (loop4 (##fixnum.+ i 2)
                               (##fixnum.+ k 2)
                               (##bignum.mdigit-div!
                                u
                                (##fixnum.+ i 1)
                                v
                                (##fixnum.+ k 1)
                                q-hat
                                (##bignum.mdigit-div!
                                 u
                                 i
                                 v
                                 k
                                 q-hat
                                 borrow)))
                        (let ((borrow
                               (if (##fixnum.< n k)
                                 borrow
                                 (##bignum.mdigit-div!
                                  u
                                  i
                                  v
                                  k
                                  q-hat
                                  borrow))))
                          (if (##fixnum.< borrow 0)
                            (let loop5 ((i j)
                                        (l 0)
                                        (carry 0))
                              (if (##fixnum.< n l)
                                (begin
                                  (##bignum.mdigit-set!
                                   q
                                   j
                                   (##fixnum.- q-hat 1))
                                  (loop3 (##fixnum.- j 1)))
                                (loop5 (##fixnum.+ i 1)
                                       (##fixnum.+ l 1)
                                       (##bignum.mdigit-mul!
                                        u
                                        i
                                        v
                                        l
                                        1
                                        carry))))
                            (begin
                              (##bignum.mdigit-set! q j q-hat)
                              (loop3 (##fixnum.- j 1))))))))
                  (##cons (##bignum.normalize! q)
                          (##exact-int.shift
                           (##bignum.normalize! u)
                           (##fixnum.- normalizing-bit-shift)))))))))))

  (define (div-one u v)
    (let ((m
           (let loop6 ((i (##fixnum.- (##bignum.mdigit-length u) 1)))
             (if (##fixnum.< 0 (##bignum.mdigit-ref u i))
               (##fixnum.+ i 1)
               (loop6 (##fixnum.- i 1))))))
      (let ((work-u (##bignum.make 1 #f #f))
            (q (##bignum.make (##bignum.adigit-length u) #f #f)))

        (##declare (not interrupts-enabled))

        (let loop7 ((i m)
                    (r-hat 0))
          (##bignum.mdigit-set!
           work-u
           1
           r-hat)
          (##bignum.mdigit-set!
           work-u
           0
           (##bignum.mdigit-ref u (##fixnum.- i 1)))
          (let ((q-hat (##bignum.mdigit-quotient work-u 1 v)))
            (let ((r-hat (##bignum.mdigit-remainder work-u 1 v q-hat)))
              (##bignum.mdigit-set! q (##fixnum.- i 1) q-hat)
              (if (##fixnum.< 1 i)
                (loop7 (##fixnum.- i 1)
                       r-hat)
                (##cons (##bignum.normalize! q)
                        r-hat))))))))
  
  (if (##fixnum? v)
    (if (##fixnum.< v ##bignum.mdigit-base)
      (div-one u v)
      (naive-div u (##bignum.<-fixnum v)))
    (naive-div u v)))

;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (##bitset-or x y)

  (define (bignum-bitset-or x x-length y y-length)

    (##declare (not interrupts-enabled))

    (if (##bignum.negative? x)
      (let ((result (##bignum.make x-length x #f)))
        (let loop1 ((i 0))
          (if (##fixnum.< i x-length)
            (begin
              (##bignum.adigit-logior! result i y i)
              (loop1 (##fixnum.+ i 1)))
            (##bignum.normalize! result))))
      (let ((result (##bignum.make y-length y #f)))
        (let loop2 ((i 0))
          (if (##fixnum.< i x-length)
            (begin
              (##bignum.adigit-logior! result i x i)
              (loop2 (##fixnum.+ i 1)))
            (##bignum.normalize! result))))))

  (cond ((##fixnum? x)
         (cond ((##fixnum? y)
                (##fixnum.logior x y))
               ((##bignum? y)
                (let* ((x-bignum (##bignum.<-fixnum x))
                       (x-length (##bignum.adigit-length x-bignum))
                       (y-length (##bignum.adigit-length y)))
                  (bignum-bitset-or x-bignum x-length y y-length)))
               (else
                '())))
        ((##bignum? x)
         (let ((x-length (##bignum.adigit-length x)))
           (cond ((##fixnum? y)
                  (let* ((y-bignum (##bignum.<-fixnum y))
                         (y-length (##bignum.adigit-length y-bignum)))
                    (bignum-bitset-or y-bignum y-length x x-length)))
                 ((##bignum? y)
                  (let ((y-length (##bignum.adigit-length y)))
                    (if (##fixnum.< x-length y-length)
                      (bignum-bitset-or x x-length y y-length)
                      (bignum-bitset-or y y-length x x-length))))
                 (else
                  '()))))
        (else
         '())))

(define-nary (bitset-or x y)
  0
  (if (macro-exact-int? x) x '())
  (##bitset-or x y)
  macro-force-vars
  macro-no-check
  (##null? ##trap-check-exact-int*))

(define (##bitset-xor x y)

  (define (bignum-bitset-xor x x-length y y-length)

    (##declare (not interrupts-enabled))

    (let ((result (##bignum.make y-length y #f)))
      (let loop1 ((i 0))
        (if (##fixnum.< i x-length)
          (begin
            (##bignum.adigit-logxor! result i x i)
            (loop1 (##fixnum.+ i 1)))
          (if (##bignum.negative? x)
            (let loop2 ((i i))
              (if (##fixnum.< i y-length)
                (begin
                  (##bignum.adigit-lognot! result i)
                  (loop2 (##fixnum.+ i 1)))
                (##bignum.normalize! result)))
            (##bignum.normalize! result))))))

  (cond ((##fixnum? x)
         (cond ((##fixnum? y)
                (##fixnum.logxor x y))
               ((##bignum? y)
                (let* ((x-bignum (##bignum.<-fixnum x))
                       (x-length (##bignum.adigit-length x-bignum))
                       (y-length (##bignum.adigit-length y)))
                  (bignum-bitset-xor x-bignum x-length y y-length)))
               (else
                '())))
        ((##bignum? x)
         (let ((x-length (##bignum.adigit-length x)))
           (cond ((##fixnum? y)
                  (let* ((y-bignum (##bignum.<-fixnum y))
                         (y-length (##bignum.adigit-length y-bignum)))
                    (bignum-bitset-xor y-bignum y-length x x-length)))
                 ((##bignum? y)
                  (let ((y-length (##bignum.adigit-length y)))
                    (if (##fixnum.< x-length y-length)
                      (bignum-bitset-xor x x-length y y-length)
                      (bignum-bitset-xor y y-length x x-length))))
                 (else
                  '()))))
        (else
         '())))

(define-nary (bitset-xor x y)
  0
  (if (macro-exact-int? x) x '())
  (##bitset-xor x y)
  macro-force-vars
  macro-no-check
  (##null? ##trap-check-exact-int*))

(define (##bitset-and x y)

  (define (bignum-bitset-and x x-length y y-length)

    (##declare (not interrupts-enabled))

    (if (##bignum.negative? x)
      (let ((result (##bignum.make y-length y #f)))
        (let loop1 ((i 0))
          (if (##fixnum.< i x-length)
            (begin
              (##bignum.adigit-logand! result i x i)
              (loop1 (##fixnum.+ i 1)))
            (##bignum.normalize! result))))
      (let ((result (##bignum.make x-length x #f)))
        (let loop2 ((i 0))
          (if (##fixnum.< i x-length)
            (begin
              (##bignum.adigit-logand! result i y i)
              (loop2 (##fixnum.+ i 1)))
            (##bignum.normalize! result))))))

  (cond ((##fixnum? x)
         (cond ((##fixnum? y)
                (##fixnum.logand x y))
               ((##bignum? y)
                (let* ((x-bignum (##bignum.<-fixnum x))
                       (x-length (##bignum.adigit-length x-bignum))
                       (y-length (##bignum.adigit-length y)))
                  (bignum-bitset-and x-bignum x-length y y-length)))
               (else
                '())))
        ((##bignum? x)
         (let ((x-length (##bignum.adigit-length x)))
           (cond ((##fixnum? y)
                  (let* ((y-bignum (##bignum.<-fixnum y))
                         (y-length (##bignum.adigit-length y-bignum)))
                    (bignum-bitset-and y-bignum y-length x x-length)))
                 ((##bignum? y)
                  (let ((y-length (##bignum.adigit-length y)))
                    (if (##fixnum.< x-length y-length)
                      (bignum-bitset-and x x-length y y-length)
                      (bignum-bitset-and y y-length x x-length))))
                 (else
                  '()))))
        (else
         '())))

(define-nary (bitset-and x y)
  -1
  (if (macro-exact-int? x) x '())
  (##bitset-and x y)
  macro-force-vars
  macro-no-check
  (##null? ##trap-check-exact-int*))

(define (##bitset-not x)

  (define (type-error)
    (##trap-check-exact-int 'bitset-not x))

  (cond ((##fixnum? x)
         (##fixnum.lognot x))
        ((##bignum? x)
         (##bignum.make (##bignum.adigit-length x) x #t))
        (else
         (type-error))))

(define (bitset-not x) (macro-force-vars (x) (##bitset-not x)))

(define (##bitset-shift x y)

  (define (type-error)
    (##trap-check-exact-int 'bitset-shift x y))

  (define (overflow)
    (##exception.heap-overflow)
    (##bitset-shift x y))

  (define (general-fixnum-fixnum-case)
    (##bignum.shift (##bignum.<-fixnum x) y))

  (cond ((##fixnum? x)
         (cond ((##fixnum? y)
                (if (##fixnum.< y 0)
                  (let ((abs-y (##fixnum.- y)))
                    (if (##fixnum.< abs-y ##fixnum-width)
                      (let ((result (##fixnum.ashr x abs-y)))
                        (if (##fixnum.= (##fixnum.shl result abs-y) x)
                          result
                          (general-fixnum-fixnum-case)))
                      (general-fixnum-fixnum-case)))
                  (if (##fixnum.< y ##fixnum-width)
                    (let ((result (##fixnum.shl x y)))
                      (if (##fixnum.= (##fixnum.ashr result y) x)
                        result
                        (general-fixnum-fixnum-case)))
                    (general-fixnum-fixnum-case))))
               ((##bignum? y)
                (cond ((##fixnum.= x 0)
                       0)
                      ((##bignum.negative? y)
                       0)
                      (else
                       (overflow))))
               (else
                (type-error))))
        ((##bignum? x)
         (cond ((##fixnum? y)
                (##bignum.shift x y))
               ((##bignum? y)
                (cond ((##bignum.negative? y)
                       0)
                      (else
                       (overflow))))
               (else
                (type-error))))
        (else
         (type-error))))

(define (bitset-shift x y) (macro-force-vars (x y) (##bitset-shift x y)))

;------------------------------------------------------------------------------

; Exact integer operations
; ------------------------

(define (##exact-int.shift x shift)
  (if (##fixnum? x)
    (##bignum.shift (##bignum.<-fixnum x) shift)
    (##bignum.shift x shift)))

(define (##exact-int.*-expt2 x y)
  (if (##fixnum.negative? y)
    (##ratnum.normalize x (##exact-int.shift 1 (##fixnum.- y)))
    (##exact-int.shift x y)))

(define (##exact-int.div x y)
  (cond ((##eq? y 1)
         (##cons x 0))
        ((##eq? y -1)
         (##cons (##negate x) 0))
        (else
         (if (##fixnum? x)
           (if (##fixnum? y)
             (##cons (##fixnum.quotient x y) ; note: y can't be -1
                     (##fixnum.remainder x y))
             (##cons 0 x))
           (let* ((x-negative?
                   (##bignum.negative? x))
                  (abs-x
                   (if x-negative? (##negate x) x))
                  (y-negative?
                   (##negative? y))
                  (abs-y
                   (if y-negative? (##negate y) y)))
             (if (##< abs-x abs-y)
               (##cons 0 x)
               (let ((result
                      (##bignum.div abs-x abs-y)))

                 (if (##not (##eq? x-negative? y-negative?))
                   (##set-car! result (##negate (##car result))))

                 (if x-negative?
                   (##set-cdr! result (##negate (##cdr result))))

                 result)))))))

(define (##exact-int.square n)
  (##* n n))

(define (##exact-int.width x)

  (define (width x n)
    (if (##fixnum.< x 65536)
      (at-most-16 x n)
      (width (##fixnum.ashr x 16) (##fixnum.+ n 16))))

  (define (at-most-16 x n)
    (if (##fixnum.< x 256)
      (at-most-8 x n)
      (at-most-8 (##fixnum.ashr x 8) (##fixnum.+ n 8))))

  (define (at-most-8 x n)
    (if (##fixnum.< x 16)
      (at-most-4 x n)
      (at-most-4 (##fixnum.ashr x 4) (##fixnum.+ n 4))))

  (define (at-most-4 x n)
    (##fixnum.+ (##u8vector-ref '#u8(0 1 2 2 3 3 3 3 4 4 4 4 4 4 4 4) x) n))

  (if (##fixnum? x)

    (if (##fixnum.< x 0)
      (width (##fixnum.- -1 x) 0)
      (width x 0))

    (let ((n (##bignum.mdigit-length x)))
      (if (##bignum.negative? x)
        (let loop1 ((i (##fixnum.- n 1)))
          (if (##fixnum.< i 0)
            0
            (let ((mdigit (##bignum.mdigit-ref x i)))
              (if (##fixnum.= mdigit ##bignum.mdigit-base-minus-1)
                (loop1 (##fixnum.- i 1))
                (width (##fixnum.- ##bignum.mdigit-base-minus-1 mdigit)
                       (##fixnum.* i ##bignum.mdigit-width))))))
        (let loop2 ((i (##fixnum.- n 1)))
          (if (##fixnum.< i 0)
            0
            (let ((mdigit (##bignum.mdigit-ref x i)))
              (if (##fixnum.= mdigit 0)
                (loop2 (##fixnum.- i 1))
                (width mdigit
                       (##fixnum.* i ##bignum.mdigit-width))))))))))

(define (##exact-int.root x y)
  (if (##eq? x 0)
    0
    (let loop ((g (##exact-int.shift
                   1
                   (##fixnum.quotient
                    (##fixnum.+ (##exact-int.width x) (##fixnum.- y 1))
                    y))))
      (let* ((a (##expt g (##fixnum.- y 1)))
             (b (##* a y))
             (c (##* a (##fixnum.- y 1)))
             (d (##quotient (##+ x (##* g c)) b)))
        (if (##< d g)
          (loop d)
          g)))))

;------------------------------------------------------------------------------

; Ratnum operations
; -----------------

(define (##ratnum.= x y)
  (and (##= (macro-ratnum-numerator x)
            (macro-ratnum-numerator y))
       (##= (macro-ratnum-denominator x)
            (macro-ratnum-denominator y))))

(define (##ratnum.< x y)
  (##< (##* (macro-ratnum-numerator x)
            (macro-ratnum-denominator y))
       (##* (macro-ratnum-denominator x)
            (macro-ratnum-numerator y))))

(define (##ratnum.+ x y)
  (##ratnum.normalize
   (##+ (##* (macro-ratnum-numerator x)
             (macro-ratnum-denominator y))
        (##* (macro-ratnum-denominator x)
             (macro-ratnum-numerator y)))
   (##* (macro-ratnum-denominator x)
        (macro-ratnum-denominator y))))

(define (##ratnum.* x y)
  (##ratnum.normalize
   (##* (macro-ratnum-numerator x)
        (macro-ratnum-numerator y))
   (##* (macro-ratnum-denominator x)
        (macro-ratnum-denominator y))))

(define (##ratnum.- x y)
  (##ratnum.normalize
   (##- (##* (macro-ratnum-numerator x)
             (macro-ratnum-denominator y))
        (##* (macro-ratnum-denominator x)
             (macro-ratnum-numerator y)))
   (##* (macro-ratnum-denominator x)
        (macro-ratnum-denominator y))))

(define (##ratnum./ x y)
  (##ratnum.normalize
   (##* (macro-ratnum-numerator x)
        (macro-ratnum-denominator y))
   (##* (macro-ratnum-denominator x)
        (macro-ratnum-numerator y))))

(define (##ratnum.normalize num den)
  (let* ((x (##gcd num den))
         (y (if (##negative? den) (##negate x) x))
         (num (##quotient num y))
         (den (##quotient den y)))
    (if (##eq? den 1)
      num
      (macro-ratnum-make num den))))

(define (##ratnum.<-exact-int x)
  (macro-ratnum-make x 1))

;------------------------------------------------------------------------------

; Flonum operations
; -----------------

(define-system (##flonum.->fixnum x))
(define-system (##flonum.<-fixnum x))

(define-nary (##flonum.max x y)
  ()
  x
  (##flonum.max x y)
  macro-no-force
  macro-no-check)

(define-nary (##flonum.min x y)
  ()
  x
  (##flonum.min x y)
  macro-no-force
  macro-no-check)

(define-nary (##flonum.+ x y)
  (macro-inexact-+0)
  x
  (##flonum.+ x y)
  macro-no-force
  macro-no-check)

(define-nary (##flonum.* x y)
  (macro-inexact-+1)
  x
  (##flonum.* x y)
  macro-no-force
  macro-no-check)

(define-nary (##flonum.- x y)
  ()
  (##flonum.- x)
  (##flonum.- x y)
  macro-no-force
  macro-no-check)

(define-nary (##flonum./ x y)
  ()
  (##flonum./ (macro-inexact-+1) x)
  (##flonum./ x y)
  macro-no-force
  macro-no-check)

(define-system (##flonum.abs x))

(define-system (##flonum.floor x))
(define-system (##flonum.ceiling x))

(define-system (##flonum.truncate x)
  (let ((y (##flonum.floor x)))
    (if (and (##flonum.negative? x) (##not (##flonum.= x y)))
      (##flonum.+ y (macro-inexact-+1))
      y)))

(define-system (##flonum.round x))

(define-system (##flonum.exp x))
(define-system (##flonum.log x))
(define-system (##flonum.sin x))
(define-system (##flonum.cos x))
(define-system (##flonum.tan x))
(define-system (##flonum.asin x))
(define-system (##flonum.acos x))
(define (##flonum.atan x #!optional (y (macro-absent-obj)))
  (if (##eq? y (macro-absent-obj))
    (##flonum.atan x)
    (##flonum.atan x y)))
(define-system (##flonum.sqrt x))
(define-system (##flonum.copysign x y))

(define-system (##flonum.zero? x)
  (##flonum.= x (macro-inexact-+0)))

(define-system (##flonum.positive? x)
  (##flonum.< (macro-inexact-+0) x))

(define-system (##flonum.negative? x)
  (##flonum.< x (macro-inexact-+0)))

(define-system (##flonum.finite? x)
  (##not (or (##flonum.nan? x)
             (##flonum.= x +inf.)
             (##flonum.= x -inf.))))

(define-system (##flonum.nan? x))

(define-nary-bool (##flonum.= x y)
  #t
  #t
  (##flonum.= x y)
  macro-no-force
  macro-no-check)

(define-nary-bool (##flonum.< x y)
  #t
  #t
  (##flonum.< x y)
  macro-no-force
  macro-no-check)

(define-nary-bool (##flonum.> x y)
  #t
  #t
  (##flonum.> x y)
  macro-no-force
  macro-no-check)

(define-nary-bool (##flonum.<= x y)
  #t
  #t
  (##flonum.<= x y)
  macro-no-force
  macro-no-check)

(define-nary-bool (##flonum.>= x y)
  #t
  #t
  (##flonum.>= x y)
  macro-no-force
  macro-no-check)

(define (##flonum.<-ratnum x)
  (let* ((num (macro-ratnum-numerator x))
         (n (##abs num))
         (d (macro-ratnum-denominator x))
         (wn (##exact-int.width n)) ; 2^(wn-1) <= n < 2^wn
         (wd (##exact-int.width d)) ; 2^(wd-1) <= d < 2^wd
         (p (##fixnum.- wn wd)))

    (define (f1 sn sd)
      (if (##< sn sd) ; n/(d*2^p) < 1 ?
        (f2 (##exact-int.shift sn 1) sd (##fixnum.- p 1))
        (f2 sn sd p)))

    (define (f2 a b p)
      ; 1 <= a/b < 2  and  n/d = (2^p*a)/b  and  n/d < 2^(p+1)
      (let* ((shift
              (##fixnum.min (macro-flonum-m-bits)
                            (##fixnum.- p (macro-flonum-e-min))))
             (abs-result
              (##flonum.*
               (##flonum.<-exact-int
                (##round
                 (##ratnum.normalize (##exact-int.shift a shift) b)))
               (##flonum.expt2 (##fixnum.- p shift)))))
        (if (##negative? num)
          (##flonum.copysign abs-result (macro-inexact--1))
          abs-result)))

    ; 2^(p-1) <= n/d < 2^(p+1)
    ; 1/2 <= n/(d*2^p) < 2  or equivalently  1/2 <= (n*2^-p)/d < 2

    (if (##fixnum.negative? p)
      (f1 (##exact-int.shift n (##fixnum.- p)) d)
      (f1 n (##exact-int.shift d p)))))

(define (##flonum.<-exact-int x)

  (define (f1 x)
    (let* ((w ; 2^(w-1) <= x < 2^w
            (##exact-int.width x))
           (p ; 2^52 <= x/2^p < 2^53
            (##fixnum.- w (macro-flonum-m-bits-plus-1))))
      (if (##fixnum.< p 1)
        (f2 x)
        (let* ((2^p (##exact-int.shift 1 p))
               (qr (##exact-int.div x 2^p))
               (q (##car qr))
               (r (##cdr qr))
               (r*2 (##exact-int.shift r 1)))
          (##flonum.*
           (##flonum.expt2 p)
           (f2 (if (or (and (##not (##odd? q))
                            (##= r*2 2^p))
                       (##< r*2 2^p))
                 q
                 (##+ q 1))))))))

  (define (f2 x) ; 0 <= x < 2^53
    (if (##fixnum? x)
      (##flonum.<-fixnum x)
      (let* ((x (if (##fixnum? x) (##bignum.<-fixnum x) x))
             (n (##bignum.mdigit-length x)))
        (let loop ((i (##fixnum.- n 1))
                   (result (macro-inexact-+0)))
          (if (##fixnum.< i 0)
            result
            (let ((mdigit (##bignum.mdigit-ref x i)))
              (loop (##fixnum.- i 1)
                    (##flonum.+ (##flonum.* result
                                            ##bignum.inexact-mdigit-base)
                                (##flonum.<-fixnum mdigit)))))))))

  (if (##fixnum? x)
    (##flonum.<-fixnum x)
    (if (##negative? x)
      (##flonum.copysign (f1 (##negate x)) (macro-inexact--1))
      (f1 x))))

(define (##flonum.expt2 n)
  (cond ((##fixnum.zero? n)
         (macro-inexact-+1))
        ((##fixnum.negative? n)
         (##expt (macro-inexact-+1/2) (##fixnum.- n)))
        (else
         (##expt (macro-inexact-+2) n))))

(define (##flonum.->exact-int x)
  (let loop1 ((z (##flonum.abs x))
              (n 1))
    (if (##flonum.< ##bignum.inexact-mdigit-base z)
      (loop1 (##flonum./ z ##bignum.inexact-mdigit-base)
             (##fixnum.+ n 1))
      (let loop2 ((result 0)
                  (z z)
                  (i n))
        (if (##fixnum.< 0 i)
          (let* ((inexact-floor-z
                  (##flonum.floor z))
                 (floor-z
                  (##flonum.->fixnum inexact-floor-z))
                 (new-z
                  (##flonum.* (##flonum.- z inexact-floor-z)
                              ##bignum.inexact-mdigit-base)))
            (loop2 (##+ floor-z
                        (##exact-int.shift result
                                           ##bignum.mdigit-width))
                   new-z
                   (##fixnum.- i 1)))
          (if (##flonum.negative? x)
            (##negate result)
            result))))))

(define (##flonum.->inexact-exponential-format x)

  (define (exp-form-pos x y i)
    (let ((i*2 (##fixnum.+ i i)))
      (let ((z (if (and (##not (##fixnum.< (macro-flonum-e-bias) i*2))
                        (##not (##flonum.< x y)))
                 (exp-form-pos x (##flonum.* y y) i*2)
                 (##vector x 0 1))))
        (let ((a (##vector-ref z 0)) (b (##vector-ref z 1)))
          (let ((i+b (##fixnum.+ i b)))
            (if (and (##not (##fixnum.< (macro-flonum-e-bias) i+b))
                     (##not (##flonum.< a y)))
              (begin
                (##vector-set! z 0 (##flonum./ a y))
                (##vector-set! z 1 i+b)))
            z)))))

  (define (exp-form-neg x y i)
    (let ((i*2 (##fixnum.+ i i)))
      (let ((z (if (and (##fixnum.< i*2 (macro-flonum-e-bias-minus-1))
                        (##flonum.< x y))
                 (exp-form-neg x (##flonum.* y y) i*2)
                 (##vector x 0 1))))
        (let ((a (##vector-ref z 0)) (b (##vector-ref z 1)))
          (let ((i+b (##fixnum.+ i b)))
            (if (and (##fixnum.< i+b (macro-flonum-e-bias-minus-1))
                     (##flonum.< a y))
              (begin
                (##vector-set! z 0 (##flonum./ a y))
                (##vector-set! z 1 i+b)))
            z)))))

  (define (exp-form x)
    (if (##flonum.< x (macro-inexact-+1))
      (let ((z (exp-form-neg x (macro-inexact-+1/2) 1)))
        (##vector-set! z 0 (##flonum.* (macro-inexact-+2) (##vector-ref z 0)))
        (##vector-set! z 1 (##fixnum.- -1 (##vector-ref z 1)))
        z)
      (exp-form-pos x (macro-inexact-+2) 1)))

  (if (##flonum.negative? (##flonum.copysign (macro-inexact-+1) x))
    (let ((z (exp-form (##flonum.copysign x (macro-inexact-+1)))))
      (##vector-set! z 2 -1)
      z)
    (exp-form x)))

(define (##flonum.->exact-exponential-format x)
  (let ((z (##flonum.->inexact-exponential-format x)))
    (let ((y (##vector-ref z 0)))
      (if (##not (##flonum.< y (macro-inexact-+2))) ; +inf. or +nan.?
        (begin
          (if (##flonum.< (macro-inexact-+0) y)
            (##vector-set! z 0 (macro-flonum-+m-min))  ; +inf.
            (##vector-set! z 0 (macro-flonum-+m-max))) ; +nan.
          (##vector-set! z 1 (macro-flonum-e-bias-plus-1)))
        (##vector-set! z 0
          (##flonum.->exact-int
           (##flonum.* (##vector-ref z 0) (macro-flonum-m-min)))))
      (##vector-set! z 1 (##fixnum.- (##vector-ref z 1) (macro-flonum-m-bits)))
      z)))

(define (##flonum.inexact->exact x)
  (let ((y (##flonum.->exact-exponential-format x)))
    (##exact-int.*-expt2
     (if (##fixnum.negative? (##vector-ref y 2))
       (##negate (##vector-ref y 0))
       (##vector-ref y 0))
     (##vector-ref y 1))))

(define (##flonum.->ratnum x)
  (let ((y (##flonum.inexact->exact x)))
    (if (macro-exact-int? y)
      (##ratnum.<-exact-int y)
      y)))

;------------------------------------------------------------------------------

; Cpxnum operations
; -----------------

(define (##cpxnum.= x y)
  (and (##= (macro-cpxnum-real x) (macro-cpxnum-real y))
       (##= (macro-cpxnum-imag x) (macro-cpxnum-imag y))))

(define (##cpxnum.+ x y)
  (let ((a (macro-cpxnum-real x)) (b (macro-cpxnum-imag x))
        (c (macro-cpxnum-real y)) (d (macro-cpxnum-imag y)))
    (##make-rectangular (##+ a c) (##+ b d))))

(define (##cpxnum.* x y)
  (let ((a (macro-cpxnum-real x)) (b (macro-cpxnum-imag x))
        (c (macro-cpxnum-real y)) (d (macro-cpxnum-imag y)))
    (##make-rectangular (##- (##* a c) (##* b d)) (##+ (##* a d) (##* b c)))))

(define (##cpxnum.- x y)
  (let ((a (macro-cpxnum-real x)) (b (macro-cpxnum-imag x))
        (c (macro-cpxnum-real y)) (d (macro-cpxnum-imag y)))
    (##make-rectangular (##- a c) (##- b d))))

(define (##cpxnum./ x y)

  (define (basic/ a b c d q)
    (##make-rectangular (##/ (##+ (##* a c) (##* b d)) q)
                        (##/ (##- (##* b c) (##* a d)) q)))

  (let ((a (macro-cpxnum-real x)) (b (macro-cpxnum-imag x))
        (c (macro-cpxnum-real y)) (d (macro-cpxnum-imag y)))
    (cond ((##eq? d 0)
           ; A normalized cpxnum can't have an imaginary part that is
           ; exact 0 but it is possible that ##cpxnum./ receives a
           ; nonnormalized cpxnum as x or y when it is called from ##/.
           (##make-rectangular (##/ a c)
                               (##/ b c)))
          ((##eq? c 0)
           (##make-rectangular (##/ b d)
                               (##negate (##/ a d))))
          ((and (##exact? c) (##exact? d)) 
           (basic/ a b c d (##+ (##* c c) (##* d d))))
          (else
           ; just coerce everything to inexact and move on
           (let ((inexact-c (##exact->inexact c))
                 (inexact-d (##exact->inexact d)))
             (if (and (##flonum.finite? inexact-c)
                      (##flonum.finite? inexact-d))
               (let ((q
                      (##flonum.+ (##flonum.* inexact-c inexact-c)
                                  (##flonum.* inexact-d inexact-d))))
                 (cond ((##not (##flonum.finite? q))
                        (let ((a
                               (if (##flonum? a)
                                 (##flonum.* a (macro-inexact-scale-down))
                                 (##* a (macro-scale-down))))
                              (b
                               (if (##flonum? b)
                                 (##flonum.* b (macro-inexact-scale-down))
                                 (##* b (macro-scale-down))))
                              (inexact-c
                               (##flonum.* inexact-c
                                           (macro-inexact-scale-down)))
                              (inexact-d
                               (##flonum.* inexact-d
                                           (macro-inexact-scale-down))))
                          (basic/ a
                                  b
                                  inexact-c
                                  inexact-d
                                  (##flonum.+
                                   (##flonum.* inexact-c inexact-c) 
                                   (##flonum.* inexact-d inexact-d)))))
                       ((##flonum.< q (macro-flonum-min-normal))
                        (let ((a
                               (if (##flonum? a)
                                 (##flonum.* a (macro-inexact-scale-up))
                                 (##* a (macro-scale-up))))
                              (b
                               (if (##flonum? b)
                                 (##flonum.* b (macro-inexact-scale-up))
                                 (##* b (macro-scale-up))))
                              (inexact-c
                               (##flonum.* inexact-c
                                           (macro-inexact-scale-up)))
                              (inexact-d
                               (##flonum.* inexact-d
                                           (macro-inexact-scale-up))))
                          (basic/ a
                                  b
                                  inexact-c
                                  inexact-d
                                  (##flonum.+
                                   (##flonum.* inexact-c inexact-c) 
                                   (##flonum.* inexact-d inexact-d)))))
                       (else
                        (basic/ a b inexact-c inexact-d q))))
               (cond ((##flonum.= inexact-c (macro-inexact-+inf))
                      (basic/ a
                              b
                              (macro-inexact-+0)
                              (if (##flonum.nan? inexact-d)
                                inexact-d
                                (##flonum.copysign (macro-inexact-+0)
                                                   inexact-d))
                              (macro-inexact-+1)))
                     ((##flonum.= inexact-c (macro-inexact--inf))
                      (basic/ a
                              b
                              (macro-inexact--0)
                              (if (##flonum.nan? inexact-d)
                                inexact-d
                                (##flonum.copysign (macro-inexact-+0)
                                                   inexact-d))
                              (macro-inexact-+1)))
                     ((##flonum.nan? inexact-c)
                      (cond ((##flonum.= inexact-d (macro-inexact-+inf))
                             (basic/ a
                                     b
                                     inexact-c
                                     (macro-inexact-+0)
                                     (macro-inexact-+1)))
                            ((##flonum.= inexact-d (macro-inexact--inf))
                             (basic/ a
                                     b
                                     inexact-c
                                     (macro-inexact--0)
                                     (macro-inexact-+1)))
                            ((##flonum.nan? inexact-d)
                             (basic/ a
                                     b
                                     inexact-c
                                     inexact-d
                                     (macro-inexact-+1)))
                            (else
                             (basic/ a
                                     b
                                     inexact-c
                                     (macro-inexact-+nan)
                                     (macro-inexact-+1)))))
                     (else
                      ; finite inexact-c
                      (cond ((##flonum.nan? inexact-d)
                             (basic/ a
                                     b
                                     (macro-inexact-+nan)
                                     inexact-d
                                     (macro-inexact-+1)))
                            (else
                             ; inexact-d is +inf. or -inf.
                             (basic/ a
                                     b
                                     (##flonum.copysign (macro-inexact-+0)
                                                        inexact-c)
                                     (##flonum.copysign (macro-inexact-+0)
                                                        inexact-d)
                                     (macro-inexact-+1))))))))))))

(define (##cpxnum.<-noncpxnum x)
  (macro-cpxnum-make x 0))

;==============================================================================
