/* file: "gambit.h" */

/*
 * Copyright (C) 1994-2000 by Marc Feeley, All Rights Reserved.
 */

#ifndef ___GAMBIT_H
#define ___GAMBIT_H

#ifndef ___VERSION_MAJOR
@error "___VERSION_MAJOR has not been set by includer"
#endif

#if ___VERSION_MAJOR != 21
@error "Incompatible version of gambit.h include file"
#endif

#define ___VERSION_MINOR 1

/* define one of ___DEBUG, ___PROFILE, ___SUPERSPARC, ___STANDARD */

#ifndef ___DEBUG
#undef ___DEBUG
#endif

#ifdef ___PROFILE
#undef ___PROFILE
#endif

#ifdef ___SUPERSPARC
#undef ___SUPERSPARC
#endif

#ifdef ___STANDARD
#undef ___STANDARD
#endif

#define ___STANDARD

/*---------------------------------------------------------------------------*/

/*
 * DETERMINE PROGRAMMING ENVIRONMENT'S CHARACTERISTICS
 *
 * The following symbols need to be defined to reflect the
 * programming environment's characteristics:
 *
 *  define one of the following symbols depending on the processor type
 *    ___CPU_alpha     DEC Alpha AXP processor
 *    ___CPU_hppa      HP PA-RISC processor
 *    ___CPU_m68k      Motorola 68k family processor
 *    ___CPU_mips      MIPS Rx000 family processor
 *    ___CPU_ppc       PowerPC processor
 *    ___CPU_rs6k      IBM RS6000 processor
 *    ___CPU_sony_news Christian Queinnec's processor!
 *    ___CPU_sparc     SUN SPARC processor
 *    ___CPU_x86       Intel x86 family processor
 *    ___CPU_cray_PVP  Cray vector processor (Cray T90, C90, J90)
 *
 *  define ___CPU_BIGEND if data is stored in "big-endian" format
 *  define ___CPU_LITEND if data is stored in "little-endian" format
 *
 *  define ___C_TYPES_USUAL_32BITS if the size in bits of the various C types
 *            is char=8 short=16 int=32 long=32 float=32 double=64 void*=32
 *  define ___C_TYPES_USUAL_64BITS if the size in bits of the various C types
 *            is char=8 short=16 int=32 long=64 float=32 double=64 void*=64
 *  define ___C_TYPES_CRAY_PVP if the size in bits of the various C types
 *            is char=8 short=64 int=64 long=64 float=64 double=64 void*=64
 *
 *  define ___BUILTIN_64BIT_INT_TYPE to the 64 bit signed integer type if
 *            one is available
 *
 *  define ___SCHAR to "char" if the C compiler does not support
 *            signed characters (e.g. SunOS 4.1 cc)
 *
 *  define ___IMPORTED_ID_PREFIX to the prefix that the C compiler puts on
 *            imported functions (for example "_" if an underscore is needed)
 *  define ___IMPORTED_ID_SUFFIX to the suffix that the C compiler puts on
 *            imported functions
 *
 *  define ___o as the extension for object files (e.g. "o", "obj")
 *  define ___CC_O as the command that compiles a C file to an object file
 *  define ___CC_O_SH like ___CC_O except that object file is for a shared lib
 *  define ___CLIBS as the libraries to use (e.g. "-lm", "-ldl", "-lncurses")
 *  define ___CC_CMD as the command that invokes the C compiler for a
 *            dynamic Scheme compilation
 *  define ___LD_CMD as the command that invokes the linker for a
 *            dynamic Scheme compilation
 *  define ___LD_FLIP to flip the two arguments to the linker
 *  define ___CANT_IMPORT_CLIB_DYNAMICALLY if the dynamic loader can not load
 *            files that import C library functions such as "sqrt"
 *  define ___CANT_IMPORT_SETJMP_DYNAMICALLY if the dynamic loader can not load
 *            files that use "setjmp"
 *  define ___CANT_IMPORT_DYNAMICALLY if the dynamic loader can not load
 *            files that import functions defined in other files
 *  define ___CANT_IMPORT_EXPORTED if exported data/functions in an
 *            executable/library can't be imported in that executable/library
 *  define ___IMPORTED_ADR_NOT_CONST if address-of operator (&) applied to
 *            imported data/functions can't be used in a constant initializer
 *  define ___SETDLPATH as the variable definition that sets up the path
 *            to search for shared libraries in "../lib"
 *  define ___USE_ISNAN if the test "x!=x" does not return true when x
 *            is a NaN
 */

#define ___FORCE_32zzzzzzzz   /* remove this line to use 64 bit words if available */

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/* SPARC running SUNOS. */

#ifdef sparc
#ifndef __linux__

#define ___CPU_sparc
#define ___CPU_BIGEND
#define ___C_TYPES_USUAL_32BITS

#ifdef __SVR4
#define ___o "o"
#ifdef __GNUC__

#ifdef ___PROFILE
#define ___CC_O "gcc -fprofile-arcs -ftest-coverage -c \$(CC_O_ARGS) -D___SINGLE_HOST"
#endif

#ifdef ___DEBUG
#define ___CC_O "gcc -g  -c \$(CC_O_ARGS) -D___SINGLE_HOST"
#define ___CC_O_SH "gcc -g -fPIC  -c \$(CC_O_SH_ARGS) -D___SINGLE_HOST"
#endif

#ifdef ___SUPERSPARC
#define ___CC_O "gcc -O1 -mcpu=supersparc -c \$(CC_O_ARGS) -D___SINGLE_HOST"
#define ___CC_O_SH "gcc -fPIC -O1 -mcpu=supersparc -c \$(CC_O_SH_ARGS) -D___SINGLE_HOST"
#endif

#ifdef ___ULTRASPARC
#define ___CC_O "gcc -m32 -O1 -mcpu=v9 -Wa,-xarch=v8plus -mtune=ultrasparc -fcall-used-g4 -fcall-used-g5 -fcall-used-g6 -fcall-used-g7 -c \$(CC_O_ARGS) -D___SINGLE_HOST"
#define ___CC_O_SH "gcc -m32 -fPIC -O1 -mcpu=v9 -Wa,-xarch=v8plus -mtune=ultrasparc -fcall-used-g4 -fcall-used-g5 -fcall-used-g6 -fcall-used-g7 -c \$(CC_O_SH_ARGS) -D___SINGLE_HOST"
#endif

#ifdef ___STANDARD
#define ___CC_O "gcc -O1 -c \$(CC_O_ARGS) -D___SINGLE_HOST"
#define ___CC_O_SH "gcc -fPIC -O1 -c \$(CC_O_SH_ARGS) -D___SINGLE_HOST"
#endif

#ifdef ___PROFILE
#define ___LINK "gcc -fprofile-arcs -ftest-coverage \\\$\$OBJECTS ../lib/\$(GAMBCLIB) \$(CLIBS) -o \\\$\$EXECUTABLE"
#define ___LINK_SH "gcc -fprofile-arcs -ftest-coverage \\\$\$OBJECTS ../lib/\$(GAMBCLIB_SH) \$(CLIBS) -o \\\$\$EXECUTABLE"
#define ___LINK_SH_INSTALL "gcc -fprofile-arcs -ftest-coverage \\\$\$OBJECTS -L\$(libdir) -lgambc \$(CLIBS) -o \\\$\$EXECUTABLE"
#define ___MAKE_GAMBCLIB_SH "gcc -fprofile-arcs -ftest-coverage -shared -o \$(GAMBCLIB_SH) \\\$\$OBJECTS \$(CLIBS)"
#endif

#ifdef ___DEBUG
#define ___LINK "gcc -g \\\$\$OBJECTS ../lib/\$(GAMBCLIB) \$(CLIBS) -o \\\$\$EXECUTABLE"
#define ___LINK_SH "gcc -g \\\$\$OBJECTS ../lib/\$(GAMBCLIB_SH) \$(CLIBS) -o \\\$\$EXECUTABLE"
#define ___LINK_SH_INSTALL "gcc -g \\\$\$OBJECTS -L\$(libdir) -lgambc \$(CLIBS) -o \\\$\$EXECUTABLE"
#define ___MAKE_GAMBCLIB_SH "gcc -g -shared -o \$(GAMBCLIB_SH) \\\$\$OBJECTS \$(CLIBS)"
#define ___CC_CMD "gcc -g -fPIC -c -D___DYNAMIC -D___SINGLE_HOST -o %s.o %s.c"
#endif

#if defined(___SUPERSPARC) || defined(___STANDARD)
#define ___LINK "gcc -O1 \\\$\$OBJECTS ../lib/\$(GAMBCLIB) \$(CLIBS) -o \\\$\$EXECUTABLE"
#define ___LINK_SH "gcc \\\$\$OBJECTS ../lib/\$(GAMBCLIB_SH) \$(CLIBS) -o \\\$\$EXECUTABLE"
#define ___LINK_SH_INSTALL "gcc \\\$\$OBJECTS -L\$(libdir) -lgambc \$(CLIBS) -o \\\$\$EXECUTABLE"
#define ___MAKE_GAMBCLIB_SH "gcc -shared -o \$(GAMBCLIB_SH) \\\$\$OBJECTS \$(CLIBS)"
#endif

#ifdef ___SUPERSPARC
#define ___CC_CMD "gcc -fPIC -O1 -mcpu=supersparc -c -D___DYNAMIC -D___SINGLE_HOST -o %s.o %s.c"
#endif


/*
#define ___CC_CMD "gcc -m32 -fPIC -O1 -mcpu=v9 -Wa,-xarch=v8plus -mtune=ultrasparc -fcall-used-g4 -fcall-used-g5 -fcall-used-g6 -fcall-used-g7 -c -D___DYNAMIC -D___SINGLE_HOST -o %s.o %s.c"
*/

#ifdef ___STANDARD
#define ___CC_CMD "gcc -fPIC -O1 -c -D___DYNAMIC -D___SINGLE_HOST -o %s.o %s.c"
#endif

#else
#define ___CC_O "cc -xO1 -c \$(CC_O_ARGS) -D___SINGLE_HOST"
#define ___CC_O_SH "cc -Kpic -xO1 -c \$(CC_O_SH_ARGS) -D___SINGLE_HOST"
#define ___LINK "cc \\\$\$OBJECTS ../lib/\$(GAMBCLIB) \$(CLIBS) -o \\\$\$EXECUTABLE"
#define ___LINK_SH "cc \\\$\$OBJECTS ../lib/\$(GAMBCLIB_SH) \$(CLIBS) -o \\\$\$EXECUTABLE"
#define ___LINK_SH_INSTALL "cc \\\$\$OBJECTS -L\$(libdir) -lgambc \$(CLIBS) -o \\\$\$EXECUTABLE"
#define ___MAKE_GAMBCLIB_SH "cc -G -o \$(GAMBCLIB_SH) \\\$\$OBJECTS \$(CLIBS)"
#define ___CC_CMD "cc -Kpic -xO1 -c -D___DYNAMIC -D___SINGLE_HOST -o %s.o %s.c"
#endif
#define ___CLIBS "-lm -ldl -lcurses -lnsl -lsocket"
#define ___GAMBCLIB "libgambc.a"
#define ___GAMBCLIB_SH "libgambc.so.1.1"
#define ___MAKE_GAMBCLIB "ar rc \$(GAMBCLIB) \\\$\$OBJECTS;ranlib \$(GAMBCLIB)"
#define ___LD_CMD "/usr/ccs/bin/ld -G %s.o -o %s"
#define ___SETDLPATH "LD_LIBRARY_PATH=../lib"

#else

#define ___o "o"
#ifdef __GNUC__
#define ___CC_O "gcc -O1 -c \$(CC_O_ARGS) -D___SINGLE_HOST"
#define ___LINK "gcc \\\$\$OBJECTS ../lib/\$(GAMBCLIB) \$(CLIBS) -o \\\$\$EXECUTABLE"
#define ___CC_CMD "gcc -fPIC -O1 -c -D___DYNAMIC -D___SINGLE_HOST -o %s.o %s.c"
#else
#define ___CC_O "cc -O1 -c \$(CC_O_ARGS) -D___SINGLE_HOST"
#define ___LINK "cc \\\$\$OBJECTS ../lib/\$(GAMBCLIB) \$(CLIBS) -o \\\$\$EXECUTABLE"
#define ___CC_CMD "cc -pic -O1 -c -D___DYNAMIC -D___SINGLE_HOST -o %s.o %s.c"
#define ___SCHAR char
#endif
#define ___LINK_SH "echo"
#define ___LINK_SH_INSTALL "echo"
#define ___CLIBS "-lm -lcurses"
#define ___GAMBCLIB "libgambc.a"
#define ___MAKE_GAMBCLIB "ar rc \$(GAMBCLIB) \\\$\$OBJECTS;ranlib \$(GAMBCLIB)"
#define ___MAKE_GAMBCLIB_SH "echo"
#define ___LD_CMD "/usr/bin/ld %s.o -o %s"
#define ___SETDLPATH "LD_LIBRARY_PATH=../lib"
#endif
#endif
#endif

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/* SGI MIPS running IRIX. */

#ifdef sgi
#ifdef mips

#define ___CPU_mips
#define ___CPU_BIGEND
#define ___C_TYPES_USUAL_32BITS

#define ___o "o"

#ifdef __GNUC__

#define ___CC_O "gcc -O1 -c \$(CC_O_ARGS) -D___SINGLE_HOST"
#define ___CC_O_SH "gcc -fpic -O1 -c \$(CC_O_SH_ARGS) -D___SINGLE_HOST"
#define ___LINK "gcc \\\$\$OBJECTS ../lib/\$(GAMBCLIB) \$(CLIBS) -o \\\$\$EXECUTABLE"
#define ___LINK_SH "gcc \\\$\$OBJECTS -L../lib -lgambc \$(CLIBS) -o \\\$\$EXECUTABLE"
#define ___LINK_SH_INSTALL "gcc \\\$\$OBJECTS -L\$(libdir) -lgambc \$(CLIBS) -o \\\$\$EXECUTABLE"
#define ___MAKE_GAMBCLIB_SH "gcc -shared -o \$(GAMBCLIB_SH) \\\$\$OBJECTS \$(CLIBS)"
#define ___CC_CMD "gcc -fPIC -O1 -c -D___DYNAMIC -D___SINGLE_HOST -o %s.o %s.c"

#else

#define ___CC_O "cc -OPT:IEEE_comparisons=ON -c \$(CC_O_ARGS)"
#define ___CC_O_SH "cc -OPT:IEEE_comparisons=ON -c \$(CC_O_SH_ARGS)"
#define ___LINK "cc \\\$\$OBJECTS ../lib/\$(GAMBCLIB) \$(CLIBS) -o \\\$\$EXECUTABLE"
#define ___LINK_SH "cc \\\$\$OBJECTS -L../lib -lgambc \$(CLIBS) -o \\\$\$EXECUTABLE"
#define ___LINK_SH_INSTALL "cc \\\$\$OBJECTS -L\$(libdir) -lgambc \$(CLIBS) -o \\\$\$EXECUTABLE"
#define ___MAKE_GAMBCLIB_SH "cc -shared -o \$(GAMBCLIB_SH) \\\$\$OBJECTS \$(CLIBS)"
#define ___CC_CMD "cc -OPT:IEEE_comparisons=ON -c -D___DYNAMIC -o %s.o %s.c"

#endif

#define ___CLIBS "-lm -lcurses"
#define ___GAMBCLIB "libgambc.a"
#define ___GAMBCLIB_SH "libgambc.so"
#define ___MAKE_GAMBCLIB "ar rc \$(GAMBCLIB) \\\$\$OBJECTS;ranlib \$(GAMBCLIB)"
#define ___LD_CMD "/usr/bin/ld -shared %s.o -o %s"
#define ___SETDLPATH "LD_LIBRARY_PATH=../lib LD_LIBRARYN32_PATH=../lib LD_LIBRARY64_PATH=../lib"

#endif
#endif

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

#ifdef __alpha

#define ___CPU_alpha
#define ___CPU_LITEND
#define ___C_TYPES_USUAL_64BITS
#define ___BUILTIN_64BIT_INT_TYPE long

#undef __GNUC__
#ifdef ___FORCE_32
#define ___WORD_IS_SMALLER_THAN_POINTERS
#undef __GNUC__
#define ___o "o"
#define ___CC_O "cc -O2 -w -ieee -c \$(CC_O_ARGS) -D___SINGLE_HOST"
#define ___CC_O_SH "cc -O2 -w -ieee -c \$(CC_O_SH_ARGS) -D___SINGLE_HOST"
#define ___LINK "cc -O2 -taso -call_shared \\\$\$OBJECTS ../lib/\$(GAMBCLIB) \$(CLIBS) -o \\\$\$EXECUTABLE"
#define ___LINK_SH "cc -O2 -taso -call_shared \\\$\$OBJECTS -L../lib -lgambc \$(CLIBS) -o \\\$\$EXECUTABLE"
#define ___LINK_SH_INSTALL "cc -O2 -taso -call_shared \\\$\$OBJECTS -L\$(libdir) -lgambc \$(CLIBS) -o \\\$\$EXECUTABLE"
#define ___CLIBS "-lm -lcurses"
#define ___GAMBCLIB "libgambc.a"
#define ___GAMBCLIB_SH "libgambc.so"
#define ___MAKE_GAMBCLIB "ar rc \$(GAMBCLIB) \\\$\$OBJECTS;ranlib \$(GAMBCLIB)"
#define ___MAKE_GAMBCLIB_SH "cc -taso -shared -call_shared -expect_unresolved \\* -o \$(GAMBCLIB_SH) \\\$\$OBJECTS \$(CLIBS)"
#define ___CC_CMD "cc -O2 -w -ieee -c -D___DYNAMIC -D___SINGLE_HOST -o %s.o %s.c"
#define ___LD_CMD "/usr/bin/ld -taso -shared -expect_unresolved \\* %s.o -o %s"
#define ___SETDLPATH "LD_LIBRARY_PATH=../lib"
#else
#define ___o "o"
#ifdef __GNUC__
#define ___CC_O "gcc -O1 -c \$(CC_O_ARGS) -D___SINGLE_HOST"
#define ___CC_O_SH "gcc -fpic -O1 -c \$(CC_O_SH_ARGS) -D___SINGLE_HOST"
#define ___LINK "gcc \\\$\$OBJECTS ../lib/\$(GAMBCLIB) \$(CLIBS) -o \\\$\$EXECUTABLE"
#define ___LINK_SH "gcc \\\$\$OBJECTS -L../lib -lgambc \$(CLIBS) -o \\\$\$EXECUTABLE"
#define ___LINK_SH_INSTALL "gcc \\\$\$OBJECTS -L\$(libdir) -lgambc \$(CLIBS) -o \\\$\$EXECUTABLE"
#define ___MAKE_GAMBCLIB_SH "gcc -shared -o \$(GAMBCLIB_SH) \\\$\$OBJECTS \$(CLIBS)"
#define ___CC_CMD "gcc -fPIC -O1 -c -D___DYNAMIC -D___SINGLE_HOST -o %s.o %s.c"
#else
#define ___CC_O "cc -O2 -ieee -c \$(CC_O_ARGS) -D___SINGLE_HOST"
#define ___CC_O_SH "cc -O2 -ieee -c \$(CC_O_SH_ARGS) -D___SINGLE_HOST"
#define ___LINK "cc -O2 -call_shared \\\$\$OBJECTS ../lib/\$(GAMBCLIB) \$(CLIBS) -o \\\$\$EXECUTABLE"
#define ___LINK_SH "cc -O2 -call_shared \\\$\$OBJECTS -L../lib -lgambc \$(CLIBS) -o \\\$\$EXECUTABLE"
#define ___LINK_SH_INSTALL "cc -O2 -call_shared \\\$\$OBJECTS -L\$(libdir) -lgambc \$(CLIBS) -o \\\$\$EXECUTABLE"
#define ___MAKE_GAMBCLIB_SH "cc -O2 -shared -call_shared -expect_unresolved \\* -o \$(GAMBCLIB_SH) \\\$\$OBJECTS \$(CLIBS)"
#define ___CC_CMD "cc -O2 -ieee -c -D___DYNAMIC -D___SINGLE_HOST -o %s.o %s.c"
#endif
#define ___CLIBS "-lm -lcurses"
#define ___GAMBCLIB "libgambc.a"
#define ___GAMBCLIB_SH "libgambc.so"
#define ___MAKE_GAMBCLIB "ar rc \$(GAMBCLIB) \\\$\$OBJECTS;ranlib \$(GAMBCLIB)"
#define ___LD_CMD "/usr/bin/ld -shared -expect_unresolved \\* %s.o -o %s"
#define ___SETDLPATH "LD_LIBRARY_PATH=../lib"
#endif

#endif

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

#ifdef hpux

#ifdef __hp9000s300

#define ___CPU_m68k
#define ___CPU_BIGEND
#define ___C_TYPES_USUAL_32BITS

#endif

#ifdef __hppa

#define ___CPU_hppa
#define ___CPU_BIGEND
#define ___C_TYPES_USUAL_32BITS

#endif

#define ___o "o"
#ifdef __GNUC__
#define ___CC_O "gcc -O1 -c \$(CC_O_ARGS) -D___SINGLE_HOST"
#define ___CC_O_SH "gcc -fPIC -O1 -c \$(CC_O_SH_ARGS) -D___SINGLE_HOST"
#define ___LINK "gcc \\\$\$OBJECTS ../lib/\$(GAMBCLIB) \$(CLIBS) -o \\\$\$EXECUTABLE"
#define ___LINK_SH "gcc \\\$\$OBJECTS -L../lib -lgambc \$(CLIBS) -o \\\$\$EXECUTABLE"
#define ___LINK_SH_INSTALL "gcc \\\$\$OBJECTS -L\$(libdir) -lgambc \$(CLIBS) -o \\\$\$EXECUTABLE"
#define ___MAKE_GAMBCLIB_SH "gcc -shared -fPIC -o \$(GAMBCLIB_SH) \\\$\$OBJECTS \$(CLIBS)"
#define ___CC_CMD "gcc -fPIC -O1 -c -D___DYNAMIC -D___SINGLE_HOST -o %s.o %s.c"
#else
#define ___CC_O "cc +O1 -c \$(CC_O_ARGS) -D___SINGLE_HOST"
#define ___CC_O_SH "cc +z +O1 -c \$(CC_O_SH_ARGS) -D___SINGLE_HOST"
#define ___LINK "cc \\\$\$OBJECTS ../lib/\$(GAMBCLIB) \$(CLIBS) -o \\\$\$EXECUTABLE"
#define ___LINK_SH "cc \\\$\$OBJECTS -L../lib -lgambc \$(CLIBS) -o \\\$\$EXECUTABLE"
#define ___LINK_SH_INSTALL "cc \\\$\$OBJECTS -L\$(libdir) -lgambc \$(CLIBS) -o \\\$\$EXECUTABLE"
#define ___MAKE_GAMBCLIB_SH "/bin/ld -b -o \$(GAMBCLIB_SH) \\\$\$OBJECTS \$(CLIBS)"
#define ___CC_CMD "cc +z -c -D___DYNAMIC -D___SINGLE_HOST -o %s.o %s.c"
#define ___SCHAR char
#endif
#define ___CLIBS "-lm -ldld -lcurses"
#define ___GAMBCLIB "libgambc.a"
#define ___GAMBCLIB_SH "libgambc.sl"
#define ___MAKE_GAMBCLIB "ar rc \$(GAMBCLIB) \\\$\$OBJECTS;ranlib \$(GAMBCLIB)"
#define ___LD_CMD "/bin/ld -b %s.o -lm -o %s"
#define ___CANT_IMPORT_DYNAMICALLY
#define ___SETDLPATH "SHLIB_PATH=../lib"
#endif

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

#ifdef _AIX

#define ___CPU_rs6k
#define ___CPU_BIGEND
#define ___C_TYPES_USUAL_32BITS

#define ___o "o"
#ifdef __GNUC__
#define ___CC_O "gcc -O1 -c \$(CC_O_ARGS) -D___SINGLE_HOST"
#define ___LINK "gcc \\\$\$OBJECTS ../lib/\$(GAMBCLIB) \$(CLIBS) -o \\\$\$EXECUTABLE"
#else
#define ___CC_O "cc -c \$(CC_O_ARGS) -D___SINGLE_HOST"
#define ___LINK "cc \\\$\$OBJECTS ../lib/\$(GAMBCLIB) \$(CLIBS) -o \\\$\$EXECUTABLE"
#endif
#define ___LINK_SH "echo"
#define ___LINK_SH_INSTALL "echo"
#define ___CLIBS "-lm -lcurses"
#define ___GAMBCLIB "libgambc.a"
#define ___MAKE_GAMBCLIB "ar rc \$(GAMBCLIB) \\\$\$OBJECTS;ranlib \$(GAMBCLIB)"
#define ___MAKE_GAMBCLIB_SH "echo"
#define ___SETDLPATH "LIBPATH=../lib"
#endif

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

#ifdef sony_news

#define ___CPU_sony_news
#define ___CPU_BIGEND
#define ___C_TYPES_USUAL_32BITS

#define ___o "o"
#define ___CC_O "gcc -O1 -c \$(CC_O_ARGS) -D___SINGLE_HOST"
#define ___LINK "gcc \\\$\$OBJECTS ../lib/\$(GAMBCLIB) \$(CLIBS) -o \\\$\$EXECUTABLE"
#define ___LINK_SH "echo"
#define ___LINK_SH_INSTALL "echo"
#define ___CLIBS "-lm -lcurses"
#define ___GAMBCLIB "libgambc.a"
#define ___MAKE_GAMBCLIB "ar rc \$(GAMBCLIB) \\\$\$OBJECTS;ranlib \$(GAMBCLIB)"
#define ___MAKE_GAMBCLIB_SH "echo"
#define ___CC_CMD "gcc -fPIC -O1 -c -D___DYNAMIC -D___SINGLE_HOST -o %s.o %s.c"
#define ___LD_CMD "/usr/ccs/bin/ld -G %s.o -o %s"
#define ___CANT_IMPORT_DYNAMICALLY
#define ___SETDLPATH "LD_LIBRARY_PATH=../lib"
#endif

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

#ifdef __linux__using_gplusplus
#ifdef __i386__

#define ___CPU_x86
#define ___CPU_LITEND
#define ___C_TYPES_USUAL_32BITS

#define ___LINK "g++ -fhandle-exceptions -rdynamic \\\$\$OBJECTS ../lib/\$(GAMBCLIB) \$(CLIBS) -o \\\$\$EXECUTABLE"
#define ___LINK_SH "g++ -fhandle-exceptions -rdynamic \\\$\$OBJECTS -L../lib -lgambc \$(CLIBS) -o \\\$\$EXECUTABLE"
#define ___LINK_SH_INSTALL "g++ -fhandle-exceptions -rdynamic \\\$\$OBJECTS -L\$(libdir) -lgambc \$(CLIBS) -o \\\$\$EXECUTABLE"
#endif
#ifdef __sparc__

#define ___CPU_sparc
#define ___CPU_BIGEND
#define ___C_TYPES_USUAL_32BITS

#define ___LINK "g++ -fhandle-exceptions -rdynamic \\\$\$OBJECTS ../lib/\$(GAMBCLIB) \$(CLIBS) -o \\\$\$EXECUTABLE"
#define ___LINK_SH "g++ -fhandle-exceptions -rdynamic \\\$\$OBJECTS -L../lib -lgambc -\$(CLIBS) o \\\$\$EXECUTABLE"
#define ___LINK_SH_INSTALL "g++ -fhandle-exceptions -rdynamic \\\$\$OBJECTS -L\$(libdir) -lgambc \$(CLIBS) -o \\\$\$EXECUTABLE"
#endif
#ifdef __powerpc__

#define ___CPU_ppc
#define ___CPU_BIGEND
#define ___C_TYPES_USUAL_32BITS

#define ___LINK "g++ -fhandle-exceptions -export-dynamic \\\$\$OBJECTS ../lib/\$(GAMBCLIB) \$(CLIBS) -o \\\$\$EXECUTABLE"
#define ___LINK_SH "g++ -fhandle-exceptions -export-dynamic \\\$\$OBJECTS -L../lib -lgambc \$(CLIBS) -o \\\$\$EXECUTABLE"
#define ___LINK_SH_INSTALL "g++ -fhandle-exceptions -export-dynamic \\\$\$OBJECTS -L\$(libdir) -lgambc \$(CLIBS) -o \\\$\$EXECUTABLE"
#endif
#define ___o "o"
#define ___CC_O "g++ -fhandle-exceptions -O1 -c \$(CC_O_ARGS) -D___SINGLE_HOST"
#define ___CC_O_SH "g++ -fhandle-exceptions -fpic -O1 -c \$(CC_O_SH_ARGS) -D___SINGLE_HOST"
#define ___CLIBS "-lm -ldl -lncurses"
#define ___GAMBCLIB "libgambc.a"
#define ___GAMBCLIB_SH "libgambc.so"
#define ___MAKE_GAMBCLIB "ar rc \$(GAMBCLIB) \\\$\$OBJECTS;ranlib \$(GAMBCLIB)"
#define ___MAKE_GAMBCLIB_SH "g++ -fhandle-exceptions -shared -o \$(GAMBCLIB_SH) \\\$\$OBJECTS \$(CLIBS)"
#define ___CC_CMD "g++ -fhandle-exceptions -fPIC -c -D___DYNAMIC -D___SINGLE_HOST -o %s.o %s.c"
#define ___LD_CMD "/usr/bin/ld -shared %s.o -o %s"
#define ___SETDLPATH "LD_LIBRARY_PATH=../lib"
#endif

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

#ifdef __linux__

#ifdef __alpha__

#define ___CPU_alpha
#define ___CPU_LITEND
#define ___C_TYPES_USUAL_64BITS
#define ___BUILTIN_64BIT_INT_TYPE long
#define HAS_fpu_control_h

#define ___LINK "gcc -g -rdynamic \\\$\$OBJECTS ../lib/\$(GAMBCLIB) \$(CLIBS) -o \\\$\$EXECUTABLE"
#define ___LINK_SH "gcc -g -rdynamic \\\$\$OBJECTS -L../lib -lgambc \$(CLIBS) -o \\\$\$EXECUTABLE"
#define ___LINK_SH_INSTALL "gcc -g -rdynamic \\\$\$OBJECTS -L\$(libdir) -lgambc \$(CLIBS) -o \\\$\$EXECUTABLE"

/*********fix*/
#define ___CC_O "gcc -Wall -Wno-unused-label -O1 -g -mieee -fno-math-errno -mcpu=ev6 -fschedule-insns2 -Wdisabled-optimization -c \$(CC_O_ARGS) -D___SINGLE_HOST"
#define ___CC_O_SH "gcc -Wall -Wno-unused-label -O1 -g -mieee -fno-math-errno -mcpu=ev6 -fschedule-insns2 -Wdisabled-optimization -c \$(CC_O_SH_ARGS) -D___SINGLE_HOST"
#define ___CC_CMD "gcc -Wall -Wno-unused-label -fPIC -O1 -g -mieee -fno-math-errno -mcpu=ev6 -fschedule-insns2 -Wdisabled-optimization -c -D___DYNAMIC -D___SINGLE_HOST -o %s.o %s.c"

#endif

#ifdef __i386__

#define ___CPU_x86
#define ___CPU_LITEND
#define ___C_TYPES_USUAL_32BITS

#define ___LINK "gcc -rdynamic \\\$\$OBJECTS ../lib/\$(GAMBCLIB) \$(CLIBS) -o \\\$\$EXECUTABLE"
#define ___LINK_SH "gcc -rdynamic \\\$\$OBJECTS -L../lib -lgambc \$(CLIBS) -o \\\$\$EXECUTABLE"
#define ___LINK_SH_INSTALL "gcc -rdynamic \\\$\$OBJECTS -L\$(libdir) -lgambc \$(CLIBS) -o \\\$\$EXECUTABLE"
#define ___CC_O "gcc -fomit-frame-pointer -O1 -c \$(CC_O_ARGS) -D___SINGLE_HOST"
#define ___CC_O_SH "gcc -fpic -fomit-frame-pointer -O1 -c \$(CC_O_SH_ARGS) -D___SINGLE_HOST"
#define ___CC_CMD "gcc -fomit-frame-pointer -fPIC -O1 -c -D___DYNAMIC -D___SINGLE_HOST -o %s.o %s.c"

#endif

#ifdef __sparc__

#define ___CPU_sparc
#define ___CPU_BIGEND
#define ___C_TYPES_USUAL_32BITS

#define ___LINK "gcc -rdynamic \\\$\$OBJECTS ../lib/\$(GAMBCLIB) \$(CLIBS) -o \\\$\$EXECUTABLE"
#define ___LINK_SH "gcc -rdynamic \\\$\$OBJECTS -L../lib -lgambc \$(CLIBS) -o \\\$\$EXECUTABLE"
#define ___LINK_SH_INSTALL "gcc -rdynamic \\\$\$OBJECTS -L\$(libdir) -lgambc \$(CLIBS) -o \\\$\$EXECUTABLE"
#define ___CC_O "gcc -fomit-frame-pointer -O1 -c \$(CC_O_ARGS) -D___SINGLE_HOST"
#define ___CC_O_SH "gcc -fpic -fomit-frame-pointer -O1 -c \$(CC_O_SH_ARGS) -D___SINGLE_HOST"
#define ___CC_CMD "gcc -fomit-frame-pointer -fPIC -O1 -c -D___DYNAMIC -D___SINGLE_HOST -o %s.o %s.c"

#endif

#ifdef __powerpc__

#define ___CPU_ppc
#define ___CPU_BIGEND
#define ___C_TYPES_USUAL_32BITS

#define ___LINK "gcc -export-dynamic \\\$\$OBJECTS ../lib/\$(GAMBCLIB) \$(CLIBS) -o \\\$\$EXECUTABLE"
#define ___LINK_SH "gcc -export-dynamic \\\$\$OBJECTS -L../lib -lgambc \$(CLIBS) -o \\\$\$EXECUTABLE"
#define ___LINK_SH_INSTALL "gcc -export-dynamic \\\$\$OBJECTS -L\$(libdir) -lgambc \$(CLIBS) -o \\\$\$EXECUTABLE"
#define ___CC_O "gcc -fomit-frame-pointer -O1 -c \$(CC_O_ARGS) -D___SINGLE_HOST"
#define ___CC_O_SH "gcc -fpic -fomit-frame-pointer -O1 -c \$(CC_O_SH_ARGS) -D___SINGLE_HOST"
#define ___CC_CMD "gcc -fomit-frame-pointer -fPIC -O1 -c -D___DYNAMIC -D___SINGLE_HOST -o %s.o %s.c"

#endif

#define ___o "o"
#define ___CLIBS "-lm -ldl -lncurses"
#define ___GAMBCLIB "libgambc.a"
#define ___GAMBCLIB_SH "libgambc.so"
#define ___MAKE_GAMBCLIB "ar rc \$(GAMBCLIB) \\\$\$OBJECTS;ranlib \$(GAMBCLIB)"
#define ___MAKE_GAMBCLIB_SH "gcc -shared -o \$(GAMBCLIB_SH) \\\$\$OBJECTS \$(CLIBS)"
#define ___LD_CMD "/usr/bin/ld -shared %s.o -o %s"
#define ___SETDLPATH "LD_LIBRARY_PATH=../lib"

#endif

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

#ifdef __OpenBSD__
#ifdef i386

#define ___CPU_x86
#define ___CPU_LITEND
#define ___C_TYPES_USUAL_32BITS

#define ___IMPORTED_ID_PREFIX "_"
#define ___o "o"
#define ___CC_O "gcc -O1 -c \$(CC_O_ARGS) -D___SINGLE_HOST"
#define ___CC_O_SH "gcc -fpic -O1 -c \$(CC_O_SH_ARGS) -D___SINGLE_HOST"
#define ___LINK "gcc \\\$\$OBJECTS ../lib/\$(GAMBCLIB) \$(CLIBS) -o \\\$\$EXECUTABLE"
#define ___LINK_SH "gcc \\\$\$OBJECTS -L../lib -lgambc \$(CLIBS) -o \\\$\$EXECUTABLE"
#define ___LINK_SH_INSTALL "gcc \\\$\$OBJECTS -L\$(libdir) -lgambc \$(CLIBS) -o \\\$\$EXECUTABLE"
#define ___CLIBS "-lm -lncurses"
#define ___GAMBCLIB "libgambc.a"
#define ___GAMBCLIB_SH "libgambc.so.1.1"
#define ___MAKE_GAMBCLIB "ar rc \$(GAMBCLIB) \\\$\$OBJECTS;ranlib \$(GAMBCLIB)"
#define ___MAKE_GAMBCLIB_SH "ld -Bshareable -o \$(GAMBCLIB_SH) \\\$\$OBJECTS \$(CLIBS)"
#define ___CC_CMD "gcc -fPIC -O1 -c -D___DYNAMIC -D___SINGLE_HOST -o %s.o %s.c"
#define ___LD_CMD "/usr/bin/ld -Bshareable %s.o -o %s"
#define ___SETDLPATH "LD_LIBRARY_PATH=../lib"
#endif
#endif

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

#ifdef __NetBSD__
#ifdef i386

#define ___CPU_x86
#define ___CPU_LITEND
#define ___C_TYPES_USUAL_32BITS

#define ___IMPORTED_ID_PREFIX "_"
#define ___o "o"
#define ___CC_O "gcc -O1 -c \$(CC_O_ARGS) -D___SINGLE_HOST"
#define ___CC_O_SH "gcc -fpic -O1 -c \$(CC_O_SH_ARGS) -D___SINGLE_HOST"
#define ___LINK "gcc \\\$\$OBJECTS ../lib/\$(GAMBCLIB) \$(CLIBS) -o \\\$\$EXECUTABLE"
#define ___LINK_SH "gcc \\\$\$OBJECTS -L../lib -lgambc \$(CLIBS) -o \\\$\$EXECUTABLE"
#define ___LINK_SH_INSTALL "gcc \\\$\$OBJECTS -L\$(libdir) -lgambc \$(CLIBS) -o \\\$\$EXECUTABLE"
#define ___CLIBS "-lm -lncurses"
#define ___GAMBCLIB "libgambc.a"
#define ___GAMBCLIB_SH "libgambc.so.1.1"
#define ___MAKE_GAMBCLIB "ar rc \$(GAMBCLIB) \\\$\$OBJECTS;ranlib \$(GAMBCLIB)"
#define ___MAKE_GAMBCLIB_SH "ld -Bshareable -o \$(GAMBCLIB_SH) \\\$\$OBJECTS \$(CLIBS)"
#define ___CC_CMD "gcc -fPIC -O1 -c -D___DYNAMIC -D___SINGLE_HOST -o %s.o %s.c"
#define ___LD_CMD "/usr/bin/ld -Bshareable %s.o -o %s"
#define ___SETDLPATH "LD_LIBRARY_PATH=../lib"
#endif
#endif

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

#ifdef __FreeBSD__
#ifdef i386

#define ___CPU_x86
#define ___CPU_LITEND
#define ___C_TYPES_USUAL_32BITS

#define ___IMPORTED_ID_PREFIX "_"
#define ___o "o"
#define ___CC_O "gcc -O1 -c \$(CC_O_ARGS) -D___SINGLE_HOST"
#define ___CC_O_SH "gcc -fpic -O1 -c \$(CC_O_SH_ARGS) -D___SINGLE_HOST"
#define ___LINK "gcc \\\$\$OBJECTS ../lib/\$(GAMBCLIB) \$(CLIBS) -o \\\$\$EXECUTABLE"
#define ___LINK_SH "gcc \\\$\$OBJECTS -L../lib -lgambc \$(CLIBS) -o \\\$\$EXECUTABLE"
#define ___LINK_SH_INSTALL "gcc \\\$\$OBJECTS -L\$(libdir) -lgambc \$(CLIBS) -o \\\$\$EXECUTABLE"
#define ___CLIBS "-lm -lncurses"
#define ___GAMBCLIB "libgambc.a"
#define ___GAMBCLIB_SH "libgambc.so.1.1"
#define ___MAKE_GAMBCLIB "ar rc \$(GAMBCLIB) \\\$\$OBJECTS;ranlib \$(GAMBCLIB)"
#define ___MAKE_GAMBCLIB_SH "ld -Bshareable -o \$(GAMBCLIB_SH) \\\$\$OBJECTS \$(CLIBS)"
#define ___CC_CMD "gcc -fPIC -O1 -c -D___DYNAMIC -D___SINGLE_HOST -o %s.o %s.c"
#define ___LD_CMD "/usr/bin/ld -Bshareable %s.o -o %s"
#define ___SETDLPATH "LD_LIBRARY_PATH=../lib"
#endif
#endif

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

#ifdef __DJGPP__
#ifdef __i386__

#define ___CPU_x86
#define ___CPU_LITEND
#define ___C_TYPES_USUAL_32BITS

#define ___o "o"
#define ___CC_O "gcc -c \$(CC_O_ARGS) -D___SINGLE_HOST"
#define ___LINK "gcc \\\$\$OBJECTS ../lib/\$(GAMBCLIB) \$(CLIBS) -o \\\$\$EXECUTABLE"
#define ___LINK_SH "echo"
#define ___LINK_SH_INSTALL "echo"
#define ___CLIBS "-lm -lemu"
#define ___GAMBCLIB "libgambc.a"
#define ___MAKE_GAMBCLIB "ar rc \$(GAMBCLIB) \\\$\$OBJECTS;ranlib \$(GAMBCLIB)"
#define ___MAKE_GAMBCLIB_SH "echo"
#define ___CC_CMD "gcc -c -D___DYNAMIC -D___SINGLE_HOST -o %s.o %s.c"
#define ___LD_CMD "dxegen %s ____linker %s.o"
#define ___LD_FLIP
#define ___SETDLPATH ""
#define ___CANT_IMPORT_DYNAMICALLY
#ifdef ___DYNAMIC
#undef ___LINKER_ID
#define ___LINKER_ID ___linker
#endif
#define ___CANT_IMPORT_CLIB_DYNAMICALLY
#define ___CANT_IMPORT_SETJMP_DYNAMICALLY
#endif
#endif

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

#ifdef _MSC_VER
#ifndef __MWERKS__
#if _MSC_VER < 1000
#ifdef _M_I86

#define ___CPU_x86
#define ___CPU_LITEND
#define ___C_TYPES_USUAL_32BITS

#define ___EXPORT_FUNC(type,name)__declspec(dllexport) type name
#define ___EXPORT_DATA(type,name)__declspec(dllexport) type name
#define ___IMPORT_FUNC(type,name)__declspec(dllimport) type name
#define ___IMPORT_DATA(type,name)__declspec(dllimport) type name
#define ___o "obj"
#define ___CC_O "cl -W0 -AH -Oitb1 -G3s \$(CFLAGS) -c \$(CC_O_ARGS) -D___SINGLE_HOST"
#define ___LINK "cl \$(LFLAGS) \\\$\$OBJECTS ../lib/\$(GAMBCLIB) \$(CLIBS) -o \\\$\$EXECUTABLE"
#define ___LINK_SH "echo"
#define ___LINK_SH_INSTALL "echo"
#define ___CLIBS ""
#define ___GAMBCLIB "libgambc.lib"
#define ___GAMBCLIB_SH "gambc.lib"
#define ___MAKE_GAMBCLIB "lib -out:\$(GAMBCLIB) \\\$\$OBJECTS"
#define ___MAKE_GAMBCLIB_SH "link -dll \$(LFLAGS) -out:gambc.dll \\\$\$OBJECTS \$(CLIBS)"
#define ___CC_CMD "cl -W0 -AH -Oitb1 -G3s -D___DYNAMIC -D___SINGLE_HOST -c -Fo%s.obj %s.c"
#define ___LD_CMD "link -noentry -dll %s.obj -out:%s"
#define ___SETDLPATH "OLDPATH=\"\$\$PATH\" PATH=\"\$\$OLDPATH;..\\\\\\\\lib\""
#define ___CANT_IMPORT_DYNAMICALLY
#define ___IMPORTED_ADR_NOT_CONST
#define ___USE_ISNAN
#endif
#else
#ifdef _M_IX86

#define ___CPU_x86
#define ___CPU_LITEND
#define ___C_TYPES_USUAL_32BITS

#define ___EXPORT_FUNC(type,name)__declspec(dllexport) type name
#define ___EXPORT_DATA(type,name)__declspec(dllexport) type name
#define ___IMPORT_FUNC(type,name)__declspec(dllimport) type name
#define ___IMPORT_DATA(type,name)__declspec(dllimport) type name
#define ___o "obj"
#define ___CC_O "cl -nologo -Oityb1 -G5s -MD \$(CFLAGS) -c \$(CC_O_ARGS) -D___SINGLE_HOST"
#define ___CC_O_SH "cl -nologo -Oityb1 -G5s -MD \$(CFLAGS) -c \$(CC_O_SH_ARGS) -D___SINGLE_HOST"
#define ___LINK "cl -nologo \$(LFLAGS) \\\$\$OBJECTS ../lib/\$(GAMBCLIB) \$(CLIBS) -o \\\$\$EXECUTABLE"
#define ___LINK_SH "cl -nologo \$(LFLAGS) \\\$\$OBJECTS ../lib/\$(GAMBCLIB_SH) -o \\\$\$EXECUTABLE"
#define ___LINK_SH_INSTALL "cl -nologo \$(LFLAGS) \\\$\$OBJECTS ../lib/\$(GAMBCLIB_SH) -o \\\$\$EXECUTABLE"
#define ___CLIBS ""
#define ___GAMBCLIB "libgambc.lib"
#define ___GAMBCLIB_SH "gambc.lib"
#define ___MAKE_GAMBCLIB "lib -out:\$(GAMBCLIB) \\\$\$OBJECTS"
#define ___MAKE_GAMBCLIB_SH "link -nologo -dll \$(LFLAGS) -out:gambc.dll \\\$\$OBJECTS \$(CLIBS)"
#define ___CC_CMD "cl -nologo -Oityb1 -G5s -MD -D___DYNAMIC -D___SINGLE_HOST -c -Fo%s.obj %s.c"
#define ___LD_CMD "link -nologo -noentry -dll %s.obj -out:%s"
#define ___SETDLPATH "OLDPATH=\"\$\$PATH\" PATH=\"\$\$OLDPATH;..\\\\\\\\lib\""
#define ___CANT_IMPORT_DYNAMICALLY
#define ___IMPORTED_ADR_NOT_CONST
#define ___USE_ISNAN
#endif
#endif
#endif
#endif

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

#ifdef __BORLANDC__
#ifdef _M_IX86

#define ___CPU_x86
#define ___CPU_LITEND
#define ___C_TYPES_USUAL_32BITS

#define ___EXPORT_FUNC(type,name)type __export name
#define ___EXPORT_DATA(type,name)type __export name
#define ___IMPORT_FUNC(type,name)type __import name
#define ___IMPORT_DATA(type,name)type __import name
#define ___IMPORTED_ID_PREFIX "_"
#define ___o "obj"
#define ___CC_O "bcc32 -w- -O2 -a4 -tWM \$(CFLAGS) -c \$(CC_O_ARGS) -D___SINGLE_HOST"
#define ___CC_O_SH "bcc32 -w- -O2 -a4 -tWM \$(CFLAGS) -c \$(CC_O_SH_ARGS) -D___SINGLE_HOST"
#define ___LINK "bcc32 -e\\\$\$EXECUTABLE \$(LFLAGS) \\\$\$OBJECTS ../lib/\$(GAMBCLIB) \$(CLIBS)"
#define ___LINK_SH "bcc32 -e\\\$\$EXECUTABLE \$(LFLAGS) \\\$\$OBJECTS ../lib/\$(GAMBCLIB_SH)"
#define ___LINK_SH_INSTALL "bcc32 -e\\\$\$EXECUTABLE \$(LFLAGS) \\\$\$OBJECTS ../lib/\$(GAMBCLIB_SH)"
#define ___CLIBS ""
#define ___GAMBCLIB "libgambc.lib"
#define ___GAMBCLIB_SH "gambc.lib"
#define ___MAKE_GAMBCLIB "tlib \$(GAMBCLIB) /p32 \\\$\$OBJECTS_PLUS"
#define ___MAKE_GAMBCLIB_SH "tlink32 -Tpd \$(LFLAGS) c0d32.obj \\\$\$OBJECTS, gambc.dll, , import32 cw32mti ; implib \$(GAMBCLIB_SH) gambc.dll"
#define ___CC_CMD "bcc32 -w- -O2 -a4 -tWM -D___DYNAMIC -D___SINGLE_HOST -c -o%s.obj %s.c"
#define ___LD_CMD "tlink32 -Tpd c0d32.obj %s.obj, %s, , import32 cw32mti"
#define ___SETDLPATH "OLDPATH=\"\$\$PATH\" PATH=\"\$\$OLDPATH;..\\\\\\\\lib\""
#define ___CANT_IMPORT_DYNAMICALLY
#define ___CANT_IMPORT_EXPORTED
#define ___IMPORTED_ADR_NOT_CONST
#define ___USE_ISNAN
#endif
#endif

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

#ifdef __WATCOMC__
#ifdef __386__

#define ___CPU_x86
#define ___CPU_LITEND
#define ___C_TYPES_USUAL_32BITS

#define ___EXPORT_FUNC(type,name)__declspec(dllexport) type name
#define ___EXPORT_DATA(type,name)__declspec(dllexport) type name
#define ___IMPORT_FUNC(type,name)__declspec(dllimport) type name
#define ___IMPORT_DATA(type,name)__declspec(dllimport) type name
#define ___IMPORTED_ID_SUFFIX "_"
#define ___o "obj"
#define ___CC_O "wcc386 -zq -w0 -obetir -zp4 -s -5r -fp5 -bm -bt=nt \$(CC_O_ARGS) -D___SINGLE_HOST"
#define ___CC_O_SH_BROKEN "wcc386 -zq -w0 -obetir -zp4 -s -5r -fp5 -bm -bt=nt -bd -br \$(CC_O_SH_ARGS) -D___SINGLE_HOST"
#define ___LINK "wlink option quiet option stack=16384 system nt file \\\$\$OBJECTS_COMMA library ..\\\\\\\\lib\\\\\\\\\$(GAMBCLIB),\$(CLIBS) name \\\$\$EXECUTABLE"
#define ___LINK_SH "wlink option quiet option stack=16384 system nt file \\\$\$OBJECTS_COMMA library ..\\\\\\\\lib\\\\\\\\\$(GAMBCLIB_SH),\$(CLIBS) name \\\$\$EXECUTABLE"
#define ___LINK_SH_INSTALL "echo"
#define ___CLIBS "kernel32,user32"
#define ___GAMBCLIB "libgambc.lib"
#define ___GAMBCLIB_SH "gambc.lib"
#define ___MAKE_GAMBCLIB "wlib -n -p=32 \$(GAMBCLIB) \\\$\$OBJECTS_PLUS"
#define ___MAKE_GAMBCLIB_SH "wlink option quiet option stack=16384 system nt_dll file \\\$\$OBJECTS_COMMA library \$(CLIBS),mthrdll name gambc.dll;wlib -n -p=32 \$(GAMBCLIB_SH) +gambc.dll"
#define ___CC_CMD "wcc386 -zq -w0 -obetir -zp4 -s -5r -fp5 -bm -bt=nt -bd -br -D___DYNAMIC -D___SINGLE_HOST -fo=%s.obj %s.c"
#define ___LD_CMD "wlink option quiet option stack=16384 system nt_dll file %s.obj name %s"
#define ___SETDLPATH "OLDPATH=\"\$\$PATH\" PATH=\"\$\$OLDPATH;..\\\\\\\\lib\""
#define ___CANT_IMPORT_CLIB_DYNAMICALLY
#define ___CANT_IMPORT_DYNAMICALLY
#define ___CANT_IMPORT_EXPORTED
#define ___USE_ISNAN
#endif
#endif

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

#ifdef __MACHTEN__

#ifdef __MACHTEN_PPC__

#define ___CPU_ppc
#define ___CPU_BIGEND
#define ___C_TYPES_USUAL_32BITS

#define ___LINK "cc -Xlstack=300000 \\\$\$OBJECTS ../lib/\$(GAMBCLIB) \$(CLIBS) -o \\\$\$EXECUTABLE"

#else

#define ___CPU_m68k
#define ___CPU_BIGEND
#define ___C_TYPES_USUAL_32BITS

#define ___LINK "cc \\\$\$OBJECTS ../lib/\$(GAMBCLIB) \$(CLIBS) -o \\\$\$EXECUTABLE"

#endif

#define ___LINK_SH "echo"
#define ___LINK_SH_INSTALL "echo"
#define ___o "o"
#define ___CC_O "cc -c \$(CC_O_ARGS)"
#define ___CLIBS "-lm -lncurses"
#define ___GAMBCLIB "libgambc.a"
#define ___MAKE_GAMBCLIB "ar rc \$(GAMBCLIB) \\\$\$OBJECTS;ranlib \$(GAMBCLIB)"
#define ___MAKE_GAMBCLIB_SH "echo"
#define ___SETDLPATH "LD_LIBRARY_PATH=../lib"
#endif

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

#ifdef NeXT

#ifdef i386

#define ___CPU_x86
#define ___CPU_LITEND
#define ___C_TYPES_USUAL_32BITS

#else

#define ___CPU_m68k
#define ___CPU_BIGEND
#define ___C_TYPES_USUAL_32BITS

#endif

#define ___o "o"
#ifdef __GNUC__
#define ___CC_O "gcc -posix -c \$(CC_O_ARGS) -D___SINGLE_HOST"
#define ___LINK "gcc -posix \\\$\$OBJECTS ../lib/\$(GAMBCLIB) \$(CLIBS) -o \\\$\$EXECUTABLE"
#else
#define ___CC_O "cc -posix -c \$(CC_O_ARGS) -D___SINGLE_HOST"
#define ___LINK "cc -posix \\\$\$OBJECTS ../lib/\$(GAMBCLIB) \$(CLIBS) -o \\\$\$EXECUTABLE"
#endif
#define ___LINK_SH "echo"
#define ___LINK_SH_INSTALL "echo"
#define ___CLIBS "-lm -lncurses"
#define ___GAMBCLIB "libgambc.a"
#define ___MAKE_GAMBCLIB "ar rc \$(GAMBCLIB) \\\$\$OBJECTS;ranlib \$(GAMBCLIB)"
#define ___MAKE_GAMBCLIB_SH "echo"
#define ___SETDLPATH "LD_LIBRARY_PATH=../lib"
#endif

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

#ifdef __MWERKS__
#ifdef __INTEL__

#define ___CPU_x86
#define ___CPU_LITEND
#define ___C_TYPES_USUAL_32BITS

#define ___SINGLE_HOST
#define ___EXPORT_FUNC(type,name)__declspec(dllexport) type name
#define ___EXPORT_DATA(type,name)__declspec(dllexport) type name
#define ___IMPORT_FUNC(type,name)__declspec(dllimport) type name
#define ___IMPORT_DATA(type,name)__declspec(dllimport) type name
#define ___CANT_IMPORT_DYNAMICALLY
#define ___CANT_IMPORT_EXPORTED
#else
#ifdef __POWERPC__

#define ___CPU_ppc
#define ___CPU_BIGEND
#define ___C_TYPES_USUAL_32BITS

#define ___SINGLE_HOST
#if __MWERKS__ == 0x1100
#define ___CC_CMD "~~:compile-file.cw11"
#else
#if __MWERKS__ == 0x1800
#define ___CC_CMD "~~:compile-file.cwpro1"
#else
#if __MWERKS__ == 0x2000
#define ___CC_CMD "~~:compile-file.cwpro2"
#else
#if __MWERKS__ == 0x2100
#define ___CC_CMD "~~:compile-file.cwpro3"
#else
#define ___CC_CMD "~~:compile-file.cwpro4"
#endif
#endif
#endif
#endif
#else

#define ___CPU_m68k
#define ___CPU_BIGEND
#define ___C_TYPES_USUAL_32BITS

#endif
#define ___EXPORT_FUNC(type,name)__declspec(export) type name
#define ___EXPORT_DATA(type,name)__declspec(export) type name
#define ___IMPORT_FUNC(type,name)__declspec(import) type name
#define ___IMPORT_DATA(type,name)__declspec(import) type name
#define ___CANT_IMPORT_DYNAMICALLY
#define ___CANT_IMPORT_EXPORTED
#endif
#if __option (warn_illpragma)
#define ___PRIMAL
#endif
#if __option (warn_unusedarg)
#define ___LIBRARY
#endif
#if __option (warn_extracomma)
#define ___SHARED
#endif
#if __option (warn_emptydecl)
#define ___DYNAMIC
#endif
#endif

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

#ifdef __MINGW32__

#define ___CPU_x86
#define ___CPU_LITEND
#define ___C_TYPES_USUAL_32BITS

#ifdef __ONLY_FOR_EXPORT__
#define ___EXPORT(type,name)EXPORTED: name :END
#define ___IMPORT(type,name)IMPORTED: name :END
#else
#define ___EXPORT(type,name)type name
#define ___IMPORT(type,name)extern type name
#endif
#define ___o "o"
#define ___CC_CMD "gambc-compile.cmd %s %s"
#define ___LD_CMD "gambc-link.cmd %s %s"
#define ___CANT_IMPORT_DYNAMICALLY
#define ___CANT_IMPORT_EXPORTED
#endif

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

#ifdef __EMX__

#define ___CPU_x86
#define ___CPU_LITEND
#define ___C_TYPES_USUAL_32BITS

#define ___EXPORT(type,name)type name
#define ___IMPORT(type,name)extern type name
#define ___o "o"
#define ___CC_CMD "gambc-compile.cmd %s %s"
#define ___LD_CMD "gambc-link.cmd %s %s"
#define ___CANT_IMPORT_DYNAMICALLY
#define ___CANT_IMPORT_EXPORTED
#endif

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

#ifdef _CRAY
#ifdef _CRAY1

#define ___CPU_cray_PVP
#define ___CPU_BIGEND
#define ___C_TYPES_CRAY_PVP

#define ___LINK "cc \\\$\$OBJECTS ../lib/\$(GAMBCLIB) \$(CLIBS) -o \\\$\$EXECUTABLE"
#define ___LINK_SH "echo"
#define ___LINK_SH_INSTALL "echo"
#define ___o "o"
#define ___CC_O "cc -c \$(CC_O_ARGS)"
#define ___CLIBS ""
#define ___GAMBCLIB "libgambc.a"
#define ___MAKE_GAMBCLIB "ar rc \$(GAMBCLIB) \\\$\$OBJECTS;ranlib \$(GAMBCLIB)"
#define ___MAKE_GAMBCLIB_SH "echo"
#define ___SETDLPATH ""
#ifndef _CRAYIEEE
#define ___CRAY_FP_FORMAT
#endif
#endif
#endif

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

#ifdef __LCC__
#ifdef _X86_
#ifdef _M_IX86

#define ___CPU_x86
#define ___CPU_LITEND
#define ___C_TYPES_USUAL_32BITS

#define ___IMPORTED_ID_PREFIX "_"
#define ___o "obj"
#define ___CC_O "c:/lcc/bin/lcc \$(CFLAGS) -c \$(CC_O_ARGS) -D___SINGLE_HOST"
#define ___CC_O_SH ""
#define ___LINK ""
#define ___LINK_SH ""
#define ___LINK_SH_INSTALL ""
#define ___CLIBS ""
#define ___GAMBCLIB "libgambc.lib"
#define ___GAMBCLIB_SH "gambc.lib"
#define ___MAKE_GAMBCLIB ""
#define ___MAKE_GAMBCLIB_SH ""
#define ___CC_CMD ""
#define ___LD_CMD ""
#define ___SETDLPATH "OLDPATH=\"\$\$PATH\" PATH=\"\$\$OLDPATH;..\\\\\\\\lib\""
#define ___CANT_IMPORT_DYNAMICALLYzzzzzzz
#define ___CANT_IMPORT_EXPORTEDzzzzzzzz
#define ___IMPORTED_ADR_NOT_CONSTzzzzzzzz
#define ___USE_ISNANzzzzzzz
#endif
#endif
#endif

/*---------------------------------------------------------------------------*/

/*
 * DETERMINE C COMPILER'S DATA TYPES
 */

#ifdef ___C_TYPES_USUAL_32BITS
#define ___CHAR_WIDTH     8
#define ___SHORT_WIDTH    16
#define ___INT_WIDTH      32
#define ___LONG_WIDTH     32
#define ___FLOAT_WIDTH    32
#define ___DOUBLE_WIDTH   64
#define ___VOIDSTAR_WIDTH 32
#endif

#ifdef ___C_TYPES_USUAL_64BITS
#define ___CHAR_WIDTH     8
#define ___SHORT_WIDTH    16
#define ___INT_WIDTH      32
#define ___LONG_WIDTH     64
#define ___FLOAT_WIDTH    32
#define ___DOUBLE_WIDTH   64
#define ___VOIDSTAR_WIDTH 64
#endif

#ifdef ___C_TYPES_CRAY_PVP
#define ___CHAR_WIDTH     8
#define ___SHORT_WIDTH    64
#define ___INT_WIDTH      64
#define ___LONG_WIDTH     64
#define ___FLOAT_WIDTH    64
#define ___DOUBLE_WIDTH   64
#define ___VOIDSTAR_WIDTH 64
#endif

#ifndef ___BUILTIN_64BIT_INT_TYPE
#ifdef __GNUC__
#define ___BUILTIN_64BIT_INT_TYPE long long
#endif
#endif

#ifndef ___SCHAR
#define ___SCHAR signed char
#endif

/*
 * The following symbols must be defined to reflect the data types of
 * the C compiler.  A symbol is only defined if the data type exists.
 * At least ___F64 and one of the pairs ___S32/___U32 and ___S64/___U64
 * must be defined.
 *
 *  define ___S64 as the signed integer type of exactly 64 bits
 *  define ___U64 as the unsigned integer type of exactly 64 bits
 *  define ___SM64 as the signed integer type of at least 64 bits
 *  define ___UM64 as the unsigned integer type of at least 64 bits
 *
 *  define ___S32 as the signed integer type of exactly 32 bits
 *  define ___U32 as the unsigned integer type of exactly 32 bits
 *  define ___SM32 as the signed integer type of at least 32 bits
 *  define ___UM32 as the unsigned integer type of at least 32 bits
 *
 *  define ___S16 as the signed integer type of exactly 16 bits
 *  define ___U16 as the unsigned integer type of exactly 16 bits
 *  define ___SM16 as the signed integer type of at least 16 bits
 *  define ___UM16 as the unsigned integer type of at least 16 bits
 *
 *  define ___S8 as the signed integer type of exactly 8 bits
 *  define ___U8 as the unsigned integer type of exactly 8 bits
 *  define ___SM8 as the signed integer type of at least 8 bits
 *  define ___UM8 as the unsigned integer type of at least 8 bits
 */

#ifndef ___S64
#undef ___FORCE_32
#define ___FORCE_32
#endif

#if ___INT_WIDTH == 64
#define ___S64 int
#define ___U64 unsigned int
#else
#if ___LONG_WIDTH == 64
#define ___S64 long
#define ___U64 unsigned long
#else
#if ___SHORT_WIDTH == 64
#define ___S64 short
#define ___U64 unsigned short
#else
#if ___CHAR_WIDTH == 64
#define ___S64 ___SCHAR
#define ___U64 unsigned char
#else
#ifdef ___BUILTIN_64BIT_INT_TYPE
#define ___S64 ___BUILTIN_64BIT_INT_TYPE
#define ___U64 unsigned ___BUILTIN_64BIT_INT_TYPE
#endif
#endif
#endif
#endif
#endif

#if ___INT_WIDTH == 32
#define ___S32 int
#define ___U32 unsigned int
#else
#if ___LONG_WIDTH == 32
#define ___S32 long
#define ___U32 unsigned long
#else
#if ___SHORT_WIDTH == 32
#define ___S32 short
#define ___U32 unsigned short
#else
#if ___CHAR_WIDTH == 32
#define ___S32 ___SCHAR
#define ___U32 unsigned char
#endif
#endif
#endif
#endif

#if ___INT_WIDTH == 16
#define ___S16 int
#define ___U16 unsigned int
#else
#if ___LONG_WIDTH == 16
#define ___S16 long
#define ___U16 unsigned long
#else
#if ___SHORT_WIDTH == 16
#define ___S16 short
#define ___U16 unsigned short
#else
#if ___CHAR_WIDTH == 16
#define ___S16 ___SCHAR
#define ___U16 unsigned char
#endif
#endif
#endif
#endif

#if ___INT_WIDTH == 8
#define ___S8 int
#define ___U8 unsigned int
#else
#if ___LONG_WIDTH == 8
#define ___S8 long
#define ___U8 unsigned long
#else
#if ___SHORT_WIDTH == 8
#define ___S8 short
#define ___U8 unsigned short
#else
#if ___CHAR_WIDTH == 8
#define ___S8 ___SCHAR
#define ___U8 unsigned char
#endif
#endif
#endif
#endif

#if ___FLOAT_WIDTH == 64
#define ___F64 float
#else
#if ___DOUBLE_WIDTH == 64
#define ___F64 double
#endif
#endif

#if ___FLOAT_WIDTH == 32
#define ___F32 float
#else
#if ___DOUBLE_WIDTH == 32
#define ___F32 double
#endif
#endif

#ifdef ___S64
#define ___SM64 ___S64
#define ___UM64 ___U64
#endif

#ifdef ___S32
#define ___SM32 ___S32
#define ___UM32 ___U32
#else
#define ___SM32 ___SM64
#define ___UM32 ___UM64
#endif

#ifdef ___S16
#define ___SM16 ___S16
#define ___UM16 ___U16
#else
#define ___SM16 ___SM32
#define ___UM16 ___UM32
#endif

#ifdef ___S8
#define ___SM8 ___S8
#define ___UM8 ___U8
#else
#define ___SM8 ___SM16
#define ___UM8 ___UM16
#endif

/*---------------------------------------------------------------------------*/

#ifdef __GNUC__
#ifndef ___I64zzz
#define ___I64zzz long long
#endif
#endif



/*---------------------------------------------------------------------------*/

/* SANITY CHECKS AND SETTING OF DEFAULT OPTIONS */

/*
 * Verify that processor type is known.
 */

#ifdef ___CPU_BIGEND
#ifdef ___CPU_LITEND
@error "Either ___CPU_BIGEND or ___CPU_LITEND must be defined"
#endif
#else
#ifndef ___CPU_LITEND
@error "Either ___CPU_BIGEND or ___CPU_LITEND must be defined"
#endif
#endif

/*
 * Generate a single host procedure or multiple C host
 * procedures (default) per Scheme module?
 */

#ifdef ___SINGLE_HOST
#ifdef ___MULTIPLE_HOSTS
@error "Define either ___SINGLE_HOST or ___MULTIPLE_HOSTS"
#endif
#else
#ifndef ___MULTIPLE_HOSTS
#define ___MULTIPLE_HOSTS
#endif
#endif

/*
 * Compiling for dynamic loading or not (default).
 */

#ifdef ___DYNAMIC
#ifdef ___NON_DYNAMIC
@error "Define either ___DYNAMIC or ___NON_DYNAMIC"
#endif
#else
#ifndef ___NON_DYNAMIC
#define ___NON_DYNAMIC
#endif
#endif

/*
 * Define symbols appropriate for dynamic loading.
 */

#ifdef ___DYNAMIC
#undef ___PRIMAL
#define ___PRIMAL
#undef ___LIBRARY
#define ___LIBRARY
#undef ___SHARED
#define ___SHARED
#undef ___BIND_LATE
#define ___BIND_LATE
#endif

/*
 * Compiling to produce a library or an application with a "main" (default
 * unless compiling a flat link file).
 */

#ifdef ___LIBRARY
#ifdef ___APPLICATION
@error "Define either ___LIBRARY or ___APPLICATION"
#endif
#else
#ifndef ___APPLICATION
#ifdef ___FLAT_LINKFILE
#define ___LIBRARY
#else
#define ___APPLICATION
#endif
#endif
#endif

/*
 * Compiling to produce a shared-library or not (default).
 */

#ifdef ___SHARED
#ifdef ___NON_SHARED
@error "Define either ___SHARED or ___NON_SHARED"
#endif
#else
#ifndef ___NON_SHARED
#define ___NON_SHARED
#endif
#endif

/* 
 * Select binding time for global variables, symbols, and keywords.
 * Early binding (default) produces faster code because it directly
 * accesses the resource.  Late binding does an indirection at run
 * time.
 */

#ifdef ___BIND_LATE
#ifdef ___BIND_EARLY
@error "Define either ___BIND_LATE or ___BIND_EARLY"
#endif
#else
#ifndef ___BIND_EARLY
#define ___BIND_EARLY
#endif
#endif

/*
 * Size in bytes of characters in Scheme strings.
 *
 * ___CS must be 1, 2 or 4, which give respectively an 8 bit, 16
 * bit, or 29 bit subset of the ISO/IEC 10646-1:1993 character set.
 * The 8 bit subset corresponds to LATIN-1 (a.k.a. ISO/IEC 8859-1) and
 * the 16 bit subset corresponds to UCS-2 (a.k.a. Unicode).
 */

#ifndef ___CS
#define ___CS 2
#endif

/*
 * Default character encoding to use for I/O.
 *
 * ___DEFAULT_IO_ENCODING should be one of:
 *    ___IO_CHAR_ENCODING    use text mode and native encoding
 *    ___IO_LATIN1_ENCODING  use text mode and LATIN-1 encoding
 *    ___IO_UTF8_ENCODING    use text mode and UTF-8 encoding (1..6 bytes/char)
 *    ___IO_BYTE_ENCODING    use binary mode and LATIN-1 encoding (1 byte/char)
 *    ___IO_UCS2_ENCODING    use binary mode and UCS-2 encoding (2 bytes/char)
 *    ___IO_UCS4_ENCODING    use binary mode and UCS-4 encoding (4 bytes/char)
 */

#ifndef ___DEFAULT_IO_ENCODING
#define ___DEFAULT_IO_ENCODING ___IO_CHAR_ENCODING
#endif

/*
 * Number of registers in the virtual machine.  These definitions must
 * agree with those in the file "gsc/_t-c-1.scm".  All Scheme sources
 * must be recompiled if these definitions are changed.
 *
 * ___NB_GVM_REGS = total number of registers available
 *                  3 <= ___NB_GVM_REGS <= 25
 * ___NB_ARG_REGS = maximum number of arguments passed in registers
 *                  1 <= ___NB_ARG_REGS <= min( 12, ___NB_GVM_REGS-2 )
 */

#define ___NB_GVM_REGS 5
#define ___NB_ARG_REGS 3

/*
 * Determine if label values (a gcc extension to C) should be used.
 */

#ifdef ___USE_LABEL_VALUES
#ifdef ___NOT_USE_LABEL_VALUES
@error "Define either ___USE_LABEL_VALUES or ___NOT_USE_LABEL_VALUES"
#endif
#else
#ifndef ___NOT_USE_LABEL_VALUES
#ifdef __GNUC__
#define ___USE_LABEL_VALUES
#else
#define ___NOT_USE_LABEL_VALUES
#endif
#endif
#endif

/*
 * Determine optimization goal.
 */

#ifdef ___OPTIMIZE_TIME
#ifdef ___OPTIMIZE_SPACE
@error "Define either ___OPTIMIZE_TIME or ___OPTIMIZE_SPACE"
#endif
#else
#ifndef ___OPTIMIZE_SPACE
#define ___OPTIMIZE_TIME
#endif
#endif

#ifdef ___OPTIMIZE_TIME
#define ___USE_INLINE_JUMP
#endif

/*---------------------------------------------------------------------------*/

/* Dependencies on language (i.e. C++, ANSI-C, and K&R C). */

#ifdef __cplusplus

#define ___P(ansi,kr)ansi
#define ___PVOID ()

#define ___BEGIN_C_LINKAGE extern "C" {
#define ___END_C_LINKAGE }

#define ___BOOL int /* should be "bool" but some compilers don't support it */

#else

#ifdef __STDC__

#define ___P(ansi,kr)ansi
#define ___PVOID (void)

#else

#define ___P(ansi,kr)kr
#define ___PVOID ()

#endif

#define ___BEGIN_C_LINKAGE
#define ___END_C_LINKAGE

#define ___BOOL int

#define ___USE_SETJMP

#endif

/*---------------------------------------------------------------------------*/

/* Visibility directives. */

#ifndef ___HIDDEN
#define ___HIDDEN static
#endif

#ifndef ___LOCAL
#define ___LOCAL static
#endif

#ifndef ___EXPORT_FUNC
#define ___EXPORT_FUNC(type,name)type name
#endif

#ifndef ___EXPORT_DATA
#define ___EXPORT_DATA(type,name)type name
#endif

#ifndef ___IMPORT_FUNC
#define ___IMPORT_FUNC(type,name)extern type name
#endif

#ifndef ___IMPORT_DATA
#define ___IMPORT_DATA(type,name)extern type name
#endif

#define ___EXP_FUNC(type,name)type name
#define ___EXP_DATA(type,name)type name
#define ___IMP_FUNC(type,name)extern type name
#define ___IMP_DATA(type,name)extern type name

#ifdef ___SHARED
#ifdef ___LIBRARY
#undef ___EXP_FUNC
#define ___EXP_FUNC(type,name)___EXPORT_FUNC(type,name)
#undef ___EXP_DATA
#define ___EXP_DATA(type,name)___EXPORT_DATA(type,name)
#endif
#ifndef ___PRIMAL
#undef ___IMP_FUNC
#define ___IMP_FUNC(type,name)___IMPORT_FUNC(type,name)
#undef ___IMP_DATA
#define ___IMP_DATA(type,name)___IMPORT_DATA(type,name)
#endif
#endif

#ifdef ___CANT_IMPORT_EXPORTED
#ifdef ___SHARED
#ifndef ___PRIMAL
#undef ___BIND_LATE
#define ___BIND_LATE
#endif
#endif
#endif

/*---------------------------------------------------------------------------*/

/*
 * IMPORTANT NOTE
 *
 * As a general principle, the macros are written in such a way that all
 * macros which expand into C expressions yield PRIMARY C expressions.
 * Also, macros assume that arguments are PRIMARY expressions.  A PRIMARY
 * expression is either:
 *
 *   an identifier, or
 *   a non-negative constant, or
 *   a parenthesized expression.
 *
 * This convention avoids many operator priority problems and helps keep
 * the number of parentheses down.  On the other hand, it is easy to
 * introduce errors when modifying these macros so be careful.
 */

/*---------------------------------------------------------------------------*/

/*
 * Macros to access ___S64 and ___U64 data types.  Note that ___S64
 * and ___U64 are either defined as a macro or as a typedef.
 */

#define ___FETCH_S64(base,i)*((___S64*)(base)+(i))
#define ___STORE_S64(base,i,val)*((___S64*)(base)+(i)) = (val)
#define ___FETCH_U64(base,i)*((___U64*)(base)+(i))
#define ___STORE_U64(base,i,val)*((___U64*)(base)+(i)) = (val)

/*
 * Macros to access ___S32 and ___U32 data types.
 */

#ifdef ___S32

#define ___FETCH_S32(base,i)*((___S32*)(base)+(i))
#define ___STORE_S32(base,i,val)*((___S32*)(base)+(i)) = (val)
#define ___FETCH_U32(base,i)*((___U32*)(base)+(i))
#define ___STORE_U32(base,i,val)*((___U32*)(base)+(i)) = (val)
#define ___CAST_S32(x)(___S32)(x)
#define ___CAST_U32(x)(___U32)(x)

#else

#define ___FETCH_S32(base,i)___CAST_S32(___FETCH_U32(base,i))
#define ___STORE_S32(base,i,val)___STORE_U32(base,i,___CAST_U32(val))

#ifdef ___CPU_BIGEND
#define ___FETCH_U32(base,i)(((i)&1) \
? (*((___U64*)(base)+((i)>>1))&0xffffffff) \
: ((*((___U64*)(base)+((i)>>1)))>>32))
#define ___STORE_U32(base,i,val)*((___U64*)(base)+((i)>>1)) = ((i)&1) \
? (*((___U64*)(base)+((i)>>1))&((___U64)0xffffffff<<32))|(val) \
: (*((___U64*)(base)+((i)>>1))&0xffffffff)|((___U64)(val)<<32)
#else
#define ___FETCH_U32(base,i)(((i)&1) \
? ((*((___U64*)(base)+((i)>>1)))>>32) \
: (*((___U64*)(base)+((i)>>1))&0xffffffff))
#define ___STORE_U32(base,i,val)*((___U64*)(base)+((i)>>1)) = ((i)&1) \
? (*((___U64*)(base)+((i)>>1))&0xffffffff)|((___U64)(val)<<32) \
: (*((___U64*)(base)+((i)>>1))&((___U64)0xffffffff<<32))|(val)
#endif

#define ___CAST_S32(x)(((___S64)(x)<<32)>>32)
#define ___CAST_U32(x)((___U64)(x)&0xffffffff)

#endif

/*
 * Macros to access ___S16 and ___U16 data types.
 */

#ifdef ___S16

#define ___FETCH_S16(base,i)*((___S16*)(base)+(i))
#define ___STORE_S16(base,i,val)*((___S16*)(base)+(i)) = (val)
#define ___FETCH_U16(base,i)*((___U16*)(base)+(i))
#define ___STORE_U16(base,i,val)*((___U16*)(base)+(i)) = (val)
#define ___CAST_S16(x)(___S16)(x)
#define ___CAST_U16(x)(___U16)(x)

#else

#define ___FETCH_S16(base,i)___CAST_S16(___FETCH_U16(base,i))
#define ___STORE_S16(base,i,val)___STORE_U16(base,i,___CAST_U16(val))

#ifdef ___S32

#ifdef ___CPU_BIGEND
#define ___FETCH_U16(base,i)(((i)&1) \
? (*((___U32*)(base)+((i)>>1))&0xffff) \
: ((*((___U32*)(base)+((i)>>1)))>>16))
#define ___STORE_U16(base,i,val)*((___U32*)(base)+((i)>>1)) = ((i)&1) \
? (*((___U32*)(base)+((i)>>1))&((___U32)0xffff<<16))|(val) \
: (*((___U32*)(base)+((i)>>1))&0xffff)|((___U32)(val)<<16)
#else
#define ___FETCH_U16(base,i)(((i)&1) \
? ((*((___U32*)(base)+((i)>>1)))>>16) \
: (*((___U32*)(base)+((i)>>1))&0xffff))
#define ___STORE_U16(base,i,val)*((___U32*)(base)+((i)>>1)) = ((i)&1) \
? (*((___U32*)(base)+((i)>>1))&0xffff)|((___U32)(val)<<16) \
: (*((___U32*)(base)+((i)>>1))&(0xffff<<16))|(___U32)(val)
#endif

#define ___CAST_S16(x)(((___S32)(x)<<16)>>16)
#define ___CAST_U16(x)((___U32)(x)&0xffff)

#else

#ifdef ___CPU_BIGEND
#define ___FETCH_S16(base,i)(((i)&1) \
? ((*((___S64*)(base)+((i)>>1)))>>16) \
: ((*((___S64*)(base)+((i)>>1))<<16)>>16))
#define ___FETCH_U16(base,i) \
((*((___U64*)(base)+((i)>>2))>>((~(i)&3)<<4))&0xffff)
#define ___STORE_U16(base,i,val)*((___U64*)(base)+((i)>>2)) = \
(*((___U64*)(base)+((i)>>2))&~((___U64)0xffff<<((~(i)&3)<<4))) | \
((___U64)(val)<<((~(i)&3)<<4))
#else
#define ___FETCH_U16(base,i) \
((*((___U64*)(base)+((i)>>2))>>(((i)&3)<<4))&0xffff)
#define ___STORE_U16(base,i,val)*((___U64*)(base)+((i)>>2)) = \
(*((___U64*)(base)+((i)>>2))&~((___U64)0xffff<<(((i)&3)<<4))) | \
((___U64)(val)<<(((i)&3)<<4))
#endif

#define ___CAST_S16(x)(((___S64)(x)<<48)>>48)
#define ___CAST_U16(x)((___U64)(x)&0xffff)

#endif

#endif

/*
 * Macros to access ___S8 and ___U8 data types.
 */

#ifdef ___S8

#define ___FETCH_S8(base,i)*((___S8*)(base)+(i))
#define ___FETCH_U8(base,i)*((___U8*)(base)+(i))
#define ___STORE_S8(base,i,val)*((___S8*)(base)+(i)) = (val)
#define ___STORE_U8(base,i,val)*((___U8*)(base)+(i)) = (val)
#define ___CAST_S8(x)(___S8)(x)
#define ___CAST_U8(x)(___U8)(x)

#else

#define ___FETCH_S8(base,i)___CAST_S8(___FETCH_U8(base,i))
#define ___STORE_S8(base,i,val)___STORE_U8(base,i,___CAST_U8(val))

#ifdef ___S32

#ifdef ___CPU_BIGEND
#define ___FETCH_U8(base,i) \
((*((___U32*)(base)+((i)>>2))>>((~(i)&3)<<3))&0xff)
#define ___STORE_U8(base,i,val)*((___U32*)(base)+((i)>>2)) = \
(*((___U32*)(base)+((i)>>2))&~((___U32)0xff<<((~(i)&3)<<3))) | \
((___U32)(val)<<((~(i)&3)<<3))
#else
#define ___FETCH_U8(base,i) \
((*((___U32*)(base)+((i)>>2))>>(((i)&3)<<3))&0xff)
#define ___STORE_U8(base,i,val)*((___U32*)(base)+((i)>>2)) = \
(*((___U32*)(base)+((i)>>2))&~((___U32)0xff<<(((i)&3)<<3))) | \
((___U32)(val)<<(((i)&3)<<3))
#endif

#define ___CAST_S8(x)(((___S32)(x)<<24)>>24)
#define ___CAST_U8(x)((___U32)(x)&0xff)

#else

#ifdef ___CPU_BIGEND
#define ___FETCH_U8(base,i) \
((*((___U64*)(base)+((i)>>3))>>((~(i)&7)<<3))&0xff)
#define ___STORE_U8(base,i,val)*((___U64*)(base)+((i)>>3)) = \
(*((___U64*)(base)+((i)>>3))&~((___U64)0xff<<((~(i)&7)<<3))) | \
((___U64)(val)<<((~(i)&7)<<3))
#else
#define ___FETCH_U8(base,i) \
((*((___U64*)(base)+((i)>>3))>>(((i)&7)<<3))&0xff)
#define ___STORE_U8(base,i,val)*((___U64*)(base)+((i)>>3)) = \
(*((___U64*)(base)+((i)>>3))&~((___U64)0xff<<(((i)&7)<<3))) | \
((___U64)(val)<<(((i)&7)<<3))
#endif

#define ___CAST_S8(x)(((___S64)(x)<<56)>>56)
#define ___CAST_U8(x)((___U64)(x)&0xff)

#endif

#endif

/*---------------------------------------------------------------------------*/

/*
 * GENERAL DEFINITIONS
 *
 * ___U64   an unsigned integer type of exactly 64 bits (only if one exists)
 * ___UM64  an unsigned integer type of at least 64 bits (only if one exists)
 * ___U32   an unsigned integer type of exactly 32 bits (only if one exists)
 * ___UM32  an unsigned integer type of at least 32 bits
 * ___U16   an unsigned integer type of exactly 16 bits (only if one exists)
 * ___UM16  an unsigned integer type of at least 16 bits
 * ___U8    an unsigned integer type of exactly 8 bits
 * ___UM8   an unsigned integer type of at least 8 bits
 * ___WORD  an integer type able to contain a pointer with no loss
 * ___LWS   is equal to the base 2 logarithm of "sizeof(___WORD)"
 */

#ifdef ___FORCE_32
#define ___WORD ___S32
#define ___LWS 2
#define ___WS 4
#define ___WORD_WIDTH 32
#define ___FLONUM_SIZE 2
#else
#define ___WORD ___S64
#define ___LWS 3
#define ___WS 8
#define ___WORD_WIDTH 64
#define ___FLONUM_SIZE 1
#endif

#define ___SCMOBJ ___WORD
#define ___SCMOBJ_WIDTH ___WORD_WIDTH

#define ___LATIN1 ___UM8
#define ___UCS4 ___UM32
#define ___UCS2 ___UM16
#define ___LATIN1STRING ___LATIN1*
#define ___UCS4STRING ___UCS4*
#define ___UCS2STRING ___UCS2*
#define ___UTF8STRING char*

#if ___CS == 1
#define ___CS_SELECT(cs1,cs2,cs4)cs1
#else
#if ___CS == 2
#define ___CS_SELECT(cs1,cs2,cs4)cs2
#else
#define ___CS_SELECT(cs1,cs2,cs4)cs4
#endif
#endif

#define ___LCS ___CS_SELECT(0,1,2)

#ifdef ___WORD_IS_SMALLER_THAN_POINTERS
#define ___FAKEWORD ___WORD
#define ___CAST_FAKEWORD_TO_WORD(x)x
#define ___CAST_WORDSTAR_TO_FAKEWORD(x)((___WORD)x)
#define ___FAKEHOST ___WORD
#define ___CAST_HOST_TO_FAKEHOST(x)((___WORD)x)
#define ___CAST_FAKEHOST_TO_HOST(x)((___host)x)
#else
#define ___FAKEWORD ___WORD*
#define ___CAST_FAKEWORD_TO_WORD(x)((___WORD)x)
#define ___CAST_WORDSTAR_TO_FAKEWORD(x)x
#define ___FAKEHOST ___host
#define ___CAST_HOST_TO_FAKEHOST(x)x
#define ___CAST_FAKEHOST_TO_HOST(x)x
#endif

/* Padding at end of objects so that they can be aligned to an address */
/* multiple of 4 or 8 by moving them up in memory.                     */

#ifdef ___FORCE_32
#define ___PADDING ,0
#define ___PADDING_LBL ,{0,0,0,0}
#define ___PADDING_64 ,0,0
#else
#define ___PADDING
#define ___PADDING_LBL
#define ___PADDING_64
#endif

/* Utility macro to select code for MULTIPLE/SINGLE host. */

#ifdef ___SINGLE_HOST
#define ___SM(s,m)s
#else
#define ___SM(s,m)m
#endif

/*---------------------------------------------------------------------------*/

/* OBJECT REPRESENTATION */

/*
 * For a complete description of the object representation, read the
 * file "lib/mem.c".  These definitions must agree with those in the
 * file "lib/header.scm".
 */

/*
 * Type tag assignment
 *
 * Type tags are located in the lower 2 bits of a ___WORD.
 *
 * ___TB = number of tag bits
 * ___tFIXNUM   = tag for fixnums (small integers), must be 0
 * ___tSPECIAL  = tag for other immediates (#f, #t, (), chars, ...)
 * ___tPAIR     = tag for pairs
 * ___tSUBTYPED = tag for other memory allocated objects
 */

#define ___TB 2
#define ___tFIXNUM   0
#define ___tSPECIAL  2
#define ___tPAIR     3
#define ___tSUBTYPED 1

#define ___MEM_ALLOCATED(obj)(obj&1)

/*
 * Subtype tag assignment
 *
 * These tags are stored in the head of memory allocated objects
 * (including pairs).
 *
 * ___SB = number of subtype tag bits
 * ___sVECTOR       = tag for vectors
 * ___sPAIR         = tag for pairs
 * ___sRATNUM       = tag for ratnums
 * ___sCPXNUM       = tag for cpxnums
 * ___sSTRUCTURE    = tag for structures
 * ___sMEROON       = tag for Meroon objects
 * ___sSYMBOL       = tag for symbols
 * ___sKEYWORD      = tag for keywords
 * ___sFRAME        = tag for continuation frames
 * ___sCONTINUATION = tag for continuation descriptors
 * ___sPROMISE      = tag for promises
 * ___sWILL         = tag for wills
 * ___sPROCEDURE    = tag for procedures
 * ___sRETURN       = tag for returns
 * ___sPOINTER      = tag for C pointers
 * ___sSTRING       = tag for strings
 * ___sS8VECTOR     = tag for 8-bit signed integer vectors
 * ___sU8VECTOR     = tag for 8-bit unsigned integer vectors
 * ___sS16VECTOR    = tag for 16-bit signed integer vectors
 * ___sU16VECTOR    = tag for 16-bit unsigned integer vectors
 * ___sS32VECTOR    = tag for 32-bit signed integer vectors
 * ___sU32VECTOR    = tag for 32-bit unsigned integer vectors
 * ___sF32VECTOR    = tag for 32-bit floating point number vectors
 * ___sS64VECTOR    = tag for 64-bit signed integer vectors
 * ___sU64VECTOR    = tag for 64-bit unsigned integer vectors
 * ___sF64VECTOR    = tag for 64-bit floating point number vectors
 * ___sFLONUM       = tag for flonums
 * ___sBIGNUM       = tag for bignums
 *
 * The five subtype tags ___sS64VECTOR, ___sU64VECTOR, ___sF64VECTOR,
 * ___sFLONUM, ___sBIGNUM must come last.  These object types are aligned on a
 * multiple of 8.
 */

#define ___SB 5
#define ___sVECTOR       0
#define ___sPAIR         1
#define ___sRATNUM       2
#define ___sCPXNUM       3
#define ___sSTRUCTURE    4
#define ___sMEROON       6
#define ___sSYMBOL       8
#define ___sKEYWORD      9
#define ___sFRAME        10
#define ___sCONTINUATION 11
#define ___sPROMISE      12
#define ___sWILL         13
#define ___sPROCEDURE    14
#define ___sRETURN       15
#define ___sPOINTER      16
#define ___sSTRING       18
#define ___sS8VECTOR     20
#define ___sU8VECTOR     21
#define ___sS16VECTOR    22
#define ___sU16VECTOR    23
#define ___sS32VECTOR    24
#define ___sU32VECTOR    25
#define ___sF32VECTOR    26
#define ___sS64VECTOR    27
#define ___sU64VECTOR    28
#define ___sF64VECTOR    29
#define ___sFLONUM       30
#define ___sBIGNUM       31

/*
 * Head type tag assignment
 *
 * Head type tags are stored in the lower 3 bits of the head of memory
 * allocated objects.
 *
 * ___HTB = number of head tag bits
 * ___MOVABLE0 = tag for movable objects in generation 0
 * ___FORW     = tag for movable objects that have been forwarded
 * ___STILL    = tag for still objects
 * ___PERM     = tag for permanent objects
 *
 * note: the tag ___FORW+(1<<___TB) is also used
 */

#define ___HTB 3
#define ___MOVABLE0 0
#define ___STILL    1
#define ___FORW     3
#define ___PERM     6

/*
 * Miscellaneous masks.
 */

#define ___LF (___HTB+___SB)

#define ___TMASK ((1<<___TB)-1)
#define ___HTMASK ((1<<___HTB)-1)
#define ___SMASK (((1<<___SB)-1)<<___HTB)
#define ___LMASK (~(unsigned ___WORD)0<<___LF)
#define ___LSMASK (~(unsigned ___WORD)0<<___HTB)

/*
 * Value constructors for object references inside executable code.
 */

#define ___FAL (((___WORD)-1<<___TB)+___tSPECIAL)
#define ___TRU (((___WORD)-2<<___TB)+___tSPECIAL)
#define ___NUL (((___WORD)-3<<___TB)+___tSPECIAL)
#define ___EOF (((___WORD)-4<<___TB)+___tSPECIAL)
#define ___VOID (((___WORD)-5<<___TB)+___tSPECIAL)
#define ___ABSENT (((___WORD)-6<<___TB)+___tSPECIAL)
#define ___UNB1 (((___WORD)-7<<___TB)+___tSPECIAL)
#define ___UNB2 (((___WORD)-8<<___TB)+___tSPECIAL)
#define ___SCRIPT (((___WORD)-10<<___TB)+___tSPECIAL)
#define ___OPTIONAL (((___WORD)-11<<___TB)+___tSPECIAL)
#define ___KEY_OBJ (((___WORD)-12<<___TB)+___tSPECIAL)
#define ___REST (((___WORD)-13<<___TB)+___tSPECIAL)
#define ___FIX(x)((___WORD)(x)<<___TB)
#if ___WS == 4
#define ___BIGFIX(i,x)___SUB(i)
#else
#define ___BIGFIX(i,x)___FIX(x)
#endif
#define ___CHR(x)(((___WORD)___CS_SELECT(___CAST_U8(x),___CAST_U16(x),___CAST_U32(x))<<___TB)+___tSPECIAL)

#define ___SYM(i,id)___CAST_FAKEWORD_TO_WORD(___sym_tbl[i])
#define ___KEY(i,id)___CAST_FAKEWORD_TO_WORD(___key_tbl[i])
#define ___CNS(i)___TAG((___ALIGNUP(___cns_tbl,___WS)+i*(___PAIR_SIZE+1)),___tPAIR)
#define ___SUB(i)___CAST_FAKEWORD_TO_WORD(___sub_tbl[i])

/*
 * Value constructors for object references inside constant
 * memory allocated objects.
 */

#define ___REF_FAL ___FAL
#define ___REF_TRU ___TRU
#define ___REF_NUL ___NUL
#define ___REF_EOF ___EOF
#define ___REF_VOID ___VOID
#define ___REF_ABSENT ___ABSENT
#define ___REF_SCRIPT ___SCRIPT
#define ___REF_OPTIONAL ___OPTIONAL
#define ___REF_KEY_OBJ ___KEY_OBJ
#define ___REF_REST ___REST
#define ___REF_FIX(x)___FIX(x)
#if ___WS == 4
#define ___REF_BIGFIX(i,x)___REF_SUB(i)
#else
#define ___REF_BIGFIX(i,x)___REF_FIX(x)
#endif
#define ___REF_CHR(x)___CHR(x)

#define ___REF_SYM(i,id)(((___WORD)(-1-i)<<___TB)+___tPAIR)
#define ___REF_KEY(i,id)(((___WORD)(-1-i)<<___TB)+___tSUBTYPED)
#define ___REF_CNS(i)(((___WORD)i<<___TB)+___tPAIR)
#define ___REF_SUB(i)(((___WORD)i<<___TB)+___tSUBTYPED)

/*---------------------------------------------------------------------------*/

/* Miscellaneous macros */

#define ___NOTHING

#define ___INT(x)((x)>>___TB)

#define ___S32UNBOX(x) \
(___TYP((___temp=x)) == ___tFIXNUM \
? ___INT(___temp) \
: ___FETCH_S32(___BODY_AS(___temp,___tSUBTYPED),0)) /******************/

#define ___U32UNBOX(x) \
(___TYP((___temp=x)) == ___tFIXNUM \
? ___INT(___temp) \
: ___FETCH_U32(___BODY_AS(___temp,___tSUBTYPED),0)) /******************/

#define ___S64UNBOX(x)___EXT(___S64_from_SM32_UM32_fn) (0,0)/************************/

#define ___U64UNBOX(x)___EXT(___U64_from_UM32_UM32_fn) (0,0)/************************/

#define ___MIN_FIX (-((___WORD)1<<((8*___WS)-___TB-1)))
#define ___MAX_FIX (((___WORD)1<<((8*___WS)-___TB-1))-1)

#define ___MAX_FIX32 (((___WORD)1<<(32-___TB-1))-1)

#if ___CS == 4
#define ___MAX_CHR ___MAX_FIX32
#else
#define ___MAX_CHR (((___WORD)1<<(8*___CS))-1)
#endif

/*
 * Module prefix and C id prefix must match the definitions
 * in the file "gsc/_parms.scm".
 */

#define ___MODULE_PREFIX " "
#define ___C_ID_PREFIX "___"

/*---------------------------------------------------------------------------*/

#define ___ALIGNUP(x,mult)((___WORD*)(((___WORD)x+mult-1)&(-mult)))

/* Type tests */

#define ___TAG(ptr,tag)((___WORD)ptr+tag)
#define ___UNTAG(obj)((___WORD*)(obj&-(1<<___TB)))
#define ___UNTAG_AS(obj,tag)((___WORD*)(obj-tag))

#define ___TYP(x)(x&___TMASK)
#define ___TESTTYPE(x,typ)(___TYP(x)==typ)

#define ___TESTSUBTYPETAG(x,s)((___HEADER(x)&___SMASK)==(s<<___HTB))
#define ___TESTSUBTYPE(x,s)(___TYP((___temp=x))==___tSUBTYPED&&___TESTSUBTYPETAG(___temp,s))
#define ___TESTLENGTHSUBTYPETAG(x,bytes,s)((___HEADER(x)&___LSMASK)==___MAKE_HD(bytes,s,0))
#define ___TESTLENGTHSUBTYPE(x,bytes,s)(___TYP((___temp=x))==___tSUBTYPED&&___TESTSUBTYPETAG(___temp,s))
#define ___HEADER(x)(*___UNTAG_AS(x,___tSUBTYPED))

#define ___BODY_OFS 1

#ifdef ___USE_HANDLES
#define ___PAIR_OVERHEAD 2
#define ___SUBTYPED_OVERHEAD 2
#define ___BODY(obj)((___WORD*)(*(___UNTAG(obj)+___BODY_OFS)+___BODY_OFS))
#define ___BODY_AS(obj,tag)((___WORD*)*(___UNTAG_AS(obj,tag)+___BODY_OFS)+___BODY_OFS)
#else
#define ___PAIR_OVERHEAD 1
#define ___SUBTYPED_OVERHEAD 1
#define ___BODY(obj)(___UNTAG(obj)+___BODY_OFS)
#define ___BODY_AS(obj,tag)(___UNTAG_AS(obj,tag)+___BODY_OFS)
#endif

#define ___MAKE_HD(bytes,subtype,tag)(((bytes)<<___LF)+((subtype)<<___HTB)+tag)
#define ___HD_TYP(head)((head)&___HTMASK)
#define ___HD_BYTES(head)((unsigned ___WORD)(head)>>___LF)
#define ___HD_WORDS(head)((((unsigned ___WORD)(head)+((___WS-1)<<___LF)))>>(___LF+___LWS))
#define ___HD_FIELDS(head)((unsigned ___WORD)(head)>>(___LF+___LWS))
#define ___HD_SUBTYPE(head)((head)>>___HTB&((1<<___SB)-1))

#define ___FIELD(obj,i)(*(___BODY_AS(obj,___tSUBTYPED)+i))

#define ___WORDS(bytes)((bytes+___WS-1)>>___LWS)

#define ___MAKE_HD_BYTES(bytes,subtype)___MAKE_HD(bytes,subtype,___MOVABLE0)
#define ___MAKE_HD_WORDS(words,subtype)___MAKE_HD((words<<___LWS),subtype,___MOVABLE0)

#define ___SYMBOL_SIZE 4
#define ___KEYWORD_SIZE 3
#define ___THREAD_SIZE 24
#define ___WILL1_SIZE 2
#define ___WILL2_SIZE 3
#define ___PROMISE_SIZE 2
#define ___RATNUM_SIZE 2
#define ___CPXNUM_SIZE 2
#define ___CONTINUATION_SIZE 2
#define ___INTRO_SIZE 3

#if ___WORD_WIDTH == 32

#define ___S32BOX(x) \
(___temp=(x), \
(___temp >= ___MIN_FIX && ___temp <= ___MAX_FIX \
? ___FIX(___temp) \
: (___hp[0]=___MAKE_HD_BYTES((2<<2),___sBIGNUM), \
  ___STORE_U32(___hp+1,0,___CAST(___UM32,___temp)), /**************/ \
  ___STORE_U32(___hp+1,1,___CAST(___UM32,(___temp<0?-1:0))), /**************/ \
  ___hp+=___WORDS((2<<2))+1, \
  ___TAG((___hp-___WORDS((2<<2))-1),___tSUBTYPED))))

#define ___U32BOX(x) \
(___temp=(x), \
(___CAST(___UM32,___temp) <= ___CAST(___UM32,___MAX_FIX) \
? ___FIX(___temp) \
: (___hp[0]=___MAKE_HD_BYTES((2<<2),___sBIGNUM), \
  ___STORE_U32(___hp+1,0,___CAST(___UM32,___temp)), /**************/ \
  ___STORE_U32(___hp+1,1,___CAST(___UM32,0)), /**************/ \
  ___hp+=___WORDS((2<<2))+1, \
  ___TAG((___hp-___WORDS((2<<2))-1),___tSUBTYPED))))

#else

#define ___S32BOX(x)___FIX(x)
#define ___U32BOX(x)___FIX(x)

#endif

#define ___S64BOX(x)___FIX(0) /*********fix me**************/
#define ___U64BOX(x)___FIX(0) /*********fix me**************/

/* Flonum boxing and unboxing */

#define ___FLONUM_VAL(obj)(*(___F64*)(___BODY_AS(obj,___tSUBTYPED)))

#define ___FLONUM_SETUP(ptr,x) \
(*ptr=___MAKE_HD_WORDS(___FLONUM_SIZE,___sFLONUM),*(___F64*)(ptr+1)=x, \
___TAG(ptr,___tSUBTYPED))

#if ___WS == 4

#define ___F64BOX(x) \
(___hp+=___FLONUM_SIZE+2,((___WORD)___hp&7)==0 \
?___FLONUM_SETUP((___hp-___FLONUM_SIZE-1),x) \
:___FLONUM_SETUP((___hp-___FLONUM_SIZE-2),x))

#else

#define ___F64BOX(x) \
(___hp+=___FLONUM_SIZE+1,___FLONUM_SETUP((___hp-___FLONUM_SIZE-1),x))

#endif

#define ___F64UNBOX(x)___FLONUM_VAL(x)

#define ___D_F64(x)___F64 x;
#define ___SET_F64(x,y)x=y;
#define ___IF(x)if (x) {
#define ___END_IF }

     /* NEW BIGNUM OPS */

#ifdef ___BUILTIN_64BIT_INT_TYPE
#define ___BIG_ABASE_WIDTH 64
#define ___BIGAFETCH ___FETCH_U64
#define ___BIGASTORE ___STORE_U64
#define ___BIGALENGTH ___U64VECTORLENGTH
#define ___BIGADIGIT ___U64
#define ___BIGADIGITSIGNED ___S64
#define ___BIGASHRINK ___U64VECTORSHRINK
#define ___BIGATEMP ___temp64
#else
#define ___BIG_ABASE_WIDTH 32
#define ___BIGAFETCH ___FETCH_U32
#define ___BIGASTORE ___STORE_U32
#define ___BIGALENGTH ___U32VECTORLENGTH
#define ___BIGADIGIT ___U32
#define ___BIGADIGITSIGNED ___S32
#define ___BIGASHRINK ___U32VECTORSHRINK
#define ___BIGATEMP ___temp
#endif

#if ___WORD_WIDTH == 64
#define ___BIG_MBASE_WIDTH 32
#else
#define ___BIG_MBASE_WIDTH 16
#endif

#define ___BIG_FBASE_WIDTH 8

#define ___BIG_ABASE_DIV_2 (___CAST(___BIGADIGIT,1) << (___BIG_ABASE_WIDTH-1))
#define ___BIG_ABASE_MIN_1 (~___CAST(___BIGADIGIT,0))

#if ___WS==4 && ___BIG_ABASE_WIDTH == 64

#define ___BIGFROMFIX(x) \
  (___hp = (___WORD *)((___WORD) (___hp +4) &(~7)), \
  ___hp[-3] = ___MAKE_HD_BYTES((1<<3),___sBIGNUM), \
  *(___BIGADIGIT *)(___hp-2) = (___BIGADIGIT)((___BIGADIGITSIGNED)___INT(x)), \
  ___TAG((___hp-3),___tSUBTYPED))

#else

#define ___BIGFROMFIX(x) \
  (___hp[0]=___MAKE_HD_BYTES((1 << (___BIG_ABASE_WIDTH == 64 ? 3 : 2)),___sBIGNUM), \
  ___BIGASTORE(___hp+1,0,___INT(x)), \
  ___hp+=2, \
  ___TAG((___hp-2),___tSUBTYPED))

#endif


/* dummy definitions... Brad will complete and rewrite most of this section */

#define ___BIGALESSP(x,y,i) \
(___BIGAFETCH(___BODY_AS(x,___tSUBTYPED),(i)>>___TB) \
<(___BIGAFETCH(___BODY_AS(y,___tSUBTYPED),(i)>>___TB)))

#define ___BIGAEQP(x,y,i) \
(___BIGAFETCH(___BODY_AS(x,___tSUBTYPED),(i)>>___TB) \
==(___BIGAFETCH(___BODY_AS(y,___tSUBTYPED),(i)>>___TB)))

#define ___BIGAONESP(x,i) \
(___BIGAFETCH(___BODY_AS(x,___tSUBTYPED),(i)>>___TB) == ___BIG_ABASE_MIN_1)

#define ___BIGAZEROP(x,i) \
(___BIGAFETCH(___BODY_AS(x,___tSUBTYPED),(i)>>___TB) == 0)

#define ___BIGANEGATIVEP(x,i) \
(___BIGAFETCH(___BODY_AS(x,___tSUBTYPED),(i)>>___TB) & ___BIG_ABASE_DIV_2)

#define ___BIGTOFIX(x) \
(___BIGATEMP=___BIGAFETCH(___BODY_AS(x,___tSUBTYPED),0), \
(___BIGATEMP == ___INT(___FIX(___BIGATEMP)) ? ___FIX(___BIGATEMP) : ___FAL))


#define ___BIGACAT(x,i,hi,j,lo,k,divider) \
___BIGATEMP=((___BIGADIGIT)___BIGAFETCH(___BODY_AS(hi,___tSUBTYPED),(j)>>___TB) << ___INT(divider)) | \
            ((___BIGADIGIT)___BIGAFETCH(___BODY_AS(lo,___tSUBTYPED),(k)>>___TB) >> (___BIG_ABASE_WIDTH - ___INT(divider))); \
___BIGASTORE(___BODY_AS(x,___tSUBTYPED),(i)>>___TB,(___BIGADIGIT)___BIGATEMP);

#define ___BIGAAND(x,i,y,j) \
___BIGATEMP= ___BIGAFETCH(___BODY_AS(x,___tSUBTYPED),(i)>>___TB) & ___BIGAFETCH(___BODY_AS(y,___tSUBTYPED),(j)>>___TB); \
___BIGASTORE(___BODY_AS(x,___tSUBTYPED),(i)>>___TB,(___BIGADIGIT)___BIGATEMP);

#define ___BIGAIOR(x,i,y,j) \
___BIGATEMP= ___BIGAFETCH(___BODY_AS(x,___tSUBTYPED),(i)>>___TB) | ___BIGAFETCH(___BODY_AS(y,___tSUBTYPED),(j)>>___TB); \
___BIGASTORE(___BODY_AS(x,___tSUBTYPED),(i)>>___TB,(___BIGADIGIT)___BIGATEMP);

#define ___BIGAXOR(x,i,y,j) \
___BIGATEMP= ___BIGAFETCH(___BODY_AS(x,___tSUBTYPED),(i)>>___TB) ^ ___BIGAFETCH(___BODY_AS(y,___tSUBTYPED),(j)>>___TB); \
___BIGASTORE(___BODY_AS(x,___tSUBTYPED),(i)>>___TB,(___BIGADIGIT)___BIGATEMP);

#define ___BIGANOT(x,i) \
___BIGATEMP= ~___BIGAFETCH(___BODY_AS(x,___tSUBTYPED),(i)>>___TB); \
___BIGASTORE(___BODY_AS(x,___tSUBTYPED),(i)>>___TB,(___BIGADIGIT)___BIGATEMP);



#define ___BIGNEGATIVEP(x) \
(___BIGAFETCH(___BODY_AS(x,___tSUBTYPED),___FIXSUB(___BIGALENGTH(x),___FIX_1)>>___TB) & ___BIG_ABASE_DIV_2)

#define ___BIGASET(result,i,y,j) \
___BIGASTORE(___BODY_AS(result,___tSUBTYPED),(i)>>___TB, \
___BIGAFETCH(___BODY_AS(y,___tSUBTYPED),(j)>>___TB));

#define ___BIGASETONES(result,i) \
___BIGASTORE(___BODY_AS(result,___tSUBTYPED),(i)>>___TB,___BIG_ABASE_MIN_1);

#define ___BIGAINC(x,i) \
(___BIGATEMP = (___BIGADIGIT)___BIGAFETCH(___BODY_AS(x,___tSUBTYPED),(i)>>___TB)+(___BIGADIGIT) 1, \
 ___BIGASTORE(___BODY_AS(x,___tSUBTYPED),(i)>>___TB,(___BIGADIGIT)___BIGATEMP), \
(___BIGATEMP == 0 ? ___FIX_1 : ___FIX_0))

#define ___BIGADEC(x,i) \
(___BIGATEMP = (___BIGADIGIT)___BIGAFETCH(___BODY_AS(x,___tSUBTYPED),(i)>>___TB)-(___BIGADIGIT) 1, \
 ___BIGASTORE(___BODY_AS(x,___tSUBTYPED),(i)>>___TB,(___BIGADIGIT)___BIGATEMP), \
(___BIGATEMP == ___BIG_ABASE_MIN_1 ? ___FIX_1 : ___FIX_0))

#define ___BIGAADD(result,i,y,j,carry) \
(carry ? (___BIGATEMP=(___BIGADIGIT)___BIGAFETCH(___BODY_AS(result,___tSUBTYPED),___INT(i))+ \
	          (___BIGADIGIT)___BIGAFETCH(___BODY_AS(     y,___tSUBTYPED),___INT(j))+ \
	          (___BIGADIGIT) 1, \
          ((___BIGADIGIT) ___BIGATEMP <= (___BIGADIGIT)___BIGAFETCH(___BODY_AS(result,___tSUBTYPED),___INT(i)) \
           ? \
	   (___BIGASTORE(___BODY_AS(result,___tSUBTYPED),___INT(i),(___BIGADIGIT)___BIGATEMP), ___FIX_1) \
           : \
	   (___BIGASTORE(___BODY_AS(result,___tSUBTYPED),___INT(i),(___BIGADIGIT)___BIGATEMP), ___FIX_0))) \
         : \
         (___BIGATEMP=(___BIGADIGIT)___BIGAFETCH(___BODY_AS(result,___tSUBTYPED),___INT(i))+ \
	          (___BIGADIGIT)___BIGAFETCH(___BODY_AS(     y,___tSUBTYPED),___INT(j)), \
          ((___BIGADIGIT) ___BIGATEMP < (___BIGADIGIT)___BIGAFETCH(___BODY_AS(result,___tSUBTYPED),___INT(i)) \
           ? \
	   (___BIGASTORE(___BODY_AS(result,___tSUBTYPED),___INT(i),(___BIGADIGIT)___BIGATEMP), ___FIX_1) \
           : \
	   (___BIGASTORE(___BODY_AS(result,___tSUBTYPED),___INT(i),(___BIGADIGIT)___BIGATEMP), ___FIX_0))))

#define ___BIGASUB(result,i,y,j,borrow) \
(borrow ? (___BIGATEMP=(___BIGADIGIT)___BIGAFETCH(___BODY_AS(result,___tSUBTYPED),___INT(i))- \
	           (___BIGADIGIT)___BIGAFETCH(___BODY_AS(     y,___tSUBTYPED),___INT(j))- \
	           (___BIGADIGIT) 1, \
           ((___BIGADIGIT) ___BIGATEMP >= (___BIGADIGIT)___BIGAFETCH(___BODY_AS(result,___tSUBTYPED),___INT(i)) \
            ? \
	    (___BIGASTORE(___BODY_AS(result,___tSUBTYPED),___INT(i),(___BIGADIGIT)___BIGATEMP), ___FIX_1) \
            : \
	    (___BIGASTORE(___BODY_AS(result,___tSUBTYPED),___INT(i),(___BIGADIGIT)___BIGATEMP), ___FIX_0))) \
          : \
          (___BIGATEMP=(___BIGADIGIT)___BIGAFETCH(___BODY_AS(result,___tSUBTYPED),___INT(i))- \
	           (___BIGADIGIT)___BIGAFETCH(___BODY_AS(     y,___tSUBTYPED),___INT(j)), \
           ((___BIGADIGIT) ___BIGATEMP > (___BIGADIGIT)___BIGAFETCH(___BODY_AS(result,___tSUBTYPED),___INT(i)) \
            ? \
	    (___BIGASTORE(___BODY_AS(result,___tSUBTYPED),___INT(i),(___BIGADIGIT)___BIGATEMP), ___FIX_1) \
            : \
	    (___BIGASTORE(___BODY_AS(result,___tSUBTYPED),___INT(i),(___BIGADIGIT)___BIGATEMP), ___FIX_0))))

#if ___BIG_MBASE_WIDTH == 16

#ifdef ___CPU_BIGEND

#if ___BIG_ABASE_WIDTH == 64
#define ___BIG_END_MFLIP(x)((x)^___FIX(3))
#else
#define ___BIG_END_MFLIP(x)((x)^___FIX(1))
#endif

#else
#define ___BIG_END_MFLIP(x)(x)
#endif

#define ___BIGMLENGTH ___U16VECTORLENGTH
#define ___BIGMBASICREF ___U16VECTORREF
#define ___BIGMBASICSET ___U16VECTORSET
#define ___BIGMFETCH ___FETCH_U16
#define ___BIGMSTORE ___STORE_U16
#define ___BIGMDIGIT ___U16
#define ___BIGMDOUBLEDIGIT ___U32

#else

#ifdef ___CPU_BIGEND
#define ___BIG_END_MFLIP(x)((x)^___FIX(1))
#else
#define ___BIG_END_MFLIP(x)(x)
#endif

#define ___BIGMLENGTH ___U32VECTORLENGTH
#define ___BIGMBASICREF ___U32VECTORREF
#define ___BIGMBASICSET ___U32VECTORSET
#define ___BIGMFETCH ___FETCH_U32
#define ___BIGMSTORE ___STORE_U32
#define ___BIGMDIGIT ___U32
#define ___BIGMDOUBLEDIGIT ___U64

#endif

#define ___BIG_MBASE ((___BIGMDOUBLEDIGIT)1 << ___BIG_MBASE_WIDTH)

#define ___BIGMREF(x,j)___BIGMBASICREF(x,___BIG_END_MFLIP(j))

#define ___BIGMSET(result,i,carry)___BIGMBASICSET(result,___BIG_END_MFLIP(i),carry)

#define ___BIGMMUL(result,i,y,j,multiplier,carry) \
     (___temp=(___BIGMDOUBLEDIGIT)___BIGMFETCH(___BODY_AS(result,___tSUBTYPED),___INT(___BIG_END_MFLIP(i))) \
      + (___BIGMDOUBLEDIGIT)___BIGMFETCH(___BODY_AS(y,___tSUBTYPED),___INT(___BIG_END_MFLIP(j))) \
      * (___BIGMDOUBLEDIGIT)___INT(multiplier) \
      + (___BIGMDOUBLEDIGIT)___INT(carry), \
      ___BIGMSTORE(___BODY_AS(result,___tSUBTYPED),___INT(___BIG_END_MFLIP(i)),(___BIGMDIGIT)___temp), \
      ___FIX((___BIGMDOUBLEDIGIT)___temp >> ___BIG_MBASE_WIDTH))
     
#define ___BIGMDIV(u,i,v,k,q,borrow) \
     (___temp = (___BIGMDOUBLEDIGIT)___BIGMFETCH(___BODY_AS(u,___tSUBTYPED),___INT(___BIG_END_MFLIP(i))) \
                -(___BIGMDOUBLEDIGIT)___BIGMFETCH(___BODY_AS(v,___tSUBTYPED),___INT(___BIG_END_MFLIP(k))) \
                 * (___BIGMDOUBLEDIGIT)___INT(q) \
                 + (___BIGMDOUBLEDIGIT)___INT(borrow),\
      ___BIGMSTORE(___BODY_AS(u,___tSUBTYPED),___INT(___BIG_END_MFLIP(i)),((___BIGMDIGIT) ___temp)), \
      ___temp >>= ___BIG_MBASE_WIDTH, \
      (___temp > 0 ? ___FIX(___temp - ___BIG_MBASE) : ___FIX(___temp)))


#define ___BIGMQUO(u,j,v) \
___FIX((((___BIGMDOUBLEDIGIT)___BIGMFETCH(___BODY_AS(u,___tSUBTYPED),___INT(___BIG_END_MFLIP(j))) << ___BIG_MBASE_WIDTH) \
+___BIGMFETCH(___BODY_AS(u,___tSUBTYPED),___INT(___BIG_END_MFLIP(j-___FIX(1)))))/___INT(v))


#define ___BIGMREM(u,j,v,q) \
___FIX((((___BIGMDOUBLEDIGIT)___BIGMFETCH(___BODY_AS(u,___tSUBTYPED),___INT(___BIG_END_MFLIP(j))) << ___BIG_MBASE_WIDTH) \
+___BIGMFETCH(___BODY_AS(u,___tSUBTYPED),___INT(___BIG_END_MFLIP(j-___FIX(1))))) \
-((___BIGMDOUBLEDIGIT)___INT(v)*(___BIGMDOUBLEDIGIT)___INT(q)))

#define ___BIGMTESTP(q,v,r,u) \
(((___BIGMDOUBLEDIGIT)___INT(v)*___INT(q))>((___BIGMDOUBLEDIGIT)___INT(r) << ___BIG_MBASE_WIDTH) +___INT(u))




#ifdef ___CPU_BIGEND

#if ___BIG_ABASE_WIDTH == 64
#define ___BIG_END_FFLIP(x)((x)^___FIX(7))
#else
#define ___BIG_END_FFLIP(x)((x)^___FIX(3))
#endif

#else

#define ___BIG_END_FFLIP(x)(x)

#endif

#define ___BIGFREF(x,j)___U8VECTORREF(x,___BIG_END_FFLIP(j))

#define ___BIGFSET(result,i,carry)___U8VECTORSET(result,___BIG_END_FFLIP(i),carry)

#define ___BIGFLENGTH(x)___U8VECTORLENGTH(x)

/*---------------------------------------------------------------------------*/

/* Inlinable operations (for 'apply' and 'ifjump' GVM instructions) */

/* APPLY-able operations */

#define ___TYPE(x)___FIX(___TYP(x))
#define ___TYPECAST(x,y)((x&~___TMASK)+___INT(y))
#define ___SUBTYPE(x)((___HEADER(x)&___SMASK)>>(___HTB-___TB))
#define ___SUBTYPESET(x,y)___temp=x;___HEADER(___temp)=(___HEADER(___temp)&~___SMASK)+((y)<<(___HTB-___TB));

/* IFJUMP-able operations */

#define ___FALSEP(x)(x==___FAL)
#define ___NULLP(x)(x==___NUL)
#define ___UNBOUNDP(x)(((___temp=x)==___UNB1)||(___temp==___UNB2))
#define ___EQP(x,y)(x==y)
#define ___EOFP(x)(x==___EOF)

/* IFJUMP-able operations */

#define ___FIXNUMP(x)___TESTTYPE(x,___tFIXNUM)
#define ___SPECIALP(x)___TESTTYPE(x,___tSPECIAL)
#define ___PAIRP(x)___TESTTYPE(x,___tPAIR)
#define ___SUBTYPEDP(x)___TESTTYPE(x,___tSUBTYPED)
#define ___SUBTYPEDVECTORP(x)___TESTSUBTYPETAG(x,___sVECTOR)
#define ___SUBTYPEDSYMBOLP(x)___TESTSUBTYPETAG(x,___sSYMBOL)
#define ___SUBTYPEDFLONUMP(x)___TESTSUBTYPETAG(x,___sFLONUM)
#define ___SUBTYPEDBIGNUMP(x)___TESTSUBTYPETAG(x,___sBIGNUM)
#define ___VECTORP(x)___TESTSUBTYPE(x,___sVECTOR)
#define ___RATNUMP(x)___TESTSUBTYPE(x,___sRATNUM)
#define ___CPXNUMP(x)___TESTSUBTYPE(x,___sCPXNUM)
#define ___STRUCTUREP(x)___TESTSUBTYPE(x,___sSTRUCTURE)
#define ___MEROONP(x)___TESTSUBTYPE(x,___sMEROON)
#define ___SYMBOLP(x)___TESTSUBTYPE(x,___sSYMBOL)
#define ___KEYWORDP(x)___TESTSUBTYPE(x,___sKEYWORD)
#define ___FRAMEP(x)___TESTSUBTYPE(x,___sFRAME)
#define ___CONTINUATIONP(x)___TESTLENGTHSUBTYPE(x,(___CONTINUATION_SIZE<<___LWS),___sCONTINUATION)
#define ___THREADP(x)___TESTLENGTHSUBTYPE(x,(___THREAD_SIZE<<___LWS),___sCONTINUATION)
#define ___PROMISEP(x)___TESTSUBTYPE(x,___sPROMISE)
#define ___WILLP(x)___TESTSUBTYPE(x,___sWILL)
#define ___PROCEDUREP(x)___TESTSUBTYPE(x,___sPROCEDURE)
#define ___RETURNP(x)___TESTSUBTYPE(x,___sRETURN)
#define ___POINTERP(x)___TESTSUBTYPE(x,___sPOINTER)
#define ___STRINGP(x)___TESTSUBTYPE(x,___sSTRING)
#define ___S8VECTORP(x)___TESTSUBTYPE(x,___sS8VECTOR)
#define ___U8VECTORP(x)___TESTSUBTYPE(x,___sU8VECTOR)
#define ___S16VECTORP(x)___TESTSUBTYPE(x,___sS16VECTOR)
#define ___U16VECTORP(x)___TESTSUBTYPE(x,___sU16VECTOR)
#define ___S32VECTORP(x)___TESTSUBTYPE(x,___sS32VECTOR)
#define ___U32VECTORP(x)___TESTSUBTYPE(x,___sU32VECTOR)
#define ___S64VECTORP(x)___TESTSUBTYPE(x,___sS64VECTOR)
#define ___U64VECTORP(x)___TESTSUBTYPE(x,___sU64VECTOR)
#define ___F32VECTORP(x)___TESTSUBTYPE(x,___sF32VECTOR)
#define ___F64VECTORP(x)___TESTSUBTYPE(x,___sF64VECTOR)
#define ___FLONUMP(x)___TESTSUBTYPE(x,___sFLONUM)
#define ___BIGNUMP(x)___TESTSUBTYPE(x,___sBIGNUM)
#define ___CHARP(x)(___TYP((___temp=x))==___tSPECIAL&&___temp>=0)

/* APPLY-able operations */

#define ___FIX_0 ___FIX(0)
#define ___FIXPOS(x)((___WORD)x)
#define ___FIXMAX(x,y)((x<y)?y:x)
#define ___FIXMIN(x,y)((x<y)?x:y)
#define ___FIXADD(x,y)((___WORD)(x+y))
#define ___FIX_1 ___FIX(1)
#define ___FIXMUL(x,y)((___WORD)(x*___INT(y)))
#define ___FIXNEG(x)((___WORD)(-x))
#define ___FIXSUB(x,y)((___WORD)(x-y))
#define ___FIXINV(x)___FIX((___FIX(1)/x))
#define ___FIXQUO(x,y)___FIX((x/y))
#define ___FIXREM(x,y)((___WORD)(x%y))/****is this ok?*******/
#define ___FIXMOD(x,y)((___temp=___FIXREM((___ps->temp1=x),(___ps->temp2=y)))==0?___FIX(0):((___ps->temp1<0)==(___ps->temp2<0)?___temp:___FIXADD(___temp,___ps->temp2)))
#define ___FIXIOR(x,y)((___WORD)(x|y))
#define ___FIXXOR(x,y)((___WORD)(x^y))
#define ___FIX_M1 ___FIX(-1)
#define ___FIXAND(x,y)((___WORD)(x&y))
#define ___FIXNOT(x)((___WORD)(x^~___TMASK))
#define ___FIXASHR(x,y)((x>>___INT(y))&~___TMASK)
#define ___FIXLSHR(x,y)(((unsigned ___WORD)x>>___INT(y))&~___TMASK)
#define ___FIXSHL(x,y)(x<<___INT(y))
#define ___FIXTOCHR(x)((x&~___TMASK)+___tSPECIAL)
#define ___FIXFROMCHR(x)(x&~___TMASK)

/* IFJUMP-able operations */

#define ___FIXZEROP(x)(x==0)
#define ___FIXPOSITIVEP(x)(x>0)
#define ___FIXNEGATIVEP(x)(x<0)
#define ___FIXODDP(x)(x&___FIX(1))
#define ___FIXEVENP(x)!(x&___FIX(1))
#define ___FIXEQ(x,y)(x==y)
#define ___FIXLT(x,y)(x<y)
#define ___FIXGT(x,y)(x>y)
#define ___FIXLE(x,y)(x<=y)
#define ___FIXGE(x,y)(x>=y)

/* APPLY-able operations */

#define ___F64TOFIX(x)___FIX(x)
#define ___F64FROMFIX(x)___INT(x)
#define ___F64_0 0.0
#define ___F64POS(x)x
#define ___F64MAX(x,y)(___F64NANP(x)?x:((x>y)?x:y))
#define ___F64MIN(x,y)(___F64NANP(x)?x:((x<y)?x:y))
#define ___F64ADD(x,y)(x+y)
#define ___F64_1 1.0
#define ___F64MUL(x,y)(x*y)
#define ___F64NEG(x)(-x)
#define ___F64SUB(x,y)(x-y)
#define ___F64INV(x)1.0/x
#define ___F64DIV(x,y)(x/y)
#define ___F64ABS(x)___CLIBEXT(fabs)(x)
#define ___F64FLOOR(x)___CLIBEXT(floor)(x)
#define ___F64CEILING(x)___CLIBEXT(ceil)(x)
#define ___F64ROUND(x)___EXT(___round)(x)
#define ___F64EXP(x)___CLIBEXT(exp)(x)
#define ___F64LOG(x)___CLIBEXT(log)(x)
#define ___F64SIN(x)___CLIBEXT(sin)(x)
#define ___F64COS(x)___CLIBEXT(cos)(x)
#define ___F64TAN(x)___CLIBEXT(tan)(x)
#define ___F64ASIN(x)___CLIBEXT(asin)(x)
#define ___F64ACOS(x)___CLIBEXT(acos)(x)
#define ___F64ATAN(x)___CLIBEXT(atan)(x)
#ifdef ___GOOD_ATAN2
#define ___F64ATAN2(y,x)___CLIBEXT(atan2)(y,x)
#else
#define ___F64ATAN2(y,x)___EXT(___atan2)(y,x)
#endif
#define ___F64SQRT(x)___CLIBEXT(sqrt)(x)
#define ___F64COPYSIGN(x,y)___EXT(___copysign)(x,y)

/* IFJUMP-able operations */

#define ___F64ZEROP(x)(x==0.0)
#define ___F64POSITIVEP(x)(x>0.0)
#define ___F64NEGATIVEP(x)(x<0.0)
#define ___F64FINITEP(x)___EXT(___isfinite)(x)
#ifdef ___USE_ISNAN
#define ___F64NANP(x)___EXT(___isnan)(x)
#else
#define ___F64NANP(x)(x!=x)
#endif
#if ___WS == 4
#define ___F64FROMFIXEXACTP(x)(1)
#else
#define ___F64FROMFIXEXACTP(x)(___INT(x)==(___WORD)(___F64)___INT(x))
#endif
#define ___F64EQ(x,y)(x==y)
#define ___F64LT(x,y)(x<y)
#define ___F64GT(x,y)(x>y)
#define ___F64LE(x,y)(x<=y)
#define ___F64GE(x,y)(x>=y)

/* IFJUMP-able operations */

#define ___CHAREQP(x,y)(x==y)
#define ___CHARLTP(x,y)(x<y)
#define ___CHARGTP(x,y)(x>y)
#define ___CHARLEP(x,y)(x<=y)
#define ___CHARGEP(x,y)(x>=y)

/* APPLY-able operations */

#define ___PAIR_SIZE 2
#define ___PAIR_CDR_OFS 0
#define ___PAIR_CAR_OFS 1

#define ___PAIR_CAR(obj)(*(___BODY_AS(obj,___tPAIR)+___PAIR_CAR_OFS))
#define ___PAIR_CDR(obj)(*(___BODY_AS(obj,___tPAIR)+___PAIR_CDR_OFS))

#define ___ALLOC_PAIR_EXPR(x,y)(___hp[0]=___MAKE_HD_WORDS(___PAIR_SIZE,___sPAIR), \
___hp[___PAIR_CAR_OFS+1]=x,___hp[___PAIR_CDR_OFS+1]=y,___hp+=___PAIR_SIZE+1)
#define ___ALLOC_PAIR(x,y){___hp[0]=___MAKE_HD_WORDS(___PAIR_SIZE,___sPAIR); \
___hp[___PAIR_CAR_OFS+1]=x;___hp[___PAIR_CDR_OFS+1]=y;___hp+=___PAIR_SIZE+1;}
#define ___CONS(x,y)(___ALLOC_PAIR_EXPR(x,y),___TAG((___hp-___PAIR_SIZE-1),___tPAIR))

#define ___SETCAR(obj,car)___PAIR_CAR(obj)=car;
#define ___SETCDR(obj,cdr)___PAIR_CDR(obj)=cdr;
#define ___CAR(obj)___PAIR_CAR(obj)
#define ___CDR(obj)___PAIR_CDR(obj)
#define ___CAAR(obj)___CAR(___CAR(obj))
#define ___CADR(obj)___CAR(___CDR(obj))
#define ___CDAR(obj)___CDR(___CAR(obj))
#define ___CDDR(obj)___CDR(___CDR(obj))
#define ___CAAAR(obj)___CAR(___CAR(___CAR(obj)))
#define ___CAADR(obj)___CAR(___CAR(___CDR(obj)))
#define ___CADAR(obj)___CAR(___CDR(___CAR(obj)))
#define ___CADDR(obj)___CAR(___CDR(___CDR(obj)))
#define ___CDAAR(obj)___CDR(___CAR(___CAR(obj)))
#define ___CDADR(obj)___CDR(___CAR(___CDR(obj)))
#define ___CDDAR(obj)___CDR(___CDR(___CAR(obj)))
#define ___CDDDR(obj)___CDR(___CDR(___CDR(obj)))
#define ___CAAAAR(obj)___CAR(___CAR(___CAR(___CAR(obj))))
#define ___CAAADR(obj)___CAR(___CAR(___CAR(___CDR(obj))))
#define ___CAADAR(obj)___CAR(___CAR(___CDR(___CAR(obj))))
#define ___CAADDR(obj)___CAR(___CAR(___CDR(___CDR(obj))))
#define ___CADAAR(obj)___CAR(___CDR(___CAR(___CAR(obj))))
#define ___CADADR(obj)___CAR(___CDR(___CAR(___CDR(obj))))
#define ___CADDAR(obj)___CAR(___CDR(___CDR(___CAR(obj))))
#define ___CADDDR(obj)___CAR(___CDR(___CDR(___CDR(obj))))
#define ___CDAAAR(obj)___CDR(___CAR(___CAR(___CAR(obj))))
#define ___CDAADR(obj)___CDR(___CAR(___CAR(___CDR(obj))))
#define ___CDADAR(obj)___CDR(___CAR(___CDR(___CAR(obj))))
#define ___CDADDR(obj)___CDR(___CAR(___CDR(___CDR(obj))))
#define ___CDDAAR(obj)___CDR(___CDR(___CAR(___CAR(obj))))
#define ___CDDADR(obj)___CDR(___CDR(___CAR(___CDR(obj))))
#define ___CDDDAR(obj)___CDR(___CDR(___CDR(___CAR(obj))))
#define ___CDDDDR(obj)___CDR(___CDR(___CDR(___CDR(obj))))

/* APPLY-able operations */

#define ___MAKECELL(x)(___hp+=2,___hp[-2]=___MAKE_HD_WORDS(1,___sVECTOR), \
___hp[-1]=x,___TAG((___hp-2),___tSUBTYPED))
#define ___CELLREF(x)___BODY_AS(x,___tSUBTYPED)[0]
#define ___CELLSET(x,y)___BODY_AS(x,___tSUBTYPED)[0]=y;

#define ___VECTORLENGTH(x)___FIX(___HD_FIELDS(___HEADER(x)))
#define ___VECTORREF(x,y)*(___WORD*)(((___WORD)___BODY_AS(x,___tSUBTYPED))+((y)<<(___LWS-___TB)))
#define ___VECTORSET(x,y,z)*(___WORD*)(((___WORD)___BODY_AS(x,___tSUBTYPED))+((y)<<(___LWS-___TB)))=z;
#define ___VECTORSHRINK(x,y)___temp=x;___HEADER(___temp)=(___HEADER(___temp)&~___LMASK)+((y)<<(___LF-___TB+___LWS));

#define ___STRINGLENGTH(x)___FIX((___HD_BYTES(___HEADER(x))>>___LCS))
#define ___STRINGREF(x,y)___CHR(___CS_SELECT( \
___FETCH_U8(___BODY_AS(x,___tSUBTYPED),(y)>>___TB), \
___FETCH_U16(___BODY_AS(x,___tSUBTYPED),(y)>>___TB), \
___FETCH_U32(___BODY_AS(x,___tSUBTYPED),(y)>>___TB)))
#define ___STRINGSET(x,y,z)___CS_SELECT( \
___STORE_U8(___BODY_AS(x,___tSUBTYPED),(y)>>___TB,___INT(z)), \
___STORE_U16(___BODY_AS(x,___tSUBTYPED),(y)>>___TB,___INT(z)), \
___STORE_U32(___BODY_AS(x,___tSUBTYPED),(y)>>___TB,___INT(z)));
#define ___STRINGSHRINK(x,y)___temp=x;___HEADER(___temp)=(___HEADER(___temp)&~___LMASK)+((y)<<(___LF-___TB+___LCS));

#define ___S8VECTORLENGTH(x)___FIX(___HD_BYTES(___HEADER(x)))
#define ___S8VECTORREF(x,y) \
___FIX(___FETCH_S8(___BODY_AS(x,___tSUBTYPED),(y)>>___TB))
#define ___S8VECTORSET(x,y,z) \
___STORE_S8(___BODY_AS(x,___tSUBTYPED),(y)>>___TB,___INT(z));
#define ___S8VECTORSHRINK(x,y)___temp=x;___HEADER(___temp)=(___HEADER(___temp)&~___LMASK)+((y)<<(___LF-___TB));

#define ___U8VECTORLENGTH(x)___FIX(___HD_BYTES(___HEADER(x)))
#define ___U8VECTORREF(x,y) \
___FIX(___FETCH_U8(___BODY_AS(x,___tSUBTYPED),(y)>>___TB))
#define ___U8VECTORSET(x,y,z) \
___STORE_U8(___BODY_AS(x,___tSUBTYPED),(y)>>___TB,___INT(z));
#define ___U8VECTORSHRINK(x,y)___temp=x;___HEADER(___temp)=(___HEADER(___temp)&~___LMASK)+((y)<<(___LF-___TB));

#define ___S16VECTORLENGTH(x)___FIX((___HD_BYTES(___HEADER(x))>>1))
#define ___S16VECTORREF(x,y) \
___FIX(___FETCH_S16(___BODY_AS(x,___tSUBTYPED),(y)>>___TB))
#define ___S16VECTORSET(x,y,z) \
___STORE_S16(___BODY_AS(x,___tSUBTYPED),(y)>>___TB,___INT(z));
#define ___S16VECTORSHRINK(x,y)___temp=x;___HEADER(___temp)=(___HEADER(___temp)&~___LMASK)+((y)<<(___LF-___TB+1));

#define ___U16VECTORLENGTH(x)___FIX((___HD_BYTES(___HEADER(x))>>1))
#define ___U16VECTORREF(x,y) \
___FIX(___FETCH_U16(___BODY_AS(x,___tSUBTYPED),(y)>>___TB))
#define ___U16VECTORSET(x,y,z) \
___STORE_U16(___BODY_AS(x,___tSUBTYPED),(y)>>___TB,___INT(z));
#define ___U16VECTORSHRINK(x,y)___temp=x;___HEADER(___temp)=(___HEADER(___temp)&~___LMASK)+((y)<<(___LF-___TB+1));

#define ___S32VECTORLENGTH(x)___FIX((___HD_BYTES(___HEADER(x))>>2))
#define ___S32VECTORREF(x,y) \
___S32BOX(___FETCH_S32(___BODY_AS(x,___tSUBTYPED),(y)>>___TB))
#define ___S32VECTORSET(x,y,z) \
___STORE_S32(___BODY_AS(x,___tSUBTYPED),(y)>>___TB,___S32UNBOX(z));
#define ___S32VECTORSHRINK(x,y)___temp=x;___HEADER(___temp)=(___HEADER(___temp)&~___LMASK)+((y)<<(___LF-___TB+2));

#define ___U32VECTORLENGTH(x)___FIX((___HD_BYTES(___HEADER(x))>>2))
#define ___U32VECTORREF(x,y) \
___U32BOX(___FETCH_U32(___BODY_AS(x,___tSUBTYPED),(y)>>___TB))
#define ___U32VECTORSET(x,y,z) \
___STORE_U32(___BODY_AS(x,___tSUBTYPED),(y)>>___TB,___U32UNBOX(z));
#define ___U32VECTORSHRINK(x,y)___temp=x;___HEADER(___temp)=(___HEADER(___temp)&~___LMASK)+((y)<<(___LF-___TB+2));

#define ___S64VECTORLENGTH(x)___FIX((___HD_BYTES(___HEADER(x))>>3))
#define ___S64VECTORREF(x,y) \
___S64BOX(___FETCH_S64(___BODY_AS(x,___tSUBTYPED),(y)>>___TB))
#define ___S64VECTORSET(x,y,z) \
___STORE_S64(___BODY_AS(x,___tSUBTYPED),(y)>>___TB,___S64UNBOX(z));
#define ___S64VECTORSHRINK(x,y)___temp=x;___HEADER(___temp)=(___HEADER(___temp)&~___LMASK)+((y)<<(___LF-___TB+3));

#define ___U64VECTORLENGTH(x)___FIX((___HD_BYTES(___HEADER(x))>>3))
#define ___U64VECTORREF(x,y) \
___U64BOX(___FETCH_U64(___BODY_AS(x,___tSUBTYPED),(y)>>___TB))
#define ___U64VECTORSET(x,y,z) \
___STORE_U64(___BODY_AS(x,___tSUBTYPED),(y)>>___TB,___U64UNBOX(z));
#define ___U64VECTORSHRINK(x,y)___temp=x;___HEADER(___temp)=(___HEADER(___temp)&~___LMASK)+((y)<<(___LF-___TB+3));

#define ___F32VECTORLENGTH(x)___FIX((___HD_BYTES(___HEADER(x))>>2))
#ifdef ___F32
#define ___F32VECTORREF(x,y)*(___F32*)(((___WORD)___BODY_AS(x,___tSUBTYPED))+((y)<<(2-___TB)))
#define ___F32VECTORSET(x,y,z)*(___F32*)(((___WORD)___BODY_AS(x,___tSUBTYPED))+((y)<<(2-___TB)))=z;
#else
#define ___F32VECTORREF(x,y)0.0
#define ___F32VECTORSET(x,y,z);
#endif
#define ___F32VECTORSHRINK(x,y)___temp=x;___HEADER(___temp)=(___HEADER(___temp)&~___LMASK)+((y)<<(___LF-___TB+2));

#define ___F64VECTORLENGTH(x)___FIX((___HD_BYTES(___HEADER(x))>>3))
#define ___F64VECTORREF(x,y)*(___F64*)(((___WORD)___BODY_AS(x,___tSUBTYPED))+((y)<<(3-___TB)))
#define ___F64VECTORSET(x,y,z)*(___F64*)(((___WORD)___BODY_AS(x,___tSUBTYPED))+((y)<<(3-___TB)))=z;
#define ___F64VECTORSHRINK(x,y)___temp=x;___HEADER(___temp)=(___HEADER(___temp)&~___LMASK)+((y)<<(___LF-___TB+3));

#define ___STRUCTREF(x,y,tag)___VECTORREF(x,y)
#define ___STRUCTSET(x,y,z,tag)___VECTORSET(x,y,z)

#define ___SYMBOLTOSTR(x)___VECTORREF(x,___FIX(___SYMBOL_NAME))

#define ___KEYWORDTOSTR(x)___VECTORREF(x,___FIX(___KEYWORD_NAME))

#define ___CLOSURECODE(x)___CLO(x,0)
#define ___CLOSUREREF(x,y)___CLO(x,___INT(y))
#define ___CLOSURESET(x,y,z)___SET_CLO(x,___INT(y),z)

#define ___GLOBALVARREF(gv)((___glo_struct*)___FIELD(gv,___SYMBOL_GLOBAL))->val
#define ___GLOBALVARSET(gv,x)((___glo_struct*)___FIELD(gv,___SYMBOL_GLOBAL))->val = x;

#define ___MAKEPROMISE(x) \
(___hp+=___PROMISE_SIZE+1, \
___hp[-3]=___MAKE_HD_WORDS(___PROMISE_SIZE,___sPROMISE), \
___hp[-2]=x, ___hp[-1]=___TAG((___hp-___PROMISE_SIZE-1),___tSUBTYPED))

#define ___MAKEWILL1(testator) \
(___hp[0]=___MAKE_HD_WORDS(___WILL1_SIZE,___sWILL), \
___hp[1]=___FLOATING_WILL, \
___hp[2]=testator, \
___hp+=___WILL1_SIZE+1, \
___TAG((___hp-___WILL1_SIZE-1),___tSUBTYPED))
#define ___MAKEWILL2(testator,action) \
(___hp[0]=___MAKE_HD_WORDS(___WILL2_SIZE,___sWILL), \
___hp[1]=___ps->non_executable_wills, \
___hp[2]=testator, \
___hp[3]=action, \
___ps->non_executable_wills=___TAG(___hp,0), \
___hp+=___WILL2_SIZE+1, \
___TAG((___hp-___WILL2_SIZE-1),___tSUBTYPED))
#define ___WILLTESTATOR(x)___BODY_AS(x,___tSUBTYPED)[1]

#define ___EXEC_WILL     1
#define ___REACH_WILL    2
#define ___FLOATING_WILL -1

#define ___CURRENTTHREAD ___ps->current_thread
#define ___RUNQUEUE ___ps->run_queue
#define ___SCHEDULERSTATE ___ps->run_queue

#define ___JUMP_THREAD_SAVE(jump) \
{ ___WORD thunk; \
  ___WORD arg1; \
  ___WORD arg2; \
  ___WORD arg3; \
  ___WORD frame; \
  ___WORD thread; \
  ___POP_ARGS4(thunk,arg1,arg2,arg3) \
  if (___R0 == ___ps->handler_break) \
    frame = ___STK(-___BREAK_FRAME_NEXT); \
  else \
    { \
      ___SET_STK(-___FRAME_STACK_RA,___R0) \
      ___R0 = ___ps->handler_break; \
      frame = ___CAST(___WORD,___fp); \
      ___ADJFP(___BREAK_FRAME_SPACE) \
      ___SET_STK(-___BREAK_FRAME_NEXT,frame) \
      ___ps->stack_break = ___fp; \
    } \
  thread = ___ps->current_thread; \
  ___FIELD(thread,0) = frame; \
  ___PUSH_ARGS4(thread,arg1,arg2,arg3) \
  jump(___SET_NARGS(4),thunk) \
}

#define ___JUMPSAFE_THREAD_SAVE \
___JUMP_THREAD_SAVE(___JUMPSAFE)

#define ___JUMPNOTSAFE_THREAD_SAVE \
___JUMP_THREAD_SAVE(___JUMPNOTSAFE)

#define ___JUMP_THREAD_RESTORE(jump) \
{ ___WORD thread; \
  ___WORD thunk; \
  ___POP_ARGS2(thread,thunk) \
  ___fp = ___ps->stack_break; \
  ___SET_STK(-___BREAK_FRAME_NEXT,___FIELD(thread,0)) \
  ___FIELD(thread,0) = ___FIX(0); \
  ___ps->current_thread = thread; \
  ___SET_R0(___ps->handler_break) \
  jump(___SET_NARGS(0),thunk) \
}

#define ___JUMPSAFE_THREAD_RESTORE \
___JUMP_THREAD_RESTORE(___JUMPSAFE)

#define ___JUMPNOTSAFE_THREAD_RESTORE \
___JUMP_THREAD_RESTORE(___JUMPNOTSAFE)


#define ___FRAME_OVERHEAD(n) (___FRAME_SPACE(n)-(n))
#define ___RETI_RA (-___FRAME_OVERHEAD(___NB_GVM_REGS+1))
#define ___FRAME_RETI_RA (1-___FRAME_RESERVE-___RETI_RA)

#define ___CAST(type,val)((type)(val))
#define ___SYMBOL_NAME   0
#define ___SYMBOL_HASH   1
#define ___SYMBOL_NEXT   2
#define ___SYMBOL_GLOBAL 3
#define ___KEYWORD_NAME  0
#define ___KEYWORD_HASH  1
#define ___KEYWORD_NEXT  2
#define ___PROMISE_THUNK  0
#define ___PROMISE_RESULT 1
#define ___CONTINUATION_FRAME 0
#define ___CONTINUATION_DENV  1
#define ___THREAD_DENV        20
#define ___THREAD_DENV_CACHE1 21
#define ___THREAD_DENV_CACHE2 22
#define ___THREAD_DENV_CACHE3 23
#define ___ENV_NAME_VAL 0
#define ___ENV_LEFT     1
#define ___ENV_RIGHT    2
#define ___DENV_GLOBAL             0
#define ___DENV_LOCAL              1
#define ___DENV_INTERRUPT_MASK     2
#define ___DENV_EXCEPTION_HANDLER  3
#define ___DENV_DEBUGGING_SETTINGS 4
#define ___DENV_INPUT_PORT         5
#define ___DENV_OUTPUT_PORT        6

#define ___FRAME_EXTRA_SLOTS 1
#define ___FRAME_STACK_RA 0
#define ___FRAME_RA 0

#define ___FRAME_STORE_RA(ra) \
if (___fp != ___ps->stack_break) \
  ___SET_STK(-___FRAME_STACK_RA,ra)

#define ___FRAME_FETCH_RA \
((___fp != ___ps->stack_break) \
? ___STK(-___FRAME_STACK_RA) \
: ___ps->handler_break)

#define ___JUMP_CONTINUATION_CAPTURE(jump) \
{ ___WORD thunk; \
  ___WORD arg; \
  ___WORD frame; \
  ___WORD cont; \
  ___POP_ARGS2(thunk,arg) \
  if (___R0 == ___ps->handler_break) \
    frame = ___STK(-___BREAK_FRAME_NEXT); \
  else \
    { \
      ___SET_STK(-___FRAME_STACK_RA,___R0) \
      ___R0 = ___ps->handler_break; \
      frame = ___CAST(___WORD,___fp); \
    } \
  ___ADJFP(___ROUND_TO_MULT(___SUBTYPED_OVERHEAD+___CONTINUATION_SIZE,___FRAME_ALIGN)) \
  ___SET_STK(0,___MAKE_HD_WORDS(___CONTINUATION_SIZE,___sCONTINUATION)) \
  ___SET_STK(-1,frame) \
  ___SET_STK(-2,___FIELD(___ps->current_thread,___THREAD_DENV)) \
  cont = ___TAG(&___STK(0),___tSUBTYPED); \
  ___ADJFP(___BREAK_FRAME_SPACE) \
  ___SET_STK(-___BREAK_FRAME_NEXT,frame) \
  ___ps->stack_break = ___fp; \
  ___PUSH_ARGS2(cont,arg); \
  jump(___SET_NARGS(2),thunk) \
}

#define ___JUMPSAFE_CONTINUATION_CAPTURE \
___JUMP_CONTINUATION_CAPTURE(___JUMPSAFE)

#define ___JUMPNOTSAFE_CONTINUATION_CAPTURE \
___JUMP_CONTINUATION_CAPTURE(___JUMPNOTSAFE)

#define ___JUMP_CONTINUATION_RETURN \
{ ___WORD cont; \
  ___WORD result; \
  ___WORD temp; \
  ___WORD thread; \
  ___POP_ARGS2(cont,result) \
  ___SET_R1(result) \
  ___fp = ___ps->stack_break; \
  ___SET_STK(-___BREAK_FRAME_NEXT,___FIELD(cont,___CONTINUATION_FRAME)) \
  temp = ___FIELD(cont,___CONTINUATION_DENV); \
  thread = ___ps->current_thread; \
  ___FIELD(thread,___THREAD_DENV) = temp; \
  temp = ___FIELD(___FIELD(temp,___DENV_LOCAL),___ENV_NAME_VAL); \
  ___FIELD(thread,___THREAD_DENV_CACHE1) = temp; \
  ___FIELD(thread,___THREAD_DENV_CACHE2) = temp; \
  ___FIELD(thread,___THREAD_DENV_CACHE3) = temp; \
  ___JUMPPRM(___NOTHING,___ps->handler_break) \
}

#define ___JUMPSAFE_CONTINUATION_RETURN \
___JUMP_CONTINUATION_RETURN

#define ___JUMPNOTSAFE_CONTINUATION_RETURN \
___JUMP_CONTINUATION_RETURN

#define ___JUMP_CONTINUATION_GRAFT(jump) \
{ ___WORD cont; \
  ___WORD thunk; \
  ___WORD temp; \
  ___WORD thread; \
  ___POP_ARGS2(cont,thunk) \
  ___fp = ___ps->stack_break; \
  ___SET_STK(-___BREAK_FRAME_NEXT,___FIELD(cont,___CONTINUATION_FRAME)) \
  temp = ___FIELD(cont,___CONTINUATION_DENV); \
  thread = ___ps->current_thread; \
  ___FIELD(thread,___THREAD_DENV) = temp; \
  temp = ___FIELD(___FIELD(temp,___DENV_LOCAL),___ENV_NAME_VAL); \
  ___FIELD(thread,___THREAD_DENV_CACHE1) = temp; \
  ___FIELD(thread,___THREAD_DENV_CACHE2) = temp; \
  ___FIELD(thread,___THREAD_DENV_CACHE3) = temp; \
  ___SET_R0(___ps->handler_break) \
  jump(___SET_NARGS(0),thunk) \
}

#define ___JUMPSAFE_CONTINUATION_GRAFT \
___JUMP_CONTINUATION_GRAFT(___JUMPSAFE)

#define ___JUMPNOTSAFE_CONTINUATION_GRAFT \
___JUMP_CONTINUATION_GRAFT(___JUMPNOTSAFE)

/*---------------------------------------------------------------------------*/

/* Stack manipulation */

#define ___FP_AFTER(fp1,fp2)((fp1) < (fp2))
#define ___FP_PUSH(fp,val)*--fp=(val);
#define ___FP_POP(fp)*fp++
#define ___FP_STK(fp,fpo)fp[-(fpo)]
#define ___FP_SET_STK(fp,fpo,val)fp[-(fpo)]=(val);
#define ___FP_ADJFP(fp,fpa)fp-=(fpa);

#define ___PUSH(val)___FP_PUSH(___fp,val)
#define ___POP ___FP_POP(___fp)
#define ___STK(fpo)___FP_STK(___fp,fpo)
#define ___SET_STK(fpo,val)___FP_SET_STK(___fp,fpo,val)
#define ___ADJFP(fpa)___FP_ADJFP(___fp,fpa)
#define ___PSSTK(fpo)___FP_STK(___PSFP,fpo)
#define ___PSADJFP(fpa)___FP_ADJFP(___PSFP,fpa)

#define ___PRM(i,glo)___GLOSTRUCT(i,glo).prm
#define ___GLO(i,glo)___GLOSTRUCT(i,glo).val
#define ___SET_GLO(i,glo,x)___GLOSTRUCT(i,glo).val=x;
#ifdef ___BIND_LATE
#define ___GLOSTRUCT(i,glo)(*(___glo_struct*)___glo_tbl[i])
#else
#define ___GLOSTRUCT(i,glo)glo
#endif

#define ___HOST_PROC ___SM(___MH_PROC,___PH_PROC)
#define ___HOST_LBL0 ___SM(1,___PH_LBL0)

#define ___BEGIN_M_COD ___SM(___BEGIN_COD,___NOTHING)
#define ___END_M_COD ___SM(___END_COD,___NOTHING)

#define ___BEGIN_M_HLBL ___SM(___BEGIN_HLBL,___NOTHING)
#define ___DEF_M_HLBL_INTRO ___SM(___DEF_HLBL_INTRO,___NOTHING)
#define ___DEF_M_HLBL(id)___SM(___DEF_HLBL(id),___NOTHING)
#define ___END_M_HLBL ___SM(___END_HLBL,___NOTHING)

#define ___BEGIN_M_SW ___SM(___BEGIN_SW,___NOTHING)
#define ___END_M_SW ___SM(___END_SW,___NOTHING)

#define ___BEGIN_P_COD ___SM(___NOTHING,___BEGIN_COD)
#define ___END_P_COD ___SM(___NOTHING,___END_COD)

#define ___BEGIN_P_HLBL ___SM(___NOTHING,___BEGIN_HLBL)
#define ___DEF_P_HLBL_INTRO ___SM(___NOTHING,___DEF_HLBL_INTRO)
#define ___DEF_P_HLBL(id)___SM(___NOTHING,___DEF_HLBL(id))
#define ___END_P_HLBL ___SM(___NOTHING,___END_HLBL)

#define ___BEGIN_P_SW ___SM(___NOTHING,___BEGIN_SW)
#define ___END_P_SW ___SM(___NOTHING,___END_SW)

#define ___D_ALL ___SM(___MD_ALL,___PD_ALL)
#define ___R_ALL ___SM(___MR_ALL,___PR_ALL)
#define ___W_ALL ___SM(___MW_ALL,___PW_ALL)

#define ___BEGIN_COD \
___HIDDEN ___WORD ___HOST_PROC ___P((___processor_state ___ps),(___ps) \
___processor_state ___ps;){ \
register ___WORD ___pc, ___start, ___temp; \
register ___U64 ___temp64; /*******************/ \
___D_ALL

#define ___END_COD \
___ps->pc=___pc; ___W_ALL return ___pc; }

#ifdef ___USE_LABEL_VALUES

#define ___SETUP_HOST_LABEL 1

#define ___BEGIN_HLBL static void *___hlbl_tbl[]={
#define ___DEF_HLBL_INTRO 0,
#define ___DEF_HLBL(id)&&id,
#define ___END_HLBL 0}; if (___ps == 0) return (___WORD)___hlbl_tbl;

#define ___BEGIN_SW \
___R_ALL ___start=___MLBL(___HOST_LBL0); ___pc=___ps->pc; \
goto *((___label_struct*)(___pc-___tSUBTYPED))->host_label;

#define ___DEF_SW(n)

#ifdef ___USE_INLINE_JUMP

#define ___END_SW \
___jumpext:

#define ___JUMP(dest) \
{void *___host_label; ___pc=dest; \
 ___host_label=((___label_struct*)(___pc-___tSUBTYPED))->host_label; \
 if (___CAST_FAKEHOST_TO_HOST(((___label_struct*)(___pc-___tSUBTYPED))->host) \
     == ___HOST_PROC) \
   goto *___host_label; \
 goto ___jumpext;}

#else

#define ___END_SW \
___jump: \
{void *___host_label=((___label_struct*)(___pc-___tSUBTYPED))->host_label; \
 if (___CAST_FAKEHOST_TO_HOST(((___label_struct*)(___pc-___tSUBTYPED))->host) \
     == ___HOST_PROC) \
   goto *___host_label;} \
___jumpext:

#define ___JUMP(dest){___pc=dest;goto ___jump;}

#endif

#define ___JUMPEXT(dest){___pc=dest;goto ___jumpext;}

#else

#define ___SETUP_HOST_LABEL 0

#define ___BEGIN_HLBL
#define ___DEF_HLBL_INTRO
#define ___DEF_HLBL(id)
#define ___END_HLBL

#define ___BEGIN_SW \
___R_ALL ___start=___MLBL(___HOST_LBL0); ___pc=___ps->pc; \
___jump: \
switch((___pc-=___start)/(___LS*___WS)) \
{

#define ___DEF_SW(n)case ___PH_LBL0-___HOST_LBL0+n:

#define ___END_SW \
}___pc+=___start;___jumpext:

#define ___JUMP(dest){___pc=dest;goto ___jump;}
#define ___JUMPEXT(dest){___pc=dest;goto ___jumpext;}

#endif

#define ___BEGIN_P_SLBL ___SM(___NOTHING,___BEGIN_SLBL)
#define ___DEF_P_SLBL(id)
#define ___END_P_SLBL ___SM(___NOTHING,___END_SLBL)

#define ___LS 4

#define ___PSHEAP ___ps->hp
#define ___D_HEAP register ___WORD *___hp;
#define ___R_HEAP ___hp=___PSHEAP;
#define ___W_HEAP ___PSHEAP=___hp;

#define ___PSFP ___ps->fp
#define ___D_FP register ___WORD *___fp;
#define ___R_FP ___fp=___PSFP;
#define ___W_FP ___PSFP=___fp;

/*---------------------------------------------------------------------------*/

/* GVM registers */

#ifndef ___MAX_LOCAL_GVM_REGS
#define ___MAX_LOCAL_GVM_REGS 25
#endif

#if ___NB_ARG_REGS == 0
#define ___PUSH_ARGS1(a)___PUSH(a)
#define ___POP_ARGS1(a)a = ___POP;
#define ___PUSH_ARGS2(a,b)___PUSH(a)___PUSH(b)
#define ___POP_ARGS2(a,b)b = ___POP; a = ___POP;
#define ___PUSH_ARGS3(a,b,c)___PUSH(a)___PUSH(b)___PUSH(c)
#define ___POP_ARGS3(a,b,c)c = ___POP; b = ___POP; a = ___POP;
#define ___PUSH_ARGS4(a,b,c,d)___PUSH(a)___PUSH(b)___PUSH(c)___PUSH(d)
#define ___POP_ARGS4(a,b,c,d)d = ___POP; c = ___POP; b = ___POP; a = ___POP;
#else
#define ___PUSH_ARGS1(a)___SET_R1(a)
#define ___POP_ARGS1(a)a = ___R1;
#if ___NB_ARG_REGS == 1
#define ___PUSH_ARGS2(a,b)___PUSH(a)___SET_R1(b)
#define ___POP_ARGS2(a,b)b = ___R1; a = ___POP;
#define ___PUSH_ARGS3(a,b,c)___PUSH(a)___PUSH(b)___SET_R1(c)
#define ___POP_ARGS3(a,b,c)c = ___R1; b = ___POP; a = ___POP;
#define ___PUSH_ARGS4(a,b,c,d)___PUSH(a)___PUSH(b)___PUSH(c)___SET_R1(d)
#define ___POP_ARGS4(a,b,c,d)d = ___R1; c = ___POP; b = ___POP; a = ___POP;
#else
#define ___PUSH_ARGS2(a,b)___SET_R1(a)___SET_R2(b)
#define ___POP_ARGS2(a,b)b = ___R2; a = ___R1;
#if ___NB_ARG_REGS == 2
#define ___PUSH_ARGS3(a,b,c)___PUSH(a)___SET_R1(b)___SET_R2(c)
#define ___POP_ARGS3(a,b,c)c = ___R2; b = ___R1; a = ___POP;
#define ___PUSH_ARGS4(a,b,c,d)___PUSH(a)___PUSH(b)___SET_R1(c)___SET_R2(d)
#define ___POP_ARGS4(a,b,c,d)d = ___R2; c = ___R1; b = ___POP; a = ___POP;
#else
#define ___PUSH_ARGS3(a,b,c)___SET_R1(a)___SET_R2(b)___SET_R3(c)
#define ___POP_ARGS3(a,b,c)c = ___R3; b = ___R2; a = ___R1;
#if ___NB_ARG_REGS == 3
#define ___PUSH_ARGS4(a,b,c,d)___PUSH(a)___SET_R1(b)___SET_R2(c)___SET_R3(d)
#define ___POP_ARGS4(a,b,c,d)d = ___R3; c = ___R2; b = ___R1; a = ___POP;
#else
#define ___PUSH_ARGS4(a,b,c,d)___SET_R1(a)___SET_R2(b)___SET_R3(c)___SET_R4(d)
#define ___POP_ARGS4(a,b,c,d)d = ___R4; c = ___R3; b = ___R2; a = ___R1;
#endif
#endif
#endif
#endif

#define ___PSR0 ___ps->r[0]
#if ___MAX_LOCAL_GVM_REGS > 0
#define ___D_R0 register ___WORD ___r0;
#define ___R_R0 ___r0=___PSR0;
#define ___W_R0 ___PSR0=___r0;
#define ___R0 ___r0
#else
#define ___D_R0
#define ___R_R0
#define ___W_R0
#define ___R0 ___PSR0
#endif
#define ___SET_R0(val)___R0=val;
#define ___PUSH_R0 ___PUSH(___R0)
#define ___POP_R0 ___SET_R0(___POP)

#if ___NB_ARG_REGS == 0
#define ___PUSH_ARGS_IN_REGS(na)
#define ___POP_ARGS_IN_REGS(na)
#define ___PSSELF ___PSR1
#define ___SELF ___R1
#define ___LD_ARG_REGS
#define ___ST_ARG_REGS
#endif

#define ___PSR1 ___ps->r[1]
#if ___MAX_LOCAL_GVM_REGS > 1
#define ___D_R1 register ___WORD ___r1;
#define ___R_R1 ___r1=___PSR1;
#define ___W_R1 ___PSR1=___r1;
#define ___R1 ___r1
#else
#define ___D_R1
#define ___R_R1
#define ___W_R1
#define ___R1 ___PSR1
#endif
#define ___SET_R1(val)___R1=(val);
#define ___PUSH_R1_TO_R1 ___PUSH(___R1)
#define ___POP_R1_TO_R1 ___SET_R1(___POP)
#define ___LD_R1_TO_R1 ___D_R1 ___R_R1
#define ___ST_R1_TO_R1 ___W_R1

#if ___NB_ARG_REGS == 1
#define ___PUSH_ARGS_IN_REGS(na) \
if ((na)>0) ___PUSH_R1_TO_R1
#define ___POP_ARGS_IN_REGS(na) \
if ((na)>0) ___POP_R1_TO_R1
#define ___PSSELF ___PSR2
#define ___SELF ___R2
#define ___LD_ARG_REGS ___LD_R1_TO_R1
#define ___ST_ARG_REGS ___ST_R1_TO_R1
#endif

#if ___NB_GVM_REGS == 2

#define ___PUSH_REGS ___PUSH_R0 ___PUSH_R1_TO_R1
#define ___POP_REGS ___POP_R1_TO_R1 ___POP_R0

#else

#define ___PSR2 ___ps->r[2]
#if ___MAX_LOCAL_GVM_REGS > 2
#define ___D_R2 register ___WORD ___r2;
#define ___R_R2 ___r2=___PSR2;
#define ___W_R2 ___PSR2=___r2;
#define ___R2 ___r2
#else
#define ___D_R2
#define ___R_R2
#define ___W_R2
#define ___R2 ___PSR2
#endif
#define ___SET_R2(val)___R2=(val);
#define ___PUSH_R1_TO_R2 ___PUSH_R1_TO_R1 ___PUSH(___R2)
#define ___POP_R2_TO_R1 ___SET_R2(___POP)___POP_R1_TO_R1
#define ___LD_R1_TO_R2 ___D_R2 ___LD_R1_TO_R1 ___R_R2
#define ___ST_R1_TO_R2 ___ST_R1_TO_R1 ___W_R2

#if ___NB_ARG_REGS == 2
#define ___PUSH_ARGS_IN_REGS(na) \
switch (na) \
  { \
  case 0:  break; \
  case 1:  ___PUSH_R1_TO_R1 break; \
  default: ___PUSH_R1_TO_R2 \
  }
#define ___POP_ARGS_IN_REGS(na) \
switch (na) \
  { \
  case 0:  break; \
  case 1:  ___POP_R1_TO_R1 break; \
  default: ___POP_R2_TO_R1 \
  }
#define ___PSSELF ___PSR3
#define ___SELF ___R3
#define ___LD_ARG_REGS ___LD_R1_TO_R2
#define ___ST_ARG_REGS ___ST_R1_TO_R2
#endif

#if ___NB_GVM_REGS == 3

#define ___PUSH_REGS ___PUSH_R0 ___PUSH_R1_TO_R2
#define ___POP_REGS ___POP_R2_TO_R1 ___POP_R0

#else

#define ___PSR3 ___ps->r[3]
#if ___MAX_LOCAL_GVM_REGS > 3
#define ___D_R3 register ___WORD ___r3;
#define ___R_R3 ___r3=___PSR3;
#define ___W_R3 ___PSR3=___r3;
#define ___R3 ___r3
#else
#define ___D_R3
#define ___R_R3
#define ___W_R3
#define ___R3 ___PSR3
#endif
#define ___SET_R3(val)___R3=(val);
#define ___PUSH_R1_TO_R3 ___PUSH_R1_TO_R2 ___PUSH(___R3)
#define ___POP_R3_TO_R1 ___SET_R3(___POP)___POP_R2_TO_R1
#define ___LD_R1_TO_R3 ___D_R3 ___LD_R1_TO_R2 ___R_R3
#define ___ST_R1_TO_R3 ___ST_R1_TO_R2 ___W_R3

#if ___NB_ARG_REGS == 3
#define ___PUSH_ARGS_IN_REGS(na) \
switch (na) \
  { \
  case 0:  break; \
  case 1:  ___PUSH_R1_TO_R1 break; \
  case 2:  ___PUSH_R1_TO_R2 break; \
  default: ___PUSH_R1_TO_R3 \
  }
#define ___POP_ARGS_IN_REGS(na) \
switch (na) \
  { \
  case 0:  break; \
  case 1:  ___POP_R1_TO_R1 break; \
  case 2:  ___POP_R2_TO_R1 break; \
  default: ___POP_R3_TO_R1 \
  }
#define ___PSSELF ___PSR4
#define ___SELF ___R4
#define ___LD_ARG_REGS ___LD_R1_TO_R3
#define ___ST_ARG_REGS ___ST_R1_TO_R3
#endif

#if ___NB_GVM_REGS == 4

#define ___PUSH_REGS ___PUSH_R0 ___PUSH_R1_TO_R3
#define ___POP_REGS ___POP_R3_TO_R1 ___POP_R0

#else

#define ___PSR4 ___ps->r[4]
#if ___MAX_LOCAL_GVM_REGS > 4
#define ___D_R4 register ___WORD ___r4;
#define ___R_R4 ___r4=___PSR4;
#define ___W_R4 ___PSR4=___r4;
#define ___R4 ___r4
#else
#define ___D_R4
#define ___R_R4
#define ___W_R4
#define ___R4 ___PSR4
#endif
#define ___SET_R4(val)___R4=(val);
#define ___PUSH_R1_TO_R4 ___PUSH_R1_TO_R3 ___PUSH(___R4)
#define ___POP_R4_TO_R1 ___SET_R4(___POP)___POP_R3_TO_R1
#define ___LD_R1_TO_R4 ___D_R4 ___LD_R1_TO_R3 ___R_R4
#define ___ST_R1_TO_R4 ___ST_R1_TO_R3 ___W_R4

#if ___NB_ARG_REGS == 4
#define ___PUSH_ARGS_IN_REGS(na) \
switch (na) \
  { \
  case 0:  break; \
  case 1:  ___PUSH_R1_TO_R1 break; \
  case 2:  ___PUSH_R1_TO_R2 break; \
  case 3:  ___PUSH_R1_TO_R3 break; \
  default: ___PUSH_R1_TO_R4 \
  }
#define ___POP_ARGS_IN_REGS(na) \
switch (na) \
  { \
  case 0:  break; \
  case 1:  ___POP_R1_TO_R1 break; \
  case 2:  ___POP_R2_TO_R1 break; \
  case 3:  ___POP_R3_TO_R1 break; \
  default: ___POP_R4_TO_R1 \
  }
#define ___PSSELF ___PSR5
#define ___SELF ___R5
#define ___LD_ARG_REGS ___LD_R1_TO_R4
#define ___ST_ARG_REGS ___ST_R1_TO_R4
#endif

#if ___NB_GVM_REGS == 5

#define ___PUSH_REGS ___PUSH_R0 ___PUSH_R1_TO_R4
#define ___POP_REGS ___POP_R4_TO_R1 ___POP_R0

#else

#define ___PSR5 ___ps->r[5]
#if ___MAX_LOCAL_GVM_REGS > 5
#define ___D_R5 register ___WORD ___r5;
#define ___R_R5 ___r5=___PSR5;
#define ___W_R5 ___PSR5=___r5;
#define ___R5 ___r5
#else
#define ___D_R5
#define ___R_R5
#define ___W_R5
#define ___R5 ___PSR5
#endif
#define ___SET_R5(val)___R5=(val);
#define ___PUSH_R1_TO_R5 ___PUSH_R1_TO_R4 ___PUSH(___R5)
#define ___POP_R5_TO_R1 ___SET_R5(___POP)___POP_R4_TO_R1
#define ___LD_R1_TO_R5 ___D_R5 ___LD_R1_TO_R4 ___R_R5
#define ___ST_R1_TO_R5 ___ST_R1_TO_R4 ___W_R5

#if ___NB_ARG_REGS == 5
#define ___PUSH_ARGS_IN_REGS(na) \
switch (na) \
  { \
  case 0:  break; \
  case 1:  ___PUSH_R1_TO_R1 break; \
  case 2:  ___PUSH_R1_TO_R2 break; \
  case 3:  ___PUSH_R1_TO_R3 break; \
  case 4:  ___PUSH_R1_TO_R4 break; \
  default: ___PUSH_R1_TO_R5 \
  }
#define ___POP_ARGS_IN_REGS(na) \
switch (na) \
  { \
  case 0:  break; \
  case 1:  ___POP_R1_TO_R1 break; \
  case 2:  ___POP_R2_TO_R1 break; \
  case 3:  ___POP_R3_TO_R1 break; \
  case 4:  ___POP_R4_TO_R1 break; \
  default: ___POP_R5_TO_R1 \
  }
#define ___PSSELF ___PSR6
#define ___SELF ___R6
#define ___LD_ARG_REGS ___LD_R1_TO_R5
#define ___ST_ARG_REGS ___ST_R1_TO_R5
#endif

#if ___NB_GVM_REGS == 6

#define ___PUSH_REGS ___PUSH_R0 ___PUSH_R1_TO_R5
#define ___POP_REGS ___POP_R5_TO_R1 ___POP_R0

#else

#define ___PSR6 ___ps->r[6]
#if ___MAX_LOCAL_GVM_REGS > 6
#define ___D_R6 register ___WORD ___r6;
#define ___R_R6 ___r6=___PSR6;
#define ___W_R6 ___PSR6=___r6;
#define ___R6 ___r6
#else
#define ___D_R6
#define ___R_R6
#define ___W_R6
#define ___R6 ___PSR6
#endif
#define ___SET_R6(val)___R6=(val);
#define ___PUSH_R1_TO_R6 ___PUSH_R1_TO_R5 ___PUSH(___R6)
#define ___POP_R6_TO_R1 ___SET_R6(___POP)___POP_R5_TO_R1
#define ___LD_R1_TO_R6 ___D_R6 ___LD_R1_TO_R5 ___R_R6
#define ___ST_R1_TO_R6 ___ST_R1_TO_R5 ___W_R6

#if ___NB_ARG_REGS == 6
#define ___PUSH_ARGS_IN_REGS(na) \
switch (na) \
  { \
  case 0:  break; \
  case 1:  ___PUSH_R1_TO_R1 break; \
  case 2:  ___PUSH_R1_TO_R2 break; \
  case 3:  ___PUSH_R1_TO_R3 break; \
  case 4:  ___PUSH_R1_TO_R4 break; \
  case 5:  ___PUSH_R1_TO_R5 break; \
  default: ___PUSH_R1_TO_R6 \
  }
#define ___POP_ARGS_IN_REGS(na) \
switch (na) \
  { \
  case 0:  break; \
  case 1:  ___POP_R1_TO_R1 break; \
  case 2:  ___POP_R2_TO_R1 break; \
  case 3:  ___POP_R3_TO_R1 break; \
  case 4:  ___POP_R4_TO_R1 break; \
  case 5:  ___POP_R5_TO_R1 break; \
  default: ___POP_R6_TO_R1 \
  }
#define ___PSSELF ___PSR7
#define ___SELF ___R7
#define ___LD_ARG_REGS ___LD_R1_TO_R6
#define ___ST_ARG_REGS ___ST_R1_TO_R6
#endif

#if ___NB_GVM_REGS == 7

#define ___PUSH_REGS ___PUSH_R0 ___PUSH_R1_TO_R6
#define ___POP_REGS ___POP_R6_TO_R1 ___POP_R0

#else

#define ___PSR7 ___ps->r[7]
#if ___MAX_LOCAL_GVM_REGS > 7
#define ___D_R7 register ___WORD ___r7;
#define ___R_R7 ___r7=___PSR7;
#define ___W_R7 ___PSR7=___r7;
#define ___R7 ___r7
#else
#define ___D_R7
#define ___R_R7
#define ___W_R7
#define ___R7 ___PSR7
#endif
#define ___SET_R7(val)___R7=(val);
#define ___PUSH_R1_TO_R7 ___PUSH_R1_TO_R6 ___PUSH(___R7)
#define ___POP_R7_TO_R1 ___SET_R7(___POP)___POP_R6_TO_R1
#define ___LD_R1_TO_R7 ___D_R7 ___LD_R1_TO_R6 ___R_R7
#define ___ST_R1_TO_R7 ___ST_R1_TO_R6 ___W_R7

#if ___NB_ARG_REGS == 7
#define ___PUSH_ARGS_IN_REGS(na) \
switch (na) \
  { \
  case 0:  break; \
  case 1:  ___PUSH_R1_TO_R1 break; \
  case 2:  ___PUSH_R1_TO_R2 break; \
  case 3:  ___PUSH_R1_TO_R3 break; \
  case 4:  ___PUSH_R1_TO_R4 break; \
  case 5:  ___PUSH_R1_TO_R5 break; \
  case 6:  ___PUSH_R1_TO_R6 break; \
  default: ___PUSH_R1_TO_R7 \
  }
#define ___POP_ARGS_IN_REGS(na) \
switch (na) \
  { \
  case 0:  break; \
  case 1:  ___POP_R1_TO_R1 break; \
  case 2:  ___POP_R2_TO_R1 break; \
  case 3:  ___POP_R3_TO_R1 break; \
  case 4:  ___POP_R4_TO_R1 break; \
  case 5:  ___POP_R5_TO_R1 break; \
  case 6:  ___POP_R6_TO_R1 break; \
  default: ___POP_R7_TO_R1 \
  }
#define ___PSSELF ___PSR8
#define ___SELF ___R8
#define ___LD_ARG_REGS ___LD_R1_TO_R7
#define ___ST_ARG_REGS ___ST_R1_TO_R7
#endif

#if ___NB_GVM_REGS == 8

#define ___PUSH_REGS ___PUSH_R0 ___PUSH_R1_TO_R7
#define ___POP_REGS ___POP_R7_TO_R1 ___POP_R0

#else

#define ___PSR8 ___ps->r[8]
#if ___MAX_LOCAL_GVM_REGS > 8
#define ___D_R8 register ___WORD ___r8;
#define ___R_R8 ___r8=___PSR8;
#define ___W_R8 ___PSR8=___r8;
#define ___R8 ___r8
#else
#define ___D_R8
#define ___R_R8
#define ___W_R8
#define ___R8 ___PSR8
#endif
#define ___SET_R8(val)___R8=(val);
#define ___PUSH_R1_TO_R8 ___PUSH_R1_TO_R7 ___PUSH(___R8)
#define ___POP_R8_TO_R1 ___SET_R8(___POP)___POP_R7_TO_R1
#define ___LD_R1_TO_R8 ___D_R8 ___LD_R1_TO_R7 ___R_R8
#define ___ST_R1_TO_R8 ___ST_R1_TO_R7 ___W_R8

#if ___NB_ARG_REGS == 8
#define ___PUSH_ARGS_IN_REGS(na) \
switch (na) \
  { \
  case 0:  break; \
  case 1:  ___PUSH_R1_TO_R1 break; \
  case 2:  ___PUSH_R1_TO_R2 break; \
  case 3:  ___PUSH_R1_TO_R3 break; \
  case 4:  ___PUSH_R1_TO_R4 break; \
  case 5:  ___PUSH_R1_TO_R5 break; \
  case 6:  ___PUSH_R1_TO_R6 break; \
  case 7:  ___PUSH_R1_TO_R7 break; \
  default: ___PUSH_R1_TO_R8 \
  }
#define ___POP_ARGS_IN_REGS(na) \
switch (na) \
  { \
  case 0:  break; \
  case 1:  ___POP_R1_TO_R1 break; \
  case 2:  ___POP_R2_TO_R1 break; \
  case 3:  ___POP_R3_TO_R1 break; \
  case 4:  ___POP_R4_TO_R1 break; \
  case 5:  ___POP_R5_TO_R1 break; \
  case 6:  ___POP_R6_TO_R1 break; \
  case 7:  ___POP_R7_TO_R1 break; \
  default: ___POP_R8_TO_R1 \
  }
#define ___PSSELF ___PSR9
#define ___SELF ___R9
#define ___LD_ARG_REGS ___LD_R1_TO_R8
#define ___ST_ARG_REGS ___ST_R1_TO_R8
#endif

#if ___NB_GVM_REGS == 9

#define ___PUSH_REGS ___PUSH_R0 ___PUSH_R1_TO_R8
#define ___POP_REGS ___POP_R8_TO_R1 ___POP_R0

#else

#define ___PSR9 ___ps->r[9]
#if ___MAX_LOCAL_GVM_REGS > 9
#define ___D_R9 register ___WORD ___r9;
#define ___R_R9 ___r9=___PSR9;
#define ___W_R9 ___PSR9=___r9;
#define ___R9 ___r9
#else
#define ___D_R9
#define ___R_R9
#define ___W_R9
#define ___R9 ___PSR9
#endif
#define ___SET_R9(val)___R9=(val);
#define ___PUSH_R1_TO_R9 ___PUSH_R1_TO_R8 ___PUSH(___R9)
#define ___POP_R9_TO_R1 ___SET_R9(___POP)___POP_R8_TO_R1
#define ___LD_R1_TO_R9 ___D_R9 ___LD_R1_TO_R8 ___R_R9
#define ___ST_R1_TO_R9 ___ST_R1_TO_R8 ___W_R9

#if ___NB_ARG_REGS == 9
#define ___PUSH_ARGS_IN_REGS(na) \
switch (na) \
  { \
  case 0:  break; \
  case 1:  ___PUSH_R1_TO_R1 break; \
  case 2:  ___PUSH_R1_TO_R2 break; \
  case 3:  ___PUSH_R1_TO_R3 break; \
  case 4:  ___PUSH_R1_TO_R4 break; \
  case 5:  ___PUSH_R1_TO_R5 break; \
  case 6:  ___PUSH_R1_TO_R6 break; \
  case 7:  ___PUSH_R1_TO_R7 break; \
  case 8:  ___PUSH_R1_TO_R8 break; \
  default: ___PUSH_R1_TO_R9 \
  }
#define ___POP_ARGS_IN_REGS(na) \
switch (na) \
  { \
  case 0:  break; \
  case 1:  ___POP_R1_TO_R1 break; \
  case 2:  ___POP_R2_TO_R1 break; \
  case 3:  ___POP_R3_TO_R1 break; \
  case 4:  ___POP_R4_TO_R1 break; \
  case 5:  ___POP_R5_TO_R1 break; \
  case 6:  ___POP_R6_TO_R1 break; \
  case 7:  ___POP_R7_TO_R1 break; \
  case 8:  ___POP_R8_TO_R1 break; \
  default: ___POP_R9_TO_R1 \
  }
#define ___PSSELF ___PSR10
#define ___SELF ___R10
#define ___LD_ARG_REGS ___LD_R1_TO_R9
#define ___ST_ARG_REGS ___ST_R1_TO_R9
#endif

#if ___NB_GVM_REGS == 10

#define ___PUSH_REGS ___PUSH_R0 ___PUSH_R1_TO_R9
#define ___POP_REGS ___POP_R9_TO_R1 ___POP_R0

#else

#define ___PSR10 ___ps->r[10]
#if ___MAX_LOCAL_GVM_REGS > 10
#define ___D_R10 register ___WORD ___r10;
#define ___R_R10 ___r10=___PSR10;
#define ___W_R10 ___PSR10=___r10;
#define ___R10 ___r10
#else
#define ___D_R10
#define ___R_R10
#define ___W_R10
#define ___R10 ___PSR10
#endif
#define ___SET_R10(val)___R10=(val);
#define ___PUSH_R1_TO_R10 ___PUSH_R1_TO_R9 ___PUSH(___R10)
#define ___POP_R10_TO_R1 ___SET_R10(___POP)___POP_R9_TO_R1
#define ___LD_R1_TO_R10 ___D_R10 ___LD_R1_TO_R9 ___R_R10
#define ___ST_R1_TO_R10 ___ST_R1_TO_R9 ___W_R10

#if ___NB_ARG_REGS == 10
#define ___PUSH_ARGS_IN_REGS(na) \
switch (na) \
  { \
  case 0:  break; \
  case 1:  ___PUSH_R1_TO_R1 break; \
  case 2:  ___PUSH_R1_TO_R2 break; \
  case 3:  ___PUSH_R1_TO_R3 break; \
  case 4:  ___PUSH_R1_TO_R4 break; \
  case 5:  ___PUSH_R1_TO_R5 break; \
  case 6:  ___PUSH_R1_TO_R6 break; \
  case 7:  ___PUSH_R1_TO_R7 break; \
  case 8:  ___PUSH_R1_TO_R8 break; \
  case 9:  ___PUSH_R1_TO_R9 break; \
  default: ___PUSH_R1_TO_R10 \
  }
#define ___POP_ARGS_IN_REGS(na) \
switch (na) \
  { \
  case 0:  break; \
  case 1:  ___POP_R1_TO_R1 break; \
  case 2:  ___POP_R2_TO_R1 break; \
  case 3:  ___POP_R3_TO_R1 break; \
  case 4:  ___POP_R4_TO_R1 break; \
  case 5:  ___POP_R5_TO_R1 break; \
  case 6:  ___POP_R6_TO_R1 break; \
  case 7:  ___POP_R7_TO_R1 break; \
  case 8:  ___POP_R8_TO_R1 break; \
  case 9:  ___POP_R9_TO_R1 break; \
  default: ___POP_R10_TO_R1 \
  }
#define ___PSSELF ___PSR11
#define ___SELF ___R11
#define ___LD_ARG_REGS ___LD_R1_TO_R10
#define ___ST_ARG_REGS ___ST_R1_TO_R10
#endif

#if ___NB_GVM_REGS == 11

#define ___PUSH_REGS ___PUSH_R0 ___PUSH_R1_TO_R10
#define ___POP_REGS ___POP_R10_TO_R1 ___POP_R0

#else

#define ___PSR11 ___ps->r[11]
#if ___MAX_LOCAL_GVM_REGS > 11
#define ___D_R11 register ___WORD ___r11;
#define ___R_R11 ___r11=___PSR11;
#define ___W_R11 ___PSR11=___r11;
#define ___R11 ___r11
#else
#define ___D_R11
#define ___R_R11
#define ___W_R11
#define ___R11 ___PSR11
#endif
#define ___SET_R11(val)___R11=(val);
#define ___PUSH_R1_TO_R11 ___PUSH_R1_TO_R10 ___PUSH(___R11)
#define ___POP_R11_TO_R1 ___SET_R11(___POP)___POP_R10_TO_R1
#define ___LD_R1_TO_R11 ___D_R11 ___LD_R1_TO_R10 ___R_R11
#define ___ST_R1_TO_R11 ___ST_R1_TO_R10 ___W_R11

#if ___NB_ARG_REGS == 11
#define ___PUSH_ARGS_IN_REGS(na) \
switch (na) \
  { \
  case 0:  break; \
  case 1:  ___PUSH_R1_TO_R1 break; \
  case 2:  ___PUSH_R1_TO_R2 break; \
  case 3:  ___PUSH_R1_TO_R3 break; \
  case 4:  ___PUSH_R1_TO_R4 break; \
  case 5:  ___PUSH_R1_TO_R5 break; \
  case 6:  ___PUSH_R1_TO_R6 break; \
  case 7:  ___PUSH_R1_TO_R7 break; \
  case 8:  ___PUSH_R1_TO_R8 break; \
  case 9:  ___PUSH_R1_TO_R9 break; \
  case 10:  ___PUSH_R1_TO_R10 break; \
  default: ___PUSH_R1_TO_R11 \
  }
#define ___POP_ARGS_IN_REGS(na) \
switch (na) \
  { \
  case 0:  break; \
  case 1:  ___POP_R1_TO_R1 break; \
  case 2:  ___POP_R2_TO_R1 break; \
  case 3:  ___POP_R3_TO_R1 break; \
  case 4:  ___POP_R4_TO_R1 break; \
  case 5:  ___POP_R5_TO_R1 break; \
  case 6:  ___POP_R6_TO_R1 break; \
  case 7:  ___POP_R7_TO_R1 break; \
  case 8:  ___POP_R8_TO_R1 break; \
  case 9:  ___POP_R9_TO_R1 break; \
  case 10:  ___POP_R10_TO_R1 break; \
  default: ___POP_R11_TO_R1 \
  }
#define ___PSSELF ___PSR12
#define ___SELF ___R12
#define ___LD_ARG_REGS ___LD_R1_TO_R11
#define ___ST_ARG_REGS ___ST_R1_TO_R11
#endif

#if ___NB_GVM_REGS == 12

#define ___PUSH_REGS ___PUSH_R0 ___PUSH_R1_TO_R11
#define ___POP_REGS ___POP_R11_TO_R1 ___POP_R0

#else

#define ___PSR12 ___ps->r[12]
#if ___MAX_LOCAL_GVM_REGS > 12
#define ___D_R12 register ___WORD ___r12;
#define ___R_R12 ___r12=___PSR12;
#define ___W_R12 ___PSR12=___r12;
#define ___R12 ___r12
#else
#define ___D_R12
#define ___R_R12
#define ___W_R12
#define ___R12 ___PSR12
#endif
#define ___SET_R12(val)___R12=(val);
#define ___PUSH_R1_TO_R12 ___PUSH_R1_TO_R11 ___PUSH(___R12)
#define ___POP_R12_TO_R1 ___SET_R12(___POP)___POP_R11_TO_R1
#define ___LD_R1_TO_R12 ___D_R12 ___LD_R1_TO_R11 ___R_R12
#define ___ST_R1_TO_R12 ___ST_R1_TO_R11 ___W_R12

#if ___NB_ARG_REGS == 12
#define ___PUSH_ARGS_IN_REGS(na) \
switch (na) \
  { \
  case 0:  break; \
  case 1:  ___PUSH_R1_TO_R1 break; \
  case 2:  ___PUSH_R1_TO_R2 break; \
  case 3:  ___PUSH_R1_TO_R3 break; \
  case 4:  ___PUSH_R1_TO_R4 break; \
  case 5:  ___PUSH_R1_TO_R5 break; \
  case 6:  ___PUSH_R1_TO_R6 break; \
  case 7:  ___PUSH_R1_TO_R7 break; \
  case 8:  ___PUSH_R1_TO_R8 break; \
  case 9:  ___PUSH_R1_TO_R9 break; \
  case 10:  ___PUSH_R1_TO_R10 break; \
  case 11:  ___PUSH_R1_TO_R11 break; \
  default: ___PUSH_R1_TO_R12 \
  }
#define ___POP_ARGS_IN_REGS(na) \
switch (na) \
  { \
  case 0:  break; \
  case 1:  ___POP_R1_TO_R1 break; \
  case 2:  ___POP_R2_TO_R1 break; \
  case 3:  ___POP_R3_TO_R1 break; \
  case 4:  ___POP_R4_TO_R1 break; \
  case 5:  ___POP_R5_TO_R1 break; \
  case 6:  ___POP_R6_TO_R1 break; \
  case 7:  ___POP_R7_TO_R1 break; \
  case 8:  ___POP_R8_TO_R1 break; \
  case 9:  ___POP_R9_TO_R1 break; \
  case 10:  ___POP_R10_TO_R1 break; \
  case 11:  ___POP_R11_TO_R1 break; \
  default: ___POP_R12_TO_R1 \
  }
#define ___PSSELF ___PSR13
#define ___SELF ___R13
#define ___LD_ARG_REGS ___LD_R1_TO_R12
#define ___ST_ARG_REGS ___ST_R1_TO_R12
#endif

#if ___NB_GVM_REGS == 13

#define ___PUSH_REGS ___PUSH_R0 ___PUSH_R1_TO_R12
#define ___POP_REGS ___POP_R12_TO_R1 ___POP_R0

#else

#define ___PSR13 ___ps->r[13]
#if ___MAX_LOCAL_GVM_REGS > 13
#define ___D_R13 register ___WORD ___r13;
#define ___R_R13 ___r13=___PSR13;
#define ___W_R13 ___PSR13=___r13;
#define ___R13 ___r13
#else
#define ___D_R13
#define ___R_R13
#define ___W_R13
#define ___R13 ___PSR13
#endif
#define ___SET_R13(val)___R13=(val);
#define ___PUSH_R1_TO_R13 ___PUSH_R1_TO_R12 ___PUSH(___R13)
#define ___POP_R13_TO_R1 ___SET_R13(___POP)___POP_R12_TO_R1
#define ___LD_R1_TO_R13 ___D_R13 ___LD_R1_TO_R12 ___R_R13
#define ___ST_R1_TO_R13 ___ST_R1_TO_R12 ___W_R13

#if ___NB_GVM_REGS == 14

#define ___PUSH_REGS ___PUSH_R0 ___PUSH_R1_TO_R13
#define ___POP_REGS ___POP_R13_TO_R1 ___POP_R0

#else

#define ___PSR14 ___ps->r[14]
#if ___MAX_LOCAL_GVM_REGS > 14
#define ___D_R14 register ___WORD ___r14;
#define ___R_R14 ___r14=___PSR14;
#define ___W_R14 ___PSR14=___r14;
#define ___R14 ___r14
#else
#define ___D_R14
#define ___R_R14
#define ___W_R14
#define ___R14 ___PSR14
#endif
#define ___SET_R14(val)___R14=(val);
#define ___PUSH_R1_TO_R14 ___PUSH_R1_TO_R13 ___PUSH(___R14)
#define ___POP_R14_TO_R1 ___SET_R14(___POP)___POP_R13_TO_R1
#define ___LD_R1_TO_R14 ___D_R14 ___LD_R1_TO_R13 ___R_R14
#define ___ST_R1_TO_R14 ___ST_R1_TO_R13 ___W_R14

#if ___NB_GVM_REGS == 15

#define ___PUSH_REGS ___PUSH_R0 ___PUSH_R1_TO_R14
#define ___POP_REGS ___POP_R14_TO_R1 ___POP_R0

#else

#define ___PSR15 ___ps->r[15]
#if ___MAX_LOCAL_GVM_REGS > 15
#define ___D_R15 register ___WORD ___r15;
#define ___R_R15 ___r15=___PSR15;
#define ___W_R15 ___PSR15=___r15;
#define ___R15 ___r15
#else
#define ___D_R15
#define ___R_R15
#define ___W_R15
#define ___R15 ___PSR15
#endif
#define ___SET_R15(val)___R15=(val);
#define ___PUSH_R1_TO_R15 ___PUSH_R1_TO_R14 ___PUSH(___R15)
#define ___POP_R15_TO_R1 ___SET_R15(___POP)___POP_R14_TO_R1
#define ___LD_R1_TO_R15 ___D_R15 ___LD_R1_TO_R14 ___R_R15
#define ___ST_R1_TO_R15 ___ST_R1_TO_R14 ___W_R15

#if ___NB_GVM_REGS == 16

#define ___PUSH_REGS ___PUSH_R0 ___PUSH_R1_TO_R15
#define ___POP_REGS ___POP_R15_TO_R1 ___POP_R0

#else

#define ___PSR16 ___ps->r[16]
#if ___MAX_LOCAL_GVM_REGS > 16
#define ___D_R16 register ___WORD ___r16;
#define ___R_R16 ___r16=___PSR16;
#define ___W_R16 ___PSR16=___r16;
#define ___R16 ___r16
#else
#define ___D_R16
#define ___R_R16
#define ___W_R16
#define ___R16 ___PSR16
#endif
#define ___SET_R16(val)___R16=(val);
#define ___PUSH_R1_TO_R16 ___PUSH_R1_TO_R15 ___PUSH(___R16)
#define ___POP_R16_TO_R1 ___SET_R16(___POP)___POP_R15_TO_R1
#define ___LD_R1_TO_R16 ___D_R16 ___LD_R1_TO_R15 ___R_R16
#define ___ST_R1_TO_R16 ___ST_R1_TO_R15 ___W_R16

#if ___NB_GVM_REGS == 17

#define ___PUSH_REGS ___PUSH_R0 ___PUSH_R1_TO_R16
#define ___POP_REGS ___POP_R16_TO_R1 ___POP_R0

#else

#define ___PSR17 ___ps->r[17]
#if ___MAX_LOCAL_GVM_REGS > 17
#define ___D_R17 register ___WORD ___r17;
#define ___R_R17 ___r17=___PSR17;
#define ___W_R17 ___PSR17=___r17;
#define ___R17 ___r17
#else
#define ___D_R17
#define ___R_R17
#define ___W_R17
#define ___R17 ___PSR17
#endif
#define ___SET_R17(val)___R17=(val);
#define ___PUSH_R1_TO_R17 ___PUSH_R1_TO_R16 ___PUSH(___R17)
#define ___POP_R17_TO_R1 ___SET_R17(___POP)___POP_R16_TO_R1
#define ___LD_R1_TO_R17 ___D_R17 ___LD_R1_TO_R16 ___R_R17
#define ___ST_R1_TO_R17 ___ST_R1_TO_R16 ___W_R17

#if ___NB_GVM_REGS == 18

#define ___PUSH_REGS ___PUSH_R0 ___PUSH_R1_TO_R17
#define ___POP_REGS ___POP_R17_TO_R1 ___POP_R0

#else

#define ___PSR18 ___ps->r[18]
#if ___MAX_LOCAL_GVM_REGS > 18
#define ___D_R18 register ___WORD ___r18;
#define ___R_R18 ___r18=___PSR18;
#define ___W_R18 ___PSR18=___r18;
#define ___R18 ___r18
#else
#define ___D_R18
#define ___R_R18
#define ___W_R18
#define ___R18 ___PSR18
#endif
#define ___SET_R18(val)___R18=(val);
#define ___PUSH_R1_TO_R18 ___PUSH_R1_TO_R17 ___PUSH(___R18)
#define ___POP_R18_TO_R1 ___SET_R18(___POP)___POP_R17_TO_R1
#define ___LD_R1_TO_R18 ___D_R18 ___LD_R1_TO_R17 ___R_R18
#define ___ST_R1_TO_R18 ___ST_R1_TO_R17 ___W_R18

#if ___NB_GVM_REGS == 19

#define ___PUSH_REGS ___PUSH_R0 ___PUSH_R1_TO_R18
#define ___POP_REGS ___POP_R18_TO_R1 ___POP_R0

#else

#define ___PSR19 ___ps->r[19]
#if ___MAX_LOCAL_GVM_REGS > 19
#define ___D_R19 register ___WORD ___r19;
#define ___R_R19 ___r19=___PSR19;
#define ___W_R19 ___PSR19=___r19;
#define ___R19 ___r19
#else
#define ___D_R19
#define ___R_R19
#define ___W_R19
#define ___R19 ___PSR19
#endif
#define ___SET_R19(val)___R19=(val);
#define ___PUSH_R1_TO_R19 ___PUSH_R1_TO_R18 ___PUSH(___R19)
#define ___POP_R19_TO_R1 ___SET_R19(___POP)___POP_R18_TO_R1
#define ___LD_R1_TO_R19 ___D_R19 ___LD_R1_TO_R18 ___R_R19
#define ___ST_R1_TO_R19 ___ST_R1_TO_R18 ___W_R19

#if ___NB_GVM_REGS == 20

#define ___PUSH_REGS ___PUSH_R0 ___PUSH_R1_TO_R19
#define ___POP_REGS ___POP_R19_TO_R1 ___POP_R0

#else

#define ___PSR20 ___ps->r[20]
#if ___MAX_LOCAL_GVM_REGS > 20
#define ___D_R20 register ___WORD ___r20;
#define ___R_R20 ___r20=___PSR20;
#define ___W_R20 ___PSR20=___r20;
#define ___R20 ___r20
#else
#define ___D_R20
#define ___R_R20
#define ___W_R20
#define ___R20 ___PSR20
#endif
#define ___SET_R20(val)___R20=(val);
#define ___PUSH_R1_TO_R20 ___PUSH_R1_TO_R19 ___PUSH(___R20)
#define ___POP_R20_TO_R1 ___SET_R20(___POP)___POP_R19_TO_R1
#define ___LD_R1_TO_R20 ___D_R20 ___LD_R1_TO_R19 ___R_R20
#define ___ST_R1_TO_R20 ___ST_R1_TO_R19 ___W_R20

#if ___NB_GVM_REGS == 21

#define ___PUSH_REGS ___PUSH_R0 ___PUSH_R1_TO_R20
#define ___POP_REGS ___POP_R20_TO_R1 ___POP_R0

#else

#define ___PSR21 ___ps->r[21]
#if ___MAX_LOCAL_GVM_REGS > 21
#define ___D_R21 register ___WORD ___r21;
#define ___R_R21 ___r21=___PSR21;
#define ___W_R21 ___PSR21=___r21;
#define ___R21 ___r21
#else
#define ___D_R21
#define ___R_R21
#define ___W_R21
#define ___R21 ___PSR21
#endif
#define ___SET_R21(val)___R21=(val);
#define ___PUSH_R1_TO_R21 ___PUSH_R1_TO_R20 ___PUSH(___R21)
#define ___POP_R21_TO_R1 ___SET_R21(___POP)___POP_R20_TO_R1
#define ___LD_R1_TO_R21 ___D_R21 ___LD_R1_TO_R20 ___R_R21
#define ___ST_R1_TO_R21 ___ST_R1_TO_R20 ___W_R21

#if ___NB_GVM_REGS == 22

#define ___PUSH_REGS ___PUSH_R0 ___PUSH_R1_TO_R21
#define ___POP_REGS ___POP_R21_TO_R1 ___POP_R0

#else

#define ___PSR22 ___ps->r[22]
#if ___MAX_LOCAL_GVM_REGS > 22
#define ___D_R22 register ___WORD ___r22;
#define ___R_R22 ___r22=___PSR22;
#define ___W_R22 ___PSR22=___r22;
#define ___R22 ___r22
#else
#define ___D_R22
#define ___R_R22
#define ___W_R22
#define ___R22 ___PSR22
#endif
#define ___SET_R22(val)___R22=(val);
#define ___PUSH_R1_TO_R22 ___PUSH_R1_TO_R21 ___PUSH(___R22)
#define ___POP_R22_TO_R1 ___SET_R22(___POP)___POP_R21_TO_R1
#define ___LD_R1_TO_R22 ___D_R22 ___LD_R1_TO_R21 ___R_R22
#define ___ST_R1_TO_R22 ___ST_R1_TO_R21 ___W_R22

#if ___NB_GVM_REGS == 23

#define ___PUSH_REGS ___PUSH_R0 ___PUSH_R1_TO_R22
#define ___POP_REGS ___POP_R22_TO_R1 ___POP_R0

#else

#define ___PSR23 ___ps->r[23]
#if ___MAX_LOCAL_GVM_REGS > 23
#define ___D_R23 register ___WORD ___r23;
#define ___R_R23 ___r23=___PSR23;
#define ___W_R23 ___PSR23=___r23;
#define ___R23 ___r23
#else
#define ___D_R23
#define ___R_R23
#define ___W_R23
#define ___R23 ___PSR23
#endif
#define ___SET_R23(val)___R23=(val);
#define ___PUSH_R1_TO_R23 ___PUSH_R1_TO_R22 ___PUSH(___R23)
#define ___POP_R23_TO_R1 ___SET_R23(___POP)___POP_R22_TO_R1
#define ___LD_R1_TO_R23 ___D_R23 ___LD_R1_TO_R22 ___R_R23
#define ___ST_R1_TO_R23 ___ST_R1_TO_R22 ___W_R23

#if ___NB_GVM_REGS == 24

#define ___PUSH_REGS ___PUSH_R0 ___PUSH_R1_TO_R23
#define ___POP_REGS ___POP_R23_TO_R1 ___POP_R0

#else

#define ___PSR24 ___ps->r[24]
#if ___MAX_LOCAL_GVM_REGS > 24
#define ___D_R24 register ___WORD ___r24;
#define ___R_R24 ___r24=___PSR24;
#define ___W_R24 ___PSR24=___r24;
#define ___R24 ___r24
#else
#define ___D_R24
#define ___R_R24
#define ___W_R24
#define ___R24 ___PSR24
#endif
#define ___SET_R24(val)___R24=(val);
#define ___PUSH_R1_TO_R24 ___PUSH_R1_TO_R23 ___PUSH(___R24)
#define ___POP_R24_TO_R1 ___SET_R24(___POP)___POP_R23_TO_R1
#define ___LD_R1_TO_R24 ___D_R24 ___LD_R1_TO_R23 ___R_R24
#define ___ST_R1_TO_R24 ___ST_R1_TO_R23 ___W_R24

#if ___NB_GVM_REGS == 25

#define ___PUSH_REGS ___PUSH_R0 ___PUSH_R1_TO_R24
#define ___POP_REGS ___POP_R24_TO_R1 ___POP_R0

#endif
#endif
#endif
#endif
#endif
#endif
#endif
#endif
#endif
#endif
#endif
#endif
#endif
#endif
#endif
#endif
#endif
#endif
#endif
#endif
#endif
#endif
#endif
#endif

/*---------------------------------------------------------------------------*/

/* Labels and switches */

#define ___NOT(x)!x
#define ___AND(x,y)(x&&y)

#define ___MLBL(n)___lp+((n)*___LS*___WS)
#define ___LBL(n)___start+((___PH_LBL0-___HOST_LBL0+n)*___LS*___WS)
#define ___DEF_SLBL(n,lbl)___DEF_SW(n)___DEF_GLBL(lbl)
#define ___DEF_GLBL(lbl)lbl:
#define ___GOTO(lbl)goto lbl;
#define ___IF_GOTO(test,lbl)if(test)___GOTO(lbl)
#define ___JUMPINT(nargs,prc,lbl)___SM(___GOTO(lbl),{nargs ___JUMP(prc)})
#define ___JUMPEXTPRM(nargs,val){nargs ___JUMPEXT(val)}
#define ___JUMPEXTNOTSAFE(nargs,val) \
{nargs ___SELF=val; \
___JUMPEXT(((___label_struct*)(___SELF-___tSUBTYPED))->entry_or_descr)}
#define ___JUMPPRM(nargs,val){nargs ___JUMP(val)}

#define ___JUMPNOTSAFE(nargs,val) \
{nargs ___SELF=val; \
___JUMP(((___label_struct*)(___SELF-___tSUBTYPED))->entry_or_descr)}

#define ___JUMPGENNOTSAFE(nargs,val)___JUMPNOTSAFE(nargs,val)
#define ___JUMPGLONOTSAFE(nargs,i,glo)___JUMPNOTSAFE(nargs,___GLO(i,glo))
#define ___JUMPGENSAFE(nargs,val) \
{nargs ___SELF=val; \
if (___PROCEDUREP(___SELF)) \
  ___JUMP(((___label_struct*)(___SELF-___tSUBTYPED))->entry_or_descr) \
___ps->temp1=___SELF;___JUMPEXTPRM(___NOTHING,___ps->handler_not_proc)}
#define ___JUMPGLOSAFE(nargs,i,glo) \
{nargs ___SELF=___GLO(i,glo); \
if (___PROCEDUREP(___SELF)) \
  ___JUMP(((___label_struct*)(___SELF-___tSUBTYPED))->entry_or_descr) \
___ps->temp4=(___WORD)&___GLOSTRUCT(i,glo);___JUMPEXTPRM(___NOTHING,___ps->handler_not_proc_glo)}

#ifdef ___NOT_SAFE_CALLS
#undef ___JUMPGENSAFE
#undef ___JUMPGLOSAFE
#define ___JUMPGENSAFE(nargs,val)___JUMPGENNOTSAFE(nargs,val)
#define ___JUMPGLOSAFE(nargs,i,glo)___JUMPGLONOTSAFE(nargs,i,glo)
#endif

#ifdef ___SAFE_CALLS
#undef ___JUMPGENNOTSAFE
#undef ___JUMPGLONOTSAFE
#define ___JUMPGENNOTSAFE(nargs,val)___JUMPGENSAFE(nargs,val)
#define ___JUMPGLONOTSAFE(nargs,i,glo)___JUMPGLOSAFE(nargs,i,glo)
#endif

/*---------------------------------------------------------------------------*/

#define ___PRC(i)___start+((i-___HOST_LBL0)*___LS*___WS)

#define ___SET_NARGS(n)___ps->na=n;

#define ___IF_NARGS_EQ(n,code)if(___ps->na==n){code}else

#define ___WRONG_NARGS(lbl,nb_req,nb_opt,nb_key) \
{___ps->temp1=___LBL(lbl); \
___JUMPEXTPRM(___NOTHING,___ps->handler_wrong_nargs)}

#define ___GET_REST(lbl,nb_req,nb_opt,nb_key) \
if(___ps->na>=0){___ps->temp1=___LBL(lbl); \
___JUMPEXTPRM(___NOTHING,___ps->handler_get_rest)}

#define ___GET_KEY(lbl,nb_req,nb_opt,nb_key,key_descr) \
if(___ps->na>=0){___ps->temp1=___LBL(lbl);___ps->temp2=nb_req+nb_opt; \
___ps->temp3=key_descr;___JUMPEXTPRM(___NOTHING,___ps->handler_get_key)}

#define ___GET_KEY_REST(lbl,nb_req,nb_opt,nb_key,key_descr) \
if(___ps->na>=0){___ps->temp1=___LBL(lbl);___ps->temp2=nb_req+nb_opt; \
___ps->temp3=key_descr;___JUMPEXTPRM(___NOTHING,___ps->handler_get_key_rest)}

#define ___BOOLEAN(x)(x)?___TRU:___FAL
#define ___EXPR(x)x;

#define ___ALLOC_CLO(n)(___hp+=n+2,___TAG((___hp-n-2),___tSUBTYPED))
#define ___BEGIN_SETUP_CLO(n,clo,lbl) \
{___WORD *___ptr=___UNTAG_AS(clo,___tSUBTYPED); \
___ptr[0]=___MAKE_HD_WORDS((n+1),___sPROCEDURE); ___ptr[1]=___LBL(lbl);
#define ___ADD_CLO_ELEM(i,val)___ptr[i+2]=(val);
#define ___END_SETUP_CLO(n) }
#define ___CLO(x,y)___BODY_AS(x,___tSUBTYPED)[y]
#define ___SET_CLO(x,y,z)___BODY_AS(x,___tSUBTYPED)[y]=z;

#define ___BEGIN_ALLOC_LIST(n,last)___ALLOC_PAIR(last,___NUL)
#define ___ADD_LIST_ELEM(i,val)___ALLOC_PAIR(val,___TAG((___hp-___PAIR_SIZE-1),___tPAIR))
#define ___END_ALLOC_LIST(n)
#define ___GET_LIST(n) ___TAG((___hp-___PAIR_SIZE-1),___tPAIR)

#define ___BEGIN_ALLOC_STRING(n) \
___hp[0]=___MAKE_HD_BYTES((n<<___LCS),___sSTRING);
#define ___ADD_STRING_ELEM(i,val)___CS_SELECT( \
___STORE_U8(___hp+1,i,___INT(val)), \
___STORE_U16(___hp+1,i,___INT(val)), \
___STORE_U32(___hp+1,i,___INT(val)));
#define ___END_ALLOC_STRING(n)___hp+=___WORDS((n<<___LCS))+1;
#define ___GET_STRING(n)___TAG((___hp-___WORDS((n<<___LCS))-1),___tSUBTYPED)

#define ___BEGIN_ALLOC_S8VECTOR(n) \
___hp[0]=___MAKE_HD_BYTES(n,___sS8VECTOR);
#define ___ADD_S8VECTOR_ELEM(i,val)___STORE_S8(___hp+1,i,___INT(val));
#define ___END_ALLOC_S8VECTOR(n)___hp+=___WORDS(n)+1;
#define ___GET_S8VECTOR(n)___TAG((___hp-___WORDS(n)-1),___tSUBTYPED)

#define ___BEGIN_ALLOC_U8VECTOR(n) \
___hp[0]=___MAKE_HD_BYTES(n,___sU8VECTOR);
#define ___ADD_U8VECTOR_ELEM(i,val)___STORE_U8(___hp+1,i,___INT(val));
#define ___END_ALLOC_U8VECTOR(n)___hp+=___WORDS(n)+1;
#define ___GET_U8VECTOR(n)___TAG((___hp-___WORDS(n)-1),___tSUBTYPED)

#define ___BEGIN_ALLOC_S16VECTOR(n) \
___hp[0]=___MAKE_HD_BYTES((n<<1),___sS16VECTOR);
#define ___ADD_S16VECTOR_ELEM(i,val)___STORE_S16(___hp+1,i,___INT(val));
#define ___END_ALLOC_S16VECTOR(n)___hp+=___WORDS((n<<1))+1;
#define ___GET_S16VECTOR(n)___TAG((___hp-___WORDS((n<<1))-1),___tSUBTYPED)

#define ___BEGIN_ALLOC_U16VECTOR(n) \
___hp[0]=___MAKE_HD_BYTES((n<<1),___sU16VECTOR);
#define ___ADD_U16VECTOR_ELEM(i,val)___STORE_U16(___hp+1,i,___INT(val));
#define ___END_ALLOC_U16VECTOR(n)___hp+=___WORDS((n<<1))+1;
#define ___GET_U16VECTOR(n)___TAG((___hp-___WORDS((n<<1))-1),___tSUBTYPED)

#define ___BEGIN_ALLOC_S32VECTOR(n) \
___hp[0]=___MAKE_HD_BYTES((n<<2),___sS32VECTOR);
#define ___ADD_S32VECTOR_ELEM(i,val)___STORE_S32(___hp+1,i,___S32UNBOX(val));
#define ___END_ALLOC_S32VECTOR(n)___hp+=___WORDS((n<<2))+1;
#define ___GET_S32VECTOR(n)___TAG((___hp-___WORDS((n<<2))-1),___tSUBTYPED)

#define ___BEGIN_ALLOC_U32VECTOR(n) \
___hp[0]=___MAKE_HD_BYTES((n<<2),___sU32VECTOR);
#define ___ADD_U32VECTOR_ELEM(i,val)___STORE_U32(___hp+1,i,___U32UNBOX(val));
#define ___END_ALLOC_U32VECTOR(n)___hp+=___WORDS((n<<2))+1;
#define ___GET_U32VECTOR(n)___TAG((___hp-___WORDS((n<<2))-1),___tSUBTYPED)

#if ___WS == 4

#define ___BEGIN_ALLOC_S64VECTOR(n) \
{___WORD *___ptr = (___WORD*)((___WORD)(___hp+2)&~7); \
___ptr[-1]=___MAKE_HD_BYTES((n<<3),___sS64VECTOR);
#define ___ADD_S64VECTOR_ELEM(i,val)___STORE_S64(___ptr,i,___S64UNBOX(val));
#define ___END_ALLOC_S64VECTOR(n)___hp+=___WORDS((n<<3))+2;}
#define ___GET_S64VECTOR(n)___TAG((((___WORD*)((___WORD)(___hp-___WORDS((n<<3)))&~7))-1),___tSUBTYPED)

#define ___BEGIN_ALLOC_U64VECTOR(n) \
{___WORD *___ptr = (___WORD*)((___WORD)(___hp+2)&~7); \
___ptr[-1]=___MAKE_HD_BYTES((n<<3),___sU64VECTOR);
#define ___ADD_U64VECTOR_ELEM(i,val)___STORE_U64(___ptr,i,___U64UNBOX(val));
#define ___END_ALLOC_U64VECTOR(n)___hp+=___WORDS((n<<3))+2;}
#define ___GET_U64VECTOR(n)___TAG((((___WORD*)((___WORD)(___hp-___WORDS((n<<3)))&~7))-1),___tSUBTYPED)

#else

#define ___BEGIN_ALLOC_S64VECTOR(n) \
___hp[0]=___MAKE_HD_BYTES((n<<3),___sS64VECTOR);
#define ___ADD_S64VECTOR_ELEM(i,val)___STORE_S64(___hp+1,i,___S64UNBOX(val));
#define ___END_ALLOC_S64VECTOR(n)___hp+=___WORDS((n<<3))+1;
#define ___GET_S64VECTOR(n)___TAG((___hp-___WORDS((n<<3))-1),___tSUBTYPED)

#define ___BEGIN_ALLOC_U64VECTOR(n) \
___hp[0]=___MAKE_HD_BYTES((n<<3),___sU64VECTOR);
#define ___ADD_U64VECTOR_ELEM(i,val)___STORE_U64(___hp+1,i,___U64UNBOX(val));
#define ___END_ALLOC_U64VECTOR(n)___hp+=___WORDS((n<<3))+1;
#define ___GET_U64VECTOR(n)___TAG((___hp-___WORDS((n<<3))-1),___tSUBTYPED)

#endif

#define ___BEGIN_ALLOC_F32VECTOR(n) \
___hp[0]=___MAKE_HD_BYTES((n<<2),___sF32VECTOR);
#ifdef ___F32
#define ___ADD_F32VECTOR_ELEM(i,val)((___F32*)(___hp+1))[i]=(val);
#else
#define ___ADD_F32VECTOR_ELEM(i,val);
#endif
#define ___END_ALLOC_F32VECTOR(n)___hp+=___WORDS((n<<2))+1;
#define ___GET_F32VECTOR(n)___TAG((___hp-___WORDS((n<<2))-1),___tSUBTYPED)

#if ___WS == 4

#define ___BEGIN_ALLOC_F64VECTOR(n) \
{___WORD *___ptr = (___WORD*)((___WORD)(___hp+2)&~7); \
___ptr[-1]=___MAKE_HD_BYTES((n<<3),___sF64VECTOR);
#define ___ADD_F64VECTOR_ELEM(i,val)((___F64*)___ptr)[i]=(val);
#define ___END_ALLOC_F64VECTOR(n)___hp+=___WORDS((n<<3))+2;}
#define ___GET_F64VECTOR(n)___TAG((((___WORD*)((___WORD)(___hp-___WORDS((n<<3)))&~7))-1),___tSUBTYPED)

#else

#define ___BEGIN_ALLOC_F64VECTOR(n) \
___hp[0]=___MAKE_HD_BYTES((n<<3),___sF64VECTOR);
#define ___ADD_F64VECTOR_ELEM(i,val)((___F64*)(___hp+1))[i]=(val);
#define ___END_ALLOC_F64VECTOR(n)___hp+=___WORDS((n<<3))+1;
#define ___GET_F64VECTOR(n)___TAG((___hp-___WORDS((n<<3))-1),___tSUBTYPED)

#endif

#define ___BEGIN_ALLOC_VECTOR(n)___hp[0]=___MAKE_HD_WORDS(n,___sVECTOR);
#define ___ADD_VECTOR_ELEM(i,val)___hp[i+1]=(val);
#define ___END_ALLOC_VECTOR(n)___hp+=n+1;
#define ___GET_VECTOR(n)___TAG((___hp-n-1),___tSUBTYPED)

#ifdef ___DEBUG_STACK_LIMIT
#define ___POLL(n) \
if(___fp<___ps->stack_trip) \
{___ps->temp1=___LBL(n); \
___ps->poll_file=__FILE__; \
___ps->poll_line=__LINE__; \
___JUMPEXTPRM(___NOTHING,___ps->handler_stack_limit) \
}
#else
#define ___POLL(n) \
if(___fp<___ps->stack_trip) \
{___ps->temp1=___LBL(n); \
___JUMPEXTPRM(___NOTHING,___ps->handler_stack_limit) \
}
#endif

#define ___TASK_PUSH(i)
#define ___TASK_POP(n)

#ifdef ___DEBUG_HEAP_LIMIT
#define ___CHECK_HEAP(n,m) \
if(___hp>___ps->heap_limit) \
{___ps->temp1=___LBL(n); \
___ps->check_heap_file=__FILE__; \
___ps->check_heap_line=__LINE__; \
___JUMPEXTPRM(___NOTHING,___ps->handler_heap_limit) \
}
#else
#define ___CHECK_HEAP(n,m) \
if(___hp>___ps->heap_limit) \
{___ps->temp1=___LBL(n); \
___JUMPEXTPRM(___NOTHING,___ps->handler_heap_limit) \
}
#endif

#define ___FORCE1(n,src) \
if (___TYP((___temp=(src)))==___tSUBTYPED&&___SUBTYPE(___temp)==___sPROMISE<<___TB){___ps->temp1=___LBL(n);___ps->temp2=___temp;___JUMPEXTPRM(___NOTHING,___ps->handler_force);
#define ___FORCE2 ___temp=___ps->temp2;}
#define ___FORCE3 ___temp

#define ___NB_INTRS 3
#define ___INTR_USER      0
#define ___INTR_HEARTBEAT 1
#define ___INTR_GC        2

#define ___IO_DEFAULT_ENCODING 0
#define ___IO_CHAR_ENCODING    1
#define ___IO_LATIN1_ENCODING  2
#define ___IO_UTF8_ENCODING    3
#define ___IO_BYTE_ENCODING    4
#define ___IO_UCS2_ENCODING    5
#define ___IO_UCS4_ENCODING    6

#define ___MODULE_KIND 0
#define ___LINKFILE_KIND 1

#define ___CLIBEXT(f)f
#define ___SETJMP(x)setjmp(x)
#define ___EXT(f)f

#ifdef ___BIND_LATE

#ifdef ___CANT_IMPORT_CLIB_DYNAMICALLY
#undef ___CLIBEXT
#define ___CLIBEXT(f)___GSTATE->f
#undef ___LOCAL_GSTATE
#define ___LOCAL_GSTATE
#endif

#ifdef ___CANT_IMPORT_SETJMP_DYNAMICALLY
#undef ___SETJMP
#define ___SETJMP(x)___GSTATE->setjmp(x)
#endif

#ifdef ___CANT_IMPORT_DYNAMICALLY
#undef ___EXT
#define ___EXT(f)___GSTATE->f
#undef ___LOCAL_GSTATE
#define ___LOCAL_GSTATE
#endif

#endif

#ifdef ___LOCAL_GSTATE

#define ___GSTATE ___local_gstate
#define ___GSTATE_DECL ___LOCAL ___global_state_struct *___local_gstate;
#define ___SET_LOCAL_GSTATE(gs) ___local_gstate = gs;

#else

#define ___GSTATE (&___gstate)
#define ___GSTATE_DECL ___IMP_DATA(___global_state_struct,___gstate);
#define ___SET_LOCAL_GSTATE(gs)

#endif

#define ___PSTATE (&___GSTATE->pstate)

#ifdef ___BIND_LATE

#define ___NEED_SYM(sym)
#define ___NEED_KEY(key)
#define ___NEED_GLO(glo)

#define ___BEGIN_SYM1 ___LOCAL ___UTF8STRING ___sym_names[]={
#define ___DEF_SYM1(i,sym,str)str,
#define ___END_SYM1 0};

#define ___DEF_SYM2(i,sym,str)

#define ___BEGIN_KEY1 ___LOCAL ___UTF8STRING ___key_names[]={
#define ___DEF_KEY1(i,key,str)str,
#define ___END_KEY1 0};

#define ___DEF_KEY2(i,key,str)

#define ___BEGIN_GLO ___LOCAL ___UTF8STRING ___glo_names[]={
#define ___DEF_GLO(i,str)str,
#define ___END_GLO 0};

#define ___BEGIN_OLD_KEY
#define ___DEF_OLD_KEY(key)
#define ___END_OLD_KEY

#define ___BEGIN_OLD_SYM_GLO
#define ___DEF_OLD_SYM_GLO(sym,glo)
#define ___END_OLD_SYM_GLO

#define ___BEGIN_NEW_KEY
#define ___DEF_NEW_KEY(prevkey,key,str,hash)
#define ___END_NEW_KEY

#define ___BEGIN_NEW_SYM_GLO
#define ___DEF_NEW_SYM_GLO_SUP(prevsym,sym,str,hash,glo)
#define ___DEF_NEW_SYM_GLO(prevsym,sym,str,hash,glo)
#define ___END_NEW_SYM_GLO(prevsym,prevkey)___LINKFILE_DESCR(0,0)

#define ___DEF_SUB_NSTR(id,n)
#define ___NSTR0
#define ___NSTR1(a)
#define ___NSTR2(a,b)
#define ___NSTR3(a,b,c)
#define ___NSTR4(a,b,c,d)
#define ___NSTR5(a,b,c,d,e)
#define ___NSTR6(a,b,c,d,e,f)
#define ___NSTR7(a,b,c,d,e,f,g)
#define ___NSTR8(a,b,c,d,e,f,g,h)

#else

#define ___NEED_SYM(sym)___IMP_DATA(___symkey_struct,sym);
#define ___NEED_KEY(key)___IMP_DATA(___symkey_struct,key);
#define ___NEED_GLO(glo)___IMP_DATA(___glo_struct,glo);

#ifdef ___IMPORTED_ADR_NOT_CONST

#define ___BEGIN_SYM1 ___LOCAL ___FAKEWORD ___sym_tbl[___SYM_COUNT];
#define ___DEF_SYM1(i,sym,str)
#define ___END_SYM1

#define ___DEF_SYM2(i,sym,str)___sym_tbl[i]=(___FAKEWORD)(sym);

#define ___BEGIN_KEY1 ___LOCAL ___FAKEWORD ___key_tbl[___KEY_COUNT];
#define ___DEF_KEY1(i,key,str)
#define ___END_KEY1

#define ___DEF_KEY2(i,key,str)___key_tbl[i]=(___FAKEWORD)(key);

#else

#define ___BEGIN_SYM1 ___LOCAL ___FAKEWORD ___sym_tbl[]={
#define ___DEF_SYM1(i,sym,str)(___FAKEWORD)(sym),
#define ___END_SYM1 0};

#define ___DEF_SYM2(i,sym,str)

#define ___BEGIN_KEY1 ___LOCAL ___FAKEWORD ___key_tbl[]={
#define ___DEF_KEY1(i,key,str)(___FAKEWORD)(key),
#define ___END_KEY1 0};

#define ___DEF_KEY2(i,key,str)

#endif

#define ___BEGIN_GLO
#define ___DEF_GLO(i,str)
#define ___END_GLO

#define ___BEGIN_OLD_KEY
#define ___DEF_OLD_KEY(key)
#define ___END_OLD_KEY

#define ___BEGIN_OLD_SYM_GLO
#define ___DEF_OLD_SYM_GLO(sym,glo)
#define ___END_OLD_SYM_GLO

#define ___BEGIN_NEW_KEY
#define ___DEF_NEW_KEY(prevkey,key,str,hash) \
___EXP_DATA(___symkey_struct,key)={(___FAKEWORD)prevkey,(___FAKEWORD)str,0,0 ___PADDING};
#define ___END_NEW_KEY

#define ___BEGIN_NEW_SYM_GLO
#define ___DEF_NEW_SYM_GLO_SUP(prevsym,sym,str,hash,glo) \
___EXP_DATA(___glo_struct,glo) = {___UNB2, ___FIX(hash), 0}; \
___EXP_DATA(___symkey_struct,sym)={(___FAKEWORD)prevsym,(___FAKEWORD)str,0,0,(___FAKEWORD)&glo ___PADDING};
#define ___DEF_NEW_SYM_GLO(prevsym,sym,str,hash,glo) \
___EXP_DATA(___glo_struct,glo) = {___UNB1, ___FIX(hash), 0}; \
___EXP_DATA(___symkey_struct,sym)={(___FAKEWORD)prevsym,(___FAKEWORD)str,0,0,(___FAKEWORD)&glo ___PADDING};
#define ___END_NEW_SYM_GLO(prevsym,prevkey)___LINKFILE_DESCR(prevsym,prevkey)

#define ___DEF_SUB_NSTR(id,n) ___DEF_SUB_STR(id,n)
#define ___NSTR0 ___STR0
#define ___NSTR1(a) ___STR1(a)
#define ___NSTR2(a,b) ___STR2(a,b)
#define ___NSTR3(a,b,c) ___STR3(a,b,c)
#define ___NSTR4(a,b,c,d) ___STR4(a,b,c,d)
#define ___NSTR5(a,b,c,d,e) ___STR5(a,b,c,d,e)
#define ___NSTR6(a,b,c,d,e,f) ___STR6(a,b,c,d,e,f)
#define ___NSTR7(a,b,c,d,e,f,g) ___STR7(a,b,c,d,e,f,g)
#define ___NSTR8(a,b,c,d,e,f,g,h) ___STR8(a,b,c,d,e,f,g,h)

#endif

#define ___BEGIN_CNS \
___LOCAL ___WORD ___cns_tbl[]={
#define ___DEF_CNS(car,cdr)___MAKE_HD((___PAIR_SIZE<<___LWS),___sPAIR,___PERM),cdr,car
#define ___END_CNS ___PADDING};

#define ___BEGIN_SUB \
___LOCAL ___FAKEWORD ___sub_tbl[]={
#define ___DEF_SUB(id)___CAST_WORDSTAR_TO_FAKEWORD(id)
#define ___END_SUB };

#define ___DEF_SUB_VEC(id,n) \
___LOCAL ___WORD id[]={___MAKE_HD((n<<___LWS),___sVECTOR,___PERM)

#define ___DEF_SUB_STR(id,n) \
___LOCAL ___WORD id[]={___MAKE_HD((n<<___LCS),___sSTRING,___PERM)

#define ___DEF_SUB_RAT(id,num,den) \
___LOCAL ___WORD id[]={___MAKE_HD((___RATNUM_SIZE<<___LWS),___sRATNUM,___PERM),num,den ___PADDING};

#define ___DEF_SUB_FLO(id,hi,lo) \
___LOCAL ___WORD id[]={___MAKE_HD((___FLONUM_SIZE<<___LWS),___sFLONUM,___PERM) \
___F64VEC1(hi,lo)___F64VEC0

#define ___DEF_SUB_CPX(id,real,imag) \
___LOCAL ___WORD id[]={___MAKE_HD((___CPXNUM_SIZE<<___LWS),___sCPXNUM,___PERM),real,imag ___PADDING};

#define ___DEF_SUB_S8VEC(id,n) \
___LOCAL ___WORD id[]={___MAKE_HD(n,___sS8VECTOR,___PERM)

#define ___DEF_SUB_U8VEC(id,n) \
___LOCAL ___WORD id[]={___MAKE_HD(n,___sU8VECTOR,___PERM)

#define ___DEF_SUB_S16VEC(id,n) \
___LOCAL ___WORD id[]={___MAKE_HD((n<<1),___sS16VECTOR,___PERM)

#define ___DEF_SUB_U16VEC(id,n) \
___LOCAL ___WORD id[]={___MAKE_HD((n<<1),___sU16VECTOR,___PERM)

#define ___DEF_SUB_S32VEC(id,n) \
___LOCAL ___WORD id[]={___MAKE_HD((n<<2),___sS32VECTOR,___PERM)

#define ___DEF_SUB_U32VEC(id,n) \
___LOCAL ___WORD id[]={___MAKE_HD((n<<2),___sU32VECTOR,___PERM)

#define ___DEF_SUB_S64VEC(id,n) \
___LOCAL ___WORD id[]={___MAKE_HD((n<<3),___sS64VECTOR,___PERM)

#define ___DEF_SUB_U64VEC(id,n) \
___LOCAL ___WORD id[]={___MAKE_HD((n<<3),___sU64VECTOR,___PERM)

#define ___DEF_SUB_F32VEC(id,n) \
___LOCAL ___WORD id[]={___MAKE_HD((n<<2),___sF32VECTOR,___PERM)

#define ___DEF_SUB_F64VEC(id,n) \
___LOCAL ___WORD id[]={___MAKE_HD((n<<3),___sF64VECTOR,___PERM)

#if ___WS == 4

#ifdef ___CPU_BIGEND

#define ___P4U8(a,b,c,d)((((___WORD)___CAST_U8(a)<<8)+(___WORD)___CAST_U8(b)<<8)+(___WORD)___CAST_U8(c)<<8)+(___WORD)___CAST_U8(d)

#define ___P2U16(a,b)((___WORD)___CAST_U16(a)<<16)+(___WORD)___CAST_U16(b)

#define ___P1U32(a)a

#define ___P1U64(hi,lo)hi,lo

#else

#define ___P4U8(d,c,b,a)((((___WORD)___CAST_U8(a)<<8)+(___WORD)___CAST_U8(b)<<8)+(___WORD)___CAST_U8(c)<<8)+(___WORD)___CAST_U8(d)

#define ___P2U16(b,a)((___WORD)___CAST_U16(a)<<16)+(___WORD)___CAST_U16(b)

#define ___P1U32(a)a

#define ___P1U64(hi,lo)lo,hi

#endif

#define ___S8VEC0 ___PADDING};
#define ___S8VEC1(a),___P4U8(a,0,0,0) ___PADDING};
#define ___S8VEC2(a,b),___P4U8(a,b,0,0) ___PADDING};
#define ___S8VEC3(a,b,c),___P4U8(a,b,c,0) ___PADDING};
#define ___S8VEC4(a,b,c,d),___P4U8(a,b,c,d) ___PADDING};
#define ___S8VEC5(a,b,c,d,e),___P4U8(a,b,c,d),___P4U8(e,0,0,0) ___PADDING};
#define ___S8VEC6(a,b,c,d,e,f),___P4U8(a,b,c,d),___P4U8(e,f,0,0) ___PADDING};
#define ___S8VEC7(a,b,c,d,e,f,g),___P4U8(a,b,c,d),___P4U8(e,f,g,0) ___PADDING};
#define ___S8VEC8(a,b,c,d,e,f,g,h),___P4U8(a,b,c,d),___P4U8(e,f,g,h)

#define ___U8VEC0 ___PADDING};
#define ___U8VEC1(a),___P4U8(a,0,0,0) ___PADDING};
#define ___U8VEC2(a,b),___P4U8(a,b,0,0) ___PADDING};
#define ___U8VEC3(a,b,c),___P4U8(a,b,c,0) ___PADDING};
#define ___U8VEC4(a,b,c,d),___P4U8(a,b,c,d) ___PADDING};
#define ___U8VEC5(a,b,c,d,e),___P4U8(a,b,c,d),___P4U8(e,0,0,0) ___PADDING};
#define ___U8VEC6(a,b,c,d,e,f),___P4U8(a,b,c,d),___P4U8(e,f,0,0) ___PADDING};
#define ___U8VEC7(a,b,c,d,e,f,g),___P4U8(a,b,c,d),___P4U8(e,f,g,0) ___PADDING};
#define ___U8VEC8(a,b,c,d,e,f,g,h),___P4U8(a,b,c,d),___P4U8(e,f,g,h)

#define ___S16VEC0 ___PADDING};
#define ___S16VEC1(a),___P2U16(a,0) ___PADDING};
#define ___S16VEC2(a,b),___P2U16(a,b) ___PADDING};
#define ___S16VEC3(a,b,c),___P2U16(a,b),___P2U16(c,0) ___PADDING};
#define ___S16VEC4(a,b,c,d),___P2U16(a,b),___P2U16(c,d)

#define ___U16VEC0 ___PADDING};
#define ___U16VEC1(a),___P2U16(a,0) ___PADDING};
#define ___U16VEC2(a,b),___P2U16(a,b) ___PADDING};
#define ___U16VEC3(a,b,c),___P2U16(a,b),___P2U16(c,0) ___PADDING};
#define ___U16VEC4(a,b,c,d),___P2U16(a,b),___P2U16(c,d)

#define ___S32VEC0 ___PADDING};
#define ___S32VEC1(a),___P1U32(a) ___PADDING};
#define ___S32VEC2(a,b),___P1U32(a),___P1U32(b)

#define ___U32VEC0 ___PADDING};
#define ___U32VEC1(a),___P1U32(a) ___PADDING};
#define ___U32VEC2(a,b),___P1U32(a),___P1U32(b)

#else

#ifdef ___CPU_BIGEND

#define ___P8U8(a,b,c,d,e,f,g,h)((((((((((((((___WORD)___CAST_U8(a)<<8)+(___WORD)___CAST_U8(b))<<8)+(___WORD)___CAST_U8(c))<<8)+(___WORD)___CAST_U8(d))<<8)+(___WORD)___CAST_U8(e))<<8)+(___WORD)___CAST_U8(f))<<8)+(___WORD)___CAST_U8(g))<<8)+(___WORD)___CAST_U8(h)

#define ___P4U16(a,b,c,d)((((((___WORD)___CAST_U16(a)<<16)+(___WORD)___CAST_U16(b))<<16)+(___WORD)___CAST_U16(c))<<16)+(___WORD)___CAST_U16(d)

#define ___P2U32(a,b)((___WORD)___CAST_U32(a)<<32)+(___WORD)___CAST_U32(b)

#else

#define ___P8U8(h,g,f,e,d,c,b,a)((((((((((((((___WORD)___CAST_U8(a)<<8)+(___WORD)___CAST_U8(b))<<8)+(___WORD)___CAST_U8(c))<<8)+(___WORD)___CAST_U8(d))<<8)+(___WORD)___CAST_U8(e))<<8)+(___WORD)___CAST_U8(f))<<8)+(___WORD)___CAST_U8(g))<<8)+(___WORD)___CAST_U8(h)

#define ___P4U16(d,c,b,a)((((((___WORD)___CAST_U16(a)<<16)+(___WORD)___CAST_U16(b))<<16)+(___WORD)___CAST_U16(c))<<16)+(___WORD)___CAST_U16(d)

#define ___P2U32(b,a)((___WORD)___CAST_U32(a)<<32)+(___WORD)___CAST_U32(b)

#endif

#define ___P1U64(hi,lo)((___WORD)(hi)<<32)+___CAST_U32(lo)

#define ___S8VEC0 ___PADDING};
#define ___S8VEC1(a),___P8U8(a,0,0,0,0,0,0,0) ___PADDING};
#define ___S8VEC2(a,b),___P8U8(a,b,0,0,0,0,0,0) ___PADDING};
#define ___S8VEC3(a,b,c),___P8U8(a,b,c,0,0,0,0,0) ___PADDING};
#define ___S8VEC4(a,b,c,d),___P8U8(a,b,c,d,0,0,0,0) ___PADDING};
#define ___S8VEC5(a,b,c,d,e),___P8U8(a,b,c,d,e,0,0,0) ___PADDING};
#define ___S8VEC6(a,b,c,d,e,f),___P8U8(a,b,c,d,e,f,0,0) ___PADDING};
#define ___S8VEC7(a,b,c,d,e,f,g),___P8U8(a,b,c,d,e,f,g,0) ___PADDING};
#define ___S8VEC8(a,b,c,d,e,f,g,h),___P8U8(a,b,c,d,e,f,g,h)

#define ___U8VEC0 ___PADDING};
#define ___U8VEC1(a),___P8U8(a,0,0,0,0,0,0,0) ___PADDING};
#define ___U8VEC2(a,b),___P8U8(a,b,0,0,0,0,0,0) ___PADDING};
#define ___U8VEC3(a,b,c),___P8U8(a,b,c,0,0,0,0,0) ___PADDING};
#define ___U8VEC4(a,b,c,d),___P8U8(a,b,c,d,0,0,0,0) ___PADDING};
#define ___U8VEC5(a,b,c,d,e),___P8U8(a,b,c,d,e,0,0,0) ___PADDING};
#define ___U8VEC6(a,b,c,d,e,f),___P8U8(a,b,c,d,e,f,0,0) ___PADDING};
#define ___U8VEC7(a,b,c,d,e,f,g),___P8U8(a,b,c,d,e,f,g,0) ___PADDING};
#define ___U8VEC8(a,b,c,d,e,f,g,h),___P8U8(a,b,c,d,e,f,g,h)

#define ___S16VEC0 ___PADDING};
#define ___S16VEC1(a),___P4U16(a,0,0,0) ___PADDING};
#define ___S16VEC2(a,b),___P4U16(a,b,0,0) ___PADDING};
#define ___S16VEC3(a,b,c),___P4U16(a,b,c,0) ___PADDING};
#define ___S16VEC4(a,b,c,d),___P4U16(a,b,c,d)

#define ___U16VEC0 ___PADDING};
#define ___U16VEC1(a),___P4U16(a,0,0,0) ___PADDING};
#define ___U16VEC2(a,b),___P4U16(a,b,0,0) ___PADDING};
#define ___U16VEC3(a,b,c),___P4U16(a,b,c,0) ___PADDING};
#define ___U16VEC4(a,b,c,d),___P4U16(a,b,c,d)

#define ___S32VEC0 ___PADDING};
#define ___S32VEC1(a),___P2U32(a,0) ___PADDING};
#define ___S32VEC2(a,b),___P2U32(a,b)

#define ___U32VEC0 ___PADDING};
#define ___U32VEC1(a),___P2U32(a,0) ___PADDING};
#define ___U32VEC2(a,b),___P2U32(a,b)

#endif

#define ___S64VEC0 ___PADDING_64};
#define ___S64VEC1(hi,lo),___P1U64(hi,lo)

#define ___U64VEC0 ___PADDING_64};
#define ___U64VEC1(hi,lo),___P1U64(hi,lo)

#define ___STR0 ___CS_SELECT( \
___U8VEC0, \
___U16VEC0, \
___U32VEC0)

#define ___STR1(a)___CS_SELECT( \
___U8VEC1(a), \
___U16VEC1(a), \
___U32VEC1(a))

#define ___STR2(a,b)___CS_SELECT( \
___U8VEC2(a,b), \
___U16VEC2(a,b), \
___U32VEC2(a,b)___U32VEC0)

#define ___STR3(a,b,c)___CS_SELECT( \
___U8VEC3(a,b,c), \
___U16VEC3(a,b,c), \
___U32VEC2(a,b)___U32VEC1(c))

#define ___STR4(a,b,c,d)___CS_SELECT( \
___U8VEC4(a,b,c,d), \
___U16VEC4(a,b,c,d)___U16VEC0, \
___U32VEC2(a,b)___U32VEC2(c,d)___U32VEC0)

#define ___STR5(a,b,c,d,e)___CS_SELECT( \
___U8VEC5(a,b,c,d,e), \
___U16VEC4(a,b,c,d)___U16VEC1(e), \
___U32VEC2(a,b)___U32VEC2(c,d)___U32VEC1(e))

#define ___STR6(a,b,c,d,e,f)___CS_SELECT( \
___U8VEC6(a,b,c,d,e,f), \
___U16VEC4(a,b,c,d)___U16VEC2(e,f), \
___U32VEC2(a,b)___U32VEC2(c,d)___U32VEC2(e,f)___U32VEC0)

#define ___STR7(a,b,c,d,e,f,g)___CS_SELECT( \
___U8VEC7(a,b,c,d,e,f,g), \
___U16VEC4(a,b,c,d)___U16VEC3(e,f,g), \
___U32VEC2(a,b)___U32VEC2(c,d)___U32VEC2(e,f)___U32VEC1(g))

#define ___STR8(a,b,c,d,e,f,g,h)___CS_SELECT( \
___U8VEC8(a,b,c,d,e,f,g,h), \
___U16VEC4(a,b,c,d)___U16VEC4(e,f,g,h), \
___U32VEC2(a,b)___U32VEC2(c,d)___U32VEC2(e,f)___U32VEC2(g,h))

#if ___BIG_ABASE_WIDTH == 32

#define ___DEF_SUB_BIG(id,n) \
___LOCAL ___WORD id[]={___MAKE_HD((n<<2),___sBIGNUM,___PERM)

#define ___BIG0 ___U32VEC0
#define ___BIG1(a)___U32VEC1(a)
#define ___BIG2(a,b)___U32VEC2(a,b)

#else

#define ___DEF_SUB_BIG(id,n) \
___LOCAL ___WORD id[]={___MAKE_HD((((n+1)>>1)<<3),___sBIGNUM,___PERM)

#define ___BIG0 ___U64VEC0
#define ___BIG1(a)___BIG2(a,(___WORD)((a) & -0x80000000)?-1:0)___U64VEC0
#define ___BIG2(a,b)___U64VEC1(b,a)

#endif

#define ___DEF_SUB_BIGFIX(id,n)___DEF_SUB_BIG(id,n)

#define ___BIGFIX0 ___BIG0
#define ___BIGFIX1(a)___BIG1(a)
#define ___BIGFIX2(a,b)___BIG2(a,b)

#define ___F32VEC0 ___U32VEC0 
#define ___F32VEC1(a)___U32VEC1(a)
#define ___F32VEC2(a,b)___U32VEC2(a,b)

#define ___F64VEC0 ___U64VEC0
#ifdef ___CRAY_FP_FORMAT
#define ___F64VEC1(hi,lo)___U64VEC1( \
((hi&0x7ff00000==0)?hi: \
((hi&0x80000000)+((hi>>4)&0x7ff0000)+0x3c028000+((hi>>5)&0x7fff))), \
((hi&0x7ff00000==0)?0: \
(((hi<<27)&0xf8000000)|((lo>>5)&0x7ffffff))))
#else
#define ___F64VEC1(hi,lo)___U64VEC1(hi,lo)
#endif

#define ___VEC0 ___PADDING};
#define ___VEC1(a),a

#define ___CEILING_DIV(n,m)(((n)+(m)-1)/(m))
#define ___ROUND_TO_MULT(n,m)(___CEILING_DIV(n,m)*(m))
#define ___BREAK_FRAME_NEXT 1/********************should be 0*/
#define ___BREAK_FRAME_SPACE ___ROUND_TO_MULT(1,___FRAME_ALIGN)

#define ___FRAME_SPACE(n)___ROUND_TO_MULT((n)+___FRAME_RESERVE,___FRAME_ALIGN)
#define ___FRAME_RESERVE 1
#define ___FRAME_ALIGN 4

#define ___BEGIN_LBL \
___LOCAL ___label_struct ___lbl_tbl[]={
#define ___DEF_LBL_INTRO(id,info,cproc){___MAKE_HD((___INTRO_SIZE<<___LWS),___sVECTOR,___PERM),info,0,___CAST_HOST_TO_FAKEHOST((___host)cproc)}
#define ___DEF_LBL_PROC(id,nbp,nbc){___PRD(nbp,nbc),0,0,___CAST_HOST_TO_FAKEHOST(___SM(___MH_PROC,id))}
#define ___DEF_LBL_RET(id,frm){___MAKE_HD((3<<___LWS),___sRETURN,___PERM),frm,0,___CAST_HOST_TO_FAKEHOST(___SM(___MH_PROC,id))}

#define ___END_LBL ___PADDING_LBL};

#ifdef ___OFD_COUNT
#define ___OFD_TBL ___ofd_tbl,sizeof(___ofd_tbl)/sizeof(___WORD)
#else
#define ___OFD_TBL 0,0
#endif

#define ___BEGIN_OFD \
___LOCAL ___WORD ___ofd_tbl[]={
#define ___END_OFD ___PADDING};

#define ___DEF_OFD(kind,fs,link)(link<<16)+(fs<<2)+kind
#define ___OFD_KIND(x)((x)&((1<<2)-1))
#define ___OFD_FS(x)(((x)>>2)&((1<<14)-1))
#define ___OFD_LINK(x)(((x)>>16)&((1<<14)-1))

#define ___GCMAP0
#define ___GCMAP1(a),a
#if ___WS == 4
#define ___GCMAP2(a,b),a,b
#else
#define ___GCMAP2(a,b),((___WORD)b<<32)+___CAST_U32(a)
#endif

#define ___U64VEC0 ___PADDING_64};
#define ___U64VEC1(hi,lo),___P1U64(hi,lo)

#define ___PRD(nbparms,nbclosed) \
___MAKE_HD((nbclosed<<12)+nbparms,___sPROCEDURE,___PERM)
#define ___PRD_NBPARMS(x)(((x)>>(___SB+___HTB))&((1<<12)-1))
#define ___PRD_NBCLOSED(x)(((x)>>(___SB+___HTB+12))&((1<<12)-1))

#define ___IFD(kind,fs,link,gcmap)(gcmap<<12)+(link<<7)+(fs<<2)+kind
#define ___OFD(kind,fs,link,gcmap)___IFD(kind,0,0,0)
#define ___IS_IFD(x)((x)&3)

#define ___IFD_KIND(x)((x)&((1<<2)-1))
#define ___IFD_FS(x)(((x)>>2)&((1<<5)-1))
#define ___IFD_LINK(x)(((x)>>7)&((1<<5)-1))
#define ___IFD_GCMAP(x)(((x)>>12)&((1<<20)-1))


#define ___RETI_CFS_TO_FS(fs) \
___ROUND_TO_MULT(fs,___FRAME_ALIGN)+___FRAME_SPACE(___NB_GVM_REGS+1)-___FRAME_RESERVE

#define ___RETI_GET_CFS(ra,fs) \
{ ___WORD descr = ((___label_struct*)((ra)-___tSUBTYPED))->entry_or_descr; \
if (___IS_IFD(descr)) \
fs = ___IFD_FS(descr); \
else \
fs = ___OFD_FS(*(___WORD*)descr); }

#define ___RETI_GET_FS(ra,fs) \
{ ___WORD descr = ((___label_struct*)((ra)-___tSUBTYPED))->entry_or_descr; \
if (___IS_IFD(descr)) \
fs = ___RETI_CFS_TO_FS(___IFD_FS(descr)); \
else \
fs = ___RETI_CFS_TO_FS(___OFD_FS(*(___WORD*)descr)); }

#define ___RETN_GET_FS(ra,fs) \
{ ___WORD descr = ((___label_struct*)((ra)-___tSUBTYPED))->entry_or_descr; \
if (___IS_IFD(descr)) \
fs = ___IFD_FS(descr); \
else \
fs = ___OFD_FS(*(___WORD*)descr); }

#define ___RETI_GET_FS_LINK(ra,fs,link) \
{ ___WORD descr = ((___label_struct*)((ra)-___tSUBTYPED))->entry_or_descr; \
if (___IS_IFD(descr)) \
{ fs = ___RETI_CFS_TO_FS(___IFD_FS(descr)); link = ___IFD_LINK(descr); } \
else \
{ descr = *(___WORD*)descr; \
fs = ___RETI_CFS_TO_FS(___OFD_FS(descr)); link = ___OFD_LINK(descr); } }

#define ___RETN_GET_FS_LINK(ra,fs,link) \
{ ___WORD descr = ((___label_struct*)((ra)-___tSUBTYPED))->entry_or_descr; \
if (___IS_IFD(descr)) \
{ fs = ___IFD_FS(descr); link = ___IFD_LINK(descr); } \
else \
{ descr = *(___WORD*)descr; \
fs = ___OFD_FS(descr); link = ___OFD_LINK(descr); } }

#define ___RETI_GET_FS_LINK_GCMAP(ra,fs,link,firstgcmap,nextgcmap) \
{ ___WORD descr = ((___label_struct*)((ra)-___tSUBTYPED))->entry_or_descr; \
if (___IS_IFD(descr)) \
{ fs = ___RETI_CFS_TO_FS(___IFD_FS(descr)); link = ___IFD_LINK(descr); \
firstgcmap = ___IFD_GCMAP(descr); } \
else \
{ nextgcmap = (___WORD*)descr; descr = *nextgcmap++; firstgcmap = *nextgcmap++; \
fs = ___RETI_CFS_TO_FS(___OFD_FS(descr)); link = ___OFD_LINK(descr); } }

#define ___RETN_GET_FS_LINK_GCMAP(ra,fs,link,firstgcmap,nextgcmap) \
{ ___WORD descr = ((___label_struct*)((ra)-___tSUBTYPED))->entry_or_descr; \
if (___IS_IFD(descr)) \
{ firstgcmap = ___IFD_GCMAP(descr); \
fs = ___IFD_FS(descr); link = ___IFD_LINK(descr); } \
else \
{ nextgcmap = (___WORD*)descr; descr = *nextgcmap++; firstgcmap = *nextgcmap++; \
fs = ___OFD_FS(descr); link = ___OFD_LINK(descr); } }




#define ___RETN 1
#define ___RETI 2
#define ___RETT 3

#ifdef ___GLO_COUNT
#ifdef ___BIND_LATE
#define ___GLO_DECL ___LOCAL ___FAKEWORD ___glo_tbl[___GLO_COUNT];
#define ___GLO_TBL ___glo_tbl,___GLO_COUNT,___SUP_COUNT,___glo_names
#else
#define ___GLO_DECL
#define ___GLO_TBL 0,0,0,0
#endif
#else
#define ___GLO_DECL
#define ___GLO_TBL 0,0,0,0
#endif

#ifdef ___SYM_COUNT
#ifdef ___BIND_LATE
#define ___SYM_DECL ___LOCAL ___FAKEWORD ___sym_tbl[___SYM_COUNT];
#define ___SYM_TBL ___sym_tbl,___SYM_COUNT,___sym_names
#else
#define ___SYM_DECL
#define ___SYM_TBL ___sym_tbl,___SYM_COUNT,0
#endif
#else
#define ___SYM_DECL
#define ___SYM_TBL 0,0,0
#endif

#ifdef ___KEY_COUNT
#ifdef ___BIND_LATE
#define ___KEY_DECL ___LOCAL ___FAKEWORD ___key_tbl[___KEY_COUNT];
#define ___KEY_TBL ___key_tbl,___KEY_COUNT,___key_names
#else
#define ___KEY_DECL
#define ___KEY_TBL ___key_tbl,___KEY_COUNT,0
#endif
#else
#define ___KEY_DECL
#define ___KEY_TBL 0,0,0
#endif

#ifdef ___LBL_COUNT
#define ___LBL_TBL &___lp,___lbl_tbl,___LBL_COUNT
#else
#define ___LBL_TBL 0,0,0
#endif

#ifdef ___CNS_COUNT
#define ___CNS_TBL ___cns_tbl,___CNS_COUNT
#else
#define ___CNS_TBL 0,0
#endif

#ifdef ___SUB_COUNT
#define ___SUB_TBL ___sub_tbl,___SUB_COUNT
#else
#define ___SUB_TBL 0,0
#endif

#define ___BEGIN_MOD1 ___LOCAL void ___init_proc ___PVOID{
#define ___DEF_PRM(i,glo,n)___PRM(i,glo)=___GLO(i,glo)=___MLBL(n);
#define ___END_MOD1 }

#define ___BEGIN_MOD2 ___LOCAL ___module_struct ___module_descr = { \
  ___VERSION_MAJOR, \
  ___VERSION_MINOR, \
  ___MODULE_KIND, \
  ___MODULE_NAME, \
  ___SETUP_HOST_LABEL, \
  ___GLO_TBL, \
  ___SYM_TBL, \
  ___KEY_TBL, \
  ___LBL_TBL, \
  ___OFD_TBL, \
  ___CNS_TBL, \
  ___SUB_TBL, \
  ___init_proc, \
  0,\
  0,0,0,0,0,0,0,0 \
}; \
___BEGIN_C_LINKAGE \
___EXP_FUNC(___mod_or_lnk,___LINKER_ID) \
   ___P((___global_state_struct *___gs),(___gs) \
___global_state_struct *___gs;) \
{ \
  ___SET_LOCAL_GSTATE(___gs)
#define ___END_MOD2 \
  return (___mod_or_lnk)&___module_descr; \
} \
___END_C_LINKAGE

#ifndef ___LIBRARY
#define ___MAIN_PROC \
___IMP_FUNC(int,___main) \
  ___P((int argc, \
        char **argv, \
        ___mod_or_lnk (*linker)(___global_state_struct*)), \
       ()); \
int main ___P((int argc, char **argv),(argc, argv) \
int argc; \
char **argv;) \
{ \
  return ___main (argc, argv, ___LINKER_ID); \
}
#else
#define ___MAIN_PROC
#endif

#define ___LINKFILE_DESCR(sym_list,key_list) \
___LOCAL ___linkfile_struct ___linkfile_descr = { \
  ___VERSION_MAJOR, \
  ___VERSION_MINOR, \
  ___LINKFILE_KIND, \
  ___LINKFILE_NAME, \
  sym_list, \
  key_list, \
  ___linker_tbl, \
  0,0,0,0,0,0,0,0 \
}; \
___BEGIN_C_LINKAGE \
___EXP_FUNC(___mod_or_lnk,___LINKER_ID) \
   ___P((___global_state_struct *___gs),(___gs) \
___global_state_struct *___gs;) \
{ \
  ___SET_LOCAL_GSTATE(___gs) \
  return (___mod_or_lnk)&___linkfile_descr; \
} \
___END_C_LINKAGE \
___MAIN_PROC

#define ___BEGIN_OLD_LNK
#define ___DEF_OLD_LNK(id)___BEGIN_C_LINKAGE ___IMP_FUNC(___mod_or_lnk,id) \
___P((___global_state_struct *___gs),()); ___END_C_LINKAGE
#define ___END_OLD_LNK

#define ___BEGIN_NEW_LNK
#define ___DEF_NEW_LNK(id)___BEGIN_C_LINKAGE ___mod_or_lnk id \
___P((___global_state_struct *___gs),()); ___END_C_LINKAGE
#define ___END_NEW_LNK

#define ___BEGIN_LNK ___LOCAL void *___linker_tbl[]={
#define ___DEF_LNK(id)(void*)id
#define ___END_LNK ,0};

/*---------------------------------------------------------------------------*/

/* 
 * ___ERR_CODE is a type which represents the union of error codes
 * from various subsystems.  Currently 4 types of error codes are
 * supported: Windows HRESULT, MACOS errors, ANSI-C errno, and system
 * specific error codes.  A Windows HRESULT error code is a 32 bit
 * signed integer value composed of (starting with the low bits) a 16
 * bit status code, a 11 bit facility code, a 4 bit reserved facility
 * code and a 1 bit severity code (negative means error).  It seems
 * that only the 4 lower bits of the facility code are really used,
 * the remaining 7 bits are zero.  This allows us to compact this 32
 * bit integer value to a 30 bit integer value (which fits in a
 * fixnum) by dropping the top 2 bits of the facility code.  Of the
 * 512 possible facility codes, the codes 448 to 511 are assigned to
 * Gambit specific error codes, the codes 384 to 447 are assigned to
 * ANSI-C errno error codes and the codes 320 to 383 are assigned to
 * MACOS errors (so 2^22 = 4194304 different error codes are available
 * in each case).  The ``no error'' error code of each subsystem is
 * always mapped to ___NO_ERR which is equal to 0, all other error
 * codes are thus negative values.
 */

#define ___ERR_CODE int

#define ___ERR_CODE_FACILITY_SYSTEM 448
#define ___ERR_CODE_FACILITY_ERRNO  384
#define ___ERR_CODE_FACILITY_MACOS  320

#define ___ERR_CODE_BUILD(facility,code) \
((___CAST(___ERR_CODE,-1)<<29)+(___CAST(___ERR_CODE,facility)<<16)+(code))

#define ___ERR_CODE_FACILITY(err)(((err)>>16)&~(~___CAST(___ERR_CODE,0)<<9))

#define ___ERR_CODE_FROM_HRESULT(hresult) \
((hresult)&~(~___CAST(___ERR_CODE,0)<<25) | \
(((hresult)&~___CAST(___ERR_CODE,0)<<27)>>2))

#define ___ERR_CODE_TO_HRESULT(err) \
((err)&~(~___CAST(___ERR_CODE,0)<<25) | \
(((err)&~___CAST(___ERR_CODE,0)<<25)<<2))

#define ___ERR_CODE_FROM_ERRNO(errno) \
((errno)==0?___NO_ERR:___ERR_CODE_BUILD(___ERR_CODE_FACILITY_ERRNO,errno))

#define ___ERR_CODE_TO_ERRNO(err) \
((err)==___NO_ERR?0:((err)-___ERR_CODE_BUILD(___ERR_CODE_FACILITY_ERRNO,0)))

#define ___ERR_CODE_FROM_MACOS(oserr) \
((oserr)==0?___NO_ERR:___ERR_CODE_BUILD(___ERR_CODE_FACILITY_MACOS,oserr))

#define ___ERR_CODE_TO_MACOS(err) \
((err)==___NO_ERR?0:((err)-___ERR_CODE_BUILD(___ERR_CODE_FACILITY_MACOS,0)))

#define ___ERR_BASE ___ERR_CODE_BUILD(___ERR_CODE_FACILITY_SYSTEM,0)

#define ___NO_ERR                  0
#define ___UNWIND_C_STACK          (___ERR_BASE+0)
#define ___CDEF_HEAP_OVERFLOW_ERR  (___ERR_BASE+1)
#define ___IMPL_LIMIT_ERR          (___ERR_BASE+2)
#define ___UNKNOWN_ERR             (___ERR_BASE+3)
#define ___HEAP_OVERFLOW_ERR       (___ERR_BASE+4)
#define ___CLOSED_DEVICE_ERR       (___ERR_BASE+5)
#define ___INVALID_OP_ERR          (___ERR_BASE+6)
#define ___KILL_PUMP               (___ERR_BASE+7)
#define ___SELECT_SETUP_DONE       (___ERR_BASE+7)
#define ___STOC_BASE ___STOC_S8_ERR
#define ___STOC_S8_ERR             (___ERR_BASE+(1<<7))
#define ___STOC_U8_ERR             (___ERR_BASE+(2<<7))
#define ___STOC_S16_ERR            (___ERR_BASE+(3<<7))
#define ___STOC_U16_ERR            (___ERR_BASE+(4<<7))
#define ___STOC_S32_ERR            (___ERR_BASE+(5<<7))
#define ___STOC_U32_ERR            (___ERR_BASE+(6<<7))
#define ___STOC_S64_ERR            (___ERR_BASE+(7<<7))
#define ___STOC_U64_ERR            (___ERR_BASE+(8<<7))
#define ___STOC_F32_ERR            (___ERR_BASE+(9<<7))
#define ___STOC_F64_ERR            (___ERR_BASE+(10<<7))
#define ___STOC_CHAR_ERR           (___ERR_BASE+(11<<7))
#define ___STOC_SCHAR_ERR          (___ERR_BASE+(12<<7))
#define ___STOC_UCHAR_ERR          (___ERR_BASE+(13<<7))
#define ___STOC_LATIN1_ERR         (___ERR_BASE+(14<<7))
#define ___STOC_UCS2_ERR           (___ERR_BASE+(15<<7))
#define ___STOC_UCS4_ERR           (___ERR_BASE+(16<<7))
#define ___STOC_SHORT_ERR          (___ERR_BASE+(17<<7))
#define ___STOC_USHORT_ERR         (___ERR_BASE+(18<<7))
#define ___STOC_INT_ERR            (___ERR_BASE+(19<<7))
#define ___STOC_UINT_ERR           (___ERR_BASE+(20<<7))
#define ___STOC_LONG_ERR           (___ERR_BASE+(21<<7))
#define ___STOC_ULONG_ERR          (___ERR_BASE+(22<<7))
#define ___STOC_FLOAT_ERR          (___ERR_BASE+(23<<7))
#define ___STOC_DOUBLE_ERR         (___ERR_BASE+(24<<7))
#define ___STOC_POINTER_ERR        (___ERR_BASE+(25<<7))
#define ___STOC_NONNULLPOINTER_ERR (___ERR_BASE+(26<<7))
#define ___STOC_FUNCTION_ERR       (___ERR_BASE+(27<<7))
#define ___STOC_BOOL_ERR           (___ERR_BASE+(28<<7))
#define ___STOC_CHARSTRING_ERR     (___ERR_BASE+(29<<7))
#define ___STOC_LATIN1STRING_ERR   (___ERR_BASE+(30<<7))
#define ___STOC_UTF8STRING_ERR     (___ERR_BASE+(31<<7))
#define ___STOC_UCS2STRING_ERR     (___ERR_BASE+(32<<7))
#define ___STOC_UCS4STRING_ERR     (___ERR_BASE+(33<<7))
#define ___STOC_VARIANT_ERR        (___ERR_BASE+(34<<7))
#define ___STOC_HEAP_OVERFLOW_ERR  (___ERR_BASE+(35<<7))
#define ___STOC_MAX                (___ERR_BASE+(36<<7)-1)
#define ___CTOS_BASE ___CTOS_S8_ERR
#define ___CTOS_S8_ERR             (___ERR_BASE+(51<<7))
#define ___CTOS_U8_ERR             (___ERR_BASE+(52<<7))
#define ___CTOS_S16_ERR            (___ERR_BASE+(53<<7))
#define ___CTOS_U16_ERR            (___ERR_BASE+(54<<7))
#define ___CTOS_S32_ERR            (___ERR_BASE+(55<<7))
#define ___CTOS_U32_ERR            (___ERR_BASE+(56<<7))
#define ___CTOS_S64_ERR            (___ERR_BASE+(57<<7))
#define ___CTOS_U64_ERR            (___ERR_BASE+(58<<7))
#define ___CTOS_F32_ERR            (___ERR_BASE+(59<<7))
#define ___CTOS_F64_ERR            (___ERR_BASE+(60<<7))
#define ___CTOS_CHAR_ERR           (___ERR_BASE+(61<<7))
#define ___CTOS_SCHAR_ERR          (___ERR_BASE+(62<<7))
#define ___CTOS_UCHAR_ERR          (___ERR_BASE+(63<<7))
#define ___CTOS_LATIN1_ERR         (___ERR_BASE+(64<<7))
#define ___CTOS_UCS2_ERR           (___ERR_BASE+(65<<7))
#define ___CTOS_UCS4_ERR           (___ERR_BASE+(66<<7))
#define ___CTOS_SHORT_ERR          (___ERR_BASE+(67<<7))
#define ___CTOS_USHORT_ERR         (___ERR_BASE+(68<<7))
#define ___CTOS_INT_ERR            (___ERR_BASE+(69<<7))
#define ___CTOS_UINT_ERR           (___ERR_BASE+(70<<7))
#define ___CTOS_LONG_ERR           (___ERR_BASE+(71<<7))
#define ___CTOS_ULONG_ERR          (___ERR_BASE+(72<<7))
#define ___CTOS_FLOAT_ERR          (___ERR_BASE+(73<<7))
#define ___CTOS_DOUBLE_ERR         (___ERR_BASE+(74<<7))
#define ___CTOS_POINTER_ERR        (___ERR_BASE+(75<<7))
#define ___CTOS_NONNULLPOINTER_ERR (___ERR_BASE+(76<<7))
#define ___CTOS_FUNCTION_ERR       (___ERR_BASE+(77<<7))
#define ___CTOS_BOOL_ERR           (___ERR_BASE+(78<<7))
#define ___CTOS_CHARSTRING_ERR     (___ERR_BASE+(79<<7))
#define ___CTOS_LATIN1STRING_ERR   (___ERR_BASE+(80<<7))
#define ___CTOS_UTF8STRING_ERR     (___ERR_BASE+(81<<7))
#define ___CTOS_UCS2STRING_ERR     (___ERR_BASE+(82<<7))
#define ___CTOS_UCS4STRING_ERR     (___ERR_BASE+(83<<7))
#define ___CTOS_VARIANT_ERR        (___ERR_BASE+(84<<7))
#define ___CTOS_HEAP_OVERFLOW_ERR  (___ERR_BASE+(85<<7))
#define ___CTOS_MAX                (___ERR_BASE+(86<<7)-1)

#define ___RETURN_POS 127

/*---------------------------------------------------------------------------*/

#ifdef ___USE_SETJMP
#define ___BEGIN_CATCH \
___jmpbuf_struct ___jbuf, *___old_catcher = ___ps->catcher; \
___ps->catcher = &___jbuf; ___err = ___SETJMP(___jbuf.buf); \
if (___err==___NO_ERR) {
#define ___END_CATCH } ___ps->catcher = ___old_catcher;
#define ___THROW(err)longjmp (___ps->catcher->buf, err)
#else
#define ___BEGIN_CATCH ___err = ___NO_ERR; try {
#define ___END_CATCH } catch (___ERR_CODE err) { ___err = err; }
#define ___THROW(err)throw (err)
#endif

#define ___CLAM_CONV_ERROR if ((___err!=___NO_ERR)&&(___err!=___UNWIND_C_STACK)) {___ps->temp1=___err;___ps->temp2=___errmsg;___ps->temp3=___LBL(0);___JUMPEXTPRM(___SET_NARGS(___NARGS),___ps->handler_clam_conv_error)}

#define ___STOC3(f,s,c,i)___err=___EXT(f)(s,c,i)
#define ___STOC4(f,s,c,t,i)___err=___EXT(f)(s,c,t,i)
#define ___CTOS3(f,c,s,i)___err=___EXT(f)(c,s,i)
#define ___CTOS5(f,c,t,r,s,i)___err=___EXT(f)(c,t,r,s,i)
#define ___IF_STOC3(f,s,c,i)if ((___STOC3(f,s,c,i))==___NO_ERR)
#define ___IF_STOC4(f,s,c,t,i)if ((___STOC4(f,s,c,t,i))==___NO_ERR)
#define ___IF_CTOS3(f,c,s,i)if ((___CTOS3(f,c,s,i))==___NO_ERR)
#define ___IF_CTOS5(f,c,t,r,s,i)if ((___CTOS5(f,c,t,r,s,i))==___NO_ERR)

#define ___BEGIN_CDEF_BODY
#define ___END_CDEF_BODY

#define ___BEGIN_CDEF_VOID ___processor_state ___ps=___PSTATE;___WORD ___marker;___ERR_CODE ___err;___WORD ___errmsg=___FAL;if ((___err=___EXT(___make_cdef_stack_marker) (&___marker))==___NO_ERR) {
#define ___CDEF_CALL_VOID(nargs,lbl)___err=___EXT(___call) (nargs,lbl,___marker);
#define ___CDEF_SET_RESULT_VOID /* no result to set */
#define ___CDEF_ERROR_VOID(lbl) while ((___err!=___NO_ERR)&&(___err!=___UNWIND_C_STACK)) {___ps->temp1=___err;___ps->temp2=___errmsg;___ps->temp3=lbl;___CDEF_CALL_VOID(0,___ps->handler_cdef_conv_error)
#define ___END_CDEF_VOID } ___EXT(___kill_cdef_stack_marker) (___marker); } ___EXT(___propagate_error) (___err);

#define ___BEGIN_CDEF_SCMOBJ ___processor_state ___ps=___PSTATE;___WORD ___marker;___WORD ___result;___ERR_CODE ___err;___WORD ___errmsg=___FAL;if ((___err=___EXT(___make_cdef_stack_marker) (&___marker))==___NO_ERR) {
#define ___CDEF_CALL_SCMOBJ(nargs,lbl)___err=___EXT(___call) (nargs,lbl,___marker);
#define ___CDEF_SET_RESULT_SCMOBJ ___result = ___CDEF_RESULT;
#define ___CDEF_ERROR_SCMOBJ(lbl) while ((___err!=___NO_ERR)&&(___err!=___UNWIND_C_STACK)) {___ps->temp1=___err;___ps->temp2=___errmsg;___ps->temp3=lbl;___CDEF_CALL_SCMOBJ(0,___ps->handler_cdef_conv_error)
#define ___END_CDEF_SCMOBJ } ___EXT(___kill_cdef_stack_marker) (___marker); } ___EXT(___propagate_error) (___err);

#define ___BEGIN_CDEF(decl)___processor_state ___ps=___PSTATE;___WORD ___marker;decl;___ERR_CODE ___err;___WORD ___errmsg=___FAL;if ((___err=___EXT(___make_cdef_stack_marker) (&___marker))==___NO_ERR) {
#define ___CDEF_CALL(nargs,lbl)if ((___err=___EXT(___call) (nargs,lbl,___marker))==___NO_ERR)
#define ___CDEF_SET_RESULT /* done in conversion function */
#define ___CDEF_ERROR(lbl) while ((___err!=___NO_ERR)&&(___err!=___UNWIND_C_STACK)) {___ps->temp1=___err;___ps->temp2=___errmsg;___ps->temp3=lbl;___CDEF_CALL(0,___ps->handler_cdef_conv_error)
#define ___END_CDEF } ___EXT(___kill_cdef_stack_marker) (___marker); } ___EXT(___propagate_error) (___err);

#define ___BEGIN_CDEF_ARG(i,arg)___WORD arg;
#define ___END_CDEF_ARG(i)
#define ___BEGIN_CDEF_ARG_SCMOBJ(i)
#define ___END_CDEF_ARG_SCMOBJ(i)
#define ___CDEF_ARG(i,arg)___PSSTK(___FRAME_SPACE(2)+i)=arg;
#define ___CDEF_RESULT ___PSR1

#define ___BEGIN_CDEF_SCMOBJ_TO_S8(s,c)___STOC3(___SCMOBJ_to_S8,s,&c,___RETURN_POS);
#define ___END_CDEF_SCMOBJ_TO_S8(s,c)
#define ___BEGIN_CDEF_SCMOBJ_TO_U8(s,c)___STOC3(___SCMOBJ_to_U8,s,&c,___RETURN_POS);
#define ___END_CDEF_SCMOBJ_TO_U8(s,c)
#define ___BEGIN_CDEF_SCMOBJ_TO_S16(s,c)___STOC3(___SCMOBJ_to_S16,s,&c,___RETURN_POS);
#define ___END_CDEF_SCMOBJ_TO_S16(s,c)
#define ___BEGIN_CDEF_SCMOBJ_TO_U16(s,c)___STOC3(___SCMOBJ_to_U16,s,&c,___RETURN_POS);
#define ___END_CDEF_SCMOBJ_TO_U16(s,c)
#define ___BEGIN_CDEF_SCMOBJ_TO_S32(s,c)___STOC3(___SCMOBJ_to_S32,s,&c,___RETURN_POS);
#define ___END_CDEF_SCMOBJ_TO_S32(s,c)
#define ___BEGIN_CDEF_SCMOBJ_TO_U32(s,c)___STOC3(___SCMOBJ_to_U32,s,&c,___RETURN_POS);
#define ___END_CDEF_SCMOBJ_TO_U32(s,c)
#define ___BEGIN_CDEF_SCMOBJ_TO_S64(s,c)___STOC3(___SCMOBJ_to_S64,s,&c,___RETURN_POS);
#define ___END_CDEF_SCMOBJ_TO_S64(s,c)
#define ___BEGIN_CDEF_SCMOBJ_TO_U64(s,c)___STOC3(___SCMOBJ_to_U64,s,&c,___RETURN_POS);
#define ___END_CDEF_SCMOBJ_TO_U64(s,c)
#define ___BEGIN_CDEF_SCMOBJ_TO_F32(s,c)___STOC3(___SCMOBJ_to_F32,s,&c,___RETURN_POS);
#define ___END_CDEF_SCMOBJ_TO_F32(s,c)
#define ___BEGIN_CDEF_SCMOBJ_TO_F64(s,c)___STOC3(___SCMOBJ_to_F64,s,&c,___RETURN_POS);
#define ___END_CDEF_SCMOBJ_TO_F64(s,c)
#define ___BEGIN_CDEF_SCMOBJ_TO_CHAR(s,c)___STOC3(___SCMOBJ_to_CHAR,s,&c,___RETURN_POS);
#define ___END_CDEF_SCMOBJ_TO_CHAR(s,c)
#define ___BEGIN_CDEF_SCMOBJ_TO_SCHAR(s,c)___STOC3(___SCMOBJ_to_SCHAR,s,&c,___RETURN_POS);
#define ___END_CDEF_SCMOBJ_TO_SCHAR(s,c)
#define ___BEGIN_CDEF_SCMOBJ_TO_UCHAR(s,c)___STOC3(___SCMOBJ_to_UCHAR,s,&c,___RETURN_POS);
#define ___END_CDEF_SCMOBJ_TO_UCHAR(s,c)
#define ___BEGIN_CDEF_SCMOBJ_TO_LATIN1(s,c)___STOC3(___SCMOBJ_to_LATIN1,s,&c,___RETURN_POS);
#define ___END_CDEF_SCMOBJ_TO_LATIN1(s,c)
#define ___BEGIN_CDEF_SCMOBJ_TO_UCS2(s,c)___STOC3(___SCMOBJ_to_UCS2,s,&c,___RETURN_POS);
#define ___END_CDEF_SCMOBJ_TO_UCS2(s,c)
#define ___BEGIN_CDEF_SCMOBJ_TO_UCS4(s,c)___STOC3(___SCMOBJ_to_UCS4,s,&c,___RETURN_POS);
#define ___END_CDEF_SCMOBJ_TO_UCS4(s,c)
#define ___BEGIN_CDEF_SCMOBJ_TO_SHORT(s,c)___STOC3(___SCMOBJ_to_SHORT,s,&c,___RETURN_POS);
#define ___END_CDEF_SCMOBJ_TO_SHORT(s,c)
#define ___BEGIN_CDEF_SCMOBJ_TO_USHORT(s,c)___STOC3(___SCMOBJ_to_USHORT,s,&c,___RETURN_POS);
#define ___END_CDEF_SCMOBJ_TO_USHORT(s,c)
#define ___BEGIN_CDEF_SCMOBJ_TO_INT(s,c)___STOC3(___SCMOBJ_to_INT,s,&c,___RETURN_POS);
#define ___END_CDEF_SCMOBJ_TO_INT(s,c)
#define ___BEGIN_CDEF_SCMOBJ_TO_UINT(s,c)___STOC3(___SCMOBJ_to_UINT,s,&c,___RETURN_POS);
#define ___END_CDEF_SCMOBJ_TO_UINT(s,c)
#define ___BEGIN_CDEF_SCMOBJ_TO_LONG(s,c)___STOC3(___SCMOBJ_to_LONG,s,&c,___RETURN_POS);
#define ___END_CDEF_SCMOBJ_TO_LONG(s,c)
#define ___BEGIN_CDEF_SCMOBJ_TO_ULONG(s,c)___STOC3(___SCMOBJ_to_ULONG,s,&c,___RETURN_POS);
#define ___END_CDEF_SCMOBJ_TO_ULONG(s,c)
#define ___BEGIN_CDEF_SCMOBJ_TO_FLOAT(s,c)___STOC3(___SCMOBJ_to_FLOAT,s,&c,___RETURN_POS);
#define ___END_CDEF_SCMOBJ_TO_FLOAT(s,c)
#define ___BEGIN_CDEF_SCMOBJ_TO_DOUBLE(s,c)___STOC3(___SCMOBJ_to_DOUBLE,s,&c,___RETURN_POS);
#define ___END_CDEF_SCMOBJ_TO_DOUBLE(s,c)
#define ___BEGIN_CDEF_SCMOBJ_TO_POINTER(s,c,t)___STOC4(___SCMOBJ_to_POINTER,s,(void**)&c,t,___RETURN_POS);
#define ___END_CDEF_SCMOBJ_TO_POINTER(s,c,t)
#define ___BEGIN_CDEF_SCMOBJ_TO_NONNULLPOINTER(s,c,t)___STOC4(___SCMOBJ_to_NONNULLPOINTER,s,(void**)&c,t,___RETURN_POS);
#define ___END_CDEF_SCMOBJ_TO_NONNULLPOINTER(s,c,t)
#define ___BEGIN_CDEF_SCMOBJ_TO_FUNCTION(s,c)___STOC3(___SCMOBJ_to_FUNCTION,s,(void**)&c,___RETURN_POS);
#define ___END_CDEF_SCMOBJ_TO_FUNCTION(s,c)
#define ___BEGIN_CDEF_SCMOBJ_TO_BOOL(s,c)___STOC3(___SCMOBJ_to_BOOL,s,&c,___RETURN_POS);
#define ___END_CDEF_SCMOBJ_TO_BOOL(s,c)
#define ___BEGIN_CDEF_SCMOBJ_TO_CHARSTRING(s,c)___STOC3(___SCMOBJ_to_CHARSTRING,s,&c,___RETURN_POS);
#define ___END_CDEF_SCMOBJ_TO_CHARSTRING(s,c)
#define ___BEGIN_CDEF_SCMOBJ_TO_LATIN1STRING(s,c)___STOC3(___SCMOBJ_to_LATIN1STRING,s,&c,___RETURN_POS);
#define ___END_CDEF_SCMOBJ_TO_LATIN1STRING(s,c)
#define ___BEGIN_CDEF_SCMOBJ_TO_UTF8STRING(s,c)___STOC3(___SCMOBJ_to_UTF8STRING,s,&c,___RETURN_POS);
#define ___END_CDEF_SCMOBJ_TO_UTF8STRING(s,c)
#define ___BEGIN_CDEF_SCMOBJ_TO_UCS2STRING(s,c)___STOC3(___SCMOBJ_to_UCS2STRING,s,&c,___RETURN_POS);
#define ___END_CDEF_SCMOBJ_TO_UCS2STRING(s,c)
#define ___BEGIN_CDEF_SCMOBJ_TO_UCS4STRING(s,c)___STOC3(___SCMOBJ_to_UCS4STRING,s,&c,___RETURN_POS);
#define ___END_CDEF_SCMOBJ_TO_UCS4STRING(s,c)
#define ___BEGIN_CDEF_SCMOBJ_TO_VARIANT(s,c)___STOC3(___SCMOBJ_to_VARIANT,s,&c,___RETURN_POS);
#define ___END_CDEF_SCMOBJ_TO_VARIANT(s,c)

#define ___BEGIN_CDEF_S8_TO_SCMOBJ(c,s,i)___IF_CTOS3(___S8_to_SCMOBJ,c,&s,i){
#define ___END_CDEF_S8_TO_SCMOBJ(c,s,i)}
#define ___BEGIN_CDEF_U8_TO_SCMOBJ(c,s,i)___IF_CTOS3(___U8_to_SCMOBJ,c,&s,i){
#define ___END_CDEF_U8_TO_SCMOBJ(c,s,i)}
#define ___BEGIN_CDEF_S16_TO_SCMOBJ(c,s,i)___IF_CTOS3(___S16_to_SCMOBJ,c,&s,i){
#define ___END_CDEF_S16_TO_SCMOBJ(c,s,i)}
#define ___BEGIN_CDEF_U16_TO_SCMOBJ(c,s,i)___IF_CTOS3(___U16_to_SCMOBJ,c,&s,i){
#define ___END_CDEF_U16_TO_SCMOBJ(c,s,i)}
#define ___BEGIN_CDEF_S32_TO_SCMOBJ(c,s,i)___IF_CTOS3(___S32_to_SCMOBJ,c,&s,i){
#define ___END_CDEF_S32_TO_SCMOBJ(c,s,i)___EXT(___release_scmobj)(s);}
#define ___BEGIN_CDEF_U32_TO_SCMOBJ(c,s,i)___IF_CTOS3(___U32_to_SCMOBJ,c,&s,i){
#define ___END_CDEF_U32_TO_SCMOBJ(c,s,i)___EXT(___release_scmobj)(s);}
#define ___BEGIN_CDEF_S64_TO_SCMOBJ(c,s,i)___IF_CTOS3(___S64_to_SCMOBJ,c,&s,i){
#define ___END_CDEF_S64_TO_SCMOBJ(c,s,i)___EXT(___release_scmobj)(s);}
#define ___BEGIN_CDEF_U64_TO_SCMOBJ(c,s,i)___IF_CTOS3(___U64_to_SCMOBJ,c,&s,i){
#define ___END_CDEF_U64_TO_SCMOBJ(c,s,i)___EXT(___release_scmobj)(s);}
#define ___BEGIN_CDEF_F32_TO_SCMOBJ(c,s,i)___IF_CTOS3(___F32_to_SCMOBJ,c,&s,i){
#define ___END_CDEF_F32_TO_SCMOBJ(c,s,i)___EXT(___release_scmobj)(s);}
#define ___BEGIN_CDEF_F64_TO_SCMOBJ(c,s,i)___IF_CTOS3(___F64_to_SCMOBJ,c,&s,i){
#define ___END_CDEF_F64_TO_SCMOBJ(c,s,i)___EXT(___release_scmobj)(s);}
#define ___BEGIN_CDEF_CHAR_TO_SCMOBJ(c,s,i)___IF_CTOS3(___CHAR_to_SCMOBJ,c,&s,i){
#define ___END_CDEF_CHAR_TO_SCMOBJ(c,s,i)}
#define ___BEGIN_CDEF_SCHAR_TO_SCMOBJ(c,s,i)___IF_CTOS3(___SCHAR_to_SCMOBJ,c,&s,i){
#define ___END_CDEF_SCHAR_TO_SCMOBJ(c,s,i)}
#define ___BEGIN_CDEF_UCHAR_TO_SCMOBJ(c,s,i)___IF_CTOS3(___UCHAR_to_SCMOBJ,c,&s,i){
#define ___END_CDEF_UCHAR_TO_SCMOBJ(c,s,i)}
#define ___BEGIN_CDEF_LATIN1_TO_SCMOBJ(c,s,i)___IF_CTOS3(___LATIN1_to_SCMOBJ,c,&s,i){
#define ___END_CDEF_LATIN1_TO_SCMOBJ(c,s,i)}
#define ___BEGIN_CDEF_UCS2_TO_SCMOBJ(c,s,i)___IF_CTOS3(___UCS2_to_SCMOBJ,c,&s,i){
#define ___END_CDEF_UCS2_TO_SCMOBJ(c,s,i)}
#define ___BEGIN_CDEF_UCS4_TO_SCMOBJ(c,s,i)___IF_CTOS3(___UCS4_to_SCMOBJ,c,&s,i){
#define ___END_CDEF_UCS4_TO_SCMOBJ(c,s,i)}
#define ___BEGIN_CDEF_SHORT_TO_SCMOBJ(c,s,i)___IF_CTOS3(___SHORT_to_SCMOBJ,c,&s,i){
#define ___END_CDEF_SHORT_TO_SCMOBJ(c,s,i)___EXT(___release_scmobj)(s);}
#define ___BEGIN_CDEF_USHORT_TO_SCMOBJ(c,s,i)___IF_CTOS3(___USHORT_to_SCMOBJ,c,&s,i){
#define ___END_CDEF_USHORT_TO_SCMOBJ(c,s,i)___EXT(___release_scmobj)(s);}
#define ___BEGIN_CDEF_INT_TO_SCMOBJ(c,s,i)___IF_CTOS3(___INT_to_SCMOBJ,c,&s,i){
#define ___END_CDEF_INT_TO_SCMOBJ(c,s,i)___EXT(___release_scmobj)(s);}
#define ___BEGIN_CDEF_UINT_TO_SCMOBJ(c,s,i)___IF_CTOS3(___UINT_to_SCMOBJ,c,&s,i){
#define ___END_CDEF_UINT_TO_SCMOBJ(c,s,i)___EXT(___release_scmobj)(s);}
#define ___BEGIN_CDEF_LONG_TO_SCMOBJ(c,s,i)___IF_CTOS3(___LONG_to_SCMOBJ,c,&s,i){
#define ___END_CDEF_LONG_TO_SCMOBJ(c,s,i)___EXT(___release_scmobj)(s);}
#define ___BEGIN_CDEF_ULONG_TO_SCMOBJ(c,s,i)___IF_CTOS3(___ULONG_to_SCMOBJ,c,&s,i){
#define ___END_CDEF_ULONG_TO_SCMOBJ(c,s,i)___EXT(___release_scmobj)(s);}
#define ___BEGIN_CDEF_FLOAT_TO_SCMOBJ(c,s,i)___IF_CTOS3(___FLOAT_to_SCMOBJ,c,&s,i){
#define ___END_CDEF_FLOAT_TO_SCMOBJ(c,s,i)___EXT(___release_scmobj)(s);}
#define ___BEGIN_CDEF_DOUBLE_TO_SCMOBJ(c,s,i)___IF_CTOS3(___DOUBLE_to_SCMOBJ,c,&s,i){
#define ___END_CDEF_DOUBLE_TO_SCMOBJ(c,s,i)___EXT(___release_scmobj)(s);}
#define ___BEGIN_CDEF_POINTER_TO_SCMOBJ(c,t,r,s,i)___IF_CTOS5(___POINTER_to_SCMOBJ,(void*)c,t,r,&s,i){
#define ___END_CDEF_POINTER_TO_SCMOBJ(c,t,r,s,i)___EXT(___release_scmobj)(s);}
#define ___BEGIN_CDEF_NONNULLPOINTER_TO_SCMOBJ(c,t,r,s,i)___IF_CTOS5(___NONNULLPOINTER_to_SCMOBJ,(void*)c,t,r,&s,i){
#define ___END_CDEF_NONNULLPOINTER_TO_SCMOBJ(c,t,r,s,i)___EXT(___release_scmobj)(s);}
#define ___BEGIN_CDEF_FUNCTION_TO_SCMOBJ(c,s,i)___IF_CTOS3(___FUNCTION_to_SCMOBJ,(void*)c,&s,i){
#define ___END_CDEF_FUNCTION_TO_SCMOBJ(c,s,i)___EXT(___release_scmobj)(s);}
#define ___BEGIN_CDEF_BOOL_TO_SCMOBJ(c,s,i)___IF_CTOS3(___BOOL_to_SCMOBJ,c,&s,i){
#define ___END_CDEF_BOOL_TO_SCMOBJ(c,s,i)}
#define ___BEGIN_CDEF_CHARSTRING_TO_SCMOBJ(c,s,i)___IF_CTOS3(___CHARSTRING_to_SCMOBJ,c,&s,i){
#define ___END_CDEF_CHARSTRING_TO_SCMOBJ(c,s,i)___EXT(___release_scmobj)(s);}
#define ___BEGIN_CDEF_LATIN1STRING_TO_SCMOBJ(c,s,i)___IF_CTOS3(___LATIN1STRING_to_SCMOBJ,c,&s,i){
#define ___END_CDEF_LATIN1STRING_TO_SCMOBJ(c,s,i)___EXT(___release_scmobj)(s);}
#define ___BEGIN_CDEF_UTF8STRING_TO_SCMOBJ(c,s,i)___IF_CTOS3(___UTF8STRING_to_SCMOBJ,c,&s,i){
#define ___END_CDEF_UTF8STRING_TO_SCMOBJ(c,s,i)___EXT(___release_scmobj)(s);}
#define ___BEGIN_CDEF_UCS2STRING_TO_SCMOBJ(c,s,i)___IF_CTOS3(___UCS2STRING_to_SCMOBJ,c,&s,i){
#define ___END_CDEF_UCS2STRING_TO_SCMOBJ(c,s,i)___EXT(___release_scmobj)(s);}
#define ___BEGIN_CDEF_UCS4STRING_TO_SCMOBJ(c,s,i)___IF_CTOS3(___UCS4STRING_to_SCMOBJ,c,&s,i){
#define ___END_CDEF_UCS4STRING_TO_SCMOBJ(c,s,i)___EXT(___release_scmobj)(s);}
#define ___BEGIN_CDEF_VARIANT_TO_SCMOBJ(c,s,i)___IF_CTOS3(___VARIANT_to_SCMOBJ,c,&s,i){
#define ___END_CDEF_VARIANT_TO_SCMOBJ(c,s,i)___EXT(___release_scmobj)(s);}


#define ___BEGIN_CLAM_BODY
#define ___END_CLAM_BODY ___AT_END
#define ___BEGIN_CLAM_BODY_CLEANUP ___W_ALL {___BEGIN_CATCH
#define ___END_CLAM_BODY_CLEANUP ___AT_END ___END_CATCH} ___R_ALL

#define ___BEGIN_CLAM_VOID *___fp=___R0; ___W_ALL {___ERR_CODE ___err=___NO_ERR;___WORD ___errmsg=___FAL;
#define ___CLAM_SET_RESULT_VOID ___CLAM_RESULT = ___VOID;
#define ___CLAM_ERROR_VOID ___CLAM_CONV_ERROR
#define ___CLAM_ERROR_CLEANUP_VOID ___CLAM_CONV_ERROR ___EXT(___propagate_error) (___err);
#define ___END_CLAM_VOID } ___R_ALL

#define ___BEGIN_CLAM_SCMOBJ *___fp=___R0; ___W_ALL {___ERR_CODE ___err=___NO_ERR;___WORD ___result;___WORD ___errmsg=___FAL;
#define ___CLAM_SET_RESULT_SCMOBJ ___CLAM_RESULT = ___result;
#define ___CLAM_ERROR_SCMOBJ ___CLAM_CONV_ERROR
#define ___CLAM_ERROR_CLEANUP_SCMOBJ ___CLAM_CONV_ERROR ___EXT(___propagate_error) (___err);
#define ___END_CLAM_SCMOBJ } ___R_ALL

#define ___BEGIN_CLAM(decl)*___fp=___R0; ___W_ALL {___ERR_CODE ___err=___NO_ERR;___WORD ___errmsg=___FAL;decl;
#define ___CLAM_SET_RESULT /* done in conversion function */
#define ___CLAM_ERROR ___CLAM_CONV_ERROR
#define ___CLAM_ERROR_CLEANUP ___CLAM_CONV_ERROR ___EXT(___propagate_error) (___err);
#define ___END_CLAM } ___R_ALL

#define ___BEGIN_CLAM_ARG(i,decl)decl;
#define ___END_CLAM_ARG(i)
#define ___BEGIN_CLAM_ARG_SCMOBJ(i)
#define ___END_CLAM_ARG_SCMOBJ(i)
#define ___CLAM_ARG(i)___PSSTK((i-___FRAME_SPACE(___NARGS+1)))
#define ___CLAM_RESULT ___PSR1

#define ___BEGIN_CLAM_SCMOBJ_TO_S8(s,c,i)___IF_STOC3(___SCMOBJ_to_S8,s,&c,i){
#define ___END_CLAM_SCMOBJ_TO_S8(s,c,i)}
#define ___BEGIN_CLAM_SCMOBJ_TO_U8(s,c,i)___IF_STOC3(___SCMOBJ_to_U8,s,&c,i){
#define ___END_CLAM_SCMOBJ_TO_U8(s,c,i)}
#define ___BEGIN_CLAM_SCMOBJ_TO_S16(s,c,i)___IF_STOC3(___SCMOBJ_to_S16,s,&c,i){
#define ___END_CLAM_SCMOBJ_TO_S16(s,c,i)}
#define ___BEGIN_CLAM_SCMOBJ_TO_U16(s,c,i)___IF_STOC3(___SCMOBJ_to_U16,s,&c,i){
#define ___END_CLAM_SCMOBJ_TO_U16(s,c,i)}
#define ___BEGIN_CLAM_SCMOBJ_TO_S32(s,c,i)___IF_STOC3(___SCMOBJ_to_S32,s,&c,i){
#define ___END_CLAM_SCMOBJ_TO_S32(s,c,i)}
#define ___BEGIN_CLAM_SCMOBJ_TO_U32(s,c,i)___IF_STOC3(___SCMOBJ_to_U32,s,&c,i){
#define ___END_CLAM_SCMOBJ_TO_U32(s,c,i)}
#define ___BEGIN_CLAM_SCMOBJ_TO_S64(s,c,i)___IF_STOC3(___SCMOBJ_to_S64,s,&c,i){
#define ___END_CLAM_SCMOBJ_TO_S64(s,c,i)}
#define ___BEGIN_CLAM_SCMOBJ_TO_U64(s,c,i)___IF_STOC3(___SCMOBJ_to_U64,s,&c,i){
#define ___END_CLAM_SCMOBJ_TO_U64(s,c,i)}
#define ___BEGIN_CLAM_SCMOBJ_TO_F32(s,c,i)___IF_STOC3(___SCMOBJ_to_F32,s,&c,i){
#define ___END_CLAM_SCMOBJ_TO_F32(s,c,i)}
#define ___BEGIN_CLAM_SCMOBJ_TO_F64(s,c,i)___IF_STOC3(___SCMOBJ_to_F64,s,&c,i){
#define ___END_CLAM_SCMOBJ_TO_F64(s,c,i)}
#define ___BEGIN_CLAM_SCMOBJ_TO_CHAR(s,c,i)___IF_STOC3(___SCMOBJ_to_CHAR,s,&c,i){
#define ___END_CLAM_SCMOBJ_TO_CHAR(s,c,i)}
#define ___BEGIN_CLAM_SCMOBJ_TO_SCHAR(s,c,i)___IF_STOC3(___SCMOBJ_to_SCHAR,s,&c,i){
#define ___END_CLAM_SCMOBJ_TO_SCHAR(s,c,i)}
#define ___BEGIN_CLAM_SCMOBJ_TO_UCHAR(s,c,i)___IF_STOC3(___SCMOBJ_to_UCHAR,s,&c,i){
#define ___END_CLAM_SCMOBJ_TO_UCHAR(s,c,i)}
#define ___BEGIN_CLAM_SCMOBJ_TO_LATIN1(s,c,i)___IF_STOC3(___SCMOBJ_to_LATIN1,s,&c,i){
#define ___END_CLAM_SCMOBJ_TO_LATIN1(s,c,i)}
#define ___BEGIN_CLAM_SCMOBJ_TO_UCS2(s,c,i)___IF_STOC3(___SCMOBJ_to_UCS2,s,&c,i){
#define ___END_CLAM_SCMOBJ_TO_UCS2(s,c,i)}
#define ___BEGIN_CLAM_SCMOBJ_TO_UCS4(s,c,i)___IF_STOC3(___SCMOBJ_to_UCS4,s,&c,i){
#define ___END_CLAM_SCMOBJ_TO_UCS4(s,c,i)}
#define ___BEGIN_CLAM_SCMOBJ_TO_SHORT(s,c,i)___IF_STOC3(___SCMOBJ_to_SHORT,s,&c,i){
#define ___END_CLAM_SCMOBJ_TO_SHORT(s,c,i)}
#define ___BEGIN_CLAM_SCMOBJ_TO_USHORT(s,c,i)___IF_STOC3(___SCMOBJ_to_USHORT,s,&c,i){
#define ___END_CLAM_SCMOBJ_TO_USHORT(s,c,i)}
#define ___BEGIN_CLAM_SCMOBJ_TO_INT(s,c,i)___IF_STOC3(___SCMOBJ_to_INT,s,&c,i){
#define ___END_CLAM_SCMOBJ_TO_INT(s,c,i)}
#define ___BEGIN_CLAM_SCMOBJ_TO_UINT(s,c,i)___IF_STOC3(___SCMOBJ_to_UINT,s,&c,i){
#define ___END_CLAM_SCMOBJ_TO_UINT(s,c,i)}
#define ___BEGIN_CLAM_SCMOBJ_TO_LONG(s,c,i)___IF_STOC3(___SCMOBJ_to_LONG,s,&c,i){
#define ___END_CLAM_SCMOBJ_TO_LONG(s,c,i)}
#define ___BEGIN_CLAM_SCMOBJ_TO_ULONG(s,c,i)___IF_STOC3(___SCMOBJ_to_ULONG,s,&c,i){
#define ___END_CLAM_SCMOBJ_TO_ULONG(s,c,i)}
#define ___BEGIN_CLAM_SCMOBJ_TO_FLOAT(s,c,i)___IF_STOC3(___SCMOBJ_to_FLOAT,s,&c,i){
#define ___END_CLAM_SCMOBJ_TO_FLOAT(s,c,i)}
#define ___BEGIN_CLAM_SCMOBJ_TO_DOUBLE(s,c,i)___IF_STOC3(___SCMOBJ_to_DOUBLE,s,&c,i){
#define ___END_CLAM_SCMOBJ_TO_DOUBLE(s,c,i)}
#define ___BEGIN_CLAM_SCMOBJ_TO_POINTER(s,c,t,i)___IF_STOC4(___SCMOBJ_to_POINTER,s,(void**)&c,t,i){
#define ___END_CLAM_SCMOBJ_TO_POINTER(s,c,t,i)}
#define ___BEGIN_CLAM_SCMOBJ_TO_NONNULLPOINTER(s,c,t,i)___IF_STOC4(___SCMOBJ_to_NONNULLPOINTER,s,(void**)&c,t,i){
#define ___END_CLAM_SCMOBJ_TO_NONNULLPOINTER(s,c,t,i)}
#define ___BEGIN_CLAM_SCMOBJ_TO_FUNCTION(s,c,i)___IF_STOC3(___SCMOBJ_to_FUNCTION,s,(void**)&c,i){
#define ___END_CLAM_SCMOBJ_TO_FUNCTION(s,c,i)___EXT(___free_function)(c);}
#define ___BEGIN_CLAM_SCMOBJ_TO_BOOL(s,c,i)___IF_STOC3(___SCMOBJ_to_BOOL,s,&c,i){
#define ___END_CLAM_SCMOBJ_TO_BOOL(s,c,i)}
#define ___BEGIN_CLAM_SCMOBJ_TO_CHARSTRING(s,c,i)___IF_STOC3(___SCMOBJ_to_CHARSTRING,s,&c,i){
#define ___END_CLAM_SCMOBJ_TO_CHARSTRING(s,c,i)___EXT(___free_string)(c);}
#define ___BEGIN_CLAM_SCMOBJ_TO_LATIN1STRING(s,c,i)___IF_STOC3(___SCMOBJ_to_LATIN1STRING,s,&c,i){
#define ___END_CLAM_SCMOBJ_TO_LATIN1STRING(s,c,i)___EXT(___free_string)(c);}
#define ___BEGIN_CLAM_SCMOBJ_TO_UTF8STRING(s,c,i)___IF_STOC3(___SCMOBJ_to_UTF8STRING,s,&c,i){
#define ___END_CLAM_SCMOBJ_TO_UTF8STRING(s,c,i)___EXT(___free_string)(c);}
#define ___BEGIN_CLAM_SCMOBJ_TO_UCS2STRING(s,c,i)___IF_STOC3(___SCMOBJ_to_UCS2STRING,s,&c,i){
#define ___END_CLAM_SCMOBJ_TO_UCS2STRING(s,c,i)___EXT(___free_string)(c);}
#define ___BEGIN_CLAM_SCMOBJ_TO_UCS4STRING(s,c,i)___IF_STOC3(___SCMOBJ_to_UCS4STRING,s,&c,i){
#define ___END_CLAM_SCMOBJ_TO_UCS4STRING(s,c,i)___EXT(___free_string)(c);}
#define ___BEGIN_CLAM_SCMOBJ_TO_VARIANT(s,c,i)___IF_STOC3(___SCMOBJ_to_VARIANT,s,&c,i){
#define ___END_CLAM_SCMOBJ_TO_VARIANT(s,c,i)___EXT(___free_variant)(c);}

#define ___BEGIN_CLAM_S8_TO_SCMOBJ(c,s)___CTOS3(___S8_to_SCMOBJ,c,&s,___RETURN_POS);
#define ___END_CLAM_S8_TO_SCMOBJ(c,s)
#define ___BEGIN_CLAM_U8_TO_SCMOBJ(c,s)___CTOS3(___U8_to_SCMOBJ,c,&s,___RETURN_POS);
#define ___END_CLAM_U8_TO_SCMOBJ(c,s)
#define ___BEGIN_CLAM_S16_TO_SCMOBJ(c,s)___CTOS3(___S16_to_SCMOBJ,c,&s,___RETURN_POS);
#define ___END_CLAM_S16_TO_SCMOBJ(c,s)
#define ___BEGIN_CLAM_U16_TO_SCMOBJ(c,s)___CTOS3(___U16_to_SCMOBJ,c,&s,___RETURN_POS);
#define ___END_CLAM_U16_TO_SCMOBJ(c,s)
#define ___BEGIN_CLAM_S32_TO_SCMOBJ(c,s)___IF_CTOS3(___S32_to_SCMOBJ,c,&s,___RETURN_POS){
#define ___END_CLAM_S32_TO_SCMOBJ(c,s)___EXT(___release_scmobj)(s);}
#define ___BEGIN_CLAM_U32_TO_SCMOBJ(c,s)___IF_CTOS3(___U32_to_SCMOBJ,c,&s,___RETURN_POS){
#define ___END_CLAM_U32_TO_SCMOBJ(c,s)___EXT(___release_scmobj)(s);}
#define ___BEGIN_CLAM_S64_TO_SCMOBJ(c,s)___IF_CTOS3(___S64_to_SCMOBJ,c,&s,___RETURN_POS){
#define ___END_CLAM_S64_TO_SCMOBJ(c,s)___EXT(___release_scmobj)(s);}
#define ___BEGIN_CLAM_U64_TO_SCMOBJ(c,s)___IF_CTOS3(___U64_to_SCMOBJ,c,&s,___RETURN_POS){
#define ___END_CLAM_U64_TO_SCMOBJ(c,s)___EXT(___release_scmobj)(s);}
#define ___BEGIN_CLAM_F32_TO_SCMOBJ(c,s)___IF_CTOS3(___F32_to_SCMOBJ,c,&s,___RETURN_POS){
#define ___END_CLAM_F32_TO_SCMOBJ(c,s)___EXT(___release_scmobj)(s);}
#define ___BEGIN_CLAM_F64_TO_SCMOBJ(c,s)___IF_CTOS3(___F64_to_SCMOBJ,c,&s,___RETURN_POS){
#define ___END_CLAM_F64_TO_SCMOBJ(c,s)___EXT(___release_scmobj)(s);}
#define ___BEGIN_CLAM_CHAR_TO_SCMOBJ(c,s)___CTOS3(___CHAR_to_SCMOBJ,c,&s,___RETURN_POS);
#define ___END_CLAM_CHAR_TO_SCMOBJ(c,s)
#define ___BEGIN_CLAM_SCHAR_TO_SCMOBJ(c,s)___CTOS3(___SCHAR_to_SCMOBJ,c,&s,___RETURN_POS);
#define ___END_CLAM_SCHAR_TO_SCMOBJ(c,s)
#define ___BEGIN_CLAM_UCHAR_TO_SCMOBJ(c,s)___CTOS3(___UCHAR_to_SCMOBJ,c,&s,___RETURN_POS);
#define ___END_CLAM_UCHAR_TO_SCMOBJ(c,s)
#define ___BEGIN_CLAM_LATIN1_TO_SCMOBJ(c,s)___CTOS3(___LATIN1_to_SCMOBJ,c,&s,___RETURN_POS);
#define ___END_CLAM_LATIN1_TO_SCMOBJ(c,s)
#define ___BEGIN_CLAM_UCS2_TO_SCMOBJ(c,s)___CTOS3(___UCS2_to_SCMOBJ,c,&s,___RETURN_POS);
#define ___END_CLAM_UCS2_TO_SCMOBJ(c,s)
#define ___BEGIN_CLAM_UCS4_TO_SCMOBJ(c,s)___CTOS3(___UCS4_to_SCMOBJ,c,&s,___RETURN_POS);
#define ___END_CLAM_UCS4_TO_SCMOBJ(c,s)
#define ___BEGIN_CLAM_SHORT_TO_SCMOBJ(c,s)___IF_CTOS3(___SHORT_to_SCMOBJ,c,&s,___RETURN_POS){
#define ___END_CLAM_SHORT_TO_SCMOBJ(c,s)___EXT(___release_scmobj)(s);}
#define ___BEGIN_CLAM_USHORT_TO_SCMOBJ(c,s)___IF_CTOS3(___USHORT_to_SCMOBJ,c,&s,___RETURN_POS){
#define ___END_CLAM_USHORT_TO_SCMOBJ(c,s)___EXT(___release_scmobj)(s);}
#define ___BEGIN_CLAM_INT_TO_SCMOBJ(c,s)___IF_CTOS3(___INT_to_SCMOBJ,c,&s,___RETURN_POS){
#define ___END_CLAM_INT_TO_SCMOBJ(c,s)___EXT(___release_scmobj)(s);}
#define ___BEGIN_CLAM_UINT_TO_SCMOBJ(c,s)___IF_CTOS3(___UINT_to_SCMOBJ,c,&s,___RETURN_POS){
#define ___END_CLAM_UINT_TO_SCMOBJ(c,s)___EXT(___release_scmobj)(s);}
#define ___BEGIN_CLAM_LONG_TO_SCMOBJ(c,s)___IF_CTOS3(___LONG_to_SCMOBJ,c,&s,___RETURN_POS){
#define ___END_CLAM_LONG_TO_SCMOBJ(c,s)___EXT(___release_scmobj)(s);}
#define ___BEGIN_CLAM_ULONG_TO_SCMOBJ(c,s)___IF_CTOS3(___ULONG_to_SCMOBJ,c,&s,___RETURN_POS){
#define ___END_CLAM_ULONG_TO_SCMOBJ(c,s)___EXT(___release_scmobj)(s);}
#define ___BEGIN_CLAM_FLOAT_TO_SCMOBJ(c,s)___IF_CTOS3(___FLOAT_to_SCMOBJ,c,&s,___RETURN_POS){
#define ___END_CLAM_FLOAT_TO_SCMOBJ(c,s)___EXT(___release_scmobj)(s);}
#define ___BEGIN_CLAM_DOUBLE_TO_SCMOBJ(c,s)___IF_CTOS3(___DOUBLE_to_SCMOBJ,c,&s,___RETURN_POS){
#define ___END_CLAM_DOUBLE_TO_SCMOBJ(c,s)___EXT(___release_scmobj)(s);}
#define ___BEGIN_CLAM_POINTER_TO_SCMOBJ(c,t,r,s)___IF_CTOS5(___POINTER_to_SCMOBJ,(void**)c,t,r,&s,___RETURN_POS){
#define ___END_CLAM_POINTER_TO_SCMOBJ(c,t,r,s)___EXT(___release_scmobj)(s);}
#define ___BEGIN_CLAM_NONNULLPOINTER_TO_SCMOBJ(c,t,r,s)___IF_CTOS5(___NONNULLPOINTER_to_SCMOBJ,(void**)c,t,r,&s,___RETURN_POS){
#define ___END_CLAM_NONNULLPOINTER_TO_SCMOBJ(c,t,r,s)___EXT(___release_scmobj)(s);}
#define ___BEGIN_CLAM_FUNCTION_TO_SCMOBJ(c,s)___IF_CTOS3(___FUNCTION_to_SCMOBJ,(void**)c,&s,___RETURN_POS){
#define ___END_CLAM_FUNCTION_TO_SCMOBJ(c,s)___EXT(___release_scmobj)(s);}
#define ___BEGIN_CLAM_BOOL_TO_SCMOBJ(c,s)___CTOS3(___BOOL_to_SCMOBJ,c,&s,___RETURN_POS);
#define ___END_CLAM_BOOL_TO_SCMOBJ(c,s)
#define ___BEGIN_CLAM_CHARSTRING_TO_SCMOBJ(c,s)___IF_CTOS3(___CHARSTRING_to_SCMOBJ,c,&s,___RETURN_POS){
#define ___END_CLAM_CHARSTRING_TO_SCMOBJ(c,s)___EXT(___release_scmobj)(s);}
#define ___BEGIN_CLAM_LATIN1STRING_TO_SCMOBJ(c,s)___IF_CTOS3(___LATIN1STRING_to_SCMOBJ,c,&s,___RETURN_POS){
#define ___END_CLAM_LATIN1STRING_TO_SCMOBJ(c,s)___EXT(___release_scmobj)(s);}
#define ___BEGIN_CLAM_UTF8STRING_TO_SCMOBJ(c,s)___IF_CTOS3(___UTF8STRING_to_SCMOBJ,c,&s,___RETURN_POS){
#define ___END_CLAM_UTF8STRING_TO_SCMOBJ(c,s)___EXT(___release_scmobj)(s);}
#define ___BEGIN_CLAM_UCS2STRING_TO_SCMOBJ(c,s)___IF_CTOS3(___UCS2STRING_to_SCMOBJ,c,&s,___RETURN_POS){
#define ___END_CLAM_UCS2STRING_TO_SCMOBJ(c,s)___EXT(___release_scmobj)(s);}
#define ___BEGIN_CLAM_UCS4STRING_TO_SCMOBJ(c,s)___IF_CTOS3(___UCS4STRING_to_SCMOBJ,c,&s,___RETURN_POS){
#define ___END_CLAM_UCS4STRING_TO_SCMOBJ(c,s)___EXT(___release_scmobj)(s);}
#define ___BEGIN_CLAM_VARIANT_TO_SCMOBJ(c,s)___IF_CTOS3(___VARIANT_to_SCMOBJ,c,&s,___RETURN_POS){
#define ___END_CLAM_VARIANT_TO_SCMOBJ(c,s)___EXT(___release_scmobj)(s);}

#endif

/*---------------------------------------------------------------------------*/

#ifndef ___DEFINES_ONLY

#ifdef _MSC_VER
#ifdef ___DYNAMIC
int _fltused; /* needed if floating-point used. */
#endif
#endif

#ifdef __BORLANDC__
#ifdef ___SHARED
#define _RTLDLL
#endif
#endif

#include <math.h>

#ifdef ___USE_SETJMP

#include <setjmp.h>

typedef struct ___jmpbuf_struct
  {
    jmp_buf buf;
  } ___jmpbuf_struct;

typedef void *___VARIANT; /* not yet implemented */

#endif

#ifdef ___BUILTIN_64BIT_INT_TYPE

/*
 * If the symbol ___BUILTIN_64BIT_INT_TYPE is defined then the data
 * types ___S64 and ___U64 are builtin integer types.
 */

#define ___S64_from_SM32(val) \
___CAST(___S64,___CAST_S32(val))

#define ___S64_from_SM32_UM32(hi32,lo32) \
((___CAST(___S64,___CAST_S32(hi32)) << 32) + ___CAST(___S64,___CAST_U32(lo32)))

#define ___S64_from_LONG(val) \
___CAST(___S64,val)

#define ___S64_to_LONG(val) \
___CAST(long,val)

#define ___S64_fits_in_width(val,width) \
((((val) >> ((width)-1)) == 0) || (((val) >> ((width)-1)) == -1))

#define ___S64_hi32(val)___CAST_S32((val) >> 32)
#define ___S64_lo32(val)___CAST_U32(val)

#define ___U64_from_UM32(val) \
___CAST(___U64,___CAST_U32(val))

#define ___U64_from_UM32_UM32(hi32,lo32) \
((___CAST(___U64,___CAST_U32(hi32)) << 32) + ___CAST(___U64,___CAST_U32(lo32)))

#define ___U64_from_ULONG(val) \
___CAST(___U64,val)

#define ___U64_to_ULONG(val) \
___CAST(unsigned long,val)

#define ___U64_fits_in_width(val,width) \
(((val) >> (width)) == 0)

#define ___U64_hi32(val)___CAST_U32((val) >> 32)
#define ___U64_lo32(val)___CAST_U32(val)

#else

/*
 * If the symbol ___BUILTIN_64BIT_INT_TYPE is not defined then the
 * data types ___S64 and ___U64 are structures.
 */

typedef struct
  {
#ifdef ___CPU_BIGEND
    ___S32 hi32;
    ___U32 lo32;
#else
    ___U32 lo32;
    ___S32 hi32;
#endif
  } ___S64;

typedef struct
  {
#ifdef ___CPU_BIGEND
    ___U32 hi32;
    ___U32 lo32;
#else
    ___U32 lo32;
    ___U32 hi32;
#endif
  } ___U64;

#define ___S64_from_SM32(val) \
___EXT(___S64_from_SM32_fn) (val)

#define ___S64_from_SM32_UM32(hi32,lo32) \
___EXT(___S64_from_SM32_UM32_fn) (hi32,lo32)

#define ___S64_from_LONG(val) \
___EXT(___S64_from_LONG_fn) (val)

#define ___S64_to_LONG(val) \
___EXT(___S64_to_LONG_fn) (val)

#define ___S64_fits_in_width(val,width) \
___EXT(___S64_fits_in_width_fn) (val,width)

#define ___S64_hi32(val)val.hi32
#define ___S64_lo32(val)val.lo32

#define ___U64_from_UM32(val) \
___EXT(___U64_from_UM32_fn) (val)

#define ___U64_from_UM32_UM32(hi32,lo32) \
___EXT(___U64_from_UM32_UM32_fn) (hi32,lo32)

#define ___U64_from_ULONG(val) \
___EXT(___U64_from_ULONG_fn) (val)

#define ___U64_to_ULONG(val) \
___EXT(___U64_to_ULONG_fn) (val)

#define ___U64_fits_in_width(val,width) \
___EXT(___U64_fits_in_width_fn) (val,width)

#define ___U64_hi32(val)val.hi32
#define ___U64_lo32(val)val.lo32

#endif

/* Symbol/keyword structure */

typedef ___FAKEWORD ___symkey_struct[];

/* Global variable structure */

typedef struct ___glo_struct
  {
    ___WORD val, prm, next;
  } ___glo_struct;

/* Processor structure */

typedef struct ___processor_state_struct
  {
    ___WORD *stack_trip;
    ___WORD *stack_limit;
    ___WORD *fp;
    ___WORD *stack_start;
    ___WORD *stack_break;

    ___WORD *heap_limit;
    ___WORD *hp;

    ___WORD current_thread;
    ___WORD run_queue;

    ___WORD r[___NB_GVM_REGS];
    ___WORD pc;
    ___WORD temp1;
    ___WORD temp2;
    ___WORD temp3;
    ___WORD temp4;

    int na;
    ___BOOL intr_enabled;
    ___BOOL intr_flag[___NB_INTRS];

    ___WORD glo_list_head;
    ___WORD glo_list_tail;

    ___WORD handler_break;
    ___WORD handler_stack_limit;
    ___WORD handler_heap_limit;
    ___WORD handler_not_proc;
    ___WORD handler_not_proc_glo;
    ___WORD handler_wrong_nargs;
    ___WORD handler_get_rest;
    ___WORD handler_get_key;
    ___WORD handler_get_key_rest;
    ___WORD handler_force;
    ___WORD handler_clam_conv_error;
    ___WORD handler_cdef_conv_error;
    ___WORD handler_return_to_c;

    ___WORD executable_wills;
    ___WORD non_executable_wills;

    void (*dummy1) ___PVOID;
    void (*dummy2) ___PVOID;
    void (*dummy3) ___PVOID;
    void (*dummy4) ___PVOID;
    void (*dummy5) ___PVOID;
    void (*dummy6) ___PVOID;
    void (*dummy7) ___PVOID;
    void (*dummy8) ___PVOID;

#ifdef ___USE_SETJMP
    ___jmpbuf_struct *catcher;
#endif

#ifdef ___DEBUG_STACK_LIMIT
    int poll_line;
    char *poll_file;
    int stack_limit_line;
    char *stack_limit_file;
#endif

#ifdef ___DEBUG_HEAP_LIMIT
    int check_heap_line;
    char *check_heap_file;
    int heap_limit_line;
    char *heap_limit_file;
#endif
  } ___processor_state_struct, *___processor_state;

typedef ___WORD (*___host) ___P((___processor_state),());

typedef struct ___label_struct
  {
    ___WORD header;
    ___WORD entry_or_descr;
    void *host_label;
    ___FAKEHOST host;
  } ___label_struct;

/* Module structure */

typedef struct ___module_struct
  {
    int version_major;
    int version_minor;
    int kind;
    ___UTF8STRING name;
    int flags;
    ___FAKEWORD *glo_tbl; int glo_count; int sup_count;
    ___UTF8STRING *glo_names;
    ___FAKEWORD *sym_tbl; int sym_count;
    ___UTF8STRING *sym_names;
    ___FAKEWORD *key_tbl; int key_count;
    ___UTF8STRING *key_names;
    ___WORD *lp;
    ___label_struct *lbl_tbl; int lbl_count;
    ___WORD *ofd_tbl; int ofd_length;
    ___WORD *cns_tbl; int cns_count;
    ___FAKEWORD *sub_tbl; int sub_count;
    void (*init_proc) ___PVOID;
    struct ___module_struct *next;

    void (*dummy1) ___PVOID;
    void (*dummy2) ___PVOID;
    void (*dummy3) ___PVOID;
    void (*dummy4) ___PVOID;
    void (*dummy5) ___PVOID;
    void (*dummy6) ___PVOID;
    void (*dummy7) ___PVOID;
    void (*dummy8) ___PVOID;
  } ___module_struct;

/* Linkfile structure */

typedef struct ___linkfile_struct
  {
    int version_major;
    int version_minor;
    int kind;
    ___UTF8STRING name;
    ___FAKEWORD *sym_list;
    ___FAKEWORD *key_list;
    void **linker_tbl;

    void (*dummy1) ___PVOID;
    void (*dummy2) ___PVOID;
    void (*dummy3) ___PVOID;
    void (*dummy4) ___PVOID;
    void (*dummy5) ___PVOID;
    void (*dummy6) ___PVOID;
    void (*dummy7) ___PVOID;
    void (*dummy8) ___PVOID;
  } ___linkfile_struct;

/* Module or linkfile structure */

typedef union
  {
    ___module_struct module;
    ___linkfile_struct linkfile;
  } *___mod_or_lnk;

/* Global state structure */

typedef struct ___global_state_struct
  {
    ___processor_state_struct pstate;

    ___F64 nb_gcs;
    ___F64 gc_user_time;
    ___F64 gc_sys_time;
    ___F64 gc_real_time;
    ___F64 bytes_allocated_minus_occupied;

    void (*dummy1) ___PVOID;
    void (*dummy2) ___PVOID;
    void (*dummy3) ___PVOID;
    void (*dummy4) ___PVOID;
    void (*dummy5) ___PVOID;
    void (*dummy6) ___PVOID;
    void (*dummy7) ___PVOID;
    void (*dummy8) ___PVOID;

#ifdef ___CANT_IMPORT_CLIB_DYNAMICALLY
    double (*fabs)
       ___P((double x),
            ());
    double (*floor)
       ___P((double x),
            ());
    double (*ceil)
       ___P((double x),
            ());
    double (*exp)
       ___P((double x),
            ());
    double (*log)
       ___P((double x),
            ());
    double (*sin)
       ___P((double x),
            ());
    double (*cos)
       ___P((double x),
            ());
    double (*tan)
       ___P((double x),
            ());
    double (*asin)
       ___P((double x),
            ());
    double (*acos)
       ___P((double x),
            ());
    double (*atan)
       ___P((double x),
            ());
#ifdef ___GOOD_ATAN2
    double (*atan2)
       ___P((double y,
             double x),
            ());
#endif
    double (*sqrt)
       ___P((double x),
            ());
#endif

#ifdef ___USE_SETJMP
#ifdef ___CANT_IMPORT_SETJMP_DYNAMICALLY
    int (*setjmp)
       ___P((jmp_buf env),
            ());
#endif
#endif

#ifdef ___CANT_IMPORT_DYNAMICALLY
    double (*___copysign)
       ___P((double x,
             double y),
            ());
    int (*___isfinite)
       ___P((double x),
            ());
    int (*___isnan)
       ___P((double x),
            ());
    double (*___round)
       ___P((double x),
            ());
#ifndef ___GOOD_ATAN2
    double (*___atan2)
       ___P((double y,
             double x),
            ());
#endif
    ___S64 (*___S64_from_SM32_fn)
       ___P((___SM32 val),
            ());
    ___S64 (*___S64_from_SM32_UM32_fn)
       ___P((___SM32 hi32,
             ___UM32 lo32),
            ());
    ___S64 (*___S64_from_LONG_fn)
       ___P((long val),
            ());
    long (*___S64_to_LONG_fn)
       ___P((___S64 val),
            ());
    ___BOOL (*___S64_fits_in_width_fn)
       ___P((___S64 val,
             int width),
            ());
    ___U64 (*___U64_from_UM32_fn)
       ___P((___UM32 val),
            ());
    ___U64 (*___U64_from_UM32_UM32_fn)
       ___P((___UM32 hi32,
             ___UM32 lo32),
            ());
    ___U64 (*___U64_from_ULONG_fn)
       ___P((unsigned long val),
            ());
    unsigned long (*___U64_to_ULONG_fn)
       ___P((___U64 val),
            ());
    ___BOOL (*___U64_fits_in_width_fn)
       ___P((___U64 val,
             int width),
            ());
    ___ERR_CODE (*___SCMOBJ_to_S8)
       ___P((___SCMOBJ obj,
             ___S8 *x,
             int arg_num),
            ());
    ___ERR_CODE (*___SCMOBJ_to_U8)
       ___P((___SCMOBJ obj,
             ___U8 *x,
             int arg_num),
            ());
    ___ERR_CODE (*___SCMOBJ_to_S16)
       ___P((___SCMOBJ obj,
             ___S16 *x,
             int arg_num),
            ());
    ___ERR_CODE (*___SCMOBJ_to_U16)
       ___P((___SCMOBJ obj,
             ___U16 *x,
             int arg_num),
            ());
    ___ERR_CODE (*___SCMOBJ_to_S32)
       ___P((___SCMOBJ obj,
             ___S32 *x,
             int arg_num),
            ());
    ___ERR_CODE (*___SCMOBJ_to_U32)
       ___P((___SCMOBJ obj,
             ___U32 *x,
             int arg_num),
            ());
    ___ERR_CODE (*___SCMOBJ_to_S64)
       ___P((___SCMOBJ obj,
             ___S64 *x,
             int arg_num),
            ());
    ___ERR_CODE (*___SCMOBJ_to_U64)
       ___P((___SCMOBJ obj,
             ___U64 *x,
             int arg_num),
            ());
    ___ERR_CODE (*___SCMOBJ_to_F32)
       ___P((___SCMOBJ obj,
             ___F32 *x,
             int arg_num),
            ());
    ___ERR_CODE (*___SCMOBJ_to_F64)
       ___P((___SCMOBJ obj,
             ___F64 *x,
             int arg_num),
            ());
    ___ERR_CODE (*___SCMOBJ_to_CHAR)
       ___P((___SCMOBJ obj,
             char *x,
             int arg_num),
            ());
    ___ERR_CODE (*___SCMOBJ_to_SCHAR)
       ___P((___SCMOBJ obj,
             ___SCHAR *x,
             int arg_num),
            ());
    ___ERR_CODE (*___SCMOBJ_to_UCHAR)
       ___P((___SCMOBJ obj,
             unsigned char *x,
             int arg_num),
            ());
    ___ERR_CODE (*___SCMOBJ_to_LATIN1)
       ___P((___SCMOBJ obj,
             ___LATIN1 *x,
             int arg_num),
            ());
    ___ERR_CODE (*___SCMOBJ_to_UCS2)
       ___P((___SCMOBJ obj,
             ___UCS2 *x,
             int arg_num),
            ());
    ___ERR_CODE (*___SCMOBJ_to_UCS4)
       ___P((___SCMOBJ obj,
             ___UCS4 *x,
             int arg_num),
            ());
    ___ERR_CODE (*___SCMOBJ_to_SHORT)
       ___P((___SCMOBJ obj,
             short *x,
             int arg_num),
            ());
    ___ERR_CODE (*___SCMOBJ_to_USHORT)
       ___P((___SCMOBJ obj,
             unsigned short *x,
             int arg_num),
            ());
    ___ERR_CODE (*___SCMOBJ_to_INT)
       ___P((___SCMOBJ obj,
             int *x,
             int arg_num),
            ());
    ___ERR_CODE (*___SCMOBJ_to_UINT)
       ___P((___SCMOBJ obj,
             unsigned int *x,
             int arg_num),
            ());
    ___ERR_CODE (*___SCMOBJ_to_LONG)
       ___P((___SCMOBJ obj,
             long *x,
             int arg_num),
            ());
    ___ERR_CODE (*___SCMOBJ_to_ULONG)
       ___P((___SCMOBJ obj,
             unsigned long *x,
             int arg_num),
            ());
    ___ERR_CODE (*___SCMOBJ_to_FLOAT)
       ___P((___SCMOBJ obj,
             float *x,
             int arg_num),
            ());
    ___ERR_CODE (*___SCMOBJ_to_DOUBLE)
       ___P((___SCMOBJ obj,
             double *x,
             int arg_num),
            ());
    ___ERR_CODE (*___SCMOBJ_to_POINTER)
       ___P((___SCMOBJ obj,
             void **x,
             ___SCMOBJ tag,
             int arg_num),
            ());
    ___ERR_CODE (*___SCMOBJ_to_NONNULLPOINTER)
       ___P((___SCMOBJ obj,
             void **x,
             ___SCMOBJ tag,
             int arg_num),
            ());
    ___ERR_CODE (*___SCMOBJ_to_FUNCTION)
       ___P((___SCMOBJ obj,
             void **x,
             int arg_num),
            ());
    ___ERR_CODE (*___SCMOBJ_to_BOOL)
       ___P((___SCMOBJ obj,
             ___BOOL *x,
             int arg_num),
            ());
    ___ERR_CODE (*___SCMOBJ_to_CHARSTRING)
       ___P((___SCMOBJ obj,
             char **x,
             int arg_num),
            ());
    ___ERR_CODE (*___SCMOBJ_to_LATIN1STRING)
       ___P((___SCMOBJ obj,
             ___LATIN1STRING *x,
             int arg_num),
            ());
    ___ERR_CODE (*___SCMOBJ_to_UTF8STRING)
       ___P((___SCMOBJ obj,
             ___UTF8STRING *x,
             int arg_num),
            ());
    ___ERR_CODE (*___SCMOBJ_to_UCS2STRING)
       ___P((___SCMOBJ obj,
             ___UCS2STRING *x,
             int arg_num),
            ());
    ___ERR_CODE (*___SCMOBJ_to_UCS4STRING)
       ___P((___SCMOBJ obj,
             ___UCS4STRING *x,
             int arg_num),
            ());
    ___ERR_CODE (*___SCMOBJ_to_VARIANT)
       ___P((___SCMOBJ obj,
             ___VARIANT *x,
             int arg_num),
            ());
    void (*___free_function)
       ___P((void *x),
            ());
    void (*___free_string)
       ___P((void *x),
            ());
    void (*___free_variant)
       ___P((___VARIANT x),
            ());
    ___ERR_CODE (*___S8_to_SCMOBJ)
       ___P((___S8 x,
             ___SCMOBJ *obj,
             int arg_num),
            ());
    ___ERR_CODE (*___U8_to_SCMOBJ)
       ___P((___U8 x,
             ___SCMOBJ *obj,
             int arg_num),
            ());
    ___ERR_CODE (*___S16_to_SCMOBJ)
       ___P((___S16 x,
             ___SCMOBJ *obj,
             int arg_num),
            ());
    ___ERR_CODE (*___U16_to_SCMOBJ)
       ___P((___U16 x,
             ___SCMOBJ *obj,
             int arg_num),
            ());
    ___ERR_CODE (*___S32_to_SCMOBJ)
       ___P((___S32 x,
             ___SCMOBJ *obj,
             int arg_num),
            ());
    ___ERR_CODE (*___U32_to_SCMOBJ)
       ___P((___U32 x,
             ___SCMOBJ *obj,
             int arg_num),
            ());
    ___ERR_CODE (*___S64_to_SCMOBJ)
       ___P((___S64 x,
             ___SCMOBJ *obj,
             int arg_num),
            ());
    ___ERR_CODE (*___U64_to_SCMOBJ)
       ___P((___U64 x,
             ___SCMOBJ *obj,
             int arg_num),
            ());
    ___ERR_CODE (*___F32_to_SCMOBJ)
       ___P((___F32 x,
             ___SCMOBJ *obj,
             int arg_num),
            ());
    ___ERR_CODE (*___F64_to_SCMOBJ)
       ___P((___F64 x,
             ___SCMOBJ *obj,
             int arg_num),
            ());
    ___ERR_CODE (*___CHAR_to_SCMOBJ)
       ___P((char x,
             ___SCMOBJ *obj,
             int arg_num),
            ());
    ___ERR_CODE (*___SCHAR_to_SCMOBJ)
       ___P((___SCHAR x,
             ___SCMOBJ *obj,
             int arg_num),
            ());
    ___ERR_CODE (*___UCHAR_to_SCMOBJ)
       ___P((unsigned char x,
             ___SCMOBJ *obj,
             int arg_num),
            ());
    ___ERR_CODE (*___LATIN1_to_SCMOBJ)
       ___P((___LATIN1 x,
             ___SCMOBJ *obj,
             int arg_num),
            ());
    ___ERR_CODE (*___UCS2_to_SCMOBJ)
       ___P((___UCS2 x,
             ___SCMOBJ *obj,
             int arg_num),
            ());
    ___ERR_CODE (*___UCS4_to_SCMOBJ)
       ___P((___UCS4 x,
             ___SCMOBJ *obj,
             int arg_num),
            ());
    ___ERR_CODE (*___SHORT_to_SCMOBJ)
       ___P((short x,
             ___SCMOBJ *obj,
             int arg_num),
            ());
    ___ERR_CODE (*___USHORT_to_SCMOBJ)
       ___P((unsigned short x,
             ___SCMOBJ *obj,
             int arg_num),
            ());
    ___ERR_CODE (*___INT_to_SCMOBJ)
       ___P((int x,
             ___SCMOBJ *obj,
             int arg_num),
            ());
    ___ERR_CODE (*___UINT_to_SCMOBJ)
       ___P((unsigned int x,
             ___SCMOBJ *obj,
             int arg_num),
            ());
    ___ERR_CODE (*___LONG_to_SCMOBJ)
       ___P((long x,
             ___SCMOBJ *obj,
             int arg_num),
            ());
    ___ERR_CODE (*___ULONG_to_SCMOBJ)
       ___P((unsigned long x,
             ___SCMOBJ *obj,
             int arg_num),
            ());
    ___ERR_CODE (*___FLOAT_to_SCMOBJ)
       ___P((float x,
             ___SCMOBJ *obj,
             int arg_num),
            ());
    ___ERR_CODE (*___DOUBLE_to_SCMOBJ)
       ___P((double x,
             ___SCMOBJ *obj,
             int arg_num),
            ());
    ___ERR_CODE (*___POINTER_to_SCMOBJ)
       ___P((void *x,
             ___SCMOBJ tag,
             ___SCMOBJ (*release_fn) ___P((void *ptr),()),
             ___SCMOBJ *obj,
             int arg_num),
            ());
    ___ERR_CODE (*___NONNULLPOINTER_to_SCMOBJ)
       ___P((void *x,
             ___SCMOBJ tag,
             ___SCMOBJ (*release_fn) ___P((void *ptr),()),
             ___SCMOBJ *obj,
             int arg_num),
            ());
    ___ERR_CODE (*___FUNCTION_to_SCMOBJ)
       ___P((void *x,
             ___SCMOBJ *obj,
             int arg_num),
            ());
    ___ERR_CODE (*___BOOL_to_SCMOBJ)
       ___P((___BOOL x,
             ___SCMOBJ *obj,
             int arg_num),
            ());
    ___ERR_CODE (*___CHARSTRING_to_SCMOBJ)
       ___P((char *x,
             ___SCMOBJ *obj,
             int arg_num),
            ());
    ___ERR_CODE (*___LATIN1STRING_to_SCMOBJ)
       ___P((___LATIN1STRING x,
             ___SCMOBJ *obj,
             int arg_num),
            ());
    ___ERR_CODE (*___UTF8STRING_to_SCMOBJ)
       ___P((___UTF8STRING x,
             ___SCMOBJ *obj,
             int arg_num),
            ());
    ___ERR_CODE (*___UCS2STRING_to_SCMOBJ)
       ___P((___UCS2STRING x,
             ___SCMOBJ *obj,
             int arg_num),
            ());
    ___ERR_CODE (*___UCS4STRING_to_SCMOBJ)
       ___P((___UCS4STRING x,
             ___SCMOBJ *obj,
             int arg_num),
            ());
    ___ERR_CODE (*___VARIANT_to_SCMOBJ)
       ___P((___VARIANT x,
             ___SCMOBJ *obj,
             int arg_num),
            ());
    ___SCMOBJ (*___release_pointer)
       ___P((___SCMOBJ obj),
            ());
    ___ERR_CODE (*___make_cdef_stack_marker)
       ___P((___SCMOBJ *marker),
            ());
    void (*___kill_cdef_stack_marker)
       ___P((___SCMOBJ marker),
            ());
    void (*___release_scmobj)
       ___P((___SCMOBJ obj),
            ());
    ___SCMOBJ (*___alloc_scmobj)
       ___P((int subtype,
             long bytes,
             int kind),
            ());
    ___SCMOBJ (*___make_pair)
       ___P((___SCMOBJ car,
             ___SCMOBJ cdr,
             int kind),
            ());
    ___SCMOBJ (*___make_vector)
       ___P((long length,
             ___SCMOBJ init,
             int kind),
            ());
    void (*___still_obj_refcount_inc)
       ___P((___SCMOBJ obj),
            ());
    void (*___still_obj_refcount_dec)
       ___P((___SCMOBJ obj),
            ());
    void (*___cleanup) ___PVOID;
    void (*___cleanup_and_exit_process)
       ___P((int status),
            ());
    ___ERR_CODE (*___call)
       ___P((int nargs,
             ___SCMOBJ proc,
             ___SCMOBJ marker),
            ());
    void (*___propagate_error)
       ___P((___ERR_CODE err),
            ());
    void (*___raise_interrupt)
       ___P((int code),
            ());
#endif
  } ___global_state_struct;

___LOCAL ___WORD ___lp; /* Module's label table pointer */
___GLO_DECL
___SYM_DECL
___KEY_DECL

/* Setup parameters structure */

typedef struct ___setup_params_struct
  {
    int version_major;
    int version_minor;
    int argc;
    char **argv;
    long min_heap;
    long max_heap;
    int live_percent;
    long (*gc_hook) ___P((long avail, long live),());
    void (*fatal_error) ___P((char *msg),());
    ___BOOL standard;
    int debug_level;
    ___mod_or_lnk (*linker) ___P((___global_state_struct*),());
    void (*dummy1) ___PVOID;
    void (*dummy2) ___PVOID;
    void (*dummy3) ___PVOID;
    void (*dummy4) ___PVOID;
    void (*dummy5) ___PVOID;
    void (*dummy6) ___PVOID;
    void (*dummy7) ___PVOID;
    void (*dummy8) ___PVOID;
  } ___setup_params_struct;

#ifndef ___INCLUDED_FROM_C_INTF
___IMP_FUNC(___S64,___S64_from_SM32_fn)
   ___P((___SM32 val),
        ());
___IMP_FUNC(___S64,___S64_from_SM32_UM32_fn)
   ___P((___SM32 hi32,
         ___UM32 lo32),
        ());
___IMP_FUNC(___S64,___S64_from_LONG_fn)
   ___P((long val),
        ());
___IMP_FUNC(long,___S64_to_LONG_fn)
   ___P((___S64 val),
        ());
___IMP_FUNC(___BOOL,___S64_fits_in_width_fn)
   ___P((___S64 val,
         int width),
        ());
___IMP_FUNC(___U64,___U64_from_UM32_fn)
   ___P((___UM32 val),
        ());
___IMP_FUNC(___U64,___U64_from_UM32_UM32_fn)
   ___P((___UM32 hi32,
         ___UM32 lo32),
        ());
___IMP_FUNC(___U64,___U64_from_ULONG_fn)
   ___P((unsigned long val),
        ());
___IMP_FUNC(unsigned long,___U64_to_ULONG_fn)
   ___P((___U64 val),
        ());
___IMP_FUNC(___BOOL,___U64_fits_in_width_fn)
   ___P((___U64 val,
         int width),
        ());
___IMP_FUNC(___ERR_CODE,___SCMOBJ_to_S8)
   ___P((___SCMOBJ obj,
         ___S8 *x,
         int arg_num),
        ());
___IMP_FUNC(___ERR_CODE,___SCMOBJ_to_U8)
   ___P((___SCMOBJ obj,
         ___U8 *x,
         int arg_num),
        ());
___IMP_FUNC(___ERR_CODE,___SCMOBJ_to_S16)
   ___P((___SCMOBJ obj,
         ___S16 *x,
         int arg_num),
        ());
___IMP_FUNC(___ERR_CODE,___SCMOBJ_to_U16)
   ___P((___SCMOBJ obj,
         ___U16 *x,
         int arg_num),
        ());
___IMP_FUNC(___ERR_CODE,___SCMOBJ_to_S32)
   ___P((___SCMOBJ obj,
         ___S32 *x,
         int arg_num),
        ());
___IMP_FUNC(___ERR_CODE,___SCMOBJ_to_U32)
   ___P((___SCMOBJ obj,
         ___U32 *x,
         int arg_num),
        ());
___IMP_FUNC(___ERR_CODE,___SCMOBJ_to_S64)
   ___P((___SCMOBJ obj,
         ___S64 *x,
         int arg_num),
        ());
___IMP_FUNC(___ERR_CODE,___SCMOBJ_to_U64)
   ___P((___SCMOBJ obj,
         ___U64 *x,
         int arg_num),
        ());
___IMP_FUNC(___ERR_CODE,___SCMOBJ_to_F32)
   ___P((___SCMOBJ obj,
         ___F32 *x,
         int arg_num),
        ());
___IMP_FUNC(___ERR_CODE,___SCMOBJ_to_F64)
   ___P((___SCMOBJ obj,
         ___F64 *x,
         int arg_num),
        ());
___IMP_FUNC(___ERR_CODE,___SCMOBJ_to_CHAR)
   ___P((___SCMOBJ obj,
         char *x,
         int arg_num),
        ());
___IMP_FUNC(___ERR_CODE,___SCMOBJ_to_SCHAR)
   ___P((___SCMOBJ obj,
         ___SCHAR *x,
         int arg_num),
        ());
___IMP_FUNC(___ERR_CODE,___SCMOBJ_to_UCHAR)
   ___P((___SCMOBJ obj,
         unsigned char *x,
         int arg_num),
        ());
___IMP_FUNC(___ERR_CODE,___SCMOBJ_to_LATIN1)
   ___P((___SCMOBJ obj,
         ___LATIN1 *x,
         int arg_num),
        ());
___IMP_FUNC(___ERR_CODE,___SCMOBJ_to_UCS2)
   ___P((___SCMOBJ obj,
         ___UCS2 *x,
         int arg_num),
        ());
___IMP_FUNC(___ERR_CODE,___SCMOBJ_to_UCS4)
   ___P((___SCMOBJ obj,
         ___UCS4 *x,
         int arg_num),
        ());
___IMP_FUNC(___ERR_CODE,___SCMOBJ_to_SHORT)
   ___P((___SCMOBJ obj,
         short *x,
         int arg_num),
        ());
___IMP_FUNC(___ERR_CODE,___SCMOBJ_to_USHORT)
   ___P((___SCMOBJ obj,
         unsigned short *x,
         int arg_num),
        ());
___IMP_FUNC(___ERR_CODE,___SCMOBJ_to_INT)
   ___P((___SCMOBJ obj,
         int *x,
         int arg_num),
        ());
___IMP_FUNC(___ERR_CODE,___SCMOBJ_to_UINT)
   ___P((___SCMOBJ obj,
         unsigned int *x,
         int arg_num),
        ());
___IMP_FUNC(___ERR_CODE,___SCMOBJ_to_LONG)
   ___P((___SCMOBJ obj,
         long *x,
         int arg_num),
        ());
___IMP_FUNC(___ERR_CODE,___SCMOBJ_to_ULONG)
   ___P((___SCMOBJ obj,
         unsigned long *x,
         int arg_num),
        ());
___IMP_FUNC(___ERR_CODE,___SCMOBJ_to_FLOAT)
   ___P((___SCMOBJ obj,
         float *x,
         int arg_num),
        ());
___IMP_FUNC(___ERR_CODE,___SCMOBJ_to_DOUBLE)
   ___P((___SCMOBJ obj,
         double *x,
         int arg_num),
        ());
___IMP_FUNC(___ERR_CODE,___SCMOBJ_to_POINTER)
   ___P((___SCMOBJ obj,
         void **x,
         ___SCMOBJ tag,
         int arg_num),
        ());
___IMP_FUNC(___ERR_CODE,___SCMOBJ_to_NONNULLPOINTER)
   ___P((___SCMOBJ obj,
         void **x,
         ___SCMOBJ tag,
         int arg_num),
        ());
___IMP_FUNC(___ERR_CODE,___SCMOBJ_to_FUNCTION)
   ___P((___SCMOBJ obj,
         void **x,
         int arg_num),
        ());
___IMP_FUNC(___ERR_CODE,___SCMOBJ_to_BOOL)
   ___P((___SCMOBJ obj,
         ___BOOL *x,
         int arg_num),
        ());
___IMP_FUNC(___ERR_CODE,___SCMOBJ_to_CHARSTRING)
   ___P((___SCMOBJ obj,
         char **x,
         int arg_num),
        ());
___IMP_FUNC(___ERR_CODE,___SCMOBJ_to_LATIN1STRING)
   ___P((___SCMOBJ obj,
         ___LATIN1STRING *x,
         int arg_num),
        ());
___IMP_FUNC(___ERR_CODE,___SCMOBJ_to_UTF8STRING)
   ___P((___SCMOBJ obj,
         ___UTF8STRING *x,
         int arg_num),
        ());
___IMP_FUNC(___ERR_CODE,___SCMOBJ_to_UCS2STRING)
   ___P((___SCMOBJ obj,
         ___UCS2STRING *x,
         int arg_num),
        ());
___IMP_FUNC(___ERR_CODE,___SCMOBJ_to_UCS4STRING)
   ___P((___SCMOBJ obj,
         ___UCS4STRING *x,
         int arg_num),
        ());
___IMP_FUNC(___ERR_CODE,___SCMOBJ_to_VARIANT)
   ___P((___SCMOBJ obj,
         ___VARIANT *x,
         int arg_num),
        ());
___IMP_FUNC(void,___free_function)
   ___P((void *x),
        ());
___IMP_FUNC(void,___free_string)
   ___P((void *x),
        ());
___IMP_FUNC(void,___free_variant)
   ___P((___VARIANT x),
        ());
___IMP_FUNC(___ERR_CODE,___S8_to_SCMOBJ)
   ___P((___S8 x,
         ___SCMOBJ *obj,
         int arg_num),
        ());
___IMP_FUNC(___ERR_CODE,___U8_to_SCMOBJ)
   ___P((___U8 x,
         ___SCMOBJ *obj,
         int arg_num),
        ());
___IMP_FUNC(___ERR_CODE,___S16_to_SCMOBJ)
   ___P((___S16 x,
         ___SCMOBJ *obj,
         int arg_num),
        ());
___IMP_FUNC(___ERR_CODE,___U16_to_SCMOBJ)
   ___P((___U16 x,
         ___SCMOBJ *obj,
         int arg_num),
        ());
___IMP_FUNC(___ERR_CODE,___S32_to_SCMOBJ)
   ___P((___S32 x,
         ___SCMOBJ *obj,
         int arg_num),
        ());
___IMP_FUNC(___ERR_CODE,___U32_to_SCMOBJ)
   ___P((___U32 x,
         ___SCMOBJ *obj,
         int arg_num),
        ());
___IMP_FUNC(___ERR_CODE,___S64_to_SCMOBJ)
   ___P((___S64 x,
         ___SCMOBJ *obj,
         int arg_num),
        ());
___IMP_FUNC(___ERR_CODE,___U64_to_SCMOBJ)
   ___P((___U64 x,
         ___SCMOBJ *obj,
         int arg_num),
        ());
___IMP_FUNC(___ERR_CODE,___F32_to_SCMOBJ)
   ___P((___F32 x,
         ___SCMOBJ *obj,
         int arg_num),
        ());
___IMP_FUNC(___ERR_CODE,___F64_to_SCMOBJ)
   ___P((___F64 x,
         ___SCMOBJ *obj,
         int arg_num),
        ());
___IMP_FUNC(___ERR_CODE,___CHAR_to_SCMOBJ)
   ___P((char x,
         ___SCMOBJ *obj,
         int arg_num),
        ());
___IMP_FUNC(___ERR_CODE,___SCHAR_to_SCMOBJ)
   ___P((___SCHAR x,
         ___SCMOBJ *obj,
         int arg_num),
        ());
___IMP_FUNC(___ERR_CODE,___UCHAR_to_SCMOBJ)
   ___P((unsigned char x,
         ___SCMOBJ *obj,
         int arg_num),
        ());
___IMP_FUNC(___ERR_CODE,___LATIN1_to_SCMOBJ)
   ___P((___LATIN1 x,
         ___SCMOBJ *obj,
         int arg_num),
        ());
___IMP_FUNC(___ERR_CODE,___UCS2_to_SCMOBJ)
   ___P((___UCS2 x,
         ___SCMOBJ *obj,
         int arg_num),
        ());
___IMP_FUNC(___ERR_CODE,___UCS4_to_SCMOBJ)
   ___P((___UCS4 x,
         ___SCMOBJ *obj,
         int arg_num),
        ());
___IMP_FUNC(___ERR_CODE,___SHORT_to_SCMOBJ)
   ___P((short x,
         ___SCMOBJ *obj,
         int arg_num),
        ());
___IMP_FUNC(___ERR_CODE,___USHORT_to_SCMOBJ)
   ___P((unsigned short x,
         ___SCMOBJ *obj,
         int arg_num),
        ());
___IMP_FUNC(___ERR_CODE,___INT_to_SCMOBJ)
   ___P((int x,
         ___SCMOBJ *obj,
         int arg_num),
        ());
___IMP_FUNC(___ERR_CODE,___UINT_to_SCMOBJ)
   ___P((unsigned int x,
         ___SCMOBJ *obj,
         int arg_num),
        ());
___IMP_FUNC(___ERR_CODE,___LONG_to_SCMOBJ)
   ___P((long x,
         ___SCMOBJ *obj,
         int arg_num),
        ());
___IMP_FUNC(___ERR_CODE,___ULONG_to_SCMOBJ)
   ___P((unsigned long x,
         ___SCMOBJ *obj,
         int arg_num),
        ());
___IMP_FUNC(___ERR_CODE,___FLOAT_to_SCMOBJ)
   ___P((float x,
         ___SCMOBJ *obj,
         int arg_num),
        ());
___IMP_FUNC(___ERR_CODE,___DOUBLE_to_SCMOBJ)
   ___P((double x,
         ___SCMOBJ *obj,
         int arg_num),
        ());
___IMP_FUNC(___ERR_CODE,___POINTER_to_SCMOBJ)
   ___P((void *x,
         ___SCMOBJ tag,
         ___SCMOBJ (*release_fn) ___P((void *ptr),()),
         ___SCMOBJ *obj,
         int arg_num),
        ());
___IMP_FUNC(___ERR_CODE,___NONNULLPOINTER_to_SCMOBJ)
   ___P((void *x,
         ___SCMOBJ tag,
         ___SCMOBJ (*release_fn) ___P((void *ptr),()),
         ___SCMOBJ *obj,
         int arg_num),
        ());
___IMP_FUNC(___ERR_CODE,___FUNCTION_to_SCMOBJ)
   ___P((void *x,
         ___SCMOBJ *obj,
         int arg_num),
        ());
___IMP_FUNC(___ERR_CODE,___BOOL_to_SCMOBJ)
   ___P((___BOOL x,
         ___SCMOBJ *obj,
         int arg_num),
        ());
___IMP_FUNC(___ERR_CODE,___CHARSTRING_to_SCMOBJ)
   ___P((char *x,
         ___SCMOBJ *obj,
         int arg_num),
        ());
___IMP_FUNC(___ERR_CODE,___LATIN1STRING_to_SCMOBJ)
   ___P((___LATIN1STRING x,
         ___SCMOBJ *obj,
         int arg_num),
        ());
___IMP_FUNC(___ERR_CODE,___UTF8STRING_to_SCMOBJ)
   ___P((___UTF8STRING x,
         ___SCMOBJ *obj,
         int arg_num),
        ());
___IMP_FUNC(___ERR_CODE,___UCS2STRING_to_SCMOBJ)
   ___P((___UCS2STRING x,
         ___SCMOBJ *obj,
         int arg_num),
        ());
___IMP_FUNC(___ERR_CODE,___UCS4STRING_to_SCMOBJ)
   ___P((___UCS4STRING x,
         ___SCMOBJ *obj,
         int arg_num),
        ());
___IMP_FUNC(___ERR_CODE,___VARIANT_to_SCMOBJ)
   ___P((___VARIANT x,
         ___SCMOBJ *obj,
         int arg_num),
        ());
___IMP_FUNC(___SCMOBJ,___release_pointer)
   ___P((___SCMOBJ obj),
        ());
___IMP_FUNC(___ERR_CODE,___make_cdef_stack_marker)
   ___P((___SCMOBJ *marker),
        ());
___IMP_FUNC(void,___kill_cdef_stack_marker)
   ___P((___SCMOBJ marker),
        ());
___IMP_FUNC(void,___release_scmobj)
   ___P((___SCMOBJ obj),
        ());
#endif

#ifndef ___INCLUDED_FROM_MEM
___IMP_FUNC(___SCMOBJ,___alloc_scmobj)
   ___P((int subtype,
         long bytes,
         int kind),
        ());
___IMP_FUNC(___SCMOBJ,___make_pair)
   ___P((___SCMOBJ car,
         ___SCMOBJ cdr,
         int kind),
        ());
___IMP_FUNC(___SCMOBJ,___make_vector)
   ___P((long length,
         ___SCMOBJ init,
         int kind),
        ());
___IMP_FUNC(void,___still_obj_refcount_inc)
   ___P((___SCMOBJ obj),
        ());
___IMP_FUNC(void,___still_obj_refcount_dec)
   ___P((___SCMOBJ obj),
        ());
#endif

#ifndef ___INCLUDED_FROM_SETUP
___GSTATE_DECL
___IMP_FUNC(double,___copysign)
   ___P((double x,
         double y),
        ());
___IMP_FUNC(int,___isfinite)
   ___P((double x),
        ());
___IMP_FUNC(int,___isnan)
   ___P((double x),
        ());
___IMP_FUNC(double,___round)
   ___P((double x),
        ());
#ifndef ___GOOD_ATAN2
___IMP_FUNC(double,___atan2)
   ___P((double y,
         double x),
        ());
#endif
___IMP_FUNC(void,___setup_params_reset)
   ___P((struct ___setup_params_struct *setup_params),
        ());
___IMP_FUNC(void,___setup)
   ___P((struct ___setup_params_struct *setup_params),
        ());
___IMP_FUNC(void,___cleanup) ___PVOID;
___IMP_FUNC(void,___cleanup_and_exit_process)
   ___P((int status),
        ());
___IMP_FUNC(___ERR_CODE,___call)
   ___P((int nargs,
         ___SCMOBJ proc,
         ___SCMOBJ marker),
        ());
___IMP_FUNC(void,___propagate_error)
   ___P((___ERR_CODE err),
        ());
___IMP_FUNC(void,___raise_interrupt)
   ___P((int code),
        ());
___IMP_FUNC(void,___begin_interrupt_service) ___PVOID;
___IMP_FUNC(___BOOL,___check_interrupt)
    ___P((int code),
         ());
___IMP_FUNC(void,___end_interrupt_service)
   ___P((int code),
        ());
___IMP_FUNC(void,___disable_interrupts) ___PVOID;
___IMP_FUNC(void,___enable_interrupts) ___PVOID;
#endif

#endif
